.H 1 "INTEGRATING THE DRIVER INTO THE 4.3 BSD SYSTEM"
Quite a lot of work had to be done in order to get the driver into the
system. The work has little to do with driving hardware, but has to be done
or otherwise debugging the driver and porting to other systems would be very
difficult. Parts of the work is essential for proper operation.
.H 2 "Conf.c \- Major device numbers"
There are two tables in the system that describe all device interfacing. These
tables are located in the file 
.I /sys/machine/conf.c.
The
.I cdevsw
table describe all character (raw) devices and the
.I bdevsw
table describe all block devices.
.P
The 
.I major
number of a device specifies an index into one of these tables and selects
a
.I class
of devices.
Raw and block devices do
.I not
share the same number space, and thus raw device 10 isn't necessarily the
same device class as block device 10.
.P
The tables describe a list of 
.I possible
devices, but does not list what is actually present on a system.
.P
There are no entries for the SCSI interface cards, only for the devices that
the cards control. In the IBM AOS4.3 system, major device numbers 0-24  for character devices, and 0-9 for block devices are 
reserved for IBM. 
.P
In this implementation, the tape major numbers are 25 for raw I/O and 10 for block I/O.
The disk major numbers are 26 for raw I/O and 11 for block I/O.
.H 3 "Layout of the cdevsw table"
The cdevsw table has the following layout. Each entry is the only link between the main
UNIX code and a driver.
.Sd
struct cdevsw
{
        int     (*d_open)();    \fR/* Called at every user open, used in order ti initialise the device */\fP
        int     (*d_close)();	\fR/* Routine to shut down a device, called only on last user close */\fP
        int     (*d_read)();	\fR/* Device specific routine to collect data from a device */\fP
        int     (*d_write)();	\fR/* device specific routine to send data to a device */\fP
        int     (*d_ioctl)();	\fR/* Routine to manipulate device specific functions*/\fP
        int     (*d_stop)();	\fR/* If the driver is a terminal interface, this is the device specific routine to suspend output */\fP
        int     (*d_reset)();	\fR/* This routine will be called to reset a device if the IOCC adapter is also reset */\fP
        struct tty *d_ttys;	\fR/* If driver is a terminal ifc, this is the address of an array of tty structures, one per line. */\fP
        int     (*d_select)();	\fR/* Device specific support to augment the select system call. */\fP
        int     (*d_mmap)();	\fR/* Never used on the RT. Maybe used on other UNIX systems, probably to do some kind of memory mapping. */\fP
};
.Se
Any routine entry may be
.I nulldev
or
.I nodev.
Nulldev means that there is no action for this function, nodev means that
an error action should be performed for this function.
.br
.H 3 "Layout of the bdevsw table"
The cdevsw table has the following layout. As above, each entry is the 
only link between the main code and a device driver.
.Sd
struct bdevsw
{
        int     (*d_open)();
        int     (*d_close)();
        int     (*d_strategy)();
        int     (*d_dump)();
        int     (*d_psize)();
        int     d_flags;
        int     (*d_minphys)();
};
.Se
The various entries have the following meaning:
.H 4 "Open routine"
This routine is called at every user
.I open.
It is used in order to initialise the device.
.H 4 "Close routine"
Routine to shut down a device. Called only on last user
.I close.
.H 4 "Strategy routine"
Performs all block I/O for a device. 
Responsible for finding the best way to order I/O requests.
.H 4 "Dump routine"
Called by the system to perform a post mortem crash dump of the system.
.H 4 "Psize routine"
This routine should return the number os blocks in a device partition, or -1
if the partition isn't allowed. The pointer should be 0
.B not
.I nodev,
if the routine is undefined.
.H 4 "Flags"
Buf style flags passed into block I/O requests, typically B_TAPE, B_CALL etc.
.H 4 "Minphys routine"
This is a very unfortunately named routine since it
calculates the
.I maximum
number of physical bytes to be transferred during a single DMA operation
for a device. The kernel supplies a routine called
.I minphys
that returns the maximum transfer length that the 
.I kernel
can handle.
.H 2 "Ioconf.c \- configuration file"
This file is automatically generated by the config(8) program. It contains 
initialised structures according to the configuration script for the system
generated. The following structures are initialised:
.H 3 "Iocccinit"
This structure contain information about all 
.I configured
controllers (adapters) on the I/O channel. It has the type
.I "struct iocc_ctlr."
See below for detailed information about this structure.
.H 3 "Ioccdinit"
This structure contain information about all
.I configured
devices that connect to any of the controllers above or that connect
directly to the IOCC. The type is
.I "struct iocc_device."
See below for detailed information about this structure.
.H 3 "Xhdinit"
This structure contain information about all
.I configured
SCSI targets 
.I configured
to be connected to any of the \fIx\fPha controllers found in iocccinit.
See below for detailed information about this structure.
.H 3 "Xhsinit"
This structure contain information about all
.I configured
SCSI logical units (LUN's)
.I configured
to be connected to any of the SCSI targets mentioned above.
See below for detailed information about this structure.
.H 2 "Kernel structures for autoconfiguration and driver use"
The kernel contain many data structures usable by the device drivers. 
These structures are also used when autoconfiguring the system. The form
a tree which resembles the hardware configuration of the particular machine.
.H 3 "Iocc_hd"
This structure hold pointers to dma registers, the state of the channel etc.
There is one structure per dma channel, the layout is:
.Sd
struct  iocc_hd {
        caddr_t hd_addr;                	/* dummy to keep config happy */
        caddr_t dh_8237_reg;            	/* hardware register location */
        unsigned dh_pagebitmask[IOCC_PAGE_SIZE]; /* bitmap to allocate page modeTCW's */
        unsigned dh_region_map[IOCC_REGION_SIZE]; /* region mode map */
        short   dh_channel;             	/* channel number */
        short   dh_subchan;             	/* subchannel on the 8237 chip */
        short   dh_state;               	/* state of this channel (see dmavar.h)*/
        struct  dma_callback *dh_callfwd; 	/* head of callback queue */
        struct  dma_callback *dh_calllst;	/* tail of callback queue */
        struct  iocc_ctlr *dh_actf;     	/* head of queue to transfer */
        struct  iocc_ctlr *dh_actl;     	/* tail of queue to transfer */
};
.Se
The field hd_addr is initialised in ioconf.c which really is a dummy,
but it makes it possible to keep the config(8) program structure almost as it
was on the VAX!
.H 3 "Iocc_ctlr"
This is the "per-controller" structure. There is one for each controller
(adapter) which control more than one device. If an adapter has devices attached
there are iocc_device structures that reference the corresponding iocc_ctlr structure. This structure isqueued in iocc resource wait and saves information about
iocc resources which are used. The queue of devices waiting to transfer is also
attached here. The layout is:
.Sd
struct iocc_ctlr {
        struct  iocc_driver *ic_driver;
        short   ic_ctlr;        	/* controller index in driver */
        short   ic_alive;       	/* controller exists */
        caddr_t ic_addr;        	/* address of device in i/o space */
        int     ic_irq;         	/* the IRQ level of the device */
        struct  iocc_hd *ic_hd; 	/* link back to the header */
        struct  buf ic_tab;     	/* queue of devices for this controller */
        short   ic_dmachannel;          /* dma channel number */
        int     ic_dmaflags;            /* dma transfer flags */
        struct  buf *ic_dmabuf;         /* buffer to transfer */
        struct iocc_ctlr *ic_dmaforw;	/* Next ctlr on the dma channel */
};
.Se
The fields ic_driver to ic_irq are initialised in ioconf.c. The rest
is set up at boot time or when the system is running.
.H 3 "Iocc_device"
This is the "per-device" structure. A controller has devices or uses DMA, 
everything else is a "device". If a controller has many drives attached, 
then there will be several iocc_device structures associated with a 
single iocc_ctlr structure.
.P
This structure contains all the information necessary to run
a non-dma device such as an async line card.  It also contains information
for slaves of iocc controllers as to which device on the slave
this is.  A flags field here can also be given in the system specification
and is used to tell which asynchronous lines are hard wired or other device
specific parameters. The layout is as follows:
.Sd
struct iocc_device {
        struct  iocc_driver *iod_driver;/* corresponding driver struct */
        short   iod_unit;       	/* unit number on the system */
        short   iod_ctlr;       	/* mass ctlr number; -1 if none */
        short   iod_slave;      	/* slave on controller */
        caddr_t iod_addr;       	/* address of device in i/o space */
        short   iod_irq;        	/* interrupt request level */
        short   iod_dk;         	/* if init 1 set to number for iostat */
        int     iod_flags;      	/* parameter from system specification */
        short   iod_alive;      	/* device exists */
        short   iod_type;       	/* driver specific type information */
        caddr_t iod_physaddr;   	/* phys addr, for standalone (dump) code */
        struct  iocc_device *iod_forw;	/* Next device on this ctlr */
        struct  iocc_ctlr *iod_mi;	/* Connected ctlr, if any */
        struct  iocc_hd *iod_hd;	/* Ptr to dma channel header struct */
};
.Se
The fields iod_driver to iod_flags are initialised in the ioconf.c file. 
The rest is set up at boot time.
.H 3 "Iocc_driver"
This structure, one per device driver, define entries for a set of routines 
in the device driver code that are used at boot time by the configuration
routine.
The following layout is used:
.Sd
struct iocc_driver {
        int     (*idr_probe)();         /* see if a device is really there */
        int     (*idr_slave)();         /* see if a slave is there */
        int     (*idr_attach)();        /* setup device for a slave */
        int     (*idr_dgo)();           /* controller specific dma setup */
        caddr_t *idr_addr;              /* device csr addresses */
        char    *idr_dname;             /* name of a device */
        struct  iocc_device **idr_dinfo;/* backpointers to iodinit structs */
        char    *idr_mname;             /* name of a controller */
        struct  iocc_ctlr **idr_minfo;  /* backpointers to iominit structs */
        int     (*idr_intr)();          /* interrupt service routine */
#define INT_SERVICED    0
#define INT_NOT_MINE    1
        int     idr_csr;                /* offset to read/write csr */
        int     (*idr_chanrelse)();     /* dma channel release request */
        char    idr_flags;              /* all zero in old drivers! */
#define EARLY_INT       0x01    /* interrupt handler called on initial
                                 * (probe) interrupt and int_8259 does the
                                 * shared interrupt reset */
#define DRIVER_SUSPEND  0x02    /* driver wants call on suspend */
        int     (*idr_suspend)();       /* device suspend */
#define SUSPEND_START   0               /* device suspend start */
#define SUSPEND_DONE    1               /* device suspend done */
};
.Se
The contents of this structure is initialised in the driver code.
.H 3 "Xha_hd"
Now we comes to the structures that have been added for the SCSI driver.
The first structure is much alike the iocc_hd structure. There is one such
for every SCSI host adapter on the system. It contain information about
the host adapter, also there are pointers to xha_device structures which
describe configured SCSI targets. The layout is as follows:
.Sd
/* Per SCSI controller structure */
struct xha_hd
{
        struct  iocc_driver *xh_driver;     /* driver structure */
        short                xh_hatype;     /* Type of ctrlr */
#define HATYPE_NONE     0                   /* Initial value, structure unused */
#define HATYPE_AHA      1                   /* Adaptec AHA-1540 */
#define HATYPE_IBM      2                   /* IBM SCSI Adapter */
        short                xh_active;     /* set if xha is active */
        short                xh_ndrive;     /* number of devices */
        caddr_t              xh_xha;        /* virt addr of xha */
        char                *xh_haname;     /* Name of host adapter */
        struct  xha_device  *xh_xhip[8];    /* attached ctrlrs */
        struct  iocc_ctlr   *xh_iocc;       /* Backptr */
        struct  xha_device  *xh_actf;       /* xfer q head */
        struct  xha_device  *xh_actl;       /* xfer q tail */
};
.Se
The fields in xha_hd are all initialised at boot time by the autoconfig
routine.
.H 3 "Xha_device"
The xha_device structure is a "one-per-target" structure. They contain
information about a SCSI target configured to a be connected to a 
SCSI host adapter described in an xha_hd structure. Since each target
may control many logical units, there is no LUN info in this structure.
Info about LUN's are in the xha_slave structure. The xha_device
structure looks like this:
.Sd
/* Per target structure */

struct xha_device
{
        struct  xha_driver  *xi_driver;
        short                xi_unit;   /* System's unit no */
        short                xi_ctlr;   /* SCSI ctrl it is on */
        short                xi_drive;  /* Ctrlr on scsi bus */
        short                xi_dk;     /* Driver no for iostat */
        short                xi_hatype; /* Type of ctlr it is on */
        short                xi_alive;  /* Target exists */
        u_long               xi_type;   /* Unit type */
        struct buf           xi_tab;    /* Head of Q for this dev */
        struct xha_device   *xi_forw;
        struct xha_slave    *xi_slaves[8]; /* Pointers to slave structs */
        struct xhac_driver  *xi_hadriver;  /* Pointers to ha's driver routines *
/
        /* Short cuts */
        caddr_t              xi_xha;
        struct  xha_hd      *xi_hd;

};
.Se
Of these fields, the xi_driver to xi_hatype are initialised in ioconf.c. The 
rest are set up at boot time or simply used during system operation.
.H 3 "Xha_slave"
This structure describe logical SCSI units, see above.
Layout:
.Sd
struct  xha_slave
{
        struct  xha_driver  *xs_driver;
        short                xs_ctlr;   /* Which formatter */
        short                xs_unit;   /* Which unit to system */
        short                xs_slave;  /* Which slave to formatter (LUN)*/
        short                xs_alive;  /* Device alive */
        /* Short cut back */
        struct xha_device   *xs_device; /* Pointer to formatter device struct */
}
.Se
Of these fields, the xs_driver to xs_slave are initialised in ioconf.c. The
rest are set up at boot time.
.H 3 "Xha_driver"
For each SCSI device type supported, i.e. for each SCSI device driver, there
is an xha_driver structure.
The structure, define entries for a set of routines 
in the device driver code that are used at boot time by the configuration
routine or by the host adapter driver. It looks as follows:
.Sd
/* Per device type structure */

struct xha_driver
{
    int                 (*xd_probe)();          /* See if target exists */
    int                 (*xd_slave)();          /* See if a slave is there */
    int                 (*xd_attach)();         /* Attach a device */
    int                 (*xd_ustart)();         /* Start unit */
    int                 (*xd_start)();          /* Setup data xfer */
    int                 (*xd_int)();            /* Data xfer complete */
    short                *xd_type;              /* Drive type codes */
    char                 *xd_dname;             /* Device name */
    char                 *xd_sname;             /* Slave name */
    struct  xha_device  **xd_dinfo;             /* Backpointers */
    struct  xha_slave   **xd_sinfo;             /* Backpointers */
    int                 (*xd_ttl)();            /* Find out time to live */
    int                   xd_flags;             /* Driver flags */
};
.Se
These fields are all initialised in the corresponding driver's code.
.H 3 "Xhac_driver"
For each type of SCSI host adapter and thus for each host adapter
driver there will be an xhac_driver structure. It is used by the SCSI device
drivers to access certain defined functions in the host adapter driver.
The structure looks like this:
.Sd
/* Per host adapter type structure */
struct xhac_driver
{
        int     (*xhc_tur)();           /* Test if unit ready */
        int     (*xhc_doscsi)();        /* Send SCSI command */
        int     (*xhc_ustart)();        /* Unit start routine */
};
.Se
These fields are all initialised in the corresponding host adapter
driver's code.
.H 2 "The config(8) program"
The config(8) program builds a set of system-configuration files from a
short file that describes the system being configured. Config also takes
as input a file that tells it which files are needed to generate a
system.  This can be augmented by a configuration-specific set of files
that gives alternate files for a specific machine.
The files generated are the makefile, ioconf.c and swapvmunix.c who are
generated into the directory where the system is built. Also, for each device,
an \fIxx\fP.h is built which only contains a single 
.I #define
which tells the compiler how many devices of the kind
.I xx
that has been configured.
.H 3 "Syntax of the configuration file"
The configuration file contains three logical pieces of information:
.BL
.LI 
parameters global to all system images
.LI
parameters specific to each system image to be generated
.LI
device specifications
.LE 1
I won't discuss the first two items on the list above since they
doesn't have anything to do with the device drivers.
.P
For the RT, a device specification takes one of the following forms:
.BL
.LI
	\fBcontroller	\fIdevice-name device-info\fR
.LI
	\fBdevice	\fIdevice-name device-info\fR
.LI
	\fBdisk	\fIdevice-name device-info\fR
.LI
	\fBtape	\fIdevice-name device-info\fR
.LE 1
For the SCSI devices the following item should be added to this list:
.BL
.LI
	\fBmaster	\fIdevice-name device-info\fR
.LE 1
A controller is an adapter that controls one or more disks or tapes
or it uses DMA. Everything else is a device.
The
.I device-name
is one of the standard device names concatenated with the 
.I logical
unit number assigned to the device. (This sould not be confused with
the logical unit numbers, LUN's used in the SCSI context.) Standard device names
are documented in the User's Reference Manual, chapter 4.
For our SCSI system the follownig names have been used:
.BL
.LI
	controller	\fBaha\fIn\fR	\- Adaptec SCSI host adapter.
.LI
	master	\fBxtc\fIn\fR	\- SCSI disk target controller.
.LI
	master	\fBcsc\fIn\fR	\- SCSI tape target controller.
.LI
	disk	\fBxt\fIn\fR	\- SCSI disk.
.LI
	tape	\fBcs\fIn\fR	\- SCSI streamer tape.
.LE 1
.P
The 
.I device-info
describes how the hardware is connected in the interconnection hierarchy. The
beginning of the hierarchy is defined as follows:
.TS 
tab(;);
lw(0.8i) lfB l l.
 ;controller;iocc0;\fBat nexus\fP ?
.TE
.ta 0.8i +0.8i +0.8i +0.8i
The remaining legal interconnections are:
.BL
.LI
A controller can be connected to another controller.
.LI
A master is connected to a controller.
.LI
A disk or tape is always attached to a controller or a master.
.LI
A device is always connected to a controller.
.LE 1
On the RT, the controller specification takes the form:
.TS 
tab(;);
lw(0.8i) lfB l l.
 ;controller;xxcn;\fBat\fP iocc0 \fBcsr\fP addr \fBpriority\fP irq
.TE
.ta 0.8i +0.8i +0.8i +0.8i
where
.DS
.nf
	"xx"	is the two or three letter name of the device.
	"n"	is the controller number.
	"addr"	is the controller adapter base address.
	"irq"	is the PC/AT IRQ level for the adapter.
.DE
For the 
.B aha
adapter, the 
.I addr
value may be specified as 0. This will cause the autoconfig code to try
the standard addresses defined for the Adaptec adapter.
.P
The master specification take the form:
.TS 
tab(;);
lw(0.8i) lfB l l.
 ;master;xxcn;\fBat\fP \fIx\fPham \fBdrive\fP target
.TE
.ta 0.8i +0.8i +0.8i +0.8i
where
.DS
.nf
	"xx"	is the two letter name of the device type.
	"n"	is the target controller number.
	"m"	is the host adapter number.
	"target"	is the SCSI target number for the device.
.DE
.P
The disk and tape specification takes one of the following forms:
.TS 
tab(;);
lw(0.8i) lfB l l.
 ;disk;xxn;\fBat\fP xxcz \fBdrive\fP z
 ;disk;xxn;\fBat\fP xxcz \fBslave\fP lun
 ;tape;xxn;\fBat\fP xxcz \fBdrive\fP z
 ;tape;xxn;\fBat\fP xxcz \fBslave\fP lun
.TE
.ta 0.8i +0.8i +0.8i +0.8i
where
.DS
.nf
	"xx"	is the name of the device
	"n"	is the unit number
	"y"	is the controller number
	"z"	is the slave unit number
	"lun"	is the SCSI logical unit number.
.DE
.H 3 "Changes to config(8)"
The config(8) program had to be changed in order to make it generate systems
with the SCSI driver. The main reason for this is that new kernel structures
are introduced that have to be initialised in 
.I ioconf.c.
The changes are introduced in the source files 
.I config.h,
.I config.y,
.I mkioconf.c
and
.I mkswapconf.c.
.H 4 "Change in config.h"
This header file is only changed in order to introduce two new global
variables
.I seen_aha
and
.I seen_iha
which are booleans set to TRUE by the parser when it sees the keywords
.I aha
or
.I iha
in the config file.
The value of these variables are used later in the program in order to
determine if the new structures should be generated or not.
.H 4 "Change in config.y"
The change in config.y introduces the recognition of the keywords 
.I aha
and
.I iha
into the parser. The variables mentioned above are set to TRUE if those
keywords are recognised.
.H 4 "Changes in mkioconf.c"
The changes here introduces code to generate the appropriate
.I #include
statements and to declare the 
.I xhdinit
and 
.I xhsinit
variables initialised with the appropriate values. There is also a change
that allows the 
.I aha
adapter to be configured with the csr value of 0. This value causes the
autoconfiguration code to ask the driver's probe routine to use addresses 
predefined in the driver thus enabling the config script to wildcard the csr
address. The code in autoconfig works for other devices too, but I restricted
the use of csr 0 to the 
.I aha
adapter since only that driver code is under my control.
.H 4 "Changes to mkswapconf.c"
Traditionally, the UNIX operating system has only allowed a disk to use
a maximum of 8 partitions per disk. As an interesting exercise I wanted to
extend the maximum number of partitions to be a device type specific figure, 
different on different device types. This exercise involved changes in 
mkswapconf.c since minor device numbers are generated here. If I want to
do swapping on 
.I /dev/xt1b, 
the minor device number should be 17 for this 
partition if we allow 16 partitions on an xt disk. If (as normally) only 
8 partitions are allowed, the figure would have been 9.
The maximum number of partitions allowed for a disk type is specified in the
.I devices.ca
file in /sys/conf as an extension to the syntax of this file. A third parameter
now may be given, specifying the maximum number of partitions for this disk
type.
.H 2 "Adding source files to the kernel"
When new source files are added to the kernel, these sources must be specified
in the
.I /sys/conf/files
or the
.I /sys/conf/files.ca
files.
For the SCSI system the following lines had to be added in /sys/conf/files.ca
.Sd
machineio/iha.c         optional iha device-driver
machineio/aha.c         optional aha device-driver
machineio/cs.c          optional cs device-driver
machineio/xt.c          optional xt device-driver
.Se
Which describe the various driver source code files to the config(8) program.
.H 2 "Defining new swap devices"
If a new device type is going to be used as a swap device, it must be defined
in the
.I devices.ca
file. The file consist of a number of lines with the format
.BL
.LI
\fIxx	majdev#	\(u[maxpart\(u]\fP
.LE 1
Where
.DS
.nf
	"xx"	is the two letter code for the disk type.
	"majdev#"	is the major-device-number for the disk type.
	"maxpart"	is the optional max-partitions parameter.
			This an extension to the original syntax. If this 
			parameter is not specified, it's default value is 8.
.DE
