.H 1 "DETAILED DESCRIPTION OF THE DRIVER CODE"
The following chapter discuss in detail what happens whithin the SCSI driver.
.H 2 "Definitions"
.BL
.LI
.B Device
the hardware interfacing the real world e.g. tape drives, serial 
line interfaces, disks, etc.
.LI
.B Controller
a piece of hardware that communicate with devices which cannot talk to the
computer directly. A controller commonly have more than one device of the same
or different kinds connected to itself.
.LI
.B Adapter
is the IBM name for a controller (more or less).
.LI 
.B "Host adapter"
in this context means a controller that connects between the internal bus
of a computer and a SCSI I/O bus.
.LI 
.B "DMA \- Direct Memory Access"
DMA is a method of transferring large blocks of sequential data between a
device and the processor memory with minimal interference 
to the system processor. DMA permits simultaneous use of I/O adapters and 
the system processor. 
.LI 
.B "Anternate controller"
The IBM RT supports two types of DMA adapters. One is an adapter
.I without
the ability of handling addresses or I/O control signals. It uses the
RT system's DMA controller for addressing and for I/O channel control.
.P
The other type is called anternate controller, or sometimes controller
using "3:rd party DMA", is an adapter 
.I with
the ability of handling its own addressing and I/O channel control. This means
that the adapter may have a DMA chip onboard. The system DMA channel does not
generate addresses, these are instead generated by the I/O adapter itself.
All alternate controllers must be 16-bit devices.
.LI
.B "TCW \- Translation Control Word"
The addresses generated by either the system DMA or the alternate controller are
translated using one or more TCW's. The TCW's must be set up before starting 
the DMA transfer. Each TCW maps a 2 Kbyte memory area on the I/O channel to a
2 kByte page in the system memory real map. The I/O addresses FE0000 to FFFFFF
in the memory map are used as a window into system memory.
.LE
.H 2 "Kernel routines commonly used in a device driver"
In the common part of the kernel, there are a lot of routines that device
drivers use. Many of them are self-evident to the average C programmer and 
some are quite special to the kernel. In order to be as complete as possible,
I will deal with all routines that have been used in the SCSI driver.
.H 3 "I/O routines"
These are routines directly involved with the process of doing I/O.
.H 4 "Physio"
This routine is used by device drivers to do character I/O to block devices.
It is called from the 
\fI\fPread 
and
\fI\fPwrite
routines in the device driver. It's synopsis is:
.Sy
physio(int (*strat)(), struct buf *bp, dev_t dev,int rw,unsigned (*mincnt)(),struct uio *uio)
.Se
Where:
.BL
.LI
.I strat
is a pointer to the 
.I strategy
routine of the driver, see below.
.LI
.I buf
is a pointer to a driver supplied 
.I buf
structure.
.LI
.I dev
is the major/minor numbers of the device in question.
.LI
.I rw
is a buf-style flag that denote the type of operation, either B_WRITE or B_READ.
.LI
.I mincnt
is a pointer to a kernel or driver supplied 
.I minphys
routine. See [4.1.2.7] and below.
.LI
.I uio
is the reference to the user I/O buffer structure that contain information
about the data to be transferred.
.LE 1
The routine loops through the I/O vector found in 
.I uio
doing the requested I/O operation on all the data spaces found in the I/O vector.
For each transfer, 
.I physio
will:
.AL A
.LI
Verify the user's access priviledges.
.LI
Then; while there is more data to be transferred
.AL 1
.LI
Call the minphys routine indicated by
.I mincnt.
The size of the transfer can be decreased by that operation, therefore
it has to loop as long as there is any data left.
.LI
Lock all affected virtual memory pages in real memory.
.LI
Call the driver supplied strategy routine which perform the actual transfer.
.LI
Unlock the virtual memory pages.
.LE
.LE
.H 4 "Minphys"
This kernel routine effectively limits all transfers to a maximum of 
.I MAXPHYS
bytes.
On the RT, MAXPHYS is defined as 128 kBytes. Other systems commonly use
63 kBytes.
The synopsis is:
.Sy
minphys(struct buf *bp)
.Se
Where
.I bp
points to the buffer structure associated with the transfer to be done.
If the driver supplies a specialized \fIxx\fPminphys routine, that routine must call
the kernel routine
.I minphys.
.H 4 "Disksort"
Is a routine used by disk drivers. It is used in order so do seek sorting of 
the transfer queue.
It is declared as:
.Sy
disksort(struct buf *dp, struct buf *bp)
.Se
The argument dp structure holds an activity chain pointer
on which two queues are kept, sorted in ascending cylinder order.
The first queue holds those requests which are positioned after
the current cylinder. The second queue holds
requests which came in after their cylinder number was passed.
Thus a one way scan is implemented, retracting after reaching the
end of the drive to the first request on the second queue,
The argument bp is a pointer to the buf structure associated with the 
transfer to be done. The buf structure variable b_resid should contain
the number of the cylinder that the request will cause a seek to.
.H 4 "Bsdtofsb"
Is a macro that maps a "block device block" number to the
file system block number. This translation really should be device dependent
but that hasn't been implemented yet.
The macro is simple and looks like this:
.Sy
#define bdbtofsb(bn)    ((bn) / (BLKDEV_IOSIZE/DEV_BSIZE))
.Se
The argument 
.I bn 
is the UNIX internal block number. 
On the RT, BLKDEV_IOSIZE is defined as 2048 bytes, DEV_BSIZE is a fixed constant
with the value 512 which is equal to the common sector size of the disks.
.H 3 "Process control"
Sometimes in device drivers it is necessary to change the status of the process
that is executing. The routines below are used in these cases.
.H 4 "Sleep"
Sleep is used in order to give up the processor until something happens
that makes it possible to continue. 
It is the primary means of synchronization of processes within
the kernel.
Sleep is defined thus:
.Sy
sleep(caddr_t wchan, int pri)
.Se
The wait channel
.I wchan
denotes an event for which the process is to wait. By convention, this event
is represented by the address to a relevant data structure.
When the event occurs, the process enters the scheduling queue at priority 
.I pri.
The most important effect of pri is that when
pri is less than or equal to PZERO a signal cannot disturb the sleep;
if pri is greater than PZERO signals will be processed.
Callers of this routine must be prepared for
premature return, and check that the reason for
sleeping has gone away.
A typical sleep sequence therefore looks like this:
.Sd
s = splbio()		/* Disable interrupts */
while(\f(COdevice_in_use\fP)
{
	sleep(&\f(COdevice_data_structure\fP, priority);
}
\f(COdevice_in_use\fP = 1;
splx(s);		/* Enable interrupt */
.Se
It is
.I illegal
to call sleep from within an interrupt handler, 
such a call will crash the system immediately. Also, wchan equal to 0 will
also crash the system. Every device driver writer should be aware of these
facts.
Another warning here is that a wakeup that occurs before the sleep won't be
remembered, this is why the interrupts are shut off in the example above.
If an interrupt was allowed to occur berween the 
.I while
and the
.I sleep
statements, the process would probably sleep forever.
.H 4 "Wakeup"
A 
.I wakeup
on a wait channel awakens 
.I all
processes waiting on that channel.
The corresponding wakeup sequence to the example above is executed 
by some other process or an interrupt handler and could look like:
.Sd
\f(COdevice_in_use\fP = 0;
wakeup(&\f(COdevice_data_structure\fP);
.Se
.H 4 "Iowait"
Is a macro that only renames biowait to iowait (or the other way round).
Biowait waits for I/O completion on a buffer and return errors to the user.
It's definition is:
.Sy
biowait(struct buf *bp)
.Se
.I Bp 
is a pointer to the buffer structure for which transfer we want to wait for.
.H 4 "Iodone"
Is also a macro that renames biodone to iodone (wonder why they have done it
this way?)
Biodone's definition looks like:
.Sy
biodone(struct buf *bp)
.Se
The routine will mark I/O complete on the buffer pointed to by
.I bp.
If some routine should be called (this can be specified in the buf structure)
that routine will be called.
Otherwise, biodone will wake up anyone waiting for the buffer.
.H 3 "DMA routines"
The IBM AOS contain some routines to do DMA on the RT. These SHOULD be used.
Doing DMA by oneself isn't healthy.
.H 4 "Dma_setup"
.Sy
int
dma_setup(struct iocc_ctlr *ic)
.Se
Dma_setup is called whenever a device first wants to initiate a transfer. 
.I Ic
is a pointer to a 
.I iocc_ctlr
structure which has:
.BL
.LI
The channel number 
.I ic_dmachannel.
.LI
Flags
.I ic_dmaflags
.LI
A buffer describing the DMA transfer
.I ic_dmabuf.
This (pointer) can be set to 0 to get the channel only.
.LE 1
.H 4 "Dma_map"
.Sy
unsigned 
dma_map(short chan, struct buf *bp)
.Se
Dma_map is used to map TCW's for DMA transfers. Dma_setup automatically call
this routine for the buffer pointer passed in ic_dmabuf, if any.
.I Chan
is the channel for the transfer.
.I Bp is a pointer to the buf structure that describe the transfer.
Dma_map returns the I/O address for the transfer, i.e. the AT-bus address
that shoud be used. If there aren't enough contiguous free TCW's to support
the transfer, dma_map will return DMA_INV_IOADDR.
.H 4 "Dma_free_map"
.Sy
void
dma_free_map(short chan, unsigned ioaddr, int len)
.Se
Dma_free_map frees TCW's allocated by dma_map. 
.I Chan
is the cannel as aboce.
.I Ioaddr
is the I/O address returned by dma_map.
.I Len is the length of the transfer in bytes.
.H 4 "Dma_go"
.Sy
void
dma_go(short chan)
.Se
Dma_go enables the DMA channel for a transfer.
.H 4 "Dma_done"
.Sy
void
dma_done(short chan)
.Se
Dma_done frees the channel for other devices to use. All TCW's are freed.
.H 4 "Dma_get_minphys"
.Sy
void
(*dma_get_minphys(struct iocc_ctlr *ic))()
.Se
Dma_get_minphys returns a pointer to the proper
.I minphys
routine to be called. 
.H 3 "Hardware interfacing routines"
These routines deal with the processor itself such as altering the interrupt 
priority or doing address translations.
.H 4 "Spl\fRn\fP, splx"
These routines are used for changing the processor's interrupt priority.
The RT privides eight priority levels, ranging from 0 to 7. In the UNIX
kernel, priorities also ranges from 0 to 7. However, on the RT, 
level 0 is the highest but in the UNIX system, level 7 is. For this reason,
2 sets of routines are provided for changing the system priority level.
.SP
spl0, spl1, through spl7 set the system priority level to UNIX level 0 through 7
respectively.
.SP
_spl0, _spl1, through _spl7 set the system priority level to RT level 0 
through 7 respectively.

These routine return an integer that is the priority level before they
were called. The return value can be used in the 
.I splx
routine to restore the previous level.
A typical example is:
.Sd
int	s;
.
.
s = splN();	/* change to UNIX priority level N */
.
. 		/* critical region of code */
.
splx(s);	/* restore to previous level */
.Se
.H 4 "Real_buf_addr"
.Sy
caddr_t
real_buf_addr(struct buf *bp,char *bufaddr)
.Se
Real_buf_addr translate a virtual address into a real memory address.
.I Bp
is a pointer to the buffer structure associated with the dataa.
.I Bbufaddr
is the real or virtual address for which the real address is sought.
.H 4 "GET_VR0 macro"
.H 4 "SET_VR macro"
.H 3 "Routines useful for debugging"
Some debugging of device drivers are usually neccessary. While in kernel, some
routines and macros are particularly useful.
.H 4 "Printf"
.H 4 "Uprintf"
.H 4 "DEBUGF macro"
.H 3 "Bulk data manipulation routines"
Doing I/O means moving data. Usage of the CPU for doing this isn't very 
effective, but sometimes necessary. The routines below are designed to 
manipulate lots of data, but also useful for small pieces.
.H 4 "Bzero"
.H 4 "Bcopy"
.H 4 "Clrbuf"
.H 4 "Strncpy, strcpy"
.H 4 "Strncmp"
.H 3 "Other routines"
Some other kernel routines have also been used in the SCSI driver. I didn't
manage to put any nice label on these:
.H 4 "Geterror"
.H 4 "DELAY macro"
.H 4 "PROBE_DELAY macro"
.H 4 "Timeout"
.H 2 "Routines commonly found in a device driver"
Now that we have looked at the routines in the kernel that aren't part of any
device driver, we'll have a look into what we could find in a typical
(disk) device driver.
.H 3 "Configuration routines"
These routines are used during the autoconfiguration process. The autoconfig
program will reach these routines through the 
.I iocc_driver
or the
.I xha_driver
structures.
.H 4 "Probe routine"
This routine is used in order to verify the existence of a device. It also
determine the interrupt vector for the device. The probe routine should verify
that the type of the device is what is expected for the driver in case that
there may be different device types at the same logical location.
The possible return values are:
.BL
.LI
PROBE_BAD \- The probe failed.
.LI
PROBE_NOINT \- Probe was OK without interrupt.
.LI
PROBE_OK \- Probe was OK with interrupt.
.LI
PROBE_BAD_INT \- We lost or didn't get interrupt even 
if the device seemed to be there.
.LE 1
The SCSI device probe routines only return PROBE_NOINT or PROBE_BAD since
they doesn't try to cause an interrupt, however, the probe routine for the
SCSI host adapter uses PROBE_OK since that card uses interrupts to the CPU.
.H 4 "Slave routine"
This routine should verify the existence of specific slave devices on a
controller. If the slave device exists, 1 should be returned, otherwise 0.
.H 4 "Attach routine"
This routine is called by the configuration routines when a slave is located.
The attach routine is used to initialize the driver variables for the
slave. The initialization process may require inquiring of the devices certain
specifics such as disk geometry, tape sizes etc. This routine shouldn't return
anything.
.H 3 "Raw I/O entry points"
These are the routines called from the upper parts of the kernel when doing
character (raw) I/O.
.H 4 "Open routine"
The open routine is used to mark the device as open to the rest of the 
system. If the open is successful, a file descriptor is returned (not
by this routine) to the caller and all subsequent read, write, ioctl and
close operations using that descriptor are accepted as valid.
.P
Before an open can complete this routine must verify that the device was 
properly autoconfigured. It should also verify that the minor number 
of the device are 
within acceptable limits. On certain devices it should check that the open 
.I mode
is acceptable. Tapes should be marked for exclusive use. Device registers
and driver variables may be initialized. 
.P
If the open meets all applicable criteria, zero should be returned, otherwise
an appropriate error code shold be returned, typically 
.I "ENXIO, EIO"
or 
.I EBUSY.
.P
The open routine shold be declared as:
.Sy
\fIxx\fPopen(dev_t dev,int mode)
.Se
.BL
.LI
.I dev
is the major/minor device number of the device. Of these the major number
usually isn't used by the driver code. Major and minor numbers 
.I should
be extracted using the macros 
.I major(dev)
and
.I minor(dev).
.LI
.I mode
is the mode flag passed by the user to the open system call.
.LE 
.H 4 "Close routine"
The close routine is called on the last user close of the device. If there are
any open file descriptors that remain associated with the device after the user
close, the close routine is
.I not
called.
The routine should unlock the device from any exclusive access, disable 
interrupts, tapes should be rewound (if specified by the minor device number)
after any eventual end-of-tape marks have been written. Phone lines should be 
hung up etc.
.br
The syntax at the driver level is:
.Sy
\fIxx\fPclose(dev_t dev,int mode, struct file *fp)
.Se
.BL
.LI
.I dev
is the major/minor device numbers as above.
.LI
.I mode
is the same mode field with which the device was opened, usable for flushing
buffers, rewinding tapes etc.
.LI
.I fp
is the address of the system file table descriptor that corresponds to
the user file descriptor number. This parameter is very seldom used or even
declared but usable i.e. to find inodes.
.LE
.H 4 "Read/write routines"
These routine transfer data from/to the device to user space. They should be
declared as:
.Sy
\fIxx\fPread(dev_t dev,struct uio *uio)
\fIxx\fPwrite(dev_t dev,struct uio *uio)
.Se
.BL
.LI
.I dev
is the major/minor device number.
.LI
.I uio
is the reference to the user I/O buffer structure that contain information
about the data to be transferred.
.LE 1
If the operation is successful return 0, otherwise the error code should be
returned.
.H 4 "Ioctl routine"
This routine should perform any device specific functions that aren't covered by
read, write, open and close. It should be declared as:
.Sy
\fIxx\fPiotcl(dev_t dev, int cmd, caddr_t data, int flag)
.Se
.BL
.LI
.I dev
is the major/minor device number.
.LI
.I cmd
specified which action to be performed, see ioctl.h.
.LI
.I data
is an address to a buffer either containing parameters for the command or
space where results could be placed.
.LI
.I flag
is a parameter usually only used in terminal drivers where it is sent along
to the line discipline's ioctl routine.
.LE
.H 3 "Cooked I/O entry points"
.H 4 "Open routine"
.H 4 "Close routine"
.H 4 "Strategy routine"
.H 3 "I/O routines in general"
.H 4 "Interrupt routine"
.H 2 "What the device driver is supposed to do"
.H 2 "The interface block"
.H 2 "The disk driver"
.H 2 "The tape driver"
.H 2 "The interface between the device driver and the interface block"
.H 2 "Performance issues"
