/*
 * dk --
 *   disk manipulation program
 */
#include <stdio.h>
#include <ctype.h>
#include </sys/h/types.h>
#include </sys/machine/dkio.h>
#include </sys/machineio/scsi.h>
#include </sys/h/ioctl.h>
#include </sys/h/file.h>

#define	equal(s1,s2)	(strcmp(s1, s2) == 0)
#define DKSTART	0
#define DKSTOP	1
#define DKNOP	2
#define DKAS	3
#define DKNAS	4
#define DKCE    5
#define DKCD    6
#define DKMS    7
#define DKPART  8
#define DKHELP  9
#define DKFORMAT 10
#define MTSETDENS 17
#define DEFDISK "/dev/rxt0c"

struct commands {
	char *c_name;
	int c_code;
	int c_ronly;
} com[] = {
	{ "start",	DKSTART,	0 },
	{ "stop",	DKSTOP,		0 },
	{ "autospin",	DKAS,		0 },
	{ "noautospin", DKNAS,		0 },
        { "cacheenable",        DKCE,           0 },
        { "cachedisable",       DKCD,           0 },
        { "modesense",          DKMS,           0 },
        { "partition",          DKPART,         0 },
        { "help",               DKHELP,         0 },
	{ "format",		DKFORMAT,	0 },
	{ "mtsetdens",		MTSETDENS,		0 },
	{ 0 }
};

int dkfd,ret;
char *disk;
char *progname;
struct dkpart dk;
char msbuf[127];
struct scsi_ms ms;
struct scsi_msc msc;
struct scsi_modesense mtms;
struct commands *comp;


main(argc, argv)
	char **argv;
{
	char *getenv();
	register char *cp;
	register struct commands *comp;
	static int spin;

	progname = argv[0];

	if (argc > 2 && (equal(argv[1], "-t") || equal(argv[1], "-f"))) {
		argc -= 2;
		disk = argv[2];
		argv += 2;
	} else
		if ((disk = getenv("DISK")) == NULL)
			disk = DEFDISK;
	if (argc < 2) {
		fprintf(stderr, "usage: %s [ -f device ] command [ count ]\n",progname);
		exit(1);
	}
	cp = argv[1];
	for (comp = com; comp->c_name != NULL; comp++)
		if (strncmp(cp, comp->c_name, strlen(cp)) == 0)
			break;
	if (comp->c_name == NULL) {
		fprintf(stderr, "%s: doesn't understand \"%s\"\n",progname, cp);
		exit(1);
	}
	if ((dkfd = open(disk, comp->c_ronly ? 0 : 2)) < 0) {
		perror(disk);
		exit(1);
	}
	switch(comp->c_code)
	{
	case DKSTART:
		if (ioctl(dkfd, DKIOCSTART) < 0) {
			fprintf(stderr, "%s %s", disk, comp->c_name);
			perror("failed ");
			exit(2);
		}
		break;
		
	case DKSTOP:
		if (ioctl(dkfd, DKIOCSTOP) < 0) {
			fprintf(stderr, "%s %s", disk, comp->c_name);
			perror("failed ");
			exit(2);
		}
		break;
	case DKAS:
		if (ioctl(dkfd, DKIOCSAUTOSPIN) < 0) {
			fprintf(stderr, "%s %s", disk, comp->c_name);
			perror("failed ");
			exit(2);
		}
		break;
	case DKNAS:
		if (ioctl(dkfd, DKIOCSNOAUTOSPIN) < 0) {
			fprintf(stderr, "%s %s", disk, comp->c_name);
			perror("failed ");
			exit(2);
		}
		break;
        case DKCE:
                cache_onoff(1);
                break;
        case DKCD:
                cache_onoff(0);
                break;
        case DKMS:
		printf("Sizeof ms = %d\n",sizeof ms);
                ret = ioctl(dkfd, DKIOCSMODESENSE,&msbuf);
                if (ret < 0) {
                        fprintf(stderr, "%s %s", disk, comp->c_name);
                        perror(" failed ");
                        exit(2);
                }
		xxfillmspages(&msbuf,&ms);
                print_modesense_pages(&ms);
                break;
        case DKPART:
                ret = ioctl(dkfd, DKIOCGPART,&dk) ;
                if (ret  < 0) {
                        fprintf(stderr, "ioctl returned %d for %s %s ",
                                ret,disk, comp->c_name);
                        perror(" failed ");
                        exit(2);
                }
                if( ret !=0 ) printf("Return code = %d\n", ret );
                printf("partition       size    blksize start   ncyl    ntrack	nsector name\n");

                printf("%s:     %d      %d      %d      %d      %d      %d	 %s\n",
                        disk,
                        dk.dk_size,
                        dk.dk_blocksize,
                        dk.dk_start,
                        dk.dk_ncyl,
                        dk.dk_ntrack,
                        dk.dk_nsector,
                        dk.dk_name );

                break;
        case DKHELP:
                int i;
                for (i=0 ; *com[i].c_name != 0 ; i++){
                        printf("        %s\n",com[i].c_name);
                }
                break;

	case DKNOP:
		break;
	case DKFORMAT:
	        dk_format(disk, dkfd);
	        break;
	case MTSETDENS:
		mt_setdens(atoi(argv[2]));
                break;
	}
	close(dkfd);
}
struct scsi_msel_cdb
{
    u_char  sc_op;
    u_char  sc_lun:3;
    u_char  sc_pf:1;
    u_char  sc_resvd:3;
    u_char  sc_smp:1;
    u_char  sc_resvd2[2];
    u_char  sc_pmlen;
    sccb    sc_control;
};
    
struct scsi_fmt_cdb
{
    u_char  sc_op;

    u_char  sc_lun:3;
    u_char  sc_FmtData:1;
    u_char  sc_CmpLst:1;
    u_char  sc_dlf:3;
#define	DLF_BLKADDR	0	/* Block address format	*/
#define	DLF_BYTEINDEX	4	/* Byte distance from index */
#define	DLF_PHYSICAL	5	/* Physical sector address */

    u_char  sc_ipattern;

    u_char  sc_ilve[2];

    sccb    sc_control;
};
struct scsi_fmt_hd
{
    u_char  scfh_code;
    u_char  scfh_fov:1;			/* Format option valid */
    u_char  scfh_dpry:1;		/* Disable primary */
    u_char  scfh_dcrt:1;		/* Disable certification */
    u_char  scfh_stpf:1;		/* Stop format */
    u_char  scfh_resvd:4;		/* Reserved bits */
    u_short scfh_defllen;		/* defect list length */
};

struct scsi_fmt_ddesc_ba
{
    unsigned	scdd_blkaddr;
};

struct scsi_fmt_ddecs_bix
{
    u_char  scdd_cylno[3];
    u_char  scdd_head;
    unsigned scdd_byteaddr;
};

struct scsi_fmt_ddesc_phys
{
    u_char  scdd_cylno[3];
    u_char  scdd_head;
    unsigned  scdd_physsect;
};
struct scsi_msfmt
{
    struct scsi_modesense sms_hd;
    struct scsi_msdadfpage sms_dadf;
};
    
dk_format( char *disk, int dkfd)
{
    struct scsi_fmt_cdb cdb, *cdbp=&cdb;
    struct scsi_msel_cdb mscdb, *mscdbp=&mscdb;
    struct scsi_fmt_hd hd, *hdp=&hd;
    struct scsi_msfmt mspg, *mspgp=&mspg;
    int ret;

    close(dkfd); /* close it since we want it exclusively */

    if ((dkfd = open(disk, O_RDWR  | O_EXCL)) < 0)
    {
	perror(disk);
	exit(1);
    }
    printf("%s exclusively opened\n",disk);

    cdbp->sc_lun = 1;	/* Should be filled by the device driver */
    cdbp->sc_op = SC0_MODE_SENSE;
    cdbp->sc_FmtData = 0;   /* Page format OK */
    cdbp->sc_CmpLst = 0;    /* */
    cdbp->sc_dlf = 0;	/* */
    cdbp->sc_ipattern =  /* 0x40+ */ 0x3F;	/* */
    cdbp->sc_ilve[0] = (sizeof msbuf) >>8 & 0xFF;
    cdbp->sc_ilve[1] = (sizeof msbuf) & 0xFF;

    ioctl(dkfd, DKIOCSSCSI, (char *) cdbp);
    read(dkfd,&msbuf,sizeof msbuf);

    xxfillmspages(&msbuf,&ms);
    print_modesense_pages(&ms);

    mspgp->sms_hd = ms.sms_hd;    /* initialise mspg from ms */
    mspgp->sms_hd.sm_length = 0;  /* Should be set to zero ? */
    mspgp->sms_hd.sm_mtype = 0;   /* Should be set to zero too ? */
    mspgp->sms_hd.sm_wp	= 0;	  /* Should be set ot zero 3 ? */
    mspgp->sms_dadf = ms.sms_daf;

    bzero(&mspgp->sms_dadf, sizeof(mspgp->sms_dadf));
    mspgp->sms_dadf.daf_pcode = 3;
    mspgp->sms_dadf.daf_plen = 22;

    mspgp->sms_dadf.daf_altspz=3;
    mspgp->sms_dadf.daf_tskew=1;
    mspgp->sms_dadf.daf_cskew=15;
  
    bzero(mscdbp,sizeof *mscdbp);
    mscdbp->sc_lun = 0;	/* Should be filled by the device driver */
    mscdbp->sc_op = SC0_MODE_SELECT;
    mscdbp->sc_pf = 1;   /* Page format OK */

    mscdbp->sc_pmlen = (sizeof *mspgp);
   

    if(ioctl(dkfd, DKIOCSSCSI, (char *) mscdbp) < 0)
    {
	perror("Mode select IOCTL:");
	exit(1);
    }
     xxhdump(mspgp, sizeof(*mspgp));
    if(write(dkfd, (char *) mspgp, sizeof(*mspgp)) < 0)
    {
	perror("Mode select Write:");
	exit(1);
    }

    printf("Really format %s ?\n",disk);
    if (getchar()!='Y')
	exit(1);

    bzero(cdbp,sizeof cdbp);
    cdbp->sc_lun = 0;	/* Should be filled by the device driver */
    cdbp->sc_op = SC0_FORMAT_UNIT;
    cdbp->sc_FmtData = 1;   /* User supplied format list header */
    cdbp->sc_CmpLst = 0;    /* won't replace G list */
    cdbp->sc_dlf = DLF_BLKADDR;	/* Block address format */
    cdbp->sc_ipattern=0;
    cdbp->sc_ipattern=0;
    cdbp->sc_ilve[0]=0;
    cdbp->sc_ilve[1]=1;


    hdp->scfh_code = 0;
    hdp->scfh_fov = 1;	/* No format options valid */
    hdp->scfh_dpry = 0;	/* Don't disable primary defect list */
    hdp->scfh_dcrt = 1;	/* Do certification */
    hdp->scfh_stpf = 0; /* Don't stop on error */
    hdp->scfh_defllen = 0; /* No defect list */

    ioctl(dkfd, DKIOCSSCSI, (char *) cdbp);

    write(dkfd, hdp, sizeof *hdp); /* Do the thing */

    close(dkfd);
}


    
    
mt_setdens(int dcode)
{
ret = ioctl(dkfd, DKIOCSMODESENSE,&mtms);
if (ret < 0) {
	fprintf(stderr, "%s %s", disk, comp->c_name);
	perror(" failed ");
	exit(2);
}
mtms.sm_length = 0;
mtms.sm_wp=0;
mtms.sm_speed=0;
mtms.sm_bdarr[0].sbd_density=dcode;
printf("Setting density code=%X\n",dcode);
ret = ioctl(dkfd, DKIOCSMODESELECT,&mtms);
if (ret < 0) {
	fprintf(stderr, "%s %s", disk, comp->c_name);
	perror(" failed ");
	exit(2);
}
ret = ioctl(dkfd, DKIOCSMODESENSE,&mtms);
if (ret < 0) {
        fprintf(stderr, "%s %s", disk, comp->c_name);
        perror(" failed ");
        exit(2);
}
printf(" Mode sense AFTER mode select \n");
print_modesense(&mtms);
}
cache_onoff(int onoff)
{
    ret = ioctl(dkfd, DKIOCSMODESENSE,&ms);
    if (ret < 0) {
	fprintf(stderr, "%s %s", disk, comp->c_name);
	perror(" failed ");
	exit(2);
    }
    msc.sms_hd = ms.sms_hd;    /* initialise msc from ms */
    msc.sms_rac = ms.sms_rac;
    msc.sms_rac.rac_suppr = onoff ? 0xFF:0xFFFF ;
    ret = ioctl(dkfd, DKIOCSMODESELECT,&msc);
    if (ret < 0) {
	    fprintf(stderr, "%s %s", disk, comp->c_name);
	    perror(" failed ");
	    exit(2);
    }
    ret = ioctl(dkfd, DKIOCSMODESENSE,&ms);
    if (ret < 0) {
        fprintf(stderr, "%s %s", disk, comp->c_name);
        perror(" failed ");
        exit(2);
    }
    printf(" Mode sense AFTER mode select \n");
    print_msracache(&ms.sms_rac);
}

#define MIN(x,y) ((x) < (y) ? (x) : (y))

			
xxfillmspages(u_char *from, struct scsi_ms *to)
{
        u_char *fromstart = from;
        struct scsi_mspghd *pghdp;
        int bavail = 0;

        bcopy(from,(char *) to, 4); /* Copy header, beleived to be constant */
        from += 4;

        /* Copy one Block descriptor */
        bcopy(from, (char *) to->sms_hd.sm_bdarr,
                MIN(to->sms_hd.sm_bdl,sizeof(struct scsi_bd)));

        from += to->sms_hd.sm_bdl;

        while(from < (fromstart + to->sms_hd.sm_length))
        {
                pghdp = (struct scsi_mspghd *) from;
                bavail = pghdp->pghd_plen + sizeof (struct scsi_mspghd);

                switch(pghdp->pghd_pcode)
                {
                case 1:         /* errpage */
                        bcopy(from, &to->sms_erp,
                                MIN(sizeof(struct scsi_mserppage), bavail));
                        from += bavail;
                        continue;
                case 2:         /* connpage */
                        bcopy(from, &to->sms_cop,
                                MIN(sizeof(struct scsi_msconnpage), bavail));
                        from += bavail;
                        continue;
                case 3:         /* dadfpage */
                        bcopy(from, &to->sms_daf,
                                MIN(sizeof(struct scsi_msdadfpage), bavail));
                        from += bavail;
                        continue;
                case 4:         /* rigidpage */
                        bcopy(from, &to->sms_rdg,
                                MIN(sizeof(struct scsi_msrigidpage), bavail));
                        from += bavail;
                        continue;
                case 8:         /* Ra cache page */
                        bcopy(from, &to->sms_rac,
                                MIN(sizeof(struct scsi_msracache), bavail));
                        from += bavail;
                        continue;
                case 56:        /* CDC Wren cache page */
                        bcopy(from, &to->sms_csh,
                                MIN(sizeof(struct scsi_mscachepage), bavail));
                        from += bavail;
                        continue;
                case 34:        /* reconnection timing page */
                        bcopy(from, &to->sms_rtc,
                                MIN(sizeof(struct scsi_rtcpage), bavail));
                        from += bavail;
                        continue;
                case 33:        /* additional error recovery page */
                        bcopy(from, &to->sms_aer,
                                MIN(sizeof(struct scsi_msaerpage), bavail));
                        from += bavail;
                        continue;
                case 32:        /* Adaptec dis/reconecct page  */
                        bcopy(from, &to->sms_atc,
                                MIN(sizeof(struct scsi_msatconnpage), bavail));
                        from += bavail;
                        continue;
                case 7:         /* Verify error recovery page */
                        bcopy(from, &to->sms_ver,
                                MIN(sizeof(struct scsi_msverpage), bavail));
                        from += bavail;
                        continue;
                default:
                        printf("Unknown mode sense page code 0x%x, length=%d bytes\n",
                                pghdp->pghd_pcode, pghdp->pghd_plen);
                        from += bavail;
                        continue;
                }
        }

}


print_modesense_pages(struct scsi_ms *p)
{
	if(p->sms_hd.sm_length >0) print_modesense(&p->sms_hd);
	if(p->sms_erp.erp_plen >0) print_mserppage(&p->sms_erp);
	if(p->sms_cop.cop_plen >0) print_msconnpage(&p->sms_cop);
	if(p->sms_daf.daf_plen >0) print_msdadfpage(&p->sms_daf);
	if(p->sms_rdg.rdg_plen >0) print_msrigidpage(&p->sms_rdg);
	if(p->sms_csh.csh_plen >0) print_mscachepage(&p->sms_csh);
	if(p->sms_rac.rac_plen >0) print_msracache(&p->sms_rac);
	if(p->sms_rtc.rtc_plen >0) print_msrtcpage(&p->sms_rtc);
	if(p->sms_aer.aer_plen >0) print_msaerpage(&p->sms_aer);
	if(p->sms_atc.atc_plen >0) print_msatcpage(&p->sms_atc);
	if(p->sms_ver.ver_plen >0) print_msverpage(&p->sms_ver);
}

print_modesense(struct scsi_modesense *p) 
{
printf("Mode Sense, Sense data length 		%d\n",p->sm_length);  
printf("Mode Sense, Medium type 		%d\n",p->sm_mtype);   
printf("Mode Sense, Write-protect 		%x\n",p->sm_wp);    
printf("Mode Sense, Maxtor Buffered mode 	%x\n",p->sm_bufmode);
printf("Mode Sense, Maxtor Speed code 		%x\n",p->sm_speed); 
printf("Mode Sense, Block descriptor length 	%d\n",p->sm_bdl);     
printf("Mode Sense Block descriptor, density code 	%d\n",
	p->sm_bdarr[0].sbd_density);     
printf("Mode Sense Block descriptor, Number of blocks 	%d\n",
	p->sm_bdarr[0].sbd_nblocks);     
printf("Mode Sense Block descriptor, Block length 	%d\n\n",
	p->sm_bdarr[0].sbd_blklen);     

};
print_mserppage(struct scsi_mserppage *p) 
{ 
printf("Error recovery page, parameter savable		%x\n", p->erp_ps);
printf("Error recovery page, Page code			%x\n", p->erp_pcode);
printf("Error recovery page, Page length		%d\n", p->erp_plen);
printf("Error recovery page, Automatic write reallocation enable	%x\n", p->erp_awre);
printf("Error recovery page, Automatic read reallocation enable		%x\n", p->erp_arre);
printf("Error recovery page, Transfer block with error	%x\n", p->erp_tb);
printf("Error recovery page, Read continous		%x\n", p->erp_rc);
printf("Error recovery page, Enable early correction	%x\n", p->erp_eec);
printf("Error recovery page, Post error			%x\n", p->erp_per);
printf("Error recovery page, Disable transfer on error	%x\n", p->erp_dte);
printf("Error recovery page, Disable correction		%x\n", p->erp_dcr);
printf("Error recovery page, Retry count		%d\n", p->erp_rrtcnt);
printf("Error recovery page, Correction span		%d\n", p->erp_cspan);
printf("Error recovery page, Head offset count		%d\n", p->erp_hoffset);
printf("Error recovery page, ata strobe offset count	%d\n", p->erp_dsoffset);
printf("Error recovery page, Recovery time limit	%d\n\n", p->erp_rtl);

};
print_msconnpage(struct scsi_msconnpage *p) 
{
printf("Connect Page , parameter savable	%x\n", p->cop_ps);
printf("Connect Page , Page code		%x\n", p->cop_pcode);
printf("Connect Page , Page length		%d\n", p->cop_plen);
printf("Connect Page , Buffer full ratio	%d\n", p->cop_bfr);
printf("Connect Page , Buffer empty ratio	%d\n", p->cop_ber);
printf("Connect Page , Bus inactivity limit	%d\n", p->cop_bil);
printf("Connect Page , Disconnect time limit	%d\n", p->cop_dtl);
printf("Connect Page , Connect time limit	%d\n\n", p->cop_ctl);

};
print_msdadfpage(struct scsi_msdadfpage *p) 
{ 
   printf("Format Page, Parameter Savable		%x \n",p->daf_ps); 
   printf("Format Page, Page code			%x \n",p->daf_pcode); 
   printf("Format Page, Page length			%d \n",p->daf_plen);    
   printf("Format Page, Tracks per zone			%d \n",p->daf_tpz);     
   printf("Format Page, Alternate sectors per zone	%d \n",p->daf_altspz);  
   printf("Format Page, Alternate tracks per zone	%d \n",p->daf_atttpz);  
   printf("Format Page, Alternate tracks per volume	%d \n",p->daf_alttpv);  
   printf("Format Page, Sectors per track		%d \n",p->daf_spt);     
   printf("Format Page, No of bytes per sector		%d \n",p->daf_nbps);    
   printf("Format Page, Interleave factor		%d \n",p->daf_intrlve); 
   printf("Format Page, Track skew			%d \n",p->daf_tskew);   
   printf("Format Page, Cylinder skew			%d \n",p->daf_cskew);   
   printf("Format Page, Soft sector format		%x \n",p->daf_ssec);  
   printf("Format Page, Hard sector format		%x \n",p->daf_hsec);  
   printf("Format Page, Removable media			%x \n",p->daf_rmb);   
   printf("Format Page, Format by surface		%x \n",p->daf_surf);  
   printf("Format Page, Inhibit Save			%x \n\n",p->daf_ins);   

};
print_msrigidpage(struct scsi_msrigidpage *p) 
{ 
 printf("Rigid page,  parameter savable            %x \n",p->rdg_ps);      
 printf("Rigid page,  Page code                    %x \n",p->rdg_pcode);   
 printf("Rigid page,  Page length                  %d \n",p->rdg_plen);      
 printf("Rigid page,  Maximum no of cylinders      %d \n",
	(p->rdg_maxcyl[0]<<16) |
	(p->rdg_maxcyl[1]<<8 ) |
	(p->rdg_maxcyl[2] )
	); 
 printf("Rigid page,  Maximum no of heads          %d \n\n",p->rdg_hmaxheads); 
};
print_mscachepage(struct scsi_mscachepage *p) 
{
	printf("Wren Cache Parameter Savable:	%s\n",
	       p->csh_ps ? "Yes" : "No" );
	printf("Wren Cache page code:		%x\n", p->csh_pcode);
	printf("Wren Cache page length:		%d\n",p->csh_plen);
	printf("Wren Write Index enabled :	%s\n",
	       p->csh_wie ? "Yes" : "No" );
	printf("Wren Caching enabled :		%s\n",
	       p->csh_ce ? "Yes" : "No" );
	printf("Wren Cache table size:		%d\n",p->csh_cts);
	printf("Wren Cache Prefetch Threshold:	%x\n\n",
                        p->csh_pt);
}
print_msracache(struct scsi_msracache *p) 
{
	printf("Cache Parameter Savable:	%s\n",
	       p->rac_ps ? "Yes" : "No" );
	printf("Cache page code:		%x\n", p->rac_pcode);
	printf("Cache page length:		%d\n",p->rac_plen);
	printf("Cache Multiple:	       		%s\n", p->rac_ms ? "Yes":"No" );
	printf("Read Cache disabled :		%s\n", p->rac_rcd ? "Yes":"No");
	printf("Cache Suppression :		%x\n", p->rac_suppr );
	printf("Cache Min prefetch :		%d\n", p->rac_min_pre );
	printf("Cache Max prefetch :		%d\n", p->rac_max_pre );
	printf("Cache Max prefetch restriction :%d\n", p->rac_max_pre_restr );
}

print_msrtcpage(struct scsi_rtcpage *p) 
{
	printf("RC timing Parameter Savable:	%s\n",
	       p->rtc_ps ? "Yes" : "No" );
	printf("RC timing page code:		%x\n", p->rtc_pcode);
	printf("RC timing page length:		%d\n",p->rtc_plen);
	printf("RC timing reserved byte 1:	%d\n",p->rtc_resv1);
	printf("RC timing reserved byte 2:	%d\n",p->rtc_resv2);
}

print_msaerpage(struct scsi_msaerpage *p) 
{
	printf("Addl err rec Parameter Savable:	%s\n",
	       p->aer_ps ? "Yes" : "No" );
	printf("Addl err rec page code:		%x\n", p->aer_pcode);
	printf("Addl err rec page length:		%d\n",p->aer_plen);
	printf("Addl err rec cmd exec delay:	%sabled\n",p->aer_dced?"Dis":"En");
	printf("Addl err rec Post SCSI err:	%sabled\n",p->aer_pser?"En":"Dis");
	printf("Addl err rec Seek retry cnt:	%d\n",p->aer_srtc);
}

print_msverpage(struct scsi_msverpage *p) 
{
	printf("Vfy err rec Parameter Savable:	%s\n",
	       p->ver_ps ? "Yes" : "No" );
	printf("Vfy err rec page code:		%x\n", p->ver_pcode);
	printf("Vfy err rec page length:		%d\n",p->ver_plen);
	printf("Vfy err rec Early recovery:	%sabled\n",p->ver_eer?"En":"Dis");
	printf("Vfy err rec Post SCSI err:	%sabled\n",p->ver_per?"En":"Dis");
	printf("Vfy err rec Xfer on error:	%sabled\n",p->ver_dte?"Dis":"En");
	printf("Vfy err rec Error correction:	%sabled\n",p->ver_dcr?"Dis":"En");
	printf("Vfy err rec Retry count:	%d\n",p->ver_vrtc);
	printf("Vfy err rec Correctable bit len:%d\n",p->ver_cbl);
}

print_msatcpage(struct scsi_msatconnpage *p) 
{
	printf("Adaptec conn Parameter Savable:	%s\n",
	       p->atc_ps ? "Yes" : "No" );
	printf("Adaptec conn page code:		%x\n", p->atc_pcode);
	printf("Adaptec conn page length:		%d\n",p->atc_plen);
	printf("Adaptec conn Host rec overhead:	%d uS\n",p->atc_hroh*100);
	printf("Adaptec conn Buffer pre-fill:	%d bytes\n",p->atc_bpfill*256);
	printf("Adaptec conn Modift data ptr:	%sabled\n",p->atc_mdp?"En":"Dis");
}
char *
xxcstr(c)
u_char c;
{
    static char     outstr[3];
    char           *convstr = "0123456789abcdef";

    outstr[0] = convstr[(c >> 4) & 0xf];
    outstr[1] = convstr[c & 0xf];
    outstr[2] = '\0';

    return outstr;
}


xxhdump(cp, n)
u_char         *cp;
int             n;
{
    int             i = 0;

    while (n--)
    {
        printf("%s", xxcstr(*cp++));
        ++i;
        if ((i % 32) == 0)
        {
            printf("\n");
        }
        else
        {
            if ((i % 4) == 0)
            {
                printf(" ");
            }
        }
    }
    if (i % 32)
    {
        printf("\n");
    }
}

