#include "dk.h"

struct scsi_fmt_cdb
{
    u_char  sc_op;

    u_char  sc_lun:3;
    u_char  sc_FmtData:1;
    u_char  sc_CmpLst:1;
    u_char  sc_dlf:3;
#define	DLF_BLKADDR	0	/* Block address format	*/
#define	DLF_BYTEINDEX	4	/* Byte distance from index */
#define	DLF_PHYSICAL	5	/* Physical sector address */

    u_char  sc_ipattern;

    u_char  sc_ilve[2];

    sccb    sc_control;
};

struct scsi_fmt_hd
{
    u_char  scfh_code;
    u_char  scfh_fov:1;			/* Format option valid */
    u_char  scfh_dpry:1;		/* Disable primary */
    u_char  scfh_dcrt:1;		/* Disable certification */
    u_char  scfh_stpf:1;		/* Stop format */
    u_char  scfh_resvd:4;		/* Reserved bits */
    u_short scfh_defllen;		/* Defect list length */
    struct  defdesc scfh_def[512];	/* Defect list */
};

struct scsi_msfmt
{
    struct scsi_modesense sms_hd;
    struct scsi_msdadfpage sms_dadf;
};

dk_format( char *disk, int dkfd)
{
    struct scsi_fmt_cdb cdb, *cdbp=&cdb;
    struct scsi_msel_cdb mscdb, *mscdbp=&mscdb;
    struct scsi_fmt_hd hd, *hdp=&hd;
    struct scsi_msfmt mspg, *mspgp=&mspg;
    struct scsi_ms ms;
    char flawfile[512];
    FILE 	*ff;

    close(dkfd); /* close it since we want it exclusively */

    if ((dkfd = open(disk, O_RDWR  | O_EXCL)) < 0)
    {
	perror(disk);
	exit(1);
    }
    printf("%s exclusively opened\n",disk);

    ioctl(dkfd, DKIOCDSR, 0);	/* Disable driver retries */
    cdbp->sc_lun = 1;	/* Should be filled by the device driver */
    cdbp->sc_op = SC0_MODE_SENSE;
    cdbp->sc_FmtData = 0;   /* Page format OK */
    cdbp->sc_CmpLst = 0;    /* */
    cdbp->sc_dlf = 0;	/* */
    cdbp->sc_ipattern =  /* 0x40+ */  0x3F;	/* */
    cdbp->sc_ilve[0] = (sizeof msbuf) >>8 & 0xFF;
    cdbp->sc_ilve[1] = (sizeof msbuf) & 0xFF;

    ioctl(dkfd, DKIOCSSCSI, (char *) cdbp);
    read(dkfd,&msbuf,sizeof msbuf);

    xxfillmspages(&msbuf,&ms);
    print_modesense_pages(&ms);

    mspgp->sms_hd = ms.sms_hd;    /* initialise mspg from ms */
    mspgp->sms_hd.sm_length = 0;  /* Should be set to zero ? */
    mspgp->sms_hd.sm_mtype = 0;   /* Should be set to zero too ? */
    mspgp->sms_hd.sm_wp	= 0;	  /* Should be set ot zero 3 ? */
    mspgp->sms_dadf = ms.sms_daf;

    bzero(&mspgp->sms_dadf, sizeof(mspgp->sms_dadf));
    mspgp->sms_dadf.daf_pcode = 3;
    mspgp->sms_dadf.daf_plen = 22;

    mspgp->sms_dadf.daf_altspz=6;
    mspgp->sms_dadf.daf_tskew=1;
    mspgp->sms_dadf.daf_cskew=15;
  
    bzero(mscdbp,sizeof *mscdbp);
    mscdbp->sc_lun = 0;	/* Should be filled by the device driver */
    mscdbp->sc_op = SC0_MODE_SELECT;
    mscdbp->sc_pf = 1;   /* Page format OK */

    mscdbp->sc_pmlen = (sizeof *mspgp);
   

    if(ioctl(dkfd, DKIOCSSCSI, (char *) mscdbp) < 0)
    {
	perror("Mode select IOCTL:");
	exit(1);
    }
     xxhdump(mspgp, sizeof(*mspgp));
    if(write(dkfd, (char *) mspgp, sizeof(*mspgp)) < 0)
    {
	perror("Mode select Write:");
	exit(1);
    }

    cdbp->sc_lun = 1;	/* Should be filled by the device driver */
    cdbp->sc_op = SC0_MODE_SENSE;
    cdbp->sc_FmtData = 0;   /* Page format OK */
    cdbp->sc_CmpLst = 0;    /* */
    cdbp->sc_dlf = 0;	/* */
    cdbp->sc_ipattern =  /* 0x40+ */ 0x03;	/* */
    cdbp->sc_ilve[0] = (sizeof msbuf) >>8 & 0xFF;
    cdbp->sc_ilve[1] = (sizeof msbuf) & 0xFF;

    bzero(&msbuf, sizeof msbuf);
    ioctl(dkfd, DKIOCSSCSI, (char *) cdbp);
    read(dkfd,&msbuf,sizeof msbuf);
    bzero(&ms, sizeof ms);

    xxfillmspages(&msbuf,&ms);
    print_modesense_pages(&ms);
    printf("Flaw map file ?");
    scanf("%s",flawfile);
    if((ff=fopen(flawfile,"r")) != NULL)
    {
	int fi;
	int length,i;
	struct defdesc *defp;

        hdp->scfh_defllen = 0;
	for(fi=0; fi<512; fi++)
	{
	    char line[512];
	    int	h,c,b,l;
	    int eof=0;
	    do
	    {
		if(fgets(line,512,ff) == NULL)
		{
		    eof=1;
		    break;
		}
	    } while(line[0]=='#');

	    if (eof) break;

	    sscanf(line,"%d %d %d %d",&h,&c,&b,&l);
	    hdp->scfh_def[fi].headno=h;
	    hdp->scfh_def[fi].cylno=c;
	    hdp->scfh_def[fi].bytesno=b;
	    hdp->scfh_defllen += sizeof(struct defdesc);
	}

	printf("User supplied defect list: ");
	length=hdp->scfh_defllen;
	printf("length=%d containing %d defects\n",length, length/sizeof (struct defdesc));
	printf("No.     Cyl     Hd      Bytes\n");

	for(i=1,defp=hdp->scfh_def; length > 0; i++,defp++)
	{
		printf("%d.     %d      %d      %d\n",i,
			defp->cylno,defp->headno,defp->bytesno);
		length -= sizeof (struct defdesc);
	}

    }
		
			
    fflush(stdin);

    bzero(cdbp,sizeof cdbp);
    cdbp->sc_lun = 0;	/* Should be filled by the device driver */
    cdbp->sc_op = SC0_FORMAT_UNIT;
    cdbp->sc_FmtData = 1;   /* User supplied format list header */
    cdbp->sc_CmpLst = 1;    /* Will replace G list */
    cdbp->sc_dlf = DLF_BYTEINDEX;	/* Block address format */
    cdbp->sc_ipattern=0;
    cdbp->sc_ipattern=0;
    cdbp->sc_ilve[0]=0;
    cdbp->sc_ilve[1]=1;


    hdp->scfh_code = 0;
    hdp->scfh_fov = 1;	/* Format options valid */
    hdp->scfh_dpry = 1;	/* Don't disable primary defect list */
    hdp->scfh_dcrt = 1;	/* Don't do certification */
    hdp->scfh_stpf = 1; /* Don't stop on error */

    if(ioctl(dkfd, DKIOCSSCSI, (char *) cdbp))
	perror("Format ioctl");

    if(write(dkfd, hdp, 4+hdp->scfh_defllen) < 0); /* Do the thing */
    {
	struct scsi_esense esense;
	perror("Format write");
	ioctl(dkfd,DKIOCGSCSISENSE,(char *) &esense);
	
	xxhdump(&esense,sizeof esense);
    }

    ioctl(dkfd, DKIOCESR, 0);	/* Enable driver retries */

    close(dkfd);
}

