#if	CMU
/* 
 * Mach Operating System
 * Copyright (c) 1986 Carnegie-Mellon University
 * All rights reserved.  The CMU software License Agreement specifies
 * the terms and conditions for use and redistribution.
 *
 **********************************************************************
 * HISTORY
 * 11-april-89 Hakan Winbom, IBM Watson
 *      Incorporated SCSI stuff from Chalmers
 *      aha driver for adaptech AHA 1540A/1542 scsi board
 *      xt         for SCSI disks (Maxtor xt4380S)
 *      cs         for SCSI tape (Perrtec CS240 3480 cassete tape )
 *
 * 16-Oct-87  Mike Accetta (mja) at Carnegie-Mellon University
 *	Added new SCSI driver linkages.
 *	[ V5.1(XF19) ]
 *
 * 27-Aug-87  Richard Sanzi (sanzi) at Carnegie-Mellon University
 *	Changed all declarations of debugging variables to integers, and
 *	added the include of "debug.h".
 *
 * 04-May-87  Mike Accetta (mja) at Carnegie-Mellon University
 *	CS_GENERIC: Added kernel log device.
 *	[ V5.1(F9) ]
 *
 *  9-Mar-87  Avadis Tevanian (avie) at Carnegie-Mellon University
 *	No longer support swstrategy, swread and swwrite.
 *
 * 05-Feb-87  Mike Accetta (mja) at Carnegie-Mellon University
 *	Completed installation of CMU pseudo-terminal hooks; fixed to
 *	correctly initialize console numbers for the terminal
 *	location table.
 *	[ V5.1(F2) ]
 *
 * 29-Jan-87  Bill Bolosky (bolosky) at Carnegie-Mellon University
 *	Changed for X.
 *
 * 16-Jul-86  Bill Bolosky (bolosky) at Carnegie-Mellon University
 *	Removed definition of swapdev.
 *
 * 25-Feb-86  Bill Bolosky (bolosky) at Carnegie-Mellon University
 *	Added include of ioctl.h.
 *
 **********************************************************************
 */

#include "cs_generic.h"
#include "cs_ttyloc.h"
#endif	CMU

/*     conf.c  6.1     83/07/29        */
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/buf.h"
#include "../h/ioctl.h"
#include "../h/tty.h"
#include "../h/conf.h"
#include "../ca/debug.h"

int nulldev();
int nodev();
#if	CS_GENERIC
int	logopen(),logclose(),logread(),logioctl(),logselect();
#endif	CS_GENERIC


#include "dk.h"
#if NDK > 0
int dkinit(), dkstrategy();
#else NDK
#define dkstrategy nodev
#define dkinit	nodev
#endif NDK

#include "hd.h"
#if NHD > 0
int hdinit(), hdstrategy(), hdread(), hdwrite(), hdopen(), hddump(), hdsize();
#else NHD
#define hdinit	nodev
#define hdstrategy	nodev
#define hdread	nodev
#define hdwrite	nodev
#define hdopen	nodev
#define hddump	nodev
#define hdsize	nodev
#endif NHD

#include "ud.h"
#if NUD > 0
int udinit(), udstrategy(), udread(), udwrite(), udopen();
#else NUD
#define udinit	nodev
#define udstrategy	nodev
#define udread	nodev
#define udwrite	nodev
#define udopen	nodev
#endif NUD

#include "fd.h"
#if NFD > 0
int fdinit(), fdstrategy(), fdread(), fdwrite(), fdopen(), fdclose(), fdsize(),fdioctl();
#else NFD
#define fdinit	nodev
#define fdstrategy	nodev
#define fdread	nodev
#define fdwrite	nodev
#define fdopen	nodev
#define fdclose	nodev
#define fdioctl	nodev
#define fdsize	nodev
#endif NFD

#include "sc.h"
#if NSC > 0
int scsiinit(), scsistrategy(), scsiread(), scsiwrite(), scsiopen(), scsisize(), scsiioctl();
#else NSC
#define scsiinit	nodev
#define scsistrategy	nodev
#define scsiread	nodev
#define scsiwrite	nodev
#define scsiopen	nodev
#define scsisize	nodev
#define scsiioctl	nodev
#endif NSC
/* HW SCSI stuff */
#include "cs.h"
#include "xt.h"
#if NCSC > 0
int   csopen(), csclose(), csread(), cswrite(), csioctl(), csstrategy(),
      csminphys();
#define               csselect        seltrue
#else
#define               csopen          nodev
#define               csclose         nodev
#define               csread          nodev
#define               cswrite         nodev
#define               csioctl         nodev
#define               csselect        nodev
#define               csstrategy      nodev
#define               csminphys       nodev
#endif NCS
#if NXTC > 0
int   xtopen(),xtclose(),xtread(),xtwrite(),xtioctl(),xtstrategy()    ,
      xtminphys(); xtsize();
#define xtselect      seltrue
#else
#define               xtopen          nodev
#define               xtclose         nodev
#define               xtread          nodev
#define               xtwrite         nodev
#define               xtioctl         nodev
#define               xtselect        nodev
#define               xtstrategy      nodev
#define               xtminphys       nodev
#define               xtsize          nodev
#endif NXT

/* HW end */

#if	MACH
#define swstrategy	nodev
#define swread		nodev
#define swwrite		nodev
#else	MACH
int swstrategy(), swread(), swwrite();
#endif	MACH

#include "st.h"
#if NST > 0
int stinit(), ststrategy(), stread(), stwrite(), stopen(), stclose(), stsize(), stioctl();
#else NST
#define stinit	nodev
#define ststrategy	nodev
#define stread	nodev
#define stwrite	nodev
#define stopen	nodev
#define stclose	nodev
#define stioctl	nodev
#define stsize	nodev
#endif NST

struct bdevsw bdevsw[] = {
	/* open		close		strat		dump
	  size		flags 	minphys */
	{ nulldev,	nulldev,	dkstrategy,	nodev,	/*0*/
	  0,		0 ,	nodev},
	{ hdopen,	nulldev,	hdstrategy,	hddump,	/*1*/
	  hdsize,	0 ,	nodev},
	{ udopen,	nulldev,	udstrategy,	nodev,	/*2*/
	  0,		0 ,	nodev},
	{ fdopen,	fdclose,	fdstrategy,	nodev,	/*3*/
	  0,		0 ,	nodev},
	{ nodev,	nodev,		swstrategy,	nodev,	/*4*/ 
	  0,		0 ,	nodev},
	{ stopen,	stclose,	ststrategy,	nodev,	/*5*/
	  0,		B_TAPE ,	nodev},
#if	CS_GENERIC
	{ nodev,	nodev,		nodev,		nodev,	/*6*/
	  0,		0 ,	nodev},
	{ scsiopen,	nulldev,	scsistrategy,	nodev,	/*7*/
	  scsisize,	0 ,	nodev},
#endif	CS_GENERIC
	{ nodev,        nodev,          nodev,          nodev,  /*8*/
          0,            0,      nodev },
        { nodev,        nodev,          nodev,          nodev,  /*9*/
          0,            0,      nodev },
/* 0-9 reserved to IBM */
  /* 10-19 reserved to local sites */
/* HW Chalmers SCSI stuff */
       { csopen,       csclose,        csstrategy,     nodev,  /*10*/
         0,            B_TAPE, csminphys },
       { xtopen,       nulldev,        xtstrategy,     nodev,  /*11*/
         xtsize,       0,      xtminphys },

/* HW Chalmers SCSI stuff end */
};
int nblkdev = sizeof(bdevsw) / sizeof(bdevsw[0]);

int cnopen(), cnclose(), cnread(), cnwrite(), cnioctl(), cnselect();
struct tty cons;
#if	CS_TTYLOC
int	ncons = 1;
#define	cnntty	(&ncons)
#endif	CS_TTYLOC

#include "asy.h"
#if NASY > 0
int asyopen(), asyclose(), asyread(), asywrite(), asyioctl(), asyselect();
struct tty asy[];
#else
#define		asyopen		nodev
#define		asyclose	nodev
#define		asystrategy	nodev
#define		asyread		nodev
#define		asywrite	nodev
#define		asyioctl	nodev
#define		asyselect	nodev
#define		asy	0
#endif

int syopen(), syread(), sywrite(), syioctl(), syselect();

int mmread(), mmwrite();
#define        mmselect        seltrue

#include "pty.h"
#if	NPTY > 0
int ptsopen(), ptsclose(), ptsread(), ptswrite(), ptsstop();
int ptcopen(), ptcclose(), ptcread(), ptcwrite(), ptcselect();
int ptyioctl();
struct tty pt_tty[];
#if	CS_TTYLOC
int	npty;
#define	ptsntty	(&npty)
#endif	CS_TTYLOC
#else	NPTY > 0
#define ptsopen         nodev
#define ptsclose        nodev
#define ptsread         nodev
#define ptswrite        nodev
#define ptcopen         nodev
#define ptcclose        nodev
#define ptcread         nodev
#define ptcwrite        nodev
#define ptyioctl        nodev
#define pt_tty          0
#define ptcselect       nodev
#define ptsstop         nulldev
#if	CS_TTYLOC
#define	ptsntty		0
#endif	CS_TTYLOC
#endif

#include "lp.h"
#if NLP > 0
int lpopen(), lpclose(), lpwrite();
#else
#define lpopen nodev
#define lpclose nodev
#define lpwrite nodev
#endif

int busopen(), busclose();

#include "psp.h"
#if NPSP > 0
int pspopen(), pspclose(), pspread(), pspwrite(), pspioctl(), pspselect();
struct tty psp[];
#else
#define		pspopen		nodev
#define		pspclose	nodev
#define		pspstrategy	nodev
#define		pspread		nodev
#define		pspwrite	nodev
#define		pspioctl	nodev
#define		pspselect	nodev
#define		psp	0
#endif

int spkopen(), spkwrite(), spkclose();

#include "cmupty.h"
#if	NCMUPTY > 0
int	cmuptyopen(),cmuptyclose(),cmuptyread(),cmuptywrite(),cmuptyioctl();
int	cmuptycopen(),cmuptycclose(),cmuptycread(),cmuptycwrite(),cmuptycioctl();
int	cmuptycselect();
struct tty cmupty_tty[];
#if	CS_TTYLOC
int	ncmupty;
#define	cmuptyntty	(&ncmupty)
#endif	CS_TTYLOC
#else	NCMUPTY
#define	cmuptyopen	nodev
#define	cmuptyclose	nodev
#define	cmuptyread	nodev
#define	cmuptywrite	nodev
#define	cmuptyioctl	nodev
#define	cmuptycopen	nodev
#define	cmuptycclose	nodev
#define	cmuptycread	nodev
#define	cmuptycwrite	nodev
#define	cmuptycioctl	nodev
#define	cmuptycselect	nodev
#define	cmupty_tty	0
#if	CS_TTYLOC
#define	cmuptyntty	0
#endif	CS_TTYLOC
#endif	NCMUPTY

#include "ms.h"
#if NMS > 0
int msopen(), msclose(), msread(), msioctl(), msselect();
struct  tty ms_tty[];
#else
#define		msopen		nodev
#define		msclose		nodev
#define		msread		nodev
#define		msioctl		nodev
#define 	ms_tty		0
#define 	msselect	nodev
#endif

#include "vice.h"
#if	VICE
#include	"vicefs.h"
#if	NVICEFS > 0
int	rmtopen(), rmtclose(), rmtread(), rmtwrite(), rmtselect();
#else
#define		rmtopen	nodev
#define		rmtclose	nodev
#define		rmtread	nodev
#define		rmtwrite	nodev
#define		rmtselect	nodev
#endif
int	rmtopen(), rmtclose(), rmtread(), rmtwrite(), rmtselect();
#endif	VICE

int ttselect(), seltrue();

struct cdevsw cdevsw[] = {
	/* open		close		read		write
	  ioctl		stop		reset		tty
	  select	mmap	*/
	{ cnopen,	cnclose,	cnread,		cnwrite,	/*0*/
	  cnioctl,	nulldev,	nulldev,	&cons,
	  cnselect,	nodev },
	{ asyopen,	asyclose,	asyread,	asywrite,	/*1*/
	  asyioctl,	nulldev,	nulldev,	asy,
	  asyselect,	nodev },
	{ syopen,	nulldev,	syread,		sywrite,	/*2*/
	  syioctl,	nulldev,	nulldev,	0,
	  syselect,	nodev },
	{ nulldev,	nulldev,	mmread,		mmwrite,	/*3*/
	  nodev,	nulldev,	nulldev,	0,
	  mmselect,	nodev },
	{ hdopen,	nulldev,	hdread,		hdwrite,	/*4*/
	  nulldev,	nulldev,	nulldev,	0,
	  seltrue,	nodev },
	{ udopen,	nulldev,	udread,		udwrite,	/*5*/
	  nulldev,	nulldev,	nulldev,	0,
	  seltrue,	nodev },
	{ ptsopen,	ptsclose,	ptsread,	ptswrite,	/*6*/
	  ptyioctl,	ptsstop,	nodev,		pt_tty,
	  ttselect,	nodev },
	{ ptcopen,	ptcclose,	ptcread,	ptcwrite,	/*7*/
	  ptyioctl,	nulldev,	nodev,		pt_tty,
	  ptcselect,	nodev },
	{ lpopen,	lpclose,	nodev,		lpwrite,	/*8*/
	  nodev,	nulldev,	nodev,		0,
	  seltrue,	nodev },
	{ nulldev,	nulldev,	swread,		swwrite,	/*9*/
	  nodev,	nodev,		nulldev,	0,
	  nodev,	nodev },
	{ fdopen,	fdclose,	fdread,		fdwrite,	/*10*/
	  fdioctl,	nulldev,	nulldev,	0,
	  seltrue,	nodev },
	{ stopen,	stclose,	stread,		stwrite,	/*11*/
	  stioctl,	nulldev,	nulldev,	0,
	  seltrue,	nodev },
	{ pspopen,	pspclose,	pspread,	pspwrite,	/*12*/
	  pspioctl,	nulldev,	nulldev,	psp,
	  pspselect,	nodev },
	{ busopen,	busclose,	nodev,		nodev,		/*13*/
	  nodev,	nodev,		nodev,		0,
	  nodev,	nodev },
	{ spkopen,	spkclose,	nulldev,	spkwrite,	/*14*/
	  nulldev,	nulldev,	nulldev,	0,
	  seltrue,	nodev },
	{ msopen,	msclose,	msread,		nodev,		/*15*/
	  msioctl,	nulldev,	nulldev,	ms_tty,
	  msselect,	nodev },
#if	VICE
	{ rmtopen,	rmtclose,	rmtread,	rmtwrite,	/*16*/
	  nodev,	nodev,		nodev,		0,
  	  rmtselect,	nodev},
#else	VICE
  	{ nodev,	nodev,		nodev,		nodev,		/*16*/
	  nodev,	nodev,		nodev,		0,
	  nodev,	nodev },
#endif	VICE
#if	CS_GENERIC
	{ cmuptyopen,	cmuptyclose,	cmuptyread,	cmuptywrite,	/*17*/
	  cmuptyioctl,	nulldev,	nulldev,	cmupty_tty,
	  ttselect,	nodev },
	{ cmuptycopen,	cmuptycclose,	cmuptycread,	cmuptycwrite,	/*18*/
	  cmuptycioctl,	nodev,		nulldev,	0,
	  cmuptycselect,nodev }, 
	{ scsiopen,	nulldev,	scsiread,	scsiwrite,	/*19*/
	  scsiioctl,	nulldev,	nulldev,	0,
	  seltrue,	nodev },
	{ logopen,	logclose,	logread,	nodev,		/*20*/
	  logioctl,	nodev,		nulldev,	0,
	  logselect,	nodev },
#endif	CS_GENERIC
        { nodev,        nodev,          nodev,          nodev,          /*21*/
          nodev,        nodev,          nodev,          0,
          nodev,        nodev },
        { nodev,        nodev,          nodev,          nodev,          /*22*/
          nodev,        nodev,          nodev,          0,
          nodev,        nodev },
        { nodev,        nodev,          nodev,          nodev,          /*23*/
          nodev,        nodev,          nodev,          0,
          nodev,        nodev },
        { nodev,        nodev,          nodev,          nodev,          /*24*/
          nodev,        nodev,          nodev,          0,
          nodev,        nodev },

  /* 0-24 reserved to IBM */
  /* 25-39 reserved to local sites */
 /* Pertec CS240 streamer on a SCSI Adapter */
       { csopen,       csclose,        csread,         cswrite,        /*25*/
         csioctl,      nulldev,        nulldev,        0,
         csselect,     nodev },
       { xtopen,       nulldev,        xtread,         xtwrite,        /*26*/
         xtioctl,      nulldev,        nulldev,        0,
         xtselect,     nodev },


};
int nchrdev = sizeof(cdevsw) / sizeof(cdevsw[0]);

#if	CS_TTYLOC
int *nttysw[] =
{
    cnntty,	/*  0 */
    0,		/*  1 */
    0,		/*  2 */
    0,		/*  3 */
    0,		/*  4 */
    0,		/*  5 */
    ptsntty,	/*  6 */
    0,		/*  7 */
    0,		/*  8 */
    0,		/*  9 */
    0,		/* 10 */
    0,		/* 11 */
    0,		/* 12 */
    0,		/* 13 */
    0,		/* 14 */
    0,		/* 15 */
    0,		/* 16 */
    cmuptyntty,	/* 17 */
};
int	ntty = sizeof (nttysw) / sizeof (nttysw[0]);

#endif	CS_TTYLOC
int mem_no = 3;		/* major device number of memory special file */


/* Romp/XT debugging switches -- may be patched during execution */

int vmdebug = 0x00;		  /* virtual memory */
int trdebug = 0x00;		  /* trap traces */
int svdebug = 0x00;		  /* svc traces */
int ttydebug = 0;		  /* TTY (async) debug */
int pspdebug = 0;		  /* psp ( planar async ) */
int cndebug = 0;		  /* CONSOLE (mono/keybrd) debug */
int iodebug = 0x00;		  /* low level i/o */
int fsdebug = 0x00;		  /* Unix file system */
int swdebug = 0x00;		  /* swapping system */
int padebug = 0x00;		  /* paging system */
int sydebug = 0x00;		  /* syncronization */

int ecdebug = 0x00;
int hddebug = 0x00;
int aeddebug = 0x00;
int autodebug = 0x00;		  /* autoconfig debug */
static int z = 0x01010101;	  /* end of debugging switches */

#ifdef ROROOT
dev_t rorootdev = ROROOT;  /* if root make it read-only */
#endif
