#include "aha.h"
#if NAHA > 0
#define		DEBUG	1 /* HW */
#define		PARANOID	1 /* HW */
/*
 * Copyright (c) 1988, Christer Bernerus, Chalmers University of Technology,
 *                     Gothenburg, SWEDEN
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms as is or as a source
 * for derivative work are permitted provided that the above copyright 
 * notice and this paragraph are  duplicated in all such forms 
 * and that any documentation, advertising materials,
 * and other materials related to such distribution
 * and use acknowledge that the software was developed
 * at Chalmers University of Technology, SWEDEN.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Adaptec AHA-1540A SCSI adapter driver 
 * 
 * $Source: /afs/sys/rt-ibm4.3-r3/usr/sys/caio/aha.c,v $ 
 * $Revision: 2.4 $ $Date: 89/05/23 12:55:27 $ 
 * $Author: bernerus $ $State: Exp $
 *
 * $Log:	aha.c,v $
 * Revision 2.4  89/05/23  12:55:27  bernerus
 * Mostly a cleanup of the code. The copyright notice has changed.
 * A bug fix has been installed that fixed a bug in the device queuing
 * routines.
 * 
 * Revision 2.3  89/02/08  13:01:03  bernerus
 * Ahatur now does real things, operates as a device probe routine.
 * Ahacstart is now split into two routines: ahaqdevice and ahastartdevices
 * A new ahacstart is written that calls the two routines in sequence. However
 * the ahacstart call at the end of ahascanmbi was split into two parts so
 * that ahastartdevices is only called once per interrupt.
 * 
 * Revision 2.2  89/02/06  08:13:57  bernerus
 * Hords of changes, mostly due to restructuring of the source.
 * Copyright notice has changed a bit.
 * A new function ahadoscsi() permits scsi calls during autoconfig.
 * Adapter initialization on attach nor tries 10 times before giving up.
 * Ahatur() is now a dummy function, probing should be done by the device's
 * probe or attach routine by using SCSI inquiry thru the ahadoscsi() function.
 * More error messages now have adapter tag: "aha0: This message should never occur"
 * 
 * Revision 2.1  89/01/30  07:59:34  bernerus
 * Performed name changes of some variables.
 * 
 * Revision 2.0  89/01/16  10:50:40  bernerus
 * Changed Copyright notice
 * Changed names cu to cs, cs to csc, xu to xt, xt to xtc.
 * Reformatted function predeclarations.
 * Moved some variables into the ahaadm structure.
 * Removed display() calls and last CDF check in ahakick().
 * Renames ahaxstart() to ahaxsetup(), xi_xhanum to xi_ctlr, 
 * ahaclearmbo() to ahafreembo().
 * Ahatstart now reserves an mbo b4 building link chain.
 * Shut off MBO interrupt.
 * 
 */

/*********************** STATIC DATA SECTION **********************************/
/* #define BUFMODE */

#include "../h/param.h"
#include "../h/dk.h"
#include "../h/vm.h"
#include "../h/buf.h"
#include "../h/types.h"

#include "../machineio/ioccvar.h"
#include "../machineio/scsi.h"
#include "../machineio/xhavar.h"
#include "../machineio/ahareg.h"
#include "../machineio/dmavar.h"
#ifdef BUFMODE
#include "../machineio/dmareg.h"
#endif
#include "../machine/debug.h"

#include "cs.h"
#include "xt.h"

int aha_bus_on	=	AHABUS_ON ; /* HW for debugger use */
int aha_bus_off	=	AHABUS_OFF ;
int aha_speed	=	AHASPEED8 ;


#ifdef DEBUG
extern int      dmadebug;
int             ahadebug = 	 0;
int		ahaturdebug = 	 0;
int             ahanotice = 	 0;
int             ahacmddebug = 	 0;
int             ahakickdebug = 	 0;
int             ahaintdebug = 	 0;
int             ahascandebug = 	 0;
int             ahainitdebug = 	 0;
int             ahacstartdebug = 0;
int             ahatstartdebug = 0;
int             ahaxsetupdebug = 0;
int		aharetrydebug =  0;
int             ahaerrdebug = 	 0;
int             ahadmaerrdebug = 0;
int             ahausage = 	 0;
int             ahasleepdebug =  0;
int		ahambdebug = 	 0;
int		ahadmadebug =	 0;
int		aharequeuedebug = 0;
int		ahadoscsidebug = 0;
int		ahafilldebug = 0; /* HW to se dlen of actual transfers */
void		printmbox(), printccb();
void		uprintmbox(), uprintccb();
#endif
int		ahalinkflag	= 1;


typedef long	ioaddr_t;

#define AHAREG(y) ((struct aha_regs *) y->xi_xha )

struct iocc_ctlr *ahainfo[NAHA];

int 		ahaprobe(),	ahaint(),	ahaget(),	ahatur();
unsigned	ahaget3();
void 		ahadgo(),	ahaattach(),	ahacstart(),	asgn3(),
		ahafillmbo(),	ahafillccb(),	ahatout(),	ahainitmbox(),
		ahafreeccb(),	ahafreembi(),	ahafreembo(),	ahascreset(),
		ahacommand(),	ahakick(),	ahasetspeed(),	ahawritefifo(),
		ahaqdevice(),	ahastartdevices(),
		ahareadfifo(),	ahassto(),	ahasetmboint(),	ahasbontime();
#ifdef DEBUG
void		printccb(),	uprintccb(),	printcdb(),	printesense(),
		uprintcdb(),	uprintesense(),	printmbox(),	uprintmbox(),
		uxhdump(),	xhdump();
#endif DEBUG
struct ahaccb	*ahagetccb(),	*ahaincccbp(),	*ahasetupccb(),	*ahaxsetup(),
		*ahasccom();
struct ahambox *ahagetmbo(),	*ahagetmbi(), 	*ahaincmbop();
caddr_t         ahvaddr();
ioaddr_t	ahioaddr();
u_short		ahadoscsi();

#ifdef FBQ
int		ahatocheck = 0;		/* Disable timeout code */
#else
int		ahatocheck = 0;
#endif
/* Standard I/O addresses for the controller */
caddr_t         ahastd[] = 
{
    (caddr_t) 0xf0000330,
    (caddr_t) 0xf0000334,
    (caddr_t) 0xf0000230,
    (caddr_t) 0xf0000234,
    (caddr_t) 0xf0000130,
    (caddr_t) 0xf0000134,
    (caddr_t) 0
};


struct xhac_driver ahacdriver = 
{
                  ahatur,    	/* Test unit ready */
     (int (*) ()) ahadoscsi,  	/* Send SCSI command */
     (int (*) ()) ahacstart,	/* AHA unit start routine */
};

struct iocc_driver ahadriver = 
{
    			    ahaprobe,	/* probe     */
    			    0,		/* slave     */
    (int (*) ())	    ahaattach,	/* attach    */
    (int (*) ())	    ahadgo,	/* dgo       */
    			    ahastd,	/* addr      */
    (char *) 		    0,	    	/* dname     */
    (struct iocc_device **) 0,		/* dinfo     */
    			    "aha",	/* mname     */
    			    ahainfo,	/* minfo     */
    			    ahaint,	/* intr      */
    			    0x0,	/* csr       */
    			    0,		/* chanrelse */
    			    0		/* flags     */
};

/*
 * Calculation of necessary number of Mboxes. We assume that a tape needs only
 * one mbox at a time.
 */

#define NTAPES NCS
#define NDISKS NXT
#define NMBOX (NTAPES+NDISKS)*4
#if NMBOX > 255
#define NMBOX 255
#endif

#define NCCB	1*NMBOX 	/* 1 links per MBOX */

#define HZ 100
#define TOTIME 4*HZ

#ifdef DEBUG
int             aha_nmbox = NMBOX;	/* Info point for debugger */
int             aha_nccb = NCCB;	/* Info point for debugger */
#endif

/* The variables below should be moved into the ahaadm struct */
/* The variables above should be moved into the ahaadm struct */

/*
 * Administrative area for the controller. it is constantly mapped and never
 * released 
 */
struct ahaadm
{
    ioaddr_t        ioaddr;		/* I/O address of this member 	    */
    int		    move:1;		/* Interrupt has occurred	    */
    int             nokick:1;		/* Resetting aha, dont start it yet */
    int		    topending:1;	/* Timeout is pending 		    */
    int		    kicking:1;		/* Kick in progress		    */
    int		    skippedkick:1;	/* A kick was skipped last time	    */
    struct ahambox *freembo;		/* May be pointing to next free MBO */
    struct ahaccb  *freeccb;		/* May be pointing to next free CCB */
    int             mbousage;		/* Current MBO usage		    */	
    int             mbiusage;		/* Last MBI count within a mbi scan */
    int             ccbusage;		/* Current CCB usage		    */
    int             maxmbousage;	/* Maximum MBO usage		    */
    int             maxmbiusage;	/* Maximum MBI count		    */
    int             maxccbusage;	/* Maximum CCB usage		    */
    int		    maxbufq;		/* Longest buffer queue		    */
    int             cmdcmpl;    	/* Indicates HACC to ahacmd         */
#define CMPL_CC		1	    /* Command completion   	*/
#define	CMPL_INVD	-1	    /* INVDCMD			*/
#define	CMPL_NOTYET	0	    /* NOT COMPLETED YET 	*/
#define CMPL_ABORT	-2	    /* Command aborted 	 	*/
    u_char          instdev[8];		/* Vector showing installed devices */
    struct buf	    iobuf;		/* Dummy i/o buffer for DMA code    */
    struct ahambox  mbo[NMBOX];		/* Outgoing Mailboxes		    */
    struct ahambox  mbi[NMBOX];		/* Ingoing Mailboxes		    */
    struct ahaccb   ccb[NCCB];		/* Command Control blocks	    */
}               ahaadm[NAHA];


/*************** CONFIGURATION FUNCTIONS SECTION *********************/

ahaprobe(reg, ic)
register caddr_t reg;
register struct iocc_ctlr *ic;
{
    register struct aha_regs *addr = (struct aha_regs *) reg;
    extern int      int_level;

    /* First, issue a reset command that also clears the interrupt port */

    addr->aha_csr = AHCSR_HRST | AHCSR_SCRST | AHCSR_IRST | AHCSR_SRST;

    DELAY(10000);

    addr->aha_csr = AHCSR_IRST ;

    /*
     * Now, the command port should be empty. So we can issue a NOP command
     * that will give us an interrupt 
     */

    ahacmd(addr, AHACMD_NOP,ic->ic_ctlr);
    PROBE_DELAY(10000);
    if (int_level < 0)
    {
	printf("Ahaprobe: no interrupt from aha at %X! \n", addr);
	return PROBE_BAD;
    }


    if(aharesetha(addr))
    {
	return PROBE_BAD;
    }
    return PROBE_OK;
}

/*
/*************/

void
ahaattach(ic)
register struct iocc_ctlr *ic;
{
    int             		i;
    struct buf     	       *bp;
    int             		dmasetuprv;
    ioaddr_t        		ioaddr;
    register int		ctlr = ic->ic_ctlr;
    struct ahaadm	       *admp = &ahaadm[ic->ic_ctlr];
    int				initcnt = 0;

    /* 
     * Clear adm area, this will as a side effect mark all mailboxes as free 
     * The CCB's though, should be set free with special code.
     */
    bzero(admp, sizeof ahaadm[0]);

    while(ahainit(ic))
    {
	if(aharesetha(ic->ic_addr))
	{
		ic->ic_alive = 0;
		printf("aha%d: attach reset failed\n",ic->ic_ctlr);
		return;
	}
        if(++initcnt>10)
	{
		ic->ic_alive = 0;
		printf("aha%d: attach init failed\n",ic->ic_ctlr);
		return;
	}
    }


    /* Init next-free pointers */
    admp->freembo = admp->mbo;
    admp->freeccb = admp->ccb;
    admp->nokick = 1;  /* Don't start anything yet */
    for (i = 0; i < NMBOX; i++)
	admp->mbi[i].aha_ccbp = 1; /* An illegal value */
    for (i = 0; i < NCCB; i++)
	admp->ccb[i].aha_op = AHAOP_CCB_FREE;

    /* Configure DMA stuff */
    ic->ic_dmaflags = 
	(DMA_CASCADE | DMA_PAGE | DMA_PHYSICAL | DMA_EXCLUSIVE | DMA_CANTWAIT);
    ic->ic_dmabuf = 0;

    /* 
     * Allocate and set up an io buffer for mailbox usage 
     */
    bp = &(admp->iobuf);
    bp->b_flags |= B_LOCKED;			    /* We keep it forever     */
    bp->b_dev = (dev_t) NODEV;			    /* No device associated   */
    bp->b_error = 0;				    /* No errors yet 	      */
    bp->b_un.b_addr=(caddr_t) admp; 		    /* Data space address     */
    bp->b_bcount = sizeof(ahaadm[ctlr]);     	    /* Byte count	      */
    DEBUGF(ahadmadebug,dmadebug=1;);
    if ((dmasetuprv = dma_setup(ic)) != DMA_OK_RET)
    {
	DEBUGF(ahadmadebug,dmadebug=0;);
	printf("Ahaattach: DMA setup failed, rv=%d\n", dmasetuprv);
	return;
    }

#ifdef BUFMODE
    *(char volatile *)DM_BUFFER |= 0x80>>ic->ic_dmachannel; /* Buffer mode ON*/
#endif

    ioaddr = (ioaddr_t) dma_map(ic->ic_dmachannel, bp);
    if (ioaddr == (ioaddr_t) DMA_INV_IOADDR)
    {
	DEBUGF(ahadmadebug,dmadebug=0;);
	printf("Ahaattach: dma_map failed\n");
	return;
    }
    printf("Ahaattach: adm area at 0x%X (%X) %d bytes\n", admp, ioaddr,bp->b_bcount);
    admp->ioaddr = ioaddr;
    dma_go(ic->ic_dmachannel);
    DEBUGF(ahadmadebug,dmadebug=0;);
    ahainitmbox(NMBOX, (struct ahambox *) admp->mbo, ctlr); /* Init Mailboxes   */
    DELAY(100); /* HW ? */
    admp->nokick = 0;		    /* Allow kicks 	      */
}

ahatur(xi)
struct xha_device *xi;
{
    struct scsi_cdb6 turcmd;
    int	scts,rv;

    bzero(&turcmd,sizeof turcmd);

    turcmd.sc_op = SC0_TEST_UNIT_READY;
    
    DEBUGF(ahaturdebug,printf("aha%d: Ahatur ENTRY, target= %d\n",
	   xi->xi_ctlr,xi->xi_drive););

    scts = ahadoscsi(xi->xi_ctlr,xi->xi_drive,0,&turcmd,0,0, B_READ | B_WANTED);

    DEBUGF(ahaturdebug,printf("aha%d: Ahatur target %d, scts=%d\n",xi->xi_ctlr,xi->xi_drive,scts););

    rv = !(((scts >> 8) & 0xFF) == XHI_SELTO);

    DEBUGF(ahaturdebug,printf("aha%d: Ahatur returning %d\n",xi->xi_ctlr,rv););
    return rv;
}

/*
/*************** TOP I/O FUNCTIONS SECTION *********************/
/*
 * The Channel start routine. This one will enter the device xi into the
 * device queue and then start all devices on the corresponding channel.
 */
void 
ahacstart(struct xha_device *xi)
{
	ahaqdevice(xi);
	ahastartdevices(xi->xi_hd);
}

void
ahaqdevice(struct xha_device *xi)
{
    register struct xha_hd     *xhp = xi->xi_hd;
    register int                s;

    /*
     * Queue the device on the xha_hd structure for processing
     * Since this device could already be queued, we check the
     * queue first.
     */
    s = splbio();
    if(!xi->xi_inq)
    {
	DEBUGF(ahacstartdebug,
	    printf("Ahacstart: xi=%X entered\n",xi););
	xi->xi_forw = NULL;
	if (xhp->xh_actf == NULL)	/* I.e. first item in Q */
	{
	    xhp->xh_actf = xi;		/* Q header points to item */
	    xi->xi_back = NULL;		/* Back ptr indicates head of Q */
	}
	else				/* Not first item in q : */
	{
	    xhp->xh_actl->xi_forw = xi;	/* Fwd ptr of last elem points to new */
	    xi->xi_back = xhp-> xh_actl;/* Back ptr of item points to q tail */
	}
	xhp->xh_actl = xi;		/* Q tail is now the new item */
	xi->xi_inq = 1;			/* Device is now in q */
    }
    splx(s);
}


/*
 * Devices start routine, this one will start all devices on the device q.
 */
void
ahastartdevices(struct xha_hd *xhp)
{
    register struct xha_device *xip;
    register int                s;
    register int		loopcnt;

    /* Loop through the device queue and start all devices */
    s = splbio();
    /* 
     * We first check if there is anything to do or if someone else is
     * already working on it, if so we have nothing to do 
     */
    if(((xip = xhp->xh_actf) == NULL) || xhp->xh_active ) /* Anything to do ? */
    {
	splx(s);
        DEBUGF(ahaerrdebug,printf("Ahacstart: nothing to do:\n"););
	return;			 /* NO, skip it */
    }
    xhp->xh_active = 1;	/* Note to others that I'm working on it */
    splx(s);

    DEBUGF(ahacstartdebug,printf("Ahacstart: starting targets: xip=%X\n",xip););
    /* Now start working through the targets */
    for(loopcnt=0;loopcnt<100;loopcnt++)
    {
	switch(ahatstart(xip))	/* Start all requests for the target */
	{
	case AHS_DONE:		/* I.E. all requests started OK */
	    /*
	     * Here we know that all outstanding requests have
	     * been started so the device can be unqueued from the
	     * device queue
	     */
	    s=splbio();
	    if(! xip->xi_tab.b_actf)	/* Nothing there now ? */
	    {
		/* Unqueue the device */
		if(xip->xi_back)/* dependin on whether wer'e the first element*/
		{
		   xip->xi_back->xi_forw =  /* Previous elements fwd ptr      */
		     xip->xi_forw;	    /* inherits my old fwd pointer    */
		}
		else			    /* or */
		{
		   xhp->xh_actf =	    /* the Q header pointer 	      */
		     xip->xi_forw;	    /* inherits my old fwd pointer    */
		}

		if(xip->xi_forw)/* depending on whether wer'e the last element*/
		{
		   xip->xi_forw->xi_back =  /* Next elem's back pointer or    */
		     xip->xi_back;	    /* inherits my old back pointer   */
		}
		else			    /* or */
		{
		   xhp->xh_actl =	    /* the Q tail pointer 	      */
		     xip->xi_back;	    /* inherits my old back pointer   */
		}

		xip->xi_inq = 0;	    /* Now wer'e no longer in the Q   */
	    }

	    xip = xip->xi_forw;		    /* Pick the next target */

	    if(xip)			    /* Was there anything ? */
	    {
		splx(s);		    /* Yes, reset priority and */	
		continue;	       	    /* process it	       */
	    }
	    
	    /* Nothing left in the Q	*/

	    xhp->xh_active = 0; /* Note to others that we aren't processing
				 * this channel anymore 
				 */
	    splx(s);
	    return;
	case AHS_RETRY:		/* Too bad, hope they will be started later */
	case AHS_FAIL:
	    /* In the RETRY case we know that a request is outstanding so 
	     * we do not unqueue the device
	     * In the FAIL case the request is removed from the request
	     * queue, but there might be other requests further down the
	     * queue so we can't unqueue the device.
	     * In the SKIP case, we have to clear the active bit, or 
	     * otherwise, the driver might hang forever.
	     */
	    xhp->xh_active = 0; /* Note to others that we aren't processing
				 * this channel anymore 
				 */
	    return;
	}
    }
    printf("%s%d: ahastartdevices: Max loop count exceeded:xip=%X, forw=%X\n",
	   xip->xi_driver->xd_dname,xip->xi_drive,
		xip,xip->xi_forw);
    xhp->xh_active = 0;
}
/*
/* 
 * Target start routine, this will attempt to start the available requests
 * for the target indicated by xi
 */
ahatstart(xi)
register struct xha_device *xi;
{
    register struct buf	*bp;
    register int	 s;
    register int	 bufq=0;
    struct ahaccb	*prevccbp = 0;
    struct ahaccb	*firstccbp = 0;
    struct ahambox	*mbop = 0;

    /* 
     * We first check if there is anything to do or if someone else is
     * already working on it, if so we have nothing to do
     */
    s = splbio();
    if((bp = xi->xi_tab.b_actf) == NULL) /* Anything in Q ? */
    {
	splx(s);			 /* NO, bye */
	return AHS_DONE;
    } 
    if(xi->xi_tab.b_active )		/* Somebody already doing this ? */
    {					
	splx(s);			/* Yes, bye */
	return AHS_RETRY;
    }
    /* Yes, there is something to do */
    xi->xi_tab.b_active = 1;    /* Note to others that I'm working on it */
    splx(s);

    DEBUGF(ahatstartdebug,printf("Ahatstart: starting target %X\n",xi););

    for(;;)
    {
	switch ((*xi->xi_driver->xd_ustart) (xi))
	{
	case XHU_NEXT:
	    s = splbio();
	    if (bp->b_flags & B_DONE)
	    {
		printf("DUP BIODONE NR 1T\n");
	    }
	    else
	    {
		iodone(bp);
	    }
	    if((xi->xi_tab.b_actf = bp->av_forw) != 0) /* Remove it from Q */
	    {
		bp = xi->xi_tab.b_actf;
		splx(s);
		continue;
	     }
	     xi->xi_tab.b_active = 0;
	     splx(s);
	     return AHS_DONE;
	case XHU_DODATA:
	    DEBUGF(ahatstartdebug,printf("Ahatstart: starting xfer %X\n",bp););
	    if((ahaxsetup(xi,bp,0,AHX_START)) == 0)
	    {
		xi->xi_tab.b_active = 0;
		return AHS_RETRY;
            }
	    else
	    {
		/* Flush the request from the queue, if it fails and should be
		 * retried, it will be requeued by the interrupt routine */
		xi->xi_tab.b_actf = bp->av_forw;
		return AHS_DONE;
	    }
	case XHU_DOLINKS:
	    DEBUGF(ahatstartdebug,printf("Ahatstart: starting link %X\n",bp););
	    /*
	     * First, reserve an MBO so that we're sure to have one for the
	     * chain since backing off a chain is difficult
	     */
	    if(mbop == 0)
	    {
		if((mbop = ahagetmbo(xi->xi_ctlr)) == 0)
		{
		    xi->xi_tab.b_active = 0;
		    return AHS_RETRY;
		}
	    }
	    /* 
             * Here we set up CCB's as long as possible
	     */
	    if((prevccbp = ahaxsetup(xi,bp,prevccbp,AHX_LINK)) == 0)
	    {
		/* 
		 * We get here if ahaxsetup fails. If we were unable to
		 * set up anything, we'll have to defer all actions to a
		 * later time
		 */
		if(firstccbp == 0) 
		{
		    /* 
		     * If no chain header could be set up, we defer action to
		     * a later time 
		     */
		    DEBUGF(aharetrydebug,ahaintdebug=1;);
		    /*
		     * If we've reserved an MBO, we must free it here
		     */
		    if(mbop) 
		    {
			bzero(mbop,sizeof *mbop);
		    }
		    xi->xi_tab.b_active = 0;
		    return AHS_RETRY;
		}
		/* 
		 * Else CCB's were set up.
		 * Start the chain built so far, return that the Q isn't empty 
		 */
		ahastartmbo(xi->xi_ctlr,firstccbp,mbop);
		return AHS_RETRY;
            }
	    else /* I.e. No problems setting up CCB */
	    {
		if(firstccbp == 0) 
		{
			firstccbp = prevccbp;
		}
		/* 
		 * Shut off interrupts here so that a request can't be entered
		 * without getting processed, only requeued requests could
		 * be entered while running here as a process, these requests
		 * are always entered at the end.
		 */
		DEBUGF(ahatstartdebug,printf("Ahatstart: xfer %X DONE\n",bp););
		/* Get new bp, unqueue old one and see if we hit the end */
		s = splbio();
		if((bp = xi->xi_tab.b_actf = bp->av_forw) != 0)
		{
		    splx(s);
		    if(++bufq > ahaadm[xi->xi_ctlr].maxbufq)
		    {
			DEBUGF(ahanotice,printf("Max Link length = %d\n",bufq););
			ahaadm[xi->xi_ctlr].maxbufq = bufq;
		    }
		    continue;
		}
		splx(s);
		ahastartmbo(xi->xi_ctlr,firstccbp,mbop);
		return AHS_DONE;
	    }
	}
    }
}
/*	
/* This routine sets up ccb's for a xfer, if action is AHX_START, the xfer
 * is started. If prevccbp is nonzero, a link will be set up to prevccbp.
 * return value is ccb set up or 0 if xfer couldn't be started */
struct ahaccb *
ahaxsetup(xi,bp,prevccbp,action)
register struct xha_device      *xi;
register struct buf *bp;
register struct ahaccb *prevccbp;
{
    register struct iocc_ctlr   *ic = xi->xi_hd->xh_iocc;
    register struct scsi_cdb6   *com;
    register ioaddr_t            ioaddr;
    struct scsi_cdb12		 combuf;
    register struct ahaccb 	*ccbp;

    if (xi->xi_driver->xd_start == (int (*) ()) 0 ||
        (com = (struct scsi_cdb6 *) (*xi->xi_driver->xd_start) (xi,bp,&combuf)) == 0)
    {
        printf("%s%d: SCSI command pointer zero or no driver start routine\n",
                xi->xi_driver->xd_dname, XHAUNIT(bp->b_dev));
        return 0;

    }
    /*
     * This is the place where things really happen
     * First we set up the DMA mapping, next we get a CCB and fill it in.
     * Then we get a free MBO, fills it with the address of the CCB, and
     * lastly we kick the host adapter telling it that there is work to do.
     */
    DEBUGF(ahaxsetupdebug,printf("Ahaxstart: %s%d, bp=%X, com=%X,prio=%d\n",
	xi->xi_driver->xd_dname,XHAUNIT(bp->b_dev),bp,com,getspl()););
    
    ioaddr = (ioaddr_t) dma_map(ic->ic_dmachannel, bp); /* Set up dma mapping */
    if (ioaddr == (ioaddr_t) DMA_INV_IOADDR)
    {
	DEBUGF(ahadmaerrdebug,printf("%s%d: dma_map failed\n",
	    xi->xi_driver->xd_dname, XHAUNIT(bp->b_dev)););
            return 0;
    }
    /* If command should be started immediately, send the SCSI command */
    if(action == AHX_START)
    {
	if((ccbp = ahasccom(xi->xi_ctlr, xi->xi_drive, XHAMIN_LUN(bp->b_dev), 
			    com, ioaddr, xi, bp)) == 0)
	{
	    DEBUGF(aharetrydebug,printf("Ahaxstart: bp=%X RETRY\n",bp););
	    dma_free_map(ic->ic_dmachannel, ioaddr,bp->b_bcount);
	    return 0;
	}
	else
	    return ccbp;
    }
    /* Set up ccb and links */
    ccbp = ahasetupccb(xi->xi_ctlr, xi->xi_drive, XHAMIN_LUN(bp->b_dev),
		       com, ioaddr, xi, bp);
    if(ccbp == 0)
    {
	    dma_free_map(ic->ic_dmachannel, ioaddr,bp->b_bcount);
	    return 0;
    }
    if(prevccbp != 0)
    {
	ahalinkccb(xi->xi_ctlr,prevccbp,ccbp);
    }
    return ccbp;
}

aharequeue(xi,bp)
struct xha_device *xi;
struct buf *bp;
{
	struct buf *dp,*tp;
	int	s;
	int	loopcnt;

	s = splbio();
	DEBUGF(aharequeuedebug,printf("Aharequeue: xi=%X, bp=%X",xi, bp););
	tp = &xi->xi_tab;
	if(!bp)
	{
		DEBUGF(aharequeuedebug,printf(" cannot be entered.\n"););
		splx(s);
		return 1;
	}
	if (tp->b_actf == NULL)
	{
	    DEBUGF(aharequeuedebug,printf(", entered into empty Q\n"););
	    bp->av_forw = NULL;
	    tp->b_actf = bp;
	}
	else
	{
	    for(dp=tp,loopcnt=0;dp;dp = dp->av_forw)
	    {
	        if(++loopcnt > 1000000)
		{
			splx(s);
			printf("%s%d: aharequeue: max loop count exceeded\n",
			    xi->xi_driver->xd_dname,XHAUNIT(bp->b_dev));
			return 1;
		}
				
		if(dp == bp)
		{
			DEBUGF(aharequeuedebug,
				printf("Aharequeue: Buffer not reentered\n"););
			splx(s);
			return 1;
		}
 	    }
	    bp->av_forw = NULL;
	    DEBUGF(aharequeuedebug,printf(" entered into Q after %X\n",
		   tp->b_actl););
	    tp->b_actl->av_forw = bp;
	}
	tp->b_actl = bp;
	splx(s);
	return 0;
}
/*
/*************** INTERRUPT FUNCTIONS SECTION *********************/
/*
 * Interrupt routine, most of it's work is done in
 * the ahascanmbi() function which takes care of all MBI's
 */
ahaint(ctlr)
int             ctlr;
{
    register struct iocc_ctlr *ic = ahainfo[ctlr];
    u_char          iflgs,
                    csr;
    struct ahaadm  *admp = &ahaadm[ctlr];
    register struct aha_regs *addr = (struct aha_regs *) ic->ic_addr;

    csr   = addr->aha_csr;			/* Read status register 	  */
    iflgs = addr->aha_iflgs;			/* Read the interrupt port  	  */
    if(!(iflgs & AHIFLGS_INT))			/* If no interrupt, skip the rest */
    {
	DEBUGF(ahanotice,printf("Ahaint, stray interrupt ... iflgs=%b\n",
		iflgs,AHIFLGS_BITS););
	return 1;
    }

    addr->aha_csr = AHCSR_IRST;	 	/* Clear the interrupt port */
    admp->move = 1;		/* Note to ahatout that int occurred */

    if ((csr & AHCSR_INIT) && (csr & AHCSR_INVDCMD)) /* Need mailbox init */
    {
	ahainitmbox(NMBOX, (struct ahambox *) admp->mbo,ctlr);
	DEBUGF(ahanotice,printf("Ahaint: reinitialized mboxes\n"););
	DELAY(1000);
    }
    if (iflgs & AHIFLGS_HACC)		/* If command complete, set indicator */
    {
	admp->cmdcmpl = CMPL_CC;
    }

    DEBUGF(ahaintdebug, printf("Ahaint: csr=%b, iflgs=%b, prio=%d\n",
		                csr, AHCSR_BITS, iflgs, AHIFLGS_BITS,getspl()););

    if ((iflgs & AHIFLGS_MBIF) || (iflgs & AHIFLGS_SCRD))    /* In mailbox filled */
    {
	admp->mbiusage=ahascanmbi(ctlr,iflgs & AHIFLGS_SCRD);
	if(admp->mbiusage > admp->maxmbiusage)
	{
	    admp->maxmbiusage = admp->mbiusage;
	    DEBUGF(ahausage, printf("Max MBI usage = %d\n", admp->maxmbiusage););
	}
    }

    if(!(csr & AHCSR_INIT) && (!admp->nokick))
    {
	ahakick(ctlr);
    }
    return 0;
}

/*
 * Scan all MBI's and take appropriate actions
 */
ahascanmbi(ctlr,scsireset)
{
    /* The variables below are  ordered in # of occurrances in the routine    */
    register struct xha_device *xi;		    /* Ptr to device 	      */
    register struct ahaccb     *ccbp;		    /* Ptr to CCB involved    */
    register struct buf        *bp;		    /* Ptr to buffer involved */
    register struct ahambox    *mp;		    /* Ptr to mbox in question*/
    register u_long             ioccbp;		    /* CCB part of MBI	      */
    register int    		dccode;    /* Code sent to driver int routine */
    register u_char 		csb;		    /* Status byte of MBI     */
    	     struct iocc_ctlr   *ic = ahainfo[ctlr];
    	     struct scsi_esense	*sdp;		    /* Ptr to sense info      */
    	     u_char 		tarstat;	    /* Target status          */
    	     u_char 		hastat;		    /* Host adapter status    */
    	     int    		nmbi = 0;	    /* Count used mbi's	      */
	     int		mbirescan = 1;	    /* Times to go thru MBI's */
	     struct ahaadm     *admp = &ahaadm[ctlr];
	     struct xha_hd     *xhp = 0;	    /* Ptr to xha_hd struct   */

    /* Scan the IN mailboxes and see what should be done to them */

    DEBUGF(ahascandebug, printf("Ahaint: Scanning in mailboxes\n"););
    while(mbirescan--)
    {
	for(mp = admp->mbi; 		/* Scan all MBI's, free them on */
	    mp < &admp->mbi[NMBOX];	/* the end of the loop		*/
		mp++)
	{
		/* Don't change the order of the following three statements */
	    if((csb = mp->aha_csb) == MBI_MAILBOX_FREE )
	    {
		continue;
	    }
	    ioccbp = mp->aha_ccbp;

	    /*
	     * Have a look at the status byte
	     */

	    if((dccode = ahacsbtodccode(csb,mp,ctlr)) == XHI_NEXT)
	    {
		continue;
	    }
#ifdef PARANOID
	    if(ioccbp & 0x01)	/* Mailbox being filled ?	   */
	    {
		    printf("Ahascanmbi: partially filled MBI, enqueued\n");
		    mbirescan = 1;
		    continue;
	    }
#endif
	    nmbi++;
	    /*
	     * Zero out corresponding MBO so that sane things could be
	     * done if the adapter hangs up, I.E. we here mark that the
	     * corresponding MBO isn't active anymore within the host adapter.
	     */
	    ahafreembo(ctlr, ioccbp);

#ifdef PARANOID
	    /*
	     * Guard against invalid ccb pointers
	     */
	    if (ioccbp < 0xfe0000)
	    {
		printf("Ahascanmbi: Strange! ccb pointer = %X\n", ioccbp);
		DELAY(10000000);
		ahafreembi(mp);
		continue;
	    }
#endif PARANOID

	    /* 
	     * Pick up some facts from the CCB now
	     */
	    ccbp = (struct ahaccb *)
		    ahvaddr(ioccbp, ctlr); /* CCB pointer 		      */
#ifdef PARANOID
	    if((ccbp < admp->ccb) || (ccbp >= &admp->ccb[NCCB]))
	    {
		    printf("Invalid CCB pointer = %X ioccbp=%X\n",ccbp, ioccbp);
		    DELAY(1000000);
		    ahafreembi(mp);
		    continue;
	    }
	    if((((int)ccbp - (int)admp->ccb) % 
		((int)&admp->ccb[1] - (int) &admp->ccb[0])) != 0)
	    {
		    printf("CCB pointer points bn ccb's ccbp = %X ioccbp=%X\n",ccbp, ioccbp);
		    DELAY(1000000);
		    ahafreembi(mp);
		    continue;
	    }
#endif
	    DEBUGF(ahascandebug,printf("Ahascanmbi: ccbp=%X\n",ccbp););
	    DEBUGF(ahascandebug,printccb(ccbp););
	    sdp  = (struct scsi_esense *)		/* Sense info 	      */
		(((char *) &ccbp->aha_cdb) + ccbp->aha_cmdlen);
	    tarstat = ccbp->aha_tarstat;	        /* Target status      */
	    hastat  = ccbp->aha_hastat;			/* Host status	      */
	    xi      = ccbp->aha_xi;			/* Device pointer     */
	    bp      = ccbp->aha_bp;			/* Buffer pointer     */

	    /* 
	     * The if stmt below is a guard against zero pointers
	     */
	    if (xi && xi->xi_driver && xi->xi_driver->xd_int)
	    {

#ifdef DEBUG
		if (csb != MBI_NOERROR)
		{
		    DEBUGF(ahaerrdebug, printf("Ahaint: MBI ERROR csb=%d\n",csb););
		    DEBUGF(ahaerrdebug, printmbox(mp,1,1););
		    DEBUGF(ahaerrdebug, printccb(ccbp););
		    DEBUGF(ahaerrdebug, DELAY(1000000););
		}
#endif
		/*
		 * Check Host status and scsireset to set dccode accordingly
		 */
		dccode = ahastattodccode(dccode, hastat, scsireset);

		if(bp)
		{
		    /* Let the target driver's int. routine decide what to do */
		    switch ((*xi->xi_driver->xd_int) (xi, bp, dccode, sdp, tarstat))
		    {
		    case XHD_RESETRETRY:
			DEBUGF(ahanotice, 
			    printf("Ahaint: int routine returned RESETRETRY\n"););
			ahasccom(ctlr, xi->xi_drive, /* Send BDR */
			         XHAMIN_LUN(bp->b_dev), NULL, 0, 0, 0, 0); 
			aharequeue(xi,bp);
			break;
		    case XHD_SCRESETRETRY:
			DEBUGF(ahanotice,
			    printf("Ahaint: int routine returned SCRESETRETRY\n"););
			ahascreset(xi->xi_xha);
			aharequeue(xi,bp);
			break;
		    case XHD_DONE:
			DEBUGF(ahascandebug, 
				printf("Ahaint: int routine returned DONE\n"););

			xi->xi_tab.b_errcnt = 0;

#ifdef PARANOID
			if (bp->b_flags & B_DONE)
			{
			    printf("DUP BIODONE nr 3: bp=%X\n", bp);
			    /* IN THIS CASE THERE ARE PROBABLY NO
			       TCW'S MAPPED SO WE DISABLE THE FREEING BELOW
			       BY CLEARING THE dptr VARIABLE */
			 
			    asgn3(ccbp->aha_dptr, 0);
			}
			else
			{
#endif
			    iodone(bp);
			    DEBUGF(ahascandebug,
				printf("IO done on buffer 0x%X\n",bp););
#ifdef PARANOID
			}
#endif
			break;

		    case XHD_RETRY:
			DEBUGF(ahaerrdebug, 
				printf("Ahaint: int routine returned RETRY\n"););
			aharequeue(xi,bp);
			break;
		    default:
			panic("ahaint: device int. routine return code error");
		    } /* End switch(xi-int()) */
		} /* End if(bp) */
	    } /* End if(xi ... ) */

	    /*
	     * Free the DMA maps
	     */
	    if (ahaget3(ccbp->aha_dptr))
	    {
		dma_free_map(ic->ic_dmachannel, ahaget3(ccbp->aha_dptr),
			     ccbp->aha_dlen);
	    }
	    else
	    {
		DEBUGF(ahaerrdebug,
		    printf("Ahaint: No data pointer, ccbp=%X\n",ccbp););
	    }
	    ahafreeccb(ctlr, ccbp); /* Release the ccb */
	    ahafreembi(mp);	    /* Free the MBI    */
	    /* 
	     * If not an intermediate interrupt, we 
	     * mark the device inactive and call the channel start routine 
	     */

	    if (xi && xi->xi_driver)
	    {
		if((hastat != AHAST_LKCPL) && (hastat != AHAST_LKCPL_INT))
		{
		    xi->xi_tab.b_active = 0; /* We ain't active anymore */
		    ahaqdevice(xi);	/* Q device for later start */
		    xhp = xi->xi_hd;	/* Remember ptr to xha_hd */
		}
	    } /* End if(xi ... ) */
	} /* End for() ... */
    } /* End while(mbirescan) */
    if(xhp)
    {
	ahastartdevices(xhp);		/* Start all queued devices */
    }
    return nmbi;
}

ahacsbtodccode(csb,mp,ctlr)
struct ahambox	*mp;
{
	switch (csb)
	{
	case MBI_HOST_TIMEOUT:
	    printf("Ahaint:  CCB timeout aborted: %X\n", mp->aha_ccbp);
	    return XHI_TIMEOUT;
	case MBI_HOST_ABORTED:
	    printf("Ahaint:  CCB aborted: %X\n", mp->aha_ccbp);
	    return XHI_ERR;
	case MBI_NO_CCB:
	    printf("Ahaint:  CCB not found: %X\n", ahvaddr(mp->aha_ccbp, ctlr));
	    ahafreembo(ctlr, mp->aha_ccbp);
	    ahafreembi(mp);
	    return XHI_NEXT;
	case MBI_MAILBOX_FREE:
	    return XHI_NEXT;
	case MBI_NOERROR:
	    DEBUGF(ahascandebug, printf("Ahaint: In mailbox:\n"););
	    DEBUGF(ahascandebug, printmbox(mp, 1, 1););
	    return XHI_CMPL;
	case MBI_ERROR:
	    DEBUGF(ahascandebug, printf("Ahaint: In mailbox:\n"););
	    DEBUGF(ahascandebug, printmbox(mp, 1, 1););
            return XHI_ERR;
	default:

	    printf("ahaint: invalid mailbox-in status byte code %d at %X\n",
		   csb, mp);
	    return XHI_ERR;
	}
}

/*
/* Decode host adapter status to a device completion code */
ahastattodccode(dccode,hastat,scsireset)
{
    if(scsireset)
    {
	return XHI_SCSIRESET;
    }
    switch(hastat)
    {
    case AHAST_NOERROR:		/* No error */
    case AHAST_LKCPL:		/* Linked cmd cmpl, no error */
    case AHAST_LKCPL_INT:	/* Linked cmd cmpl, no error, intrpt */
	return dccode;
    case AHAST_SELTO:		/* Selection Timeout */
	return XHI_SELTO;
    case AHAST_OUVR:		/* Data over/under run */
	return XHI_OUVR;
    case AHAST_BUSFREE:		/* Unexpected bus free */
    case AHAST_TGT_FAIL:	/* Target bus phase sequence failure */
	return XHI_BUSFAIL;
    case AHAST_MBO_CMD:		/* MBO command was invalid */
    case AHAST_CCB_OP:		/* CCB opcode was invalid */
    case AHAST_INVD_LINK:	/* Linked CCB have uneq status bytes */
	return XHI_HASWFAIL;
    }
}

/*
/************** BOTTOM I/O DEVICE DRIVER ROUTINES SECTION ********************/

ahalinkccb(ctlr,fromccb,toccb)
struct ahaccb  *fromccb;
struct ahaccb  *toccb;
{
	sccb	*ccp;
	
	ccp = (sccb *)(((char *)&(fromccb->aha_cdb))+(fromccb->aha_cmdlen - 1));
	
	ccp -> sc_link = 1;
	ccp -> sc_flag = ahalinkflag;

	asgn3(fromccb->aha_lptr,ahioaddr(toccb,ctlr));

	fromccb->aha_fwd = toccb;
	toccb->aha_bwd = fromccb;
}


u_short
ahadoscsi(ctlr, target, lun, command, addr, dlen, flags)
caddr_t	addr;
struct scsi_cdb6  *command;
{
    struct buf *bp,buf;
    ioaddr_t	ioaddr = 0;
    volatile	struct 	ahaccb *ccbp; /* HW at least one of the fields it is 
					pointing at is volatile */
    int rv;

    DEBUGF(ahadoscsidebug,printf("aha%d:Ahadoscsi ENTRY: target=%d, lun=%d, addr=%X, dlen=%d\n",
		ctlr,target,lun,addr,dlen););
    bp = &buf;

    bp->b_un.b_addr = addr;
    bp->b_bcount = dlen;
    bp->b_flags = flags | B_BUSY;
	
    if(bp->b_flags & B_READ)
    {
	bzero(bp->b_un.b_addr,bp->b_bcount);
    }
    
    if(dlen != 0)
    {
	ioaddr = dma_map(ahainfo[ctlr]->ic_dmachannel,bp);
	if (ioaddr == (ioaddr_t) DMA_INV_IOADDR)
	{
	    printf("aha%d:Ahadoscsi: target=%d, lun=%d. Dma_map failed\n",
		ctlr,target,lun);
	    return SCTS_DRV_FAIL;
	}
    }

    if((ccbp = ahasccom(ctlr, target, lun, command, ioaddr, 0, bp)) == 0)
    {
	if(dlen != 0)
	{
	    dma_free_map(ahainfo[ctlr]->ic_dmachannel,ioaddr,bp->b_bcount);
	}
	DEBUGF(ahadoscsidebug,printf("aha%d:Ahadoscsi: target=%d, lun=%d. Ahasccom FAILED\n",
		ctlr,target,lun););
	return SCTS_DRV_FAIL;
    }

    if(flags & B_WANTED)
    {
	DEBUGF(ahadoscsidebug,printf("aha%d:ahadoscsi: waiting for completion\n",
		ctlr););

	while(ccbp->aha_op != AHAOP_CCB_FREE)
	{
            DEBUGF(ahadoscsidebug,printf("aha%d:ahadoscsi: op=%d,hastat=%d\n",
		ctlr,ccbp->aha_op,ccbp->aha_hastat););
		/* DELAY(10);  HW */
	} /* HW hc might remove the entire loop ! */

	/*DELAY(1);  HW 100 ? */
	DEBUGF(ahadoscsidebug,printf("aha%d:ahadoscsi: completed: op=%d,hastat=%d\n",
	    ctlr,ccbp->aha_op,ccbp->aha_hastat););
    }
    rv = (ccbp->aha_tarstat + 
	    (ahastattodccode(XHI_CMPL, ccbp->aha_hastat, 0) << 8));

    DEBUGF(ahadoscsidebug,printf("aha%d: ahadoscsi: tarstat=%d, hastat=%d, rv=%d\n",
		ctlr,ccbp->aha_tarstat,ccbp->aha_hastat,rv););
#ifdef DEBUG
    if(ahadoscsidebug)
    {
	if(ccbp->aha_tarstat == SCTS_CHECK_CONDITION)
	{
		printccb(ccbp);
	}
    }
#endif DEBUG
    return rv;
}

struct ahaccb *
ahasccom(ctlr, target, lun, com, ioaddr, xi, bp)    /* Send SCSI command  */
struct xha_device *xi;
struct scsi_cdb6  *com;
ioaddr_t	   ioaddr;
struct buf	  *bp;
{
    struct ahaccb  *ccbp;

    if((ccbp = ahasetupccb(ctlr, target, lun, com, ioaddr, xi, bp)) == 0)
    {
	return 0;
    }

    if(ahastartmbo(ctlr,ccbp,0) == AHS_RETRY)
    {
	ahafreeccb(ctlr,ccbp);
	return 0;
    }

    return ccbp;
}

/*
/* Set up a CCB */
struct ahaccb *
ahasetupccb(ctlr, target, lun, com, ioaddr, xi, bp)
struct xha_device *xi;
struct scsi_cdb6  *com;
ioaddr_t	   ioaddr;
struct buf	  *bp;
{
    struct ahaccb  *ccbp;

    if((ccbp = ahagetccb(ctlr)) == 0)
    {
	return 0;
    }

    if (com)
    {
	ahafillccb(ccbp, com, ioaddr, bp ? bp->b_bcount:0, target, lun, xi, bp);
    }
    else
    {
	/* We assume a bus device reset */
	ccbp->aha_op = AHAOP_BDR;
	ccbp->aha_id = target;
	ccbp->aha_lun = lun;
    }
    return ccbp;
}

ahastartmbo(ctlr, ccbp, mbop)
struct ahaccb  *ccbp;
    register struct ahambox *mbop;
{
    register struct ahaadm  *admp = &ahaadm[ctlr];
    register struct ahaccb  *clccbp;
    register int    clid;
    register int    loopcnt;

    if(mbop == 0)
    {
	if((mbop = ahagetmbo(ctlr)) == 0)
	{
	    return AHS_RETRY;
	}
    }

    clid = mbop - admp->mbo; /* Get MBO index */
	
    /* Now fill in command link ID's of all CCB's */
    for(clccbp = ccbp, loopcnt=0; clccbp; loopcnt++ )
    {
	if(loopcnt > 1000000)
	{
		printf("Ahastartmbo: loop count exceeded\n");
		break;
	}
	clccbp->aha_clid = clid;
	if((clccbp = (struct ahaccb *) ahaget3(clccbp->aha_lptr)) == 0)
	{
	    break;
	}
	clccbp = (struct ahaccb *) ahvaddr((ioaddr_t) clccbp, ctlr);
    }


    ahafillmbo(ctlr, mbop, ccbp);	/* Fill the MBO and	*/

    if (!admp->nokick) 
    {
	ahakick(ctlr);			/* start the controller	*/
    }

    if(admp->topending == 0)		/* Start a timeout 	*/
    {
	admp->move = 0;
	timeout(ahatout,ctlr,TOTIME);
	admp->topending = 1;
    }
    return AHS_DONE;
}

void
ahafillmbo(ctlr, mbop, ccbp)
struct ahambox *mbop;
struct ahaccb  *ccbp;
{

    struct ahaadm *admp = &ahaadm[ctlr];

    /* Things MUST happen in the order below */
    mbop->aha_ccbp = ahioaddr((caddr_t) ccbp, ctlr);
    admp->mbousage++;
    if (admp->mbousage > admp->maxmbousage)
    {
	admp->maxmbousage = admp->mbousage;
	DEBUGF(ahausage, printf("Max MBO usage = %d\n", admp->maxmbousage););
    }
    mbop->aha_csb = MBO_START_SCSI;
}



void
ahafillccb(ccbp, com, ioaddr, dlen, target, lun, xi, bp)
struct xha_device *xi;
struct ahaccb  *ccbp;
struct scsi_cdb6 *com;
ioaddr_t        ioaddr;
struct buf     *bp;
{
    ccbp->aha_op = AHAOP_SCSI;
    ccbp->aha_id = target;
    switch (com->sc_op)
    {
    case SC0_READ:
    case SC0_MODE_SENSE:
    case SC0_REQUEST_SENSE:
    case SC0_READ_REVERSE:
    case SC1_READ:
	ccbp->aha_din = 1;
	ccbp->aha_dout = 0;
	break;
    case SC0_WRITE:
    case SC0_MODE_SELECT:
    case SC1_WRITE:
	ccbp->aha_dout = 1;
	ccbp->aha_din = 0;
	break;
    default:
	ccbp->aha_dout = 0;
	ccbp->aha_din = 0;
    }
    ccbp->aha_lun = lun;
    ccbp->aha_dlen = dlen;
    asgn3(ccbp->aha_dptr, ioaddr);
    asgn3(ccbp->aha_lptr, 0);
    ccbp->aha_clid = 0;
    ccbp->aha_hastat = AHAST_STARTED;
    ccbp->aha_tarstat = 0;
    switch (com->sc_op >> 5 & 7)
    {
    case 0:
	ccbp->aha_cmdlen = 6;
	break;
    case 1:
	ccbp->aha_cmdlen = 10;
	break;
    case 5:
	ccbp->aha_cmdlen = 12;
	break;
    default:
	panic("ahafillccb: invalid SCSI op code\n");
	break;
    }
    ccbp->aha_senselen = AHA_SENSELEN;
    bcopy((char *) com, (char *) &ccbp->aha_cdb, ccbp->aha_cmdlen);
    ccbp->aha_xi = xi;
    ccbp->aha_bp = bp;
    if(xi && xi->xi_driver->xd_ttl != (int (*) () ) 0 )
    {
	ccbp->aha_ttl = (*xi->xi_driver->xd_ttl)(com);
    }
    DEBUGF(ahafilldebug,printf("Ahafill: ccbp=%X\n",ccbp););
    DEBUGF(ahafilldebug,printccb(ccbp););

}
/*
/*
 * Find a free MBI for simulated usage.
 */
struct ahambox *
ahagetmbi(ctlr)
{
    struct ahambox *mbip;

    for (mbip = ahaadm[ctlr].mbi; mbip->aha_csb != MBI_MAILBOX_FREE; mbip++)
    {
	if (mbip >= &ahaadm[ctlr].mbi[NMBOX])
	{
	    return 0;
	}
    }
    return mbip;
}

/*
 * Round robin increment of a pointer to a CCB
 * *wrap is increased if wrapping occurred
 */
struct ahaccb *
ahaincccbp(admp,ccbp,wrap) 
struct ahaccb *ccbp;
struct ahaadm *admp;
int *wrap;
{
	if(++ccbp >= &admp->ccb[NCCB]) /* Hit the edge yet ? */
	{
		(*wrap)++; /* Tell caller that we wrapped around the edge */
		return admp->ccb; /* Return start of array */
	}
	return ccbp;	/* Just return next one */
}

/*
 * Round robin increment of a pointer to a MBO
 * *wrap is increased if wrapping occurred
 */
struct ahambox *
ahaincmbop(admp,mbop,wrap) 
register struct ahambox	*mbop;
register struct ahaadm	*admp;
register int *wrap;
{
	if(++mbop >= &admp->mbo[NMBOX]) /* Hit the edge yet ? */
	{
		(*wrap)++; /* Tell caller that we wrapped around the edge */
		return admp->mbo; /* Return start of array */
	}
	return mbop;	/* Just return next one */
}

/*
/* Pick a free CCB */
struct ahaccb *
ahagetccb(ctlr)
{
    register struct ahaccb	*ccbp;
    register struct ahaadm	*admp = &ahaadm[ctlr];
    register struct ahaccb	*ccbpinit;
             int		 wrap = 0;
	     int                 s;

agloop:
    ccbpinit = admp->freeccb;

    if((ccbpinit < admp->ccb) || (ccbpinit >= &admp->ccb[NCCB]))
    {
	printf("ahagetccb: ccbpinit out of range: %X set to %X\n",
		ccbpinit,admp->ccb);
	ccbpinit = admp->ccb;
    }

    s = splbio();
    for (ccbp = ccbpinit;
	(ccbp->aha_op != AHAOP_CCB_FREE) ; 
	 ccbp=ahaincccbp(admp,ccbp,&wrap))
    {
	if (wrap && (ccbp == ccbpinit))
	{
	    DEBUGF(ahasleepdebug, printf("Ahagetccb, no CCB avail.\n"););
	    splx(s);
	    return 0;
	}
    }

    bzero(ccbp, sizeof *ccbp ); 	/* Zero out the ccb */
    ccbp->aha_op = AHAOP_CCB_RESERVED;	/* Mark it as reserved */
    admp->ccbusage++;
    splx(s);

    if (admp->ccbusage > admp->maxccbusage)
    {
	admp->maxccbusage = admp->ccbusage;
	DEBUGF(ahausage, printf("Max CCB usage = %d\n", admp->maxccbusage););
    }

    admp->freeccb = ahaincccbp(admp,ccbp,&wrap);

    return ccbp;
}

/* Get a free MBO */
struct ahambox *
ahagetmbo(ctlr)
{
    register struct ahaadm	*admp     = &ahaadm[ctlr];
    register struct ahambox	*mbopinit = admp->freembo ;
    register struct ahambox	*mbop;
             int		 wrap     = 0;
	     int		 s;

    if((mbopinit < admp->mbo) || (mbopinit >= &admp->mbo[NMBOX]))
    {
	printf("ahagetmbo: mbopinit out of range: %X set to %X\n",
		mbopinit,admp->mbo);
	mbopinit = admp->mbo;
    }

    s = splbio();
    for (mbop = mbopinit;
	(mbop->aha_csb != MBO_MAILBOX_FREE) || (mbop->aha_ccbp) ; 
	 mbop=ahaincmbop(admp,mbop,&wrap))
    {
	if (wrap && (mbop == mbopinit))
	{
	    DEBUGF(ahasleepdebug, printf("Ahagetmbo: No mbo avail.\n"););
	    splx(s);
	    return 0;
	}
    }
    mbop->aha_ccbp = 1;	/* Reserve the Box */
    splx(s);

    admp->freembo = ahaincmbop(admp,mbop,&wrap);
    return mbop;
}

/*
/*
 * Free a CCB
 */
void
ahafreeccb(ctlr, ccbp)
struct ahaccb  *ccbp;
{
    struct ahaadm        *admp = &ahaadm[ctlr];
    if (ccbp)
    {
	ccbp->aha_op = AHAOP_CCB_FREE;
	admp->freeccb = ccbp;
	if (--admp->ccbusage < 0)
	    admp->ccbusage = 0;
    }
    else
    {
	uprintf("ahafreeccb: ccbp=0\n");
    }
}

void
ahafreembi(mp)
struct ahambox *mp;
{
	/* We set it to an obviously illegal value so that we can be sure
	   that the last byte is really written by the AHA */
	mp->aha_ccbp = 1;	/* Don't change order of these 2 stmts */
	mp->aha_csb = MBI_MAILBOX_FREE;
}


/* 
 * Free and clear MBO entries so that the timeout routine can 
 * recover outstanding requests.
 */
void
ahafreembo(ctlr, ccbp)
ioaddr_t        ccbp;
{
    struct ahambox *mbop;
    struct ahaadm  *admp = &ahaadm[ctlr];

    DEBUGF(ahascandebug,printf("Ahafreembo: ctlr=%d, ccbp=%X\n",ctlr,ccbp););
    for (mbop = admp->mbo; mbop < &admp->mbo[NMBOX]; mbop++)
    {
	if (mbop->aha_ccbp == ccbp)
	{
	    bzero(mbop, sizeof *mbop);
	    DEBUGF(ahascandebug,
		printf("MBO at %X cleared\n",mbop););
	    admp->freembo = mbop;
	    if (--admp->mbousage < 0)
	    {
		admp->mbousage = 0;
	    }
	    return;
	}
    }
}

/*
/***************** HARDWARE CONTROL ROUTINES SECTION *************************/
/*
 * These routines have the purpose of initailizing the hardware and checking
 * hardware operation.
 */
aharesetha(addr)
register struct aha_regs   *addr;
{
	register char dummy;
	int	s;
	register char csr;

    /* Empty Status and Data-In port */
    s = splhigh();

    /* 
     * In the stmt's below the register reading is the important thing, the
     * rest is there in order to fool the optimizer
     */
    if((dummy = addr->aha_csr) != 0)
    {
	dummy++; 
    }
    if((dummy = addr->aha_cd) != 0)
    {
	dummy--;
    }
    if((dummy = addr->aha_iflgs) != 0)
    {
	dummy++;
    }

    DELAY(1000);

    /* Reset everything in it */
    DEBUGF(ahainitdebug, printf("Aharesetha: AHA general reset\n"););
    addr->aha_csr = AHCSR_HRST | AHCSR_SRST | AHCSR_IRST | AHCSR_SCRST;

    DELAY(100000);

    /* Await completion of the self test */
    printf("AHA reset ...");
    while (addr->aha_csr & (AHCSR_STST))
	DELAY(1000);
    DEBUGF(ahainitdebug, printf("Aharesetha: AHA self test done \n"););

    /* Wait for either DIAGF or INIT to be set */
    while (!((csr = addr->aha_csr) & (AHCSR_DIAGF | AHCSR_INIT)))
	DELAY(100);
    printf(" done.\n");

    DEBUGF(ahainitdebug,printf("Aharesetha: AHA csr=%d\n",(short) csr););
    if (csr & AHCSR_DIAGF)
    {
	printf("Aharesetha: internal diagnostics failed\n");
	splx(s);
	return 1;
    }
    splx(s);
    return 0;
}

ahareinit(ic)
register struct iocc_ctlr *ic;
{
    register struct aha_regs   *addr = (struct aha_regs *) ic->ic_addr;

    if(aharesetha(addr))
    {
	printf("Ahareinit: Aha reset failed (Internal diag failure)\n");
	return 1;
    }
    if(ahainit(ic))
    {
	return 1;
    }
    ahainitmbox(NMBOX,(struct ahambox *) ahaadm[ic->ic_ctlr].mbo,ic->ic_ctlr);
    return 0;
}

/*
/* Initialize the host adapter */
ahainit(ic)
register struct iocc_ctlr *ic;
{
    int             		i;
    int         		rv;
    u_int           		intchan;
    register struct aha_regs   *addr = (struct aha_regs *) ic->ic_addr;
    char			swvers[5];

    /* Get SW revision */
    ahacmd(addr, AHACMD_INQ,ic->ic_ctlr);
    DELAY(10000);
    for(i=0; i < 4; i++)
    {
	if ((rv = ahaget(addr,ic->ic_ctlr)) < 0)
	{
	    printf("Ahainit: init failed(INQ, i = %d)\n",i);
	    return 1;
	}
	swvers[i] = rv;
    }
    swvers[4] = '\0';
    printf("aha%d: Firmware Revision '%s'\n",ic->ic_ctlr,swvers);

    DEBUGF(ahainitdebug,printf("aha%d:Installed: ", ic->ic_ctlr););
    DEBUGF(ahainitdebug,xhdump(ahaadm[ic->ic_ctlr].instdev,8););


    /* Get the configuration data */
    ahacmd(addr, AHACMD_RCD,ic->ic_ctlr);
    if ((rv = ahaget(addr,ic->ic_ctlr)) < 0)
    {
	printf("Ahainit: Init failed\n");
	return 1;
    }
    ic->ic_dmachannel -= 1;
    if (rv)
    {
	for (ic->ic_dmachannel = 0; !(rv & 1); ic->ic_dmachannel++)
	{
	    rv >>= 1;
	}
    }
    DEBUGF(ahainitdebug,
	printf("aha%d: Dma channel %d\n", ic->ic_ctlr, ic->ic_dmachannel););

    if ((rv = ahaget(addr,ic->ic_ctlr)) < 0)
    {
	printf("Ahainit: Init failed\n");
	return 1;
    }
    intchan = 0;
    if (rv)
    {
	for (intchan = 9; !(rv & 1); intchan++)
	{
	    rv >>= 1;
	}
    }
    DEBUGF(ahainitdebug,
	printf("aha%d: Interrupt channel %d\n", ic->ic_ctlr, intchan););

    if ((rv = ahaget(addr,ic->ic_ctlr)) < 0)
    {
	printf("Ahainit: Init failed\n");
	return 1;
    }
    DEBUGF(ahainitdebug,printf("aha%d: SCSI ID=%d\n", ic->ic_ctlr, rv & 7););
    ahasbontime(aha_bus_on, ic->ic_ctlr);	  /* Set Max BUSON time       */
    ahasbofftime(aha_bus_off, ic->ic_ctlr);	  /* Set BUS OFF time         */
    ahasetspeed(aha_speed, ic->ic_ctlr);	  /* Set Transfer speed       */
    ahassto(250, ic->ic_ctlr);		  /* Selection timeout 250 mS */
    /* ahasetmboint(AHAMBOINT_ENABLE,ic->ic_ctlr); /* Enable mailbox-out int   */
    return 0;
}

/*
/* Initialize the mailbox pointer */
void
ahainitmbox(n, amo, ctlr)
int             n;
struct ahambox *amo;
{
    ioaddr_t        mbp = ahioaddr((caddr_t) amo, ctlr);

    ahacommand(AHACMD_MBINIT, ctlr,AHA_CMD);
    ahacommand(n, ctlr,AHA_DATA);			    /* # of mailboxes */
    ahacommand((long) mbp >> 16 & 0xFF, ctlr,AHA_DATA);/* MSB  */
    ahacommand((long) mbp >> 8 & 0xFF, ctlr,AHA_DATA); /* MISB */
    ahacommand((long) mbp & 0xFF, ctlr,AHA_DATA);	    /* LSB  */
}

void
ahatout(ctlr)
{
    struct aha_regs *addr = (struct aha_regs *) ahainfo[ctlr]->ic_addr;
    struct ahambox *mbop;
    struct ahambox *mbip;
    struct ahaccb  *ccbp;
    struct ahaadm  *admp;
    u_char          iflgs;
    int	i,s,nmboxes;

    /* First note to driver that no timeout is pending */
    s = splbio();
    admp = &ahaadm[ctlr];
    admp->topending = 0;
    /* See if a kick was skipped */
    if(admp->skippedkick)
    {
	ahakick(ctlr);
	if(admp->skippedkick)
	{
	    admp->topending = 1;
	    timeout(ahatout,ctlr,TOTIME);
	}
	splx(s);
	return;
    }
    if(!ahatocheck)
    {
	splx(s);
	return;
    }

    /* Are there currently any active MBO's ? */
    if(admp->mbousage == 0)
    {
	splx(s);
	return;
    }
    /* OK, there are active MBO's, are we hung ? */
    /* See if there has been any interrupts since last Timeout activation */
    if(admp->move != 0)
    {
	char	dummy;
	/* Yes, interrupts has been taken, clear the move bit and issue 
	 * another timeout to see if the ctrlr REALLY have been waiting
	 * for a whole timeout period (with MBO's active).
	 */
	admp->move = 0;
	admp->topending = 1;
	dummy = addr->aha_cd;
	/* printf("Aha pending hang ? dummy=%x\n",dummy); */
	timeout(ahatout,ctlr,TOTIME);
	splx(s);
	return;
    }
    /* We get here when there are outstanding MBO's and there have been no
     * interrupt for a whole timeout period
     */

    DEBUGF(ahaerrdebug,
	printf("Ahatout: csr=%b, iflgs=%b, prio=%d mbo=%d\n",
	    addr->aha_csr,AHCSR_BITS,addr->aha_iflgs,AHIFLGS_BITS, getspl(),
	    admp->mbousage);
    );

    iflgs = addr->aha_iflgs;		/* Read the interrupt port  */
    if(iflgs & AHIFLGS_INT)
    {
	(void) spl3();
	DEBUGF(ahanotice,printf("Ahatout: Lost interrupts, Simulated one\n"););
	ahaint(ctlr);
	splx(s);
	return;
    }
#ifdef DEBUG
    if(ahaerrdebug)
    {
	printmbox(&admp->mbo,0,NMBOX);
	DELAY(3000000);
	printmbox(&admp->mbi,1,NMBOX);
	DELAY(3000000);
    }
#endif
    if (addr->aha_csr & AHCSR_IDLE)
    {
	int kickit=0;
	for(i=0;i<NMBOX;i++)
	{
	    mbop = &admp->mbo[i];

	    if(mbop->aha_ccbp)
	    {
		ccbp = (struct ahaccb *)
		    ahvaddr(mbop->aha_ccbp, ctlr); 
		if(ccbp < admp->ccb || 
		   ccbp >= &admp->ccb[NCCB])
		{
		    printf("Ahatout: Invalid CCB pointer = %X,ioccbp=%X\n",ccbp,mbop->aha_ccbp);
		    DELAY(1000000);
		    continue;
		}
		printf("Ahatout: forgotten mbox=%X\n",mbop);
		if(ccbp->aha_hastat == AHAST_STARTED)
		{
			printf("Ahatout: Hastat == STARTED, mbox restarted\n");
			mbop->aha_csb = MBO_START_SCSI;
			kickit++;
		}
		else
		{
			printf("Ahatout:Hastat = %d, Tarstat=%d\n",ccbp->aha_hastat,
				ccbp->aha_tarstat);
			mbip = ahagetmbi(ctlr);
			if(ccbp->aha_hastat || ccbp->aha_tarstat)
			{
			    mbip->aha_csb = MBI_ERROR;
			}
			else
			{
			    mbip->aha_csb = MBI_NOERROR;
			}

			mbip->aha_ccbp = mbop->aha_ccbp;
			printf("Simulated MBI intr for MBI=%X, ccbp=%X\n",
			    mbip, mbop->aha_ccbp);
			ahascanmbi(ctlr,0);
		}
	    }
	}
	if(kickit)
	{
	    ahakick(ctlr);
	}
	splx(s);
	return;
    }
    /* See if there is any CCB that isn't timed out yet */
    for(i=0;i<NMBOX;i++)
    {
	mbop = &admp->mbo[i];

	if(mbop->aha_ccbp)
	{
	    ccbp = (struct ahaccb *)
		ahvaddr(mbop->aha_ccbp, ctlr); 
	    if(ccbp < admp->ccb || 
		   ccbp >= &admp->ccb[NCCB])
	    {
		printf("Ahatout: Invalid CCB pointer = %X,ioccbp=%X\n",ccbp,mbop->aha_ccbp);
		DELAY(1000000);
		continue;
	    }
	    ccbp->aha_ttl -= TOTIME/HZ;
	    if(ccbp->aha_ttl >= 0)
	    /* This one didnt time out, so there is really no controller hang */	    /* we just restart another timeout period */
	    {
		admp->move = 0;
		admp->topending = 1;
		DEBUGF(ahaerrdebug,printf("No timeout yet \n"); );
		timeout(ahatout,ctlr,TOTIME);
		splx(s);
		return;
	    }
	}
    }
    printf("aha%d is hung\n", ctlr);
    /* ahatstartdebug=1; */
    /* ahacstartdebug=1; */
    splx(s);
    ahareinit(xha_hd[ctlr].xh_iocc);
    admp->nokick = 1;	    /* Don't start anything yet */

    nmboxes=0;
    for (mbop = admp->mbo;
	 mbop < &admp->mbo[NMBOX]; mbop++)
    {
	if (mbop->aha_ccbp)
	{
	    mbip = ahagetmbi(ctlr);
	    mbip->aha_csb = MBI_HOST_TIMEOUT;
	    mbip->aha_ccbp = mbop->aha_ccbp;
	    nmboxes++;
	    DEBUGF(ahanotice,printf("Simulated MBI:%X=%x.%X, for mbo:%X=%x.%X\n",
		    mbip, mbip->aha_csb, mbip->aha_ccbp,
		    mbop, mbop->aha_csb, mbip->aha_ccbp););
	}
    }
    if(nmboxes)
    {
	ahascanmbi(ctlr,0);
    }
    admp->nokick = 0;		    /* Let's go again */
    /* Kick the controller */
    ahakick(ctlr);
}

ahaget(addr,ctlr)
struct aha_regs *addr;
{
    int             waittime = 0;
    int             rv;
#ifdef DEBUG
    short           csr = -1;
#endif
    u_char          ahcsr;

    for (ahcsr = addr->aha_csr; !(ahcsr & AHCSR_DF); ahcsr = addr->aha_csr)
    {
	if ((ahaadm[ctlr].cmdcmpl != CMPL_NOTYET) && (ahaadm[ctlr].cmdcmpl != CMPL_CC))
	{
	    printf("aha%d: AHACMD ABORTED cc=%d\n",ctlr,ahaadm[ctlr].cmdcmpl);
	    ahaadm[ctlr].cmdcmpl = CMPL_ABORT;
	    return -1;
	}
	if (ahcsr & AHCSR_INVDCMD)
	{
	    printf("aha%d: AHACMD INVALID\n",ctlr);
	    ahaadm[ctlr].cmdcmpl = CMPL_INVD;
	    return -1;
	}
#ifdef DEBUG
	if (ahadebug)
	{
	    if (ahcsr != csr)
	    {
		if (csr != -1)
		    printf("\n");
		printf("aha%d: Ahaget: csr=%b", ctlr, (short) ahcsr, AHCSR_BITS);
		csr = ahcsr;
	    }
	    else
	    {
		printf(".");
	    }
	}
#endif DEBUG
	DELAY(100000);
	if (++waittime > 40)
	{
	    printf("aha%d: Ahaget: timeout\n",ctlr);
	    return -1;
	}
    }
#ifdef DEBUG
    if (ahadebug)
    {
	if (csr != -1)
	{
	    printf("\n");
	}
    }
#endif DEBUG
    rv = addr->aha_cd;
    DEBUGF(ahadebug, printf("Ahaget: returning 0x%x\n", rv););
    return rv;
}

void
ahascreset(addr)
struct aha_regs  *addr;
{
	addr->aha_csr = AHCSR_SCRST;
}

void

ahacommand(cmd, ctlr,cmd_op)
int cmd_op;
{
if(cmd_op){
    ahacmd((struct aha_regs *) ahainfo[ctlr]->ic_addr, cmd, ctlr);
}else{
    ahaop((struct aha_regs *) ahainfo[ctlr]->ic_addr, cmd, ctlr);
}
}

/* Send START SCSI command to the HA, Don't have to wait for IDLE */
void
ahakick(ctlr)
{
    u_char          cmd;
    u_char          csr;
    u_char	    dummy,dummyread=0;
    int		    loops=0;
    register int	s;
    struct aha_regs *addr=(struct aha_regs *)ahainfo[ctlr]->ic_addr;
    register struct ahaadm *admp=&ahaadm[ctlr];

    cmd = AHACMD_START_SCSI;

    s = splbio();
    if(admp->kicking)
    {
    	splx(s);
	return;
    }
    admp->kicking = 1;

again:
    if( (csr = addr->aha_csr) & (AHCSR_CDF))
    {
	admp->skippedkick = 1;
	admp->kicking=0;
	splx(s);
	return;
    }
	
    for (csr = addr->aha_csr; (csr & (AHCSR_STST | AHCSR_CDF))
	 ; csr = addr->aha_csr)
    {
	DEBUGF(ahakickdebug,printf("Ahakick: csr=%b loops=%d\n",csr,AHCSR_BITS,loops++););
	loops++;
	if(loops > 5000000)
	{
		if(dummyread==0)
		{
		    printf("Ahakick: Almost infinite loop\n");
		    dummy = addr->aha_cd;
		    dummyread++;
		    printf("ahakick: Read dummy data = %d\n",dummy);
		    loops = 0;
		    goto again;
		}
		break;
	}
	DELAY(1);
    }
    if((csr & AHCSR_DF) && (dummyread == 0))
    {
	dummy = addr->aha_cd;
	printf("ahakick: Read dummy data = %d\n",dummy);
	goto again;
    }
    addr->aha_cd = cmd;
    admp->kicking=0;
    admp->skippedkick = 0;
    splx(s);
    return;
}

ahacmd(addr, cmd, ctlr)
struct aha_regs *addr;
u_char          cmd;
{
    int             waittime = 0;
    int             waitmax = 300;
    u_short         csr;
    u_short         ir;
    int		    s; 

    DEBUGF(ahacmddebug, printf("Ahacmd: Addr=%X, cmd=%X\n", addr, cmd););
    s = splbio();
    for (csr = addr->aha_csr;
	 (csr & (AHCSR_STST | AHCSR_CDF)) ||
	 !(csr & AHCSR_IDLE); csr = addr->aha_csr)
    {
	ir = addr->aha_iflgs;
	DEBUGF(ahacmddebug, printf("Ahacmd: Aha_csr = %b aha_iflgs=%b\n",
		                   csr, AHCSR_BITS, ir, AHIFLGS_BITS););
	DELAY(1000);
	if (++waittime > waitmax)
	{
            printf("Ahacmd: Addr=%X, cmd=%X timeout\n", addr, cmd);
	    /*
	     * Primitive recovery action, can't be used for real system
	     * operation 
	     */
	    if (waitmax >= 300) /* HW == changed to >= */
	    {
		if (csr & AHCSR_CDF)
		{
		    printf("Warning... aha reset due to too many timeouts\n");
		    addr->aha_csr = AHCSR_IRST | AHCSR_SRST;
		}

		waitmax = 600;
	    }
	    /* We get here if reset didn't work */

	    splx(s);
	    return -1;
	}
    }
    ahaadm[ctlr].cmdcmpl = CMPL_NOTYET;
    DEBUGF(ahacmddebug, printf("Ahacmd: Command=0x%x\n", (short) cmd););
    addr->aha_cd = cmd;
    splx(s);
    return 0;
}
ahaop(addr, cmd, ctlr) /* added by HW */
struct aha_regs *addr;
u_char          cmd;
{
    int             waittime = 0;
    int             waitmax = 300;
    u_short         csr;
    u_short         ir;
    int             s;

    DEBUGF(ahacmddebug, printf("Ahaop: Addr=%X, cmd=%X\n", addr, cmd););
    s = splbio();
    for (csr = addr->aha_csr;
         (csr & (AHCSR_STST | AHCSR_CDF)) ; csr = addr->aha_csr)
    {
        ir = addr->aha_iflgs;
        DEBUGF(ahacmddebug, printf("Ahaop: Aha_csr = %b aha_iflgs=%b\n",
                                   csr, AHCSR_BITS, ir, AHIFLGS_BITS););
        DELAY(1000);
        if (++waittime > waitmax)
        {
            printf("Ahaop: Addr=%X, op=%X timeout\n", addr, cmd);
            /*
             * Primitive recovery action, can't be used for real system
             * operation
             */
            if (waitmax >= 300)
            {
                if (csr & AHCSR_CDF)
                {
                    printf("Warning... aha reset due to too many timeouts\n");
                    addr->aha_csr = AHCSR_IRST | AHCSR_SRST;
                }

                waitmax = 600;
            }
            /* We get here if reset didn't work */

            splx(s);
            return -1;
        }
    }
    ahaadm[ctlr].cmdcmpl = CMPL_NOTYET;
    DEBUGF(ahacmddebug, printf("Ahaop: Command=0x%x\n", (short) cmd););
    addr->aha_cd = cmd;
    splx(s);
    return 0;
}


void
ahasetspeed(i, ctlr) 
{
    ahacommand(AHACMD_SXFER_SPEED, ctlr,AHA_CMD);
    ahacommand(i, ctlr,AHA_DATA);
}

void
ahawritefifo(cp, ctlr) 
char * cp;
{
    ioaddr_t        iop = ahioaddr(cp, ctlr);

    ahacommand(AHACMD_WFIFO, ctlr,AHA_CMD);
    ahacommand(iop >> 16 & 0xFF, ctlr,AHA_DATA);	    /* MSB  */
    ahacommand(iop >> 8 & 0xFF, ctlr,AHA_DATA);	    /* MISB */
    ahacommand(iop & 0xFF, ctlr,AHA_DATA);		    /* LSB  */
}

void
ahareadfifo(cp, ctlr) 
caddr_t cp;
{
    ioaddr_t        iop = ahioaddr(cp, ctlr);

    ahacommand(AHACMD_RFIFO, ctlr,AHA_CMD);
    ahacommand(iop >> 16 & 0xFF, ctlr,AHA_DATA);	    /* MSB  */
    ahacommand(iop >> 8 & 0xFF, ctlr,AHA_DATA);	    /* MISB */
    ahacommand(iop & 0xFF, ctlr,AHA_DATA);		    /* LSB  */
}

void
ahassto(time, ctlr)
{
    ahacommand(AHACMD_SSTO, ctlr,AHA_CMD);
    ahacommand(time == 0 ? AHASSTO_DISABLE : AHASSTO_ENABLE, ctlr,AHA_DATA);
    ahacommand(0, ctlr,AHA_DATA);
    ahacommand(time >> 16 & 0xFF, ctlr,AHA_DATA);
    ahacommand(time & 0xFF, ctlr,AHA_DATA);
}

void
ahasetmboint(ena, ctlr)
{
    ahacommand(AHACMD_MBOINT, ctlr,AHA_CMD);
    ahacommand(ena, ctlr,AHA_DATA);
}

void
ahasbontime(time, ctlr)
{
    ahacommand(AHACMD_SBON_TIME, ctlr,AHA_CMD);
    ahacommand(time & 0xFF, ctlr,AHA_DATA);
}

ahasbofftime(time, ctlr)
{
    ahacommand(AHACMD_SBOFF_TIME, ctlr,AHA_CMD);
    ahacommand(time & 0xFF, ctlr,AHA_DATA);
}
/*
/******************** DMA INTERFACE ROUTINES **********************************/
/*
 * Convert I/O address to virtual memory address, 
 * The host adapter can either be addressed with 
 * an xha_device struct pointer xi
 * or directly with a host adapter number
 */
caddr_t
ahvaddr(ioaddr, ctlr)
ioaddr_t        ioaddr;
{
	return (caddr_t) ioaddr -
	    (caddr_t) ahaadm[ctlr].ioaddr +
	    (caddr_t) &ahaadm[ctlr];
}

/*
 * Convert vitrual memory address to an I/O address.
 * The host adapter can either be addressed with 
 * an xha_device struct pointer xi
 * or directly with a host adapter number
 */
ioaddr_t
ahioaddr(addr, ctlr)
caddr_t         addr;
{
	return (ioaddr_t) ((long) ahaadm[ctlr].ioaddr +
			   (addr - (caddr_t) &ahaadm[ctlr]));
}

/* 
 * Combine three bytes into an unsigned
 */
unsigned
ahaget3(p)
char           *p;
{
    return ((*p & 0xFF) << 16) | ((p[1] & 0xFF) << 8) | (p[2] & 0xFF);
}

/*
 * Split an unsigned long into three bytes
 */
void
asgn3(p, v)
char           *p;
u_long          v;
{
    *p++ = v >> 16 & 0xFF;
    *p++ = v >> 8 & 0xFF;
    *p = v & 0xFF;
}
/*
 * The dgo routine doesn't have to do anything since we're using
 * 3'rd party DMA
 */
void
ahadgo(dm, len, ioaddr, bp) 
struct iocc_ctlr *dm; unsigned ioaddr; struct buf *bp;
{
}
/*
/***************** DEBUGGING ROUTINES *****************************************/

#ifdef DEBUG
/* Some debugging aids */
void
printccb(cp)
struct ahaccb  *cp;
{
    printf("CCB    at %X:", cp);
    if (!cp)
    {
	return;
    }
    xhdump(cp, 16);
    printf("CDB    at %X:", &cp->aha_cdb);
    xhdump(&cp->aha_cdb, cp->aha_cmdlen);
    printf("Esense at %X:", ((char *) (&cp->aha_cdb)) + cp->aha_cmdlen);
    xhdump(((char *) (&cp->aha_cdb)) + cp->aha_cmdlen, AHA_SENSELEN);
    printf("Xi & bp  are:");
    xhdump(&cp->aha_xi, 8);
    printf("\n\tOp\t= 0x%x", cp->aha_op);
    printf("\tTarget\t= %d", cp->aha_id);
    printf("\tDout\t= %d, Din= %d", cp->aha_dout, cp->aha_din);
    printf("\tLun\t= %d", cp->aha_lun);
    printf("\tCmdlen\t= %d\n", cp->aha_cmdlen);
    printf("\tDlen\t= %d", cp->aha_dlen);
    printf("\tDptr\t= 0x%x%x%x", cp->aha_dptr[0], cp->aha_dptr[1], cp->aha_dptr[2]);
    printf("\tLptr\t= 0x%x%x%x", cp->aha_lptr[0], cp->aha_lptr[1], cp->aha_lptr[2]);
    printf("\tClid\t= %d", cp->aha_clid);
    printf("\tHastat\t= %d", cp->aha_hastat);
    printf("\tTarstat\t= %d\n", cp->aha_tarstat);
    printcdb(&cp->aha_cdb);
    printesense(((char *) &cp->aha_cdb) + cp->aha_cmdlen);
}
void
uprintccb(cp)
struct ahaccb  *cp;
{
    uprintf("CCB    at 0x%X:", cp);
    if (!cp)
    {
	return;
    }
    uxhdump(cp, 16);
    uprintf("CDB    at 0x%X:", &cp->aha_cdb);
    uxhdump(&cp->aha_cdb, cp->aha_cmdlen);
    uprintf("Esense at 0x%X:", ((char *) (&cp->aha_cdb)) + cp->aha_cmdlen);
    uxhdump(((char *) (&cp->aha_cdb)) + cp->aha_cmdlen, AHA_SENSELEN);
    uprintf("Xi & bp  are:");
    uxhdump(&cp->aha_xi, 8);
    uprintf("\n\tOp\t= 0x%x", cp->aha_op);
    uprintf("\tTarget\t= %d", cp->aha_id);
    uprintf("\tDout\t= %d, Din= %d", cp->aha_dout, cp->aha_din);
    uprintf("\tLun\t= %d", cp->aha_lun);
    uprintf("\tCmdlen\t= %d\n", cp->aha_cmdlen);
    uprintf("\tDlen\t= %d", cp->aha_dlen);
    uprintf("\tDptr\t= 0x%x%x%x", cp->aha_dptr[0], cp->aha_dptr[1], cp->aha_dptr[2]);
    uprintf("\tLptr\t= 0x%x%x%x", cp->aha_lptr[0], cp->aha_lptr[1], cp->aha_lptr[2]);
    uprintf("\tClid\t= %d", cp->aha_clid);
    uprintf("\tHastat\t= %d", cp->aha_hastat);
    uprintf("\tTarstat\t= %d\n", cp->aha_tarstat);
    uprintcdb(&cp->aha_cdb);
    uprintesense(((char *)&cp->aha_cdb) + cp->aha_cmdlen);
}

void
printcdb(cdbp)
struct scsi_cdb *cdbp;
{
}

void
printesense(sep)
struct scsi_esense *sep;
{
}

void
uprintcdb(cdbp)
struct scsi_cdb *cdbp;
{
}

void
uprintesense(sep)
struct scsi_esense *sep;
{
}

void
printmbox(box, io, n)
struct ahambox *box;
int             io;
{
    if (!box)
    {
	return;
    }
    if(n<=0)
    {
	return;
    }

    printf("%s Mbox%s at %X:\n", io ? "In " : "Out", n>1 ? "es":"", box);
	xhdump(box,4*n); 
}
void
uprintmbox(box, io, n)
struct ahambox *box;
int             io;
{

    if (!box)
    {
	return;
    }
    if(n<=0)
    {
	return;
    }

    uprintf("%s Mbox%s at %X:\n", io ? "In " : "Out", n>1 ? "es":"", box);
	uxhdump(box,4*n); 
}

char *
xcstr(c) 
u_char c;
{
    static char     outstr[3];
    char           *convstr = "0123456789abcdef";

    outstr[0] = convstr[(c >> 4) & 0xf];
    outstr[1] = convstr[c & 0xf];
    outstr[2] = '\0';

    return outstr;
}

void
uxhdump(cp, n)
u_char         *cp;
int             n;
{
    int             i = 0;

    while (n--)
    {
	uprintf("%s", xcstr(*cp++));
	++i;
	if ((i % 32) == 0)
	{
	    uprintf("\n");
	}
	else
	if ((i % 4) == 0)
	{
	    uprintf(" ");
	}
    }
    if (i % 32)
    {
	uprintf("\n");
    }
}

void
xhdump(cp, n)
u_char         *cp;
int             n;
{
    int             i = 0;

    while (n--)
    {
	printf("%s", xcstr(*cp++));
	++i;
	if ((i % 32) == 0)
	{
	    printf("\n");
	}
	else
	{
	    if ((i % 4) == 0)
	    {
		printf(" ");
	    }
	}
    }
    if (i % 32)
    {
	printf("\n");
    }
}

#endif DEBUG
#endif NAHA
/* HW return current spl level without seting it  */
int getspl()
{
int s;
s = splhigh();
splx(s);
return(s);
}

