#include "cs.h"
#if NCS > 0
#define	DEBUG 1 /* HW */
/*
 * Copyright (c) 1988, Christer Bernerus, Chalmers University of Technology,
 *                     Gothenburg, SWEDEN
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms as is or as a source
 * for derivative work are permitted provided that the above copyright 
 * notice and this paragraph are  duplicated in all such forms 
 * and that any documentation, advertising materials,
 * and other materials related to such distribution
 * and use acknowledge that the software was developed
 * at Chalmers University of Technology, SWEDEN.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Pertec CS240 SCSI Tape Streamer driver 
 *
 * $Source: /afs/sysrw/rt-ibm4.3-r3/usr/sys/caio/cs.c,v $
 * $Revision: 2.3 $ $Date: 89/02/24 13:10:06 $
 * $Author: bernerus $ $State: Exp $
 *
 * $Log:	cs.c,v $
 * Revision 2.3  89/02/24  13:10:06  bernerus
 * Changed the copyright notice.
 * 
 * Revision 2.2  89/02/08  13:09:36  bernerus
 * Made some ANSI-C constructions.
 * Now we do a SCSI Inquiry to get the tape type.
 * The probe routine now really does something.
 * 
 * Revision 2.1  89/01/30  08:01:57  bernerus
 * Removed some display() calls.
 * Added initial flag value in xha_driver structure.
 * 
 * Revision 2.0  89/01/16  11:16:50  bernerus
 * Added copyright notice.
 * Changed names cu->cs, cs->csc.
 * Added ioctl routine.
 * 
 *
 */
#include "../machine/pte.h"

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/vm.h"
#include "../h/buf.h"
#include "../h/file.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/ioctl.h"
#include "../h/cmap.h"
#include "../h/uio.h"
#include "../h/tty.h"
#include "../h/mtio.h"
#include "../machineio/ioccvar.h"
#include "../machineio/scsi.h"
#include "../machineio/xhavar.h"
#include "../machine/debug.h"

#define INF             1000000L		    /* a block number that
						     * won't exist */

/* Bits in minor device */

#define CSUNIT(dev)	XHAUNIT(dev)
#define CSNOREW(dev)	XHAMIN_NOREW(dev)
#define CSCUNIT(dev)	XHAMIN_TARGET(dev)

struct xha_device *cscinfo[NCSC];
struct xha_slave *csinfo[NCS];

short           cstypes[] = {1, 0};		    

#ifdef DEBUG
int             csdebug = 0;
int             csintdebug = 0;
int             csmodeselectdebug = 0;
int		cssccomdebug = 0;
int		cssleepdebug = 0;
int		cserrdebug = 0;
#endif

/* Per unit status information */

struct cs_softc
{
    char               scf_open:1;	/* unit is open 		      */
    char               scf_written:1;	/* last operation was a write 	      */
    char               scf_eot:1;	/* end of tape encountered 	      */
    char               scf_ieot:1;	/* ignore EOT condition 	      */
    char               scf_selectmode:1;/* Mode select needed b4 1:st op      */
    char               scf_busy:1;	/* Unit busy detected 		      */
    int                sc_busyretries;	/* Busy retry count	 	      */
    daddr_t            sc_blkno;	/* current physical block number      */
    daddr_t            sc_nxrec;	/* firewall input block number 	      */
    short              sc_resid;	/* residual function count for ioctl  */
    struct xha_device *sc_xi;		/* massbus structure for unit 	      */
    int                sc_slave;	/* slave number for unit 	      */
    struct tty        *sc_ttyp;		/* record user's tty for errors       */
    union _cdb				/* Gen purp SCSI CDB struct 	      */
    {
	struct scsi_cdb6 cdb6blk;
	struct scsi_cdb10 cdb10blk;
	struct scsi_cdb12 cdb12blk;
    }               sc_cdb;

}               cs_softc[NCS];

char sensebuf[NCS][SIZEOF_MODESENSE];	/* MODE SENSE buffer		      */
char selectbuf[NCS][SIZEOF_MODESELECT];	/* MODE SELECT buffer		      */

static struct scsi_cdb6 *
getsc6buf(dev_t dev)
{
    struct scsi_cdb6 *scp;

    scp = &cs_softc[CSUNIT(dev)].sc_cdb.cdb6blk;
    bzero(scp, 6);
    return scp;
}

static void
fillsc6buf(struct scsi_cdb6 *scp, dev_t dev, u_char op, u_char xflen)
{

    bzero(scp,6);
    scp->sc_op = op;
    scp->sc_lun = XHAMIN_LUN(dev);
    scp->sc_xflen[0] = xflen;
}

struct buf      rcsbuf[NCSC];	    /* data transfer buffer structures */
struct buf      ccsbuf[NCSC];	    /* tape command buffer structures */

#define b_command b_iodone
#define b_repcnt  b_bcount

struct scsi_cdb6 *csstart();

int csprobe(caddr_t, struct xha_device *), 
    csattach(struct xha_slave *), 
    csslave(struct xha_slave *, caddr_t),
    csustart(), 
    csint();
int             csstrategy();
int		csttl();
void            csminphys();
int 		csinquiry(struct xha_device *,short , struct scsi_inqblk *);

#define spl_tape() spl4()	     /* cpu level 4 */

struct xha_driver cscdriver =
{
    csprobe, 
    csslave,  
    csattach,  
    csustart,
    (int (*) ()) csstart,
    csint,
    cstypes, 
    "csc", 
    "cs", 
    cscinfo, 
    csinfo,
    csttl,
    EARLY_INT,
};

csprobe(caddr_t reg, struct xha_device *xi)
{
    struct scsi_inqblk 	iqbuf;
    struct scsi_inqblk 	*iq = &iqbuf;
    short lun;

    for(lun=0;lun<8;lun++)
    {
	if(csinquiry(xi,lun,iq) != 0)
	{
	    return PROBE_BAD;
	}
	if(iq->sinq_pdt == SCPDT_SASD)
	{
	    return PROBE_NOINT;
	}
    }	
    return PROBE_BAD;
}

/* ARGSUSED */
csattach(struct xha_slave *xs)
{
    if (xs->xs_unit > NCSC)
    {
	printf("cs%d: Not configured\n", xs->xs_unit);
	return 0;
    }
    bzero(&ccsbuf[xs->xs_unit], sizeof ccsbuf[xs->xs_unit]);
    bzero(&rcsbuf[xs->xs_unit], sizeof rcsbuf[xs->xs_unit]);
    bzero(&cs_softc[xs->xs_unit], sizeof cs_softc[xs->xs_unit]);
    bzero(sensebuf[xs->xs_unit], SIZEOF_MODESENSE);
    bzero(selectbuf[xs->xs_unit], SIZEOF_MODESELECT);

    return 1;
}

csinquiry(struct xha_device *xi,short lun, struct scsi_inqblk *iq)
{
	static struct scsi_cdb6 sc6blk;
	register struct scsi_cdb6 *sc6buf = &sc6blk;

	fillsc6buf(sc6buf,lun,SC0_INQUIRY,sizeof *iq);
        iq->sinq_pdt = SCPDT_NO_LUN;

	/*
	 * Now do the Inquiry 
	 */
	if(((*xi->xi_hadriver->xhc_doscsi)(xi->xi_ctlr,xi->xi_drive,lun,
	     sc6buf,iq,sizeof *iq,B_READ | B_WANTED)) == 
	     ((XHI_CMPL << 8) + SCTS_GOOD))
	{
		return 0;
	}
	return -1;
}

csslave(struct xha_slave *xs, caddr_t addr)
{
    struct xha_device	*xi = xs->xs_device;
    struct scsi_inqblk 	iqbuf;
    struct scsi_inqblk 	*iq = &iqbuf;
/*    static char inqname[] = "92170"; HW */
    static char inqname[] = "EXB-8200"; 
    static char		vid[sizeof(iq->sinq_vid)+1];
    static char		pid[sizeof(iq->sinq_pid)+1];
    static char		fwrev[sizeof(iq->sinq_fwrev)+1];

    if(xs->xs_driver != &cscdriver) 
    {
	return 0;
    }

    if(xi)
    {
	if(csinquiry(xi,xs->xs_slave,iq) == 0)
	{
		if(iq->sinq_pdt != SCPDT_SASD)
		{
		    return 0;
		}

		csidcpy(iq->sinq_vid,vid,sizeof vid -1);
		csidcpy(iq->sinq_pid,pid,sizeof pid -1);
		csidcpy(iq->sinq_fwrev,fwrev,sizeof fwrev -1);

		printf("%s%d:%s %s Rev. %s ANSI V.%d\n",
			xi->xi_driver->xd_sname,xi->xi_unit,
			vid,pid,fwrev,iq->sinq_ansiversion);

	}
	else
	{
	    return 0;
	}
    }
    else
    {
	printf("csslave: xs->xs_device not loaded, fix autoconfig\n");
	return 0;
    }
    return 1;
}

csidcpy(char *f,char *t,int size)
{
    register int	i;

    bcopy(f,t,size-1);
    t[size-1] = '\0';

    for(i=size-1;i>=0;i--)
    {
	if((t[i] != ' ') && (t[i] != '\0'))
	{
	    break;
	}
	t[i] = '\0';
    }
}

/* Tell controller routine expected max lifetime of */
/* a command (time in seconds)			    */
csttl(struct scsi_cdb6 *com)
{
	switch(com->sc_op)
	{
	    case SC0_REWIND:		return 1000/24;
	    case SC0_SPACE:		return 1000;
	    case SC0_TRACK_SELECT:	return 5;
	    case SC0_LOAD_UNLOAD:	return 1000/24+30;
	    case SC0_ERASE:		return 1000;
	    default:			return 1;
	}
}

csopen(dev_t dev, int flag)				    /* Open routine */
{
    struct xha_device *xi;
    struct cs_softc *sc;
    int             csunit;
    int             cscunit;
    struct scsi_modesense *msenp;

    cscunit = CSCUNIT(dev);
    csunit = CSUNIT(dev);

    /* Validate unit and aliveness */
    xi = cscinfo[cscunit];
    if ((csunit >= NCS)
	|| (xi == 0)
	|| (xi->xi_alive == 0))
    {
	uprintf("Csopen: returning ENXIO: xi=0x%X,alive=0x%X\n",
		xi, xi ? xi->xi_alive : -1);
	return (ENXIO);
    }

    /*
     * Check if already open, no spl's needed since we don't voluntarily give
     * up the cpu and the system will nor reschedule while in kernel 
     */
    if ((sc = &cs_softc[csunit])->scf_open)
	return (EBUSY);
    bzero(sc, sizeof *sc);
    sc->scf_open = 1;

    /* OK, now get mode status from the unit */

    msenp = (struct scsi_modesense *) sensebuf[csunit];
    if ((flag & FWRITE))
	if (csmodesense(xi, dev, msenp) != 0)
	{
	    uprintf("cs%s: not online ?\n", csunit);
	    sc->scf_open = 0;
	    return EIO;
	}

    if ((flag & FWRITE) && (msenp->sm_wp))
    {
	uprintf("cs%s: write protected\n", csunit);
	sc->scf_open = 0;
	return EIO;
    }

    sc->scf_selectmode = 1;

    sc->sc_blkno = (daddr_t) 0;

    /* Since cooked I/O may do a read-ahead before a write, trash   */
    /* on a tape can make the first write fail.  Suppress the first */
    /* read-ahead unless definitely doing read-write                */

    sc->sc_nxrec = ((flag & (FTRUNC | FWRITE)) == (FTRUNC | FWRITE))
	? (daddr_t) 0
	: (daddr_t) INF;

    sc->sc_ttyp = u.u_ttyp;
    return 0;
}

/*
 * Code to execute mode select command 
 */
csmodeselect(xi, dev, mselp)
struct scsi_modeselect	*mselp;
struct xha_device	*xi;
dev_t			 dev;
{
    register struct scsi_cdb6 *scp;

    /* Get a SCSI command buffer  and fill it in */
    scp = getsc6buf(dev);
    fillsc6buf(scp, dev, SC0_MODE_SELECT, SIZEOF_MODESELECT);

    return (cssccom(dev,scp,1,mselp,SIZEOF_MODESELECT));
}

/*
 * Code to execute SCSI Mode sense command 
 */
csmodesense(xi, dev, msenp)
struct scsi_modesense	*msenp;
struct xha_device	*xi;
dev_t			 dev;
{
    register struct scsi_cdb6 *scp;

    /* Get a SCSI command buffer  and fill it in */
    scp = getsc6buf(dev);
    fillsc6buf(scp, dev, SC0_MODE_SENSE, SIZEOF_MODESENSE);

    return (cssccom(dev,scp,1,msenp,SIZEOF_MODESENSE));
}


csclose(dev, flag)
dev_t           dev;
int             flag;
{
    struct xha_device *xi;
    struct cs_softc *sc = &cs_softc[CSUNIT(dev)];
    struct scsi_cdb6 *scp;

    xi = cscinfo[CSCUNIT(dev)];

    /*
     * Now we must shut off selectmode if still set, or else the 
     * process will hang up since csstrategy will call cssccom while
     * csstrategy is called from cssccom ....
     */
    sc->scf_selectmode = 0;

    if (((flag & (FREAD | FWRITE)) == FWRITE)
	|| ((flag & FWRITE) && (sc->scf_written)))
    {
	scp = getsc6buf(dev);
	fillsc6buf(scp, dev, SC0_WRITE_FILEMARKS, 2);
	cssccom(dev,scp,1,0,0);		    /* Write 2 filemarks */
    }

    /*
     * If we should rewind the tape, issue that command, else we just close it 
     */

    if (!CSNOREW(dev))
    {
	scp = getsc6buf(dev);
	fillsc6buf(scp, dev, SC0_REWIND, 0);
	scp->sc_lbaMSB |= 1;			    /* IMMEDIATE bit set */
	cssccom(dev,scp,1,0,0);		    /* Rewind the tape */
    }
    sc->scf_open = 0;
    return 0;
}

cssccom(dev,scp,wait,spacep,size)
register dev_t			 dev;
register struct scsi_cdb6	*scp;
register int			 wait,
				 size;
register caddr_t		 spacep;

{
    register struct buf *bp;

    int             s;

    /* Get command buffer */
    bp = &ccsbuf[CSCUNIT(dev)];
    s = spl5();
    while (bp->b_flags & B_BUSY)
    {
	if ((bp->b_repcnt == 0) && (bp->b_flags & B_DONE))
	    break;
	bp->b_flags |= B_WANTED;
	DEBUGF(cssleepdebug,
	    printf("Waiting for command buffer %X *scp=%x. zzzz\n",
		bp,scp->sc_op););
	sleep((caddr_t) bp, PRIBIO);
	DEBUGF(cssleepdebug,
	    printf("Aaah got command buffer %X, *scp=%x ?\n",bp,scp->sc_op););
    }
    bp->b_flags = B_BUSY | B_READ;
    splx(s);

    /* Fill in the buf variables */
    bp->b_dev = dev;
    bp->b_bufsize = size;
    bp->b_command = (int (*) ()) scp;
    bp->b_bcount = size;
    bp->b_blkno = 0;
    bp->b_error = 0;
    bp->b_errcnt = 0;
    bp->b_un.b_addr = spacep;
    DEBUGF(cssccomdebug, printf("Cssccom: buf=%X,spacep=0x%X, size=%d, op=%x,wait=%d\n",
				      bp,
				      bp->b_un.b_addr,bp->b_bufsize,
				      scp->sc_op,wait);); 

    /* Call strategy routine */
    csstrategy(bp);

    /* Await completion */
    if(wait)
    {
	DEBUGF(cssleepdebug,
	    uprintf("Waiting for command bfr %X op=%x to complete\n",
		     bp,scp->sc_op););
	iowait(bp);
	DEBUGF(cssleepdebug,
	    uprintf("Command bfr %X, op=%x done\n",bp,scp->sc_op););
    }

    /* If anyone wants this buffer, wake them up */
    if(bp->b_flags & B_WANTED)
	wakeup((caddr_t) bp);

    bp->b_flags &= ~(B_BUSY | B_WANTED | B_PHYS);
    return geterror(bp);
}

csustart(xi)
struct xha_device *xi;
{
    register struct buf *bp = xi->xi_tab.b_actf;
    register struct cs_softc *sc = &cs_softc[CSUNIT(bp->b_dev)];
    int             cscunit = CSCUNIT(bp->b_dev);

    if (sc->scf_open == 0)
    {
	bp->b_flags |= B_ERROR;
	return (XHU_NEXT);
    }
    if (bp != &ccsbuf[cscunit])
    {

	if ((sc->scf_eot && !sc->scf_ieot) &&
	    ((bp->b_flags & B_READ) == 0))
	{
	    bp->b_flags |= B_ERROR;
	    bp->b_error = ENOSPC;
	    return (XHU_NEXT);
	}

	/* special case tests for cooked mode */

	if (bp != &rcsbuf[cscunit])
	{

	    /* seek beyond end of file */

	    if (bdbtofsb(bp->b_blkno) > sc->sc_nxrec)
	    {
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		return (XHU_NEXT);
	    }

	    /* This should be end of file, but the buffer      */
	    /* system wants a one-block look-ahead.  Humor it. */

	    if ((bdbtofsb(bp->b_blkno) == sc->sc_nxrec)
		&& (bp->b_flags & B_READ))
	    {
		clrbuf(bp);
		return (XHU_NEXT);
	    }

	    /* If writing, mark the next block invalid. */

	    if ((bp->b_flags & B_READ) == 0)
		sc->sc_nxrec = bdbtofsb(bp->b_blkno) + 1;
	}
    }
    /* If raw I/O, or if the tape is positioned correctly for       */
    /* cooked I/O, set the byte count, unit number and repeat count */
    /* then tell the SCSI to proceed.  Note that a negative      */
    /* bcount tells mbstart to map the buffer for "read backwards". */

    /*
     * if (   (bp == &rcsbuf[cscunit]) || ((blkno = sc->sc_blkno) ==
     * bdbtofsb(bp->b_blkno)) ) { if (mi->mi_tab.b_errcnt == 2) { mtaddr->mtbc
     * = -(bp->b_bcount); mtaddr->mtca = MUUNIT(bp->b_dev); } else {
     * mtaddr->mtbc = bp->b_bcount; mtaddr->mtca = (1<<2)|MUUNIT(bp->b_dev); }
     * return (XHU_DODATA); } 
     */
    return (XHU_DODATA);
}

struct scsi_cdb6 *
csstart(xi,bp,combuf)
struct xha_device *xi;
struct scsi_cdb12 *combuf;
register struct buf *bp;
{
    register struct scsi_cdb6 *scbuf;
    register int    cscunit;

    cscunit = CSCUNIT(bp->b_dev);
    if (bp == &ccsbuf[cscunit])
	return (struct scsi_cdb6 *) bp->b_command;


    if(combuf)
	scbuf = (struct scsi_cdb6 *) combuf;
    else
	scbuf = getsc6buf(bp->b_dev);

    fillsc6buf(scbuf, bp->b_dev, bp->b_flags & B_READ ? SC0_READ : SC0_WRITE, 1);

    if (bp->b_flags & B_PHYS)
    {
	scbuf->sc_lbarest[0] = bp->b_bcount >> 16 & 0xFF;
	scbuf->sc_lbarest[1] = bp->b_bcount >> 8 & 0xFF;
	scbuf->sc_xflen[0] = bp->b_bcount & 0xFF;
    }
    else
	scbuf->sc_lbaMSB = 1;			    /* I.e. fixed mode */
    return scbuf;
}

csint(xi, bp, ccode, sdp, tarstat)
struct xha_device *xi;
struct buf     *bp;
struct scsi_esense *sdp;
u_char          tarstat;
{
    struct cs_softc *sc = &cs_softc[xi->xi_unit];

    DEBUGF(csintdebug, printf("Csint: xi=0x%X, bp=0x%X, ccode=%d, sdp=0x%X\n",
		               xi, bp, ccode, sdp););

    if (tarstat == SCTS_BUSY)
    {
	if (sc->scf_busy)
	{
	    if (++sc->sc_busyretries > 10)
	    {
		bp->b_error = EIO;
		bp->b_flags |= B_ERROR;
		sc->scf_busy = 0;
		return XHD_DONE;
	    }
	    return XHD_RETRY;
	}
	sc->scf_busy = 1;
	sc->sc_busyretries = 1;
	return XHD_RETRY;
    }
    sc->scf_busy = 0;


    switch (ccode)
    {
    case XHI_CMPL:
	sc->scf_written = !(bp->b_flags & B_READ);
	sc->sc_blkno++;
	bp->b_resid = 0;
	return XHD_DONE;
    case XHI_TIMEOUT:
	if (++bp->b_errcnt <= 1)
	    return XHD_RETRY;
	bp->b_error = EIO;
	bp->b_flags |= B_ERROR;
	return XHD_DONE;

    case XHI_ERR:
	/* Analyze the error */
	if (sdp)
	    switch (sdp->es_key)
	    {
	    case SCES_KEY_NO_SENSE:
		DEBUGF(cserrdebug, printf("Csint: NO SENSE KEY\n"););
		/* Check FMK, ILI and EOM bits here */
		sc->sc_blkno++;
		if (sdp->es_fmark)
		{
		    DEBUGF(cserrdebug, printf("Csint: File mark detected\n"););
		    if (bp != &rcsbuf[CSUNIT(bp->b_dev)])
			sc->sc_nxrec = bdbtofsb(bp->b_blkno);
		}
		if (sdp->es_ili)
		{
		    int             res;

		    /*
		     * Get residual count, hope this will do expected things
		     * for most compilers 
		     */
		    res = (sdp->es_info[0] << 24) +
			(sdp->es_info[1] << 16) +
			(sdp->es_info[2] << 8) +
			sdp->es_info[3];
		    DEBUGF(cserrdebug, 
			   printf("Csint: Invalid length indicated, res=%d\n",
			   res););
		    if (res < 0)		    /* I.E. Long record */
		    {
			bp->b_bcount = bp->b_resid;
			bp->b_resid = 0;
			bp->b_error = ENOMEM;
			bp->b_flags |= B_ERROR;
		    }
		    else			    /* Short record */
		    {
			bp->b_bcount = bp->b_resid;
			bp->b_resid += res;
			if (bp != &rcsbuf[CSUNIT(bp->b_dev)])
			    bp->b_flags |= B_ERROR;
		    }
		}
		if (sdp->es_eom)
		{
		    DEBUGF(cserrdebug, 
		   	   printf("Csint: End of medium detected\n"););
		    sc->scf_eot = 1;
		    bp->b_resid = 0;
		}
		return XHD_DONE;
	    case SCES_KEY_RECOVERED_ERROR:
		DEBUGF(csintdebug, printf("Csint: RECOVERED ERROR KEY\n"););
		sc->scf_written = !(bp->b_flags & B_READ);
		sc->sc_blkno++;
		bp->b_resid = 0;
		return XHD_DONE;
	    case SCES_KEY_NOT_READY:
		DEBUGF(cserrdebug, printf("Csint: NOT READY KEY\n"););
		bp->b_error = EIO;
		bp->b_flags |= B_ERROR;
		break;
	    case SCES_KEY_MEDIUM_ERROR:
		DEBUGF(cserrdebug, printf("Csint: MEDIUM ERROR KEY\n"););
		bp->b_error = EIO;
		bp->b_flags |= B_ERROR;
		break;
	    case SCES_KEY_HARDWARE_ERROR:
		DEBUGF(cserrdebug, printf("Csint: HARDWARE ERROR KEY\n"););
		bp->b_error = EIO;
		bp->b_flags |= B_ERROR;
		break;
	    case SCES_KEY_ILLEGAL_REQUEST:
		DEBUGF(cserrdebug, printf("Csint: ILLEGAL REQUEST KEY\n"););
		if (++bp->b_errcnt <= 1)
		    return XHD_RETRY;
		else
		{
		    bp->b_bcount = bp->b_resid;
		    bp->b_resid = 0;
		    bp->b_error = EIO;
		    bp->b_flags |= B_ERROR;
		    break;
		}
	    case SCES_KEY_UNIT_ATTENTION:
		DEBUGF(cserrdebug, printf("Csint: UNIT ATTENTION KEY\n"););
		if (++bp->b_errcnt <= 1)
		    return XHD_RETRY;
		else
		{
		    bp->b_bcount = bp->b_resid;
		    bp->b_resid = 0;
		    bp->b_error = EIO;
		    bp->b_flags |= B_ERROR;
		    break;
		}
	    case SCES_KEY_DATA_PROTECT:
		DEBUGF(cserrdebug, printf("Csint: DATA PROTECT KEY\n"););
		bp->b_error = EIO;
		bp->b_flags |= B_ERROR;
		break;
	    case SCES_KEY_BLANK_CHECK:
		DEBUGF(cserrdebug, printf("Csint: BLANK CHECK KEY\n"););
		bp->b_flags |= B_ERROR;
		break;
	    case SCES_KEY_COPY_ABORTED:
		DEBUGF(csintdebug, printf("Csint: COPY ABORTED KEY\n"););
		bp->b_flags |= B_ERROR;
		break;
	    case SCES_KEY_ABORTED_COMMAND:
		DEBUGF(csintdebug, printf("Csint: ABORTED COMMAND KEY\n"););
		bp->b_flags |= B_ERROR;
		break;
	    case SCES_KEY_VOLUME_OVERFLOW:
		DEBUGF(cserrdebug, printf("Csint: VOLUME OVERFLOW KEY\n"););
		bp->b_error = ENOSPC;
		bp->b_flags |= B_ERROR;
		break;
	    default:
		DEBUGF(cserrdebug, 
		       printf("Csint: Unknown KEY=%d\n", sdp->es_key););
		break;
	    }
	else
	    printf("Csint: Sense data pointer is zero\n");
	break;
    }

    /* Make sure the byte count is reasonable.                           */
    if (bp->b_bcount < 0)
	bp->b_bcount = bp->b_resid;

    return XHD_DONE;
}

csread(dev, uio)
dev_t           dev;
struct uio     *uio;
{
    int             errno;

    errno = csphys(dev, uio);
    if (errno)
	return (errno);
    return (physio(csstrategy, &rcsbuf[CSCUNIT(dev)], dev, B_READ, csminphys, uio));
}

cswrite(dev, uio)
dev_t           dev;
struct uio     *uio;
{
    int             errno;

    DEBUGF(csdebug, printf("Cswrite: dev=0x%x,uio=0x%X\n", dev, uio););
    errno = csphys(dev, uio);
    if (errno)
	return (errno);
    return (physio(csstrategy, &rcsbuf[CSCUNIT(dev)], dev, B_WRITE, csminphys, uio));

}

csphys(dev, uio)
dev_t           dev;
struct uio     *uio;
{
    register int    target,
                    lun;
    struct xha_device *xi;
    struct xha_slave *xs;
    register int    bsize = uio->uio_iov->iov_len;

    DEBUGF(csdebug, printf("Csphys: dev=0x%x,uio=0x%X\n", dev, uio););

    target = CSCUNIT(dev);
    lun = CSUNIT(dev);
    if (((xi = cscinfo[target]) == 0)
	|| (xi->xi_alive == 0))
	return (ENXIO);
    if (((xs = csinfo[lun]) == 0)
	|| (xs->xs_alive == 0))
	return (ENXIO);
    if ((bsize > 0xffff)			    /* blocksize limit */
	|| (bsize <= 0))
	return (EINVAL);
    return (0);
}

csstrategy(bp)
struct buf     *bp;
{
    register struct xha_device *xi = cscinfo[CSCUNIT(bp->b_dev)];
    register struct buf *dp;
    register int    s;
    register struct xhac_driver *xc;
    register struct cs_softc *sc;
    int             csunit;
    struct scsi_modeselect *mselp;

    DEBUGF(csdebug, printf("Csstrategy: bp=0x%X\n", bp););
    csunit = CSUNIT(bp->b_dev);
    sc = &cs_softc[csunit];
    xc = xi->xi_hadriver;
    if (sc->scf_selectmode)
    {
	mselp = (struct scsi_modeselect *) selectbuf[csunit];
	bzero(mselp, SIZEOF_MODESELECT);
	mselp->sms_bdl = sizeof mselp->sms_bdarr[0];
	if (!(bp->b_flags & B_PHYS))
	{
	    mselp->sms_bufmode = 1;
	    mselp->sms_bdarr[0].sbd_blklen = bp->b_bcount;
	}
	mselp->sms_bufmode = 1;
	sc->scf_selectmode = 0;			    /* have to clear this bit
						     * first, or else an
						     * infinite loop will start */
	if (csmodeselect(xi, bp->b_dev, mselp) != 0)
	{
	    printf("MODE SELECT FAILED\n");
	    bp->b_flags |= B_ERROR;
	    bp->b_error = EIO;
	    return (XHU_NEXT);
	}
    }
/*
    if (bp != &ccsbuf[CSCUNIT(bp->b_dev)]) 
	bp->b_resid = bp->b_bcount;
*/
    bp->av_forw = NULL;
    dp = &xi->xi_tab;
    s = spl_tape();
    if (dp->b_actf == NULL)
	dp->b_actf = bp;
    else
	dp->b_actl->av_forw = bp;
    dp->b_actl = bp;
	(*xc->xhc_ustart) (xi);
    splx(s);
}

csioctl(dev, cmd, data, flag)
dev_t           dev;
int             cmd;
caddr_t         data;
int             flag;
{
	register struct cs_softc *sc = &cs_softc[CSUNIT(dev)];
	int fcount;
	struct mtop *mtop;
	struct mtget *mtget;
	struct scsi_cdb6 *scp;
	int	smode,rv;
	static u_char csops[] =
	{
		SC0_WRITE_FILEMARKS,
		SC0_SPACE,
		SC0_SPACE,
		0xFF,
		0xFF,
		SC0_REWIND,
		SC0_LOAD_UNLOAD,
		SC0_TEST_UNIT_READY,
		SC0_ERASE,
		SC0_LOAD_UNLOAD,
	};	
	DEBUGF( (csdebug),
		printf("stioctl: entered\n"));


	switch (cmd) {

	case MTIOCTOP:	/* tape operation */
		mtop = (struct mtop *)data;
		switch (mtop->mt_op) {

                case MTWEOF:
                        fcount = mtop->mt_count;
                        break;
                case MTFSR:
                case MTBSR:
                        fcount =  -mtop->mt_count;
                        break;
                case MTFSF:
                case MTBSF:
                        fcount = mtop->mt_count;
                        break;
                case MTERASE:
                        fcount = 0; /* HW EXAbyte  (1 for cs )*/
                        break;
                case MTREW:
                        fcount = 0; /* HW EXAbyte  (1 for cs )*/
                        break;
                case MTOFFL:
                        fcount = 0;  /* Becomes !Re-Ten & !Load */
                        break;
                case MTNOP:
                        fcount = 0;
                        break; /* HW */
                case MTRETENSION:
                        fcount = 1; /* HW (0x3) Becomes Re-Ten | Load */
                        break;

                default:
                        return (ENXIO);

		}
		DEBUGF(csdebug,
			printf("cmd=0x%x fcount=%d ", csops[mtop->mt_op], fcount));
		scp = getsc6buf(dev);
		fillsc6buf(scp,dev,csops[mtop->mt_op], fcount);
		switch(csops[mtop->mt_op])
		{
			case SC0_SPACE:
				if(fcount < 0)
				{
					scp->sc_lbarest[0] = 0xff;
					scp->sc_lbarest[1] = 0xff;
				}
				scp->sc_lbaMSB |= 0x1; /* Filemarks */
				break;
			case SC0_WRITE_FILEMARKS:
			case SC0_REWIND:
			case SC0_LOAD_UNLOAD:
			case SC0_TEST_UNIT_READY:
				break;
			case SC0_ERASE:
				scp->sc_lbaMSB |= 0x1; /* Long erase */
				break;
		}
		smode = sc->scf_selectmode;
		sc->scf_selectmode = 0; /* Disable mode select */
		rv = cssccom(dev,scp,1,0,0);
		sc->scf_selectmode = smode;
		return rv;

	case MTIOCGET:
		mtget = (struct mtget *)data;
		mtget->mt_dsreg = 0;
		/*
		mtget->mt_erreg = sc->sc_data[0] + (sc->sc_data[1] << 8);
		*/
		mtget->mt_resid = sc->sc_resid;
		mtget->mt_type = MT_ISCS;
		break;

	default:
		return (ENXIO);
	}
	return (0);
}

void
csminphys(bp)
struct buf     *bp;
{
    register int    cscunit = CSCUNIT(bp->b_dev);
    register struct iocc_ctlr *ic;
    register struct xha_device *xi;
    extern unsigned (*dma_get_minphys()) ();


    /* if(csphys(bp->b_dev)) return;	/* Check existence and aliveness */

    minphys(bp);				    /* Enforce kernel-wide
						     * b_bcount restriction */

    if (bp->b_bcount > 65536)
	bp->b_bcount = 65536;

    xi = cscinfo[cscunit];

    ic = xi->xi_hd->xh_iocc;
    (*dma_get_minphys(ic)) (bp);		    /* dma's bcount
						     * restrictions */

}

#endif
