#if !defined(lint) && !defined(LOCORE) && defined(RCS_HDRS)
static char *rcsidscsi = "$Header:scsi.h 1.1$";
#endif
/*
 * Copyright (c) 1988, Christer Bernerus, Chalmers University of Technology,
 *                     Gothenburg, SWEDEN
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms as is or as a source
 * for derivative work are permitted provided that the above copyright 
 * notice and this paragraph are  duplicated in all such forms 
 * and that any documentation, advertising materials,
 * and other materials related to such distribution
 * and use acknowledge that the software was developed
 * at Chalmers University of Technology, SWEDEN.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * COMMON DEFINITIONS FOR SCSI AS OF ANSI X2.131-1986
 */

/* SCSI	COMMANDS		     COMMON/DASD    SEQ    PRT*/
#define	SC0_TEST_UNIT_READY		    0x00
#define	SC0_REZERO_UNIT			    0x01
#define	SC0_REWIND				    0x01
#define	SC0_REQUEST_SENSE		    0x03
#define	SC0_FORMAT_UNIT			    0x04
#define	SC0_FORMAT					    0x04
#define	SC0_READ_BLOCK_LIMITS				    0x05
#define	SC0_REASSIGN_BLOCKS		    0x07
#define	SC0_READ			    0x08
#define	SC0_WRITE			    0x0A
#define	SC0_PRINT					    0x0A
#define	SC0_SEEK			    0x0B
#define	SC0_TRACK_SELECT			    0x0B
#define	SC0_SLEW_AND_PRINT				    0x0B
#define	SC0_READ_REVERSE			    0x0F
#define	SC0_WRITE_FILEMARKS			    0x10
#define	SC0_FLUSH_BUFFER				    0x10
#define	SC0_SPACE				    0x11
#define	SC0_INQUIRY			    0x12
#define	SC0_VERIFY				    0x13
#define	SC0_RECOVER_BUFFERED_DATA		    0x14
#define	SC0_MODE_SELECT			    0x15
#define	SC0_RESERVE			    0x16
#define	SC0_RESERVE_UNIT			    0x16
#define	SC0_RELEASE			    0x17
#define	SC0_RELEASE_UNIT			    0x17
#define	SC0_COPY			    0x18
#define	SC0_ERASE				    0x19
#define	SC0_MODE_SENSE			    0x1A
#define	SC0_START_STOP_UNIT		    0x1B
#define	SC0_LOAD_UNLOAD				    0x1B
#define	SC0_STOP_PRINT					    0x1B
#define	SC0_RECEIVE_DIAGNOSTIC_RESULTS	    0x1C
#define	SC0_SEND_DIAGNOSTIC		    0x1D
#define	SC0_PREVENT_ALLOW_MEDIUM_REMOVAL    0x1E
#define	SC1_READ_CAPACITY		    0x25
#define	SC1_READ			    0x28
#define	SC1_WRITE			    0x2A
#define	SC1_SEEK			    0x2B
#define	SC1_WRITE_AND_VERIFY		    0x2E
#define	SC1_VERIFY			    0x2F
#define	SC1_SEARCH_DATA_HIGH		    0x30
#define	SC1_SEARCH_DATA_EQUAL		    0x31
#define	SC1_SEARCH_DATA_LOW		    0x32
#define	SC1_SET_LIMITS			    0x33
#define	SC1_COMPARE			    0x39
#define	SC1_COPY_AND_VERIFY		    0x3A

typedef struct _int24 { u_char _val[3]; } int24;
typedef struct _int21 { u_char :3; u_char _lba2:5; u_char lba1, lba0; } int21;

/* CDB control byte definition */

typedef struct _sccb {
	u_char	sc_vu1:1;
	u_char	sc_vu0:1;
	u_char	sc_resvd:4;
	u_char	sc_flag:1;
	u_char	sc_link:1;
} sccb;

struct scsi_cdb6
{
	u_char  sc_op;
	u_char	sc_lun:3;
	u_char	sc_lbaMSB:5;
	u_char  sc_lbarest[2];
	u_char	sc_xflen[1];
	sccb	sc_control;
};

struct scsi_cdb10
{
	u_char	sc_op;
	u_char  sc_lun:3;
        u_char  sc_resvd1:4;
        u_char  sc_reladr:1;
        u_char  sc_lba[4];
	u_char	sc_resvd2;
	u_char	sc_xflen[2];
	sccb	sc_control;
};

struct scsi_cdb12
{
	u_char	sc_op;
	u_char   sc_lun:3;
        u_char   sc_resvd1:4;
        u_char   sc_reladr:1;
        u_char  sc_lba[4];
	u_char	sc_resvd2;
	u_char	sc_resvd3;
	u_char	sc_resvd4;
	u_char	sc_xflen[2];
	sccb	sc_control;
};
struct scsi_bd
{
	u_char	sbd_density;	/* Density code */
	u_long	sbd_nblocks:24;	/* No of blocks */
	u_char	sbd_resvd1;	/* Reserved */
	u_long	sbd_blklen:24;	/* Block length */
};

struct scsi_modeselect
{
	u_char	sms_resvd1;	/* Reserved */
	u_char	sms_resvd2;	/* Reserved */
	u_char	sms_resvd3:1;	/* Reserved */
	u_char	sms_bufmode:3;	/* Buffered mode code */
	u_char	sms_speed:4;	/* Speed code */
	u_char	sms_bdl;	/* Block descriptor length */
	struct	scsi_bd	sms_bdarr[1];	/* Block descriptor 1 */
	/* Could be more block descriptors here */
};
#define SIZEOF_MODESELECT	12	/* Since the compiler can't figure it out OK */
#define SIZEOF_MODESENSE	12	/* Since the compiler can't figure it out OK */

struct scsi_modesense
{
	u_char	sm_length;	/* Sense data length */
	u_char	sm_mtype;	/* Medium type */
	u_char	sm_wp:1;	/* Write-protect */
	u_char  sm_bufmode:3;	/* Buffered mode */
	u_char	sm_speed:4;	/* Speed code */
	u_char	sm_bdl;		/* Block descriptor length */
	struct	scsi_bd sm_bdarr[1];	/* Block descriptor 1 */
	/* Could be more block descriptors here */
};
#define SCMS_CURRENT 	0x00		/* Report current values */
#define SCMS_CHANGEABLE	0x40		/* Report changeable values	*/
#define SCMS_DEFAULT	0x80		/* Report default values	*/
#define SCMS_SAVED	0xE0		/* Report saved values		*/

#define SCMS_ERP	0x01		/* Error recovery page		*/
#define SCMS_COP	0x02		/* Disconnect/Reconnect page	*/
#define SCMS_DAF	0x03		/* Direct access format page	*/
#define SCMS_RDG	0x04		/* Rigid disk geometries page	*/
#define SCMS_CSH	0x38		/* HW Wren Cache control page   */
#define SCMS_ALL	0x3F		/* All pages			*/

struct scsi_esense 
{
	u_int	es_valid:1;	/* Data valid		*/
	u_int	es_eclass:3;	/* Error class		*/
	u_int	es_ecode:4;	/* Error code		*/
	u_char	es_segno;	/* Segment number	*/
	u_int	es_fmark:1;	/* File mark		*/
	u_int	es_eom:1;	/* End of medium	*/
	u_int	es_ili:1;	/* Illegal length ind   */
	u_int	es_res1:1;	/* Reserved		*/
	u_int	es_key:4;	/* Sense key		*/
	u_char	es_info[4];	/* Info bytes		*/
	u_char	es_addlen;	/* Additional sense len	*/
	u_char	es_add[0];	/* Additional sense     */
};
/* HW definitions moved from xt.c */
struct scsi_mserppage           /* Error recovery page                  */
{
        u_char  erp_ps:1;       /* parameter savable    */
        u_char  erp_rsrv:1;     /* reserved             */
        u_char  erp_pcode:6;    /* Page code            */
        u_char  erp_plen;       /* Page length          */
        u_char  erp_awre:1;     /* Automatic write reallocation enable  */
        u_char  erp_arre:1;     /* Automatic read reallocation enable   */
        u_char  erp_tb:1;       /* Transfer block with error            */
        u_char  erp_rc:1;       /* Read continous                       */
        u_char  erp_eec:1;      /* Enable early correction              */
        u_char  erp_per:1;      /* Post error                           */
        u_char  erp_dte:1;      /* Disable transfer on error            */
        u_char  erp_dcr:1;      /* Disable correction                   */
        u_char  erp_rtcnt;      /* Retry count          */
        u_char  erp_cspan;      /* Correction span      */
        u_char  erp_hoffset;    /* Head offset count    */
        u_char  erp_dsoffset;   /* Data strobe offset count             */
        u_char  erp_rtl;        /* Recovery time limit  */
};
struct scsi_msconnpage          /* Disconnect/Reconnect page            */
{
        u_char  cop_ps:1;       /* parameter savable    */
        u_char  cop_rsrv:1;     /* reserved             */
        u_char  cop_pcode:6;    /* Page code            */
        u_char  cop_plen;       /* Page length          */
        u_char  cop_bfr;        /* Buffer full ratio    */
        u_char  cop_ber;        /* Buffer empty ratio   */
        u_short cop_bil;        /* Bus inactivity limit */
        u_short cop_dtl;        /* Disconnect time limit        */
        u_short cop_ctl;        /* Connect time limit   */
        u_char  cop_resvd[2];   /* Reserved             */
};
struct scsi_msdadfpage          /* Direct access device format parameter page */
{
        u_char  daf_ps:1;       /* parameter savable    */
        u_char  daf_rsrv:1;     /* reserved             */
        u_char  daf_pcode:6;    /* Page code            */
        u_char  daf_plen;       /* Page length          */
        u_short daf_tpz;        /* Tracks per zone      */
        u_short daf_altspz;     /* Alternate sectors per zone   */
        u_short daf_atttpz;     /* Alternate tracks per zone    */
        u_short daf_alttpv;     /* Alternate tracks per volume  */
        u_short daf_spt;        /* Sectors per track            */
        u_short daf_nbps;       /* No of bytes per sector       */
        u_short daf_intrlve;    /* Interleave factor            */
        u_short daf_tskew;      /* Track skew                   */
        u_short daf_cskew;      /* Cylinder skew                */
        u_char  daf_ssec:1;     /* Soft sector format           */
        u_char  daf_hsec:1;     /* Hard sector format           */
        u_char  daf_rmb:1;      /* Removable media              */
        u_char  daf_surf:1;     /* Format by surface            */
        u_char  daf_ins:1;      /* ????                         */
        u_char  daf_resvd:3;    /* reserved                     */
        u_char  daf_resv2[3];   /* Reserved                     */
};

struct scsi_msrigidpage         /* Rigid disk drive geometry page       */
{

        u_char  rdg_ps:1;       /* parameter savable            */
        u_char  rdg_rsrv:1;     /* Reserved                     */
        u_char  rdg_pcode:6;    /* Page code                    */
        u_char  rdg_plen;       /* Page length                  */
        u_char  rdg_maxcyl[3];  /* Maximum no of cylinders      */
        u_char  rdg_hmaxheads;  /* Maximum no of heads          */
        u_char  rdg_scwpc[3];   /* Starting cylinder - write precompensation */
        u_char  rdg_scrwc[3];   /* Starting cylinder - reduced write current */
        u_short rdg_dsr;        /* Drive step rate              */
        u_char  rdg_lzone[3];   /* landing zone cylinder        */
        u_char  rdg_resvd[3];   /* Reserved                     */

};
/* HW Wren cache control */
struct scsi_mscachepage         /* Cache control page */
{
        u_char  csh_ps:1;       /* parameter savable            */
        u_char  csh_pcode:7;    /* Page code                    */
        u_char  csh_plen;       /* Page length                  */
        u_char  csh_rsv1:1;     /* reserved                     */
        u_char  csh_wie:1;      /* Write Index Enable           */
        u_char  csh_rsv2:1;     /* reserved                     */
        u_char  csh_ce:1;       /* Cache Enable                 */
        u_char  csh_cts:4;      /* Cache Table Size             */
        u_char  csh_pt:8;       /* Prefetch Threshold           */
        u_char  csh_max_pre:8;  /* Maximum Prefetch             */
        u_char  csh_max_pre_m:8; /* Maximum Prefetch multiplier */
        u_char  csh_min_pre:8;  /* Miniumum Prefetch             */
        u_char  csh_min_pre_m:8; /* Miniumum Prefetch multiplier */
        u_char  rsv3[8];         /* reserved */

};
struct  scsi_ms /* HW SCSI Mode sense all pages */
{
    struct scsi_modesense       sms_hd;
    struct scsi_mserppage       sms_erp;
    struct scsi_msconnpage      sms_cop;
    struct scsi_msdadfpage      sms_daf;
    struct scsi_msrigidpage     sms_rdg;
    struct scsi_mscachepage     sms_csh
} ;
struct  scsi_msc /* HW SCSI Mode sense/select cache  page */
{
    struct scsi_modesense       sms_hd;
    struct scsi_mscachepage     sms_csh
} ;


#define SCES_KEY_NO_SENSE		0
#define SCES_KEY_RECOVERED_ERROR	1
#define SCES_KEY_NOT_READY		2
#define SCES_KEY_MEDIUM_ERROR		3
#define SCES_KEY_HARDWARE_ERROR		4
#define SCES_KEY_ILLEGAL_REQUEST	5
#define SCES_KEY_UNIT_ATTENTION		6
#define SCES_KEY_DATA_PROTECT		7
#define SCES_KEY_BLANK_CHECK		8
#define SCES_KEY_VENDOR_UNIQUE		9
#define SCES_KEY_COPY_ABORTED		10
#define SCES_KEY_ABORTED_COMMAND	11
#define SCES_KEY_EQUAL			12
#define SCES_KEY_VOLUME_OVERFLOW	13
#define SCES_KEY_MISCOMPARE		14
#define SCES_KEY_RESERVED		15

#define SCTS_GOOD			0x00
#define SCTS_CHECK_CONDITION		0x02
#define	SCTS_CONDITION_MET		0x04
#define SCTS_BUSY			0x08
#define SCTS_INT_GOOD			0x10
#define SCTS_INT_CONDITION_MET		0x14
#define SCTS_RES_CONFL			0x18
#define SCTS_DRV_FAIL			0x7F

struct scsi_inqblk
{
	u_char	sinq_pdt;
#define SCPDT_DASD	0x00
#define SCPDT_SASD	0x01
#define SCPDT_PRINTER	0x02
#define SCPDT_PROCESSOR	0x03
#define SCPDT_WORM	0x04
#define SCPDT_RODA	0x05
#define SCPDT_NO_LUN	0x7F
	u_int	sinq_rmb:1;
	u_int	sinq_dtq:7;
	u_char	sinq_ansiversion;
	u_char	sinq_resvd:4;
	u_char	sinq_rdd:4;
	u_char	sinq_addlen;
	u_char	sinq_rqsenselen;
	u_char	sinq_resvd2[2];
	u_char	sinq_vid[8];
	u_char	sinq_pid[16];
	u_char	sinq_fwrev[4];
};
