#include "xt.h"
#if NXT > 0 
#define 	DEBUG	1 /* HW */
/*
 * Copyright (c) 1988, Christer Bernerus, Chalmers University of Technology,
 *                     Gothenburg, SWEDEN
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms as is or as a source
 * for derivative work are permitted provided that the above copyright 
 * notice and this paragraph are  duplicated in all such forms 
 * and that any documentation, advertising materials,
 * and other materials related to such distribution
 * and use acknowledge that the software was developed
 * at Chalmers University of Technology, SWEDEN.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * Maxtor XT-4380S SCSI Disk driver
 *
 * $Source: /afs/sys/rt-ibm4.3-r3/usr/sys/caio/xt.c,v $
 * $Revision: 2.6 $ $Date: 89/05/23 12:52:08 $
 * $Author: bernerus $ $State: Exp $
 *
 * $Log:	xt.c,v $
 * Revision 2.6  89/05/23  12:52:08  bernerus
 * The code for starting disks didn't work as expected. Now it does.
 * 
 * Revision 2.5  89/05/23  09:26:57  bernerus
 * Now handles sector sizes different from 512 (untested).
 * Spin up/down and automatic spin up capability added.
 * Probe routine now uses SCSI Inquiry instead of Test Unit Ready
 * to determine existence of a disk. It also only acctpts DASD types.
 * The slave routine now attempts to start a drive that isn't ready
 * before doing mode sensing.
 * 
 * Revision 2.4  89/02/24  13:09:42  bernerus
 * Changed the copyright notice.
 * 
 * Revision 2.3  89/02/08  13:07:11  bernerus
 * Partial sectors queue was horribly large, now a lot smaller.
 * We now do a SCSI Inquiry to get the disk type, and a Mode sense to
 * get the disk's geometries, thus making the driver less device dependent.
 * 
 * Revision 2.2  89/01/30  08:01:10  bernerus
 * Added minidisk support.
 * 
 * Revision 2.1  89/01/18  09:18:31  bernerus
 * Added some minidisk support.
 * 
 * Revision 2.0  89/01/16  11:11:58  bernerus
 * Added copyright notice.
 * Changed names xu->xt, xt->xtc.
 * 
 *
 */
#include "../h/param.h"
#include "../h/vm.h"
#include "../h/buf.h"
#include "../h/time.h"
#include "../h/proc.h"
#include "../h/errno.h"
/* #include "../machine/pte.h" HW */
#include "../machine/io.h"
#include "../h/kernel.h"
#include "../h/systm.h"
#include "../h/cmap.h"
#include "../h/dk.h"
#include "../h/ioctl.h"
#include "../machine/mmu.h"
#include "../machine/dkio.h"

#include "../machineio/ioccvar.h"
#include "../machineio/scsi.h"
#include "../machineio/xhavar.h"
#include "../machine/debug.h"
#include "../machineio/scsiconfig.h"

#define MAXNPART 8
#define MAXQPERUNIT 2
#define MAXNBPS	        4096     

/* THIS SHOULD BE READ OFF THE PACK, PER DRIVE */
/* 4380S Geometries :
	36	Sectors/Track 512 bytes/sector
	18360   tracks totally and
	540     sectors/cylinder
*/
struct size {
	struct partentry {
	    daddr_t nblocks;
	    int     blkoff;
	}
	partition[MAXNPART];
};

struct size
def_sizes = {
        15884,  0,              /* A=cyl 0 thru 29 */
        33480,  16200,          /* B=cyl 30 thru 91 */
        621180,   0,            /* C=cyl 0 thru 1217 */
        16740,  49680,          /* D=cyl 92 thru 122 */
	0,	0,		/* E = nothing */
	0,	0,		/* F = nothing */
	82620, 538380,		/* G = cyl 997 thru 1149 */
        471960,  66420,		/* H = cyl 123 thru  996 */
};

struct size
xt43_sizes = {
        15884,  0,              /* A=cyl 0 thru 29 */
        33480,  16200,          /* B=cyl 30 thru 91 */
        621180,   0,            /* C=cyl 0 thru 1217 */
        16740,  49680,          /* D=cyl 92 thru 122 */
	0,	0,		/* E = nothing */
	0,	0,		/* F = nothing */
	82620, 538380,		/* G = cyl 997 thru 1149 */
        471960,  66420,		/* H = cyl 123 thru  996 */
};
struct size
xt42_sizes = {
        16236,  0,              /* A=cyl 0 thru 40 */
        33660,  16236,          /* B=cyl 41 thru 125 */
        484704,   0,            /* C=cyl 0 thru 1223 */
        16236,  49896,          /* D=cyl 125 thru 165 */
	0,	0,		/* E = nothing */
	0,	0,		/* F = nothing */
	82368, 373032,		/* G = cyl 941 thru 1148 */
};
struct size
xt41_sizes = {
        16128,  0,              /* A=cyl 0 thru 63 */
        33516,  16128,          /* B=cyl 64 thru 196 */
        308448,   0,            /* C=cyl 0 thru 1223 */
        16128,  49680,          /* D=cyl 197 thru 260 */
	0,	0,		/* E = nothing */
	0,	0,		/* F = nothing */
	82404, 538380,		/* G = cyl 821 thru 1147 */
        141120,  66420,		/* H = cyl 261 thru  820 */

};
/* HW added Wren default partitions for wren (minidisk will override ) */
/* Note these partition sizes are not clever, just somewhat compatibel
   with history .... */
/* Wren V Geometries :
        1541    Cylinders
        15      Tracks
        51      Sectors/Track 512 bytes/sector
        23115   tracks totally and
        765     sectors/cylinder
*/
struct size
wv15_sizes = {
#if     CS_GENERIC
        (  55-  1)*51*15,   1*15*51,  /* A=cyl   1 thru   54 */
                      0,    0*15*51,  /* B */
        (1540-  0)*51*15,   0*15*51,  /* C=cyl   0 thru 1540 */
        ( 162- 55)*51*15,  55*15*51,  /* D=cyl  55 thru  161 */
        ( 265-162)*51*15, 162*15*51,  /* E=cyl 162 thru  264 */
                      0,    0*15*51,  /* F */
        (1534-  1)*51*15,   1*15*51,  /* G=cyl   1 thru 1534 */
        (1534-265)*51*15, 265*15*51,  /* H=cyl 265 thru 1534 */
#else   CS_GENERIC
/*
------------------------------------------------------------------------------
| a |  b  | d |    h     |                      g                            |
------------------------------------------------------------------------------
|                                c                                           |
------------------------------------------------------------------------------
*/
        (  29-  1)*51*15,   1*15*51, /* A=cyl 1 thru 29 */
        (91-   30)*51*15,  30*15*51, /* B=cyl 30 thru 91 */
        (1540-  1)*51*15,   1*15*51, /* C=cyl 1 thru 1540 */
        ( 122- 92)*51*15,  55*15*51, /* D=cyl 92 thru 122 */
                       0,   0*15*51, /* E = nothing */
                       0,   0*15*51, /* F = nothing */
        (1540-997)*51*15, 997*15*51, /* G = cyl 997 thru 1540 */
        (966- 123)*51*15, 123*15*51, /* H = cyl 123 thru  996 */
#endif  CS_GENERIC
};
struct size
wvi15_sizes = {
/*
------------------------------------------------------------------------------
| a |  b  | d |    h     |                      g                            |
------------------------------------------------------------------------------
|                                c                                           |
------------------------------------------------------------------------------
*/
        (  29-  1)*53*15,   1*15*53, /* A=cyl 1 thru 29 */
        (91-   30)*53*15,  30*15*53, /* B=cyl 30 thru 91 */
        (1627-  1)*53*15,   1*15*53, /* C=cyl 1 thru 1627 */
        ( 122- 92)*53*15,  55*15*53, /* D=cyl 92 thru 122 */
                       0,   0*15*53, /* E = nothing */
                       0,   0*15*53, /* F = nothing */
        (1627-997)*53*15, 997*15*53, /* G = cyl 997 thru 1627 */
        (966- 123)*53*15, 123*15*53, /* H = cyl 123 thru  996 */
};
/* The structure below is copied into the softc structure on attach */
struct  xtst {
	char	inqname[17];
        short   nsect;
        short   ntrak;
        short   nspc;
        short   ncyl;
	short	nbps;
        struct  size *defsizes;
	char	typename[8];
	short	flags;
#define XTFL_LINK	1	/* SCSI Command linking supported */
#define XTFL_NO_LINK	0	/* SCSI Command linking NOT supported HW */
} xtst[] = {
/* inqname,nsect,ntrak, nspc,ncyl, nbps, def.sizes, typename    */
{"Default ",  34,  15, 34*15,1218, 512, &def_sizes,  "default", 0},
{"XT-4380S",  34,  15, 34*15,1218, 512, &xt43_sizes, "xt4380s", XTFL_LINK},
{"XT-4280S",  34,  11, 34*11,1218, 512, &xt42_sizes, "xt4280s", XTFL_LINK},
{"XT-4170S",  34,   7, 34* 7,1218, 512, &xt41_sizes, "xt4170s", XTFL_LINK},
{"94181-15",  51,  15, 51*15,1541, 512, &wv15_sizes, "wv15", XTFL_NO_LINK},
{"94191-15",  53,  15, 53*15,1627, 512, &wvi15_sizes,"wvi15",XTFL_NO_LINK},
{"scsi0240"},
{"scsi0260"},
{"scsi0440"},
{"scsi0460"},
};

struct  xt_softc {
	int	sc_busy:1;		/* Set when BUSY is detected, reset otherwise; */
	int	sc_partitioninited:1;	/* Set when partition init is done */
	int	sc_nomap:1;	        /* Don't map or check partitions */
	int	sc_cinited:1;		/* True if c part'n is initialized */
	int	sc_spundown:1;		/* Drive has been spun down */
	int	sc_autospinup:1;	/* Spin up automatically	*/
	int	sc_busyretries;		/* No of retries for busy drive */
	int	sc_queuedepth;		/* Current queue depth */
        union _cdb                      /* Gen purp SCSI CDB struct */
        {
                struct scsi_cdb6 cdb6blk;
                struct scsi_cdb10 cdb10blk;
                struct scsi_cdb12 cdb12blk;
        } sc_cdb;
	struct partsec
	{
		char	ps_phys:1;	/* Xfer had PHYS flag set in buffer */
		struct buf *	ps_bp;	/* Associated buffer zero if unused */
		caddr_t	ps_addr;	/* Original b_addr in buf */
		long	ps_bcount;	/* Original bcount in buf */
		char	ps_space[MAXNBPS];	/* Data space */
	} sc_part[MAXQPERUNIT];
	struct xtst sc_xtst;
	struct size sc_size;		/* Actual partitioning */
} xt_softc[NXT];

/* HW mode sense pages struct defs moved to scsi.h */

struct scsi_cdb6 * getsc6buf();
struct scsi_cdb10 * getsc10buf();
void fillsc6buf(),
     fillsc10buf();
unsigned	xtget3();

struct partsec *
xtgetps(xi,bp)
struct xha_device *xi;
struct buf *bp;
{
        struct xt_softc *sc = &xt_softc[xi->xi_unit];
	
	struct partsec *ps;

	for(ps = sc->sc_part; ps < &(sc->sc_part[MAXQPERUNIT]); ps++)
		if(ps->ps_bp == bp) return ps;
	return NULL;
}
#define PARTIAL(bp)	(((bp)->b_un.b_addr > (caddr_t) xt_softc) && ((bp)->b_un.b_addr < (caddr_t) &xt_softc[NXT]))

/*
 * Since we want to have more than one transfer on each unit going on
 * "simultaneously", we have to track retry counts in each buffer instead
 * of doing it in the device's buffer queue header. This means that a suitable
 * variable must be found that isn't needed during I/O. The only one we cound
 * find is b_error. Unfortunately this variable isn't cleared for transfers
 * coming from the swap system so we clear it here in the device driver since
 * we didn't want to touch the swap system's code.
 * In ordinary device drivers the retry counting is done in the b_resid variable
 * in the buffer queue head (which is just a header). The definition for this is
 * in ../h/buf.h which defines b_errcnt to be b_resid. The b_resid variable
 * can't be used in the transfer buffers since we use it for indicating cylinder
 * number to disksort(), and thus can't easily be changed.
 * When a transfer is completed, b_error will regain its usual function.
 */ 
#define b_cylin b_resid
#undef b_errcnt
#define b_errcnt b_error

#define dkblock(bp)	((bp)->b_blkno)


struct size xtsizes[NXT];


/* Bits in minor device */

#define XTUNIT(dev)	XHAUNIT(dev)
#define XTCUNIT(dev)	XHAMIN_TARGET(dev)
caddr_t real_buf_addr();

struct xha_device *xtcinfo[NXTC];
struct xha_slave *xtinfo[NXT];
struct buf rxtbuf[NXTC];


short xttypes [] = { 1,0 };	/* FIX THIS LATER */

#ifdef DEBUG
int     xtdebug		= 0;
int     xtustartdebug	= 0;
int     xtstartdebug	= 0;
int     xtintdebug	= 0;
int	xtstrategydebug = 0;
int	xtminphysdebug  = 0;
int	xtpartial	= 0;
int	xtgetpartdebug  = 0;
int	xtioctldebug    = 0;
#endif
int	xtdisksort 	= 1;	/* Use disksort */
int	xtdolinks	= 0;	/* Use SCSI linked commands */


int xtprobe(),xtslave(),xtustart(),xtint(),xtttl();
int xtminphys();
int xtinquiry(struct xha_device *,short , struct scsi_inqblk *);
void xtattach(struct xha_slave *);
struct scsi_cdb6 *xtstart();

struct xha_driver xtcdriver =
{
	xtprobe, xtslave, 
	(int (*)()) xtattach, 
	xtustart,
	(int (*)()) xtstart,
	xtint,
	xttypes,"xtc","xt",xtcinfo,xtinfo,
	xtttl,
	EARLY_INT,
};

struct	scsi_ms  xtmsbuf;

xtprobe(reg,xi)
register caddr_t	    reg;
struct xha_device *xi;
{
    struct scsi_inqblk 	iqbuf;
    struct scsi_inqblk 	*iq = &iqbuf;
    short lun;

    for(lun=0;lun<8;lun++)
    {
	if(xtinquiry(xi,lun,iq) != 0)
	{
	    return PROBE_BAD;
	}
	if(iq->sinq_pdt == SCPDT_DASD)
	{
	    return PROBE_NOINT;
	}
    }	
    return PROBE_BAD;
}

void 
xtattach(struct xha_slave *xs)
{

    if(xs->xs_unit > NXTC)
    {
	printf("xt%d: Not configured\n",xs->xs_unit);
    }
}

xtstartstop(struct xha_device *xi,short start)
{
	static struct scsi_cdb6 sc6blk;
	register struct scsi_cdb6 *sc6buf = &sc6blk;

	int lun = 0; /* For now */

	fillsc6buf(sc6buf,0,SC0_START_STOP_UNIT,start);

	if(((*xi->xi_hadriver->xhc_doscsi)(xi->xi_ctlr,xi->xi_drive,lun,
	     sc6buf,0,0,start ? B_WANTED : 0)) == 
	     ((XHI_CMPL << 8) + SCTS_GOOD))
	{
		return 0;
	}
	return -1;
}
/* HW hacked IOCTL routine to turn on/off wren V VI caching by mode select */
xtmodeselect(struct xha_device *xi,caddr_t data)
{
        static struct scsi_cdb6 sc6blk;
        register struct scsi_cdb6 *sc6buf = &sc6blk;
	int ret;
	struct scsi_msc     *mscp = (struct scsi_msc * ) data ;
		/* this is a  kludge! oriented to support mode select for 
		   cache page */

        DEBUGF(xtioctldebug,printf("xtmodeselect entered \n"););
        int lun = 0; /* For now */

        fillsc6buf(sc6buf,0,SC0_MODE_SELECT,sizeof *mscp );
	sc6buf->sc_lbaMSB = 0x1 ; /* turn on SMP Save Mode Page bit */
	DEBUGF(xtioctldebug,printf("xtmodeselect before doscsi \n"););
        ret = (*xi->xi_hadriver->xhc_doscsi)
		(xi->xi_ctlr,
		xi->xi_drive,
		lun, 
		sc6buf,
		mscp,
		sizeof *mscp ,
		B_WRITE | B_WANTED ) ;
	DEBUGF(xtioctldebug,printf("xtmodeselect doscsi returned %d\n",ret););
	if ( ret ==
             ((XHI_CMPL << 8) + SCTS_GOOD))
        {
                return 0;
        }
        return -1;
}
/* HW hacked IOCTL routine to do dynamic mode sense command*/
xtmodesense(struct xha_device *xi,caddr_t data)
{
        static struct scsi_cdb6 sc6blk;
        register struct scsi_cdb6 *sc6buf = &sc6blk;
	struct scsi_ms      *msp = (struct scsi_ms * ) data ;
	int ret;
	DEBUGF(xtioctldebug,printf("xtmodesense entered \n"););
	fillsc6buf(sc6buf,0,SC0_MODE_SENSE,sizeof *msp);
	sc6buf->sc_lbarest[0] = SCMS_CURRENT | SCMS_ALL; 
	DEBUGF(xtioctldebug,printf("xtmodesense before doscsi \n"););

	ret =	(*xi->xi_hadriver->xhc_doscsi)	(
		xi->xi_ctlr,
		xi->xi_drive,
		0,
		sc6buf,
		msp,
		sizeof *msp,
		B_READ | B_WANTED);
	DEBUGF(xtioctldebug,printf("xtmodesense doscsi returned %d\n",ret););
	if(ret == ((XHI_CMPL << 8) + SCTS_GOOD)) {
		return 0;
	}
	return -1;
}

/* Tell controller routine expected max lifetime of */
/* a command (time in seconds)			    */
xtttl(struct scsi_cdb6 *com)
{
	switch(com->sc_op)
	{
	    case SC0_START_STOP_UNIT:		return 60;
	    default:			return 1;
	}
}

xttur(struct xha_device *xi,short lun)
{
	static struct scsi_cdb6 sc6blk;
	register struct scsi_cdb6 *sc6buf = &sc6blk;

	fillsc6buf(sc6buf,lun,SC0_TEST_UNIT_READY,0);

	/*
	 * Now do the test
	 */
	if(((*xi->xi_hadriver->xhc_doscsi)(xi->xi_ctlr,xi->xi_drive,lun,
	     sc6buf,0,0,B_READ | B_WANTED)) == 
	     ((XHI_CMPL << 8) + SCTS_GOOD))
	{
		return 0;
	}
	return -1;
}

xtinquiry(struct xha_device *xi,short lun, struct scsi_inqblk *iq)
{
	static struct scsi_cdb6 sc6blk;
	register struct scsi_cdb6 *sc6buf = &sc6blk;

	fillsc6buf(sc6buf,lun,SC0_INQUIRY,sizeof *iq);
        iq->sinq_pdt = SCPDT_NO_LUN;

	/*
	 * Now do the Inquiry 
	 */
	if(((*xi->xi_hadriver->xhc_doscsi)(xi->xi_ctlr,xi->xi_drive,lun,
	     sc6buf,iq,sizeof *iq,B_READ | B_WANTED)) == 
	     ((XHI_CMPL << 8) + SCTS_GOOD))
	{
		return 0;
	}
	return -1;
}

xtslave(xs,addr)
struct xha_slave *xs;
caddr_t addr;
{
    struct xha_device	*xi = xs->xs_device;
    struct xt_softc	*sc = &xt_softc[xi->xi_unit];
    char		*typep;
    struct scsi_inqblk 	iqbuf;
    struct scsi_inqblk 	*iq = &iqbuf;
    struct scsi_cdb6	*sc6buf;
    static char inqname[] = "XT-4380S";
    struct scsi_ms	*msp = &xtmsbuf;
    static char		vid[sizeof(iq->sinq_vid)+1];
    static char		pid[sizeof(iq->sinq_pid)+1];
    static char		fwrev[sizeof(iq->sinq_fwrev)+1];
    int	type,scts;

    if(xs->xs_driver != &xtcdriver) 
    {
	return 0;
    }

    if(xi)
    {
	if(xtinquiry(xi,xs->xs_slave,iq) == 0)
	{
		if(iq->sinq_pdt != SCPDT_DASD)
		{
		    return 0;
		}

		xttur(xi,xs->xs_slave); /* Clear any unit attn condition */

		if(xttur(xi,xs->xs_slave) < 0 ) /* Isn't ready ? */
		{
			/* NO, start it first */
			printf("%s%d: Attempting drive start\n",
			xi->xi_driver->xd_sname,xi->xi_unit);
			if(xtstartstop(xi,1) < 0)
			{
				printf("%s%d: Couldn't start drive\n",
				    xi->xi_driver->xd_sname,xi->xi_unit);
				return 0;
			}
			xtinquiry(xi,xs->xs_slave,iq); /* Rerun the inquiry */
		}

		xtidcpy(iq->sinq_vid,vid,sizeof vid -1);
		xtidcpy(iq->sinq_pid,pid,sizeof pid -1);
		xtidcpy(iq->sinq_fwrev,fwrev,sizeof fwrev -1);

		type = xtgettype(xs,&typep,pid);

		if(type < 0)
		{
		    type = 0;
/*		    printf("%s%d: Unknown disk type: %s using %s\n", */
/*			xi->xi_driver->xd_sname,xi->xi_unit,typep,   */
/*			xtst[type].typename);  			     */
		}

		xi->xi_type = (u_long) type;
		printf("%s%d: Disk type %s name=%s\n",
			xi->xi_driver->xd_sname,xi->xi_unit,
			xtst[xi->xi_type].typename,
			typep);
		/* 
		 * Fill in default values for disk info. 
		 */
		bcopy(&(xtst[xi->xi_type]),&(sc->sc_xtst),sizeof(struct xtst));
		/* 
		 * Clear size structure.
		 */
		bzero(&(sc->sc_size),sizeof(struct size));
		sc->sc_partitioninited = 0; /* Mark partition init to be done */
		    
		printf("%s%d:%s %s Rev. %s ANSI V.%d\n",
			xi->xi_driver->xd_sname,xi->xi_unit,
			vid,pid,fwrev,iq->sinq_ansiversion);

		sc6buf = getsc6buf(xs->xs_unit);
		fillsc6buf(sc6buf,xs->xs_slave,SC0_MODE_SENSE,sizeof xtmsbuf);
		sc6buf->sc_lbarest[0] = SCMS_CURRENT | SCMS_ALL;

		if((scts=(*xi->xi_hadriver->xhc_doscsi)(xi->xi_ctlr,
		      xi->xi_drive, xs->xs_slave, sc6buf,msp,sizeof xtmsbuf,
		      B_READ | B_WANTED)) == 
		      ((XHI_CMPL << 8) + SCTS_GOOD))
		{
		    strncpy(sc->sc_xtst.inqname,pid,
				sizeof sc->sc_xtst.inqname);
		    if(msp->sms_daf.daf_tpz == 0)
		    {
			msp->sms_daf.daf_tpz = 1;
		    }
		    sc->sc_xtst.nsect = msp->sms_daf.daf_spt -
				        (msp->sms_daf.daf_altspz/
				         msp->sms_daf.daf_tpz);
		    sc->sc_xtst.ntrak = msp->sms_rdg.rdg_hmaxheads;

		    if(sc->sc_xtst.ntrak == 0)
		    {
			sc->sc_xtst.ntrak = 1;
		    }

		    sc->sc_xtst.nspc  = sc->sc_xtst.nsect * sc->sc_xtst.ntrak;
		    sc->sc_xtst.ncyl  = xtget3(msp->sms_rdg.rdg_maxcyl) -
				        (msp->sms_daf.daf_alttpv/
				        sc->sc_xtst.ntrak) - 3;
		    sc->sc_xtst.nbps  = msp->sms_daf.daf_nbps;
			
 		    sc->sc_size.partition[2].nblocks = 
			sc->sc_xtst.ncyl * sc->sc_xtst.nspc;
 		    sc->sc_size.partition[2].blkoff = 0;
		    sc->sc_cinited = 1;

		    printf("nsect=%d, ntrak=%d, nspc=%d, ncyl=%d, nbps=%d, csize=%d\n",
			sc->sc_xtst.nsect, sc->sc_xtst.ntrak, sc->sc_xtst.nspc,
			sc->sc_xtst.ncyl, sc->sc_xtst.nbps,
			sc->sc_size.partition[2].nblocks);
/* HW try to print Wren cach status */
		printf("Wren Caching enabled : %s\n",
			msp->sms_csh.csh_ce ? "Yes" : "No" );

		    sc->sc_partitioninited = 0;  /* Mark partition init to be done */
		    return 1;
		}
		else
		{
			printf("%s%d: Mode sense failed, scts=%d\n",
			xi->xi_driver->xd_sname,xi->xi_unit,scts);
		}
	}
	else
	{
	    return 0;
	}
    }
    else
    {
	printf("xtslave: xs->xs_device not loaded, fix autoconfig\n");
	return 0;
    }
    return 1;
}

xtgettype(xs,typep,inqname)
struct xha_slave *xs;
char **typep;
char *inqname;
{
    /* This should be read from disk */
    struct xtst *st;
    
    *typep = inqname;

    for (st = xtst; st < xtst + (sizeof xtst/sizeof (struct xtst)); st++)
    {
	if(strncmp(st->inqname,inqname,sizeof (st->inqname)) == 0)
	{
	    return (st - xtst);
	}
    }
    return -1;
}

xtsize(dev)
	dev_t				dev;
{
	register int			un = XTCUNIT(dev);
	register int			pa = XHAMIN_PART(dev);
	register struct xha_device	*xi;
	struct xt_softc 		*sc;

	if (un >= NXTC || pa < 0 || pa >= MAXNPART ||
	    (xi = xtcinfo[un]) == 0 || xi->xi_alive == 0)
		return(-1);
        sc = &xt_softc[xi->xi_unit];
	return(sc->sc_size.partition[pa].nblocks);
}


xtopen(dev,flag)
dev_t dev;
int flag;
{

        register int xtunit = XTCUNIT(dev);
        register struct xha_device *xi;
        struct xt_softc *sc;

        if (xtunit >= NXT || (xi = xtcinfo[xtunit]) == 0 || xi->xi_alive == 0)
                return (ENXIO);
        sc = &xt_softc[xi->xi_unit];
	if(sc->sc_spundown && sc->sc_autospinup)
	{
		xtstartstop(xi,1);
	}
        return (0);

}

xtclose(dev,flag)
dev_t dev;
int flag;
{
}

xtpartitioninit(struct xha_device *xi, struct xt_softc *sc, dev_t dev)
{
	sc->sc_partitioninited = 1; /* Must do this first or else an 
			        * infinite recursion will begin */
	sc->sc_nomap = 1;      /* Open whole disk, no mapping */

	if(xtgetpartitions(xi,dev) == 0)
	{
		/* No minidisk info, use default */
		bcopy(sc->sc_xtst.defsizes,
		      &sc->sc_size,sizeof( struct size ));
	}
	else
	{
	    /* 
	     * NOTE! The minidisk utilities doesn't set up the c partition
	     * If this isn't initializes to sane things, lots of other things
	     * will break, so we always copy the default c partition from
	     * the standard table
	     */
	    if(sc->sc_cinited == 0)
	    {
		bcopy(&(sc->sc_xtst.defsizes->partition[2]),
		      &(sc->sc_size.partition[2]),
		      sizeof (struct partentry));
		sc->sc_cinited = 1;
	    }
	}
	sc->sc_nomap = 0;      /* Map again */
}

xtstrategy(bp)
register struct buf *bp;
{
    register struct xha_device *xi;
    register struct xtst *st;
    register struct xt_softc *sc;
    register int xtunit;
    int pa = XHAMIN_PART(bp->b_dev);
    long bn, sz;
    int s;
    register struct xhac_driver *xc;
    struct buf *dp;



    sz = (bp->b_bcount+511) >> 9;
    xtunit = XTCUNIT(bp->b_dev);
    if (xtunit >= NXTC) 
    {
	bp->b_error = ENXIO;
	DEBUGF(xtstrategydebug,
	    printf("Xtstrategy(BAD): bp=0x%X, sz=%d, blkno=%d\n",
		bp,sz,bp->b_blkno););
	bp->b_flags |= B_ERROR;
	iodone(bp);
	return;
    }
    xi = xtcinfo[xtunit];
    if (xi == 0 || xi->xi_alive == 0) 
    {
	bp->b_error = ENXIO;
	DEBUGF(xtstrategydebug,
	    printf("Xtstrategy(BAD): bp=0x%X, sz=%d, blkno=%d\n",
		bp,sz,bp->b_blkno););
	bp->b_flags |= B_ERROR;
	iodone(bp);
	return;
    }
   
    xc  = xi->xi_hadriver;
    sc = &xt_softc[xi->xi_unit];
    st = &(sc->sc_xtst);
    if(sc->sc_spundown && sc->sc_autospinup)
    {
	xtstartstop(xi,1);
    }
    if(sc->sc_partitioninited == 0)
    {
	xtpartitioninit(xi,sc,bp->b_dev);
    }
    DEBUGF(xtstrategydebug,printf("Xtstrategy: bp=0x%X, sz=%d, blkno=%d\n",
	    bp,sz,bp->b_blkno););
    bn = bp->b_blkno;
    if ((bn < 0) || ((sc->sc_nomap == 0) && 
		     (bn+sz > sc->sc_size.partition[pa].nblocks)))
    {
	if (bn == sc->sc_size.partition[pa].nblocks)
	{
	    bp->b_resid = bp->b_bcount;
	    iodone(bp);
	    return;
	}
	bp->b_error = EINVAL;
	DEBUGF(xtstrategydebug,
	    printf("Xtstrategy(BAD): bp=0x%X, sz=%d, blkno=%d\n",
		bp,sz,bp->b_blkno););
	bp->b_flags |= B_ERROR;
	iodone(bp);
	return;
    }

#ifdef DEBUG
    for(dp=&xi->xi_tab;dp;dp = dp->b_actf)
	if(dp == bp)
		printf("Xtstrategy: bp=%X already in Q\n");
#endif

    s = splbio();
    if(!xtdisksort)
    {
	bp->av_forw = NULL;
	dp = &xi->xi_tab;
	if (dp->b_actf == NULL)
	    dp->b_actf = bp;
	else
	    dp->b_actl->av_forw = bp;
	dp->b_actl = bp;
	bp->b_errcnt = 0;
    }
    else
    {
	if(sc->sc_nomap)
	{
	    bp->b_cylin = bn / st->nspc;
	}
	else
	{
	    bp->b_cylin =  (bn + sc->sc_size.partition[pa].blkoff)/ st->nspc;
	}
	bp->b_errcnt = 0;
	disksort(&xi->xi_tab, bp);
    }
    splx(s);
    (*xc->xhc_ustart)(xi);
}

xtustart(xi)
struct xha_device *xi;
{
    DEBUGF(xtustartdebug,printf("Xtustart: xi=0x%X\n",xi););
    struct xt_softc *sc = &xt_softc[xi->xi_unit];

    if(xtdolinks && (sc->sc_xtst.flags & XTFL_LINK) )
	return XHU_DOLINKS;
    else
	return XHU_DODATA;
}

static struct scsi_cdb10 *
getsc10buf(unit)
{
    struct scsi_cdb10 *scp;
    scp = &xt_softc[unit].sc_cdb.cdb10blk;
    return scp;
}

static void
fillsc10buf(scp,lun,op,xflen)
struct scsi_cdb10 *scp;
{
    bzero(scp,10);
    scp->sc_op = op;
    scp->sc_lun = lun;
    scp->sc_xflen[0]=xflen >> 8 & 0xFF;
    scp->sc_xflen[1]=xflen      & 0xFF;
}

static
struct scsi_cdb6 *
getsc6buf(unit)
{
    struct scsi_cdb6 *scp;

    scp = &xt_softc[unit].sc_cdb.cdb6blk;
    return scp;
}

static void
fillsc6buf(scp,lun,op,xflen)
struct scsi_cdb6 *scp;
{

    bzero(scp,6);
    scp->sc_op = op;
    scp->sc_lun = lun;
    scp->sc_xflen[0]=xflen;
}

void
xtpartsetup(xi,bp,bn,sz)
register struct xha_device *xi;
register struct buf *bp;
int	bn,sz;
{
	register struct partsec *ps;
	int	s;
	struct	xt_softc *sc = &xt_softc[xi->xi_unit];

	DEBUGF(xtpartial,printf("Xtstart(Partial): xi=0x%X, bn=%d, sz=%d, bcount=%d\n",
		xi,bn,sz,bp->b_bcount););
	s = splbio();
	ps = xtgetps(xi,NULL); /* NULL will get us a free entry */
	if(ps == NULL)
	    panic("Xtstart: partial sector init failed");
	
	/* Save interesting facts in buf structure */
	ps->ps_bp = bp;
	ps->ps_addr = bp->b_un.b_addr;
	ps->ps_bcount = bp->b_bcount;
	ps->ps_phys = bp->b_flags & B_PHYS;
	splx(s);
	/* Now let's modify the buf struct a bit */
	bp->b_un.b_addr = ps->ps_space;	/* link i/o from buffer to this space */
	bp->b_bcount = sc->sc_xtst.nbps;
	bp->b_flags &= ~B_PHYS;
	
	if(!(bp->b_flags & B_READ))
	    xtfillpart(ps);
}

struct scsi_cdb6 *
xtstart(xi,bp,combuf)
struct xha_device *xi;
register struct buf *bp;
struct scsi_cdb12 *combuf;
{
    register struct scsi_cdb6 *sc6buf;
    register struct scsi_cdb10 *sc10buf;
    register int bn;
    register int sz;
    register struct xt_softc *sc = &xt_softc[xi->xi_unit];
    dev_t 	dev;

    dev = bp->b_dev;
    if(sc->sc_nomap)
    {
	bn = dkblock(bp);
    }
    else
    {
	bn = dkblock(bp) + sc->sc_size.partition[XHAMIN_PART(dev)].blkoff;
    }
    sz = (bp->b_bcount + (sc->sc_xtst.nbps-1)) / sc->sc_xtst.nbps;	/* Compute no of blks */
    /*
     * Code here to handle partial sector transfers 
     */
    if(bp->b_bcount < sc->sc_xtst.nbps)
    {
	xtpartsetup(xi,bp,bn,sz);
    }

    DEBUGF(xtstartdebug,printf("Xtstart: xi=0x%X, bn=%d, sz=%d\n",xi,bn,sz););
    if(bn < 0x1FFFFF ) 
    {
	if(combuf)
	    sc6buf = (struct scsi_cdb6 *) combuf;
	else
	    sc6buf = getsc6buf(XTUNIT(dev));

	fillsc6buf(sc6buf,XHAMIN_LUN(dev),bp->b_flags & B_READ ? SC0_READ:SC0_WRITE,sz);
	sc6buf->sc_lbaMSB     = bn >> 16 & 0x1F;
	sc6buf->sc_lbarest[0] = bn >> 8  & 0xFF;
	sc6buf->sc_lbarest[1] = bn       & 0xFF;
	DEBUGF(xtdebug,printf("Xtstart: Return SC0: bn=%d, sz=%d\n",bn,sz););
	return sc6buf;
    }
    else
    {
	if(combuf)
	    sc10buf = (struct scsi_cdb10 *) combuf;
	else
	    sc10buf = getsc10buf(XTUNIT(dev));

	fillsc10buf(sc10buf,XHAMIN_LUN(dev),bp->b_flags & B_READ ? SC1_READ:SC1_WRITE,sz);
	sc10buf->sc_lba[0]    = bn >> 24 & 0xFF;
	sc10buf->sc_lba[1]    = bn >> 16 & 0xFF;
	sc10buf->sc_lba[2]    = bn >>  8 & 0xFF;
	sc10buf->sc_lba[3]    = bn       & 0xFF;
	DEBUGF(xtstartdebug,printf("Xtstart: Return SC1: bn=%d, bno=%d\n",bn,sz););
	return ((struct scsi_cdb6 *) sc10buf);
    }
}

xtfillpart(ps) struct partsec *ps;
{
    DEBUGF(xtdebug,printf("Xtfillpart: filling %d bytes\n",ps->ps_bcount););
    if(!ps->ps_phys)
    {
	DEBUGF(xtdebug,printf("bcopy(%X,%X,%d) (COOKED)\n",ps->ps_addr,ps->ps_space,ps->ps_bcount););
	bcopy(ps->ps_addr,ps->ps_space,ps->ps_bcount);
    }
    else
    {
	register caddr_t        from, from2;
	register int            s, v;
	register u_int          n;

	ps->ps_bp->b_flags |= B_PHYS;
	from = real_buf_addr(ps->ps_bp, ps->ps_addr);
	n = PGOFSET + 1 - ((u_int)from & PGOFSET);
	if (n >= ps->ps_bcount) 
	{
	    DEBUGF(xtdebug,printf("bcopy(%X,%X,%d) (RAW1)\n",from,ps->ps_space,ps->ps_bcount););
	    s = splhigh();
	    GET_VR0(v);
	    bcopy(from, ps->ps_space, ps->ps_bcount);
	} 
	else 
	{
	    from2 = real_buf_addr(ps->ps_bp, ps->ps_addr + n);
	    DEBUGF(xtdebug,printf("bcopy(%X,%X,%d) (RAW2)\n",from,ps->ps_space,ps->ps_bcount););
	    DEBUGF(xtdebug,printf("bcopy(%X,%X,%d) (RAW2)\n",from2,ps->ps_space + n,ps->ps_bcount - n););
	    s = splhigh();
	    GET_VR0(v);
	    bcopy(from, ps->ps_space, n);
	    bcopy(from2, ps->ps_space + n, ps->ps_bcount - n);
	}
	SET_VR(v);
	splx(s);
	ps->ps_bp->b_flags &= ~B_PHYS;
    }
    DEBUGF(xtdebug,printf("bzero(%X,%d)\n",
	  ps->ps_space + ps->ps_bcount,sizeof ps->ps_space - ps->ps_bcount););
    bzero(ps->ps_space + ps->ps_bcount,sizeof ps->ps_space - ps->ps_bcount);
}

xtint(xi,bp,ccode,sdp,tarstat)
struct xha_device *xi;
struct buf *bp;
struct scsi_esense *sdp;
u_char tarstat;
{
    struct xt_softc *sc = &xt_softc[xi->xi_unit];

    DEBUGF(xtintdebug,printf("Xtint: xi=0x%X, bp=0x%X, ccode=%d, sdp=0x%X\n",
        xi,bp,ccode,sdp););
    if(tarstat == SCTS_BUSY)
    {
	if(sc->sc_busy)	/* I.e. we were here last time too */
	{
	    if(++sc->sc_busyretries > 5)
	    {
		bp->b_error=EIO;
		bp->b_flags |= B_ERROR;
		sc->sc_busy = 0;	/* Next time we'll find out ... */
		xtpartdone(xi,bp);
		return XHD_DONE;
	    }
	    return XHD_RETRY;
	}
	sc->sc_busy = 1;
	sc->sc_busyretries=1;
	return XHD_RETRY;
    }
    sc->sc_busy = 0;
    switch(ccode)
    {
    case XHI_OUVR:
    case XHI_CMPL:
        bp->b_resid = 0;
	break;
    case XHI_TIMEOUT:
	return xttryretry(xi,bp,1,EIO);
    case XHI_SELTO:
	if(++bp->b_errcnt <= 1)
	    return XHD_RESETRETRY;
	if(bp->b_errcnt <= 2 )
	    return XHD_SCRESETRETRY;
	bp->b_bcount = bp->b_resid;
	bp->b_resid = 0;
	bp->b_error = EIO;;
	bp->b_flags |= B_ERROR;
	xtpartdone(xi,bp);
	return XHD_DONE;
    case XHI_SCSIRESET:
	return XHD_RETRY;
    case XHI_ERR:
        /* Analyze the error */
        if(sdp)
            switch (sdp->es_key)
            {
	        case SCES_KEY_RECOVERED_ERROR:
                    DEBUGF(xtintdebug,printf("Xtint: RECOVERED ERROR KEY\n"););
                    break;
                case SCES_KEY_NOT_READY:
                    DEBUGF(xtintdebug,printf("Xtint: NOT READY KEY\n"););
		    printf("%s%d: not ready\n", xi->xi_driver->xd_sname,xi->xi_unit);
		    bp->b_error = EIO;
		    bp->b_flags |= B_ERROR;
		    break;
                case SCES_KEY_MEDIUM_ERROR:
                    DEBUGF(xtintdebug,printf("Xtint: MEDIUM ERROR KEY\n"););
		    bp->b_error = EIO;
		    bp->b_flags |= B_ERROR;
                    break;
                case SCES_KEY_HARDWARE_ERROR:
                    DEBUGF(xtintdebug,printf("Xtint: HARDWARE ERROR KEY\n"););
		    bp->b_error = EIO;
		    bp->b_flags |= B_ERROR;
		    break;
                case SCES_KEY_ILLEGAL_REQUEST:
                    DEBUGF(xtintdebug,printf("Xtint: ILLEGAL REQUEST KEY\n"););
		    return xttryretry(xi,bp,1,EIO);
                case SCES_KEY_UNIT_ATTENTION:
                    DEBUGF(xtintdebug,printf("Xtint: UNIT ATTENTION KEY\n"););
		    return xttryretry(xi,bp,2,EIO);
                case SCES_KEY_ABORTED_COMMAND:
                    DEBUGF(xtintdebug,printf("Xtint: ABORTED COMMAND KEY\n"););
		    return xttryretry(xi,bp,1,EIO);
                default:
		    DEBUGF(xtintdebug,printf("Xtint: Unknown KEY=%d\n",sdp->es_key););
                    break;
            }
            else
                printf("Xtint: Sense data pointer is zero\n");
            break;
        }

        /* Make sure the byte count is reasonable.                           */
        if(bp->b_bcount < 0)
            bp->b_bcount = bp->b_resid;

	xtpartdone(xi,bp);
        return XHD_DONE;
}

xtpartdone(xi,bp)
struct xha_device *xi;
struct buf *bp;
{
	register struct partsec *ps;
	if(!PARTIAL(bp)) 
		return;
	ps = xtgetps(xi,bp);
	/* Now restore buf header variables */
	if(ps->ps_phys)
		bp->b_flags |= B_PHYS;
	bp->b_un.b_addr = ps->ps_addr;
	/* If we were reading, copy out the stuff */
 	if(bp->b_flags & B_READ)
		xtemptypart(ps);
	ps->ps_addr = NULL;	
	ps->ps_bcount = 0;
	ps->ps_bp = NULL;	/* Free this entry */
}

xtemptypart(ps) struct partsec *ps;
{
    DEBUGF(xtdebug,printf("Xtemptypart(%X): copying %d bytes\n",ps,ps->ps_bcount););
    if (!ps->ps_phys) 
    {
	DEBUGF(xtdebug,printf("bcopy(%X,%X,%d) (COOKED)\n",ps->ps_space,ps->ps_addr,ps->ps_bcount););
	bcopy(ps->ps_space, ps->ps_addr, ps->ps_bcount);
    }
    else 
    {
	register caddr_t        to, to2;
	register int            s, v;
	register u_int          n;

	to = real_buf_addr(ps->ps_bp, ps->ps_addr);
	n = PGOFSET + 1 - ((u_int)to & PGOFSET);
	if (n >= ps->ps_bcount) 
	{
	    DEBUGF(xtdebug,printf("bcopy(%X,%X,%d) (RAW1)\n",ps->ps_space,
			to,ps->ps_bcount););
	    s = splhigh();
	    GET_VR0(v);
	    bcopy(ps->ps_space, to, ps->ps_bcount);
	} 
	else 
	{
	    DEBUGF(xtdebug,printf("bcopy(%X,%X,%d) (RAW2)\n",ps->ps_space,
			to,n););
	    DEBUGF(xtdebug,printf("bcopy(%X,%X,%d) (RAW2)\n",ps->ps_space + n,
			to2,ps->ps_bcount - n););
	    to2 = real_buf_addr(ps->ps_bp, ps->ps_addr + n);
	    s = splhigh();
	    GET_VR0(v);
	    bcopy(ps->ps_space, to, n);
	    bcopy(ps->ps_space + n, to2, ps->ps_bcount - n);
	}
	SET_VR(v);
	splx(s);
    }
    DEBUGF(xtdebug,printf("Xtemptypart(%X) DONE\n",ps););
}


xtread(dev,uio)
dev_t dev;
struct uio *uio;
{
 int errno;

    DEBUGF(xtdebug,printf("Xtread: dev=0x%x,uio=0x%X\n",dev,uio););
    errno = xtphys(dev);
    if (errno)
        return (errno);
    return (physio(xtstrategy, &rxtbuf[XTCUNIT(dev)], dev, B_READ, xtminphys, uio))
;
}

xtwrite(dev,uio)
dev_t dev;
struct uio *uio;
{
    int errno;

    DEBUGF(xtdebug,printf("Xtwrite: dev=0x%x,uio=0x%X\n",dev,uio););
    errno = xtphys(dev, uio);
    if (errno)
        return (errno);
    return (physio(xtstrategy, &rxtbuf[XTCUNIT(dev)], dev, B_WRITE, xtminphys, uio)
);
}

xtphys(dev)
dev_t dev;
{
        register int target,lun;
        struct xha_device *xi;
        struct xha_slave *xs;

        DEBUGF(xtdebug,printf("Xtphys: dev=0x%x\n",dev););

        target = XTCUNIT(dev);
        lun = XTUNIT(dev);
        if (((xi = xtcinfo[target]) == 0)
            || (xi->xi_alive == 0) )
                return (ENXIO);
        if (((xs = xtinfo[lun]) == 0 )
            || (xs->xs_alive == 0) )
                return (ENXIO);
        return (0);
}

xtioctl(dev,cmd,data,flag)
dev_t dev;
int cmd;
caddr_t data;
int flag;
{
    register int xtunit = XTCUNIT(dev);
    register struct xha_device *xi;
    register struct xt_softc *sc;
    register struct dkpart *dk = (struct dkpart *) data;
    register struct partentry *pep;
    register int pa;

    if(xtunit >= NXTC || (xi = xtcinfo[xtunit]) == 0 || xi->xi_alive == 0)
    {
	DEBUGF(xtioctldebug,printf("Xtioctl: bad device %d/%d\n",major(dev),minor(dev)););
	return ENODEV;
    }
    DEBUGF(xtioctldebug,printf("%s%d: Xtioctl: cmd = %d\n, data=%x\n",
	xi->xi_driver->xd_sname, xi->xi_unit, cmd););

    sc = &xt_softc[xi->xi_unit];

    switch(cmd)
    {
    case DKIOCGPART:
	
	xtpartitioninit(xi,sc,dev);

	pa = XHAMIN_PART(dev);
	pep = &sc->sc_size.partition[pa];
	dk->dk_size = pep->nblocks;		/* Fill in size */
	dk->dk_start = pep->blkoff;		/* Fill in starting block */
	dk->dk_blocksize = sc->sc_xtst.nbps;	/* Fill in blocksize */
	dk->dk_ntrack = sc->sc_xtst.ntrak;	/* Fill in track count */
	dk->dk_nsector = sc->sc_xtst.nsect;	/* Fill in sector count */
	dk->dk_ncyl = pep->nblocks / sc->sc_xtst.nspc;	/* Fill in cyl count */
	strcpy(dk->dk_name,"xt");		/* Fill in name */
	    DEBUGF(xtioctldebug,printf("%s%d: Xtioctl: DKIOCGPART done\n",
		xi->xi_driver->xd_sname, xi->xi_unit););
	return 0;

    case DKIOCSTART:
	if(xtstartstop(xi,1) < 0)			/* Start unit */
	{
		return -1;
	}
	sc->sc_spundown = 0;
	return 0;

    case DKIOCSTOP:
	if(xtstartstop(xi,0) < 0)			/* Stop unit */
	{
		return -1;
	}
	sc->sc_spundown = 1;
	return 0;

    case DKIOCSAUTOSPIN:
	sc->sc_autospinup = 1;
	return 0;

    case DKIOCSNOAUTOSPIN:
	sc->sc_autospinup = 0;
	return 0;

    case DKIOCSMODESELECT:
        if(xtmodeselect(xi,data) < 0)	 /* SCSI Mode Select */
        {
                return -1;
        }
        return 0;


    case DKIOCSMODESENSE:
        if(xtmodesense(xi,data) < 0)	/* SCSI Mode Sense */
        {
                return -1;
        }
        return 0;
    default:
	return ENOTTY;
    }
}

xtminphys(bp)
struct buf *bp;
{
    register int xtunit = XTCUNIT(bp->b_dev);
    register struct iocc_ctlr     *ic;
    register struct xha_device    *xi;
    register struct xt_softc	  *sc;
    extern unsigned                 (*dma_get_minphys())();
	int	ret; /* HW */
#ifdef DEBUG
    long	bcount;
#endif

    xi = xtcinfo[xtunit];
    sc = &xt_softc[xi->xi_unit];
#ifdef 0 /* HW */
	ret = xtphys(bp->b_dev);	/* Check existence and aliveness */
	if(ret <= 0 ) {
		return;	
	}					/* HW added 0 <=  */
#endif
#ifdef DEBUG
    bcount = bp->b_bcount;
    if(xtminphysdebug && (bp->b_bcount & (sc->sc_xtst.nbps-1)))
    {
	printf("Xtminphys: bp=%X, bp->bcount=%d not modulo %d on entry\n",
		bp,bp->b_bcount,sc->sc_xtst.nbps);
    }
#endif

    minphys(bp);		/* Enforce kernel-wide b_bcount restriction */
#ifdef DEBUG
	if(xtminphysdebug && (bp->b_bcount != bcount))
	{
	    printf("Xtminphys bp=%X,bcount=%d changed by minphys() to %d\n",
		    bp,bcount,bp->b_bcount);
	    bcount = bp->b_bcount;
	}
#endif

    if(bp->b_bcount > 65536)
	bp->b_bcount = 65536 ; 
#ifdef DEBUG
	if(xtminphysdebug && (bp->b_bcount != bcount))
	{
	    printf("Xtminphys bp=%X,bcount=%d changed by xtminphys() to %d\n",
		    bp,bcount,bp->b_bcount);
	    bcount = bp->b_bcount;
	}
#endif


    ic = xi->xi_hd->xh_iocc;
    (*dma_get_minphys(ic))(bp);     /* dma's bcount restrictions */
#ifdef DEBUG
	if(xtminphysdebug && (bp->b_bcount != bcount))
	{
	    printf("Xtminphys bp=%X,bcount=%d changed by dma_get_minphys() to %d\n",
		    bp,bcount,bp->b_bcount);
	    bcount = bp->b_bcount;
	}
#endif
    /*
     * Dma minphys routines can leave us w/ a partial
     * xfer *and* leftovers.  We can't handle that so
     * we truncate any partial (pushing it into leftovers).
     */
    if (bp->b_bcount >= sc->sc_xtst.nbps)
	bp->b_bcount &= ~(sc->sc_xtst.nbps - 1);
#ifdef DEBUG
	if(xtminphysdebug && (bp->b_bcount != bcount))
	{
	    printf("Xtminphys bp=%X,bcount=%d corrected by xtminphys() to %d\n",
		    bp,bcount,bp->b_bcount);
	    bcount = bp->b_bcount;
	}
    if(xtminphysdebug && (bp->b_bcount & (sc->sc_xtst.nbps-1)))
    {
	printf("Xtminphys: bp=%X, bp->bcount=%d not modulo %d on exit.\n",
		bp,bp->b_bcount,sc->sc_xtst.nbps);
	bcount = bp->b_bcount;
    }
#endif
	if(xtminphysdebug ) {
		printf("Xtminphys: bp=%X, bp->bcount=%d \n",
			bp,bp->b_bcount);
	} /* HW */


}

xttryretry(xi,bp,maxretries,errcode)
struct xha_device *xi;
struct buf *bp;
{
    if(++bp->b_errcnt <= maxretries)
	return XHD_RETRY;
    else
    {
	bp->b_bcount = bp->b_resid;
	bp->b_resid = 0;
	bp->b_error = errcode;
	bp->b_flags |= B_ERROR;
	xtpartdone(xi,bp);
	return XHD_DONE;
    }
}

xtgetpartitions(xi,dev)
struct xha_device *xi;
dev_t dev;
{
	struct	buf	minibuf;
	char	mmbuf[1024];
	struct	minidirectory	*mdp= (struct minidirectory *) &mmbuf ;
	register struct minidisk *mdiskp;
	register int m,nxt;
	int	i=0;
	register int pa,nparts = 0;
	register struct xt_softc *sc = &xt_softc[xi->xi_unit];

	/* Set up a xfer buffer */
	
	minibuf.b_un.b_addr =  (caddr_t)mdp;
	minibuf.b_bcount = sizeof mmbuf;
        minibuf.b_flags = B_BUSY | B_READ;
	minibuf.b_dev = dev;
	minibuf.b_blkno = MINIDISK_BLOCK;

	/* Start the xfer */
	DEBUGF(xtgetpartdebug,printf("%s%d: Starting minidisk read\n",
		xi->xi_driver->xd_sname,xi->xi_unit););
	xtstrategy(&minibuf);

	/* 
	 * Await completion, need busy wait here in order to aviod
         * reentrance 
         */
	DEBUGF(xtgetpartdebug,printf("%s%d: Waiting for minidisk read\n",
		xi->xi_driver->xd_sname,xi->xi_unit););

	while((minibuf.b_flags & B_DONE) == 0)
		if (++i > 10000000) break;
	if(minibuf.b_flags & B_DONE == 0)
	{
	
	    DEBUGF(xtgetpartdebug,printf("%s%d: Minidisk read timed out\n",
		    xi->xi_driver->xd_sname,xi->xi_unit););
	    return 0;
	}
	DEBUGF(xtgetpartdebug,printf("%s%d: Minidisk read done\n",
		xi->xi_driver->xd_sname,xi->xi_unit););

	if (minibuf.b_flags & B_ERROR)
	{
	    DEBUGF(xtgetpartdebug,printf("%s%d: Minidisk read failed\n",
		xi->xi_driver->xd_sname,xi->xi_unit););
	    return 0;
	}
	if ( mdp->header.number <= 0 ||
            mdp->header.number >= MAXDISKS)
	{
	    DEBUGF(xtgetpartdebug,printf("%s%d: Minidisk contained bad data\n",
		xi->xi_driver->xd_sname,xi->xi_unit););
	    return 0;
	}

	DEBUGF(xtgetpartdebug,printf("%s%d: Minidisk read OK\n",
		xi->xi_driver->xd_sname,xi->xi_unit););

	m = mdp->header.number;
	
	for(nxt = mdp->header.first; nxt >= 0; nxt = mdiskp->next)
	{
	    int nlen = strlen(xi->xi_driver->xd_sname);
	    if(--m < 0 || nxt > MAXDISKS)
	    {
		DEBUGF(xtgetpartdebug,printf("%s%d: Minidisk corrupted\n",
		    xi->xi_driver->xd_sname,xi->xi_unit););
		return 0;
	    }
	    mdiskp = &(mdp->minidisk[nxt]);
	    if (ISFREE(mdiskp))
	    {
		continue;
	    }
	    pa = mdiskp->name[nlen+1] - 'a';
	    if((strncmp(xi->xi_driver->xd_sname, mdiskp->name,nlen) == 0 &&
		pa >= 0 && pa < MAXNPART ) ||
	       (mdiskp->type == TYPE_PAGE && (pa == 1) ) )
	    {
		nparts++;
		DEBUGF(xtgetpartdebug,
		    printf("%s%d: Partition %s start at %d, size = %d\n",
			    xi->xi_driver->xd_sname,xi->xi_unit,
			    mdiskp->name,mdiskp->start,mdiskp->size););
		sc->sc_size.partition[pa].blkoff = mdiskp->start;
		sc->sc_size.partition[pa].nblocks = mdiskp->size;
	    }
	}
	DEBUGF(xtgetpartdebug,printf("%s%d: xtgetpart done, nparts=%d\n",
	    xi->xi_driver->xd_sname,xi->xi_unit,nparts););
	return nparts;
}
/* 
 * Combine three bytes into an unsigned
 */
unsigned
xtget3(p)
char           *p;
{
    return ((*p & 0xFF) << 16) | ((p[1] & 0xFF) << 8) | (p[2] & 0xFF);
}

xtidcpy(f,t,size)
char	*f,*t;
{
    register int	i;

    bcopy(f,t,size-1);
    t[size-1] = '\0';

    for(i=size-1;i>=0;i--)
    {
	if((t[i] != ' ') && (t[i] != '\0'))
	{
	    break;
	}
	t[i] = '\0';
    }
}
#endif
