/*
 * dk --
 *   disk manipulation program
 */
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include "/usr2/rt6.00/machine/dkio.h"
#include "/usr2/rt6.00/machineio/scsi.h"
#include <sys/ioctl.h>

#define	equal(s1,s2)	(strcmp(s1, s2) == 0)
#define DKSTART	0
#define DKSTOP	1
#define DKNOP	2
#define DKAS	3
#define DKNAS	4
#define DKCE	5
#define DKCD	6
#define DKMS  	7
#define DKPART	8
#define DKHELP	9
#define DEFDISK "/dev/rxt0c"

struct commands {
	char *c_name;
	int c_code;
	int c_ronly;
} com[] = {
	{ "start",		DKSTART,	0 },
	{ "stop",		DKSTOP,		0 },
	{ "autospin",		DKAS,		0 },
	{ "noautospin", 	DKNAS,		0 },
	{ "cacheenable", 	DKCE,		0 },
	{ "cachedisable", 	DKCD,		0 },
	{ "modesense",	 	DKMS,		0 },
	{ "partition",	 	DKPART,		0 },
	{ "help",	 	DKHELP,		0 },
	{ 0 }
};

int dkfd,ret;
char *disk;
char *progname;
struct dkpart dk;
struct scsi_ms ms;
struct scsi_msc msc;
struct commands *comp;

main(argc, argv)
	char **argv;
{
	char *getenv();
	register char *cp;

	progname = argv[0];

	if (argc > 2 && (equal(argv[1], "-t") || equal(argv[1], "-f"))) {
		argc -= 2;
		disk = argv[2];
		argv += 2;
	} else
		if ((disk = getenv("DISK")) == NULL)
			disk = DEFDISK;
	if (argc < 2) {
		fprintf(stderr, 
		"usage: %s [ -f device ] command [ count ]\n",progname);
		exit(1);
	}
	cp = argv[1];
	for (comp = com; comp->c_name != NULL; comp++){
		if (strncmp(cp, comp->c_name, strlen(cp)) == 0)
			break;
	}
	if (comp->c_name == NULL) {
		fprintf(stderr, "%s: doesn't understand \"%s\"\n",progname, cp);
		exit(1);
	}
	/* if ((dkfd = open(disk, comp->c_ronly ? 0 : 2)) < 0) { */
	if ((dkfd = open(disk, 0)) < 0) {
		perror(disk);
		exit(1);
	}
	switch(comp->c_code)
	{
	case DKSTART:
		if (ioctl(dkfd, DKIOCSTART) < 0) {
			fprintf(stderr, "%s %s", disk, comp->c_name);
			perror(" failed ");
			exit(2);
		}
		break;
		
	case DKSTOP:
		if (ioctl(dkfd, DKIOCSTOP) < 0) {
			fprintf(stderr, "%s %s", disk, comp->c_name);
			perror(" failed ");
			exit(2);
		}
		break;
	case DKAS:
		if (ioctl(dkfd, DKIOCSAUTOSPIN) < 0) {
			fprintf(stderr, "%s %s", disk, comp->c_name);
			perror(" failed ");
			exit(2);
		}
		break;
	case DKNAS:
		if (ioctl(dkfd, DKIOCSNOAUTOSPIN) < 0) {
			fprintf(stderr, "%s %s", disk, comp->c_name);
			perror(" failed ");
			exit(2);
		}
		break;
	case DKCE:
		cache_onoff(1);
		break;
	case DKCD:
		cache_onoff(0);
		break;
	case DKMS:
                ret = ioctl(dkfd, DKIOCSMODESENSE,&ms);
                if (ret < 0) {
                        fprintf(stderr, "%s %s", disk, comp->c_name);
                        perror(" failed ");
                        exit(2);
                }
		print_modesense_pages(&ms);
		break;
	case DKPART:
                ret = ioctl(dkfd, DKIOCGPART,&dk) ;
                if (ret  < 0) {
                        fprintf(stderr, "ioctl returned %d for %s %s ", 
				ret,disk, comp->c_name);
			perror(" failed ");
			exit(2);
                }
		if( ret !=0 ) printf("Return code = %d\n", ret );
		printf("partition	size	blksize	start	ncyl	ntrack	nsector	name\n");

		printf("%s:	%d	%d	%d	%d	%d	%d	%s\n",
                        disk,
                        dk.dk_size,
                        dk.dk_blocksize,
                        dk.dk_start,
                        dk.dk_ncyl,
                        dk.dk_ntrack,
                        dk.dk_nsector,
                        dk.dk_name );

		break;
	case DKHELP:
		int i;
		for (i=0 ; *com[i].c_name != 0 ; i++){
			printf("	%s\n",com[i].c_name);
		}
		break;
	case DKNOP:
		break;
	}
}
cache_onoff(int onoff)
{
ret = ioctl(dkfd, DKIOCSMODESENSE,&ms);
if (ret < 0) {
	fprintf(stderr, "%s %s", disk, comp->c_name);
	perror(" failed ");
	exit(2);
}
msc.sms_hd = ms.sms_hd;    /* initialise msc from ms */
msc.sms_csh = ms.sms_csh;
msc.sms_csh.csh_ce = onoff ; /* turn cache bit on/off */
ret = ioctl(dkfd, DKIOCSMODESELECT,&msc);
if (ret < 0) {
	fprintf(stderr, "%s %s", disk, comp->c_name);
	perror(" failed ");
	exit(2);
}
ret = ioctl(dkfd, DKIOCSMODESENSE,&ms);
if (ret < 0) {
        fprintf(stderr, "%s %s", disk, comp->c_name);
        perror(" failed ");
        exit(2);
}
printf(" Mode sense AFTER mode select \n");
print_mscachepage(&ms.sms_csh);
}

print_modesense_pages(struct scsi_ms *p)
{
	print_modesense(&p->sms_hd);
	print_mserppage(&p->sms_erp);
	print_msconnpage(&p->sms_cop);
	print_msdadfpage(&p->sms_daf);
	print_msrigidpage(&p->sms_rdg);
	print_mscachepage(&p->sms_csh);
}

print_modesense(struct scsi_modesense *p) 
{
printf("Mode Sense, Sense data length 		%d\n",p->sm_length);  
printf("Mode Sense, Medium type 		%d\n",p->sm_mtype);   
printf("Mode Sense, Write-protect 		%x\n",p->sm_wp);    
printf("Mode Sense, Maxtor Buffered mode 	%x\n",p->sm_bufmode);
printf("Mode Sense, Maxtor Speed code 		%x\n",p->sm_speed); 
printf("Mode Sense, Block descriptor length 	%d\n",p->sm_bdl);     
printf("Mode Sense Block descriptor, density code 	%d\n",
	p->sm_bdarr[0].sbd_density);     
printf("Mode Sense Block descriptor, Number of blocks 	%d\n",
	p->sm_bdarr[0].sbd_nblocks);     
printf("Mode Sense Block descriptor, Block length 	%d\n\n",
	p->sm_bdarr[0].sbd_blklen);     

};
print_mserppage(struct scsi_mserppage *p) 
{ 
printf("Error recovery page, parameter savable		%x\n", p->erp_ps);
printf("Error recovery page, Page code			%x\n", p->erp_pcode);
printf("Error recovery page, Page length		%d\n", p->erp_plen);
printf("Error recovery page, Automatic write reallocation enable	%x\n", p->erp_awre);
printf("Error recovery page, Automatic read reallocation enable		%x\n", p->erp_arre);
printf("Error recovery page, Transfer block with error	%x\n", p->erp_tb);
printf("Error recovery page, Read continous		%x\n", p->erp_rc);
printf("Error recovery page, Enable early correction	%x\n", p->erp_eec);
printf("Error recovery page, Post error			%x\n", p->erp_per);
printf("Error recovery page, Disable transfer on error	%x\n", p->erp_dte);
printf("Error recovery page, Disable correction		%x\n", p->erp_dcr);
printf("Error recovery page, Retry count		%d\n", p->erp_rtcnt);
printf("Error recovery page, Correction span		%d\n", p->erp_cspan);
printf("Error recovery page, Head offset count		%d\n", p->erp_hoffset);
printf("Error recovery page, ata strobe offset count	%d\n", p->erp_dsoffset);
printf("Error recovery page, Recovery time limit	%d\n\n", p->erp_rtl);

};
print_msconnpage(struct scsi_msconnpage *p) 
{
printf("Connect Page , parameter savable	%x\n", p->cop_ps);
printf("Connect Page , Page code		%x\n", p->cop_pcode);
printf("Connect Page , Page length		%d\n", p->cop_plen);
printf("Connect Page , Buffer full ratio	%d\n", p->cop_bfr);
printf("Connect Page , Buffer empty ratio	%d\n", p->cop_ber);
printf("Connect Page , Bus inactivity limit	%d\n", p->cop_bil);
printf("Connect Page , Disconnect time limit	%d\n", p->cop_dtl);
printf("Connect Page , Connect time limit	%d\n\n", p->cop_ctl);

};
print_msdadfpage(struct scsi_msdadfpage *p) 
{ 
   printf("Format Page, Parameter Savable		%x \n",p->daf_ps); 
   printf("Format Page, Page code			%x \n",p->daf_pcode); 
   printf("Format Page, Page length			%d \n",p->daf_plen);    
   printf("Format Page, Tracks per zone			%d \n",p->daf_tpz);     
   printf("Format Page, Alternate sectors per zone	%d \n",p->daf_altspz);  
   printf("Format Page, Alternate tracks per zone	%d \n",p->daf_atttpz);  
   printf("Format Page, Alternate tracks per volume	%d \n",p->daf_alttpv);  
   printf("Format Page, Sectors per track		%d \n",p->daf_spt);     
   printf("Format Page, No of bytes per sector		%d \n",p->daf_nbps);    
   printf("Format Page, Interleave factor		%d \n",p->daf_intrlve); 
   printf("Format Page, Track skew			%d \n",p->daf_tskew);   
   printf("Format Page, Cylinder skew			%d \n",p->daf_cskew);   
   printf("Format Page, Soft sector format		%x \n",p->daf_ssec);  
   printf("Format Page, Hard sector format		%x \n",p->daf_hsec);  
   printf("Format Page, Removable media			%x \n",p->daf_rmb);   
   printf("Format Page, Format by surface		%x \n",p->daf_surf);  
   printf("Format Page, Inhibit Save			%x \n\n",p->daf_ins);   

};
print_msrigidpage(struct scsi_msrigidpage *p) 
{ 
 printf("Rigid page,  parameter savable            %x \n",p->rdg_ps);      
 printf("Rigid page,  Page code                    %x \n",p->rdg_pcode);   
 printf("Rigid page,  Page length                  %d \n",p->rdg_plen);      
 printf("Rigid page,  Maximum no of cylinders      %d \n",
	(p->rdg_maxcyl[0]<<16) |
	(p->rdg_maxcyl[1]<<8 ) |
	(p->rdg_maxcyl[2] )
	); 
 printf("Rigid page,  Maximum no of heads          %d \n\n",p->rdg_hmaxheads); 
};
print_mscachepage(struct scsi_mscachepage *p) 
{
	printf("Wren Cache Parameter Savable:	%s\n",
	       p->csh_ps ? "Yes" : "No" );
	printf("Wren Cache page code:		%x\n", p->csh_pcode);
	printf("Wren Cache page length:		%d\n",p->csh_plen);
	printf("Wren Write Index enabled :	%s\n",
	       p->csh_wie ? "Yes" : "No" );
	printf("Wren Caching enabled :		%s\n",
	       p->csh_ce ? "Yes" : "No" );
	printf("Wren Cache table size:		%d\n",p->csh_cts);
	printf("Wren Cache Prefetch Threshold:	%x\n\n",
                        p->csh_pt);
}
