static char sccsid[] = "@(#)45	1.9  R2/sysx/cdrm/cdromb.c, bos, bos320 5/28/91 15:59:25";
/*
 * COMPONENT_NAME: (SYSXCDRM) SCSI CDROM Device Driver Bottom Half Routines
 *
 * FUNCTIONS:
 *		cd_strategy
 *              cd_release
 *		cd_pass_thru
 *              cd_device_reset
 *		cd_mode_select
 *              cd_read_capacity
 * 		cd_reserve
 *              cd_start_stop
 * 		cd_start
 *              cd_busy
 *		cd_iowait
 *              cd_watchdog
 *              cd_scsi_error
 * 		cd_adapter_error
 *              cd_iodone
 *
 * ORIGINS: 27
 *
 * (C) COPYRIGHT International Business Machines Corp. 1988, 1989
 * All Rights Reserved
 * Licensed Materials - Property of IBM
 *
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */

/* INCLUDED SYSTEM FILES */
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/sysmacros.h>
#include <sys/syspest.h>
#include <sys/dma.h>
#include <sys/buf.h>
#include <sys/file.h>
#include <sys/pin.h>
#include <sys/sleep.h>
#include <sys/ioctl.h>
#include <sys/intr.h>
#include <sys/i_machine.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/conf.h>
#include <sys/devinfo.h>
#include <sys/lockl.h>
#include <sys/device.h>
#include <sys/uio.h>
#include <sys/watchdog.h>
#include <sys/errids.h>
#include <sys/ddtrace.h>
#include <sys/trchkid.h>
#include <sys/priv.h>
#include <sys/iostat.h>

#include <sys/scsi.h>
#include <sys/cdrom.h>
#include "cdromdd.h"

/* END OF INCLUDED SYSTEM FILES  */

/**************************************************************************/
/*                                                                        */
/*                             Static Structures                          */
/*                                                                        */
/*   The disk_list variable is an array of linked lists of disk           */
/*   structures.  All of the defined devices that hash to a specific      */
/*   array element are linked together.  In addition, all of the opened   */
/*   devices are connected using seperate links. The disk structure for a */
/*   specific device is accessed by hashing to the correct array element  */
/*   based on the last 4 bits of the devno, and then sequentially         */
/*   searching either the 'defined' or the 'opened list for the correct   */
/*   devno.  The disk_chain_lock is a global lock for this driver that    */
/*   is used to insure that only one process is using the disk_list at    */
/*   a time.  The config_count is used to count the number of devices     */
/*   configured.  This info is needed for the devswadd and devswdel calls.*/
/*                                                                        */
/**************************************************************************/

struct cd_hash_table_df   cd_disk_list[CD_HASHSIZE] = { NULL };
int  cd_disk_chain_lock = LOCK_AVAIL; /* lock structure for disk array    */
int  cd_config_count = 0;             /* # of devices configured          */
#ifdef DEBUG
int  cd_debug = 0;
#endif

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_strategy                                                   */
/*                                                                        */
/*   FUNCTION:  The strategy entry point.                                 */
/*                                                                        */
/*      This routine accepts linked buf structs and scans them to see     */
/*      if consecutive ones should be consolidated into a single SCSI     */
/*      command for a chain of buf structs.  The chains are added to      */
/*      the queue with the end of each chain marked by a flag in the      */
/*      b_work field of the last buf struct in each chain.  If there's no */
/*      command in progress, cd_start will be called for the first chain  */
/*      on the queue.  If consecutive buf structs on the chain are for    */
/*      different devices, the chain is broken there and the command      */
/*      is either issued or queued for the first device.                  */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver.     */
/*      It can be called on an interrupt level, is pinned in memory,      */
/*      and can not page fault.                                           */
/*                                                                        */
/*   DATA STRUCTURES:                                                     */
/*      cd_disk_list            Common device driver information.         */
/*                                                                        */
/*   INPUTS:                                                              */
/*      buf_ptr - the address of the first buf struct in the chain        */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION: The following errno values may be returned in the */
/*      buf struct b_error field:                                         */
/*      ENXIO   - The device is not opened or not defined.                */
/*              - A bad logical block address has been specified.         */
/*      EINVAL  - The device has been opened in diag mode.                */
/*              - The request is not a multiple of the blocksize.         */
/*      EACCES  - An illegal access is attempted while the device is      */
/*                opened in the wrong mode.                               */
/*      Also if a reset has failed, the same errno will be returned as    */
/*      on the command that the reset failed on.                          */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  i_disable, i_enable,                    */
/*                                iodone                                  */
/*                                                                        */
/**************************************************************************/

int cd_strategy( struct buf *buf_ptr )
{
	struct buf              *cur_bp, *next_ptr;
	struct cd_disk_df       *disk_ptr;
	int                     old_ilevel, blocksize, last_lba;

#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_strategy");
	}
#endif
	DDHKWD5(HKWD_DD_CDDD,DD_ENTRY_STRATEGY,0,buf_ptr->b_dev,buf_ptr,
		buf_ptr->b_flags,buf_ptr->b_blkno,buf_ptr->b_bcount);
	cur_bp = buf_ptr;
	/*
	 * Loop through the entire chain of buf structs passed to this routine.
	 */
	do
	{
	/*
	 * Find the disk structure for this device in the 'opened' list.
	 */
		old_ilevel = i_disable(INTIODONE);
		disk_ptr =cd_disk_list[minor(cur_bp->b_dev) & 0x0F].opened_head;
		while (disk_ptr != NULL)
		{
			if (disk_ptr->devno == cur_bp->b_dev)
				break;
			else
				disk_ptr = disk_ptr->next_opened;
		}
		if (disk_ptr == NULL)
		{
		/*
		 * The device is not opened or not even defined so
		 * return an error.
		 */
			cur_bp->b_flags |= B_ERROR;
			cur_bp->b_error = ENXIO;
			i_enable(old_ilevel);
		}
		else
		{
			disk_ptr->cmd_pending = TRUE;
			i_enable(old_ilevel);
			cur_bp->b_resid = 0;
			blocksize = disk_ptr->disk_nblks.len;
			last_lba = disk_ptr->disk_nblks.lba;
			if (disk_ptr->diag_mode)
			{
			/*
			 * The user is attempting to do a normal read
			 * while in diag mode.  This can not be
			 * allowed because the current state of the
			 * device is unknown.
			 */
				cur_bp->b_error = EACCES;
				cur_bp->b_flags |= B_ERROR;
			}
			else if (disk_ptr->reset_failed)
			{
			/*
			 * The device is flagged as 'dead' so return
			 * the last errno returned (that caused the
			 * failure in the first place).
			 */
				cur_bp->b_error = disk_ptr->failed_errno;
				cur_bp->b_flags |= B_ERROR;
			}
			else if (cur_bp->b_bcount % blocksize)
			{
			/*
			 * The request is not a multiple of the
			 * block size so return an error.
			 */
				cur_bp->b_error = EINVAL;
				cur_bp->b_flags |= B_ERROR;
			}
		}
		next_ptr = cur_bp->av_forw;
		cur_bp->av_forw = NULL;
		if ((cur_bp->b_flags & B_ERROR) ||
		   (cur_bp->b_blkno > last_lba) ||
		   (cur_bp->b_bcount == 0))
		{
		/*
		 * Either an error was flagged above, or the user
		 * is attempting to read the 1st block beyond the end
		 * of the disc.  In the case of an error, the errno
		 * is already set from above.  In the second case,
		 * no errno should be set but the resid value should
		 * be set to indicate EOF to the user.  Call iodone
		 * on the bufstruct.
		 */
			cur_bp->b_resid = cur_bp->b_bcount;
			/*
 			 * If b_blkno starts right at the end of media
 			 * (last_lba +1), then let it return with length 0
 			 * and no error
 			 */
			if ( cur_bp->b_blkno > (last_lba + 1) )
			{
				cur_bp->b_error = ENXIO;
				cur_bp->b_flags |= B_ERROR;
			}
			iodone(cur_bp);
			if (disk_ptr != NULL)
			{
				disk_ptr->cmd_pending = FALSE;
			}
		}
		else
		{ /* no error */
			old_ilevel = i_disable(INTIODONE);
			if (disk_ptr->head == NULL)
			{
				disk_ptr->head = cur_bp;
			}
			else
			{
				disk_ptr->tail->av_forw = cur_bp;
			}
			disk_ptr->tail = cur_bp;
			if (!disk_ptr->busy)
			{
			/*
			 * The device is not busy, so the command can be issued
			 * by calling cd_start.
			 */
				disk_ptr->busy = TRUE;
				cd_start(disk_ptr, (uchar)CD_ORIGINAL_REQUEST);
			}
			disk_ptr->cmd_pending = FALSE;
			i_enable(old_ilevel);
		}
		cur_bp = next_ptr;
	} while (cur_bp != NULL);
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_strategy");
	}
#endif
	DDHKWD1(HKWD_DD_CDDD, DD_EXIT_STRATEGY, 0, buf_ptr->b_dev);
	return(0);
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_release                                                    */
/*                                                                        */
/*   FUNCTION:  Builds and sends a RELEASE command to the CD ROM device.  */
/*                                                                        */
/*      This routine fills out a sc_buf struct complete with the          */
/*      SCSI command block for a RELEASE and sends it to the SCSI         */
/*      adapter driver, which then passes the command on to the device.   */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the top half of the device driver.  It    */
/*      can be called only by a process and can page fault.               */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      disk_ptr   - a pointer to the disk structure for this device      */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION:  none                                             */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  devstrat                                */
/*                                                                        */
/**************************************************************************/

void cd_release( struct cd_disk_df *disk_ptr )
{
	struct sc_buf   *scsi_buf_ptr;
#ifdef DEBUG
	if (cd_debug)
	{
	printf("\nEntering cd_release");
	}
#endif
	/*
	 * Fill in the sc_buf, set the command state, and issue the command
	 * by calling the adapter driver's strategy routine via devstrat.
	 */
	scsi_buf_ptr = &disk_ptr->norm_buf.scsi_buf;
	scsi_buf_ptr->bufstruct.b_flags                   = 0x00;
	scsi_buf_ptr->bufstruct.b_blkno                   = 0x00;
	scsi_buf_ptr->bufstruct.b_un.b_addr               = 0x00;
	scsi_buf_ptr->bufstruct.b_bcount                  = 0x00;
	scsi_buf_ptr->bufstruct.b_error                   = 0x00;
	scsi_buf_ptr->bufstruct.b_xmemd.aspace_id         = XMEM_GLOBAL;
	scsi_buf_ptr->bp                                  = NULL;
	scsi_buf_ptr->scsi_command.scsi_length            = 0x06;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_op_code  = SCSI_RELEASE_UNIT;
	scsi_buf_ptr->scsi_command.scsi_cmd.lun           = disk_ptr->disk_ddi.
								lun_id<<5;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[0] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[1] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[2] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[3] = 0x00;
	disk_ptr->cmd_state = SCSI_RELEASE_UNIT;
	disk_ptr->norm_buf.retry_count = 0;
	(void) devstrat(&scsi_buf_ptr->bufstruct);
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_release");
	}
#endif
return;
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_pass_thru                                                  */
/*                                                                        */
/*   FUNCTION:  Builds and sends a pass-thru command to the CD ROM device.*/
/*                                                                        */
/*      This routine fills out a sc_buf struct complete with the          */
/*      SCSI command block for a pass-thru command and sends it to the    */
/*      SCSI adapter driver, which then passes the command on to the      */
/*      device.                                                           */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the top half of the device driver.  It    */
/*      can be called only by a process and can page fault.               */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      disk_ptr      - a pointer to the disk structure for this device   */
/*      sc_iocmd_ptr  - pointer to the user structure with the command    */
/*      xmem_desc_ptr - pointer to xmem struct (xmattach already called)  */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION:  none                                             */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  devstrat                                */
/*                                                                        */
/**************************************************************************/

void cd_pass_thru( struct cd_disk_df *disk_ptr,
      struct sc_iocmd *sc_iocmd_ptr,
      struct xmem *xmem_desc_ptr )
{
	struct sc_buf   *scsi_buf_ptr;
	int             i;
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_pass_thru");
	}
#endif
	/*
	 * Fill in the sc_buf, set the command state, and issue the command
	 * by calling the adapter driver's strategy routine via devstrat.
	 */
	scsi_buf_ptr = &disk_ptr->ioctl_buf.scsi_buf;
	scsi_buf_ptr->bufstruct.b_flags = sc_iocmd_ptr->flags;
	scsi_buf_ptr->bufstruct.b_blkno = 0x00;
	/*
 	 * The adapter driver should never look at this field
 	 */
	scsi_buf_ptr->bufstruct.b_error = 0x00;
	scsi_buf_ptr->bufstruct.b_un.b_addr = sc_iocmd_ptr->buffer;
	scsi_buf_ptr->bufstruct.b_bcount = sc_iocmd_ptr->data_length;

	bcopy((caddr_t) (xmem_desc_ptr),(caddr_t)
		&(scsi_buf_ptr->bufstruct.b_xmemd), sizeof(struct xmem));

	scsi_buf_ptr->bp = NULL;
	scsi_buf_ptr->scsi_command.scsi_length = sc_iocmd_ptr->command_length;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_op_code = sc_iocmd_ptr->
								scsi_cdb[0];
	scsi_buf_ptr->scsi_command.scsi_cmd.lun = sc_iocmd_ptr->scsi_cdb[1];
	for (i=2 ; i < sc_iocmd_ptr->command_length ; i++)
	{
		scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[i-2] =
						sc_iocmd_ptr->scsi_cdb[i];
	}
	disk_ptr->cmd_state = CD_PASS_THRU;
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_pass_thru");
	}
#endif
return;
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_device_reset                                               */
/*                                                                        */
/*   FUNCTION:  Sends a TEST_UNIT_READY command to the CD ROM device.     */
/*                                                                        */
/*      When this command completes, the iodone routine will cause the    */
/*      device to go through the entire reset procedure:                  */
/*      RESERVE, MODE_SENSE, MODE_SELECT, READ_CAPACITY, and START_UNIT.  */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver.     */
/*      It can be called on an interrupt level, is pinned in memory,      */
/*      and can not page fault.                                           */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      disk_ptr        Pointer to the disk structure for this device.    */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION:  none                                             */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  devstrat                                */
/*                                                                        */
/**************************************************************************/

void cd_device_reset( struct cd_disk_df *disk_ptr )
{
#ifdef DEBUG
if (cd_debug)
	{
		printf("\nEntering cd_device_reset");
	}
#endif
	/*
	 * Fill in the sc_buf, set the command state, and issue the command
	 * by calling the adapter driver's strategy routine via devstrat.
	 */
	disk_ptr->reset_buf.scsi_buf.bufstruct.b_flags = B_READ;
	disk_ptr->reset_buf.scsi_buf.bufstruct.b_blkno = 0x00;
	disk_ptr->reset_buf.scsi_buf.bufstruct.b_un.b_addr = 0x00;
	disk_ptr->reset_buf.scsi_buf.bufstruct.b_bcount = 0x00;
	disk_ptr->reset_buf.scsi_buf.bufstruct.b_error = 0x00;
	disk_ptr->reset_buf.scsi_buf.bufstruct.b_xmemd.aspace_id = XMEM_GLOBAL;
	disk_ptr->reset_buf.scsi_buf.bp = NULL;
	disk_ptr->reset_buf.scsi_buf.scsi_command.scsi_length = 0x06;
	disk_ptr->reset_buf.scsi_buf.scsi_command.scsi_cmd.scsi_op_code =
							SCSI_TEST_UNIT_READY;
	disk_ptr->reset_buf.scsi_buf.scsi_command.scsi_cmd.lun = disk_ptr->
							disk_ddi.lun_id<<5;
	disk_ptr->reset_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[0] = 0x00;
	disk_ptr->reset_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[1] = 0x00;
	disk_ptr->reset_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[2] = 0x00;
	disk_ptr->reset_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[3] = 0x00;
	disk_ptr->reset_buf.retry_count = 0;
	disk_ptr->cmd_state = SCSI_TEST_UNIT_READY;
	disk_ptr->reset_pending = TRUE;
	/* call devstrat */
	(void) devstrat(&disk_ptr->reset_buf.scsi_buf.bufstruct);
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_device_reset");
	}
#endif
return;
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_mode_sense                                                 */
/*                                                                        */
/*   FUNCTION:  Builds a MODE SENSE command.                              */
/*                                                                        */
/*      This routine fills out a sc_buf struct complete with the          */
/*      SCSI command block for a MODE SENSE command.  A MODE SENSE        */
/*      requests either the changeable or current parameters on the       */
/*      device.                                                           */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver.     */
/*      It can be called on an interrupt level, is pinned in memory,      */
/*      and can not page fault.                                           */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      disk_ptr   - a pointer to the disk structure for this device      */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION:  none                                             */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  devstrat                                */
/*                                                                        */
/**************************************************************************/

void cd_mode_sense( struct cd_disk_df *disk_ptr )
{
	struct sc_buf           *scsi_buf_ptr;
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_mode_sense");
	}
#endif
	/*
	 * Fill in the sc_buf, set the command state, and issue the command
	 * by calling the adapter driver's strategy routine via devstrat.
	 */
	scsi_buf_ptr = &disk_ptr->reset_buf.scsi_buf;
	scsi_buf_ptr->bufstruct.b_flags = (B_READ | B_BUSY);
	scsi_buf_ptr->bufstruct.b_blkno = 0x00;
	scsi_buf_ptr->bufstruct.b_error = 0x00;
	scsi_buf_ptr->bufstruct.b_un.b_addr = (caddr_t)disk_ptr->sense_buf;
	scsi_buf_ptr->bufstruct.b_bcount = 255;
	scsi_buf_ptr->bufstruct.b_xmemd.aspace_id = XMEM_GLOBAL;
	scsi_buf_ptr->bp = NULL;
	scsi_buf_ptr->scsi_command.scsi_length = 0x06;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_op_code  = SCSI_MODE_SENSE;
	scsi_buf_ptr->scsi_command.scsi_cmd.lun = 
		(disk_ptr->disk_ddi.lun_id<<5);
        /*
         * Set Page Code byte to request all supported pages....set Page
         * Control bits appropriately depending if we want changeable or
         * current mode data
         */
        scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[0] = (0x3F |
                ((disk_ptr->m_sense_status == CD_SENSE_CHANGEABLE) ?
                CD_SENSE_CHANGEABLE : CD_SENSE_CURRENT));

	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[1] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[2] = 
		(uchar)scsi_buf_ptr->bufstruct.b_bcount;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[3] = 0x00;
	disk_ptr->cmd_state = SCSI_MODE_SENSE;
        /*
         * Clear out the sense data buffer
         */
        bzero(disk_ptr->sense_buf, 256);

	(void) devstrat(&scsi_buf_ptr->bufstruct);
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_mode_sense");
	}
#endif
return;
}


/************************************************************************/
/*									*/
/* NAME: cd_format_mode_data						*/
/*									*/
/* FUNCTION:  This function parses a buffer of mode data into a standard*/ 
/*            control structure to prepare the mode data for comparison */   
/*									*/
/* EXECUTION ENVIRONMENT:						*/
/*									*/
/*      This routine is  called at the interrupt level and 		*/
/*	cannot page fault.						*/
/*									*/
/* (DATA STRUCTURES:)  cd_disk_df     Disk's information		*/
/*									*/
/* INPUTS:								*/
/*                    mode_data  - Pointer to mode data			*/
/*                    mf         - Pointer to mode format structure	*/
/*									*/
/* (RECOVERY OPERATION:)  None.						*/
/*									*/
/* RETURNS:	Void.							*/
/************************************************************************/

void cd_format_mode_data(
char *mode_data,
struct cd_mode_format *mf,
int	sense_length)
{
	char	i=0,page=0;
	short	bd_length,offset,p_length;

	for (i=0;i<CD_MAX_MODE_PAGES;i++)	
		/* 
		 * initialize all page indices to -1 
		 */
		mf->page_index[i] = (signed char)-1;

	mf->sense_length = sense_length;
	/* 
	 * get length of block descriptor 
	 */
	bd_length = mode_data[CD_BLK_DESC_LN_INDEX];
	if (bd_length == 8) {
		/*
		 * if == 8, we understand this
		 */
		mf->block_length = ((mode_data[9] << 16) | (mode_data[10]<< 8) |
				  mode_data[11]) & 0x00FFFFFF; 
	} else {
		/*
		 * else we don't understand this
		 */
		mf->block_length = 0;	
	}
	/* 
	 * compute offset to first page (i.e. offset to block descriptor +
	 * length of block descriptor area )
	 */
	offset = CD_BLK_DESC_LN_INDEX+bd_length+1;   
	while (offset < mf->sense_length) {
		/*
		 * For the remainder of the sense data...
		 * Store the index into the data buffer for each page
		 */
		page = (mode_data[offset] & CD_MAX_MODE_PAGES);	
		mf->page_index[page] = offset;
		p_length = mode_data[++offset];
		for (i=0;i<=p_length;i++)
			/*
			 * step over the data bytes for this page
			 */
			++offset;
	}
	return;
}
		

/************************************************************************/
/*									*/
/* NAME: cd_mode_data_compare						*/
/*									*/
/* FUNCTION:  Parses a devices changable mode parameters, current mode 	*/
/* 	      parameters, and the using systems desired mode parameters */
/* 	      to determine if a mode select is necessary.		*/
/*									*/
/* EXECUTION ENVIRONMENT: This routine is called at the interrupt level */
/*            and can not page fault.					*/
/*									*/
/* DATA STRUCTURES: cd_disk_df		- Disk information structure	*/
/*		    cd_mode_format	- Mode data control information	*/
/*									*/
/* (RECOVERY OPERATION:)   None.					*/
/*									*/
/* RETURNS:   0 - No mode select is necessary (or wouldn't do any good)	*/
/*	      1 - Perform mode select with altered current data		*/
/*									*/
/************************************************************************/

int cd_mode_data_compare(
struct cd_disk_df *disk_ptr)
{

	int 	select = FALSE;
	char	page,i,ch_index,dd_index,df_index,made_change;
	uchar	diff,changeable,def;
	short	ch_length,dd_length,df_length;

	/*
	 * First check block length, if non-zero
	 */
	if (!((disk_ptr->dd.block_length == 0) || 
	      (disk_ptr->cd.block_length == 0)))
		/*
		 * if both the desired data and the current data contain
		 * a defined block length
		 */
		if (disk_ptr->dd.block_length != disk_ptr->cd.block_length) {
			/*
			 * if they aren't the same, then set flag to do select
			 * and alter the current data block length to reflect
			 * the desired block length
			 */
			select = TRUE;
			disk_ptr->sense_buf[9] = 
				(disk_ptr->dd.block_length >> 16) & 0x00ff;
			disk_ptr->sense_buf[10] = 
				(disk_ptr->dd.block_length >> 8) & 0x00ff;
			disk_ptr->sense_buf[11] = 
				disk_ptr->dd.block_length & 0x00ff;
		}

	disk_ptr->sense_buf[0] = 0;  /* set reserved byte 0 of header to 0 */
	/*
	 * Now check each changeable page
	 */
	for (page=0;page<CD_MAX_MODE_PAGES;page++) {
	    made_change = FALSE;
	    if (disk_ptr->ch.page_index[page] != (signed char) -1) {
		/*
		 * if this page is supported
		 */
		ch_index = disk_ptr->ch.page_index[page];
		ch_length = disk_ptr->ch_data[ch_index+1];
		/* 
		 * mask off reserved bits 6 and 7 of the page code
		 */
		disk_ptr->sense_buf[ch_index] &= CD_MAX_MODE_PAGES; 
		if (disk_ptr->dd.page_index[page] != (signed char) -1) {
			/*
			 * if this page is in our data base pages, 
			 * then we can and may potentially change this page 
			 * of data
			 */
			dd_index = disk_ptr->dd.page_index[page];
			dd_length = disk_ptr->disk_ddi.
				mode_select_data[dd_index+1];
			if (dd_length < ch_length)
				/*
				 * if our data base has fewer bytes for this
				 * page than are changeable, only look at
				 * the ones the data base wants to change
				 */
				ch_length = dd_length;
			for (i=2;i < ch_length+2; i++) {
				/*
				 * for each changeable byte of this page
				 */
                                def = 0;
                                if (disk_ptr->df.page_index[page] !=
				    (signed char) -1) {
                                    /*
                                     * if there is default data specified
                                     * for this page
                                     */
                                    df_index = disk_ptr->df.page_index[page];
                                    df_length = disk_ptr->disk_ddi.
					mode_default_data[df_index +1];
                                    if (i < (df_length+2))
                                          /*
                                           * if there is default data specified
                                           * for this byte
                                           */
                                          def = disk_ptr->disk_ddi.
						mode_default_data[df_index+i];
                                }
				/*
				 * diff = a bitmask of bits we would like to
				 * change. (i.e. current_data XOR desired_data)
				 */
				diff = disk_ptr->sense_buf[ch_index + i] ^ 
				       disk_ptr->disk_ddi.
				       mode_select_data[dd_index + i];
				/*
				 * changeable = a bitmask of bits we would like
				 * to change and can change. (i.e. difference
				 * between current and desired ANDed with 
				 * changeable bits)
				 */
				changeable = 
				    diff & disk_ptr->ch_data[ch_index + i];
				/*
				 * Now, make sure we don't clobber any bits
				 * that the data base has specified that we
				 * should use device default parameters
				 */
				changeable &= ~def;
				if (changeable) {
				    /*
				     * means we would like to make a change
				     */
				    /*
				     * Clear and Set desired bits
				     * Current Data XOR changeable
				     */
				    disk_ptr->sense_buf[ch_index + i] ^= 
				      changeable;
				    select = TRUE;
				    made_change = TRUE;
				} /* if changeable */
			} /* for each byte of this page */
		} /* if this page in data base */	
            	if (!made_change) {
			/*
			 * we didn't change anything in this
			 * page, so shift it out of the current and
			 * changeable data
			 */
			ch_index = disk_ptr->ch.page_index[page];
			ch_length = disk_ptr->ch_data[ch_index+1];
			 for (i=ch_index;i<disk_ptr->ch.sense_length;i++)
			 	disk_ptr->ch_data[i] = 
			 		disk_ptr->ch_data[i+ch_length+2];
			 for (i=ch_index;i< disk_ptr->cd.sense_length;i++)
			 	disk_ptr->sense_buf[i] = 
			 		disk_ptr->sense_buf[i+ch_length+2];
			 cd_format_mode_data(disk_ptr->ch_data,
			    &disk_ptr->ch,(int)(disk_ptr->ch.sense_length - 
			    (ch_length + 2)));
			 cd_format_mode_data(disk_ptr->sense_buf, 
			    &disk_ptr->cd, (int)(disk_ptr->cd.sense_length - 
			    (ch_length + 2)));
		}
	    } /* if this page supported */					
	} /* for each possible page */
	return(select);
}



/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_mode_select                                                */
/*                                                                        */
/*   FUNCTION:  Builds a MODE SELECT command.                             */
/*                                                                        */
/*      This routine fills out a sc_buf struct complete with the          */
/*      SCSI command block for a MODE SELECT command.  A MODE SELECT      */
/*      loads the default parameters to be used initially by the body of  */
/*      the code.                                                         */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver.     */
/*      It can be called on an interrupt level, is pinned in memory,      */
/*      and can not page fault.                                           */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      disk_ptr   - a pointer to the disk structure for this device      */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION:  none                                             */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  devstrat                                */
/*                                                                        */
/**************************************************************************/

void cd_mode_select( struct cd_disk_df *disk_ptr )
{
	struct sc_buf           *scsi_buf_ptr;
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_mode_select");
	}
#endif
	/*
	 * Fill in the sc_buf, set the command state, and issue the command
	 * by calling the adapter driver's strategy routine via devstrat.
	 */
	scsi_buf_ptr = &disk_ptr->reset_buf.scsi_buf;
	scsi_buf_ptr->bufstruct.b_flags = B_WRITE;
	scsi_buf_ptr->bufstruct.b_blkno = 0x00;
	scsi_buf_ptr->bufstruct.b_error = 0x00;
	scsi_buf_ptr->bufstruct.b_un.b_addr = (caddr_t)disk_ptr->sense_buf;
	scsi_buf_ptr->bufstruct.b_bcount = (unsigned)disk_ptr->cd.sense_length;
	scsi_buf_ptr->bufstruct.b_xmemd.aspace_id = XMEM_GLOBAL;
	scsi_buf_ptr->bp = NULL;
	scsi_buf_ptr->scsi_command.scsi_length = 0x06;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_op_code  = SCSI_MODE_SELECT;
	scsi_buf_ptr->scsi_command.scsi_cmd.lun = (disk_ptr->disk_ddi.
						lun_id<<5) | CD_COMPLIES_PF;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[0] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[1] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[2] = (uchar)disk_ptr->
						cd.sense_length;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[3] = 0x00;
	disk_ptr->cmd_state = SCSI_MODE_SELECT;
	(void) devstrat(&scsi_buf_ptr->bufstruct);
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_mode_select");
	}
#endif
return;
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  read_capacity                                                 */
/*                                                                        */
/*   FUNCTION:  Builds a READ CAPACITY command.                           */
/*                                                                        */
/*      This routine fills out a sc_buf struct complete with the          */
/*      SCSI command block for a READ CAPACITY command.  This finds out   */
/*      the disk's capacity for data and passes this into the structures  */
/*      for later use.                                                    */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver.     */
/*      It can be called on an interrupt level, is pinned in memory,      */
/*      and can not page fault.                                           */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      disk_ptr   - a pointer to the disk structure for this device      */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION:  none                                             */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  devstrat                                */
/*                                                                        */
/**************************************************************************/

void cd_read_capacity( struct cd_disk_df *disk_ptr )
{
	struct sc_buf           *scsi_buf_ptr;
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_read_capacity");
	}
#endif
	/*
	 * Fill in the sc_buf, set the command state, and issue the command
	 * by calling the adapter driver's strategy routine via devstrat.
	 */
	scsi_buf_ptr = &disk_ptr->reset_buf.scsi_buf;
	scsi_buf_ptr->bufstruct.b_flags = B_READ;
	scsi_buf_ptr->bufstruct.b_blkno = 0x00;
	scsi_buf_ptr->bufstruct.b_error = 0x00;
	scsi_buf_ptr->bufstruct.b_un.b_addr = (caddr_t)&disk_ptr->disk_nblks;
	scsi_buf_ptr->bufstruct.b_bcount = sizeof (struct cd_capacity);
	scsi_buf_ptr->bufstruct.b_xmemd.aspace_id = XMEM_GLOBAL;
	scsi_buf_ptr->bp = NULL;
	scsi_buf_ptr->scsi_command.scsi_length = 0x0A;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_op_code = SCSI_READ_CAPACITY;
	scsi_buf_ptr->scsi_command.scsi_cmd.lun = disk_ptr->disk_ddi.lun_id<<5;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[0] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[1] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[2] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[3] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[4] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[5] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[6] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[7] = 0x00;
	disk_ptr->cmd_state = SCSI_READ_CAPACITY;
	(void) devstrat(&scsi_buf_ptr->bufstruct);
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_read_capacity");
	}
#endif
return;
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_reserve                                                    */
/*                                                                        */
/*   FUNCTION:  Builds a RESERVE command.                                 */
/*                                                                        */
/*      This routine fills out a sc_buf struct complete with the          */
/*      SCSI command block for a RESERVE command.  This command allows the*/
/*      to shut out other initiators from trying to get control of the    */
/*      device.                                                           */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver.     */
/*      It can be called on an interrupt level, is pinned in memory,      */
/*      and can not page fault.                                           */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      disk_ptr   - a pointer to the disk structure for this device      */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION:  none                                             */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  devstrat                                */
/*                                                                        */
/**************************************************************************/

void cd_reserve( struct cd_disk_df *disk_ptr )
{
	struct sc_buf           *scsi_buf_ptr;
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_reserve");
	}
#endif
	/*
	 * Fill in the sc_buf, set the command state, and issue the command
	 * by calling the adapter driver's strategy routine via devstrat.
	 */
	scsi_buf_ptr = &disk_ptr->reset_buf.scsi_buf;
	scsi_buf_ptr->bufstruct.b_flags = B_READ;
	scsi_buf_ptr->bufstruct.b_blkno = 0x00;
	scsi_buf_ptr->bufstruct.b_error = 0x00;
	scsi_buf_ptr->bufstruct.b_un.b_addr = 0x00;
	scsi_buf_ptr->bufstruct.b_bcount = 0x00;
	scsi_buf_ptr->bufstruct.b_xmemd.aspace_id = XMEM_GLOBAL;
	scsi_buf_ptr->bp = NULL;
	scsi_buf_ptr->scsi_command.scsi_length = 0x06;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_op_code =
							SCSI_RESERVE_UNIT;
	scsi_buf_ptr->scsi_command.scsi_cmd.lun = disk_ptr->disk_ddi.lun_id<<5;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[0] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[1] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[2] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[3] = 0x00;
	disk_ptr->cmd_state = SCSI_RESERVE_UNIT;
	(void) devstrat(&scsi_buf_ptr->bufstruct);
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_reserve");
	}
#endif
return;
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_start_stop                                                 */
/*                                                                        */
/*   FUNCTION:  Builds a START_STOP_UNIT command.                         */
/*                                                                        */
/*      This routine fills out a sc_buf struct complete with the          */
/*      SCSI command block for a START_STOP_UNIT command.  This command   */
/*      spins the disk up if it is opened up for the first time or        */
/*      prepares the disk to be spun down if the device is to be closed   */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver.     */
/*      It can be called on an interrupt level, is pinned in memory,      */
/*      and can not page fault.                                           */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      disk_ptr        - a pointer to the disk structure for this device */
/*      scsi_buf_ptr    - a pointer to the sc_buf to be used for this cmd */
/*      start_stop_flag - a flag to indicate either a START or a STOP     */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION:  none                                             */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  devstrat                                */
/*                                                                        */
/**************************************************************************/

void cd_start_stop( struct cd_disk_df *disk_ptr,
    struct sc_buf *scsi_buf_ptr,
    uchar start_stop_flag )
{
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_start_stop");
	}
#endif
	/*
	 * Fill in the sc_buf, set the command state, and issue the command
	 * by calling the adapter driver's strategy routine via devstrat.
	 */
	scsi_buf_ptr->bufstruct.b_flags = B_READ;
	scsi_buf_ptr->bufstruct.b_blkno = 0x00;
	scsi_buf_ptr->bufstruct.b_error = 0x00;
	scsi_buf_ptr->bufstruct.b_un.b_addr = 0x00;
	scsi_buf_ptr->bufstruct.b_bcount = 0x00;
	scsi_buf_ptr->bufstruct.b_xmemd.aspace_id = XMEM_GLOBAL;
	scsi_buf_ptr->bp = NULL;
	scsi_buf_ptr->scsi_command.scsi_length = 0x06;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_op_code  =SCSI_START_STOP_UNIT;
	scsi_buf_ptr->scsi_command.scsi_cmd.lun = disk_ptr->disk_ddi.lun_id<<5;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[0] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[1] = 0x00;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[2] = start_stop_flag;
	scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[3] = 0x00;
	disk_ptr->cmd_state = SCSI_START_STOP_UNIT;
	(void) devstrat(&scsi_buf_ptr->bufstruct);
#ifdef DEBUG
	if (cd_debug)
	{
	printf("\nLeaving cd_start_stop");
	}
#endif
return;
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_start                                                      */
/*                                                                        */
/*   FUNCTION:  Sends a command to the CD ROM device.                     */
/*                                                                        */
/*      This routine is called when a command is to be issued to the      */
/*      specified device.  The routine begins by checking for the case    */
/*      of an ioctl.  For all ioctls other than reads, the scsi buf will  */
/*      already be filled out and only needs to be issued by calling the  */
/*      adapter driver's strategy routine.  For reads (either normal or   */
/*      ioctl read with sense),the routine begins by adjusting the length */
/*      of a read that will go past the end of the disk so that it will   */
/*      only read up to the last block.  If this is not a spanned command,*/
/*      the length is checked and adjusted so that it is not larger than  */
/*      the maximum that the adapter driver can handle.  Next the sc_buf  */
/*      is filled in for the read request and the initial retry count is  */
/*      set up.  Finally the command is issued via a call to the adapter  */
/*      driver's strategy routine.                                        */
/*                                                                        */
/*      One unique aspect of this routine is that in the case of a single */
/*      command that is too large to execute in one SCSI command, this    */
/*      routine will get called repetitively until the request is         */
/*      complete.  The active_resid variable is used to store the number  */
/*      of bytes remaining after each sub-command completes.  The blkno   */
/*      and usr_bfr_ptr paramaters are also needed exclusively for this   */
/*      situation so this routine knows where to start the read, and      */
/*      where to send the data (since these values may be different than  */
/*      what is in the buf struct).                                       */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver.     */
/*      It can be called on an interrupt level, is pinned in memory,      */
/*      and can not page fault.                                           */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      disk_ptr    - a pointer to the disk structure for this device     */
/*      bp          - first buf struct in a chain that this read is for   */
/*                    (this value will be NULL in the case of an ioctl)   */
/*      length      - total length of the SCSI command to issue (bytes)   */
/*                    (this value will be 0 in the case of an ioctl)      */
/*      blkno       - logical block address to start reading from         */
/*                    (this value will be 0 in the case of an ioctl)      */
/*      usr_bfr_ptr - where the read data is to be sent                   */
/*                    (this value will be NULL in the case of an ioctl)   */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION:  none                                             */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  devstrat                                */
/*                                                                        */
/**************************************************************************/

void cd_start( struct cd_disk_df *disk_ptr, uchar flag )
{
	struct sc_buf   *scsi_buf_ptr;
	struct buf      *bp;
	ushort          xfer_length;
	struct xmem     *xmemd_ptr;
	daddr_t         blkno;
	uint            length, transfer;
	caddr_t         usr_bfr_ptr;

#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_start");
	}
#endif

	disk_ptr->cd_dkstat.dk_status |= IOST_DK_BUSY;
	if (disk_ptr->ioctl_pending)
	{
	/*
	 * This is an ioctl operation.
	 */
		if ( (disk_ptr->ioctl_buf.scsi_buf.scsi_command.scsi_cmd.
			scsi_op_code == SCSI_READ_EXTENDED) &&
		     (!(disk_ptr->cmd_state == CD_PASS_THRU)) )
		{
		/*
	 	 * This is a read with sense operation. Set up the bp,length,and
		 * blkno values so this operation can use the same path as the
		 * regular read.
	 	 */
			bp = &disk_ptr->rdse_buf_struct;
			xmemd_ptr = &bp->b_xmemd;
		}
		else
		{
		/*
	 	 * This is a pass-through command.
	 	 */
			bp = NULL;
			xmemd_ptr = &disk_ptr->ioctl_buf.scsi_buf.bufstruct.
					b_xmemd;
		}
		/*
	 	 * For an ioctl operation, retries are not done.
	 	 * The sc_buf used is the ioctl_buf.
	 	 */
		disk_ptr->ioctl_buf.retry_count = CD_NO_RETRY;
		scsi_buf_ptr = &disk_ptr->ioctl_buf.scsi_buf;
	}
	else
	{
	/*
	 * This is a normal read command.  If in diag mode, retries
	 * are not done.  Otherwise, the retry count is reset to 0.
	 */
		if (disk_ptr->diag_mode)
		{
			disk_ptr->norm_buf.retry_count = CD_NO_RETRY;
		}
		else
		{
			disk_ptr->norm_buf.retry_count = 0;
		}
		scsi_buf_ptr = &disk_ptr->norm_buf.scsi_buf;
		bp = disk_ptr->head;
		xmemd_ptr = &bp->b_xmemd;
	}
	if (bp != NULL)
	{
		DDHKWD5(HKWD_DD_CDDD, DD_ENTRY_BSTART, 0, disk_ptr->devno, bp,
				bp->b_blkno, bp->b_bcount, bp->b_flags);
	/*
	 * This is either a normal read or an ioctl read with sense.
	 * If this read will go beyond the end of the disc, shorten
	 * the length of the read to go only to the end and set the
	 * resid value to the remainder.
	 */
		if (flag == CD_ORIGINAL_REQUEST)
		{
			if (!(disk_ptr->ioctl_pending) && (bp->b_blkno + bp->
				b_bcount / disk_ptr->disk_nblks.len > disk_ptr->
				disk_nblks.lba+1))
			{
				length = (disk_ptr->disk_nblks.lba+1 -
						bp->b_blkno) * disk_ptr->
						disk_nblks.len;
			}
			else
			{
				length = bp->b_bcount;
			}
			blkno = bp->b_blkno;
			usr_bfr_ptr = bp->b_un.b_addr;
			bp->b_resid = bp->b_bcount - length;
		}
		else
		{ /* this is the continuation of a fragmented command */
			transfer = *(int *)(disk_ptr->disk_ddi.
					mode_select_data + 8);
			blkno = scsi_buf_ptr->bufstruct.b_blkno + (disk_ptr->
					max_transfer / transfer);
			usr_bfr_ptr = scsi_buf_ptr->bufstruct.b_un.b_addr +
					disk_ptr->max_transfer;
			length = bp->b_un.b_addr + bp->b_bcount - usr_bfr_ptr -
					bp->b_resid;
		}
		if (length > disk_ptr->max_transfer)
			length = disk_ptr->max_transfer;
		/*
		 * The xfer_length is the number of blocks to read.  This is
		 * used in the scsi command block.
		 */
		transfer = *(int *)(disk_ptr->disk_ddi.mode_select_data + 8);
		xfer_length = (ushort)(length/transfer);
		disk_ptr->cd_dkstat.dk_rblks += xfer_length;
		disk_ptr->cd_dkstat.dk_xfers++;
		/*
		 * Fill out the sc_buf for the read command and set the
		 * command state.
		 */
		scsi_buf_ptr->bufstruct.b_flags = B_READ;
		scsi_buf_ptr->bufstruct.b_blkno = blkno;
		scsi_buf_ptr->bufstruct.b_error = 0x00;
		scsi_buf_ptr->bufstruct.b_un.b_addr = usr_bfr_ptr;
		scsi_buf_ptr->bufstruct.b_bcount = length;
		bcopy((caddr_t) (xmemd_ptr),(caddr_t)
		       &(scsi_buf_ptr->bufstruct.b_xmemd), sizeof(struct xmem));

		scsi_buf_ptr->scsi_command.scsi_length = 0x0A;
		scsi_buf_ptr->scsi_command.scsi_cmd.scsi_op_code =
					SCSI_READ_EXTENDED;
		scsi_buf_ptr->scsi_command.scsi_cmd.lun = disk_ptr->disk_ddi.
					lun_id<<5;
		scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[0] =
					(uchar)(blkno >> 24);
		scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[1] =
					(uchar)(blkno >> 16);
		scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[2] =
					(uchar)(blkno >> 8);
		scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[3] =
					(uchar)blkno;
		scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[4] = 0x00;
		scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[5] =
					(uchar)(xfer_length >> 8);
		scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[6] =
					(uchar)xfer_length;
		scsi_buf_ptr->scsi_command.scsi_cmd.scsi_bytes[7] = 0x00;
		disk_ptr->cmd_state = SCSI_READ_EXTENDED;
	}
	else
	{
		DDHKWD5(HKWD_DD_CDDD,DD_ENTRY_BSTART,0,disk_ptr->devno,0,0,0,0);
	}
	/*
	 * Issue the command by calling the adapter driver's strategy routine
	 * via devstrat.
	 */
	(void) devstrat(&scsi_buf_ptr->bufstruct);
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_start");
	}
#endif
	DDHKWD1(HKWD_DD_CDDD, DD_EXIT_BSTART, 0, disk_ptr->devno);
	return;
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_busy                                                       */
/*                                                                        */
/*   FUNCTION:  Called by cd_ioctl to check if a command is in progress.  */
/*                                                                        */
/*      The purpose of this routine is to provide a service to the ioctl  */
/*      routine to check if the queue is empty.  Interrupts must be       */
/*      disabled before the queue is checked by checking the busy flag.   */
/*      If it is busy, the routine sets the ioctl_rqst flag and returns   */
/*      a value of true.  If it is not busy, the busy flag and is set so  */
/*      the ioctl can issue the command, and the return value is set to   */
/*      FALSE (not busy).                                                 */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver      */
/*      because it must be pinned in memory and can not page fault.       */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      disk_ptr - a pointer to the disk structure for this device.       */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:                                            */
/*      The return value is either TRUE if a command is already in        */
/*      progress, or FALSE if not.                                        */
/*                                                                        */
/*   ERROR DESCRIPTION:  none                                             */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  i_disable, i_enable                     */
/*                                                                        */
/**************************************************************************/

uchar cd_busy( struct cd_disk_df *disk_ptr )
{
	int     old_ilevel;
	uchar   return_value;

#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_busy");
	}
#endif
	old_ilevel = i_disable(INTIODONE);
	if (disk_ptr->busy)
	{
	/*
	 * There is currently a normal read in progress, so set the
	 * flag to indicate to the iodone routine that an ioctl
	 * operation needs to be kicked off.  Return the TRUE value
	 * to indicate to the ioctl routine to not start the operation.
	 */
		disk_ptr->ioctl_rqst = TRUE;
		return_value = TRUE;
	}
	else
	{
	/*
	 * No command is in progress, so it is ok to issue the ioctl
	 * command.  The busy and ioctl_pending flags are set to
	 * prevent a normal read from being issued(they will be
	 * queued up.  The return value is set to indicate to the
	 * ioctl routine that it is ok to call start to kick off
	 * the operation.
	 */
		disk_ptr->busy = TRUE;
		disk_ptr->ioctl_pending = TRUE;
		return_value = FALSE;
	}
	i_enable(old_ilevel);
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_busy");
	}
#endif
	return(return_value);
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_iowait                                                     */
/*                                                                        */
/*   FUNCTION:  The iowait service for this driver.                       */
/*                                                                        */
/*      The purpose of this routine is to provide a service to all the    */
/*      top half routines to wait on a specified buf struct.  The routine */
/*      makes sure that the wakeup has not already been issued and then   */
/*      sleeps.  The iodone routine will wake the process back up.        */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver      */
/*      because it must be pinned in memory and can not page fault.       */
/*      It must be called, however, by a process.                         */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      buf_ptr - the address of the buf struct.                          */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION:  none                                             */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  i_disable, i_enable,                    */
/*                                e_sleep                                 */
/*                                                                        */
/**************************************************************************/

void cd_iowait( struct buf *buf_ptr )
{
	int     old_ilevel;
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_iowait");
	}
#endif
	old_ilevel = i_disable(INTIODONE);
	/*
	 * Only go to sleep if the B_DONE flag has not already been set.
	 * If B_DONE is set, processing is already complete and the wakeup
 	 * has already been issued.
	 */
	while (!(buf_ptr->b_flags & B_DONE))
	{
		(void)e_sleep(&buf_ptr->b_event, EVENT_SHORT);
	}
	/*
	 * Clear the B_DONE flag for the next iowait operation.
	 */
	buf_ptr->b_flags &= ~B_DONE;
	i_enable(old_ilevel);
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_iowait");
	}
#endif
	return;
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_watchdog                                                   */
/*                                                                        */
/*   FUNCTION:  This routine is called when the watchdog timer pops.      */
/*                                                                        */
/*      The purpose of this routine is to call the cd_device_reset        */
/*      routine when the timer pops.  The purpose of this delay is to     */
/*      wait for the device to get back in the ready state after a bus    */
/*      reset.                                                            */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver.     */
/*      It can be called on an interrupt level, is pinned in memory,      */
/*      and can not page fault.                                           */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      watchdog_timer - address of the watchdog struct (this is also     */
/*                       the address of the disk structure.               */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION:  none                                             */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  none                                    */
/*                                                                        */
/**************************************************************************/

void cd_watchdog( struct watchdog *watchdog_timer )
{
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_watchdog");
	}
#endif
	/*
	 * Call the reset procedure with the address of the disk structure.
	 * The watchdog timer structure was put at the front of the
	 * disk structure so that the addresses would be the same.
	 */
	cd_device_reset((struct cd_disk_df *)watchdog_timer);
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_watchdog");
	}
#endif
	return;
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_scsi_error                                                 */
/*                                                                        */
/*   FUNCTION:  Logs and determines how to handle SCSI errors.            */
/*                                                                        */
/*      This routine is called when a CHECK_CONDITION is detected.        */
/*      The sense key is checked to determine if the error should be      */
/*      logged when there is a CHECK_CONDITION.  If the command should    */
/*      be retried, CD_RETRY is returned.  If the device should be reset  */
/*      before a retry is attempted, CD_RESET is returned.  If a timer    */
/*      should be set before the reset is done, CD_WAIT is returned.      */
/*      Otherwise, the proper errno value to return to the user is        */
/*      returned to the caller.                                           */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver.     */
/*      It can be called on an interrupt level, is pinned in memory,      */
/*      and can not page fault.                                           */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      disk_ptr       - pointer to the disk structure for this device.   */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:                                            */
/*      The return value is either the errno value that should be         */
/*      returned to the user,  a RESET flag to indicate that the device   */
/*      should be reset before the command is retried, a RETRY flag       */
/*      to indicate that the command should be retried, a WAIT flag to    */
/*      indicate that a timer should be set before the reset is attempted,*/
/*      or a 0 to indicate a recovered error.                             */
/*                                                                        */
/*   ERROR DESCRIPTION: The following errno values may be returned to     */
/*      the calling routine (not directly to the user):                   */
/*      EIO       - Device is unable to receive cmds (stuck on unit attn.)*/
/*                - Hardware error.                                       */
/*                - Illegal request.                                      */
/*                - Aborted command.                                      */
/*      ENOTREADY - The device is not ready.                              */
/*      ESTALE    - The CD has been changed.                              */
/*      EMEDIA    - Media error.                                          */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  errsave                                 */
/*                                                                        */
/**************************************************************************/
uchar cd_scsi_error( struct cd_disk_df *disk_ptr )
{
	uchar                   return_value;
	struct cd_buf_block     *buf_ptr;
	uint                    err_blkno, transfer;
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_scsi_error");
	}
#endif
	/*
	 * The first thing that must be determined is what sc_buf was used
	 * to issue the command that caused the original check_condition.
	 */
	if (disk_ptr->reset_pending)
	{
		buf_ptr = &disk_ptr->reset_buf;
	}
	else if (disk_ptr->ioctl_pending)
	{
		buf_ptr = &disk_ptr->ioctl_buf;
	}
	else
	{
		buf_ptr = &disk_ptr->norm_buf;
	}
	/*
	 * If the block number in the sense data is valid, that is used to
	 * determine the resid count.  If an error occurs, the entire block
	 * will be transfered anyway, so the resid count returned from the
	 * adapter driver will not be valid.
	 */
	if (disk_ptr->req_sense_data.err_code & 0x80)
	{
		err_blkno = ((disk_ptr->req_sense_data.addr_byte0 << 0x18) |
     				(disk_ptr->req_sense_data.addr_byte1 << 0x10) |
     				(disk_ptr->req_sense_data.addr_byte2 << 0x08) |
     				(disk_ptr->req_sense_data.addr_byte3));
		transfer = *(int *)(disk_ptr->disk_ddi.mode_select_data + 8);
		buf_ptr->scsi_buf.bufstruct.b_resid = buf_ptr->
			scsi_buf.bufstruct.b_bcount - ((err_blkno - buf_ptr->
			scsi_buf.bufstruct.b_blkno) * transfer);
	}
	else
	{
		err_blkno = 0xffffffff;
		buf_ptr->scsi_buf.bufstruct.b_resid = buf_ptr->
			scsi_buf.bufstruct.b_bcount;
	}
	/*
	 * The error log structure is filled out if it has not already been
	 * done (if this is the second failure, it will already be filled in).
	 */
	if (!disk_ptr->error_pending)
	{
		bcopy(disk_ptr->disk_ddi.resource_name, disk_ptr->
			log_struct.resource_name, 8);
		bcopy(&buf_ptr->scsi_buf.scsi_command, &disk_ptr->
			log_struct.scsi_command, sizeof(struct scsi));
		disk_ptr->log_struct.status_validity = buf_ptr->
			scsi_buf.status_validity;
		disk_ptr->log_struct.scsi_status =buf_ptr->scsi_buf.scsi_status;
		disk_ptr->log_struct.general_card_status = buf_ptr->
			scsi_buf.general_card_status;
		disk_ptr->log_struct.reserved1 = 0x00;
		bzero(disk_ptr->log_struct.req_sense_data, 128);
		bcopy(&disk_ptr->req_sense_data, disk_ptr->
			log_struct.req_sense_data,
			sizeof(struct cd_req_sense_info));
		/*
		 * The first reserved word is used to store all the pending
		 * flags: open_pending, error_pending, ioctl_pending,
		 * reset_pending. The second reserved word is not used.
		 */
		disk_ptr->log_struct.reserved2 = 
					((uint)disk_ptr->open_pending << 24) |
       				((uint)disk_ptr->error_pending << 16) |
       				((uint)disk_ptr->ioctl_pending << 8) |
       				((uint)disk_ptr->reset_pending);
		disk_ptr->log_struct.reserved3 = 0x00;
	}
	switch (disk_ptr->req_sense_data.sense_key)
	{
		case CD_RECOVERED_ERROR:
		/*
		 * In the case of a recovered command, no recovery is necessary,
		 * but if the recovered error threshold has been reached,
		 * the error must be logged. Since the mode select paramaters
		 * have been set so recovered errors are not reported, this
		 * code should never be reached, but I have left
		 * it here in case the paramaters change.
		 */
			disk_ptr->log_struct.error_id = ERRID_CDROM_ERR7;
			errsave(&disk_ptr->log_struct,
			sizeof(struct sc_error_log_df));
			return_value = 0;
			break;
		case CD_NOT_READY:
		/*
		 * If the device is not ready, the command will be retried until
		 * the retries are exhausted.  In the case of a reset.  A delay
		 * will be used (using the watchdog timer) before the retry is
		 * attempted. In diag mode, no retries will be done.
		 */
			if (disk_ptr->reset_pending)
			{
				if (disk_ptr->reset_count < CD_MAX_RESET_RETRY)
				{
					disk_ptr->reset_count++;
					return_value = CD_WAIT;
				}
				else
				{
					return_value = ENOTREADY;
				}
			}
			else if (buf_ptr->retry_count < CD_MAX_RETRY)
			{
				buf_ptr->retry_count++;
				return_value = CD_RETRY;
			}
			else
			{
				return_value = ENOTREADY;
			}
			break;
		case CD_UNIT_ATTENTION:
		/*
		 * For a unit attention, the device will be reset
		 * until the reset count is exhausted.  If the media
		 * has been exchanged, however, a media error will
		 * be returned.  A reset will not be issued if in
		 * diag mode.
		 */
			if (disk_ptr->req_sense_data.add_sense_key !=
				CD_MEDIUM_CHANGED || disk_ptr->open_pending)
			{
				if (disk_ptr->reset_count < CD_MAX_RESET_RETRY)
				{
					disk_ptr->reset_count++;
					if (disk_ptr->reset_count == 1)
					{
						return_value = CD_RESET;
					}
					else
					{
						return_value = CD_WAIT;
					}
				}
				else
				{
					return_value = EIO;
				}
			}
			else
			{
				return_value = ESTALE;
				disk_ptr->reset_failed = TRUE;
				disk_ptr->failed_errno = ESTALE;
			}
			break;
		case CD_MEDIUM_ERROR:
		/*
		 * A medium error will be retried.  If retries are
		 * successful, the temporary error will be logged
		 * in cd_iodone.  If retries are not successful, the
		 * permanent error will the logged below.
		 */
			if (buf_ptr->retry_count < CD_MAX_RETRY)
			{
				buf_ptr->retry_count++;
				return_value = CD_RETRY;
				disk_ptr->log_struct.error_id =ERRID_CDROM_ERR2;
				disk_ptr->error_pending = TRUE;
			}
			else
			{
				return_value = EMEDIA;
				disk_ptr->log_struct.error_id =ERRID_CDROM_ERR1;
				errsave(&disk_ptr->log_struct,
					sizeof(struct sc_error_log_df));
				disk_ptr->error_pending = FALSE;
			}
			break;
		case CD_HARDWARE_ERROR:
		case CD_ABORTED_COMMAND:
		/*
		 * A hardware error or aborted command will be retried.
		 * If retries are successful, the temporary error will
		 * be logged in cd_iodone.  If retries are not
		 * successful, the permanent error will the logged below.
		 */
			if (buf_ptr->retry_count < CD_MAX_RETRY)
			{
				buf_ptr->retry_count++;
				return_value = CD_RETRY;
				disk_ptr->log_struct.error_id =ERRID_CDROM_ERR4;
				disk_ptr->error_pending = TRUE;
			}
			else
			{
				return_value = EIO;
				disk_ptr->log_struct.error_id =ERRID_CDROM_ERR3;
				errsave(&disk_ptr->log_struct,
					sizeof(struct sc_error_log_df));
				disk_ptr->error_pending = FALSE;
			}
			break;
		case CD_ILLEGAL_REQUEST:
		/*
		 * An illegal request is most likely due to an attempt
		 * to read data from an audio track or a read past the
		 * end of the disc (only possible with CDIORDSE).
		 */
			if ((disk_ptr->reset_pending) && 
			   ((disk_ptr->reset_state == SCSI_MODE_SENSE) ||
			   (disk_ptr->reset_state == SCSI_MODE_SELECT))) {
				/*
				 * this was a mode sense or select that got
				 * an illegal request. Either it doesn't
				 * support SCSI Command Command Set, or didn't
				 * like one of our parameter, so we'll ignore
				 * and run with default settings.
				 */
				return_value = CD_BYPASS;
			} else {
				if (buf_ptr->retry_count < CD_MAX_RETRY)
				{
					buf_ptr->retry_count++;
					return_value = CD_RETRY;
				}
				else
				{
					return_value = EIO;
				}
			}
			break;
		default:
		/*
		 * This case should never be reached.
		 */
		ASSERT(FALSE);
		return_value = EIO;
			disk_ptr->log_struct.error_id = ERRID_CDROM_ERR8;
			errsave(&disk_ptr->log_struct,
			sizeof(struct sc_error_log_df));
			disk_ptr->error_pending = FALSE;
			break;
	}
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_scsi_error");
	}
#endif
	return(return_value);
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_adapter_error                                              */
/*                                                                        */
/*   FUNCTION:  Logs and determines how to handle adapter errors.         */
/*                                                                        */
/*      This routine is called when a SCSI adapter error is detected and  */
/*      logs the error if necessary and determines what action to take.   */
/*      If the command should be retried, CD_RETRY is returned.  If the   */
/*      device should be reset before a retry is attempted, CD_RESET is   */
/*      returned.  If a timer should be set before the reset is done,     */
/*      CD_WAIT will be returned.  Otherwise, the proper errno value to   */
/*      return to the user is returned to the caller.                     */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver.     */
/*      It can be called on an interrupt level, is pinned in memory,      */
/*      and can not page fault.                                           */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      buf_ptr        - pointer to the buf_block used to issue the cmd.  */
/*      disk_ptr       - pointer to the disk structure for this device.   */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:                                            */
/*      The return value is either the errno value that should be         */
/*      returned to the user,  a RESET flag to indicate that the device   */
/*      should be reset before the command is retried, a RETRY flag       */
/*      to indicate that the command should be retried, a WAIT flag to    */
/*      indicate that a timer should be set before the reset is attempted,*/
/*      or a 0 to indicate a recovered error.                             */
/*                                                                        */
/*   ERROR DESCRIPTION: The following errno values may be returned to     */
/*      the calling routine (not directly to the user):                   */
/*      EIO       - SCSI bus fault.                                       */
/*                - No device response.                                   */
/*                - Host i/o bus error.                                   */
/*                - SCSI adapter hardware failure.                        */
/*                - SCSI adapter software failure.                        */
/*                - SCSI adapter fuse or terminal power.                  */
/*                - SCSI bus reset (stuck on).                            */
/*      ETIMEDOUT - The command timed out.                                */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  errsave                                 */
/*                                                                        */
/**************************************************************************/
uchar cd_adapter_error( struct cd_buf_block *buf_ptr,
				struct cd_disk_df *disk_ptr)
{
	uchar                   return_value;
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_adapter_error");
	}
#endif
	/*
	 * The error log structure is filled out if it has not already been
	 * done (if this is the second failure, it will already be filled in).
	 */
	if (!disk_ptr->error_pending)
	{
		bcopy(disk_ptr->disk_ddi.resource_name, disk_ptr->
			log_struct.resource_name, 8);
		bcopy(&buf_ptr->scsi_buf.scsi_command, &disk_ptr->
			log_struct.scsi_command, sizeof(struct scsi));
		disk_ptr->log_struct.status_validity = buf_ptr->
			scsi_buf.status_validity;
		disk_ptr->log_struct.scsi_status = buf_ptr->
			scsi_buf.scsi_status;
		disk_ptr->log_struct.general_card_status = buf_ptr->
			scsi_buf.general_card_status;
		disk_ptr->log_struct.reserved1 = 0x00;
		bzero(disk_ptr->log_struct.req_sense_data, 128);
		/*
		 * The first reserved word is used to store all the pending
		 * flags: open_pending, error_pending, ioctl_pending,
		 * reset_pending. The second reserved word is not used.
		 */
		disk_ptr->log_struct.reserved2 =
					((uint)disk_ptr->open_pending << 24) |
       				((uint)disk_ptr->error_pending << 16) |
       				((uint)disk_ptr->ioctl_pending << 8) |
       				((uint)disk_ptr->reset_pending);
		disk_ptr->log_struct.reserved3 = 0x00;
	}
	if (buf_ptr->scsi_buf.status_validity & SC_ADAPTER_ERROR)
	{
		switch(buf_ptr->scsi_buf.general_card_status)
		{
			case SC_CMD_TIMEOUT:
			/*
			 * A timed out command will be retried.
			 * If retries are successful, the temporary
			 * error will be logged in cd_iodone.  If
			 * retries are not successful, the permanent
			 * error will the logged
			 * below.
			 */
				if (buf_ptr->retry_count < CD_MAX_RETRY)
				{
					buf_ptr->retry_count++;
					return_value = CD_RETRY;
					disk_ptr->log_struct.error_id =
					ERRID_CDROM_ERR6;
					disk_ptr->error_pending = TRUE;
				}
				else
				{
					return_value = ETIMEDOUT;
					disk_ptr->log_struct.error_id =
						ERRID_CDROM_ERR5;
					errsave(&disk_ptr->log_struct,
						sizeof(struct sc_error_log_df));
					disk_ptr->error_pending = FALSE;
				}
				break;
			case SC_SCSI_BUS_FAULT:
			/*
 			 * A SCSI bus fault error will be retried.
 			 * If retries are successful, the temporary error will
 			 * be logged in cd_iodone.  If the retries are not
 			 * successful, the permanent error will be logged below.
 			 * On the very last retry allowed, the transfer mode
 			 * will be changed from synchronous (default value
			 * NULL) to the aysnchronous mode (SC_ASYNC).
 			 */
				if ( buf_ptr->retry_count < CD_MAX_RETRY )
				{
					buf_ptr->retry_count++;
					return_value = CD_RETRY;
					disk_ptr->log_struct.error_id =
						ERRID_CDROM_ERR6;
					disk_ptr->error_pending = TRUE;

#ifdef DEBUG
		if (cd_debug)
		{
		printf("case SC_SCSI_BUS_FAULT\n");
		printf("Going through retry procedures\n");
		printf("New retry_count is %d\n",(int)buf_ptr->
		retry_count);
		printf("CD_MAX_RETRY IS %d\n",(int)CD_MAX_RETRY);
		}
#endif
					if ( buf_ptr->retry_count ==
							(CD_MAX_RETRY - 1) )
					{
#ifdef DEBUG
				if (cd_debug)
				{
				printf("Setting mode flags to try in");
				printf("SC_ASYNC mode\n");
				}
#endif
						disk_ptr->rs_buf.scsi_buf.
						scsi_command.flags = SC_ASYNC;
					}

				}
				else
				{
					return_value = EIO;
					disk_ptr->log_struct.error_id =
						ERRID_CDROM_ERR5;
					errsave(&disk_ptr->
						log_struct,
						sizeof(struct sc_error_log_df));
					disk_ptr->error_pending = FALSE;
				}
				break;
			case SC_NO_DEVICE_RESPONSE:
			/*
			 * If no device response, the command will be
			 * retried. If retries are successful, the
			 * temporary error will be logged in cd_iodone.
			 * If retries are not successful, the permanent
			 * error will the logged below.
			 */
				if (buf_ptr->retry_count < CD_MAX_RETRY)
				{
					buf_ptr->retry_count++;
					return_value = CD_RETRY;
					disk_ptr->log_struct.error_id =
						ERRID_CDROM_ERR6;
					disk_ptr->error_pending = TRUE;
				}
				else
				{
					return_value = EIO;
					disk_ptr->log_struct.error_id =
						ERRID_CDROM_ERR5;
					errsave(&disk_ptr->
						log_struct,
						sizeof(struct sc_error_log_df));
					disk_ptr->error_pending = FALSE;
				}
				break;
			case SC_HOST_IO_BUS_ERR:
			case SC_ADAPTER_HDW_FAILURE:
			case SC_ADAPTER_SFW_FAILURE:
			case SC_FUSE_OR_TERMINAL_PWR:
			/*
			 * All of these adapter errors are logged at the
			 * adapter driver level.  The command will be
			 * retried until retries are exhausted.  It will
			 * not be retried if in diag mode.
			 */
				if (buf_ptr->retry_count < CD_MAX_RETRY)
				{
					buf_ptr->retry_count++;
					return_value = CD_RETRY;
				}
				else
				{
					return_value = EIO;
				}
				break;
			case SC_SCSI_BUS_RESET:
			/*
			 * The device reset procedure will be attempted
		 	 * in case of a SCSI bus reset.  The first
			 * command should result in a unit attention.
			 * If in diag mode, the reset will not be attempted.
	 		 */
				if (disk_ptr->reset_count < CD_MAX_RESET_RETRY)
				{
					disk_ptr->reset_count++;
					if (disk_ptr->reset_count == 1)
					{
						return_value = CD_RESET;
					}
					else
					{
						return_value = CD_WAIT;
					}
				}
				else
				{
					return_value = EIO;
				}
				break;
			default:
			/*
			 * This case should never be reached.
			 */
			ASSERT(FALSE);
				return_value = EIO;
				disk_ptr->log_struct.error_id =ERRID_CDROM_ERR8;
				errsave(&disk_ptr->
					log_struct, sizeof(
					struct sc_error_log_df));
				disk_ptr->error_pending = FALSE;
				break;
		}
	}
	else
	{
	/*
	 * This case is only reached when there was a resid
	 * value set and no status was valid. The command will
	 * be retried until the retry count is exhausted.
	 * If retries are successful, the temporary error will
	 * be logged in cd_iodone.  If retries are not
	 * successful, the permanent error will the logged below.
	 */
		if (buf_ptr->retry_count < CD_MAX_RETRY)
		{
			buf_ptr->retry_count++;
			return_value = CD_RETRY;
			disk_ptr->log_struct.error_id = ERRID_CDROM_ERR6;
			disk_ptr->error_pending = TRUE;
		}
		else
		{
			return_value = EIO;
			disk_ptr->log_struct.error_id = ERRID_CDROM_ERR5;
			errsave(&disk_ptr->log_struct, sizeof(
					struct sc_error_log_df));
			disk_ptr->error_pending = FALSE;
		}
	}
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_adapter_error");
	}
#endif
	return(return_value);
}

/**************************************************************************/
/*                                                                        */
/*   NAME:  cd_iodone                                                     */
/*                                                                        */
/*   FUNCTION:  The iodone entry point (called by SCSI adapter driver).   */
/*                                                                        */
/*      This routine acts as the interrupt handler for this driver.       */
/*      There are several oddities about this routine.  First in the case */
/*      of read requests that are too big, a completed SCSI command does  */
/*      not necessarily mean that the user request is complete.  The      */
/*      active_resid field in the disk structure indicates how much of    */
/*      the original command remains.  Another problem is in the case of  */
/*      spanned commands. In this case, one SCSI command completion must  */
/*      result in calling iodone on more than one buf struct.             */
/*                                                                        */
/*      Before any other processing, this routine checks for a pass-thru  */
/*      command, and wakes up the ioctl process if that is the case.      */
/*      Next, the routine considers the good completion case.  The first  */
/*      condition checked is the case where a command resulted in a check */
/*      condition, the request sense was issued and just completed.       */
/*      The cd_scsi_error routine is called to determine how to handle the*/
/*      error.  In the case of an ioctl where the request sense data is   */
/*      needed, the SC_VALID_SENSE flag is set to indicate that the data  */
/*      is valid.  The returned error value is also checked to see if a   */
/*      recovered error is indicated.  In this case the command state is  */
/*      restored to its initial value (before the request sense) so that  */
/*      normal completion handling can be done.  Next the case of a read  */
/*      is handled by checking that the entire user request has completed.*/
/*      If not (the request was to big to be done in one command), the    */
/*      start routine is called again with the remainder of the request.  */
/*      Otherwise, iodone is called on all the bufs in the active chain.  */
/*      If there is an ioctl waiting to be executed, it is kicked off by  */
/*      calling start before the next read in the queue is executed.      */
/*      If the read was from an ioctl, iodone is not called, but the      */
/*      ioctl routine is woken up.  In either the normal or ioctl case,   */
/*      if the queue is not empty for this device, the first chain in     */
/*      the queue will be issued by calling the start routine.            */
/*                                                                        */
/*      If the command wasn't a read, it was either part of the reset     */
/*      procedure, a stop unit, release, or request sense.  If in the     */
/*      middle of a device reset, the next command in the reset procedure */
/*      needs to be kicked off.  If the reset procedure is done, either   */
/*      the process needs to be woken up (in the case of an open), or the */
/*      command issued before the reset needs to be re-issued. A complete */
/*      stop unit needs to kick off a release command.  A completed       */
/*      release command simply needs to have the process woken up.  The   */
/*      case of a request sense completion occurs only after a check      */
/*      condition occured on a previous command.  The error routine was   */
/*      called to determine how to handle the error.  If the request      */
/*      was a read, successful completion handling must be done for the   */
/*      portion of the read that was successful.  Then either the command */
/*      is retried, the device is reset, or it was a fatal error and will */
/*      be handled below.                                                 */
/*                                                                        */
/*      In the case of an error condition resulting from the command, the */
/*      first check is for an adapter error.  The error routine is called */
/*      to determine how to handle the error, and the command is retried, */
/*      the device is reset, or the error is fatal.  The next check is for*/
/*      scsi status of INTMD_GOOD or DEVICE_BUSY or    */
/*      RESERVATION_CONFLICT.  Any of these conditions are fatal.  The    */
/*      next check is for a check condition.  If this is the case, a      */
/*      request sense will be issued and it will go through the path      */
/*      mentioned above.                                                  */
/*                                                                        */
/*      For a fatal error, if a reset failed, a flag is set that the      */
/*      device is dead.  If the original request wasn't a read, the error */
/*      is set in the buf struct and the process is woken up.  For an     */
/*      ioctl, if the request sense data was requested, it is copied to   */
/*      another buffer.  The next command in the queue is then kicked off.*/
/*      In the case of a failed read, if a reset failed, iodone is called */
/*      on all the buf structs in the active queue and the normal queue.  */
/*      If the reset did not fail, the error value is set and iodone      */
/*      is called on the first buf struct in the active queue.  If the    */
/*      active queue is not empty, start is called for the remaining      */
/*      chain.  If the active queue is empty, then a waiting ioctl is     */
/*      started or the next command in the normal queue is started.       */
/*                                                                        */
/*   EXECUTION ENVIRONMENT:                                               */
/*      This routine is part of the bottom half of the device driver.     */
/*      It can be called on an interrupt level, is pinned in memory,      */
/*      and can not page fault.                                           */
/*                                                                        */
/*   DATA STRUCTURES:  none                                               */
/*                                                                        */
/*   INPUTS:                                                              */
/*      buf_ptr - the address of the buf struct, which is also the        */
/*                address of the sc_buf and the buf_block                 */
/*                                                                        */
/*   RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                        */
/*   ERROR DESCRIPTION: The following errno values may be returned in the */
/*      b_error field of the buf struct:                                  */
/*      EIO     - The request sense failed with a CHECK_CONDITION.        */
/*              - The scsi status was not expected (should never happen). */
/*      EBUSY   - Device is reserved by someone else, or device is busy.  */
/*      Plus others from cd_scsi_error or cd_adapter_error                */
/*                                                                        */
/*   EXTERNAL PROCEDURES CALLED:  iodone,                                 */
/*                                e_wakeup,                               */
/*                                devstrat                                */
/*                                                                        */
/**************************************************************************/

void cd_iodone( struct buf *bp )
{
	struct cd_buf_block     *buf_ptr;
	struct buf              *cur_bp, *next_bp;
	struct sc_buf           *scsi_buf_ptr;
	struct cd_disk_df       *disk_ptr;
	int                     return_code;
	uchar                   error;

#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nEntering cd_iodone");
	}
#endif
	/*
	 * The B_DONE flag must be cleared to force cd_iowait to wait
	 * on this buf struct.  The flag will be set when processing
	 * is complete.
	 */
	bp->b_flags &= ~B_DONE;
	buf_ptr = (struct cd_buf_block *) bp;
	scsi_buf_ptr = (struct sc_buf *) bp;
	disk_ptr = buf_ptr->disk_ptr;
	DDHKWD1(HKWD_DD_CDDD, DD_ENTRY_IODONE, 0, disk_ptr->devno);
	error = FALSE;
	if (disk_ptr->ioctl_pending)
	{
		cur_bp = &disk_ptr->rdse_buf_struct;
	}
	else
	{
		cur_bp = disk_ptr->head;
	}
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\n    cmd_state   = %X   old_cmd_state       = %X",
			disk_ptr->cmd_state, disk_ptr->old_cmd_state);
		printf("\n    bp->b_flags = %X   status_validity     = %X",
			bp->b_flags, scsi_buf_ptr->status_validity);
		printf("\n    bp->b_error = %X   scsi_status         = %X",
			bp->b_error, scsi_buf_ptr->scsi_status);
		printf("\n    bp->b_resid = %d   general_card_status = %X",
			bp->b_resid, scsi_buf_ptr->general_card_status);
	}
#endif
	if (disk_ptr->cmd_state == CD_PASS_THRU)
	{
	/*
	 * In the case of a pass-through command, the waiting process
	 * is simply woken up.  No error handling is done.
	 */
		bp->b_flags |= B_DONE;
		e_wakeup(&bp->b_event);
		disk_ptr->ioctl_pending = FALSE;
	}
	else
	{
		if (!(bp->b_flags & B_ERROR) && ((bp->b_resid == 0) || 
			(disk_ptr->cmd_state == SCSI_MODE_SENSE)))
		{
		/*
		 * The command completed successfully.
		 */
			if (disk_ptr->cmd_state == SCSI_REQUEST_SENSE)
			{
			/*
			 * The command was a request sense.  If this
			 * was a read with sense, the valid_sense flag
			 * is set and the sense data is moved to a
			 * different buffer to be sent to the user.
			 * The scsi_error routine is called and
			 * if it was a recovered error, the command
			 * state is restored to its original value so
			 * that normal completion of the original
			 * command can proceed.
			 */
#ifdef DEBUG
	if (cd_debug)
	{
	printf("\nIssued a successful request sense.");
	printf("Sense_key = %X  Addl. sense = %X", disk_ptr->req_sense_data.
	sense_key, disk_ptr->req_sense_data.add_sense_key);
	}
#endif
				if (disk_ptr->ioctl_pending && disk_ptr->
					ioctl_req_sense.count > 0)
				{ /* CDIORDSE */
					return_code = xmemout(&disk_ptr->
								req_sense_data,
							disk_ptr->
								ioctl_req_sense.
								buffer,
							disk_ptr->
								ioctl_req_sense.
								count,
							&disk_ptr->
								ioctl_req_sense.
								xmemd);
					if (return_code == XMEM_SUCC)
					{
						disk_ptr->
							ioctl_buf.scsi_buf.
							status_validity |=
							SC_VALID_SENSE;
					}
				}
				error = cd_scsi_error(disk_ptr);
				if (error == 0)
				{ /* RECOVERED_ERROR */
					if (!disk_ptr->reset_pending)
					{
						disk_ptr->cmd_state = disk_ptr->
							old_cmd_state;
						if (disk_ptr->ioctl_pending)
						{
							bp = &disk_ptr->
								ioctl_buf.
								scsi_buf.
								bufstruct;
						}
						else
						{
							bp = &disk_ptr->
								norm_buf.
								scsi_buf.
								bufstruct;
						}
						buf_ptr =
						      (struct cd_buf_block *)bp;
						scsi_buf_ptr =
							(struct sc_buf *)bp;
						disk_ptr = buf_ptr->disk_ptr;
					}
					else
					{
						disk_ptr->cmd_state = disk_ptr->
								reset_state;
					}
				}
			}
			switch (disk_ptr->cmd_state)
			{
				case SCSI_READ_EXTENDED:
				/*
				 * The command was a read.  Decrement
				 * the total byte count to see if the
				 * whole command is complete or if the
				 * command had to be broken up.
				 */
					if (((uint)bp->b_un.b_addr + bp->
						b_bcount) == ((uint)cur_bp->
						b_un.b_addr + cur_bp->
						b_bcount - cur_bp->b_resid))
					{
					/*
					 * The user command is complete.
					 */
						if (disk_ptr->ioctl_pending)
						{
						/*
						 * This was a read with sense.
						 * Wake up the ioctl process and
						 * set a flag if a normal
						 * read is waiting. No error
						 * logging or handling given
						 * during an ioctl read.
						 */
							disk_ptr->ioctl_buf.
								scsi_buf.
								bufstruct.
								b_flags |=
									B_DONE;
							e_wakeup(&disk_ptr->
								ioctl_buf.
								scsi_buf.
								bufstruct.
								b_event);
							disk_ptr->
								ioctl_pending =
									FALSE;
							if (disk_ptr->head !=
								NULL)
							{
 							cd_start(
							   disk_ptr,(uchar)
							   CD_ORIGINAL_REQUEST);
							DDHKWD1(HKWD_DD_CDDD,
							    DD_EXIT_IODONE, 0,
							    disk_ptr->devno);
							return;
							}
						}
						else
						{
						/*
						 * This was a normal read. Call
						 * iodone on the buf struct and
						 * start the next ioctl if one
						 * is waiting or set the flag if
						 * there is a read command
						 * waiting. If a recovered cmd,
						 * log the error.
						 */
							if (disk_ptr->
								error_pending)
							{
							errsave(&disk_ptr->
							      log_struct,
							      sizeof(struct
							      sc_error_log_df));
							disk_ptr->
							      error_pending =
									FALSE;
							}
							disk_ptr->head =
								disk_ptr->head->
									av_forw;
							iodone(cur_bp);
							if (disk_ptr->
								ioctl_rqst)
							{
							   disk_ptr->
								ioctl_rqst =
									FALSE;
							   disk_ptr->
							 	ioctl_pending =
									TRUE;
							   cd_start(disk_ptr,
						       		(uchar)
							   CD_ORIGINAL_REQUEST);
							   DDHKWD1(HKWD_DD_CDDD,
								DD_EXIT_IODONE,
								0, disk_ptr->
								devno);
							    return;
							}
							if (disk_ptr->head !=
								NULL)
							{
							   cd_start(disk_ptr,
								(uchar)
							   CD_ORIGINAL_REQUEST);
							   DDHKWD1(HKWD_DD_CDDD,
								DD_EXIT_IODONE,
								0, disk_ptr->
								devno);
							   return;
							}
						}
					}
					else
					{
					/*
					 * This was a command that had to be
					 * broken up and is not actually
					 * complete yet.  The rest of the
					 * command is  started.
					 */
						cd_start(disk_ptr,(uchar)
							CD_CONTINUED_REQUEST);
						DDHKWD1(HKWD_DD_CDDD,
							DD_EXIT_IODONE, 0,
							disk_ptr->devno);
						return;
					}
					break;
				case SCSI_TEST_UNIT_READY:
				/*
				 * A reset is in process and the next command
				 * needs to be started. Log the error if
				 * recovery was  necessary.
				 */
					if (disk_ptr->error_pending)
					{
						errsave(&disk_ptr->log_struct,
							sizeof(struct
							sc_error_log_df));
						disk_ptr->error_pending = FALSE;
					}
					buf_ptr->retry_count = 0;
					/*
					 * If the target is to 'lock out' other
					 * initiators to the device, then the
					 * 'reserve_lock' flag will be true
					 * and cd_reserve will be called.  
					 * The default value of 'reserve_lock'
					 * is set to false in the configuration
					 * routines for CDROM, and we skip to
					 * the next routine.  So by default,
					 * 'twin-tailing' of CDROM devices is
					 * allowed.  Note that this default
					 * value makes the command 
					 * SC_RETAIN_RESERVATION useless unless
					 * the 'reserve_lock' flag is changed
					 * to true.
					 */
					if (disk_ptr->reserve_lock)
						cd_reserve(disk_ptr);
					else {
						disk_ptr->m_sense_status = 
						   CD_SENSE_CHANGEABLE;
						cd_mode_sense(disk_ptr);
					}
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
							0, disk_ptr->devno);
					return;
				case SCSI_RESERVE_UNIT:
				/*
				 * A reset is in process and the next
				 * command needs to be started.
				 * Log the error if recovery was necessary.
				 */
					if (disk_ptr->error_pending)
					{
						errsave(&disk_ptr->log_struct,
							sizeof(struct
							sc_error_log_df));
						disk_ptr->error_pending = FALSE;
					}
					buf_ptr->retry_count = 0;
					disk_ptr->m_sense_status = 
						CD_SENSE_CHANGEABLE;
					cd_mode_sense(disk_ptr);
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
							0, disk_ptr->devno);
					return;
				case SCSI_MODE_SENSE:
				/*
				 * A Mode sense command has completed
				 */
					if (disk_ptr->error_pending)
					{
						errsave(&disk_ptr->log_struct,
							sizeof(struct
							sc_error_log_df));
						disk_ptr->error_pending = FALSE;
					}
					buf_ptr->retry_count = 0;
			                if (disk_ptr->m_sense_status == 
					    CD_SENSE_CHANGEABLE) {
			                   /*
			                    * if this was the sense of 
					    * changeable attributes, then 
					    * prepare to sense current 
					    * attributes. First, copy 
					    * changeable data into separate 
					    * data buffer, and format the data
			                    */
			                   bcopy (disk_ptr->sense_buf, 
						disk_ptr->ch_data, 256);
			                   cd_format_mode_data(
					      disk_ptr->ch_data, &disk_ptr->ch,
					      (int) (disk_ptr->ch_data[0]+1));
			                   disk_ptr->m_sense_status = 
						CD_SENSE_CURRENT;
			                   cd_mode_sense(disk_ptr);
			                } else {
			                   /*
			                    * this was the sense of current 
					    * data, so format the data and then
					    * determine if a select is 
					    * necessary
			                    */
			                   cd_format_mode_data(
					      disk_ptr->sense_buf, 
					      &disk_ptr->cd,
					      (int)(disk_ptr->sense_buf[0]+1));
			                   if (cd_mode_data_compare(disk_ptr))
			                           /*
			                            * if a mode select is 
						    * necessary
			                            */
			                           cd_mode_select(disk_ptr);
			                   else
			                           /*
			                            * bypass the mode select
			                            */
			                           cd_read_capacity(disk_ptr);
			                }
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
						0, disk_ptr->devno);
			
					return;
				case SCSI_MODE_SELECT:
				/*
				 * A reset is in process and the next
				 * command needs to be started.
				 * Log the error if recovery was necessary.
				 */
					if (disk_ptr->error_pending)
					{
						errsave(&disk_ptr->log_struct,
							sizeof(struct
							sc_error_log_df));
						disk_ptr->error_pending = FALSE;
					}
					buf_ptr->retry_count = 0;
					cd_read_capacity(disk_ptr);
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
						0, disk_ptr->devno);
					return;
				case SCSI_READ_CAPACITY:
				/*
				 * A reset is in process and the next
				 * command needs to be started.
				 * Log the error if recovery was necessary.
				 */
					if (disk_ptr->error_pending)
					{
						errsave(&disk_ptr->log_struct,
							sizeof(struct
							sc_error_log_df));
						disk_ptr->error_pending = FALSE;
					}
					buf_ptr->retry_count = 0;
					cd_start_stop(disk_ptr, scsi_buf_ptr,
							(uchar)CD_START_UNIT);
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
							0, disk_ptr->devno);
					return;
				case SCSI_START_STOP_UNIT:
				/*
				 * Log the error if recovery was  necessary.
				 */
					if (disk_ptr->error_pending)
					{
						errsave(&disk_ptr->log_struct,
							sizeof(struct
							sc_error_log_df));
						disk_ptr->error_pending = FALSE;
					}
					if (disk_ptr->reset_pending)
					{
					/*
					 * A reset just completed. Either this
					 * is the open process and the open
					 * needs to be woken up, or this was a
					 * recovery reset and the original
					 * command needs to be retried.
					 */
						disk_ptr->reset_pending = FALSE;
						disk_ptr->reset_count = 0;
						/* successful reset */
						if (disk_ptr->open_pending)
						{
							disk_ptr->
							   reset_buf.scsi_buf.
							   bufstruct.b_flags |=
								B_DONE;
							e_wakeup(&disk_ptr->
							   reset_buf.scsi_buf.
							   bufstruct.b_event);
						}
						else if(disk_ptr->ioctl_pending)
						{
							error = EIO;
							disk_ptr->cmd_state =
								disk_ptr->
								old_cmd_state;
						}
						else
						{
						/*
						 * Re-issue command.
						 */
							disk_ptr->cmd_state =
								disk_ptr->
								old_cmd_state;
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nRetrying the original command that failed");
		printf("with a UNIT ATTENTION");
	}
#endif
							(void)devstrat(
							    &disk_ptr->norm_buf.
							    scsi_buf.bufstruct);
							DDHKWD1(HKWD_DD_CDDD,
							    DD_EXIT_IODONE,
							    0, disk_ptr->devno);
							return;
						}
					}
					else
					{
					/*
					 * A stop unit just completed. The
					 * release needs to be issued, unless
					 * the retain-reservation flag is set.
					 */
						if (!disk_ptr->
							retain_reservation)
						{
							cd_release(disk_ptr);
							DDHKWD1(HKWD_DD_CDDD,
								DD_EXIT_IODONE,
								0, disk_ptr->
								devno);
							return;
						}
						else
						{
							bp->b_flags |= B_DONE;
							e_wakeup(&bp->b_event);
						}
					}
					break;
				case SCSI_RELEASE_UNIT:
				/*
				 * The release just completed.  Wake up the
				 * waiting process. Log the error if recovery
				 * was necessary.
				 */
					if (disk_ptr->error_pending)
					{
						errsave(&disk_ptr->log_struct,
							sizeof(struct
							sc_error_log_df));
						disk_ptr->error_pending = FALSE;
					}
					bp->b_flags |= B_DONE;
					e_wakeup(&bp->b_event);
					break;
				case SCSI_REQUEST_SENSE:
				/*
				 * The command was a request sense (and not a
				 * recovered error).  The error routine has
				 * already been called and the return value is
				 * in 'error'.
				 */
					if (error == CD_BYPASS) {
					  /*
					   * Treat this as NO ERROR, since
					   * this is from an illegal request
					   * on a mode sense or select.  just
					   * bypass to the read capacity.
					   */
					   if (disk_ptr->error_pending)
					   {
						errsave(&disk_ptr->log_struct,
							sizeof(struct
							sc_error_log_df));
						disk_ptr->error_pending = FALSE;
					   }
					   buf_ptr->retry_count = 0;
					   cd_read_capacity(disk_ptr);
					   DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
						0, disk_ptr->devno);
					   return;
					}
					if (!disk_ptr->reset_pending)
					{
						disk_ptr->cmd_state = disk_ptr->
							old_cmd_state;
					}
					else
					{
						disk_ptr->cmd_state = disk_ptr->
							reset_state;
					}
					if (error == CD_WAIT)
					{
					/*
					 * A reset needs to be attempted
					 * but the device is not ready
					 * so timeout is called to
					 * schedule the reset in 5 secs.
					 */
						w_start(&disk_ptr->
								watchdog_timer);
						DDHKWD1(HKWD_DD_CDDD,
							DD_EXIT_IODONE,
							0, disk_ptr->devno);
						return;
					}
					if (error == CD_RESET)
					{
					/*
					 * A reset needs to be attempted
					 * so call the routine to start the
					 * procedure.
					 */
						cd_device_reset(disk_ptr);
						DDHKWD1(HKWD_DD_CDDD,
							DD_EXIT_IODONE,
							0, disk_ptr->devno);
						return;
					}
					if (error == CD_RETRY)
					{
					/*
					 * The command should be retried so call
 					 * devstrat.
					 */
						if (disk_ptr->cmd_state ==
							SCSI_READ_EXTENDED)
						{
							(void)devstrat(
							    &disk_ptr->norm_buf.
							    scsi_buf.bufstruct);
						}
						else
						{
							(void)devstrat(
							     &disk_ptr->
							     reset_buf.scsi_buf.
							     bufstruct);
						}
						DDHKWD1(HKWD_DD_CDDD,
							DD_EXIT_IODONE,
							0, disk_ptr->devno);
						return;
					}
					/*
					 * A scsi bus reset recovery failed.
					 * The cmd_state is set to the original
					 * command for fatal error processing.
					 */
					if (disk_ptr->reset_pending &&
						!disk_ptr->open_pending)
					{
						disk_ptr->cmd_state = disk_ptr->
							old_cmd_state;
					}
					break;
			}
		}
		else
		{
		/*
		 * The command resulted in some sort of error (B_ERROR
		 * is set or there is a resid value).  First check for
		 * an adapter error, and then for a SCSI error.
		 */
			bp->b_flags &= ~B_ERROR;
			if (scsi_buf_ptr->status_validity & SC_ADAPTER_ERROR)
			{
			/*
			 * An adapter error was detected.  Call the
			 * adapter_error routine to determine if the
			 * command should be retried, the device should
			 * be reset, or if the error is fatal.
			 */
				error = cd_adapter_error(buf_ptr, disk_ptr);
				if (!disk_ptr->reset_pending && disk_ptr->
					cmd_state != SCSI_REQUEST_SENSE)
				{
					disk_ptr->old_cmd_state = disk_ptr->
								     cmd_state;
				}
				if (error == CD_WAIT)
				{
					w_start(&disk_ptr->watchdog_timer);
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
						0, disk_ptr->devno);
					return;
				}
				if (error == CD_RESET)
				{
					cd_device_reset(disk_ptr);
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
						0, disk_ptr->devno);
					return;
				}
				if (error == CD_RETRY)
				{
				/* call devstrat to re-issue the command */
					(void) devstrat(bp);
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
 						0, disk_ptr->devno);
					return;
				}
			}
			else if (scsi_buf_ptr->status_validity & SC_SCSI_ERROR)
			{ /* valid SCSI status */
			/*
			 * There was no adapter error but there was
			 * a SCSI error, so check the SCSI status.
			 */
				switch (scsi_buf_ptr->scsi_status)
				{
					case SC_RESERVATION_CONFLICT:
					case SC_BUSY_STATUS:
					/*
					 * There is no reason to retry
					 * the command if it is reserved
					 * or busy (fatal error).
					 */
						error = EBUSY;
						break;
					case SC_CHECK_CONDITION:
						if (disk_ptr->cmd_state != 
							SCSI_REQUEST_SENSE)
						{
						/*
						 * The command resulted in a
						 * check condition so issue the
						 * request sense.
						 */
							disk_ptr->rs_buf.
								retry_count = 0;
							if (!disk_ptr->
								reset_pending)
							{
							/*
							 * This is not a reset
							 * attempt, so save the 
							 * command state in case
							 * it needs to be
							 * retried.
							 */
							disk_ptr->old_cmd_state=
									disk_ptr->cmd_state;
							}
							else
							{
							/*
							 * This is a reset
							 * attempt,so the
							 * old_cmd_state will
							 * contain the original
							 * command. reset_state
							 * is used to save the
							 * current position
							 * of the reset.
							 */
							disk_ptr->reset_state =
								disk_ptr->
								     cmd_state;
							}
							disk_ptr->cmd_state =
							     SCSI_REQUEST_SENSE;
#ifdef DEBUG
	if (cd_debug)
	{
	printf("\nIssuing the request sense because of a CHECK_CONDITION");
	}
#endif
							/*
 							 * call devstrat with
							 * r_s_buf_ptr
 							 */
							(void)devstrat(
								&disk_ptr->
								rs_buf.scsi_buf.
								bufstruct);
							DDHKWD1(HKWD_DD_CDDD,
								DD_EXIT_IODONE,
								0, disk_ptr->
								devno);
							return;
						}
						else
						{
						/*
						 * A request sense cmd resulted 
						 * in a check-condition.  Retry
						 * the command until retries are
						 * exhausted.  Then set the
						 * reset_failed flag to indicate
						 * the device is dead.
						 */
							if (disk_ptr->rs_buf.
								retry_count <
								CD_MAX_RETRY)
							{
								disk_ptr->
								  rs_buf.
								  retry_count++;
								/*
								 * call devstrat
								 * with 
								 * r_s_buf_ptr 
								 */
								(void)devstrat(
								     &disk_ptr->
								     rs_buf.
								     scsi_buf.
								     bufstruct);
								DDHKWD1(
								 HKWD_DD_CDDD,
								 DD_EXIT_IODONE,
								 0, disk_ptr->
								 devno);
								return;
							}
							else
							{
								error = EIO;
								disk_ptr->
								  reset_failed =
									TRUE;
								disk_ptr->
								  failed_errno =
									EIO;
							}
						}
						break;
					case SC_GOOD_STATUS:
					default:
					/*
					 * This point should never be
					 * reached. All the valid status
					 * values should be handled above.
					 */
						ASSERT(FALSE);
						error = EIO;
						break;
				}
			}
			else
			{
			/*
			 * The adapter indicated an error by setting the resid
			 * value with no other status.  Call adapter_error
			 * routine to determine if the command should be
			 * retried, the device should be reset, or if the error
			 * is fatal.
			 */
				error = cd_adapter_error(buf_ptr, disk_ptr);
				if (!disk_ptr->reset_pending && disk_ptr->
					cmd_state != SCSI_REQUEST_SENSE)
				{
					disk_ptr->old_cmd_state = disk_ptr->
								     cmd_state;
				}
				if (error == CD_WAIT)
				{
					w_start(&disk_ptr->watchdog_timer);
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
							0, disk_ptr->devno);
					return;
				}
				if (error == CD_RESET)
				{
					cd_device_reset(disk_ptr);
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
							0, disk_ptr->devno);
					return;
				}
				if (error == CD_RETRY)
				{
				/* call devstrat to re-issue the command */
					(void) devstrat(bp);
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
							0, disk_ptr->devno);
					return;
				}
			}
			/*
			 * If the command was a request sense, the command state
			 * is restored to its original value before the fatal
			 * error processing below.
			 */
			if (disk_ptr->cmd_state == SCSI_REQUEST_SENSE)
			{
				if (!disk_ptr->open_pending)
				{
				/*
				 * The original command was the one
				 * issued before the req. sense was  issued.
				 */
					disk_ptr->cmd_state = disk_ptr->
								old_cmd_state;
				}
				else
				{
				/*
				 * An open was in progress, so the "original"
				 * command was the reset procedure.
				 */
					disk_ptr->cmd_state = disk_ptr->
								reset_state;
				}
			}
		}
		if (error)
		{
		/*
		 * The error was not recoverable.
		 */
			if (disk_ptr->reset_pending)
			{
			/*
			 * A reset was in progress and could not
			 * complete.  The reset_failed flag is set
			 * to indicate that the device is dead.
			 */
				disk_ptr->reset_failed = TRUE;
				disk_ptr->failed_errno = error;
				disk_ptr->reset_pending = FALSE;
			}
			if (disk_ptr->cmd_state == SCSI_START_STOP_UNIT &&
			    disk_ptr->norm_buf.scsi_buf.scsi_command.scsi_cmd.scsi_bytes[2] ==
			    (uchar)CD_STOP_UNIT)
			{
			/*
			 * The stop unit command failed.
			 * If the retain reservation flag is
			 * not set, issue the release command
			 * to continue the close process
			 * regardless of the error.
			 */
				if (!disk_ptr->retain_reservation)
				{
					cd_release(disk_ptr);
					DDHKWD1(HKWD_DD_CDDD,
							DD_EXIT_IODONE,
							0, disk_ptr->
							devno);
					return;
				}
				else
				{
				/*
				 * If the retain reservation
				 * flag is set, set the
				 * cmd_state so execution will
				 * continue as if the release
				 * had completed.
				 */
					disk_ptr->cmd_state =
						SCSI_RELEASE_UNIT;
				}
			}
			if (disk_ptr->cmd_state == SCSI_RELEASE_UNIT)
			{
			/*
			 * The release command failed.  Wake up the
			 * waiting process (failed open or close).
			 */
				disk_ptr->norm_buf.scsi_buf.bufstruct.b_error =
					error;
				disk_ptr->norm_buf.scsi_buf.bufstruct.b_flags |=
					B_ERROR | B_DONE;
				e_wakeup(&disk_ptr->norm_buf.scsi_buf.bufstruct.
					b_event);
			}
			else if (disk_ptr->ioctl_pending)
			{
			/*
			 * The command was an ioctl.  Set the error
			 * values and wake up the ioctl process.
			 */
				disk_ptr->ioctl_buf.scsi_buf.bufstruct.b_error =
					error;
				disk_ptr->ioctl_buf.scsi_buf.bufstruct.b_flags|=
					(B_ERROR | B_DONE);
				e_wakeup(&disk_ptr->ioctl_buf.scsi_buf.
					bufstruct.b_event);
				disk_ptr->ioctl_pending = FALSE;
				if (disk_ptr->reset_failed)
				{
					while (disk_ptr->head != NULL)
					{
						next_bp = disk_ptr->head->
							     av_forw;
						disk_ptr->head->b_error = error;
						disk_ptr->head->b_resid =
							     disk_ptr->head->
								       b_bcount;
						disk_ptr->head->b_flags |=
							     (B_ERROR | B_DONE);
						iodone(disk_ptr->head);
						disk_ptr->head = next_bp;
					}
				}
				else if (disk_ptr->head != NULL)
				{
				/*
				 * There is a command waiting in the
				 * queue, so issue the command.
				 */
					cd_start(disk_ptr,(uchar)
							CD_ORIGINAL_REQUEST);
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
							0, disk_ptr->devno);
					return;
				}
			}
			else if (disk_ptr->open_pending)
			{
			/*
			 * The open failed, so set the error values and
			 * wake up the sleeping open process.
			 */
				disk_ptr->reset_buf.scsi_buf.bufstruct.b_error =
					error;
				disk_ptr->reset_buf.scsi_buf.bufstruct.b_flags|=
					(B_ERROR | B_DONE);
				e_wakeup(&disk_ptr->reset_buf.scsi_buf.
					bufstruct.b_event);
			}
			else
			{
			/*
			 * A normal read command failed.
			 * Set the error values and call iodone on the buf
			 * struct.
			 */
				cur_bp->b_error = error;
				cur_bp->b_flags |= B_ERROR;
				cur_bp->b_resid += disk_ptr->norm_buf.scsi_buf.
					bufstruct.b_resid + (cur_bp->b_un.
					b_addr + cur_bp->b_bcount) - (disk_ptr->
					norm_buf.scsi_buf.bufstruct.b_un.b_addr+
					disk_ptr->norm_buf.scsi_buf.bufstruct.
					b_bcount);
				disk_ptr->head = disk_ptr->head->av_forw;
				iodone(cur_bp);
				if (disk_ptr->reset_failed)
				{
				/*
				 * The device has been flagged as
				 * 'dead' so call iodone on all the
				 * buf structs waiting in the queue.
				 * Also check if there is an ioctl
				 * waiting to get started and wake it up if it
				 * is.
				 */
					while (disk_ptr->head != NULL)
					{
						next_bp=disk_ptr->head->av_forw;
						disk_ptr->head->b_error = error;
						disk_ptr->head->b_resid =
						       disk_ptr->head->b_bcount;
						disk_ptr->head->b_flags |=
						       B_ERROR;
						iodone(disk_ptr->head);
						disk_ptr->head = next_bp;
					}
					if (disk_ptr->ioctl_rqst)
					{
						disk_ptr->ioctl_buf.scsi_buf.
							bufstruct.b_error=error;
						disk_ptr->ioctl_buf.scsi_buf.
							bufstruct.b_flags |=
								B_ERROR;
						e_wakeup(&disk_ptr->ioctl_buf.
							scsi_buf.bufstruct.
							b_event);
						disk_ptr->ioctl_rqst = FALSE;
					}
				}
				else if (disk_ptr->ioctl_rqst)
				{
				/*
				 * There is an ioctl request waiting
				 * to get started, so start it off.
				 */
					disk_ptr->ioctl_rqst = FALSE;
					disk_ptr->ioctl_pending = TRUE;
					cd_start(disk_ptr,(uchar)
						CD_ORIGINAL_REQUEST);
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
							0, disk_ptr->devno);
					return;
				}
				else if (disk_ptr->head != NULL)
				{
				/*
				 * The device is not dead, so issue the next
				 * command on the queue
				 */
					cd_start(disk_ptr,(uchar)
						CD_ORIGINAL_REQUEST);
					DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE,
						0, disk_ptr->devno);
					return;
				}
			}
		}
	}
	/*
	 * There is no ioctl waiting to start, and no read requests waiting
	 * in the queue, so the device is no longer busy and the busy flag
	 * can be cleared.
	 */
	disk_ptr->busy = FALSE;
	disk_ptr->cd_dkstat.dk_status &= ~IOST_DK_BUSY;
#ifdef DEBUG
	if (cd_debug)
	{
		printf("\nLeaving cd_iodone");
	}
#endif
DDHKWD1(HKWD_DD_CDDD, DD_EXIT_IODONE, 0, disk_ptr->devno);
return;
}
