/* @(#)46       1.8.2.1  R2/sysx/disk/scdiskdd.h, sysxdisk, bos320, 9230320a 7/19/92 16:42:18 */
#ifndef _H_SCDISKDD
#define _H_SCDISKDD
/*
 * COMPONENT_NAME: (SYSXDISK) SCSI Disk Device Driver Include File
 *
 * FUNCTIONS:  NONE
 *
 * ORIGINS: 27
 *
 * (C) COPYRIGHT International Business Machines Corp. 1988, 1989, 1990
 *		 1992
 * All Rights Reserved
 * Licensed Materials - Property of IBM
 *
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */

/************************************************************************/
/* scdisk.h header dependencies                                         */
/************************************************************************/

#include <sys/types.h>
#include <sys/buf.h>
#include <sys/file.h>
#include <sys/conf.h>
#include <sys/lockl.h>
#include <sys/device.h>
#include <sys/scsi.h>
#include <sys/watchdog.h>
#include <sys/iostat.h>
#include <sys/dump.h>
#include <sys/scdisk.h>

#ifdef DEBUG
#define DKprintf(args) if (scdisk_debug) printf args
#endif

/*
 * Debug Trace length
 */
#ifdef DEBUG
#define TRCLNGTH                1000    /* Length of debug trace table       */
#endif

/*
 * Debug Trace Table Structure
 */
#ifdef DEBUG
struct scdisk_trace {
#define SCDISK_TRC_STR_LENGTH       8
        char    desc[SCDISK_TRC_STR_LENGTH];/* ASCII descrip of this entry   */
        char    type[3];                    /* ASCII desrip of entry or exit */
        char    count;                      /* which of several entries      */
        uint    word1;                      /*  meaning depends on the desc  */
        uint    word2;                      /*  meaning depends on the desc  */
        uint    word3;                      /*  meaning depends on the desc  */
        uint    word4;                      /*  meaning depends on the desc  */
        uint    word5;                      /*  meaning depends on the desc  */
};
#endif

/************************************************************************/
/* X3.131 SCSI Standard Sense Key Values 				*/
/************************************************************************/
#define DK_NO_SENSE		0x00
#define DK_RECOVERED_ERROR	0x01
#define DK_NOT_READY		0x02
#define DK_MEDIUM_ERROR		0x03
#define DK_HARDWARE_ERROR	0x04
#define DK_ILLEGAL_REQUEST	0x05
#define DK_UNIT_ATTENTION	0x06
#define DK_DATA_PROTECT		0x07
#define DK_BLANK_CHECK		0x08
#define DK_VENDOR_UNIQUE	0x09
#define DK_COPY_ABORTED		0x0A
#define DK_ABORTED_COMMAND	0x0B
#define DK_EQUAL_CMD		0x0C
#define DK_VOLUME_OVERFLOW	0x0D
#define DK_MISCOMPARE		0x0E

/************************************************************************/
/* Misc. Defines 							*/
/************************************************************************/
#define DK_BLOCKSIZE	512		/* Disk blocksize 		*/
#define DK_MAX_RETRY    5               /* maximum error retry count(+1)*/
#define DK_MAX_RESETS	3		/* maximum number of reset      */
					/* cycles on an error           */
#define DK_MAX_RESET_FAILURES 2		/* maximum number of reset cycle*/
					/* failures before we give up   */
#define DK_MAX_DISKS	255		/* maximum disks supported	*/
#define DK_TIMEOUT	30		/* Default Timeout period (secs)*/
#define DK_HASH		0x0F
#define DK_HASHSIZE	0x10
#define DK_REQSNS_LEN	0xFF
#define DK_ERROR	-1
#define DK_CMD_Q        0x1             /* Use the disks pending Queue  */
#define DK_QUEUE	0x1		/* Place cmd_ptr on end of Q    */
#define DK_STACK	0x2             /* Place cmd_ptr on front of Q  */

/* Queue types */
#define DK_PENDING	0x01
#define DK_INPROG	0x02

/* SCSI Disk Error types */
#define DK_MEDIA_ERR		0x01
#define DK_HARDWARE_ERR		0x02
#define DK_ADAPTER_ERR		0x03
#define DK_RECOVERED_ERR	0x04
#define DK_UNKNOWN_ERR		0x05

/* SCSI Disk Error severities */
#define DK_HARD_ERROR		0x01
#define DK_SOFT_ERROR		0x02

/* SCSI Disk Error severities */
#define DK_OVERLAY_ERROR	0x01
#define DK_NOOVERLAY_ERROR	0x02

/* SCSI Disk Mode Sense/Select defines */
#define DK_MAX_MODE_PAGES	0x3f	/* Maximum number of mode page */
#define DK_BLK_DESC_LN_INDEX	3	/* Index to block descriptor length */
#define DK_COMPLIES_PF		0x10	/* Indicates mode pages comply to*/
					/* Page Format			 */

/************************************************************************/
/* Macro definitions							*/
/************************************************************************/

#define SCDISK_IN_PROG_ENQUEUE(diskinfo, bp)				\
{									\
	if (diskinfo->dk_bp_queue.in_progress.head == NULL) {		\
		diskinfo->dk_bp_queue.in_progress.head = bp;		\
		diskinfo->dk_bp_queue.in_progress.tail = bp;		\
		bp->av_forw = NULL;					\
	} else {							\
		diskinfo->dk_bp_queue.in_progress.tail->av_forw = bp;	\
		bp->av_forw = NULL;					\
		diskinfo->dk_bp_queue.in_progress.tail = bp;		\
	}								\
}

#define SCDISK_IN_PROG_DEQUEUE(diskinfo, bp)				\
{									\
       if (diskinfo->dk_bp_queue.in_progress.head != NULL) {		\
	       diskinfo->dk_bp_queue.in_progress.head = bp->av_forw;	\
	       if (diskinfo->dk_bp_queue.in_progress.head == NULL) {	\
		       diskinfo->dk_bp_queue.in_progress.tail = NULL;   \
		       }						\
       }                                                                \
}




/************************************************************************/
/* Device capacity data block 						*/
/************************************************************************/
struct scdisk_capacity {
	int	lba;	/* last logical block address */
	int	len;	/* block length in bytes */
};

/************************************************************************/
/* Device statistics and thresholds structure                           */
/************************************************************************/
struct scdisk_stats {
	int	recovery_limit;		/* Max Allowable recovery level */
	uint	segment_size;		/* # bytes in sample rate       */
	uint	segment_count;		/* # segments processed         */
	uint	byte_count;		/* # bytes current segment      */
};

/************************************************************************/
/* Disk driver operation descriptor block                               */
/************************************************************************/
struct dk_cmd {
	struct sc_buf		scbuf;		/* buf for adapter use  */
	struct buf		*bp;		/* buf list for cmd     */
	uchar			type;		/* cmd type             */
#define DK_BUF		0x01			/* used by strategy     */
#define DK_IOCTL	0x02			/* used by ioctl        */
#define DK_REQSNS	0x04			/* used to request sense*/
#define DK_RESET	0x08			/* used during reset    */
#define DK_WRITEV	0x20			/* used for write verify*/
#define DK_Q_RECOV      0x40                    /*used for queue recovery*/
#define DK_REASSIGN	0x80			/* used for reassign op  */

	uchar			subtype;	/* cmd subtype */
#define DK_TUR		0x01
#define DK_START	0x02
#define DK_STOP		0x03
#define DK_RESERVE	0x04
#define DK_RELEASE	0x05
#define DK_MSENSE	0x06
#define DK_SELECT	0x07
#define DK_READCAP	0x08
#define DK_RDWR_SENSE	0x09
#define DK_IOCMD	0x0a

	uchar			group_type;
#define DK_SINGLE	0x01
#define DK_COALESCED	0x02

	uchar			flags;
#define DK_READY	0x01

	uchar			status;        /* State of command block */
#define DK_FREE		0x00
#define DK_IN_USE	0x01
#define DK_ACTIVE       0x02
#define DK_QUEUED	0x04
#define DK_RETRY	0x08	

	uchar			queued;       /* TRUE if queued         */
	ushort			retry_count;
	uchar			intrpt;        /* used to wait for      */
					       /* resources             */
	uchar			resvd1;
	uchar			resvd2;
	uchar			resvd3;
	uint			segment_count;
	uint			soft_resid;
	struct scdisk_diskinfo	*diskinfo;
	struct dk_cmd		*next;        
	struct dk_cmd		*prev;
};




/************************************************************************/
/* Disk bp queue control structs                                        */
/************************************************************************/

struct dk_queue_control {
	struct buf	*head;
	struct buf	*tail;
};

struct dk_bp_queue {
	struct dk_queue_control	pending;
	struct dk_queue_control	in_progress;
};

/************************************************************************/
/* Information for entire driver                                        */
/************************************************************************/
struct scdisk_info {
	int		config_cnt;	/* Configuration count          */
	int		open_cnt;	/* Number devices open		*/
	lock_t		lock;		/* Lock Word                    */
	struct cdt	*cdt;		/* Component dump table ptr	*/
};

/************************************************************************/
/* Defect list for reassignment operations                              */
/************************************************************************/
struct scdisk_def_list {
	uchar   header[4];
	uchar   descriptor[4];
};

/************************************************************************/
/* Structure used for DKIORDSE and DKIOWRSE ioctl's.                    */
/************************************************************************/
struct scdisk_ioctl_req_sense {
	struct xmem	xmemd;		/* xmemd for request sense data */
	char		*buffer;	/* user address of r.s. buf     */
	int		count;		/* length of user r.s. buf      */
};

/************************************************************************/
/* Mode Data Format Control Structure					*/
/************************************************************************/
struct scdisk_mode_format {
        signed char page_index[DK_MAX_MODE_PAGES]; /* offset to page    */
        ushort  sense_length;                      /* length of sense   */
        uint    block_length;                      /* device block lgth */
};
struct scdisk_timer {
	struct watchdog watch;			  /* system watchdog timer*/
	void   *pointer;			  /* diskinfo pointer     */
};

/************************************************************************/
/* Information on individual disks                                      */
/************************************************************************/
struct scdisk_diskinfo {
	struct scdisk_diskinfo  *next;	        /* Ptr next disk's info */
	struct scdisk_diskinfo  *next_open;     /* Ptr to next open disk*/
	dev_t			devno;          /* Device major+minor   */
						/* numbers		*/

        /*
         * NOTE: the preceding fields must remain in
         * this order to stay consistent with hardware's
	 * thresholding disablement scheme.  In other words they
	 * currently traverse scdisk_open_list of diskinfo structures
	 * via the next pointer looking for a devno match.
         */

	dev_t			adapter_devno;  /* Adapter maj+min      */
	struct scdisk_timer     watchdog_timer;	/* For reset recovery   */
	uchar			scsi_id;        /* SCSI id for device   */
	uchar			lun_id;         /* SCSI lun for device  */
	ushort			reset_count;    /* Number of reset tries*/
	struct dk_cmd           *dk_cmd_q_head; /* head of dk_cmd queue */
	struct dk_cmd           *dk_cmd_q_tail; /* tail of dk_cmd queue */
	struct dk_cmd           ioctl_cmd;      /* reserved ioctl cmd */

	struct dk_cmd	        *cmd_pool;      /* Pool of buf & ioctl  */
						/* cmds                 */
	int                     pool_index;     /* index into above pool*/

	uint			open_event;	/* open/close event word*/
	struct dk_cmd		*checked_cmd;   /* cmd with check cond. */
	struct dk_cmd		*writev_err_cmd;/* cmd for which write  */
						/* verify was issued    */
	struct dk_cmd           *reassign_err_cmd;/* cmd which reassign */
						  /* is being done for  */
	struct dk_cmd	        reset_cmd;      /* reserved reset cmd   */
	struct dk_cmd	        reqsns_cmd;     /* reserved reqsns cmd  */
	struct dk_cmd	        writev_cmd;     /* reserved write and   */
						/* verify  cmd         */
	struct dk_cmd	        q_recov_cmd;    /* reserved command for*/
						/* Q clear and Q resume */
	struct dk_cmd		reassign_cmd;	/* reserved reassing cmd*/
	struct dk_bp_queue	dk_bp_queue;    /* bp queue control     */

	uchar		mode;   /* Current op mode of device            */
#define DK_NORMAL	0x01    /* Normal IO mode for device            */
#define DK_DIAG		0x02    /* Diagnostic mode for device           */
	uchar		state;  /* Current state of device              */
#define DK_OFFLINE	0x00    /* Device is uninitialized and offline  */
#define DK_ONLINE	0x01    /* Device is initialized and online     */
#define DK_RST_FAILED	0x02    /* Device is uninitialized and offline  */
#define DK_REQSNS_PENDING   DK_REQSNS  /* Device has request sense      */
				       /* pending			*/
#define DK_RESET_PENDING    DK_RESET   /* Device has reset pending      */
#define DK_DUMP_PENDING     0x10       /* Device has dump pending       */
#define DK_WRITEV_PENDING   DK_WRITEV  /* Device has write verify 	*/
				       /* pending			*/
#define DK_Q_RECOV_PENDING  DK_Q_RECOV /* Device has q recovery pending */
#define DK_REASSIGN_PENDING DK_REASSIGN/* Device has reassign pending   */

#define DK_RECOV_MASK   (DK_REQSNS_PENDING | DK_RESET_PENDING  \
			| DK_REASSIGN_PENDING | DK_Q_RECOV_PENDING \
	                |DK_DUMP_PENDING)
				/* Mask of DK_REQSNS_PENDING	        */
				/*      DK_RESET_PENDING  		*/
				/*	DK_DUMP_PENDING			*/
				/*	DK_Q_RECOV_PENDING		*/
				/*	DK_REASSIGN_PENDING 		*/
	uchar		disk_intrpt;     /* used to wait for resources  */
        uchar           m_sense_status; /* flag if chgable or curr sns  */
#define DK_SENSE_CURRENT        0x00    /* the sense reflects current   */
#define DK_SENSE_CHANGEABLE     0x40    /* the sense reflects changeable*/
	uchar		opened;         /* device has been opened       */

	uchar		cmd_pending;    /* cmd pending in strategy      */
	uchar		errno;          /* saved error on open path     */
	uchar		error_type;     /* Type of error (for logging)  */
	uchar		retain_reservation; /* Do not release device    */
	uchar		q_type;		/* Queuing type using for normal*/
					/* operation			*/
	uchar		q_err_value;    /* Queue error recovery flag    */
	uchar 		clr_q_on_error; /* Whether or not disk clears   */
					/* its queue on error		*/   
	uchar           buffer_ratio;   /* Read/Write buffer ration     */
	uchar 		cmd_tag_q;      /* If device supports command   */
					/* tag queuing			*/
	uchar		q_status;	/* Used to determine if the     */
					/* adapter's queue was cleared  */
	uchar		q_clr;		/* TRUE if Clear Q command is   */
					/* needed			*/	
	uchar		timer_status;	/* If a watchdog timer is set   */
					/* as well as why it is set	*/
#define DK_TIMER_PENDING   0x1		/* Timer is set			*/
#define DK_TIMER_BUSY      0x2		/* Timer was set due to SCSI    */
					/* Busy status			*/
	
	uchar		restart_unit;   /* Issue SC_RESUME to adapter   */
	uchar		retry_flag;     /* Set if a retry is to be done */
	uchar		safe_relocate;  /* Is relocation safe?          */
	uchar		async_flag;     /* SC_ASYNC or NULL for sync    */

	uchar		dump_inited;    /* Dump has been inited for dev */
	uchar		extended_rw;    /* is extended r/w supported?   */
	uchar		reset_delay;    /* is delay needed after reset? */
	uchar		starting_close; /* TRUE if a close is being     */
					/* attempted for this disk.	*/
	uchar		reset_failures; /* Number of times the reset    */
					/* cycle has failed without a   */
					/* a good buf or ioctl 		*/
					/* completion.			*/	
	ushort          queue_depth;	/* maximum number of commands   */
					/* that can be queued up to a   */
					/* device			*/
	ushort	        cmds_out;       /* Number of command issued for */
					/* this device			*/

	struct buf 	*currbuf;	/* current buf of elevator      */
	struct buf 	*low;		/* low buf of elevator          */
					/* that can be queued up to a   */
					/* device			*/

	uint		max_request;    /* Maximum request to scsi dd   */
	uint		max_coalesce;	/* Maximum size to coalesce	*/
   
	lock_t		lock;           /* Lock Word                    */
	struct file	*fp;            /* filepointer for dev--- calls */

	struct sc_error_log_df error_rec; /* Error record for device    */

	struct scdisk_stats stats;      /* I/O & Soft error statistics  */

	int		mode_data_length; /* Length of useable mode data*/
	uchar		mode_buf[256];  /* Mode data buffer             */
	uchar		sense_buf[256]; /* Request Sense data buffer    */
	uchar		ch_data[256];	/* buffer of chngeable mode data*/
	uchar		df_data[256];	/* buffer of default mode data  */
        struct scdisk_mode_format dd;   /* Mode control for desired data*/
        struct scdisk_mode_format df;   /* Mode control for default data*/
        struct scdisk_mode_format ch;   /* Mode control for chngeable   */
        struct scdisk_mode_format cd;   /* Mode control for current     */
	struct scdisk_ioctl_req_sense ioctl_req_sense;/* Req Sense info */
	struct scdisk_capacity capacity; /* Disk capacity information   */
	struct scdisk_def_list def_list;
	struct dkstat	dkstat;		/* Disk iostat information      */
};

/************************************************************************/
/* Request Sense Data Block                                             */
/************************************************************************/
struct scdisk_req_sense_info  {
	uchar	err_code;	/* error class and code   */
	uchar	rsvd0;
	uchar	sense_key;
	uchar	sense_byte0;
	uchar	sense_byte1;
	uchar	sense_byte2;
	uchar	sense_byte3;
	uchar	add_sense_length;
	uchar	add_sense_byte0;
	uchar	add_sense_byte1;
	uchar	add_sense_byte2;
	uchar	add_sense_byte3;
	uchar	add_sense_key;
	uchar	add_sense_qualifier;
	uchar	fru;
	uchar	flag_byte;
	uchar	field_ptrM;
	uchar	field_ptrL;
	uchar	rsvd2;
	uchar	rsvd3;
	uchar	unit_ref_code1;
	uchar	unit_ref_code2;
	uchar	rsvd4;
	uchar	rsvd5;
	uchar	error_rec0;
	uchar	error_rec1;
	uchar	error_rec2;
	uchar	error_rec3;
	uchar	error_rec4;
	uchar	error_rec5;
	uchar	rsvd6;
	uchar	rsvd7;
};

#ifndef _NO_PROTO 
/************************************************************************/
/* scdisk functions                                                     */
/************************************************************************/

int	scdisk_config(dev_t devno, int op, struct uio *uiop);
int	scdisk_open(dev_t devno, int rwflag, int chan, int ext);
int	scdisk_close(dev_t devno, int chan, int ext);
int	scdisk_read(dev_t devno, struct uio *uiop, int chan, int ext);
int	scdisk_write(dev_t devno, struct uio *uiop, int chan, int ext);
int	scdisk_mincnt(struct buf *bp, void *minparms);
int	scdisk_ioctl(dev_t devno, int op, int arg, int flag, int chan,int ext);
int     scdisk_alloc(struct scdisk_diskinfo *diskinfo);
void    scdisk_init_cmds(struct scdisk_diskinfo *diskinfo);
int	scdisk_strategy(struct buf *bp);
int	scdisk_dump(dev_t devno, struct uio *uiop, int cmd, int arg, int chan,
	int ext);
void	scdisk_pending_enqueue(struct scdisk_diskinfo *diskinfo, 
	struct buf *bp);
void	scdisk_pending_dequeue(struct scdisk_diskinfo *diskinfo);
void	scdisk_start(struct scdisk_diskinfo *diskinfo);
void	scdisk_iodone(struct dk_cmd *cmd_ptr);
void	scdisk_process_good(struct dk_cmd *cmd_ptr);
void	scdisk_process_sense(struct dk_cmd *cmd_ptr);
void scdisk_recover_adap_q(struct scdisk_diskinfo *diskinfo);
void	scdisk_process_error(struct dk_cmd *cmd_ptr);
void	scdisk_process_diagnostic_error(struct dk_cmd *cmd_ptr);
void	scdisk_process_adapter_error(struct dk_cmd *cmd_ptr);
void	scdisk_process_scsi_error(struct dk_cmd *cmd_ptr);
void	scdisk_process_buf_error(struct dk_cmd *cmd_ptr);
void	scdisk_process_ioctl_error(struct dk_cmd *cmd_ptr);
void	scdisk_process_reset_error(struct dk_cmd *cmd_ptr);
void    scdisk_fail_disk(struct scdisk_diskinfo *diskinfo);
void	scdisk_process_reqsns_error(struct dk_cmd *cmd_ptr);
void	scdisk_process_special_error(struct dk_cmd *cmd_ptr);
void	scdisk_process_buf(struct dk_cmd *cmd_ptr);
void	scdisk_process_reset(struct dk_cmd *cmd_ptr);
void	scdisk_coalesce(struct scdisk_diskinfo *diskinfo);
struct dk_cmd	*scdisk_build_cmd(struct scdisk_diskinfo *diskinfo, 
				  struct buf **head, char good_path);
void	scdisk_write_verify(struct scdisk_diskinfo *diskinfo, 
	struct dk_cmd *prev_cmd_ptr, ulong blkno);
void	scdisk_request_sense(struct scdisk_diskinfo *diskinfo,
        struct dk_cmd *cmd_ptr);
void	scdisk_start_unit(struct scdisk_diskinfo *diskinfo, uchar
	start_stop_flag);
void	scdisk_test_unit_ready(struct scdisk_diskinfo *diskinfo);
void	scdisk_reserve(struct scdisk_diskinfo *diskinfo);
void	scdisk_mode_sense(struct scdisk_diskinfo *diskinfo);
void 	scdisk_format_mode_data(char *mode_data,struct scdisk_mode_format *mf,
        int sense_length,char over_ride,struct scdisk_diskinfo *diskinfo);
int	scdisk_mode_data_compare(struct scdisk_diskinfo *diskinfo);
void   scdisk_q_mode(struct scdisk_diskinfo *diskinfo,char  i,char ch_index,
		     char dd_index);
void	scdisk_mode_select(struct scdisk_diskinfo *diskinfo);
void	scdisk_read_cap(struct scdisk_diskinfo *diskinfo);
void	scdisk_release(struct scdisk_diskinfo *diskinfo);
void	scdisk_build_error(struct dk_cmd *cmd_ptr, int type, int priority, 
	int valid_sense);
void	scdisk_log_error(struct dk_cmd *cmd_ptr, int sev);
void	scdisk_watchdog(struct watchdog *watchdog_timer);
struct cdt 	*scdisk_cdt_func(int arg);
void scdisk_q_cmd(struct  dk_cmd *cmd_ptr,char queue,uchar  which_q);
void scdisk_d_q_cmd(struct dk_cmd *cmd_ptr,uchar which_q);
struct  dk_cmd *scdisk_cmd_alloc(struct scdisk_diskinfo *diskinfo,
				 uchar cmd_type);
void scdisk_free_cmd(struct dk_cmd *cmd_ptr);
void scdisk_sleep(uchar *intrpt,uint  *event);
void scdisk_start_watchdog(struct scdisk_diskinfo *diskinfo);
void scdisk_trc(char *desc,char *type,char count,uint word1,uint word2,uint word3,
	    uint word4,uint word5);


#else

/************************************************************************/
/* scdisk functions                                                     */
/************************************************************************/

int	scdisk_config();
int	scdisk_open();
int	scdisk_close();
int	scdisk_read();
int	scdisk_write();
int	scdisk_mincnt();
int	scdisk_ioctl();
int     scdisk_alloc();
void    scdisk_init_cmds();
int	scdisk_strategy();
int	scdisk_dump();
void	scdisk_pending_enqueue();
void	scdisk_pending_dequeue();
void	scdisk_start();
void	scdisk_iodone();
void	scdisk_process_good();
void	scdisk_process_sense();
void scdisk_recover_adap_q();
void	scdisk_process_error();
void	scdisk_process_diagnostic_error();
void	scdisk_process_adapter_error();
void	scdisk_process_scsi_error();
void	scdisk_process_buf_error();
void	scdisk_process_ioctl_error();
void	scdisk_process_reset_error();
void    scdisk_fail_disk();
void	scdisk_process_reqsns_error();
void	scdisk_process_special_error();
void	scdisk_process_buf();
void	scdisk_process_reset();
void	scdisk_coalesce();
struct dk_cmd	*scdisk_build_cmd();
void	scdisk_write_verify();
void	scdisk_request_sense();
void	scdisk_start_unit();
void	scdisk_test_unit_ready();
void	scdisk_reserve();
void	scdisk_mode_sense();
void 	scdisk_format_mode_data();
int 	scdisksd_mode_data_compare();
void    scdisk_q_mode();
void	scdisk_mode_select();
void	scdisk_read_cap();
void	scdisk_release();
void	scdisk_build_error();
void	scdisk_log_error();
void	scdisk_watchdog();
struct cdt	*scdisk_cdt_func();
void scdisk_q_cmd();
void scdisk_d_q_cmd();
struct  dk_cmd *scdisk_cmd_alloc();
void scdisk_free_cmd();

void scdisk_sleep();
void scdisk_start_watchdog();
int scdisk_finish_cmds();
void scdisk_trc();
#endif /* _NO_PROTO */
#endif /* _H_SCDISKDD */
