static char sccsid[] = "@(#)44  1.10.2.4  R2/sysx/disk/scdiskt.c, sysxdisk, bos320, 9312320 2/10/93 07:36:45";
/*
 * COMPONENT_NAME: (SYSXDISK) SCSI Disk Device Driver Top Half Routines
 *
 * FUNCTIONS:  scdisk_config
 *             scdisk_open
 *             scdisk_close
 *             scdisk_ioctl
 *             scdisk_rdwrse
 *             scdisk_read
 *             scdisk_write
 *             scdisk_mincnt
 *	       scdisk_alloc
 *	       scdisk_init_cmds
 *
 * ORIGINS: 27
 *
 * (C) COPYRIGHT International Business Machines Corp. 1988, 1989, 1990, 1992
 * All Rights Reserved
 * Licensed Materials - Property of IBM
 *
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */

/* Included System Files: */

#include <sys/types.h>
#include <sys/errno.h>
#include <sys/syspest.h>
#include <sys/buf.h>
#include <sys/file.h>
#include <sys/pin.h>
#include <sys/sleep.h>
#include <sys/ioctl.h>
#include <sys/intr.h>
#include <sys/i_machine.h>
#include <sys/systm.h>
#include <sys/m_param.h>
#include <sys/malloc.h>
#include <sys/conf.h>
#include <sys/devinfo.h>
#include <sys/lockl.h>
#include <sys/uio.h>
#include <sys/device.h>
#include <sys/watchdog.h>
#include <sys/lvdd.h>
#include <sys/errids.h>
#include <sys/dump.h>
#include <sys/ddtrace.h>
#include <sys/trchkid.h>
#include <sys/priv.h>
#include <sys/iostat.h>

#include <sys/scdisk.h>
#include "scdiskdd.h"

/* End of Included System Files */

/************************************************************************/
/*                                                                      */
/*                      Static Structures                               */
/*                                                                      */
/*      The scdisk_list data structure is an array of pointers to       */
/*      the first diskinfo structure of a NULL terminated linked        */
/*      list of diskinfo structures describing each of the disk         */
/*      device's currently configured in the system.                    */
/*                                                                      */
/*      The scdisk_open_list data structure is similar to the disk_list */
/*      the only exception being that it contains only those disk       */
/*      devices currently open on the system.                           */
/*                                                                      */
/*      The scdisk_info data structure is a structure used for storing  */
/*      general disk driver information, including global states,       */
/*      configuration counts, global locks, etc.                        */
/*                                                                      */
/*      The diskinfo data structures are allcoated on a one per device  */
/*      basis and are used for storing device specific information.     */
/*      These structures are dynamically allocated when a device is     */
/*      configured into the system, and deallocated when the device     */
/*      is deleted. These structures are linked into the disk_list      */
/*      upon configuration and linked into the disk_open_list during    */
/*      any outstanding opens.                                          */
/*                                                                      */
/************************************************************************/

extern struct scdisk_info      scdisk_info;
extern struct scdisk_diskinfo  *scdisk_list[DK_HASHSIZE];
extern struct scdisk_diskinfo  *scdisk_open_list[DK_HASHSIZE];
extern int			scdisk_threshold;
extern int			scdisk_mode_data_offset;

#ifdef DEBUG
extern int                     scdisk_debug;
extern	struct scdisk_trace	*scdisk_trace;

/*
 ******  strings for Internal Debug Trace Table *******
 * To use compile with the flags SC_GOOD_PATH and SC_ERROR_PATH
 * the variable scdisk_trace will be the beginning of the trace table in
 * memory.  The variable scdisk_trctop will point to location where the
 * next entry in the trace table will go.
 */


char     *open		= "OPEN	   ";
char	 *close		= "CLOSE   ";
char	 *read		= "READ    ";
char	 *write		= "WRITE   ";
char	 *mincnt	= "MINCNT  ";
char	 *ioctl		= "IOCTL   ";
char     *entry         = " IN";	/* Entry into routine                */
char     *exit          = " EX";	/* Exit from routine                 */
char     *trc           = " TR";	/* Trace point in routine            */

#endif


/************************************************************************/
/*									*/
/*	NAME:	scdisk_config						*/
/*									*/
/*	FUNCTION: SCSI disk device driver configuration routine.	*/
/*		During device INIT, this routine allocates and		*/
/*		initializes data structures required for processing	*/
/*		user requests. This routine will also delete an already	*/
/*		defined device and free these structures when called	*/
/*		for the TERM operation.					*/
/*									*/
/*	EXECUTION ENVIRONMENT:						*/
/*		This routine can be called by a process.		*/
/*		It can page fault only if called under a process and	*/
/*		the stack is not pinned.				*/
/*									*/
/*	NOTES:								*/
/*									*/
/*	DATA STRUCTURES:						*/
/*		diskinfo	Disk device specific information	*/
/*		scdiskinfo	General disk driver information		*/
/*		scdisk_list	List of configured devices		*/
/*									*/
/*	INPUTS:								*/
/*		devno	- device major/minor number			*/
/*		op	- what function to perform (INIT or TERM)	*/
/*		uiop	- uio pointer to configuration data		*/
/*									*/
/*	RETURN VALUE DESCRIPTION:					*/
/*		The errno values listed in the 'error description'	*/
/*		will be returned to the caller if there is an error.	*/
/*		Otherwise a value of zero will be returned to indicate	*/
/*		successful completion.					*/
/*									*/
/*	ERROR DESCRIPTION:						*/
/*		The following errno values may be returned:		*/
/*		EIO     - Minor number out of range			*/
/*			- Attempt to INIT an existing device		*/
/*			- Attempt to TERM a nonexistent device		*/
/*			- Attempt to TERM an open device		*/
/*		EINVAL  - Unknown operation				*/
/*									*/
/*	EXTERNAL PROCEDURES CALLED:					*/
/*		pincode		unpincode				*/
/*		devswadd	devswdel				*/
/*		xmalloc		xmfree					*/
/*		uiomove							*/
/*									*/
/*									*/
/************************************************************************/

int
scdisk_config(
dev_t		devno,
int		op,
struct uio	*uiop)
{
	int			errnoval, dev, rc, hash_key, idlun;
	uint			cdt_size;
	struct scdisk_diskinfo	*last_diskinfo, *diskinfo;
	struct disk_ddi		config_info;
	struct devsw		devsw_struct;
	struct cdt		*old_cdt;

	extern			nodev();

#ifdef DEBUG
	DKprintf(("Entering scdisk_config devno[0x%x] op[0x%x]\n", devno, op));
#endif

	DDHKWD5(HKWD_DD_SCDISKDD, DD_ENTRY_CONFIG, 0, devno, op, uiop, 0, 0);
	errnoval = 0;
	dev = minor(devno);
	if (dev > DK_MAX_DISKS) {
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_CONFIG, ENXIO, devno);
		return(ENXIO);
	}

	rc = lockl(&(scdisk_info.lock), LOCK_SHORT);

	switch (op) {
	case CFG_INIT:
		scdisk_info.config_cnt++;
		if (scdisk_info.config_cnt == 0x01) {

			/* Initialize devsw struct for scdisk */
			devsw_struct.d_open = (int(*)())scdisk_open;
			devsw_struct.d_close = (int(*)())scdisk_close;
			devsw_struct.d_read = (int(*)())scdisk_read;
			devsw_struct.d_write = (int(*)())scdisk_write;
			devsw_struct.d_ioctl = (int(*)())scdisk_ioctl;
			devsw_struct.d_strategy=(int(*)())scdisk_strategy;
			devsw_struct.d_ttys = (struct tty *)NULL;
			devsw_struct.d_select = (int(*)())nodev;
			devsw_struct.d_config = (int(*)())scdisk_config;
			devsw_struct.d_print = (int(*)())nodev;
			devsw_struct.d_dump = (int(*)())scdisk_dump;
			devsw_struct.d_mpx = (int(*)())nodev;
			devsw_struct.d_revoke = (int(*)())nodev;
			devsw_struct.d_dsdptr = NULL;
			devsw_struct.d_selptr = NULL;
			devsw_struct.d_opts = 0;

			/* Add devsw struct to system devsw table */
			errnoval = devswadd(devno, &devsw_struct);
			if (errnoval) {
				scdisk_info.config_cnt--;
				unlockl(&(scdisk_info.lock));
				DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_CONFIG,
				    errnoval, devno);
				return(errnoval);
			}
		}

		/* Search for diskinfo structure in configured list */
		hash_key = dev & DK_HASH;
		diskinfo = scdisk_list[hash_key];
		while (diskinfo != NULL) {
			if (diskinfo->devno == devno)
				break;
			diskinfo = diskinfo->next;
		}
		/* Verify device not already configured */
		if (diskinfo != NULL) {
			scdisk_info.config_cnt--;
			unlockl(&(scdisk_info.lock));
			DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_CONFIG, EINVAL,
			    devno);
			return(EINVAL);
		}
		/* Allocate memory for diskinfo structure and zero it */
		diskinfo = (struct scdisk_diskinfo *)
		    xmalloc((uint) sizeof(struct scdisk_diskinfo), 2, 
		    (heapaddr_t) pinned_heap);
		if (diskinfo == NULL) {
			scdisk_info.config_cnt--;
			if (scdisk_info.config_cnt == 0) {
				(void) devswdel(devno);
			}
			unlockl(&(scdisk_info.lock));
			DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_CONFIG, EINVAL,
			    devno);
			return(ENOMEM);
		}
		bzero((caddr_t) diskinfo, sizeof(struct scdisk_diskinfo));

		/*
		 *  Allocate memory for component dump table.  Three entries
		 *  are needed for the static structures, and then one
		 *  additional entry for each disk structure.
		 */
		 
		cdt_size = (uint) sizeof(struct cdt_head) + (3 + 
		    scdisk_info.config_cnt) * (uint) sizeof(struct cdt_entry);
		old_cdt = scdisk_info.cdt;
		scdisk_info.cdt = (struct cdt *) xmalloc(cdt_size, 2, 
		    (heapaddr_t) pinned_heap);
		if (scdisk_info.cdt == NULL) {
			scdisk_info.cdt = old_cdt;
			(void) xmfree((caddr_t)diskinfo, (heapaddr_t)
			    pinned_heap);
			scdisk_info.config_cnt--;
			if (scdisk_info.config_cnt == 0) {
				(void) devswdel(devno);
			}
			unlockl(&(scdisk_info.lock));
			DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_CONFIG, EINVAL,
			    devno);
			return(ENOMEM);
		}
		/*  Free the old component dump table */
		(void) xmfree((caddr_t)old_cdt, (heapaddr_t) pinned_heap);

		bzero((caddr_t) scdisk_info.cdt, cdt_size);
		scdisk_info.cdt->_cdt_head._cdt_magic = DMP_MAGIC;
		bcopy("scdisk", scdisk_info.cdt->_cdt_head._cdt_name, 7);
		scdisk_info.cdt->_cdt_head._cdt_len = cdt_size;

		/* Move config data into diskinfo structure */
		uiomove((caddr_t)&config_info, (int) sizeof(struct disk_ddi),
		    (enum uio_rw) UIO_WRITE, uiop);

		diskinfo->devno = devno;
		diskinfo->adapter_devno = config_info.adapter_devno;
		diskinfo->scsi_id = config_info.scsi_id;
		diskinfo->lun_id = config_info.lun_id;
		diskinfo->safe_relocate = config_info.safe_relocate;
		diskinfo->async_flag = config_info.async_flag;
		diskinfo->extended_rw = config_info.extended_rw;
		diskinfo->reset_delay = config_info.reset_delay;
		diskinfo->stats.recovery_limit = config_info.recovery_limit;
		diskinfo->max_coalesce = config_info.max_coalesce;
		diskinfo->queue_depth = config_info.queue_depth;
		if (diskinfo->queue_depth < 1)
			diskinfo->queue_depth = 1;
		diskinfo->cmds_out = 0;
		diskinfo->max_request = config_info.max_request;
		diskinfo->q_type = config_info.q_type;
		diskinfo->q_err_value = config_info.q_err_value;
		diskinfo->clr_q_on_error = config_info.clr_q_on_error;
		diskinfo->cmd_tag_q = config_info.cmd_tag_q;
		diskinfo->buffer_ratio = config_info.buffer_ratio;
		diskinfo->q_clr = FALSE;
		diskinfo->q_status = 0;	
		diskinfo->timer_status = 0;
				
		/* Copy mode select data from dds data into diskinfo struct*/
		bcopy(config_info.mode_data,
		    diskinfo->mode_buf,
		    config_info.mode_data_length);
		diskinfo->mode_data_length = config_info.mode_data_length;

		/* Copy mode default data from dds data into diskinfo struct*/
		bcopy(config_info.mode_default_data, 
		    diskinfo->df_data, config_info.mode_default_length);

		/*
		 * Format the data base mode data
		 */
		scdisk_format_mode_data((char *)diskinfo->mode_buf, 
				    (struct scdisk_mode_format *)&diskinfo->dd,
				    (int)diskinfo->mode_data_length,
				    (char) TRUE,
				    (struct scdisk_diskinfo *)diskinfo);

		/*
		 * Format the data base mode default data
		 */
		scdisk_format_mode_data((char *)diskinfo->df_data, 
				    (struct scdisk_mode_format *)&diskinfo->df,
				    (int)config_info.mode_default_length,
				    (char)FALSE,
				    (struct scdisk_diskinfo *)NULL);

		/* Initialize remaining sections of disk info structure */
		diskinfo->dk_cmd_q_head = NULL;
		diskinfo->dk_cmd_q_tail = NULL;

		

		diskinfo->currbuf = NULL;
		diskinfo->low = NULL;
		diskinfo->dk_bp_queue.in_progress.head = NULL;
		diskinfo->dk_bp_queue.in_progress.tail = NULL;
		diskinfo->reset_count = 0;
		diskinfo->mode = DK_NORMAL;
		diskinfo->state = DK_OFFLINE;
		diskinfo->opened = FALSE;
		diskinfo->retain_reservation = FALSE;

		diskinfo->cmd_pending = FALSE;
		diskinfo->starting_close = FALSE;
		diskinfo->lock = LOCK_AVAIL;
		diskinfo->fp = NULL;
		diskinfo->dump_inited = FALSE;
		diskinfo->m_sense_status = 0;
		diskinfo->open_event = EVENT_NULL;

		/* initialize fields for watchdog timer */
		diskinfo->watchdog_timer.watch.next = NULL;
		diskinfo->watchdog_timer.watch.prev = NULL;
		diskinfo->watchdog_timer.watch.func = (void (*)())scdisk_watchdog;
		diskinfo->watchdog_timer.watch.count = 0;
		diskinfo->watchdog_timer.watch.restart = 2;
		diskinfo->watchdog_timer.pointer = (void *)diskinfo;
		w_init(&diskinfo->watchdog_timer);

		/* Initialize cmd structs for use */
		scdisk_init_cmds(diskinfo);

		/* Initialize drive statistics */
		diskinfo->stats.recovery_limit = config_info.recovery_limit;
		diskinfo->stats.segment_size = config_info.segment_size;
		diskinfo->stats.segment_count = config_info.segment_cnt;
		diskinfo->stats.byte_count = config_info.byte_count;

		/* Initialize drive error record */
		diskinfo->error_type = 0;
		bcopy(config_info.resource_name, 
		    diskinfo->error_rec.resource_name, 8);

		/* fill out fields in io statistics structure */
		bzero (&(diskinfo->dkstat), sizeof(struct dkstat));
		diskinfo->dkstat.dk_bsize = 512;	/* block size */
		bcopy(config_info.resource_name, diskinfo->dkstat.diskname, 8);

		/* Register io statistics tracking structure */
		rc = iostadd(DD_DISK, &(diskinfo->dkstat));
		ASSERT(rc == 0);


		/* Link diskinfo structure in configured device list */
		diskinfo->next = scdisk_list[hash_key];
		if (scdisk_info.config_cnt == 0x01) {
			/*
			 * Set device driver's thresholding value.
			 * This variable is exported so that
			 * hardware can then override mode data.
			 */
			scdisk_threshold = 4;

			/*
			 * Determine offset of mode data in 
			 * diskinfo structure.  This variable
			 * is exported so that hardware can then
			 * override mode data.
			 */
			scdisk_mode_data_offset = 
				(int)((int)(&(diskinfo->mode_buf))
				      - (int)diskinfo);
		}
		scdisk_list[hash_key] = diskinfo;
#ifdef DEBUG
			/*
			 * Malloc Trace Buffer one time only, this is
			 * so we can have the buffer more readable 
			 * 32-byte aligned
			 */
			if (scdisk_info.config_cnt == 0x01) {
				scdisk_trace=(struct scdisk_trace *)xmalloc(
				(uint)(sizeof(struct scdisk_trace) * TRCLNGTH), 
				 5, (heapaddr_t)pinned_heap);
			}
#endif
		break;
	case CFG_TERM:
		/* Search for diskinfo structure in configured device list */
		hash_key = dev & DK_HASH;
		diskinfo = scdisk_list[hash_key];
		last_diskinfo = diskinfo;
		while (diskinfo != NULL) {
			if (diskinfo->devno == devno)
				break;
			last_diskinfo = diskinfo;
			diskinfo = diskinfo->next;
		}
		if (diskinfo == NULL) {
			unlockl(&(scdisk_info.lock));
			DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_CONFIG, 0, devno);
			return(0);
		}
		if (diskinfo->opened) {
			unlockl(&(scdisk_info.lock));
			DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_CONFIG, EINVAL,
			    devno);
			return(EBUSY);
		}

		/* Open adapter for device and store filepointer */
		errnoval = fp_opendev(diskinfo->adapter_devno, (int) FWRITE,
		    (caddr_t) NULL, (int) NULL, &(diskinfo->fp));
		if (errnoval == 0) {
			/* Issue a start IOCTL to the adapter driver */
			idlun = IDLUN(diskinfo->scsi_id, diskinfo->lun_id);
			errnoval = fp_ioctl(diskinfo->fp, (uint) SCIOSTART,
			    (caddr_t) idlun, (int) NULL);
			if (errnoval == 0) {
				/* pin bottom half routines */
				errnoval = pincode((int(*)())scdisk_iodone);
				if (errnoval == 0) {
					diskinfo->errno = 0x00;
					diskinfo->disk_intrpt = 1;
					scdisk_start_unit(diskinfo,
					    (uchar) DK_STOP);
					scdisk_sleep(&(diskinfo->disk_intrpt),
						     &(diskinfo->open_event));


					errnoval = diskinfo->errno;
					(void *) unpincode((int(*)())
					    scdisk_iodone);
				}
				fp_ioctl(diskinfo->fp, (uint) SCIOSTOP,
				    (caddr_t) idlun, (int) NULL);
			}
			fp_close(diskinfo->fp);
		}

		if (diskinfo == last_diskinfo) {
			scdisk_list[hash_key] = diskinfo->next;
		} else {
			last_diskinfo->next = diskinfo->next;
		}

		scdisk_info.config_cnt--;
		/* Check for last device being deleted */
		if (scdisk_info.config_cnt == 0) {
			(void) devswdel(devno);
			(void) xmfree((caddr_t)scdisk_info.cdt,
				      (heapaddr_t) pinned_heap);
			scdisk_info.cdt = NULL;
#ifdef DEBUG
				xmfree((caddr_t)scdisk_trace,
					(heapaddr_t)pinned_heap);
#endif
		} else {
			/*
			 *  Allocate memory for component dump table.  Three
			 *  entries are needed for the static structures, and
			 *  then one additional entry for each disk structure.
			 */
		 
			cdt_size = (uint) sizeof(struct cdt_head) + (3 + 
			    scdisk_info.config_cnt) * (uint) sizeof(struct 
			    cdt_entry);
			old_cdt = scdisk_info.cdt;
			scdisk_info.cdt = (struct cdt *) xmalloc(cdt_size, 2,
			    (heapaddr_t) pinned_heap);
			if (scdisk_info.cdt == NULL) {
				scdisk_info.cdt = old_cdt;
			} else {
				(void) xmfree((caddr_t)old_cdt,
				    (heapaddr_t) pinned_heap);
			}
			bzero((caddr_t) scdisk_info.cdt, cdt_size);
			scdisk_info.cdt->_cdt_head._cdt_magic = DMP_MAGIC;
			bcopy("scdisk",scdisk_info.cdt->_cdt_head._cdt_name,7);
			scdisk_info.cdt->_cdt_head._cdt_len = cdt_size;
		}

		w_clear(&diskinfo->watchdog_timer);

		/* de-register the dkstat structure */
		iostdel(&(diskinfo->dkstat));
		(void) xmfree((caddr_t)diskinfo, (heapaddr_t)pinned_heap);
		/*
		 *  The unconfigure method checks for EBUSY to see if the
		 *  device unconfigured the device or not.  If we got this
		 *  far, we must have unconfigured it, so don't return
		 *  EBUSY (EIO will basically be ignored).
		 */
		if (errnoval == EBUSY) {
			errnoval = EIO;
		}
		break;

	default:
		/* Unknown op */
		errnoval = EINVAL;
	}
#ifdef DEBUG
	DKprintf(("Exiting scdisk_config errnoval[0x%x]\n", errnoval));
#endif
	unlockl(&(scdisk_info.lock));
	DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_CONFIG, errnoval, devno);
	return(errnoval);
}

/************************************************************************/
/*									*/
/*	NAME:	scdisk_open						*/
/*									*/
/*	FUNCTION: SCSI Disk Device Driver Open Routine.			*/
/*		This routine prepares a device for utilization by the	*/
/*		system. The disk's information structure is located	*/
/*		in the scdisk_list.					*/
/*		The open is checked for validity, and if the first	*/
/*		valid open, the diskinfo structure is pinned and placed	*/
/*		on the scdisk_open_list. The adapter driver is called	*/
/*		to open and start the device.				*/
/*									*/
/*	EXECUTION ENVIRONMENT:						*/
/*		This routine can be called by a process.		*/
/*		It can page fault only if called under a process and	*/
/*		the stack is not pinned.				*/
/*									*/
/*	NOTES:								*/
/*									*/
/*	DATA STRUCTURES:						*/
/*		diskinfo		Disk device specific information*/
/*		scdiskinfo		General disk driver information	*/
/*		scdisk_list		List of configured devices	*/
/*		scdisk_open_list	List of open devices		*/
/*									*/
/*	INPUTS:								*/
/*		devno	- device major/minor number			*/
/*		rwflag	- FREAD for read only, FWRITE for read/write	*/
/*		chan	- unused (will be zero)				*/
/*		ext	- extended parameter, could be SC_DIAGNOSTIC	*/
/*									*/
/*	RETURN VALUE DESCRIPTION:					*/
/*		The errno values listed in the 'error description'	*/
/*		will be returned to the caller if there is an error.	*/
/*		Otherwise a value of zero will be returned to indicate	*/
/*		successful completion.					*/
/*									*/
/*	ERROR DESCRIPTION:						*/
/*		The following errno values may be returned:		*/
/*		EIO	- Minor number out of range			*/
/*			- Device not configured				*/
/*			- Devopen to adapter failed			*/
/*			- Devioctl to start device failed		*/
/*			- Device open cycle failed			*/
/*		EACCES  - Device is in DIAGNOSTIC mode			*/
/*			- Attempt to enter DIAGNOSTIC mode while open	*/
/*									*/
/*									*/
/*	EXTERNAL PROCEDURES CALLED:					*/
/*		lockl		unlockl					*/
/*		fp_opendev	fp_close				*/
/*		fp_ioctl						*/
/*		pincode		unpincode				*/
/*		dmp_add		dmp_del					*/
/*									*/
/*									*/
/************************************************************************/

int
scdisk_open(
dev_t	devno,
int	rwflag, 
int	chan, 
int	ext)
{
	int			errnoval, rc, dev, hash_key, idlun;
	struct scdisk_diskinfo	*diskinfo;
	struct devinfo		iocinfo_buf;




	DDHKWD5(HKWD_DD_SCDISKDD, DD_ENTRY_OPEN, 0, devno, rwflag, chan, ext,
	    0);
	errnoval = 0;
	dev = minor(devno);
	if (dev > DK_MAX_DISKS) {
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, ENODEV, devno);
		return(ENODEV);
	}
	rc = lockl(&(scdisk_info.lock), LOCK_SHORT);

#ifdef DEBUG
#ifdef SC_GOOD_PATH
   scdisk_trc(open, trc,(char)0, (uint)devno, (uint)rwflag,(uint)chan, 
	      (uint)ext,(uint)0);
#endif	
#endif
	/* Locate diskinfo structure in device list */
	hash_key = dev & DK_HASH;
	diskinfo = scdisk_list[hash_key];
	while (diskinfo != NULL) {
		if (diskinfo->devno == devno)
			break;
		diskinfo = diskinfo->next;
	}
	if (diskinfo == NULL) {
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, ENODEV, devno);
		return(ENODEV);
	}
	if (ext && (rc = privcheck(RAS_CONFIG)))  {
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, ENODEV, devno);
		return(rc);
	}
	if ((diskinfo->opened == TRUE) && (ext & SC_DIAGNOSTIC)) {
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, EINVAL, devno);
		return(EACCES);
	}
	if ((diskinfo->opened == TRUE) && (diskinfo->mode & DK_DIAG)) {
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, EINVAL, devno);
		return(EACCES);
	}

	if (diskinfo->opened == TRUE) {
		if (ext & SC_RETAIN_RESERVATION) {
			diskinfo->retain_reservation = TRUE;
		}
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, 0, devno);
		return(0);
	} else {
		diskinfo->errno = 0;
	}

	/*
	 * Allocate command block pool
	 */

	if ((rc = scdisk_alloc(diskinfo))) {
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, EINVAL, devno);
		return(rc);
	}


	/* Open adapter for device and store filepointer */
	errnoval = fp_opendev(diskinfo->adapter_devno, (int) FWRITE,
	    (caddr_t) NULL, (int) NULL, &(diskinfo->fp));
	if (errnoval != 0) {
		(void)xmfree((caddr_t)(diskinfo->cmd_pool),pinned_heap);
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, errnoval, devno);
		return(errnoval);
	}
	/*
	 * Issue IOCINFO to the adapter driver to find out the maximum request
	 * size.
	 */
	idlun = IDLUN(diskinfo->scsi_id, diskinfo->lun_id);
	errnoval = fp_ioctl(diskinfo->fp, (uint) IOCINFO,
	    (caddr_t) &iocinfo_buf, (int) NULL);
	if (errnoval != 0) {
		fp_close(diskinfo->fp);
		(void)xmfree((caddr_t)(diskinfo->cmd_pool),pinned_heap);
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, errnoval, devno);
		return(errnoval);
	} else {

		/*
                 *  Max_request can not exceed adapter's max transfer size
                 */

		if (diskinfo->max_request > iocinfo_buf.un.scsi.max_transfer) 
			diskinfo->max_request = 
				iocinfo_buf.un.scsi.max_transfer;
		

		/*
                 *  If max_request is larger than 128K and extended read
                 *  and write commands are not supported, set the max
                 *  to 128K so we will not overrun the normal r/w cmds.
                 */
		if ((diskinfo->max_request > 0x20000) && 
		    !(diskinfo->extended_rw)) {
			/*
			 *  If max_request is larger than 128K and extended read
			 *  and write commands are not supported, set the max
			 *  to 128K so we will not overrun the normal r/w cmds.
			 */
			diskinfo->max_request = 0x20000;
		}
		else if ((diskinfo->max_request > 0x333300) &&
			 (diskinfo->extended_rw)) {
			/*
			 * If max_request is larger then 32767.5 K and extended
			 * read and write commands are supported then set the 
			 * max to 32768K (the largest possible for these 
			 * commands) so we will not overrun the extended r/w 
			 * commands
			 */
			diskinfo->max_request =  0x333300;
		}
		if (diskinfo->max_coalesce > diskinfo->max_request) {
			diskinfo->max_coalesce = diskinfo->max_request;
		}
	}
	/* Issue a start IOCTL to the adapter driver */
	errnoval = fp_ioctl(diskinfo->fp, (uint) SCIOSTART, (caddr_t) idlun,
	    (int) NULL);
	if (errnoval != 0) {
		fp_close(diskinfo->fp);
		(void)xmfree((caddr_t)(diskinfo->cmd_pool),pinned_heap);
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, errnoval, devno);
		return(errnoval);
	}

	/* pin bottom half routines */
	errnoval = pincode((int(*)())scdisk_iodone);
	if (errnoval) {
		fp_ioctl(diskinfo->fp, (uint) SCIOSTOP, (caddr_t) idlun,
		    (int) NULL);
		fp_close(diskinfo->fp);
		(void)xmfree((caddr_t)(diskinfo->cmd_pool),pinned_heap);
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, errnoval, devno);
		return(errnoval);
	}

	/*  Register component dump table function on first open */
	if (scdisk_info.open_cnt == 0) {
		(void) dmp_add(scdisk_cdt_func);
	}

	/* Issue the BDR if requested */
	if (ext & SC_FORCED_OPEN) {
		idlun = IDLUN(diskinfo->scsi_id, diskinfo->lun_id);
		errnoval = fp_ioctl(diskinfo->fp, (uint) SCIORESET,
		    (caddr_t) idlun, (int) NULL);
		if (errnoval != 0) {
			if (scdisk_info.open_cnt == 0) {
				(void) dmp_del(scdisk_cdt_func);
			}
			(void *) unpincode((int(*)())scdisk_iodone);
			fp_ioctl(diskinfo->fp, (uint) SCIOSTOP,
			    (caddr_t) idlun, (int) NULL);
			fp_close(diskinfo->fp);
			(void)xmfree((caddr_t)(diskinfo->cmd_pool),pinned_heap);
			unlockl(&(scdisk_info.lock));
			DDHKWD1(HKWD_DD_SCDISKDD,DD_EXIT_OPEN,errnoval,devno);
			return(errnoval);
		}
	}
	if (ext & SC_DIAGNOSTIC) {
		if (ext & SC_RETAIN_RESERVATION) {
			diskinfo->retain_reservation = TRUE;
		}
		/* Place diskinfo struct in scdisk_open_list */
		diskinfo->next_open = scdisk_open_list[hash_key];
		scdisk_open_list[hash_key] = diskinfo;
		diskinfo->state = DK_ONLINE;
		diskinfo->mode = DK_DIAG;
		diskinfo->opened = TRUE;
		scdisk_info.open_cnt++;
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, 0, devno);
		return(0);
	}
	diskinfo->disk_intrpt = 1;
	scdisk_start_unit(diskinfo, (uchar) DK_START);
	scdisk_sleep(&(diskinfo->disk_intrpt),&(diskinfo->open_event));


	/* Check for reset cycle failure */
	if (diskinfo->errno != 0) {
		errnoval = diskinfo->errno;
		diskinfo->disk_intrpt = 1;
		scdisk_release(diskinfo);
		scdisk_sleep(&(diskinfo->disk_intrpt),
			     &(diskinfo->open_event));

		if (scdisk_info.open_cnt == 0) {
			(void) dmp_del(scdisk_cdt_func);
		}
		(void *) unpincode((int(*)())scdisk_iodone);
		fp_ioctl(diskinfo->fp, (uint) SCIOSTOP, (caddr_t) idlun,
		    (int) NULL);
		fp_close(diskinfo->fp);
		(void)xmfree((caddr_t)(diskinfo->cmd_pool),pinned_heap);
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, errnoval, devno);
		return(errnoval);
	}
	if (ext & SC_RETAIN_RESERVATION) {
		diskinfo->retain_reservation = TRUE;
	}

	/*
	 * initialize timer starts  on first open.  
	 */
	diskinfo->reset_failures = 0;

	/* Place diskinfo struct in scdisk_open_list */
	diskinfo->next_open = scdisk_open_list[hash_key];
	scdisk_open_list[hash_key] = diskinfo;
	diskinfo->state = DK_ONLINE;
	diskinfo->mode = DK_NORMAL;
	diskinfo->opened = TRUE;
	scdisk_info.open_cnt++;

#ifdef DEBUG
#ifdef SC_GOOD_PATH
   scdisk_trc(open, trc,(char)1, (uint)0, (uint)0,(uint)0, (uint)0,(uint)0);
#endif	
#endif
	unlockl(&(scdisk_info.lock));
	DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_OPEN, errnoval, devno);

	return(errnoval);
}

/************************************************************************/
/*									*/
/*	NAME:	scdisk_close						*/
/*									*/
/*	FUNCTION: SCSI Disk Device Driver Close Routine.		*/
/*		This routine searches the scdisk_open_list for		*/
/*		the specified device, and unlinks it when found.	*/
/*		The device is then released for other initiators.	*/
/*		The device's diskinfo structure is unpinned from	*/
/*		memory, device open flag is cleared, and the		*/
/*		adapter driver is called via fp_ioctl to stop the	*/
/*		device and via fp_close to close if necessary.		*/
/*									*/
/*	EXECUTION ENVIRONMENT:						*/
/*		This routine can be called by a process.		*/
/*		It can page fault only if called under a process and	*/
/*		the stack is not pinned.				*/
/*									*/
/*	NOTES:								*/
/*									*/
/*	DATA STRUCTURES:						*/
/*		diskinfo        Disk device specific information	*/
/*		scdiskinfo      General disk driver information		*/
/*		scdisk_list     List of configured devices		*/
/*		scdisk_open_list List of open devices			*/
/*									*/
/*	INPUTS:								*/
/*		devno	- device major/minor number			*/
/*		chan	- unused (will be zero)				*/
/*		ext	- extended parameter, could be SC_DIAGNOSTIC	*/
/*									*/
/*	RETURN VALUE DESCRIPTION:					*/
/*		The errno values listed in the 'error description'	*/
/*		will be returned to the caller if there is an error.	*/
/*		Otherwise a value of zero will be returned to indicate	*/
/*		successful completion.					*/
/*									*/
/*	ERROR DESCRIPTION:						*/
/*		The following errno values may be returned:		*/
/*		EIO     - Minor number out of range			*/
/*			- Device not configured				*/
/*			- Device not open				*/
/*			- Close cycle failed				*/
/*									*/
/*	EXTERNAL PROCEDURES CALLED:					*/
/*		lockl		unlockl					*/
/*				unpincode				*/
/*		fp_ioctl	fp_close				*/
/*									*/
/*									*/
/************************************************************************/

int
scdisk_close(
dev_t	devno,
int	chan, 
int	ext)
{
	int			errnoval, dev, rc, hash_key, idlun;
	struct scdisk_diskinfo	*diskinfo, *last_diskinfo;



	DDHKWD1(HKWD_DD_SCDISKDD, DD_ENTRY_CLOSE, 0, devno);
	errnoval = 0;
	dev = minor(devno);
	if (dev > DK_MAX_DISKS) {
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_CLOSE, ENXIO, devno);
		return(ENXIO);
	}

	rc = lockl(&(scdisk_info.lock), LOCK_SHORT);
#ifdef DEBUG
#ifdef SC_GOOD_PATH
   scdisk_trc(close, trc,(char)0, (uint)devno, (uint)0,(uint)chan, 
	      (uint)ext,(uint)0);
#endif	
#endif
	/* Locate diskinfo structure in scdisk_open_list */
	hash_key = dev & DK_HASH;
	diskinfo = scdisk_open_list[hash_key];
	last_diskinfo = diskinfo;
	while (diskinfo != NULL) {
		if (diskinfo->devno == devno)
			break;
		else {
			last_diskinfo = diskinfo;
			diskinfo = diskinfo->next_open;
		}
	}

	if (diskinfo == NULL) {
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_CLOSE, ENXIO, devno);
		return(ENXIO);
	}

	/* Unlink diskinfo struct from scdisk_open_list */
	if (last_diskinfo == diskinfo) {
		scdisk_open_list[hash_key] = diskinfo->next_open;
	}
	else {
		last_diskinfo->next_open = diskinfo->next_open;
	}

	/*
	 * Lock the disk lock for this device to wait for an ioctl to complete.
	 */
	rc = lockl(&(diskinfo->lock), LOCK_SHORT);

	/*
	 * Set the following flags to prevent any new commands from
	 * being issued to the device driver
	 */

	diskinfo->opened = FALSE;
	diskinfo->starting_close = TRUE;
	diskinfo->state = DK_OFFLINE;
	diskinfo->mode = DK_NORMAL;


	/* Check for requests awaiting completion */

	while (diskinfo->cmds_out || 
	       (diskinfo->dk_cmd_q_head != NULL) ||
	       (diskinfo->ioctl_cmd.flags & DK_READY) ||
	       (diskinfo->dk_bp_queue.in_progress.head != NULL) ||
	       (diskinfo->currbuf != NULL) ||
	       diskinfo->cmd_pending || 
	       diskinfo->dump_inited) {

		/*
		 * We will wait until all outstanding commands complete, 
		 * all pending commands (i.e. bufs, ioctls, and error recovery
		 * commands) get issued and complete, and for a dump to 
		 * to complete.
		 */
		delay(HZ);  /* wait one second for operation to complete */
	}

	/* Check for diagnostic mode or retain reservation, release if OK */
	if (!(diskinfo->mode & DK_DIAG) && !(diskinfo->retain_reservation)) {
		/* Release device for other initiators */
		if (diskinfo->state == DK_RST_FAILED) {
			/*
			 * Since the reset sequence failed assume that it is
			 * released (i.e. do not issue release).  The 
			 * alternative would be to do the
			 * reset cycle over and potentially hang the system.
			 * Also if this host is having difficulty opening the
			 * disk then most likely the other host will too, 
			 * regardless if it is  released  or not.
			 * We will however flag this as an error.
			 */
			diskinfo->errno = EIO;
		}
		else {
			diskinfo->errno = 0x00;

			diskinfo->disk_intrpt = 1;
			scdisk_release(diskinfo);
			scdisk_sleep(&(diskinfo->disk_intrpt),
				     &(diskinfo->open_event));

		}
		errnoval = diskinfo->errno;
	}

	diskinfo->retain_reservation = FALSE;
	diskinfo->state = DK_OFFLINE;
	diskinfo->starting_close = FALSE;

	/*
	 * initialize timer starts on close so it is ready for next open.  
	 */
	diskinfo->reset_failures = 0;

	/* Terminate adapter connection for this device */
	idlun = IDLUN(diskinfo->scsi_id, diskinfo->lun_id);
	rc = fp_ioctl(diskinfo->fp, (uint)SCIOSTOP, (caddr_t)idlun, (int)NULL);
	errnoval = (errnoval == 0) ? rc : errnoval;
	rc = fp_close(diskinfo->fp);
	errnoval = (errnoval == 0) ? rc : errnoval;
	scdisk_info.open_cnt--;
	if (scdisk_info.open_cnt == 0) {
		(void) dmp_del(scdisk_cdt_func);
	}
	(void)xmfree((caddr_t)(diskinfo->cmd_pool),pinned_heap);
	rc = unpincode((int(*)())scdisk_iodone);
	errnoval = (errnoval == 0) ? rc : errnoval;


#ifdef DEBUG
#ifdef SC_GOOD_PATH
   scdisk_trc(close, trc,(char)1, (uint)devno, (uint)errnoval,(uint)0, (uint)0,
	      (uint)0);
#endif	
#endif
	unlockl(&(diskinfo->lock));
	unlockl(&(scdisk_info.lock));
	DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_CLOSE, errnoval, devno);
	return(errnoval);
}

/************************************************************************/
/*									*/
/*	NAME:	scdisk_read						*/
/*									*/
/*	FUNCTION: SCSI Disk Device Driver Raw Read Routine.		*/
/*		This routine processes raw IO requests from the system.	*/
/*		A buf struct is allocated, and the b_options field	*/
/*		is initialized with the ext argument. Uphysio is called	*/
/*		to begin processing of the raw IO request.		*/
/*		Upon completion of the operation, the buf struct is	*/
/*		freed.							*/
/*									*/
/*	EXECUTION ENVIRONMENT:						*/
/*		This routine can be called by a process.		*/
/*		It can page fault only if called under a process and	*/
/*		the stack is not pinned.				*/
/*									*/
/*	NOTES:								*/
/*									*/
/*	DATA STRUCTURES:						*/
/*									*/
/*	INPUTS:								*/
/*		devno	- device major/minor number			*/
/*		uiop	- Address of uio struct describing request	*/
/*		chan	- unused (will be zero)				*/
/*		ext	- extended parameter				*/
/*									*/
/*	RETURN VALUE DESCRIPTION:					*/
/*		This routine returns the value received from uphysio	*/
/*									*/
/*	ERROR DESCRIPTION:						*/
/*		See above.						*/
/*									*/
/*	EXTERNAL PROCEDURES CALLED:					*/
/*		uphysio							*/
/*									*/
/*									*/
/************************************************************************/

int
scdisk_read(
dev_t		devno,
struct uio	*uiop,
int		chan, 
int		ext)
{
	int	errnoval;


#ifdef DEBUG
#ifdef SC_GOOD_PATH
   scdisk_trc(read, entry,(char)0, (uint)devno, (uint)uiop,(uint)chan, 
	      (uint)ext,(uint)0);
#endif	
#endif
	DDHKWD1(HKWD_DD_SCDISKDD, DD_ENTRY_READ, 0, devno);
	errnoval = uphysio(uiop, B_READ, 9, devno, scdisk_strategy,
	    scdisk_mincnt, ext);
#ifdef DEBUG
#ifdef SC_GOOD_PATH
   scdisk_trc(read, exit,(char)0, (uint)devno, (uint)uiop,(uint)chan, 
	      (uint)ext,(uint)errnoval);
#endif
#endif
	DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_READ, errnoval, devno);
	return(errnoval);
}

/************************************************************************/
/*									*/
/*	NAME:	scdisk_write						*/
/*									*/
/*	FUNCTION: SCSI Disk Device Driver Raw Write Routine.		*/
/*		This routine processes raw IO requests from the system.	*/
/*		A buf struct is allocated, and the b_options field	*/
/*		is initialized with the ext argument. Uphysio is called	*/
/*		to begin processing of the raw IO request.		*/
/*		Upon completion of the operation, the buf struct is	*/
/*		freed.							*/
/*									*/
/*	EXECUTION ENVIRONMENT:						*/
/*		This routine can be called by a process.		*/
/*		It can page fault only if called under a process and	*/
/*		the stack is not pinned.				*/
/*									*/
/*	NOTES:								*/
/*									*/
/*	DATA STRUCTURES:						*/
/*									*/
/*	INPUTS:								*/
/*		devno	- device major/minor number			*/
/*		uiop	- Address of uio struct describing request	*/
/*		chan	- unused (will be zero)				*/
/*		ext	- extended parameter				*/
/*									*/
/*	RETURN VALUE DESCRIPTION:					*/
/*		This routine returns the value received from uphysio	*/
/*									*/
/*	ERROR DESCRIPTION:						*/
/*		See above.						*/
/*									*/
/*	EXTERNAL PROCEDURES CALLED:					*/
/*		uphysio							*/
/*									*/
/*									*/
/************************************************************************/

int
scdisk_write(
dev_t		devno,
struct uio	*uiop,
int		chan, 
int		ext)
{
	int	errnoval;


#ifdef DEBUG
#ifdef SC_GOOD_PATH
   scdisk_trc(write, entry,(char)0, (uint)devno, (uint)uiop,(uint)chan, 
	      (uint)ext,(uint)0);
#endif	
#endif
	DDHKWD1(HKWD_DD_SCDISKDD, DD_ENTRY_WRITE, 0, devno);
	errnoval= uphysio(uiop, B_WRITE, 9, devno, scdisk_strategy,
	    scdisk_mincnt,ext);

#ifdef DEBUG
#ifdef SC_GOOD_PATH
   scdisk_trc(write, exit,(char)0, (uint)devno, (uint)uiop,(uint)chan, 
	      (uint)ext,(uint)errnoval);
#endif	
#endif
	DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_WRITE, errnoval, devno);
	return(errnoval);
}

/************************************************************************/
/*									*/
/*	NAME:	scdisk_mincnt						*/
/*									*/
/*	FUNCTION:							*/
/*		This routine is used to breakup raw requests for	*/
/*		processing by uphysio.					*/
/*									*/
/*	EXECUTION ENVIRONMENT:						*/
/*		This routine can be called by a process.		*/
/*		It can page fault only if called under a process and	*/
/*		the stack is not pinned.				*/
/*									*/
/*	NOTES:								*/
/*									*/
/*	DATA STRUCTURES:						*/
/*									*/
/*	INPUTS:								*/
/*		bp		- Pointer to buf structure for request	*/
/*		minparms	- Parameters for breakup		*/
/*									*/
/*	RETURN VALUE DESCRIPTION:					*/
/*									*/
/*	ERROR DESCRIPTION:						*/
/*									*/
/*	EXTERNAL PROCEDURES CALLED:					*/
/*									*/
/*									*/
/************************************************************************/

int
scdisk_mincnt(
struct buf	*bp,
void		*minparms)
{

	struct scdisk_diskinfo *diskinfo;        /* disk information structure*/
	int			hash_key;



#ifdef DEBUG
#ifdef SC_GOOD_PATH
   scdisk_trc(mincnt, entry,(char)0, (uint)bp, (uint)bp->b_bcount,
	      (uint)minparms, (uint)0,(uint)0);
#endif	
#endif
	diskinfo = scdisk_open_list[minor(bp->b_dev) & DK_HASH];
	while (diskinfo != NULL) {
		if (diskinfo->devno == bp->b_dev)
			break;
		diskinfo = diskinfo->next_open;
	}

	if (diskinfo == NULL) {
		return(ENXIO);
	}
	if (bp->b_bcount > diskinfo->max_request)
		/*
		 * If this buf is bigger then our max transfer
		 * size then change count to our max transfer
		 */
		bp->b_bcount = diskinfo->max_request;
	
	
	bp->b_options = (int) minparms;
#ifdef DEBUG
#ifdef SC_GOOD_PATH
   scdisk_trc(mincnt, exit,(char)0, (uint)bp, (uint)bp->b_bcount,
	      (uint)bp->b_options, (uint)0,(uint)0);
#endif	
#endif
	return(0);
}

/************************************************************************/
/*									*/
/*	NAME:   scdisk_ioctl						*/
/*									*/
/*	FUNCTION: SCSI Disk Device Driver IOCTL Routine.		*/
/*		This routine supports the following ioctl commands:	*/
/*		IOCINFO	-  Returns info about the device such as block	*/
/*			   size and number of blocks.			*/
/*		DKIORDSE - Issues a read cmd and returns sense data	*/
/*			   if an error occurs				*/
/*		DKIOWRSE - Issues a read cmd and returns sense data	*/
/*			   if an error occurs				*/
/*		DKIOCMD  - Issues the scsi cmd described in the arg	*/
/*			   and returns the cmd status.			*/
/*									*/
/*	EXECUTION ENVIRONMENT:						*/
/*		This routine can be called by a process.		*/
/*		It can page fault only if called under a process and	*/
/*		the stack is not pinned.				*/
/*									*/
/*	NOTES:								*/
/*									*/
/*	DATA STRUCTURES:						*/
/*		diskinfo	Disk device specific information	*/
/*		scdisk_open_list List of open devices			*/
/*									*/
/*	INPUTS:								*/
/*		devno	- device major/minor number			*/
/*		op	- operation to be performed			*/
/*		arg	- address of user argument structure		*/
/*		flag	- File parameter word				*/
/*		chan	- unused (will be zero)				*/
/*		ext	- extended parameter				*/
/*									*/
/*	RETURN VALUE DESCRIPTION:					*/
/*		The errno values listed in the 'error description'	*/
/*		will be returned to the caller if there is an error.	*/
/*		Otherwise a value of zero will be returned to indicate	*/
/*		successful completion.					*/
/*									*/
/*	ERROR DESCRIPTION:						*/
/*		The following errno values may be returned:		*/
/*		EIO	- device not open				*/
/*			- operation failed				*/
/*		EINVAL	- Invalid parameter in request			*/
/*									*/
/*	EXTERNAL PROCEDURES CALLED:					*/
/*		lockl		unlockl					*/
/*		copyin		copyout					*/
/*		pinu		unpinu					*/
/*		xmattach	xmdetach				*/
/*									*/
/*									*/
/*									*/
/************************************************************************/

int
scdisk_ioctl(
dev_t	devno,
int	op, 
int	arg, 
int	devflag, 
int	chan, 
int	ext)
{
	int			errnoval, rc, dev, hash_key;
	struct dk_cmd		*cmd_ptr;
	struct scdisk_diskinfo	*diskinfo;



	DDHKWD5(HKWD_DD_SCDISKDD, DD_ENTRY_IOCTL, 0, devno, op, arg, 0, 0);

	errnoval = 0;

	/* Lock the scsi disk common lock */
	(void) lockl(&(scdisk_info.lock), LOCK_SHORT);
#ifdef DEBUG
#ifdef SC_GOOD_PATH
   scdisk_trc(ioctl, trc,(char)0, (uint)devno, (uint)op,(uint)arg, 
	      (uint)devflag,(uint)chan);
#endif	
#endif
	/* Search for diskinfo structure in open list */
	dev = minor(devno);
	hash_key = dev & DK_HASH;
	diskinfo = scdisk_open_list[hash_key];
	while (diskinfo != NULL) {
		if (diskinfo->devno == devno)
			break;
		diskinfo = diskinfo->next_open;
	}

	if (diskinfo == NULL) {
		unlockl(&(scdisk_info.lock));
		DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_IOCTL, ENXIO, devno);
		return(ENXIO);
	}

	/* Lock the device's lock */
	(void) lockl(&(diskinfo->lock), LOCK_SHORT);

	/* Unlock the scsi disk common lock */
	unlockl(&(scdisk_info.lock));

	/* Determine operation to perform */
	switch (op) {
	case IOCINFO:
		{
			struct devinfo  devinfo;

			devinfo.devtype = DD_SCDISK;
			devinfo.flags = (uchar) (DF_RAND | DF_FAST | DF_FIXED);

			devinfo.devsubtype = (uchar) DS_PV;
			devinfo.un.scdk.blksize = DK_BLOCKSIZE;
			devinfo.un.scdk.numblks = diskinfo->capacity.lba + 1;
			devinfo.un.scdk.max_request = diskinfo->max_request;
			devinfo.un.scdk.segment_size = 
			    diskinfo->stats.segment_size;
			devinfo.un.scdk.segment_count = 
			    diskinfo->stats.segment_count;
			devinfo.un.scdk.byte_count = 
			    diskinfo->stats.byte_count;
			if (devflag & DKERNEL) {
				bcopy((caddr_t) &devinfo, (caddr_t) arg,
				    sizeof(struct devinfo));
			} else {
				copyout(&devinfo, arg, sizeof(struct devinfo));
			}
			break;
		}
	case DKIORDSE:
	case DKIOWRSE:
		{
			errnoval = scdisk_rdwrse(op, arg, devflag, diskinfo);
			break;
		}
	case DKIOCMD:
		{
			struct sc_buf           *scbuf;
			struct buf              *bp;
			struct scsi             *scsi;
			struct sc_iocmd         iocmd;
			int			segflag;

#ifdef DEBUG
			DKprintf(("scdisk_ioctl: DKIOCMD entered\n"));
#endif

			/*
			 *  Set up segflag for calls to pinu and xmattach and
			 *  copy in the user's structure.
			 */
			if (devflag & DKERNEL) {
				segflag = SYS_ADSPACE;
				bcopy((caddr_t) arg, (caddr_t) &iocmd,
				    sizeof(struct sc_iocmd));
			} else {
				segflag = USER_ADSPACE;
				copyin(arg, &iocmd, sizeof(struct sc_iocmd));
			}

			/* This pass thru ioctl is only available during */
			/* DIAGNOSTIC mode.                              */
			if (diskinfo->mode != DK_DIAG) {
				iocmd.status_validity = 0x00;
				if (devflag & DKERNEL) {
					bcopy((caddr_t) &iocmd, (caddr_t) arg,
					    sizeof(struct sc_iocmd));
				} else {
					copyout(&iocmd, arg,
					    sizeof(struct sc_iocmd));
				}
				unlockl(&(diskinfo->lock));
				DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_IOCTL,
				    EACCES, devno);
				return(EACCES);
			}

			/* Validate request */
			if ((int) iocmd.buffer & 0x03) {
				iocmd.status_validity = 0x00;
				if (devflag & DKERNEL) {
					bcopy((caddr_t) &iocmd, (caddr_t) arg,
					    sizeof(struct sc_iocmd));
				} else {
					copyout(&iocmd, arg,
					    sizeof(struct sc_iocmd));
				}
				unlockl(&(diskinfo->lock));
				DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_IOCTL, 
				    EINVAL, devno);
				return(EINVAL);
			}

			/* Allocate an IOCTL cmd block for operation */

			cmd_ptr = scdisk_cmd_alloc(diskinfo,(uchar) DK_IOCTL);
			ASSERT(cmd_ptr != NULL);
			scbuf = &(cmd_ptr->scbuf);
			scsi = &(scbuf->scsi_command);
			bp = &(scbuf->bufstruct);

			/* Initialize the scbuf for the operation */
			scbuf->bp = NULL;
			scbuf->status_validity = 0x00;
			scbuf->timeout_value = iocmd.timeout_value;
			scbuf->flags = SC_RESUME | diskinfo->reset_delay | 
				iocmd.resvd5;
			scbuf->q_tag_msg = iocmd.q_tag_msg;

			/* Initialize scbuf.bufstruct for the operation */
			bp->b_dev = diskinfo->adapter_devno;
			if (iocmd.flags & B_READ)
				bp->b_flags = (B_BUSY | B_READ);
			else
				bp->b_flags = B_BUSY;
			bp->b_bcount = iocmd.data_length;
			bp->b_work = 0x00;
			bp->b_error = 0x00;
			bp->b_un.b_addr = iocmd.buffer;
			bp->b_blkno = 0x00;
			bp->b_iodone = scdisk_iodone;
			bp->b_xmemd.aspace_id = XMEM_INVAL;
			bp->b_event = EVENT_NULL;

			/* Initialize SCSI cmd for operation */
			scsi->scsi_length = iocmd.command_length;
			scsi->scsi_id = diskinfo->scsi_id;
			scsi->flags = (iocmd.flags & ~(B_READ | B_WRITE));
			scsi->scsi_cmd.scsi_op_code = iocmd.scsi_cdb[0];
			scsi->scsi_cmd.lun = iocmd.scsi_cdb[1];
			scsi->scsi_cmd.scsi_bytes[0] = iocmd.scsi_cdb[2];
			scsi->scsi_cmd.scsi_bytes[1] = iocmd.scsi_cdb[3];
			scsi->scsi_cmd.scsi_bytes[2] = iocmd.scsi_cdb[4];
			scsi->scsi_cmd.scsi_bytes[3] = iocmd.scsi_cdb[5];
			scsi->scsi_cmd.scsi_bytes[4] = iocmd.scsi_cdb[6];
			scsi->scsi_cmd.scsi_bytes[5] = iocmd.scsi_cdb[7];
			scsi->scsi_cmd.scsi_bytes[6] = iocmd.scsi_cdb[8];
			scsi->scsi_cmd.scsi_bytes[7] = iocmd.scsi_cdb[9];
			scsi->scsi_cmd.scsi_bytes[8] = iocmd.scsi_cdb[10];
			scsi->scsi_cmd.scsi_bytes[9] = iocmd.scsi_cdb[11];

			/* Check for data transfer on operation */
			if (bp->b_bcount != 0) {
				if (bp->b_bcount > diskinfo->max_request) {
					scdisk_free_cmd(cmd_ptr);
					iocmd.status_validity = 0x00;
					if (devflag & DKERNEL) {
						bcopy((caddr_t) &iocmd,
						    (caddr_t) arg, sizeof
						    (struct sc_iocmd));
					} else {
						copyout(&iocmd, arg, sizeof
						    (struct sc_iocmd));
					}
					unlockl(&(diskinfo->lock));
					DDHKWD1(HKWD_DD_SCDISKDD,
					    DD_EXIT_IOCTL, EINVAL, devno);
					return(EINVAL);
				}
				/* Attach and pin user memory for transfer */
				rc = xmattach((uint)bp->b_un.b_addr,
				    (int) bp->b_bcount,
				    &(bp->b_xmemd), (int) segflag);
				if (rc != XMEM_SUCC) {
					scdisk_free_cmd(cmd_ptr);
					iocmd.status_validity = 0x00;
					if (devflag & DKERNEL) {
						bcopy((caddr_t) &iocmd,
						    (caddr_t) arg, sizeof
						    (struct sc_iocmd));
					} else {
						copyout((caddr_t) &iocmd,
						    (caddr_t) arg, sizeof
						    (struct sc_iocmd));
					}
					unlockl(&(diskinfo->lock));
					DDHKWD1(HKWD_DD_SCDISKDD,
					    DD_EXIT_IOCTL, EFAULT, devno);
					return(EFAULT);
				}
				rc = pinu((caddr_t) bp->b_un.b_addr,
				    (int) bp->b_bcount, (short) segflag);
				if (rc != PIN_SUCC) {
					xmdetach(&(bp->b_xmemd));
					scdisk_free_cmd(cmd_ptr);
					iocmd.status_validity = 0x00;
					if (devflag & DKERNEL) {
						bcopy((caddr_t) &iocmd,
						    (caddr_t) arg, sizeof
						    (struct sc_iocmd));
					} else {
						copyout(&iocmd, arg, sizeof
						    (struct sc_iocmd));
					}
					unlockl(&(diskinfo->lock));
					DDHKWD1(HKWD_DD_SCDISKDD,
					    DD_EXIT_IOCTL, rc, devno);
					return(rc);
				}
			}

			/* Start execution of operation */

			cmd_ptr->subtype = DK_IOCMD;
			cmd_ptr->retry_count = 0;
			cmd_ptr->segment_count = 0;
			cmd_ptr->soft_resid = 0;

			cmd_ptr->intrpt = 1;
			cmd_ptr->flags |= DK_READY;
			scdisk_start(diskinfo);

			/* Wait for completion */
			scdisk_sleep((uchar *)&(cmd_ptr->intrpt),
				     (uint *)&(bp->b_event));


			if (bp->b_bcount != 0) {
				/* detach and unpin user memory */
				unpinu((caddr_t) bp->b_un.b_addr,
				    (int) bp->b_bcount, (short) segflag);
				xmdetach(&(bp->b_xmemd));
			}

			errnoval = bp->b_error;
			iocmd.status_validity = scbuf->status_validity;
			iocmd.scsi_bus_status = scbuf->scsi_status;
			iocmd.adapter_status = scbuf->general_card_status;
			iocmd.adap_q_status = scbuf->adap_q_status;
			/* Deallocate cmd block from operation */
			scdisk_free_cmd(cmd_ptr);

			/* Return arg stucture to user */
			if (devflag & DKERNEL) {
				bcopy((caddr_t) &iocmd, (caddr_t) arg,
				    sizeof(struct sc_iocmd));
			} else {
				copyout(&iocmd, arg, sizeof(struct sc_iocmd));
			}

#ifdef DEBUG
			DKprintf(("scdisk_ioctl: DKIOCMD exiting\n"));
#endif
			break;
		}
	default:
		errnoval = EINVAL;
	}
#ifdef DEBUG
#ifdef SC_GOOD_PATH
   scdisk_trc(ioctl, trc,(char)1, (uint)devno, (uint)op,(uint)arg, 
	      (uint)devflag,(uint)errnoval);
#endif	
#endif


	unlockl(&(diskinfo->lock));
	DDHKWD1(HKWD_DD_SCDISKDD, DD_EXIT_IOCTL, errnoval, devno);
	return(errnoval);
}

/************************************************************************/
/*									*/
/*	NAME:   scdisk_rdwrse						*/
/*									*/
/*	FUNCTION: Read and Write with Sense Subroutine.			*/
/*		This routine supports the following ioctl commands:	*/
/*		DKIORDSE - Issues a read cmd and returns sense data	*/
/*			   if an error occurs				*/
/*		DKIOWRSE - Issues a read cmd and returns sense data	*/
/*			   if an error occurs				*/
/*									*/
/*	EXECUTION ENVIRONMENT:						*/
/*		This routine can be called by a process.		*/
/*		It can page fault only if called under a process and	*/
/*		the stack is not pinned.				*/
/*									*/
/*	NOTES:								*/
/*									*/
/*	DATA STRUCTURES:						*/
/*		diskinfo	Disk device specific information	*/
/*		scdisk_open_list List of open devices			*/
/*									*/
/*	INPUTS:								*/
/*		devno	- device major/minor number			*/
/*		op	- operation to be performed			*/
/*		arg	- address of user argument structure		*/
/*		flag	- File parameter word				*/
/*		chan	- unused (will be zero)				*/
/*		ext	- extended parameter				*/
/*									*/
/*	RETURN VALUE DESCRIPTION:					*/
/*		The errno values listed in the 'error description'	*/
/*		will be returned to the caller if there is an error.	*/
/*		a value of zero will be returned to indicate		*/
/*		successful completion.					*/
/*									*/
/*	ERROR DESCRIPTION:						*/
/*		The following errno values may be returned:		*/
/*		EIO	- device not open				*/
/*			- operation failed				*/
/*		EINVAL	- Invalid parameter in request			*/
/*									*/
/*	EXTERNAL PROCEDURES CALLED:					*/
/*		lockl		unlockl					*/
/*		copyin		copyout					*/
/*		pinu		unpinu					*/
/*		xmattach	xmdetach				*/
/*									*/
/*									*/
/*									*/
/************************************************************************/

int
scdisk_rdwrse(
int			op, 
int			arg, 
int			devflag, 
struct scdisk_diskinfo	*diskinfo)
{
	int		errnoval, rc, lba, blk_cnt, segflag;
	struct dk_cmd	*cmd_ptr;
	struct sc_rdwrt	dk_rdwrt;
	struct sc_buf	*scbuf;
	struct buf	*bp;
	struct scsi	*scsi;

#ifdef DEBUG
	DKprintf(("Entering scdisk_rdwrse op[0x%x] arg[0x%x]\n", op, arg));
#endif

	errnoval = 0;

	/*
	 *  Set up segflag for calls to pinu and xmattach and copy in the
	 *  user's structure.
	 */
	
	if (devflag & DKERNEL) {
		segflag = SYS_ADSPACE;
		bcopy((caddr_t) arg, (caddr_t) &dk_rdwrt,
		    sizeof(struct sc_rdwrt));
	} else {
		segflag = USER_ADSPACE;
		copyin(arg, &dk_rdwrt, sizeof(struct sc_rdwrt));
	}

	/* Validate request */
	if ((dk_rdwrt.data_length & 0x01ff) ||
	    (dk_rdwrt.data_length > diskinfo->max_request)) {
		dk_rdwrt.status_validity = 0x00;
		if (devflag & DKERNEL) {
			bcopy((caddr_t) &dk_rdwrt, (caddr_t) arg,
			    sizeof(struct sc_rdwrt));
		} else {
			copyout(&dk_rdwrt, arg, sizeof(struct sc_rdwrt));
		}
		return(EINVAL);
	}

	/* Allocate an IOCTL cmd block for operation */

	cmd_ptr = scdisk_cmd_alloc(diskinfo,(uchar) DK_IOCTL);
	ASSERT(cmd_ptr != NULL);
	scbuf = &(cmd_ptr->scbuf);
	scsi = &(scbuf->scsi_command);
	bp = &(scbuf->bufstruct);

	/* Initialize the scbuf for the operation */
	scbuf->bp = NULL;
	scbuf->status_validity = 0x00;
	scbuf->timeout_value = dk_rdwrt.timeout_value;
	scbuf->flags = SC_RESUME | diskinfo->reset_delay| dk_rdwrt.q_flags;
	scbuf->q_tag_msg = dk_rdwrt.q_tag_msg;
	/* Initialize scbuf.bufstruct for the operation */
	bp->b_dev = diskinfo->adapter_devno;
	bp->b_flags = B_BUSY;
	if (op == DKIORDSE) {
		bp->b_flags |= B_READ;
	}
	bp->b_bcount = (uint) dk_rdwrt.data_length;
	bp->b_work = 0x00;
	bp->b_error = 0x00;
	bp->b_un.b_addr = dk_rdwrt.buffer;
	bp->b_blkno = dk_rdwrt.logical_blk_addr;
	bp->b_iodone = scdisk_iodone;
	bp->b_xmemd.aspace_id = XMEM_INVAL;
	bp->b_event = EVENT_NULL;

	/* Initialize SCSI cmd for operation */
	/*
	 * The blocksize must be hard-coded here because if
	 * the device was opened in diag mode, the capacity
	 * data will be invalid.
	 */
	blk_cnt = bp->b_bcount / DK_BLOCKSIZE;
	lba = bp->b_blkno;
	scsi->scsi_id = diskinfo->scsi_id;
	scsi->flags = (ushort) diskinfo->async_flag;
	if ((blk_cnt & 0xffffff00) || (lba & 0xffe00000)) {
		scsi->scsi_length = 10;
		if (op == DKIORDSE) {
			scsi->scsi_cmd.scsi_op_code = SCSI_READ_EXTENDED;
		} else {
			scsi->scsi_cmd.scsi_op_code = SCSI_WRITE_EXTENDED;
		}
		scsi->scsi_cmd.lun = (diskinfo->lun_id << 5);
		scsi->scsi_cmd.scsi_bytes[0] = ((lba >> 24) & 0xff);
		scsi->scsi_cmd.scsi_bytes[1] = ((lba >> 16) & 0xff);
		scsi->scsi_cmd.scsi_bytes[2] = ((lba >> 8) & 0xff);
		scsi->scsi_cmd.scsi_bytes[3] = (lba & 0xff);
		scsi->scsi_cmd.scsi_bytes[4] = 0x00;
		scsi->scsi_cmd.scsi_bytes[5] = ((blk_cnt >> 8) & 0xff);
		scsi->scsi_cmd.scsi_bytes[6] = (blk_cnt & 0xff);
		scsi->scsi_cmd.scsi_bytes[7] = 0x00;
	} else {
		scsi->scsi_length = 0x06;
		if (op == DKIORDSE) {
			scsi->scsi_cmd.scsi_op_code=SCSI_READ;
		} else {
			scsi->scsi_cmd.scsi_op_code=SCSI_WRITE;
		}
		scsi->scsi_cmd.lun = ((diskinfo->lun_id << 5) |
		    ((lba >> 16) & 0x1f));
		scsi->scsi_cmd.scsi_bytes[0] = ((lba >> 8) & 0xff);
		scsi->scsi_cmd.scsi_bytes[1] = (lba & 0xff);
		scsi->scsi_cmd.scsi_bytes[2] = blk_cnt;
		scsi->scsi_cmd.scsi_bytes[3] = 0x00;
	}
	/*
	 * Attach and pin user memory (buffer and req. sense 
	 * buffer) for transfer from bottom half routines.
	 */
	if (bp->b_bcount != 0) {

		rc = xmattach((uint)bp->b_un.b_addr, (int) bp->b_bcount,
		    &(bp->b_xmemd), (int) segflag);
		if (rc != XMEM_SUCC) {
			scdisk_free_cmd(cmd_ptr);
			dk_rdwrt.status_validity = 0x00;
			if (devflag & DKERNEL) {
				bcopy((caddr_t) &dk_rdwrt, (caddr_t) arg, 
				    sizeof (struct sc_rdwrt));
			} else {
				copyout((caddr_t) &dk_rdwrt, (caddr_t) arg,
				    sizeof (struct sc_rdwrt));
			}
			return(EFAULT);
		}
		rc = pinu((caddr_t) bp->b_un.b_addr, (int) bp->b_bcount,
		    (short) segflag);
		if (rc != PIN_SUCC) {
			xmdetach(&(bp->b_xmemd));
			scdisk_free_cmd(cmd_ptr);
			dk_rdwrt.status_validity = 0x00;
			if (devflag & DKERNEL) {
				bcopy((caddr_t) &dk_rdwrt, (caddr_t) arg, 
				    sizeof (struct sc_rdwrt));
			} else {
				copyout((caddr_t) &dk_rdwrt, (caddr_t) arg,
				    sizeof (struct sc_rdwrt));
			}
			return(rc);
		}
	}
	diskinfo->ioctl_req_sense.xmemd.aspace_id = XMEM_INVAL;
	if (dk_rdwrt.req_sense_length != 0) {
		rc = xmattach((uint) dk_rdwrt.request_sense_ptr,
		    (int)dk_rdwrt.req_sense_length,
		    &(diskinfo->ioctl_req_sense.xmemd), (int) segflag);
		if (rc != XMEM_SUCC) {
			if (bp->b_bcount != 0) {
				unpinu((caddr_t) bp->b_un.b_addr,
				    (int) bp->b_bcount, (short) segflag);
				xmdetach(&(bp->b_xmemd));
			}
			scdisk_free_cmd(cmd_ptr);
			dk_rdwrt.status_validity = 0x00;
			if (devflag & DKERNEL) {
				bcopy((caddr_t) &dk_rdwrt, (caddr_t) arg, 
				    sizeof (struct sc_rdwrt));
			} else {
				copyout(&dk_rdwrt, arg, sizeof
				    (struct sc_rdwrt));
			}
			return((int) EFAULT);
		}
		rc = pinu((caddr_t) dk_rdwrt.request_sense_ptr,
		    (int) dk_rdwrt.req_sense_length, (short) segflag);
		if (rc != PIN_SUCC) {
			xmdetach(&(diskinfo-> ioctl_req_sense.xmemd));
			if (bp->b_bcount != 0) {
				unpinu((caddr_t) bp->b_un.b_addr,
				    (int) bp->b_bcount, (short) segflag);
				xmdetach(&(bp->b_xmemd));
			}
			scdisk_free_cmd(cmd_ptr);
			dk_rdwrt.status_validity = 0x00;
			if (devflag & DKERNEL) {
				bcopy((caddr_t) &dk_rdwrt, (caddr_t) arg, 
				    sizeof (struct sc_rdwrt));
			} else {
				copyout(&dk_rdwrt, arg, sizeof
				    (struct sc_rdwrt));
			}
			return(rc);
		}
	}
	/*
	 * Set up for xmemout of request sense data in 
	 * bottom half.
	 */
	diskinfo->ioctl_req_sense.count = dk_rdwrt.req_sense_length;
	diskinfo->ioctl_req_sense.buffer = dk_rdwrt.request_sense_ptr;

	/* Start execution of operation */

	cmd_ptr->subtype = DK_RDWR_SENSE;
	cmd_ptr->retry_count = 0;
	cmd_ptr->segment_count = 0x00;
	cmd_ptr->soft_resid = 0x00;

	cmd_ptr->intrpt = 1;
	cmd_ptr->flags |= DK_READY;
	scdisk_start(diskinfo);

	/* Wait for completion */

	scdisk_sleep((uchar *)&(cmd_ptr->intrpt),(uint *)&(bp->b_event));	

	/* detach and unpin user memory */
	if (bp->b_bcount != 0) {
		unpinu((caddr_t) bp->b_un.b_addr, (int) bp->b_bcount,
		    (short) segflag);
		xmdetach(&(bp->b_xmemd));
	}
	if (dk_rdwrt.req_sense_length != 0) {
		unpinu((caddr_t) dk_rdwrt.request_sense_ptr,
		    (int) dk_rdwrt.req_sense_length, (short) segflag);
		xmdetach(&(diskinfo->ioctl_req_sense.xmemd));
	}

	/* Check for IO error to return sense info */
	errnoval = bp->b_error;
	dk_rdwrt.status_validity = scbuf->status_validity;
	dk_rdwrt.scsi_bus_status = scbuf->scsi_status;
	dk_rdwrt.adapter_status = scbuf->general_card_status;
	dk_rdwrt.adap_q_status = scbuf->adap_q_status;
	/*
	 * For the special case where a resid is returned
	 * from the SCSI adapter driver, but no error is
	 * set, set an adapter hardware failure.
	 */
	if ((errnoval == EIO) && (dk_rdwrt.status_validity == 0x00)) {
		dk_rdwrt.status_validity = SC_ADAPTER_ERROR;
		dk_rdwrt.adapter_status = SC_ADAPTER_HDW_FAILURE;
	}
	/* Deallocate cmd block from operation */
	scdisk_free_cmd(cmd_ptr);

	/* Return arg stucture to user */
	if (devflag & DKERNEL) {
		bcopy((caddr_t) &dk_rdwrt, (caddr_t) arg,
		    sizeof(struct sc_rdwrt));
	} else {
		copyout(&dk_rdwrt,arg,sizeof(struct sc_rdwrt));
	}
	return(errnoval);
}

/************************************************************************/
/*									*/
/*	NAME:   scdisk_alloc						*/
/*									*/
/*	FUNCTION: Allocates memory from the pinned heap for		*/
/*		  the ioctl command blocks and buf command blocks.      */
/*		  It also initializes this command blocks		*/
/*									*/
/*	EXECUTION ENVIRONMENT:						*/
/*		This routine can be called by a process.		*/
/*		It can page fault only if called under a process and	*/
/*		the stack is not pinned.				*/
/*									*/
/*	NOTES:								*/
/*									*/
/*	DATA STRUCTURES:						*/
/*		diskinfo	Disk device specific information	*/
/*		dk_cmd	        Disk device command block               */
/*									*/
/*	INPUTS:								*/
/*		diskinfo	Disk device specific information	*/
/*									*/
/*	RETURN VALUE DESCRIPTION:					*/
/*		The errno values listed in the 'error description'	*/
/*		will be returned to the caller if there is an error.	*/
/*		a value of zero will be returned to indicate		*/
/*		successful completion.					*/
/*									*/
/*	ERROR DESCRIPTION:						*/
/*		The following errno values may be returned:		*/
/*		ENOMEM  - Xmalloc failed				*/
/*									*/
/*	EXTERNAL PROCEDURES CALLED:					*/
/*		xmalloc							*/
/*									*/
/*									*/
/************************************************************************/

int scdisk_alloc(struct scdisk_diskinfo *diskinfo)
{

	int 	i;				/* general counter 	*/



	/* 
	 * Allocate cmd blocks for buf's
	 */

	diskinfo->cmd_pool = 
		(struct dk_cmd *) xmalloc(
		(uint) ((diskinfo->queue_depth)*(sizeof(struct dk_cmd))),
		2,pinned_heap);

	if (diskinfo->cmd_pool == NULL) {
		/*
		 * Xmalloc failed
		 */
		return(ENOMEM);
	}
	


	/*
	 * Initialize_cmds
	 */

	bzero((caddr_t) diskinfo->cmd_pool, 
	      ((diskinfo->queue_depth)*(sizeof(struct dk_cmd))));

	for (i = 0; i< (diskinfo->queue_depth); i++) {

		diskinfo->cmd_pool[i].status = DK_FREE;
		diskinfo->cmd_pool[i].diskinfo = diskinfo;
		diskinfo->cmd_pool[i].scbuf.flags = 
		    diskinfo->reset_delay;
		diskinfo->cmd_pool[i].scbuf.q_tag_msg = SC_NO_Q;
		diskinfo->cmd_pool[i].scbuf.bufstruct.b_event = 
		    EVENT_NULL;
	}
	diskinfo->pool_index = 0;

	return(0);
}


/************************************************************************/
/*									*/
/*	NAME:   scdisk_init_cmds					*/
/*									*/
/*	FUNCTION: This routine initializes the diskinfo structure's    	*/
/*		  commands at config init time.			       	*/
/*									*/
/*	EXECUTION ENVIRONMENT:						*/
/*		This routine can be called by a process.		*/
/*		It can page fault only if called under a process and	*/
/*		the stack is not pinned.				*/
/*									*/
/*	NOTES:								*/
/*									*/
/*	DATA STRUCTURES:						*/
/*		diskinfo	Disk device specific information	*/
/*									*/
/*	INPUTS:								*/
/*		diskinfo	Disk device specific information	*/
/*									*/
/*	RETURN VALUE DESCRIPTION:					*/
/*		None							*/
/*									*/
/*	EXTERNAL PROCEDURES CALLED:					*/
/*		None							*/
/*									*/
/************************************************************************/

void scdisk_init_cmds(struct scdisk_diskinfo *diskinfo)
{

	
	diskinfo->ioctl_cmd.status = DK_FREE;
	diskinfo->ioctl_cmd.diskinfo = diskinfo;
	diskinfo->ioctl_cmd.scbuf.flags =
		diskinfo->reset_delay;
	diskinfo->ioctl_cmd.scbuf.q_tag_msg = SC_NO_Q;
	diskinfo->ioctl_cmd.scbuf.bufstruct.b_event =
		EVENT_NULL;
	
	diskinfo->reset_cmd.status = DK_FREE;
	diskinfo->reset_cmd.diskinfo = diskinfo;
	diskinfo->reset_cmd.scbuf.flags =
		diskinfo->reset_delay;
	diskinfo->reset_cmd.scbuf.q_tag_msg = SC_NO_Q;
	diskinfo->reset_cmd.scbuf.bufstruct.b_event =
		EVENT_NULL;

	diskinfo->reqsns_cmd.status = DK_FREE;
	diskinfo->reqsns_cmd.diskinfo = diskinfo;
	diskinfo->reqsns_cmd.scbuf.flags =
		diskinfo->reset_delay;
	diskinfo->reqsns_cmd.scbuf.q_tag_msg = SC_NO_Q;
	diskinfo->reqsns_cmd.scbuf.bufstruct.b_event =
		EVENT_NULL;
	
	diskinfo->q_recov_cmd.status = DK_FREE;
	diskinfo->q_recov_cmd.diskinfo = diskinfo;
	diskinfo->q_recov_cmd.scbuf.flags =
		diskinfo->reset_delay;
	diskinfo->q_recov_cmd.scbuf.q_tag_msg = SC_NO_Q;
	diskinfo->q_recov_cmd.scbuf.bufstruct.b_event =
		EVENT_NULL;
	
	diskinfo->writev_cmd.status = DK_FREE;
	diskinfo->writev_cmd.diskinfo = diskinfo;
	diskinfo->writev_cmd.scbuf.flags =
		diskinfo->reset_delay;
	diskinfo->writev_cmd.scbuf.q_tag_msg = SC_NO_Q;
	diskinfo->writev_cmd.scbuf.bufstruct.b_event =
		EVENT_NULL;

	
	diskinfo->reassign_cmd.status = DK_FREE;
	diskinfo->reassign_cmd.diskinfo = diskinfo;
	diskinfo->reassign_cmd.scbuf.flags =
		diskinfo->reset_delay;
	diskinfo->reassign_cmd.scbuf.q_tag_msg = SC_NO_Q;
	diskinfo->reassign_cmd.scbuf.bufstruct.b_event =
		EVENT_NULL;


	return;
}
