/*
 * The Western Digital controller has 8K of RAM.  We use (32 * 256)
 * for receive buffers and (6 * 256) for transmit buffers.
 */

#define		WD_NUMRBUFS   20 
#define		WD_NUMXBUFS   2
#define		WD_RBUFSIZE   256
#define		WD_XBUFSIZE   1536 

#define		ETH_ADDR_SIZE 6 

#define		ETHER_TRAILER_LOW	ETHERTYPE_TRAIL
#define		ETHER_TRAILER_HIGH	ETHERTYPE_TRAIL+ETHERTYPE_NTRAILER

#define 	MAX_DPDU		ETHERMTU+sizeof(struct ether_header)
#define		MIN_DPDU		ETHERMIN+sizeof(struct ether_header)

/*
 * NIC bit masks.
 */
#define 	STP	0x01	        /* command register bit masks */
#define		STA	0x02
#define		TXP	0x04
#define		RD0	0x08
#define		RD1	0x10
#define		RD2	0x20
#define		PS0	0x40
#define		PS1	0x80

#define		PRX	0x01		/* interrupt status bit masks */
#define		PTX	0x02
#define		RXE	0x04
#define		TXE	0x08
#define		OVW	0x10
#define		CNT	0x20
#define		RDC	0x40
#define		RST	0x80

#define		PRXE	0x01		/* iterrupt mask register */
#define		PTXE	0x02
#define		RXEE	0x04
#define		TXEE	0x08
#define		OVWE	0x10
#define		CNTE	0x20
#define		RDCE	0x40

#define		WTS	0x01		/* data configuration register */
#define		BOS	0x02
#define		LAS	0x04
#define		LS	0x08
#define		DAR	0x10
#define		FT0	0x20
#define		FT1	0x40

#define		CRC	0x01		/* transmit config register */
#define		LB0	0x02
#define		LB1	0x04
#define		ATD	0x08
#define		OFST	0x10

#define		PTXS	0x01		/* transmit status register */
#define		COL	0x02
#define		ABT	0x04
#define		CRS	0x08
#define		FU	0x10
#define		CDH	0x20
#define		OWC	0x40

#define		SEP	0x01		/* receive configuration register */
#define		RAR	0x02
#define		AB	0x04
#define		AM	0x08
#define		PRO	0x10
#define		MON	0x20

#define		PRXS	0x01		/* receive status register */
#define		CRCS	0x02
#define		FAE	0x04
#define		FO	0x08
#define		MPA	0x10
#define		PHY	0x20
#define		DIS	0x40
#define		DFR	0x80

/*
 * Western Digital 8300e map of port-mapped I/O space.
 * 0x00	      - control register
 * 0x01-0x08  - not used 
 * 0x09-0x0F  - LAN address ROM, address, id, and checksum.
 * 0x10-0x1F  - NS DP8390 NIC (net interface controller)
 *
 * NIC weird behavior description: The NIC address space 
 * is mapped into 3 pages although only page 1 and 2 are used in this 
 * driver. Some registers in page 1 behave differently when written to 
 * or read from.  The page 1 write values are declared in the struct below;
 * the page 1 read values and the page 2 values are declared with macros.
 */

struct wddevice {
	char	cntrl;			/* wd8003e control register      */
	char	hole0[7];		/* not used on this controller   */
	char	addr[ETH_ADDR_SIZE];	/* hardware address in ROM       */
	char	wdid;			/* wd identification (3 = ether) */
	char	cksum;			/* 8 bit sum of 00-0f = 0xff     */

	char	cr;			/* NIC cont reg (W/R all pages   */
	char	pstart;			/* start of recv buffer ring     */
#define clda0	pstart			/* current local DMA addr 0      */
#define par0	pstart			/* start of Phys address reg     */
	char	pstop;			/* end of recv buffer ring       */
#define clda1	pstop			/* current local DMA addr 1      */
	char 	bndry;			/* we read recv buffer here      */
	char	tpsr;			/* transmit page start address   */
#define tsr	tpsr			/* transmit status register      */
	char	tbcr0;			/* transmit byte count reg 0     */
#define ncr	tbcr0			/* number of collisions reg      */
	char 	tbcr1;			/* transmit byte count reg 1     */
#define fifo	tbcr1			/* NIC internal fifo (not used)  */
	char 	isr;			/* interrupt status register     */
#define curr	isr			/* NIC writes to recv buff here  */
	char	rsar0;			/* remote DMA start (not used)   */
#define crda0	rsar0			/* curr rem DMA addr (not used)  */
#define mar0	rsar0			/* start of multicast address    */
	char	rsar1;			/* curr rem DMA addr (not used)  */
#define crda1	rsar1			/* curr rem DMA addr (not used)  */ 
	char 	rbcr0;			/* rem byte count DMA (not used) */
	char	rbcr1;			/* rem byte count DMA (not used) */
	char	rcr;			/* receive control register      */
#define rsr	rcr			/* recieve status register       */
	char	tcr;			/* transmit control register     */
#define cntr0	tcr			/* frame alignment error count   */
	char  	dcr;			/* data control register         */
#define cntr1	dcr			/* CRC error count               */
	char 	imr;			/* iterrupt mask register        */
#define cntr2	imr			/* missed packet error count     */
} wddevice;


