.if \n(1T .bp
.NH 1
\&Network Services
..IX network services
.LP
..IX NFS
..IX "Network File System"
This guide gives an overview of the network services available in the
Sun 4.1 release.  To appreciate the design of these services, it's
necessary to see that SunOS is
.I structurally
a network UNIX system, and is designed to evolve as network technology
changes.
.LP
SunOS originally diverged from the 4.2BSD UNIX system, a system that
already strained at the limits of the UNIX system's original simplicity of
design.  It was with 4.2BSD that many of the network services found in
SunOS were first introduced.  Fortunately, the Berkeley designers found
alternatives to wedging everything into the kernel.  They implemented
network services by offloading certain jobs to specialized daemons (server
processes) working in close cooperation with the kernel, rather than by
adding all new code to the kernel itself. 
Though NFS is primarily kernel based (using a daemon only to make
system calls), SunOS has continued this line of
development.  Its expanding domain of network services is uniformly built
upon a daemon (server) based architecture.
Examples of server daemons are the
.I portmapper ,
the network naming service (NIS),
the Remote Execution Facility (REX), the Network Lock 
Manager, and the Status Monitor.
..IX REX
.SH
\&Terminology
.ft I
A machine that provides resources to the network is called
a ``server'', while a machine that employs these resources is
called a ``client''.
..IX "server machines"
..IX "client machines"
A machine may be both a server and a client, and when NFS resources
(files and directories) are at issue, often is.  A person logged in on
a client machine is a ``user'', while a program or set of programs that
run on a client is an ``application''.
..IX user
..IX application
There is a distinction between the code implementing the operations of a
filesystem, (called ``filesystem operations'') and the data making up the
filesystem's structure and contents (called ``filesystem data'').
..IX filesystem operations
..IX filesystem data
.LP
.ft R
..IX servers "network services"
.LP
Network services are added to SunOS by means of server processes that are
based upon Sun's RPC (Remote Procedure Call) mechanism.  These servers are
executed on all machines that provide the service.
Sun daemons differ significantly from
those that were inherited from Berkeley in that most of them are based on RPC.
As a consequence, they automatically benefit from the services provided by
RPC, and the External Data Representation (XDR) that it is built
upon \(em for example, the data portability provided by XDR and RPC's
authentication system.
.LP
Anything built with RPC/XDR is automatically a network application, as is
anything that stores data in NFS files, even if it doesn't use RPC directly.
Furthermore, in so far as network applications can presume the functionality
of other network applications and call upon their services, all network
applications are network services as well.  The RPC/XDR environment then,
is inherently
.I extensible .
New network services can be easily added by building upon the foundation
already in place.  In SunOS, then, network services are analogous to UNIX
commands \(em  anyone can add one, and when they do they are effectively
extending the ``system''.
.SH
NOTE
.LP
The term Open Network Computing (ONC) is based on RPC utilities only (such as
REX, NIS, Lock Manager, and Status Monitor).
The other network utilities described here are not considered part of ONC.
.LP
.NH 2
\&The Major Network Services
..IX network "major services"
.LP
The
.I "Remote Procedure Call (RPC)"
facility is a library of procedures that provide a means whereby one
process (the caller process) can have another process (the server
process) execute a procedure call, as if the caller process had executed
the procedure call in its own address space (as in the local model of a
procedure call).
..IX "caller process"
..IX "server process"
Because the caller and the server are now two separate processes, they
no longer have to live on the same physical machine.
..IX "Remote Procedure Call"
..IX RPC
.LP
The
.I "External Data Representation (XDR)" is
a specification for the portable data representation standard.
RPC uses XDR to ensure that data is represented the same on different
computers, operating systems, and computer languages.
In SunOS 4.1 XDR is implemented through the socket interface, yet
allows programmers to have a standardized
access to sockets without being concerned about the low-level
details of socket-based IPC.
.LP
..IX NFS
The
.I "Network File System (NFS)" ,
is an operating system-independent service
which allows users to mount directories, even root directories,
across the network, and then to treat those directories as if
they were local.  There is also an option for a secure mount
involving DES authentication of user and host\(emfor more
information about it, see the
.I "Secure Networking Features"
chapter of
.I "Security Features Guide" .
..IX portmapper "" \fIportmapper\fP
.LP
.I portmapper 
is a system service upon which all other RPC-based services rely.
It's a kind of registrar that keeps track of the correspondence
between ports (logical communications channels) and services
on a machine, and provides a standard way for a client to look up
the port number of any RPC program supported by the server.
But in effect, only RPC programs use it.
..IX "Network Information Service"
.LP
Sun's
.I "Network Information Service"
is a network service designed to ease the
job of administering large networks.
NIS is a replicated, read-only, distributed database service.
Network file system clients use it to access network-wide data in
a manner that is entirely independent of the relative locations
of the client and the server.  The NIS database typically provides
password, group, network, and host information.
.LP
..IX locking
..IX "Network Lock Manager"
As part of its System V compatibility program, Sun now supports
System-V (SVID) compatible advisory file and record locking for
both local and NFS mounted files.  User programs simply issue
.I lockf() 
and
.I fcntl() 
system calls to set and test file locks \(em these calls are then
processed by
.I "Network Lock Manager"
daemons, which maintain locks at
the network level, even in the face of multiple machine crashes.
.LP
..IX "Network Status Monitor"
The lock-manager daemons are able to manage machine crashes
because they are based upon a general purpose
.I "Network Status Monitor" .
This monitor provides a mechanism by which network
applications can detect machine reboots and trigger
application-specific recovery mechanisms.  The Lock Manager is therefore
equipped with a flexible fault-tolerant recovery capability.
.LP
..IX NIS
There are other network services \(em NIS and REX\**
..IX REX
.FS
These, however, are not 
.I fundamental
network services, in the same sense as NFS.  REX, for
example, cannot be guaranteed to be portable to a non-UNIX environment.
This is true because the executability of a program depends on
many environmental factors \(em from machine architecture to
operating-system services \(em that are not universally available.
.FE
are two
obvious examples \(em and there are many others that are certainly
services in the broad sense.  This section, however, is intended
as an introduction, and it covers only the fundamental services
noted above.
.NH 2
\&Network Programming Manual Overview
.LP
This
.I "Network Programming" 
manual contains this 
.I "Network Services"
overview and then three major sections.  In this overview the
fundamental network services are introduced without dealing 
with any protocol or implementation related issues.
.LP
PART ONE focuses on Sun's network programming mechanisms.  It
includes:
.IP \(bu
The
.I "\fIrpcgen\fP Programming Guide" ,
which introduces the
.I rpcgen 
protocol compiler and the C-like language that it uses to specify
RPC applications and define network data.  In almost all cases,
.I rpcgen 
will allow network applications developers to avoid the use of
lower-level RPC mechanisms.
.IP \(bu
The
.I "Remote Procedure Call Programming Guide"
is intended for programmers who wish to understand the lower-level
RPC mechanisms.  Readers are assumed to be familiar with the C
language and to have a working knowledge of network theory.
.IP \(bu
The
.I "External Data Representation: Sun Technical Notes" ,
which introduces XDR and explains the justification for its
``canonical'' approach to network data interchange.  This section
also gives Sun implementation information and a few examples of
advanced XDR usage.
.LP
PART TWO includes a number of number of protocol specifications.
Both the
.I "External Data Representation Protocol Specification"
and
.I "Remote Procedure Call Specification"
have been published as a DARPA RFC
(Request for Comments).  These protocol specifications include:
.IP \(bu
The
.I "External Data Representation Protocol Specification" ,
which includes a complete specification of XDR data types, a
discussion of the XDR approach and a number of examples of
XDR usage.  This specification is published as DARPA RFC 1014.
.IP \(bu
The
.I "Remote Procedure Call Protocol Specification" ,
which includes a discussion of the RPC model, a detailed treatment
of the RPC authentication facilities and a complete specification
of the 
.I portmapper 
Protocol.  This specification is published as DARPA RFC 1057.
.IP \(bu
The
.I "Network File System: Version 2 Protocol Specification" ,
which includes a complete specification of the Mount Protocol, as
well as the NFS specification itself.  This specification is published as
DARPA RFC 1094.
.LP
PART THREE documents Transport-Level Network Programming.  
.IP \(bu
The first chapter, 
.I "Transport Level Interface (TLI) Programming" ,
describes the TLI system interface for direct access to network 
mechanisms.
.LP
The rest of the chapters in this part document the 
Berkeley style, socket-Based Inter-Process Communications mechanisms.  
.IP \(bu
.I "A Socket-Based Interprocess Communications Tutorial" 
then introduces socket-based IPC.  It assumes little more that basic 
networking concepts on the part of its reader, and includes many 
examples.
.IP \(bu
.I "An Advanced Socket-Based Interprocess Communications Tutorial" ,
which takes up where the
.I Tutorial
leaves off.
.IP \(bu
.I "Berkeley-Style IPC Implementation Notes" ,
which describes the low-level networking primitives (e.g.
.I accept() ,
.I bind() 
and
.I select() )
which originated with the 4.2BSD UNIX system.
This document is of interest primarily to system programmers and aspiring
UNIX gurus.
.
.NH 2
\&The Network File System (NFS)
..IX "Network File System"
..IX NFS
.LP
The Network File System is a facility for sharing files in a
heterogeneous environment of machines, operating systems, and networks.
Sharing is accomplished by mounting a remote filesystem, then reading
or writing files in place.
.LP
NFS was
.I not
designed by extending SunOS onto the network \(em such an approach was
considered unacceptable because it would mean that every computer on the
network would have to run SunOS.
Instead, operating-system independence was taken as an NFS design goal,
along with machine independence, crash recovery, transparent access and
high performance.  NFS was thus designed as a collection of network services,
and not as a distributed operating system.  As such, it is able to support
distributed applications without restricting the network to a single
operating system.
.LP
Sun's implementation of NFS is integrated with the SunOS kernel
for reasons of efficiency, although such close integration is not
strictly necessary.  Other vendors will make different choices, as
dictated by their operating environments and applications.  And
because of NFS's open design, all of these applications will be able to
work together on a single network.
.
.KS
.NH 3
\&Computing Environments
..IX network "computing environments"
.LP
The traditional timesharing environment looks like this:
.PS
Abox: box wid 1.5i ht 1.5i "Mainframe"
move from Abox right 1.25i
Bbox: box "terminal2"
move from Bbox up .5i
Cbox: box "terminal1"
move from Bbox down .50i
Dbox: box "terminal3"
move from Dbox down .50i
Gbox: box "terminal4"
move from Abox.sw right .25i
line down .50i
Eell: ellipse ht .25i wid .50i
line from Eell.e down .50i
line from Eell.w down .50i
arc cw from end of 2nd line to end of 3rd line
move from Abox.sw right 1.25i
line down .50i
Fell: ellipse ht .25i wid .50i
line from Fell.e down .50i
line from Fell.w down .50i
arc cw from end of 5th line to end of 6th line
line from Abox.e to Bbox.w
line from Abox.e to Cbox.w
line from Abox.e to Dbox.w
line from Abox.e to Gbox.w
.PE
.KE
.KS
The major problem with this environment
is competition for CPU cycles.
..IX "computing environments"
The workstation environment solves that problem,
but requires more disk drives.
A network environment looks like this:
.PS
Abox: box ht .25i wid 4.5i "Network"
move from Abox.sw right 1i
line down .5i
Cbox: box wid 1i "workstation1"
line from Cbox.s down .25i
Eell: ellipse ht .2i wid .4i
line from Eell.e down .4i
line from Eell.w down .4i
arc cw from end of 3rd line to end of 4th line
move from Abox.sw right 2.5i
line down .5i
Dbox: box wid 1i "server"
move from Dbox.sw right .1i
line down .25i
Fell: ellipse ht .25i wid .5i
line from Fell.e down .5i
line from Fell.w down .5i
Aarc: arc cw from end of 7th line to end of 8th line
move from Dbox.se left .1i
line down .25i
Mell: ellipse ht .25i wid .5i
line from Mell.e down .5i
line from Mell.w down .5i
arc cw from end of 10th line to end of 11th line
move from Abox.nw right .25i
line up .5i
Hbox: box wid 1i "workstation2"
move from Abox.nw right 1.75i
line up .5i
Ibox: box wid 1i "workstation3"
move from Abox.nw right 3.25i
line up .5i
Jbox: box wid 1i "workstation4"
move from Abox.sw right 4i
line down .5i
Lastbox: box wid 1i ht 1i "printer"
.PE
.KE
The goal of the NFS design was to make all disks available as needed.
Individual workstations have access to all information residing anywhere 
on the network.  Printers and supercomputers may also be available
somewhere on the network.
.
.NH 3
\&Example NFS usage
..IX NFS "example usage"
.LP
This section gives three examples of NFS usage.
.NH 4
\&Example 1: Mounting a Remote Filesystem
..IX "mounting a remote filesystem"
..IX "remote mounting"
.LP
Suppose your machine name is
.I client ,
that you want to read some on-line manual pages, and that
these pages are not available on your server machine, named
.I server ,
but are available on another machine named
.I docserv .
Mount the directory containing the manuals as follows:
.DS
.ft CW
client# \fB/usr/etc/mount docserv:/usr/man /usr/man\fP
.DE
Note that you have to be superuser in order to do this.
..IX mount "" \fImount\fP
..IX etc/mount "" \fI/etc/mount\fP
Now you can use the
.I man 
command whenever you want.
Try running the
.I "mount\ -p"
command (on
.I client )
after you've mounted the remote filesystem.
..IX mount "" "\fImount\fP"
Its output will look something like this:
.DS
.ft CW
server:/roots/client       /                   nfs rw,hard     0 0
server:/usr                /usr                nfs ro          0 0
server:/home/server        /home/server        nfs rw,bg       0 0
server:/usr/local          /usr/local          nfs ro,soft,bg  0 0
docserv:/usr/man           /usr/man            nfs ro,soft,bg  0 0
.DE
You can remote mount not only filesystems, but also directory
hierarchies inside filesystems.  In this example,
.I /usr/man 
is not a filesystem mount point \(em it's just a subdirectory
within the
.I /usr 
filesystem.  Here's a diagram showing a few key directories of
the three machines involved in this example.
Ellipses represent machines, and NFS-mounted filesystems are
shown boxed.  There are five such boxed directories, corresponding
to the five lines shown in the
.I "mount\ -p"
output above.  The
.I docserv:/usr/man 
directory is shown mounted as the
.I /usr/man 
directory on
.I client ,
as it would be by the
.I mount 
command given above.
.KS
.LP
.ce
Figure: An Example NFS Filesystem Hierarchy
.sp 5.0i
..IX NFS "filesystem example"
.\".# copy postscript file: mountexample.ps
.\".CP mountexample.ps 5.0i "" ""
.KE
.
.NH 4
\&Example 2: Exporting a Filesystem
..IX "export a filesystem"
..IX filesystem exportation
.LP
Suppose that you and a colleague need to work together on a
programming project.  The source code is on your machine, in
the directory
.I /usr/proj .
It doesn't matter whether your workstation is a diskless node or has a 
local disk.  Suppose that after creating the proper directory your 
colleague tried to remote mount your directory.  Unless you have 
explicitly exported the directory, your colleague's remote mount will 
fail with a ``permission denied'' message.
.LP
To export a directory, first become superuser and then edit the
.I /etc/exports 
file.  If your colleague is on a machine named
.I cohort ,
then you need to run
.I exportfs(8) 
(after putting this line in
.I /etc/exports ):
.DS
.ft CW
/usr/proj		-access=cohort
.DE
If no explicit access is given for a directory, then the system allows 
anyone on the network to remote mount your directory.  By giving explicit 
access to
.I cohort ,
you have denied access to others.  (For more details about the
.I /etc/exports ,
see the
.I exports(5) 
man page).  
.I mountd ,
the NFS mount request server,
(see
.I "The NFS Interface" ,
below) reads the file
.I /etc/xtab 
whenever it receives a request for a remote mount.
The file
.I /etc/xtab 
contains the entries for directories that are currently exported.
..IX exports "" \fIexports\fP
..IX etc/exports "" \fI/etc/exports\fP
Now your cohort can remote mount the source directory by issuing this
command:
.DS
.ft CW
cohort# \fB/etc/mount client:/usr/proj /usr/proj\fP
.DE
This, however, isn't the end of the story, since NFS requests are
also checked at request time.  If you do nothing, the accesses that
you've established in your
.I /etc/exports 
file will stay in effect, but you (and your programs) are free to
change them at any time with the
.I exportfs 
command and system call.
.LP
Since both you and your colleague will be able to edit files on
.I /usr/proj ,
it would be best to use the
.I sccs 
source code control system for concurrency control.
.
.NH 4
\&Example 3: Administering a Server Machine
..IX administering servers
..IX servers administration
.LP
System administrators must know how to set up
the NFS server machine so that client workstations
can mount all the necessary filesystems.
You export filesystems (that is, make them available)
by placing appropriate lines in the
.I /etc/exports 
file.  Here is a sample
.I /etc/exports 
file for a typical server machine:
.DS
.ft CW
/                -access=systems
/exec            -access=engineering:joebob:shilling
/usr             -access=engineering
/home/server     -access=engineering
/home/local.sun2 -access=engineering:athena
/home/local.sun3 -access=engineering
.DE
Machine names or netgroups, such as
.I staff 
(see
.I netgroup(5) )
may be specified after the filesystem,
in which case remote mounts are limited to machines
that are a member of this netgroup.  For the complete syntax of
the
.I /etc/exports 
file, see
.I exports(5) .
At any time, the system administrator can see
which filesystems are remote mounted by executing the
.I showmount 
command.
..IX "administering a server"
..IX servers administration
.
.NH 3
\&NFS Architecture
..IX NFS architecture
.sp 2
.NH 4
\&Transparent Information Access
..IX NFS "transparent access"
.LP
Users are able to get directly to the files they want without knowing
the network address of the data.  To the user, all NFS-mounted
filesystems look just like private disks.  There's no apparent
difference between reading or writing a file on a local disk, and
reading or writing a file on a disk in the next building.  Information
on the network is truly distributed.
..IX "transparency of NFS"
.
.NH 4
\&Different Machines and Operating Systems
..IX NFS "different machines"
..IX NFS "different operating systems"
.LP
No single vendor can supply tools for all the work that needs to get
done, so appropriate services must be integrated on a network.  NFS
provides a flexible, operating system-independent platform for such
integration.
..IX "heterogeneity of machines"
.
.NH 4
\&Easily Extensible
..IX NFS extensibility
..IX "extensible design"
.LP
A distributed system must have an architecture that allows integration of
new software technologies without disturbing the extant software
environment.  Since the NFS network-services approach does not depend on
pushing the operating system onto the network, but instead offers an
extensible set of protocols for data exchange, it supports the flexible
integration of new software.
.
.NH 4
\&Ease of Network Administration
..IX NFS administration
.LP
..IX administering networks
..IX network administration
The administration of large networks can be complicated and time-consuming,
yet they should (ideally) be
.I "at least"
as easy to administer as a set of local filesystems on a timesharing
system.  The UNIX system has a convenient set of maintenance commands
developed over the years, and the Network Information Service,
a NFS-based network
database service, has allowed them to be adapted and extended for the
purpose of administering a network of machines.  The NIS also allows certain
aspects of network administration to be centralized onto a small number of
file servers, e.g. only server disks must be backed up in networks of
diskless clients.  An overview of the NIS facility is presented in 
.I "The Network Information Service Database Service"
section of this manual.
.LP
The NIS interface is implemented using RPC and XDR, so it is available to
non-UNIX operating systems and non-Sun machines.  NIS servers do not
interpret data, so it is easy for new databases to be added to the NIS
service without modifying the servers.
.
.NH 4
\&Reliability
..IX NFS reliability
.LP
NFS's reliability derives from the robustness of the 4.2BSD
filesystem, from the stateless NFS protocol\**,
..IX NFS "stateless protocol"
..IX NFS protocol
.FS
The NFS protocol is
.I stateless
because each transaction stands on its own.  The server doesn't have to
remember anything \(em about clients or files \(em between transactions.
.FE
and from the daemon-based methodology by which network services
like file and record locking are provided.  See
.I "The Network Lock Manager" 
for more details on locking.  In addition, the file server protocol is
designed so that client workstations can continue to operate even when the
server crashes and reboots.
.LP
The major advantage of a stateless server is robustness in the face of
client, server, or network failures.  Should a client fail, it is not
necessary for a server (or human administrator) to take any action to
continue normal operation.  Should a server or the network fail, it is only
necessary that clients continue to attempt to complete NFS operations until
the server or network gets fixed.  This robustness is especially important
in a complex network of heterogeneous systems, many of which are not under
the control of a professional operations staff, and which may be running
untested systems that are often rebooted without warning.
..IX reliability
.
.NH 4
\&High Performance
..IX NFS performance
.LP
The flexibility of NFS allows configuration for a variety of
cost and performance trade-offs.
For example, configuring servers with large, high-performance disks,
and clients with no disks,
may yield better performance at lower cost than having
many machines with small, inexpensive disks.
Furthermore, it is possible to distribute the filesystem data
across many servers and get the added benefit of multiprocessing
without losing transparency.
In the case of read-only files,
copies can be kept on several servers to avoid bottlenecks.
.LP
Sun has also added several performance enhancements to NFS, such
as ``fast paths'' for key operations, asynchronous service of multiple
requests, disk-block caching , and asynchronous read-ahead and write-behind.
The fact that caching and read-ahead occur on both client and server
effectively increases the cache size and read-ahead distance.
Caching and read-ahead do not add state to the server;
nothing (except performance) is lost if cached information is thrown away.
In the case of write-behind, both the client and server
attempt to flush critical information to disk whenever necessary,
to reduce the impact of an unanticipated failure;
clients do not free write-behind blocks
until the server confirms that the data is written.
..IX performance
.
.NH 3
\&The Sun NFS Implementation
..IX NFS "Sun implementation"
.LP
In the Sun NFS implementation, there are three entities to be
considered: the operating system interface, the
.I "virtual file system (VFS) ",
interface, and the network file system (NFS) interface.  The UNIX
operating system interface has been preserved in the Sun implementation
of NFS, thereby insuring compatibility for existing applications.
Applications will use
.I read(2) 
and
.I write(2) 
to access NFS files just as the do to access local files.
.LP
The VFS is best seen as a layer that Sun has wrapped
around the traditional UNIX filesystem.  This traditional filesystem is
composed of directories and files, each of which has a corresponding
.I inode 
(index node), containing administrative information about the file,
such as location, size, ownership, permissions, and access times.
..IX inode "" \fIinode\fP
Inodes are assigned unique numbers within a filesystem, but a file on
one filesystem could have the same number as a file on another filesystem.
This is a problem in a network environment, because remote filesystems
need to be mounted dynamically, and numbering conflicts would cause
havoc.  To solve this problem, Sun designed the VFS, which is based
on a data structure called a
.I vnode .
In the VFS, files are guaranteed to have unique numerical designators,
even within a network.
..IX vnode "" \fIinode\fP
Vnodes cleanly separate filesystem operations from the semantics of their
implementation.  Above the VFS interface, the operating system deals in
.I vnodes ;
below this interface, the filesystem may or may not implement
.I inodes .
The VFS interface can connect the operating system
to a variety of filesystems (for example, 4.2 BSD or MS-DOS).
A local VFS connects to filesystem data on a local device.
..IX "virtual file system"
..IX VFS
.KS
.LP
The remote VFS defines and implements the NFS interface on the basis of
the RPC and XDR mechanisms.  The figure below shows the flow of a request
from a client (at the top left) to a collection of filesystems.
.PS
Abox:  box invis ht .2i "sys calls"
move from Abox down .50i
Bbox:  box ht .5i wid .75i "VFS" "interface"
move from Bbox down .75i
Cbox:  box ht .5i wid .75i "4.2" "VFS"
move from Cbox left .75i
Dbox:  box ht .5i wid 1i "other VFS" "(like PC-FS)"
move from Cbox right .75i
Ebox:  box ht .5i wid .75i "NFS" "client"
move from Cbox down .75i
Fell: ellipse ht .25i wid .50i
line from Fell.e down .50i
line from Fell.w down .50i
arc cw from end of first line to end of last line
move from Ebox down .75i
Gbox: box ht .5i wid .75i "RPC/" "XDR"
Hbox: box invisible wid .75i ht 1.5i with .sw at Gbox.se
Ibox: box ht .5i wid .75i "RPC/" "XDR" with .sw at Hbox.se
move from Ibox up 2.5i
Jbox: box same "NFS" "server"
move from Jbox down 1.75i right .75i
Kbox: box same "4.2BSD" "UFS"
move from Kbox up .75i
Lbox: box same "VFS"
move from Kbox down .75i
Mell: ellipse ht .25i wid .50i
line from Mell.e down .50i
line from Mell.w down .50i
arc cw from end of 3rd line to end of last line
move from Hbox down 1.25i
Nbox: box ht .25i wid 3.5i "Network"
arrow from Abox.s to Bbox.n
arrow from Bbox.s to Cbox.n
arrow from Bbox.w to Dbox.n
arrow from Bbox.e to Ebox.n
arrow from Cbox.s to Fell.n
arrow from Ebox.s to Gbox.n
arrow from Gbox.s down .50i
line <- from Ibox.s down .50i
arrow from Ibox.n to Jbox.s
arrow from Jbox.e to Lbox.w
arrow from Lbox.s to Kbox.n
arrow from Kbox.s to Mell.n
.PE
.KE
In the case of access through a local VFS,
requests are directed to filesystem data on devices
connected to the client machine.
In the case of access through a remote VFS,
the request is passed through the RPC and XDR layers onto the net.
In the current implementation, Sun uses the UDP/IP protocols
and the Ethernet.
On the server side, requests are passed through
the RPC and XDR layers to an NFS server;
the server uses
.I vnodes 
to access one of its local VFSs and service the request.
This path is retraced to return results.
.LP
Sun's implementation of NFS provides five types of transparency:
..IX "transparency of NFS"
..IX NFS transparencies
.IP  1.
.I "Filesystem Type:"
The
.I vnode ,
in conjunction with one or more local VFSs (and possibly remote VFSs)
permits an operating system (hence client and application)
to interface transparently to a variety of filesystem types.
.IP  2.
.I "Filesystem Location:"
Since there is no differentiation between a local and a remote VFS,
the location of filesystem data is transparent.
.IP  3.
.I "Operating System Type:"
The RPC mechanism allows interconnection of a variety
of operating systems on the network,
and makes the operating system type of a remote server transparent.
.IP  4.
.I "Machine Type:"
The XDR definition facility allows a variety of machines to
communicate on the network and makes the machine type of a remote
server transparent.
.IP  5.
.I "Network Type:"
RPC and XDR can be implemented for a variety of transport protocols,
thereby making the network type transparent.
.LP
Simpler NFS implementations are possible
at the expense of some advantages of the Sun version.
In particular, a client (or server) may be added to the network by
implementing one side of the NFS interface.
An advantage of the Sun implementation
is that the client and server sides can be symmetrical;
thus, it is possible for any machine to be client, server, or both.
Users at client machines with disks can arrange to share them over NFS
without having to appeal to a system administrator
or configure a different system on their workstation.
.
.NH 3
\&The NFS Interface
..IX NFS Interface
.LP
As mentioned in the preceding section, a major advantage of NFS is
the ability to mix filesystems.  In keeping with this, Sun encourages
other vendors to develop products to interface with Sun network services.
The specifications for RPC and XDR have been placed in the public domain,
and Sun's implementation of RPC and XDR is freely licensed, whic
serves as a
standard for anyone wishing to develop applications for the network.
Furthermore, the NFS interface itself is open and can be used by anyone
wishing protocol specifications to implement an NFS client or server for the network.
.NH 4
\&The NFS and the Mount Protocol
..IX mount NFS
..IX NFS "the mount protocol"
.LP
The NFS interface defines traditional filesystem operations for reading
directories, creating and destroying files, reading and writing files,
and reading and setting file attributes.  The interface is designed so
that file operations address files with an uninterpreted identifier
called a
..IX filehandle "" \fIfilehandle\fP
.I filehandle ,
a starting byte address, and a length in bytes.  NFS
.I never
deals with pathnames, only with filehandles.
..IX NFS pathnames
(More precisely, NFS never
.I interprets
pathnames.  Some NFS procedures take pathname arguments, but they are
just strings to NFS.)
.LP
Given a filehandle for a directory, a client program can use NFS
procedures to get other filehandles and thereby navigate throughout
the directories and files of a filesystem.  A client must, however,
get its first filehandle for a filesystem by using RPC to call the
mount server.  Mount will return a filehandle that grants access to
the filesystem.  The figure 
.I "Mount and NFS Servers"
shows the interaction between a client program, a mount server, and
an NFS server.  Note that the only interface between a mount server
and an NFS server is a common filehandle.
.KS
.LP
.ce
Figure: Mount and NFS Servers
.sp 4.0i
..IX NFS "mount servers"
..IX mount "NFS servers"
.\".# copy postscript file: mountserver.ps
.\".CP mountserver.ps 4.0i "" ""
.KE
.LP
.NH 4
\&Pathname Parsing
.LP
Although many operating systems have analogs to the hierarchical NFS
directory and file structure, the conventions used by operating systems
to formulate pathnames vary considerably.  To accommodate the many
possible path naming conventions, the mount procedure is not defined in
the NFS protocol but in a separate mount protocol.  Actually the mount
protocol is the same for any Operating System. It is only the implementation
that differs between systems.
.LP
The
.I mount 
procedure in the UNIX mount protocol converts a UNIX
pathname into a filehandle.  If local pathnames can be reasonably mapped
to UNIX pathnames, an NFS server developer may wish to implement the UNIX
mount protocol, even though the server runs on a different operating
system.  This approach makes the server immediately usable by clients
that use the UNIX
protocol and eliminates the need to develop a new
.I mount 
command
for UNIX-based clients.
.LP
Alternatively, a server developer can obtain
a new remote program number from Sun and define a new mount protocol.
For example, the
.I mount 
procedure in a VMS Mount protocol would take a VMS file specification
rather than a UNIX
pathname.  Mount protocols are not mutually exclusive; a server could,
for example, support the UNIX protocol for UNIX
clients and a Multics protocol for Multics clients.  Both protocols
would return filehandles defined by the NFS implementation on their server.
.LP
The mount protocols remove pathname parsing from the NFS protocol, so that
a single NFS protocol can work with multiple operating systems.  This means
that users and client programs need to know the details of a server's path
naming conventions only when mounting a filesystem.  Different server path
naming conventions therefore typically have little impact on users.
.LP
Because mounts are relatively infrequent operations, mount servers can be
implemented outside of operating system kernels without materially affecting
overall file system performance.  Because user-level code is easier to write
and far easier to debug than kernel code, mount servers are fairly simple
to put together.
.LP
.NH 4
\&Export and Mount Lists
.LP
Technically, a mount protocol needs to define only a
.I mount 
procedure that bootstraps the first filehandle for a
filesystem.  (By convention, a mount protocol should also
define a NULL procedure).  However, adding other procedures
can simplify network management.  As a convenience to
clients, a mount protocol might provide a procedure that
returns a list of filesystems exported by a server.
Another useful item is a mount list, a list of clients and
the pathnames they have mounted from the server.  The UNIX
mount protocol defines a mount list and a procedure called
.I readmount() 
that returns the list.  With the help of
.I readmount() ,
an administrator can notify the clients of a server that is
about to be shut down.
.LP
Note that a mount list makes a mount server stateful.
Recall, however, that the business of a mount server is to
translate pathnames into filehandles; the state represented
by a mount list does not affect a server's ability to
operate correctly.  Neither servers nor clients need take
any action to update or rebuild a mount list after a
crash.  Mount server users should regard the mount and
export lists provided by a mount server as ``accessories''
that are usually, but not necessarily, accurate.
.LP
.NH 4
\&UNIX Mount Protocol Procedures
..IX "mount protocol"
.LP
The
mount protocol consists of the six remote procedures listed in the table
.I "MOUNT: Remote Procedures, Version 1" .  
The
.I mount() 
procedure transforms a UNIX
pathname into a filehandle which the client can then pass to the
associated NFS server.  The pathname passed to the mount procedure
usually refers to a directory, often the root directory of a filesystem,
but it can name a file instead.  In addition to returning the filehandle,
.I mount 
adds the client's host name and the pathname to its mount list.  The
.I readmount() 
procedure returns the server's mount
list.
.I unmount() 
removes an entry from the server's mount list and
.I unmountall() 
removes all of a client's mount list entries.  The
.I readexport() 
procedure returns the server's export list.
.KS
.LP
.ce
Table: MOUNT: Remote Procedures, Version 1
.TS
box tab (/) ;
cI | cI | cI
c | lI | l .
Number/Name/Description
_
0/null/Do nothing
1/mount/Return filehandle for pathname
2/readmount/Return mount list
3/unmount/Remove mount list entry
4/unmountall/Clear mount list
5/readexport/Return export list
.TE
.KE
.NH 4
\&A Stateless Protocol
..IX NFS "stateless protocol"
.LP
The NFS interface is defined so that a server can be
.I stateless .
..IX "statelessness of NFS"
This means that a server does not have to remember from one transaction
to the next anything about its clients, transactions completed
or files operated on.  For example, there is no
.I open() 
operation, as this would imply state in the server;  of course, the UNIX
interface uses an
.I open() 
operation, but the information in the UNIX
operation is remembered by the client for use in later NFS operations.
.LP
An interesting problem occurs when a UNIX
application unlinks an open file.  This is done to achieve the effect
of a temporary file that is automatically removed when the application
terminates.  If the file in question is served by NFS, the call to
.I unlink() 
will remove the file, since the server does not remember that the file
is open.  Thus, subsequent operations on the file will fail.
In order to avoid state on the server, the client operating system
detects the situation, renames the file rather than unlinking it, and
unlinks the file when the application terminates.  In certain failure
cases, this leaves unwanted ``temporary'' files on the server;
these files are removed as a part of periodic filesystem maintenance.
.LP
Another example of the advantages gained by having the NFS interface
to the UNIX system without introducing state is the
.I mount 
command.  A UNIX client of NFS
``builds'' its view of the filesystem on its local devices using the
.I mount 
command or via
.I automount ;
thus, it is natural for the UNIX
client to initiate its contact with NFS and build its view
of the filesystem on the network with an extended
.I mount 
command.  This
.I mount 
command does not imply state in the server,
since it only acquires information for the client
to establish contact with a server.  The
.I mount 
command may be issued at any time, but is typically
executed as a part of client initialization.
The corresponding
.I umount 
command is only an informative message to the server,
but it does change state in the client
by modifying its view of the filesystem on the network.
.LP
The major advantage of a stateless server is robustness in the
face of client, server or network failures.  Should a client
fail, it is not necessary for a server (or human administrator)
to take any action to continue normal operation.  Should a
server or the network fail, it is only necessary that clients
continue to attempt to complete NFS operations until the
server or network is fixed.  This robustness is especially important in a
complex network of heterogeneous systems, many of which are not under the
control of a professional operations staff and may be running
untested systems and/or may be rebooted without warning.
.LP
An NFS server can be a client of another NFS server.
However, it is not often that a Sun server will not act as an intermediary
between a client and
another server.  Instead, a client may ask what remote mounts the server
has and then attempt to make similar remote mounts.  The decision to disallow
intermediary servers is based on several factors.  First, the existence
of an intermediary will impact the performance characteristics of the
system; the potential performance implications are so complex that it
seems best to require direct communication between a client and server.
Second, the existence of an intermediary complicates access control;
it is much simpler to require a client and server to establish direct
agreements for service.  Finally, disallowing intermediaries prevents
cycles in the service arrangements; Sun prefers this to detection or
avoidance schemes.
.LP
NFS currently implements UNIX file protection
by making use of the authentication mechanisms built into RPC.
..IX authentication
This retains transparency for clients and applications that make use of
UNIX file protection.
Although the RPC definition allows other authentication schemes,
their use may have adverse effects on transparency.
.LP
Note that NFS, although very UNIX-like, is
.I not
a UNIX filesystem per se \(em there are cases in which its behavior
differs from that which would be expected of the UNIX system proper:
.IP \(bu
The guaranteed APPEND_MODE is the most striking of these
differences, for it simply is not supported by NFS.
.LP
..IX NFS "stateful devices"
..IX NFS "special files"
Note: Network access to devices such as tape drivers is
a good idea, but it is best implemented as a separate network service
whose requirement for stateful operation is kept separate from network
access to files.
.LP
.IP \(bu
NFS does \fInot\fP support device operation over NFS.  Support of
special files is not stateful because the device operations are carried
out locally.
.IP \(bu
There are also minor incompatibilities between NFS and UNIX
file-system interfaces that are dictated by the very nature of
remote NFS mounts.  For example, a local NFS daemon simply can't
tell that a remote disk partition is full until the remote NFS
daemon tells it so.  Rather than wait for a positive confirm on
every write \(em a strategy that would impose unacceptable
performance problems \(em the local NFS code caches writes and
returns to its caller.  If a remote error occurs, it gets
reported back as soon as possible, but not as immediately as
would a local disk.
.LP
File locking and other inherently stateful functionality has been
omitted from the base NFS definition.  In this way, Sun has been able
to preserve a simple, general interface that can be implemented by a
wide variety of customers.  File locking has been provided as a
NFS-compatible network service, and Sun is considering doing the
same for other features that inherently imply state and/or
distributed synchronization.  These features, too, will be kept
separate from the base NFS definition.  In any case, the open nature
of the RPC and NFS interfaces means that customers and users who
need stateful or complex features can implement them ``beside'' NFS.
.
.NH 3
\&Note: Miscellaneous Network Operations
.LP
Sun supports a small number of miscellaneous networking operations that
are useful for temporary inter-host connections, isolated file
transfers, and access to non-UNIX systems (e.g. VMS machines on
the Internet).  These operations include
.I rcp ,
.I rlogin ,
.I rsh ,
.I ftp ,
.I telnet ,
and
.I tftp .
..IX rcp "" \fIrcp\fP
..IX rsh "" \fIrsh\fP
..IX rlogin "" \fIrlogin\fP
..IX tftp "" \fItftp\fP
..IX ftp "" \fIftp\fP
..IX telnet "" \fItelnet\fP
.IP \(bu
.I rcp 
is a remote copy utility program that uses ``BSD networking facilities''
to copy files from one machine to another.  The
.I rcp 
user supplies the path name of a file on a remote machine, and
receives a stream of bytes in return.  Access control is based on
the client's login name and host name.
.IP
The major problem with
.I rcp 
is that it's not transparent to the user, who winds up with a
redundant copy of the transferred file.  With NFS, by contrast,
only one copy of the file is necessary.  Another problem is that
.I rcp 
does nothing but copy files.  To use it as a model for additional
network services would be to introduce a remote command for every
regular command: for example,
.I rdiff 
to perform differential file comparisons across machines.  By
providing for the sharing of filesystems, NFS makes this unnecessary.
.LP
.I rcp 
is useful for NFS servers that you have login access to but not NFS access.
Files can copied back and forth, yet you don't need any filesystem mounted.
.IP \(bu
.I rlogin 
allows the user to log into a remote machine, directly accessing
both its processor and its mounted file systems.  It remains
useful in NFS-based networks because, with it, users can directly
execute commands on remote machines over the network.
.IP \(bu
.I rsh 
allows the user to execute a command on a remote machine.  If no
command is specified,
.I rsh 
is equivalent to
.I rlogin .
Unlike the REX-based
.I on 
command,
.I rsh 
does not copy the users local environment
to the remote machine before executing the command.
This can be a benefit in situations where exporting your local environment
might cause problems.
.IP \(bu
.I ftp 
is very much like
.I rcp ,
in that it supports file copying between machines.  However,
.I ftp 
is more general that
.I rcp ,
and is not restricted to copies between two UNIX systems.
.IP \(bu
.I telnet 
communicates with another host using the TELNET protocol.  It isn't
used much because
.I rlogin 
is the standard mechanism for local inter-host communication.
But like
.I ftp ,
.I telnet 
is useful for non-Unix systems.
.IP \(bu
.I tftp 
is like
.I ftp ,
expect that it is simpler and less reliable.  This is because
.I tftp 's
transfer protocol is very simple; it is less robust that
.I ftp 's
protocol, and offers fewer options.
.I tftp 
is also used as part of the diskless NFS booting procedure (i.e. netdisk).

.NH 2
\&Remote File Sharing (RFS)
.LP
Remote File Sharing (RFS) provides a means of viewing files that physically
reside on remote machines as if they were on the local machine.  Remote
files are named using the same conventions as for local files, and all
operations on remote files work the same as they do on local files.  Like
NFS, RFS allows application programs to transparently share files across
the network.
.LP
NFS, however, is stateless, transactions are independent of each other, and
thus no recovery is required when a server or client goes down.  RFS, in
contrast, supports all UNIX semantics as defined by AT&T.  Consequently,
it saves state across transactions, and must recover when a server or
client goes down.
.LP
RFS is used in much the same way as NFS.  For both, the user accesses
remote files by mounting directories which are made available across the
network by server processes running on remote machines.  The details do
vary, though.  Machines using RFS make selected directories available for
sharing by
.I advertising
them.  Correspondingly, machines are able to augment their own file trees
with the advertised files from other machines.  This augmentation is
performed by means of a
.I "remote mount" ,
which is a direct extension of the standard mount operation.  Once remote
directories have been mounted on the local filesystem, they are
functionally part of that filesystem and are accessed in the same way
as local directories.
.NH 3
\&Advertise
.LP
To allow other machines to access a directory, its owner must advertise 
it by using the
.I adv(8) 
command.  Once advertised, the directory and all files contained in its
subtree are available for sharing by any authorized machine.  
.NH 3
\&Unadvertise
.LP
A directory can be unadvertised at any time with the 
.I unadv(8) 
command.  Unadvertising a directory has no effect on existing mounts of
that directory, but future mount requests will fail.
.NH 3
\&Remote Mounts
.LP
RFS extends the 
.I mount(8) 
operation to include a remote mount.  After a machine has advertised a
resource, another machine may remotely mount that resource in its own
file tree.  For example, to advertise a directory named
.I /fs1 ,
the administrator of a server machine would type:
.DS
.ft CW
example%\fB adv DATA /fs1\fP
.DE
This makes the 
.I /fs1 
subtree available for sharing, and specifies that other machines will
use the name 
.I DATA 
to refer to it when they mount it.  The name
.I DATA 
can be almost any name that would work as a file name as long as it
does not contain a period (``.'').  See below for the special meaning 
of the period.
.LP
Another machine (a client) gains access to the advertised subtree by 
mounting the remote subtree on the local directory.  The remote 
.I /fs1 
is mounted on the local
.I /fs1 
with the command
.DS
.ft CW
example%\fB mount -d DATA /fs1\fP
.DE
The 
.I -d 
option tells the 
.I mount(8) 
command that the resource being mounted is remote.
.LP
There is no need for the structures of the client and server file trees 
to match in any way, or for advertised subtrees to be mounted at the 
same level on the client as they occupy on the server.  If the client 
had done the remote mount onto its 
.I /usr 
directory, then its references to files under 
.I /usr 
would yield files in the server subtree under
.I /fs1 .
A client cannot get to parts of the server file tree that are not
within an advertised directory.  
.NH 3
\&Resource Naming
.LP
Resource naming is modeled after the DARPA domain naming convention,
which has a hierarchically structured name space.  A domain in this
usage is a name space that may encompass a group of machines and a set
of resources advertised by that group of machines.  
.LP
Resource names are made up of two components separated by a period 
(``.'').  For example, 
.I "isl.payroll"
might represent a resource called 
.I payroll
in domain
.I isl ,
and 
.I "isl.acctp"
might represent the machine 
.I acctp
within the same domain.  Whether a name specifies a resource or a
machine is determined by context; there is no syntactic distinction.
If a name is unqualified (i.e., if it contains no periods), the
associated domain may (in some cases) be inferred from the context.
.LP
A domain's name space is maintained by a 
.I "domain name server" ,
which insures uniqueness of names within the domain and provides a
central location for storing information about the machines and 
advertised resources in the domain.  The
.I adv(8) ,
.I unadv(8) ,
.I mount(8) ,
.I umount(8) ,
and 
.I nsquery(8) 
commands use the domain name server as a data base for information
about advertised resources, such as their names and the servers that
own them.  
.LP
As described above, each resource is assigned a symbolic name when it
is advertised, and the resource is subsequently identified (e.g. with a 
.I mount(8) 
command issued on a client) using just the domain name and that
symbolic name.  Because of this symbolic naming of resources, remote
users of resources need not know the actual position of the resources 
within the server's file tree, nor even what server within the domain
is offering the resource.  This location independence simplifies references
to resources, and allows for the transparent migration of resources
among the machines within a domain (for example, for balancing the load
among a set of server machines).
.NH 3
\&RFS Security Features
.LP
RFS contains three security features \(em client authentication, client
authorization, and user and group id mapping.  
.NH 4
\&Client Authentication
.LP
This feature associates a password with a client machine so that the
identity of a prospective client can be checked before a mount request
is serviced.  Entry and update of passwords is discussed in the 
.I rfadmin(8) ,
.I rfstart(8) ,
and
.I rfpasswd(8) 
commands.  
.NH 4
\&Client Authorization
.LP
RFS provides a means of selectively advertising directories through the
.I adv(8) 
command.  For example, if you want to advertise
.I /usr/private ,
but only want to authorize machines 
.I mach1
and
.I mach2
to mount it, you would issue the command:
.DS
.ft CW
example% \fBadv PRIVATE /usr/private mach1 mach2\fP
.DE
Without such a list of machines, the 
.I adv(8) 
command puts no restrictions on availability.
.LP
One may also choose to advertise a directory read-only by using the
.I -r 
option.  Here, a remote mount will only succeed if the mount command
also includes the
.I -r 
option.
.NH 4
\&User and Group Id Mapping
.LP
Whenever a user accesses a remote file, that user's permissions must
be checked as part of the normal processing of the request (for
example, an ``open to write'' is only valid if the user making the
request has write permissions on the file).  When accessing a file
across two machines, there is no guarantee that the user and group ids
on the local machine have the same meaning on the other machine.
.LP
Some machines handle this problem by requiring the same numeric ids
across machines and expecting the administrators to make sure that the
.I /etc/passwd 
and
.I /etc/group 
files are identical across all machines (at least the entries for all
users that access remote files).  This approach is conceptually simple,
but it is not always feasible in practice, especially in large or
already established environments.
.LP
RFS, therefore, provides a range of id mapping options through the 
.I idload(8) 
command.  Id mapping is done by a server machine on all incoming
requests, as well as in reporting file ownership ids in response to a
request from a client machine (e.g. a
.I stat(2) 
or
.I fstat(2) ).
A client machine maps ids in order to determine the effective user or
group id to use in executing a program that is stored on a server and
is ``set user id'' or ``set group id''.
.LP
On each machine, mapping can be set globally, for all remote machines,
or on a per-machine basis.  All mapping is based on one of two default
cases:
.IP Id
This case maps all incoming ids to 
.I id ,
which means that remote users will have the permissions associated
with 
.I id 
in accessing a server's files.  This mapping is the default if no
other mapping is specified.
.IP Transparent
This is a null mapping; remote user and group ids are used locally
without change.
.LP
These base mappings are augmented by two additional capabilities:
.IP Exclude
This capability excludes selected ids from the default mapping by
mapping them to an otherwise unused id.  This capability can be used
together with the transparent mapping capability to handle a network
where the 
.I /etc/passwd 
and
.I /etc/group 
files were identical, but certain permissions (e.g. root) are to be
disallowed from remote machines.
.IP Map
This capability provides arbitrary mapping between remote and local ids
that have different name or different numeric values.  It can be used
with the transparent mapping to handle exceptions to ``nearly''
identical
.I /etc/passwd 
files.
.\".H 2 "\fBportmapper\fP"
.\"..IX portmapper "" "\fIportmapper\fP" "" PAGE MAJOR
.\".LP
.\"Client programs need a way to find server programs; that is, they need a
.\"way to look up the network addresses of server programs.  Network transport
.\"services do not themselves provide such a service; they merely provide
.\"process-to-process message transfer across a network.  A message is sent to
.\"a transport-specific network address.  A network address is a logical
.\"communications channel \(em by waiting on a network address, a process
.\"receives messages from the network).
.\".LP
.\"RPC, being transport independent, makes
.\".I no
.\"assumptions about the structure of a network address.  It deals with
.\".I "universal addresses" ,
.\"specified only as null-terminated strings of characters.  Such universal
.\"addresses are translated into local transport addresses by translation
.\"routines specific to each transport provider.  (See
.\".XR  \fI@TitleOf(nametoadd)\fP\,
.\"below).
.\".LP
.\"How a process waits on a network address varies from one operating system to 
.\"the next, but all provide mechanisms that by which a process can synchronize 
.\"its activity with arriving messages. Thus, messages are not sent across 
.\"networks to receiving processes, but rather to the network address at which 
.\"receiving processes pick them up.  Network address are valuable because they 
.\"allow message receivers to be specified in a way that is independent of the 
.\"conventions of the receiving operating system.  The
.\".L rpcbind
.\"protocol defines a network service that provides a standard way for clients
.\"to look up the network address of any remote program supported by a server.
.\"Because it can be implemented on any transport, it provides a single 
.\"solution to a general problem that works for all clients, all servers and 
.\"all networks.
.\".H 3 "Port Registration"
.\"..IX rpcbind "page registration" "\fIrpcbind\fP"
.\".LP
.\"Since the
.\".L rpcbind 
.\"daemon is responsible for mapping network services to network addresses, 
.\"its address must be well known.  In the Internet domain, this problem is 
.\"solved by always assigning 
.\".L rpcbind
.\"the port number 111.  Unfortunately, this simple solution is not acceptable
.\"on all transports.  Thus 
.\".L rpcbind
.\"begins each session by registering its location on each of the transports 
.\"supported by the host upon which it finds itself running.
.\".LP
.\".L rpcbind
.\"is the only network service that must have such a well-known network address,
.\"since it is responsible for registering the addresses of other network 
.\"services and making them available to network clients, either by way of 
.\".L rpcinfo(1M) 
.\"or the library routines specified in 
.\".L rpcbind (3N).
.\"Thus, network services can be assigned network addresses statically or 
.\"dynamically \(em as long as they register them with their host's 
.\".L rpcbind 
.\"daemon.  Thus, RPC-based servers typically get mapped to network address at 
.\"run time, when they register with 
.\".L rpcbind ,
.\"and neither they nor their clients make any assumptions about what
.\"those network addresses will be.
.\".LP
.\"Delegating address-to-program mapping to 
.\".L rpcbind
.\"also automates network address administration.  Statically mapping addresses
.\"and remote programs in a file duplicated on each client would require
.\"updating all mapping files whenever a new remote program was introduced to
.\"a network.  Such updating quickly becomes impossible as networks grow.  And
.\"the alternative, placing the address-to-program mappings in a shared NFS
.\"file, is too centralized \(em if the fileserver went down the whole network
.\"would go down with it.
.\".LP
.\".L rpcbind 
.\"is started whenever a machine is booted.  As shown in 
.\"the
.\".XR \fI@TitleOf(tpsequence)\fR
.\"figure, below, both server programs and client programs call 
.\".L rpcbind
.\"procedures.\**
.\".FS
.\"..IX "clients and servers"
.\"..IX servers "and clients"
.\"Although client and server programs and client and server machines are
.\"usually distinct, they need not be.  A server program can also be a client
.\"program, as when an NFS server calls a
.\".L rpcbind
.\"server.  Likewise, when a client program directs a ``remote'' procedure
.\"call to its own machine, the machine acts as both client and server.
.\".FE
.\"As part of its initialization, a server program calls its host's 
.\".L rpcbind 
.\"daemon to register itself in the host's network-address map.  Whereas 
.\"server programs call 
.\".L rpcbind
.\"to update network address maps, clients call them to query those maps.  
.\"To find a remote program's network address, a client sends an RPC call 
.\"message to a server's 
.\".L rpcbind
.\"daemon; if the remote program is supported on the server, the daemon 
.\"returns the relevant network address in an RPC reply message.  The 
.\"client program can then send RPC call messages to that address.  A 
.\"client program can minimize its calls to 
.\".L rpcbind 
.\"by caching the network addresses of recently called remote programs.
.\".NS NOTE
.\"Note that the service provided by
.\".L rpcbind
.\"is
.\".I inherently
.\"stateful because a network-address map is a set of associations between 
.\"registrant names and network address, that is, a description of a state.
.\".NE
.\".KS
.\".FN "Typical \fIrpcbind\fP Sequence" tpsequence
.\"..IX rpcbind "typical mapping sequence" \fIrpcbind\fP
.\".CP pmapseq.ps 5.0i "" ""
.\".KE
.\"The 
.\".L rpcbind
.\"protocol (for details, see the
.\".XR \fI@TitleOf(pmapper)\fP
.\"section of the
.\".XR \fI@TitleOf(rpc_spec)\fP
.\"chapter) provides a procedure,
.\".L callit() ,
.\"by which 
.\".L rpcbind
.\"can assist a client in making a remote procedure call.  A client program 
.\"passes the target procedure's program number, version number and procedure 
.\"number (for a discussion of these numbers, see the
.\".XR \fI@TitleOf(rpcpg)\fP
.\"chapter) and arguments in an RPC call message.
.\".L callit()
.\"then looks up the target procedure's port number in the portmap and sends an 
.\"RPC call message to the target procedure including in it the arguments received
.\"from the client.  When the target procedure returns results, 
.\".L callit()
.\"passes them on to the client program.  It also returns the target procedure's 
.\"network address so the client can subsequently call it directly.
.\".LP
.\"Note that, because every instance of a remote program can be mapped to a 
.\"different port on each server on which it resides, a client has no way to 
.\"broadcast a remote procedure call directly.  However,
.\".L rpcbind 's
.\".L callit()
.\"procedure can be used to broadcast a remote procedure call indirectly, since 
.\".L rpcbind
.\"always registers itself first so that its address is well known.  One way for 
.\"a client to find a server running a remote program is to broadcast a call to
.\".L callit() ,
.\"asking it to call procedure 0 of the desired remote program.  If this call
.\"is broadcast to all servers, the first reply received is likely to be from
.\"the server with the lightest workload.
.\".LP
.\"The Sun RPC library provides an interface to all 
.\".L rpcbind 
.\"procedures.  Some of the RPC library procedures also call 
.\".L rpcbind
.\"automatically on behalf of client and server programs.
.\".H 3 "Name To Address Mapping" nametoadd
.\"..IX "Name to Address Mapping"
.\".LP
.\".ft B
.\"This seems like a good place for this stuff.  Whatever Chuck writes should
.\"probably go here.  It needs to discuss universal addresses and the generic
.\"name-to-address translation routines, at the minimum.
.\".ft P
.NH 2
\&The Portmapper
..IX portmapper "" "" "" PAGE MAJOR
.LP
Client programs need a way to find server programs; that is,
they need a way to look up and find the
.I port
numbers of server
programs.\**
.FS
The naming of services by way of the port-number segment of their IP
address is mandated by the Internet protocols.  Given this, clients face
the problem of determining which ports are associated with the services
they wish to use.
.FE
Network transport services do not provide such a service; they merely
provide process-to-process message transfer across a network.  A message
typically contains a transport address which contains a network number, a
host number, and a port number. (A port is a logical communications channel
in a host \(em by waiting on a port, a process receives messages from the
network).
.LP
How a process waits on a port varies from one operating system to
the next, but all provide mechanisms that suspend a process until a
message arrives at a port.  Thus, messages are not sent across
networks to receiving processes, but rather to the ports at which
receiving processes wait for messages.
.\"Port are valuable because
.\"they allow message receivers to be specified in a way that is
.\"independent of the conventions of the receiving operating system.
The portmapper protocol defines a network service that provides a
standard way for clients to look up the port number of any remote
program supported by a server.
.NH 3
\&Port Registration
..IX portmapper "page registration"
.LP
The portmapper on every host is associated with port number 111.
The portmapper is one of the few network services that must have such a
well-known and dedicated port.  Other network services can be
assigned port numbers statically or dynamically so long as they
register their ports with their host's portmapper.  For example, a
server program based on Sun's RPC library typically gets a port
number at run time by calling an RPC library procedure.  Note that
a given network service can be associated with port number 1256 on
one server and with port number 885 on another; on a given host,
a service can be associated with a different port every time its
server program is started.  Delegating port-to-remote program
mapping to portmappers also automates port number administration.
.\"Statically mapping ports and remote programs in a file duplicated
.\"on each client would require updating all mapping files whenever a
.\"new remote program was introduced to a network. (The alternative
.\"of placing the port-to-program mappings in a shared NFS file would
.\"be too centralized, and if the fileserver went down the whole
.\"network would go down with it).
.LP
The portmapper is started
automatically whenever a machine is booted.  As shown in the figure
.I "Typical Portmapping Sequence" 
below, both server programs and client programs call portmapper
procedures.\**
.FS
..IX "clients and servers"
..IX servers "and clients"
Although client and server programs and client and server machines are
usually distinct, they need not be.  A server program can also be a
client program, as when an NFS server calls a portmapper server.
Likewise, when a client program directs a ``remote'' procedure call to
its own machine, the machine acts as both client and server.
.FE
To find a remote program's port, a client sends
an RPC call message to a server's portmapper; if the remote program is
registered with the portmapper, it returns the relevant port number
in an RPC reply message.  The client program can then send RPC call
messages to the remote program's port.
.SH
NOTE
.LP
The portmapper provides an
.I inherently
stateful service because a portmap is a set of associations between
registrants and ports.
Hence, all the RPC services need to be reregistered if the portmap is restarted.
.LP
.KS
.LP
.ce
Figure: Typical Portmapping Sequence
.sp 5.0i
..IX portmapper "typical mapping sequence"
.\".# copy postscript file: pmapseq.ps
.\".CP pmapseq.ps 5.0i "" ""
.KE
.\"The portmapper protocol (for details, see the
.\".XR \fI@TitleOf(pmapper)\fP
.\"section of the
.\".XR \fI@TitleOf(rpc_spec)\fP
.\"chapter) provides a procedure,
.\".L PMAPPROC_CALLIT ,
.\"by which the portmapper can assist a client in making a remote procedure
.\"call.  A client program passes the target procedure's program number,
.\"version number, procedure number (for a discussion of these numbers, see
.\"the
.\".XR \fI@TitleOf(rpcpg)\fP
.\"chapter) and arguments in an RPC call message.
.\"It looks up the target procedure's port number in the portmap and sends an
.\"RPC call message to the target procedure including in it the arguments
.\"received from the client.  When the target procedure returns results to
.\"portmapper, it
.\"returns the results to the client program; also returned is the target
.\"procedure's port number so the client can subsequently call the target
.\"procedure directly
.\".LP
Note that, because every instance of a remote program can be mapped to
a different port on every server, a client has no way to broadcast a remote
procedure call directly.  However, the portmapper
.I PMAPPROC_CALLIT 
procedure can be used to broadcast a remote procedure call indirectly,
since all portmappers are associated with port number 111.  One way for a
client to find a server running a remote program is to broadcast a call
to
.I PMAPPROC_CALLIT ,
asking it to call procedure 0 of the desired remote program.
.\"If this call
.\"is broadcast to all servers, the first reply received may be from
.\"the server with the lightest workload.
.LP
The Sun RPC library provides an interface to all portmapper procedures.
Some of the RPC library procedures also call portmappers automatically on
behalf of client and server programs.
.NH 2
\&The Network Information Service Database Service
..IX "Network Information Service" "" "" "" PAGE MAJOR
.LP
This chapter explains Sun's network database mechanism,
the Network Information Service.
NIS was previously known as "Yellow Pages", which is now a trademark
of British Telecom (refer to the trademark page at the front of this manual).
Although it is not intended exclusively
for system administrators, it leans towards their
concerns.  The Network Information Service permits password information and host
addresses for an entire network to be held in a single database,
and, by so doing, greatly ease system and network administration.
.
.NH 3
\&What Is The Network Information Service?
.LP
The Network Information Service constitutes a distributed network lookup service:
.IP \(bu
NIS is a lookup service: it maintains a set of databases for querying.
Programs can ask for the value associated with a particular key,
or all the keys, in a database.
.IP \(bu
NIS is a network service: programs need not know the location of data,
or how it is stored.  Instead, they use a network protocol
to communicate with a database server that knows those details.
.IP \(bu
Network Information Service is distributed:
databases are fully replicated on several machines,
known as NIS servers.  Servers propagate updated databases among themselves,
ensuring consistency.  At steady state, it doesn't matter which server
answers a request; the answer is the same everywhere.
..IX "Network Information Service" explained
.NH 3
\&Network Information Service Maps
..IX "Network Information Service" maps
.LP
The Network Information Service serves information stored in NIS
.I maps .
Each map contains a set of keys and associated values.
For example, the
.I hosts 
map contains (as keys) all host names on a network,
and (as values) the corresponding Internet addresses.
Each NIS map has a
.I mapname ,
used by programs to access data in the map.
Programs must know the format of the data in the map.
Most maps are derived from ASCII files formerly found in
.I /etc/passwd ,
.I /etc/group ,
.I /etc/hosts ,
.I /etc/networks ,
and other files in
.I /etc .
The format of data in the NIS map is in most cases identical
to the format of the ASCII file.  Maps are implemented by
.I dbm(3X) 
..IX dbm() "" \fIdbm()\fP
files located in subdirectories of
.I /etc/yp 
on NIS server machines.
.LP
The relationship between a NIS map and the standard UNIX
.I /etc 
file which it relates to varies from map to map.  Some files (e.g.
.I /etc/hosts ,
are
.I replaced
by their corresponding NIS maps, while some (e.g.
.I /etc/passwd 
are merely
.I augmented .
.LP
Maps sometimes have nicknames.  Although the
.I ypcat 
command is a general NIS database print program,
it knows about the standard files in the NIS.  Thus
.I "ypcat hosts"
is translated into
.I "ypcat hosts.byaddr"
since there is no file called
.I hosts 
in the NIS.  The command
.I "ypcat\ -x"
furnishes a list of expanded nicknames.
.
.NH 3
\&Network Information Service Domains
..IX "Network Information Service" domain
.LP
A NIS
.I domain
is a named set of NIS maps.  Taken together, these maps define a
distinct network namespace and locate a distinct area of
administrative control.
NIS domains differ from both Internet domains and
.I sendmail 
domains, which define similar kinds of administrative loci in
their respective (IP and electronic mail) networks.  A given
host will typically fall within all three domains, but these
domains will not typically coincide.  A NIS domain is implemented
as a directory in
.I /etc/yp 
containing a set of maps.
.LP
You can determine your NIS domain by executing the
.I domainname 
..IX domainname "" "\fIdomainname\fP command"
command.  A domain name is required for retrieving data from a NIS
database.  For instance, if your NIS domain is
.I sys1 
and you want to find the Internet address of host
.I dbserver ,
you must ask NIS for the value associated with the key
.I dbserver 
in the map
.I hosts.byname 
within the NIS domain
.I sys1 .
Each machine on the network belongs to a default domain, which is
set at boot time.  Diskfull machines have their default domains
set by a call to the
.I domainname 
command made from
.I /etc/rc.local .
Diskless clients have it set as the result of a consultation with
the
.I bootparams(5) 
server.
.LP
A NIS server holds all the maps of a NIS domain in a subdirectory of
.I /etc/yp ,
named after the domain.
In the example above, maps for the
.I sys1 
domain would be held in
.I /etc/yp/sys1 .
A given host can contain maps for more than one NIS domain.
.
.\".H 3 "Servers and Clients"
.\".LP
.\"Servers provide resources, while clients consume them.
.\"..IX servers "and clients"
.\"..IX "clients and servers"
.\"The terms ``server'' and ``client'' do not necessarily
.\"indicate machines.  Consider both NFS (network file
.\"system), and the NIS:
.\".IP NFS 5
.\"NFS allows client machines to mount remote filesystems and access files
.\"in place, provided a server machine has exported the filesystem.  However,
.\"a server that exports filesystems may also mount remote filesystems
.\"exported by other machines, thus becoming a client.  So a given machine may
.\"be both server and client, or client only, or server only.
.\".IP NIS 5
.\"The NIS server, by contrast, is a process rather than a machine, A process
.\"can request information out of the NIS database, obviating the need to have
.\"such information on every machine.  All processes that make use of NIS
.\"services are NIS clients.  Sometimes clients are served by NIS servers on the
.\"same machine, but other times by NIS servers running on another machine.  If
.\"a remote machine running a NIS server process crashes, client processes can
.\"obtain NIS services from another machine.  Thus, the network NIS service will
.\"remain available even if an individual NIS host machine goes down.
.\".
.NH 3
\&Masters and Slaves
.LP
..IX "master and slave"
..IX "slave and master"
NIS servers containing copies of the same databases can be spread
throughout a network.  When an arbitrary machine wants information in
one of the NIS databases, it makes an RPC call to one of the NIS servers
to get it.  For any NIS map, one NIS server is designated as the
.I master
\(em the only one whose database may be modified.  The other NIS servers
are
.I slaves ,
and they are automatically updated from time to time to keep their
information in sync with that of the master.
.LP
All changes to a NIS map must be made on the machine which is the master
NIS server for that map.  The changes will then propagate to the slaves.  A
newly built map is timestamped internally when it's created by
.I makedbm .
If you build a NIS map on a slave server, you will temporarily break the
NIS update algorithm, and will have to get all versions in synch manually.
Moral: after you decide which server is the master, do all database
updates and builds there, not on slaves.
.LP
A given server may even be master with regard to one map, and slave with
regard to another.  This can get confusing quickly.  Thus, its
recommended that a single server be master for all maps created by
.I ypinit 
in a single domain.  Here we are assuming this simple case, in which one
server is the master for all maps in a database.
.NH 4
\&Naming
..IX "Network Information Service" naming
.LP
Imagine a company with two different networks,
each of which has its own separate list of hosts and passwords.
Within each network, user names, numerical user IDs,
and host names are unique.
However, there is duplication between the two networks.
If these two networks are ever connected, chaos could result.
The host name, returned by the
.I hostname 
..IX hostname "" \fIhostname\fP
command and the
.I gethostname() 
..IX gethostname() "" \fIgethostname()\fP
system call, may no longer uniquely identify a machine.
Thus a new command and system call,
.I domainname 
..IX domainname "" "\fIdomainname\fP command"
and
.I getdomainname() 
..IX getdomainname() "" \fIgetdomainname()\fP
have been added.  In the example above,
each of the two networks could be given a different domain name.
However, it is always simpler to use a single domain whenever possible.
.LP
The relevance of domains to NIS is that data is stored in
.I /etc/yp/ \fIdomainname\fP.
In particular, a machine can contain data for several different domains.
.
.NH 4
\&Data Storage
..IX "Network Information Service" "data storage"
.LP
The data in NIS maps is stored as
.I dbm 
..IX dbm() "" \fIdbm()\fP
format databases.
(See
.I dbm(3X) ).
Thus the database
.I hosts.byname 
for the domain
.I sys1 
is stored as
.I /etc/yp/sys1/hosts.byname.pag 
and
.I /etc/yp/sys1/hosts.byname.dir .
The command
.I makedbm 
..IX makedbm "" \fImakedbm\fP
takes an ASCII file such as
.I /etc/hosts 
and converts it into a
.I dbm 
file suitable for use by the NIS.
However, system administrators normally use the makefile in
.I /etc/yp 
to create new
.I dbm 
files (read on for details).
This makefile in turn calls
.I makedbm .
.
.NH 4
\&Servers
..IX "Network Information Service" servers
.LP
To become a server, a machine must contain the NIS databases,
and must also be running the NIS daemon
.I ypserv .
The
.I ypinit 
..IX ypinit "" "\fIypinit\fP command"
command invokes this daemon automatically.
It also takes a flag saying whether you are creating a master or a slave.
When updating the master copy of a database,
you can force the change to be propagated to all the slaves with the
.I yppush 
..IX yppush "" "\fIyppush\fP command"
command.  This pushes the information out to all the slaves.
Conversely, from a slave, the
.I ypxfr 
..IX ypxfr "" "\fIypxfr\fP command"
command gets the latest information from the master.  The makefile in
.I /etc/yp 
first executes
.I makedbm 
to make a new database, and then calls
.I yppush 
to propagate the change throughout the network.
.
.NH 4
\&Clients
..IX "Network Information Service" clients
.LP
Remember that a client machine
does not access local copies of
.I /etc 
files, but rather makes an RPC call to a NIS server
each time it needs information from a NIS database.
NIS clients on NIS servers also don't access local copies of
.I /etc 
files.  The
.I ypbind 
..IX ypbind "" "\fIypbind\fP command"
daemon remembers the name of a server.  When a client boots,
.I ypbind 
broadcasts asking for the name of the NIS server.
Similarly,
.I ypbind 
broadcasts asking for the name of a new NIS server
if the old server crashes.  The
.I ypwhich 
..IX ypwhich "" "\fIypwhich\fP command"
command gives the name of the server that
.I ypbind 
currently points at.
.LP
Since client machines don't have entire copies of files in the NIS,
the commands
.I ypcat 
and
.I ypmatch 
..IX ypcat "" "\fIypcat\fP command"
have been provided.  As you might guess,
.I "ypcat passwd"
is equivalent to
.I "cat /etc/passwd"
To look for someone's password entry,
searching through the password file no longer suffices;
you have to issue one of the following commands

.DS
.ft CW
example% \fBypcat passwd | grep username\fP
example% \fBypmatch username passwd\fP
.DE

where you replace
.I username 
with the login name you're searching for.
.
.NH 3
\&Default NIS Files
..IX "Network Information Service" "default files"
.LP
By default, Sun workstations have a number of files from
.I /etc 
in their NIS:
.I /etc/passwd ,
.I /etc/group ,
.I /etc/hosts ,
.I /etc/networks ,
.I /etc/services ,
.I /etc/protocols ,
and
.I /etc/ethers .
In addition, there is the
.I netgroup(5) ,
file, which defines network wide groups, and used for permission
checking when doing remote mounts, remote logins, and remote shells.
..IX etc/passwd "" \fI/etc/passwd\fP
..IX etc/group "" \fI/etc/group\fP
..IX etc/hosts "" \fI/etc/hosts\fP
..IX etc/networks "" \fI/etc/networks\fP
..IX etc/services "" \fI/etc/services\fP
..IX etc/protocols "" \fI/etc/protocols\fP
..IX etc/ethers "" \fI/etc/ethers\fP
..IX etc/netgroup "" \fI/etc/netgroup\fP
.LP
In SunOS 4.0, the library routines
.I getpwent() ,
..IX getpwent() "" \fIgetpwent()\fP
.I getgrent() ,
..IX getgrent() "" \fIgetgrent()\fP
and
.I gethostent() 
..IX gethostent() "" \fIgethostent()\fP
were rewritten to take advantage of the NIS.  Thus, C programs that
call these library routines may have to be relinked in order to
function correctly.
.
.NH 4
\&Hosts
..IX "Network Information Service" "hosts database"
.LP
The hosts file is stored as two different NIS maps.
..IX etc/hosts "" \fI/etc/hosts\fP
The first,
.I hosts.byname ,
is indexed by hostname.
The second,
.I hosts.byaddr ,
is indexed by Internet address.
Remember that this actually expands into four files, with suffixes
.I .pag ,
and
.I .dir .
When a user program calls the library routine
.I gethostbyname() ,
..IX gethostbyname() "" \fIgethostbyname()\fP
a single RPC call to a server retrieves the entry from the
.I hosts.byname 
file.  Similarly,
.I gethostbyaddr() 
..IX gethostbyaddr() "" \fIgethostbyaddr()\fP
retrieves the entry from the
.I hosts.byaddr 
file.  If the NIS is not running (which is caused by commenting
.I ypbind 
out of the
.I /etc/rc 
file), then
.I gethostbyname() 
will read the
.I /etc/hosts 
files, just as it always has.
.LP
Normally, the hosts file for the NIS will be the same as the
.I /etc/hosts 
file on the machine serving as a NIS master.
In this case, the makefile in
.I /etc/yp 
will check to see if
.I /etc/hosts 
is newer than the
.I dbm 
file.  If it is, it will use a simple
.I sed 
script to recreate
.I hosts.byname 
and
.I hosts.byaddr ,
run them through
.I makedbm 
and then call
.I yppush 
See
.I ypmake 
for details.
..IX ypmake "" "\fIypmake\fP command"
.
.NH 4
\&Passwd
..IX "Network Information Service" "password database"
.LP
The
.I passwd 
file is similar to the
.I hosts 
file.  It exists as two separate files,
.I passwd.byname 
and
.I passwd.byuid .
..IX etc/passwd "" \fI/etc/passwd\fP
The
.I ypcat 
program prints it, and
.I ypmake 
updates it.  However, if
.I getpwent 
always went directly to the NIS as does
.I gethostent ,
then everyone would be forced to have an identical password file.
Consequently,
.I getpwent 
reads the local
.I /etc/passwd 
file, just as it always did.
But now it interprets ``+'' entries in the password file to mean,
interpolate entries from the NIS database.
If you wrote a simple program using
.I getpwent 
to print out all the entries from your password file,
it would print out a virtual password file:
rather than printing out + signs,
it would print out whatever entries the local password file
included from the NIS database.
.
.NH 4
\&Others
.LP
Of the other files in
.I /etc ,
.I /etc/group 
is treated like
.I /etc/passwd ,
in that
.I getgrent() 
will only consult the NIS if explicitly told to do so by the
.I /etc/group 
file.  The files
.I /etc/networks ,
.I /etc/services ,
.I /etc/protocols ,
.I /etc/ethers ,
and
.I /etc/netgroup 
are treated like
.I /etc/hosts :
for these files, the library routines go directly to the NIS,
without consulting the local files.
.
.NH 4
\&Changing your passwd
..IX "changing passwords"
..IX "passwords, changing"
.LP
To change data in the NIS,
the system administrator must log into the master machine,
and edit databases there;
.I "ypwhich\ -m"
tells where the master server is.
However, since changing a password is so commonly done, the
.I yppasswd 
command has been provided to change your NIS password.
..IX yppasswd "" "\fIyppasswd\fP command"
It has the same user interface as the
.I passwd 
..IX passwd "" \fIpasswd\fP
command.  This command will only work if the
.I yppasswdd 
..IX yppasswdd "" "\fIyppasswdd\fP command"
server has been started up on the NIS master server machine.
.
.NH 2
\&The Network Lock Manager
..IX "Network Lock Manager" "" "" "" PAGE MAJOR
..IX "Lock Manager" "" "" "" PAGE MAJOR
.LP
SunOS includes an NFS-compatible Network Lock Manager (see the
.I lockd(8C)
man page for more details) that supports the
.I lockf()/fcntl() ,
System V style of advisory file and record locking over the
network.  System V locks are generally considered superior to
4.3BSD locks, implemented with the
.I flock() 
system call, for they provide record level, and not merely file
level, locking.  Record level locking is essential for database
systems.  Sun does support
.I flock() 
for use on individual machines, but
.I flock() 
is not intended to be used across the network.
.I flock() 
locks exclude only other processes on the same machine.
There is no interaction between
.I flock() 
and
.I lockf() .
.LP
Locking prevents multiple processes from modifying the same file
at the same time, and allows cooperating processes to synchronize
access to shared files.  The user interfaces with Sun's network
locking service by way of the standard
.I lockf() 
system-call interface, and rarely requires any detailed knowledge
of how it works.  The kernel maps user calls to
.I flock() 
and
.I fcntl() 
into RPC-based messages to the local lock manager (or, if the files
in question are on RFS-mounted filesystems, into calls to RFS).
..IX "Remote File Sharing (RFS)"
..IX RFS
The fact that the file system may be spread across multiple machines
is really not a complication \(em until a crash occurs.
.LP
..IX "Lock Manager" crashing
All computers crash from time to time, and in an NFS environment,
where multiple machines can have access to the same file at the
same time, the process of recovering from a crash is necessarily
more complex than in a non-network environment.  Furthermore,
locking is
.I "inherently stateful" .
If a server crashes, clients with locked files must be able to
..IX "Lock Manager" state
recover their locks.  If a client crashes, its servers must have
the sense to hold the client's locks while it recovers.  And, to
preserve NFS's overall transparency, the recovery of lost locks
must not require the intervention of the applications themselves.
This is accomplished as follows:
.IP \(bu
Basic file access operations, such as read and write, use a
stateless protocol (the NFS protocol).  All interactions between
NFS servers and clients are atomic \(em the server doesn't
remember anything about its clients from one interaction to the
next.  In the case of a server crash, client applications will
simply sleep until it comes back up and their NFS operations
can complete.
.IP \(bu
.I "Stateful services"
..IX "stateful services"
(those that require the server to maintain client information from
one transaction to the next) such as the locking service, are not
part of NFS per se.  They are separate services that use the
status monitor (see
.I "The Network Status Monitor" )
to ensure that their implicit network state information remains
consistent with the real state of the network.  There are two
specific state-related problems involved in providing locking in
a network context:
.RS
.IP
1) if the client has crashed, the lock can be held forever by
the server
.IP
2) if the server has crashed, it loses its state (including all
its lock information) when it recovers.
.RE
.IP
..IX "Network Status Monitor"
..IX "Status Monitor"
The Network Lock Manager solves both of these problems by
cooperating with the Network Status Monitor to ensure that it's
notified of relevant machine crashes.  Its own protocol then
allows it to recover the lock information it needs when crashed
machines recover.
.LP
The lock manager and the status monitor are both network-service
daemons \(em they run at user level, but they are essential to
the kernel's ability to provide fundamental network services, and
they are therefore run on all network machines.
They are best seen as extensions
to the kernel which, for reasons of space, efficiency and
organization, are implemented as daemons.  Most application programs
will request the network service through a system call to the kernel (like
.I lockf() ),
though it is possible to interact with the service directly with RPC.
With
.I lockf() 
the kernel uses RPC
to call the daemon.  The network daemons communicate among themselves
with RPC (see
.I "The Locking Protocol"
for some details of the lock manager protocol).  It should be
noted that the daemon-based approach to network services allows
for tailoring by users who need customized services.
.\"It's possible, for example, for users to alter the
.\"lock manager to provide locking in a different style.
.LP
The following figure depicts the overall architecture of the
locking service.
.KS
.LP
.ce
Figure: Architecture of the NFS Locking Service
.PS
ellipse at  6.800, 6.762 wid  0.900 ht  0.900
arrowht =  0.1000i
arrowwid =  0.0500i
line ->  from  5.488, 7.512 to  6.613, 7.200
dashwid =  0.0500i
line <-> dashed from  2.737, 6.762 to  4.925, 6.762
line <->  from  2.737, 8.012 to  4.925, 8.012
line   from  4.925, 7.262 to  6.175, 7.262\
	to  6.175, 6.263\
	to  4.925, 6.263\
	to  4.925, 7.263
line <-  from  5.488, 8.512 to  5.488, 9.012
line <->  from  5.488, 7.512 to  5.488, 7.262
line   from  4.925, 8.512 to  6.175, 8.512\
	to  6.175, 7.513\
	to  4.925, 7.513\
	to  4.925, 8.512
line <-  from  2.050, 8.512 to  2.050, 9.012
line <->  from  2.050, 7.512 to  2.050, 7.262
line  dashed from  2.050, 7.512 to  2.050, 7.262
line  dashed from  4.113, 9.512 to  7.363, 9.512\
	to  7.362, 6.013\
	to  4.112, 6.013\
	to  4.112, 9.512
line  dashed from  0.237, 9.512 to  3.487, 9.512\
	to  3.487, 6.013\
	to  0.237, 6.013\
	to  0.237, 9.512
line   from  1.487, 8.512 to  2.737, 8.512\
	to  2.737, 7.513\
	to  1.488, 7.513\
	to  1.488, 8.512
line   from  1.487, 7.262 to  2.737, 7.262\
	to  2.737, 6.263\
	to  1.488, 6.263\
	to  1.488, 7.263
"Status Messages" at  3.237, 6.825 ljust
"RPC Lock Requests" at  3.112, 8.075 ljust
"monitor" at  5.175, 6.637 ljust
"monitor" at  1.737, 6.637 ljust
"status" at  5.175, 6.825 ljust
"status" at  1.737, 6.825 ljust
"files" at  6.613, 6.650 ljust
"local" at  6.613, 6.825 ljust
"manager" at  5.175, 7.800 ljust
"manager" at  1.737, 7.800 ljust
"lock" at  5.175, 8.000 ljust
"lock" at  1.737, 8.000 ljust
"lock requests" at  5.600, 8.762 ljust
"lock requests" at  2.100, 8.762 ljust
"local applications" at  4.988, 9.100 ljust
"local applications" at  1.487, 9.100 ljust
"Machine B" at  5.050, 9.600 ljust
"Machine A" at  1.237, 9.600 ljust
.PE
.KE
.\""files" at  0.613, 6.650 ljust
.\""local" at  0.613, 6.825 ljust
.\"ellipse at  0.863, 6.762 wid  0.900 ht  0.900
.\"line ->  from  2.050, 7.512 to  1.050, 7.200
.LP
At each server site, a lock manager process accepts lock requests,
made on behalf of client processes by a remote lock manager, or on
behalf of local processes by the kernel.  The client and server
lock managers communicate with RPC calls.  Upon receiving a
lock request for a machine that it doesn't already hold a lock on,
the lock manager registers its interest in that machine with the
local status monitor, and waits for that monitor to notify it that
the machine is up.  The monitor continues to watch the status of
registered machines, and notifies the lock manager is one of them
is rebooted (after a crash).  If the lock request is for a local
file, the lock manager tries to satisfy it, and communicates back
to the application along the appropriate RPC path.
.LP
The crash recovery procedure is very simple.  If the failure of a
client is detected, the server releases the failed client's locks,
on the assumption that the client application will request locks
again as needed.  If the recovery (and, by implication, the crash)
of a server is detected, the client lock manager retransmits all
lock requests previously granted by the recovered server.  This
retransmitted information is used by the server to reconstruct its
locking state.  See below for more details.
.LP
The locking service recovers from failure in a stateless manner.
Its state information is carefully circumscribed within a
pair of system daemons that are set up for automatic,
application-transparent crash recovery.  If a server crashes, and
thus loses its state, it expects that its clients will be notified
of the crash and send it the information that it needs to reconstruct
its state.  The key in this approach is the status monitor, which the
lock manager uses to detect both client and server failures.
.
.NH 3
\&The Locking Protocol
..IX "Lock Manager" protocol
.LP
The lock style implemented by the network lock manager is that
specified in the
.I "AT&T System V Interface Definition" ,
(see the
.I lockf(2) 
and
.I fcntl(2) 
man pages for details).  There is no interaction between the lock
manager's locks and
.I flock() -style
locks, which remain supported, but which should be used for non-network
applications only.
.LP
..IX "locks, advisory"
..IX "advisory locks"
Locks are presently advisory only, on the (well supported) assumption
that cooperating processes can do whatever they wish without mandatory
locks. (See the
.I fcntl(2) 
man page for more information about advisory locks).
.LP
There are four basic Lock Manager requests that are made by the kernel
in response to various
.I ioctl()/fcntl() 
calls:
.IP \fIKLM_LOCK\fP
Lock the specified record.
.IP \fIKLM_UNLOCK\fP
Unlock the specified record.
.IP \fIKLM_TEST\fP
Test if the specified record is locked.
.IP \fIKLM_CANCEL\fP
Cancel an outstanding lock request. 
.LP
Despite the fact that the network lock manager adheres to the
.I lockf()/fcntl() 
semantics, there are a few subtle points about its behavior that
deserve mention.  These arise directly from the nature of the network:
.IP \(bu
The first and most important of these has to do with crashes.  When
an NFS-client goes down, the lock managers on all of its servers are
notified by their status monitors, and they simply releases its locks,
on the assumption that it will request them again when it wants them.
When a server crashes, however, matters are different: the clients
will wait for it to come back up, and when it does, its lock manager
will give the client lock managers a grace period to submit lock reclaim
requests, and during this period will accept only reclaim requests.
The client status monitors will notify their respective lock managers
when the server recovers.  The default grace period is 45 seconds.
.IP \(bu
It is possible that, after a server crash, a client will not be able
to recover a lock that it had on a file on that server.  This can
happen for the simple reason that another process may have beaten the
recovering application process to the lock.  In this case the
.I SIGLOST 
signal will be sent to the process (the default action for this signal
is to kill the application).
.IP \(bu
The local lock manager does not reply to the kernel lock request
until the server lock manager has gotten back to it.  Further, if the
lock request is on a server new to the local lock manager, the
lock manager registers its interest in that server with the
local status monitor and waits for its reply.  Thus, if either the
status monitor or the server's lock manager are unavailable, the
reply to a lock request for remote data is delayed until it
becomes available.
.
.NH 2
\&The Network Status Monitor
..IX "Status Monitor" "" "" "" PAGE MAJOR
..IX "Network Status Monitor" "" "" "" PAGE MAJOR
..IX statd "" \fIstatd\fP
.LP
The Network Status Monitor (see the
.I statd(8C)
man page for more details) was
introduced with the lock manager, which relies heavily on it to
maintain the inherently stateful locking service within the
stateless NFS environment.  However, the status monitor is very
general, and can also be used to support other kinds of stateful
network services and applications.  Normally, crash recovery is
one of the most difficult aspects of network application
development, and requires a major design and installation
effort.  The status monitor makes it more or less routine.
.LP
It is anticipated that, in the future, new network services, some
of them stateful, will be introduced into the Sun system.  These
services will use the status monitor to keep up with the state of
the network and to cope with machine crashes.
.LP
The status monitor works by providing a general framework for
collecting network status information.  Implemented as a daemon
that runs on all network machines, it implements a simple
protocol which allows applications to easily monitor the status of
other machines.  Its use improves overall robustness, and
avoids situations in which applications
running of different machines (or even on the same machine)
come to disagree about the status of a site \(em a potentially
dangerous situation that can lead to inconsistencies in many
applications.
.LP
Applications using the status monitor do so by registering with
it the machines that they are interested in.  The monitor then
tracks the status of those machines, and when one of them
crashes\**
.FS
Actually, when one of them recovers from a crash.
.FE
it notifies the interested applications to that effect, and
they then take whatever actions are necessary to reestablish a
consistent state.
.LP
There are several major advantages to this approach:
.IP \(bu
Only applications that use stateful services must pay the
overhead \(em in time and in code \(em of dealing with the
status monitor.
.IP \(bu
The implementation of stateful network applications is eased,
since the status monitor shields application developers from
the complexity of the network.
