.if \n(1T .bp
.NH 1
\&Introduction to Remote Procedure Calls
.nr OF 1
..IX RPC introduction
.NH 2
\&Overview
.LP
What are Remote Procedure Calls?  Simply put, they are a high-level
communications paradigm which allows network applications to be developed
by way of specialized kinds of procedure calls designed to hide the
details of the underlying networking mechanisms.
..IX "Remote Procedure Call"
..IX RPC
.LP
RPC implements a logical client to server communications system designed
specifically for the support of network applications. With RPC, the client
makes a procedure call which sends requests to the server as necessary.
When these requests arrive, the server calls a dispatch routine, performs
whatever service is requested, sends back the reply, and the procedure 
call returns to the client.
..IX "caller process"
..IX "server process"
.NH 3
\&How it is useful
.LP
The net effect of programming with RPC is that programs are designed to
run within a client/server network model.  Such programs use RPC mechanisms
to avoid the details of interfacing to the network, and provide network
services to their callers without even requiring that they be aware of the
existence and function of the underlying network.
.LP
This mechanism solves the tedious issues of programming by making the calls
transparent.  For example, a program can simply make a call to
.I rnusers() ,
a C routine which returns the number of users on a remote machine.  The caller
is not explicitly aware of using RPC \(em they simply call a procedure,
much like making a system call to
.I malloc() .
.LP
Even though this discussion only mentions the interface to C, Remote Procedure
Calls can be made from any language.  Additionally even though this discussion
refers to RPC only as it is used to communicate
between processes on different machines, it also works for communication
between different processes on the same machine.
.NH 3
\&Terminology
.LP
This chapter discusses servers, services, programs, procedures,
clients, and versions.  A server provides network services and a network
service is a
collection of one or more remote programs.  A remote program
implements one or more remote procedures; the procedures, their
parameters, and results are documented in the specific program's
protocol specification.
Network clients initiate remote procedure calls
to services.  A server may support more than one version of a
remote program in order to be forward compatible with changing
protocols.
.LP
For example, a network file service may be composed of two
programs.  One program may deal with high-level applications such
as file system access control and locking.  The other may deal with
low-level file IO and have procedures like "read" and "write".  A
client machine of the network file service would call the
procedures associated with the two programs of the service on
behalf of some user on the client machine.
.NH 3
\&The RPC Model
.LP
The remote procedure call model is similar to the local procedure
call model.  In the local case, the caller places arguments to a
procedure in some well-specified location (such as a result
register).  It then transfers control to the procedure, and
eventually gains back control.  At that point, the results of the
procedure are extracted from the well-specified location, and the
caller continues execution.
.LP
The remote procedure call is similar, in that one thread of control
logically winds through two processes\(emone is the caller's
process, the other is a server's process.  That is, the caller
process sends a call message to the server process and waits
(blocks) for a reply message.  The call message contains the
procedure's parameters, among other things.  The reply message
contains the procedure's results, among other things.  Once the
reply message is received, the results of the procedure are
extracted, and caller's execution is resumed.
.LP
On the server side, a process is dormant awaiting the arrival of a
call message.  When one arrives, the server process extracts the
procedure's parameters, computes the results, sends a reply
message, and then awaits the next call message.
Please refer to the figure \fINetwork Communication with the
Remote Procedure Call\fP.  
.LP
Note that in this model, only one of the two processes is active at
any given time.
The RPC protocol makes no restrictions on the concurrency model
implemented, and others are possible.  For example, an
implementation may choose to have RPC calls be asynchronous, so
that the client may do useful work while waiting for the reply from
the server.  Another possibility is to have the server create a
task to process an incoming request, so that the server can be free
to receive other requests.
For a more detailed discussion on the RPC protocol, see the chapter
\fIRemote Procedure Calls:  Protocol Specification\fP.  
.KS
.LP
.ce
Figure: Network Communication with the Remote Procedure Call
.sp
.PS
L1: arrow down 1i "client " rjust "program " rjust
L2: line right 1.5i "\fIRPC Call\fP" ""
move up 1.5i; line dotted down 6i; move up 4.5i
arrow right 1i
L3: arrow down 1i "invoke " rjust "service " rjust
L4: arrow right 1.5i "call" "service"
L5: arrow down 1i " service" ljust " executes" ljust
L6: arrow left 1.5i "\fIAnswer Return\fP" ""
L7: arrow down 1i "request " rjust "completed " rjust
L8: line left 1i
arrow left 1.5i "\fIreturn\fP" "reply"
L9: arrow down 1i "program " rjust "continues " rjust
line dashed down from L2 to L9
line dashed down from L4 to L7
line dashed up 1i from L3 "service " rjust "daemon " rjust
arrow dashed down 1i from L8
move right 1i from L3
box invis "Machine B"
move left 1.2i from L2; move down
box invis "Machine A"
.PE
.KE
.LP
In the above diagram, the details of the network transport are hidden
within the Remote Procedure Call.  Note, however, that the RPC would not
be very useful if those details were entirely unavailable to user and
programmers who required access to them.
.NH 2
\&Versions and Numbers
.LP
Each RPC procedure is uniquely defined by a program number and procedure
number.  The program number specifies a group of related remote procedures,
each of which has a different procedure number.  Each program also has a
version number, so when a minor change is made to a remote service (adding
a new procedure, for example), a new program number doesn't have to be
assigned.  For example, when you want to call a procedure to find the number
of remote users, you look up the appropriate program, version and procedure
numbers in a manual, just as you look up the name of a memory allocator when
you want to allocate memory.
.NH 2
\&Portmap
.LP
The 
.I portmap 
is the only network service that must have such a
well-known (dedicated) port.  Other network services can be
assigned port numbers statically or dynamically so long as they
register their ports with their host's 
.I portmap .
The
.I portmap 
is started
automatically whenever a machine is booted.
As part of its initialization, a server program calls its
host's
.I portmap 
to create a portmap entry for its program and version number.
To find a remote program's port, a client sends
an RPC call message to a server's
.I portmap ;
if the remote program is
registered with the
.I portmap ,
it returns the relevant port number
in an RPC reply message.  The client program can then send RPC call
messages to the remote program's port.
.NH 2
\&Transports and Semantics
.LP
The RPC protocol is independent of transport protocols.  That is,
RPC does not care how a message is passed from one process to
another.  The protocol deals only with specification and
interpretation of messages.
.LP
It is important to point out that RPC does not try to implement any
kind of reliability and that the application must be aware of the
type of transport protocol underneath RPC.  If it knows it is
running on top of a reliable transport such as TCP/IP[6], then most
of the work is already done for it.  On the other hand, if it is
running on top of an unreliable transport such as UDP/IP[7], it
must implement is own retransmission and time-out policy as the RPC
layer does not provide this service.
.LP
Because of transport independence, the RPC protocol does not attach
specific semantics to the remote procedures or their execution.
Semantics can be inferred from (but should be explicitly specified
by) the underlying transport protocol.  For example, consider RPC
running on top of an unreliable transport such as UDP/IP.  If an
application retransmits RPC messages after short time-outs, the
only thing it can infer if it receives no reply is that the
procedure was executed zero or more times.  If it does receive a
reply, then it can infer that the procedure was executed at least
once.
.LP
On the other hand, if using a reliable transport such as TCP/IP,
the application can infer from a reply message that the procedure
was executed exactly once, but if it receives no reply message, it
cannot assume the remote procedure was not executed.  Note that
even if a connection-oriented protocol like TCP is used, an
application still needs time-outs and reconnection to handle server
crashes.
.NH 3
\&Transport Selection
.LP
Sun RPC is currently supported on both UDP/IP and TCP/IP transports.
The selection of the transport depends upon the requirements of the
application.  UDP (connection less) may be the transport of choice if
the application has all of the following characteristics:
.IP 1.
The procedures are idempotent.  i.e. the same procedure can 
be executed more than once without any harmful side-effects.
For example, reading a block of data is idempotent, while
creating a file is a non-idempotent operation.
.IP 2.
The size of both the arguments and results is smaller than the
UDP packet size (8 Kbytes for Sun UDP implementation).
.IP 3.
The server is required to handle many (hundreds) of clients.
Since the UDP server does not keeps any state about the client, it
can potentially handle many clients.  On the other hand, TCP
server keeps state for each open client connection and hence the
number of clients is limited by the machine resources.
.LP
TCP (connection oriented) may be the transport of choice if
the application has any of the following requirements and characteristics:
.IP 1.
The application needs to maintain a high degree of reliability.
.IP 2.
The procedures are non-idempotent and at-most-once
semantics are required.
.IP 3.
The size of either the arguments or the results exceeds 8 Kbytes.
.NH 2
\&External Data Representation
.LP
RPC presumes the existence of the eXternal Data Representation (XDR),
a standard for the machine-independent description and encoding of data.
XDR is useful for transferring data between different computer architectures,
and has been used to communicate data between such divers machines as the
Sun Workstation, VAX, IBM-PC, and Cray.
.LP
RPC can handle arbitrary data structures, regardless of different machines'
byte orders or structure layout conventions, by always converting them to
a network standard called \fIExternal Data Representation\fP (XDR) before
sending them over the wire.  The process of converting from a particular 
machine representation to XDR format is called \fIserializing\fP,
and the reverse process is called \fIdeserializing\fP.
For a detailed discussion of XDR, see the chapter
\fIExternal Data Representation Standard:  Protocol Specification\fP.  
.NH 2
\&rpcinfo
.LP
.I rpcinfo 
is a command that reports current RPC registration information
known to
.I portmap 
(and can be used by administrators to delete registrations).
.I rpcinfo 
can be used to find all the RPC services registered on a specified host
and to report their port numbers and the transports for which they are
registered.  It can also be used to call (ping) a specific version of a
specific program on a specific host using TCP or UDP transport, and to
report whether the response was received.  For details, see the
.I rpcinfo(8C) 
manual pages.
.NH 2
\&Assigning Program Numbers
..IX "program number assignment"
..IX "assigning program numbers"
.LP
Program numbers are assigned in groups of
.I 0x20000000 
according to the following chart:

.DS
.ft CW
       0x0 - 0x1fffffff	\fRDefined by Sun\fP
0x20000000 - 0x3fffffff	\fRDefined by user\fP
0x40000000 - 0x5fffffff	\fRTransient\fP
0x60000000 - 0x7fffffff	\fRReserved\fP
0x80000000 - 0x9fffffff	\fRReserved\fP
0xa0000000 - 0xbfffffff	\fRReserved\fP
0xc0000000 - 0xdfffffff	\fRReserved\fP
0xe0000000 - 0xffffffff	\fRReserved\fP
.ft R
.DE

Sun Microsystems administers the first group of numbers, which
should be identical for all Sun customers.  If a customer
develops an application that might be of general interest, that
application should be given an assigned number in the first
range.  The second group of numbers is reserved for specific
customer applications.  This range is intended primarily for
debugging new programs.  The third group is reserved for
applications that generate program numbers dynamically.  The
final groups are reserved for future use, and should not be
used.
.LP
To register a protocol specification, send a request by network
mail to
.I rpc@sun.com ,
or write to:
.DS
RPC Administrator
Sun Microsystems
2550 Garcia Ave.
Mountain View, CA 94043
.DE
Please include a compilable
.I rpcgen 
.I ``.x'' 
file describing your protocol.
You will be given a unique program number in return.
..IX RPC administration
..IX administration "of RPC"
.LP
Some of the RPC program numbers can be found in /etc/rpc. Protocol specifications
of standard Sun RPC services can be
found in the include files in
.I "/usr/include/rpcsvc" .
These services, however, constitute only a small subset
of those which have been registered.  A list of
some of the registered programs is:
.LP
.ce
Table: Registered RPC Program Numbers
.TS H
box tab (&) ;
lB lB lB
l l l .
RPC Number&Program&Description
_
.TH
.sp.5
100000&PMAPPROG&portmap
100001&RSTATPROG&remote stats
100002&RUSERSPROG&remote users
100003&NFSPROG&nfs
100004&YPPROG&NIS
100005&MOUNTPROG&mount daemon
100006&DBXPROG&remote dbx
100007&YPBINDPROG&NIS binder
100008&WALLPROG&shutdown msg
100009&YPPASSWDPROG&yppasswd server
100010&ETHERSTATPROG&ether stats
100011&RQUOTAPROG&disk quotas
100012&SPRAYPROG&spray packets
100013&IBM3270PROG&3270 mapper
100014&IBMRJEPROG&RJE mapper
100015&SELNSVCPROG&selection service
100016&RDATABASEPROG&remote database access
100017&REXECPROG&remote execution
100018&ALICEPROG&Alice Office Automation
100019&SCHEDPROG&scheduling service
100020&LOCKPROG&local lock manager
100021&NETLOCKPROG&network lock manager
100022&X25PROG&x.25 inr protocol
100023&STATMON1PROG&status monitor 1
100024&STATMON2PROG&status monitor 2
100025&SELNLIBPROG&selection library
100026&BOOTPARAMPROG&boot parameters service
100027&MAZEPROG&mazewars game
100028&YPUPDATEPROG&NIS update
100029&KEYSERVEPROG&key server
100030&SECURECMDPROG&secure login
100031&NETFWDIPROG&nfs net forwarder init
100032&NETFWDTPROG&nfs net forwarder trans
100033&SUNLINKMAP_PROG&sunlink MAP
100034&NETMONPROG&network monitor
100035&DBASEPROG&lightweight database
100036&PWDAUTHPROG&password authorization
100037&TFSPROG&translucent file svc
100038&NSEPROG&nse server
100039&NSE_ACTIVATE_PROG&nse activate daemon
100043&SHOWHFD&showfh
.sp .2i
150001&PCNFSDPROG&pc passwd authorization
.sp .2i
200000&PYRAMIDLOCKINGPROG&Pyramid-locking
200001&PYRAMIDSYS5&Pyramid-sys5
200002&CADDS_IMAGE&CV cadds_image
.sp .2i
300001&ADT_RFLOCKPROG&ADT file locking
.TE
