/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)readdir.c	1.4 90/07/20 4.1NFSSRC; from	5.2 (Berkeley) 3/9/86";
#endif LIBC_SCCS and not lint

/*
 *	readdir()
 *
 *	readdir() returns a pointer to the next directory entry.
 *	It returns NULL upon reaching the end of the directory or
 *	detecting an invalid seekdir() operation.
 */

#include <sys/param.h>
#include <dirent.h>

struct dirent *
readdir(dirp)
register DIR	*dirp;
{
	/*
	 * We loop through, skipping empty entries, and refilling
	 * our buffer with a block of entries as needed.
	 */

	for (;;)
	{
		register struct dirent	*p;
		long	lastloc;

		/*
		 * The logic is inverted here to leave the
		 * dd_loc pointing at the entry we return,
		 * more convoluted than otherwise necessary.
		 */

		if (dirp->dd_size > 0)
		{
			lastloc = dirp->dd_loc;
			dirp->dd_loc += 
			    ((struct dirent *)(dirp->dd_buf
					+ dirp->dd_loc))->d_reclen;
		}

		if (dirp->dd_loc >= dirp->dd_size) {
			dirp->dd_size = 0;
			dirp->dd_loc = 0;;
		}

		/*
		 * If dd_size is 0, we have come from opendir or have
		 * exhausted entries in this block, read another.
		 *
		 * We leave dd_loc pointing at entry we return.
		 */
	
		if (dirp->dd_size == 0)
		{
			if ((dirp->dd_size = getdents(dirp->dd_fd,
				dirp->dd_buf, (int)dirp->dd_bsize)) <= 0)
			{
				/* EOF or ERROR defined in errno */
				dirp->dd_loc = lastloc;
				return NULL;
			}
		}
	
		/*
		 * We have an entry.
		 */

		p = (struct dirent *)(dirp->dd_buf + dirp->dd_loc);

		/*
		 * Prevent obvious hosed directory errors?
		 */

		if (p->d_reclen <= 0
			|| p->d_reclen > (dirp->dd_bsize + 1 - dirp->dd_loc))
			return NULL;

		/*
		 * The first entry of a block is marked free by setting
		 * d_fileno to be zero. All other freed entries are
		 * added onto the previous entry, making for large d_reclen's
		 * sometimes.
		 */

		if (p->d_fileno == 0)
			continue;

		dirp->dd_off = p->d_off;
		return p;
	}
}

