/*
 *  Program to generate xdr'ed table of contents for distribution
 *  test tapes.
 *
 *  @(#)gen_toc.c	1.2 90/07/25 4.1NFSSRC SMI; from 1.2 87/10/23 D/NFS
 *  Copyright 1987 Sun Microsystems, Inc.
 */

#include "mktp.h"
#define MALLOC(x) (x *)mymalloc(sizeof(x))

#define SZ 100
char Arch[SZ];
char *Prog, *Infile, *Outfile;
char buf[BUFSIZ];

main(argc, argv)
int argc;
char **argv;
{
	register toc_entry *eptr;
	Address *ap;
	Information *ip;
	contents *con;
	amorphous *amor;
	Installable *ins;
	FILE *infile, *outfile;
	char ofilename[SZ];
	char os[SZ], release[SZ], date[SZ], source[SZ];
	char name[SZ], type[SZ], loc[SZ], movable[SZ];
	char swtype[SZ];
	char *p1, *p2, *index();
	char junk[SZ];
	int tapeno, filenum, size;
	int maxvol = 1;
	int ret, cnt;
	file_type typen;
	char *mymalloc(), *mstr();
	char *nullstr = "";

	Prog = argv[0];
	if (argc != 3 && argc != 4) {
		fprintf(stderr, "usage: %s infile outfile [archname]\n", Prog);
		exit(1);
	} 
	Infile = argv[1];
	Outfile = argv[2];
	if (argc == 4)
		strcpy(Arch, argv[3]);
	else
		Arch[0] = '\0';
	if ((infile = fopen(Infile, "r")) == NULL) {
		perror(Infile);
		exit(1);
	}

	if (fgets(buf, BUFSIZ, infile) == NULL) {
		fprintf(stderr, "%s: Premature EOF on %s\n", Prog, Infile);
		exit(1);
	}
	if (p1 = index(buf, '\n'))
		*p1 = '\0';
	/* os */
	for (p1 = buf, p2 = buf; *p2 && *p2 != ' '; p2++)
		;
	if (*p2 == '\0') {
		fprintf(stderr, "Bad format first line\n");
		exit(1);
	}
	*p2++ = '\0';
	strcpy(os, p1);
	/* release */
	for (p1 = p2; *p2 && *p2 != ' '; p2++)
		;
	if (*p2 == '\0') {
		fprintf(stderr, "Bad format first line\n");
		exit(1);
	}
	*p2++ = '\0';
	strcpy(release, p1);
	/* date */
	if (strncmp(p2, "of ", 3)) {
		fprintf(stderr, "Bad format first line\n");
		exit(1);
	}
	p2 += 3;
	for (p1 = p2; *p2 && strncmp(p2, " from ", 6); )
		for (p2++; *p2 && *p2 != ' '; p2++)
			;
	if (*p2 == '\0') {
		fprintf(stderr, "Bad format first line\n");
		exit(1);
	}
	*p2 = '\0';
	p2 += 6;
	strcpy(date, p1);
	strcpy(source, p2);

	if (fgets(buf, BUFSIZ, infile) == NULL) {
		fprintf(stderr, "%s: Premature EOF on %s\n", Prog, Infile);
		exit(1);
	}
	if ((ret = sscanf(buf, "ARCH %s", junk)) != 1) {
		fprintf(stderr, "ARCH sscanf ret %d exp 1\n", ret);
		exit(1);
	}
	if (Arch[0] == '\0')
		strcpy(Arch, junk);
	if (fgets(buf, BUFSIZ, infile) == NULL) {
		fprintf(stderr, "%s: Premature EOF on %s\n", Prog, Infile);
		exit(1);
	}
	if ((ret = sscanf(buf, "VOLUME %s", junk)) != 1) {
		fprintf(stderr, "VOLUME sscanf ret %d exp 1\n", ret);
		exit(1);
	}
	if (fgets(buf, BUFSIZ, infile) == NULL) {
		fprintf(stderr, "%s: Premature EOF on %s\n", Prog, Infile);
		exit(1);
	}
	if ((ret = sscanf(buf, " Vol File %s", junk)) != 1) {
		fprintf(stderr, "hdr sscanf ret %d exp 1\n", ret);
		exit(1);
	}
	
	dinfo.Title = os;
	dinfo.Source = source;
	dinfo.Version = release;
	dinfo.Date = date;
	dinfo.nentries = 0;		/* filled in by write_xdr_toc */
	dinfo.dstate = GONE;

	for (eptr=entries, cnt=0; fgets(buf, BUFSIZ, infile); eptr++, cnt++) {
		ret = sscanf(buf, "%d %d %s %d %s %s %s %s",
			&tapeno, &filenum, name, &size, type,
			loc, swtype, movable);
		switch (ret) {
		    case 5:
		    case 8:
			ap = &eptr->Where;
			ip = &eptr->Info;
			if (!strcmp(type, "toc")) {
				typen = TOC;
				ip->kind = CONTENTS;
			} else if (!strcmp(type, "image")) {
				typen = IMAGE;
				ip->kind = AMORPHOUS;
			} else if (!strcmp(type, "tar")) {
				typen = TAR;
				ip->kind = INSTALLABLE;
				if (ret == 5) {
					strcpy(loc, "/usr");
					strcpy(swtype, "desirable");
					strcpy(movable, "not_movable");
				}
			} else  {
				typen = UNKNOWN;
				ip->kind = UNDEFINED;
			}
			break;
		    default:
			fprintf(stderr, "sscanf ret %d exp 5 or 8\n", ret);
			fprintf(stderr, "buf: %s\n", buf);
			exit(1);
		}
		if (tapeno > maxvol)
			maxvol = tapeno;
		eptr->Name = mstr(name);
		eptr->LongName = eptr->Name;
		eptr->Command = nullstr;
		eptr->Type = typen;
		ap->dtype = TAPE;
		ap->Address_u.tape.volno = tapeno;
		ap->Address_u.tape.file_number = filenum;
		ap->Address_u.tape.label = nullstr;
		ap->Address_u.tape.size = size;
		ap->Address_u.tape.volsize = 0;
		ap->Address_u.tape.dname = nullstr;
		switch(ip->kind) {
		    case CONTENTS:
			ip->Information_u.Toc.junk = 0;
			break;
		    case AMORPHOUS:
			ip->Information_u.File.size = size;
			break;
		    case INSTALLABLE:
			mstrlist(Arch, &ip->Information_u.Install.arch_for);
			mstrlist("", &ip->Information_u.Install.soft_depends);
			if (!strcmp(swtype, "required")) {
				ip->Information_u.Install.required = TRUE;
				ip->Information_u.Install.desirable = FALSE;
				ip->Information_u.Install.common = FALSE;
			} else if (!strcmp(swtype, "desirable")) {
				ip->Information_u.Install.required = FALSE;
				ip->Information_u.Install.desirable = TRUE;
				ip->Information_u.Install.common = FALSE;
			} else if (!strcmp(swtype, "common")) {
				ip->Information_u.Install.required = FALSE;
				ip->Information_u.Install.desirable = FALSE;
				ip->Information_u.Install.common = TRUE;
			} else {
				ip->Information_u.Install.required = FALSE;
				ip->Information_u.Install.desirable = FALSE;
				ip->Information_u.Install.common = FALSE;
			}
			ip->Information_u.Install.loadpoint = mstr(loc);
			ip->Information_u.Install.moveable =
				strcmp(movable, "movable") == 0? TRUE : FALSE;
			ip->Information_u.Install.size = size;
			ip->Information_u.Install.pre_install = nullstr;
			ip->Information_u.Install.install = nullstr;
			break;
		    default:
			fprintf(stderr, "unknown file_type %d\n", typen);
			exit(1);
		}
	}
	vinfo.vaddr.dtype = TAPE;
	vinfo.vaddr.Address_u.tape.file_number = 0;
	vinfo.vaddr.Address_u.tape.size = 0;
	vinfo.vaddr.Address_u.tape.volsize = 0;
	vinfo.vaddr.Address_u.tape.dname = nullstr;
	fprintf(stderr, "Constructed %d entries\n", cnt);
	for (tapeno = 1; tapeno <= maxvol; tapeno++) {
		char label[SZ];

		sprintf(label, "VOLUME %d", tapeno);
		vinfo.vaddr.Address_u.tape.volno = tapeno;
		vinfo.vaddr.Address_u.tape.label = label;

		sprintf(ofilename, "%s.vol%d", Outfile, tapeno);
		if ((outfile = fopen(ofilename, "w")) == NULL) {
			perror(ofilename);
			continue;
		}
		ret = write_xdr_toc(outfile, &dinfo, &vinfo, entries, cnt);
		if (ret == TRUE)
			fprintf(stderr, "%d records written to %s\n",
				cnt, ofilename);
		else
			fprintf(stderr, "write_xdr_toc ret FALSE for %s\n",
				ofilename);
		fclose(outfile);
	}
	exit(0);
}

char *
mstr(s)
char *s;
{
	char *p;

	p = mymalloc(strlen(s) + 1);
	strcpy(p, s);
	return p;
}

mstrlist(s, list)
char *s;
string_list *list;
{
	char *p;
	stringp *pp;

	if (s == NULL || *s == '\0') {
		list->string_list_len = 0;
		list->string_list_val = NULL;
	} else {
		p = mstr(s);
		pp = MALLOC(stringp);
		*pp = p;
		list->string_list_len = 1;
		list->string_list_val = pp;
	}
}

char *
mymalloc(size)
int size;
{
	char *p;
	if ((p = malloc(size)) == NULL) {
		fprintf(stderr, "mymalloc: trying to malloc %d bytes\n", size);
		perror("malloc");
		exit(1);
	}
	return p;
}
