
%/*
% * Copyright (c) 1985, 1990 by Sun Microsystems, Inc.
% */
%
%/* from @(#)rpcb_prot.x	1.2 90/05/24 TIRPC 1.0 */

/*
 * rpcb_prot.x
 * RPCBIND protocol in rpc language
 */

#ifdef RPC_HDR
%
%#ifndef _rpcsvc_rpcb_prot_h
%#define _rpcsvc_rpcb_prot_h
%
#endif
/*
 * The following procedures are supported by the protocol:
 *
 * RPCBPROC_NULL() returns ()
 * 	takes nothing, returns nothing
 *
 * RPCBPROC_SET(RPCB) returns (bool_t)
 * 	TRUE is success, FALSE is failure.  Registers the tuple
 *	[prog, vers, address, netconfig].
 *
 * RPCBPROC_UNSET(RPCB) returns (bool_t)
 *	TRUE is success, FALSE is failure.  Un-registers tuple
 *	[prog, vers, netconfig->nc_netid].  addressss is ignored.
 *
 * RPCBPROC_GETADDR(RPCB) returns (Universal address).
 *	0 is failure.  Otherwise returns the universal address where the
 *	triple [prog, vers, nconf] is registered.
 *
 * RPCBPROC_DUMP() RETURNS (RPCBLIST *)
 *	used to dump the entire rpcbind maps
 * RPCBPROC_CALLIT(rpcb_rmtcallargs)
 * 	RETURNS (rpcb_rmtcallres);
 * 	Calls the procedure on the remote machine.  If it is not registered,
 *	this procedure is quiet; i.e. it does not return error information!!!
 *	This routine only passes null authentication parameters.
 *	It has no interface to xdr routines for RPCBPROC_CALLIT.
 *
 * RPCBPROC_GETTIME() returns (time).
 *	Gets the remote machines time
 *
 */

/*
 * A mapping of (program, version, network ID) to address
 */
struct rpcb {
	u_long r_prog;			/* program number */
	u_long r_vers;			/* version number */
	string r_netid<>;		/* network id */
	string r_addr<>;		/* universal address */
};

/*
 * A list of mappings
 */
struct rpcblist {
	rpcb rpcb_map;
	struct rpcblist *rpcb_next;
};

/*
 * Arguments of remote calls
 */
struct rpcb_rmtcallargs {
	u_long prog;			/* program number */
	u_long vers;			/* version number */
	u_long proc;			/* procedure number */
	opaque args_ptr<>;		/* argument */
};

/*
 * Results of the remote call
 */
struct rpcb_rmtcallres {
	string addr_ptr<>;		/* remote universal address */
	opaque results_ptr<>;		/* result */
};

/*
 * rpcbind procedures
 */
program RPCBPROG {
	version RPCBVERS {
		bool
		RPCBPROC_SET(rpcb) = 1;

		bool
		RPCBPROC_UNSET(rpcb) = 2;

		string
		RPCBPROC_GETADDR(rpcb) = 3;

		rpcblist
		RPCBPROC_DUMP(void) = 4;

		rpcb_rmtcallres
		RPCBPROC_CALLIT(rpcb_rmtcallargs) = 5;

		unsigned int
		RPCBPROC_GETTIME(void) = 6;

		struct netbuf
		RPCBPROC_UADDR2TADDR(string) = 7;

		string
		RPCBPROC_TADDR2UADDR(struct netbuf) = 8;
	} = 3;
} = 100000;

#ifdef RPC_HDR
%
%#endif /* ! _rpcsvc_rpcb_prot_h */
#endif
