#ifndef lint
static  char sccsid[] = "@(#)lodg_tli.c 1.7 90/08/29 Copyr 1989 Sun Micro";
#endif
/*
 * Copyright (c) 1988 by Sun Microsystems, Inc.
 */

/*
 * Streams based loopback datagram transport provider.
 * Compatible with the AT&T Transport Level Interface.
 *
 * General notes:
 * 1)	Addresses are maximum of 128 bytes, default addresses are
 *	4 bytes long.
 *
 * 2)   LODGTLI_PRINTF debug statements are switched by the global
 *	variable lodgtli_debug.
 */

#include "lodgtli.h"

#if	NLODGTLI > 0

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/debug.h>
#include <sys/file.h>
#include <sys/user.h>
#include <sys/errno.h>
#include <sys/vnode.h>
#include <sys/ucred.h>

#include <nettli/tihdr.h>
#include <nettli/tiuser.h>

#include <nettli/lodg_tli.h>

#define		LODG_MAXADDR		10000

#define		_RECOVER(A, B, C) \
	{ \
		putbq((A), (B)); \
		if (!bufcall((C), BPRI_MED, qenable, (caddr_t)(A))) \
			(void)timeout(qenable, (caddr_t)(A), hz); \
	}
#define		MSGBLEN(A)	((A)->b_wptr-(A)->b_rptr)
#define		MBLKLEN(A)	((A)->b_datap->db_lim-(A)->b_datap->db_base)

#define		LODGTLI_PRINTF	if (lodgtli_debug) (void)printf

#define		TT_OPEN		0x1
#define		TT_CLOSE	0x2
#define		TT_BLOCKED	0x4

extern int	hz;

#ifdef DYNAMIC
/******************************************************************************/
/*
 * Wrapper code to facilitate making this driver a loadable
 * module.
 */
#include	"sys/conf.h"
#include	"sun/autoconf.h"
#include	"sun/vddrv.h"

extern int 		nodev();
extern struct streamtab	lodgtli_info;
extern int		opencount;
extern int		lodgtli_debug;

struct cdevsw lodg_cdev = {
	nodev, nodev, nodev, nodev,
	nodev, nodev, nodev, 0,
	&lodgtli_info, 0
};

struct vdldrv vd = {
	VDMAGIC_PSEUDO,
	"lodgdrv",
#ifndef sun4c
	NULL,
	NULL,
	NULL,
	0,
	0,
#else sun4c
	NULL,
#endif sun4c
	NULL,
	&lodg_cdev,
	0,
	0
};

lodgtli_init(function_code, vdp, vdi, vds)
	unsigned int	function_code;
	struct vddrv	*vdp;
	addr_t		vdi;
	struct 	vdstat	*vds;
{
	switch (function_code) {
		case VDLOAD:
			printf("lodg driver loaded\n");
			vdp->vdd_vdtab = (struct vdlinkage *)&vd;
			return (0);

		case VDUNLOAD:
			printf("lodg driver unloaded\n");
			if (opencount == 0)
				return (0);
			else	return (-1);

		case VDSTAT:
			LODGTLI_PRINTF(
			"lodg: VDSTAT: opencount %d\n", opencount);
			return (0);
	}
}

/*****************************************************************************/
#endif DYNAMIC

/*
 * Stereotyped streams glue.
 */

int	lodg_open();
int	lodg_close();
int	lodg_wput();
int	lodg_wsrv();
int	lodg_rsrv();

struct module_info	lodg_minfo = {
	51,			/* module id */
	"lodg",
	0,
	INFPSZ,			/* max msg size */
	8096,			/* high water: 6 x RFS's */
	STRLOW			/* low water XXX: make reasonable */
};

struct qinit	lodg_rinit = {
	NULL,
	lodg_rsrv,
	lodg_open,
	lodg_close,
	NULL,
	&lodg_minfo,
	NULL
};

struct qinit	lodg_winit = {
	lodg_wput,
	lodg_wsrv,
	lodg_open,
	lodg_close,
	NULL,
	&lodg_minfo,
	NULL
};

struct streamtab	lodgtli_info = {
	&lodg_rinit,
	&lodg_winit,
	NULL,
	NULL
};

/*
 * link structure used when flow controlled.
 */
struct backwq {
	queue_t		*wrq;
	struct backwq	*next;
};

/*
 * Linked list of bound endpoints.
 */
struct lodg_link {
	struct lodg_lodg	*next;
	struct lodg_lodg	*prev;
};

/*
 * Pointer to head of linked list.
 */
static struct lodg_lodg		*lodg_listhead = (struct lodg_lodg *)NULL;

struct lodg_lodg {
	long			flags;
	int			state;
	queue_t			*rdq;
	char			*laddr;
	int			laddrlen;
	struct lodg_link	lodg_link;
	struct backwq		*oqptr;
	struct lodg_lodg	*dest_lodg;
	int			idflag;
	u_short			cr_uid;
	u_short			cr_gid;
	u_short			cr_ruid;
	u_short			cr_rgid;
};

/*
 * Global variables:
 *
 * 1) Per-device instance state information.
 *
 * 2) Debug flag.
 *
 * 3) Transport provider selected address.
 */
static struct lodg_lodg	lodg_cnt[NLODGTLI];
static u_int		gport;
int			lodgtli_debug = 0;

static void 		del_backq(/* queue_t *, struct lodg_lodg * */);
static void		add_backq(/* queue_t *, mblk_t *,
					struct lodg_lodg * */);
static int		snd_UDERR(/* queue_t *,
					struct T_unitdata_req *, int */);
static int		snd_ERRACK(/* queue_t *, mblk_t *, int, int, int */);
static int		snd_OKACK(/* queue_t *, mblk_t *, int */);
static void		lodg_addlink(/* struct lodg_lodg * */);
static void		lodg_dellink(/* struct lodg_lodg * */);
static struct lodg_lodg	*lodg_findlink(/* caddr_t, int */);
static int		chk_udata(/* mblk_t * */);
static int		chk_bind(/* mblk_t * */);
static mblk_t		*mk_udata_ind(/* queue_t *, mblk_t * */);
static void		lodg_fill_opt(/* struct lodg_lodg, struct lodg_lodg,
					char * */);
static int		do_default_ops(/* queue_t *, mblk_t * */);
static int		option_check(/* mblk_t * */);
static int		lodg_optmgmt(/* queue_t *, mblk_t * */);
static int		opencount;


/* ARGSUSED */
int
lodg_open(q, dev, oflag, sflag)
	struct queue	*q;
	dev_t		dev;
	int		oflag;
	int		sflag;
{
	register int			unit;
	struct lodg_lodg		*lodg;

	LODGTLI_PRINTF(
		"lodg_open: q: %x, WR(q): %x, dev: %x, oflag: %x, sflag: %x\n",
		q, WR(q), dev, oflag, sflag);

	/*
	 * Get our minor device number, verifying while we're at it
	 * that nobody's trying to open us as a module (as opposed
	 * to a device).
	 */
	if (sflag == CLONEOPEN) {
		/*
		 * Find an unused minor device instance.
		 * XXX -- Skip returning instance 0, until
		 * strclose() code is fixed
		 */
		for (unit = 1 /* XXX -- 0 */; unit < NLODGTLI; unit++)
			if (!(lodg_cnt[unit].flags & (TT_OPEN | TT_CLOSE)))
				break;
	} else if (sflag)
		return (OPENFAIL);
	else
		unit = minor(dev);

	if (unit < 0 || unit >= NLODGTLI) {
		u.u_error = ENXIO;
		return (OPENFAIL);
	}

	lodg = &lodg_cnt[unit];
	LODGTLI_PRINTF("lodg_open: lodg: %x index %d\n", lodg, lodg-lodg_cnt);
	if (lodg->flags & TT_OPEN) {
		u.u_error = EBUSY;		/* bletch! */
		return (OPENFAIL);
	}

	/*
	 * Set up cross-links between device and queues.
	 */
	lodg->rdq = q;
	q->q_ptr = (char *)lodg;
	WR(q)->q_ptr = (char *)lodg;

	/*
	 * Perform tli automaton state transition to the initial state.
	 */
	lodg->state = TS_UNBND;
	lodg->flags = TT_OPEN;

	lodg->oqptr = (struct backwq *)NULL;
	lodg->laddr = (caddr_t)NULL;
	lodg->laddrlen = 0;
	lodg->lodg_link.next = (struct lodg_lodg *)NULL;
	lodg->lodg_link.prev = (struct lodg_lodg *)NULL;

	/*
	 * Save ident information.
	 */
	lodg->idflag = 0;
	lodg->cr_uid = u.u_uid;
	lodg->cr_gid = u.u_gid;
	lodg->cr_ruid = u.u_ruid;
	lodg->cr_rgid = u.u_rgid;

	opencount++;

	return (unit);
}


lodg_close(q)
	register queue_t	*q;
{
	register struct lodg_lodg	*lodg;

	lodg = (struct lodg_lodg *)q->q_ptr;

	LODGTLI_PRINTF("lodg_close-%d: q: %x, WR(q): %x \n",
		lodg-lodg_cnt, q, WR(q));

	if (lodg->flags & TT_BLOCKED && lodg->dest_lodg)
		del_backq(q, lodg->dest_lodg);
	/*
	 * Flush outstanding traffic on its way
	 * downstream to us.
	 */
	flushq(WR(q), FLUSHALL);

	lodg->flags = TT_CLOSE;

	/*
	 * Delete from the linked list.
	 */
	if (lodg->state & TS_IDLE)
		lodg_dellink(lodg);

	/*
	 * Perform tli automaton state transition.
	 *
	 */
	lodg->state = 0;

	/*
	 * Break association between queue and driver instance.
	 */
	LODGTLI_PRINTF("lodg-%d: lodg: %x, rdq: %x closed\n",
		lodg-lodg_cnt, lodg, lodg->rdq);
	lodg->rdq = (queue_t *)NULL;
	lodg->flags = 0;

	opencount--;
}

/*
 * Only entered when back enabled by flow control.
 * That is a module upstream has taken messages of
 * it's queue so that the total number is below the
 * lo water mark for that queue, and so this service
 * procedure is called.
 *
 * A linked list contains all the clients blocked.
 */
static int
lodg_rsrv(q)
	register queue_t	*q;
{
	register struct lodg_lodg	*lodg;
	register struct backwq		*qptr;

	lodg = (struct lodg_lodg *)q->q_ptr;

	LODGTLI_PRINTF("lodg-%d: lodg: %x, lodg_rsrv\n",
		lodg-lodg_cnt, lodg);

	if (lodg->state == TS_IDLE && ((qptr = lodg->oqptr) !=
						(struct backwq *)NULL))  {
		while (qptr) {
			qenable(qptr->wrq);
			qptr = qptr->next;
		}
	}
	return (0);
}

/*
 * Write-side put procedure.
 */
lodg_wput(q, mp)
	register queue_t	*q;
	register mblk_t		*mp;
{
	register struct lodg_lodg	*lodg;

	lodg = (struct lodg_lodg *)q->q_ptr;

	switch (mp->b_datap->db_type) {
	case M_FLUSH:
		LODGTLI_PRINTF("lodg_wput-%d: M_FLUSH q: %x, mp: %x\n",
			lodg-lodg_cnt, q, mp);
		if (*mp->b_rptr & FLUSHW) {
			flushq(q, FLUSHALL);
			*mp->b_rptr &= ~FLUSHW;
		}
		if (*mp->b_rptr & FLUSHR) {
			/*
			 * N.B.: FLUSHW has been turned off by the
			 * time we get here.  (This is necessary to
			 * prevent a FLUSHW from circling back through
			 * the stream head to hit us a second time.)
			 */
			qreply(q, mp);
		} else	freemsg(mp);
		return (0);

	case M_PCPROTO:
		/*
		 * Priority request: validity check the message type,
		 * putting it onto the service queue if it's a legitimate
		 * tli message.
		 */
		LODGTLI_PRINTF("lodg_wput-%d: M_PCPROTO q: %x, mp: %x\n",
			lodg-lodg_cnt, q, mp);
		switch (((union T_primitives *)mp->b_rptr)->type) {
		case T_INFO_REQ:
			putq(q, mp);
			return (0);

		default:
			/* Unknown message type. */
			break;
		}

		break;

	case M_PROTO:
		LODGTLI_PRINTF("lodg_wput-%d: M_PROTO q: %x, mp: %x\n",
			lodg-lodg_cnt, q, mp);

		switch (((union T_primitives *)mp->b_rptr)->type) {
		case T_BIND_REQ:
		case T_UNBIND_REQ:
		case T_OPTMGMT_REQ:
			putq(q, mp);
			return (0);

		case T_UNITDATA_REQ: {
			register struct T_unitdata_req	*udata_req;
			register char			*addr;
			register int			addrlen;
			int				error;
			struct lodg_lodg		*olodg;

			LODGTLI_PRINTF(
				"lodg_wput-%d: T_UNITDATA_REQ q: %x, mp: %x\n",
				lodg-lodg_cnt, q, mp);

			if (lodg->flags & TT_BLOCKED) {
				putq(q, mp);
				return (0);
			}

			/*
			 * Handle directly
			 */
			error = chk_udata(mp);
			if (error > 0) {
				/*
				 * Let the service procedure
				 * generate the error.
				 */
				putq(q, mp);
				return (0);
			}
			if (error == -1) {
				freemsg(mp);
				return (0);
			}

			/*
			 * Lookup destination.
			 */
			udata_req = (struct T_unitdata_req *)mp->b_rptr;
			addr = (caddr_t)(mp->b_rptr + udata_req->DEST_offset);
			addrlen = udata_req->DEST_length;
			if ((olodg = lodg_findlink(addr, addrlen)) ==
				(struct lodg_lodg *)NULL) {
			LODGTLI_PRINTF("lodg_wput-%d: bad dest: %x\n",
					lodg-lodg_cnt, *(long *)addr);
				if (snd_UDERR(RD(q), udata_req,
							ECONNREFUSED) < 0)
					putq(q, mp);
				return (0);
			}
			if (olodg->state != TS_IDLE) {
				if (snd_UDERR(RD(q), udata_req,
							ECONNREFUSED) < 0)
					putq(q, mp);
				return (0);
			}

			if (canput(olodg->rdq->q_next) == 0)
				putq(q, mp);
			else	{
				if ((mp = mk_udata_ind(q, mp, olodg->rdq)) ==
						(mblk_t *)NULL)
					putq(q, mp);
				else	putnext(olodg->rdq, mp);
			}
			return (0);
		}

		default:
			/* Unknown message type. */
			break;
		}		/* union T_primitives */
		break;
	default:
		/* Unknown message type. */
		break;
	}			/* mp->b_datap->db_type */
	freemsg(mp);
	return (0);
}


/*
 * Write side service procedure. Handles the connection management
 * primitives.
 */
int
lodg_wsrv(q)
	register queue_t	*q;
{
	register struct lodg_lodg	*lodg;
	register mblk_t			*mp;
	register int			error;
	register int			size;

	lodg = (struct lodg_lodg *)q->q_ptr;

wgetnext:
	if ((mp = getq(q)) == (mblk_t *)NULL)
		return (0);

	switch (mp->b_datap->db_type) {
	case M_PCPROTO:
		switch (((union T_primitives *)mp->b_rptr)->type) {
		case T_INFO_REQ: {
			register struct T_info_ack	*info_ack;
			register mblk_t			*bp;

			LODGTLI_PRINTF(
				"lodg_wsrv-%d: T_INFO_REQ q: %x, mp: %x\n",
				lodg-lodg_cnt, q, mp);
			if ((bp = allocb(sizeof (struct T_info_ack),
					BPRI_MED)) == (mblk_t *)NULL) {
				_RECOVER(q, mp, sizeof (struct T_info_ack));
				return (0);
			}
			bp->b_datap->db_type = M_PCPROTO;
			bp->b_wptr += sizeof (struct T_info_ack);

			info_ack = (struct T_info_ack *)bp->b_rptr;
			info_ack->PRIM_type = T_INFO_ACK;
			info_ack->TSDU_size = strmsgsz;
			info_ack->ETSDU_size = -2;
			info_ack->CDATA_size = -2;
			info_ack->DDATA_size = -2;
			info_ack->ADDR_size = LODG_ADDRSZ;
			info_ack->OPT_size = LODG_OPTSZ;
			info_ack->TIDU_size = LODG_TIDUSZ;
			info_ack->SERV_type = T_CLTS;
			info_ack->CURRENT_state = lodg->state;

			qreply(q, bp);
			freemsg(mp);
			goto wgetnext;
		}

		default:
			freemsg(mp);
			goto wgetnext;
		}

	case M_PROTO:
		switch (((union T_primitives *)mp->b_rptr)->type) {
		case T_BIND_REQ: {
			register struct T_bind_req	*bind_req;
			register struct T_bind_ack	*bind_ack;
			register int			addrlen;
			register char			*addr;
			mblk_t				*bp;

			LODGTLI_PRINTF(
				"lodg_wsrv-%d: T_BIND_REQ q: %x, mp: %x\n",
				lodg-lodg_cnt, q, mp);
			/*
			 * Check state.
			 */
			if (lodg->state != TS_UNBND) {
				if (snd_ERRACK(RD(q), mp, T_BIND_REQ,
						TOUTSTATE, 0) < 0) {
				_RECOVER(q, mp, sizeof (struct T_error_ack));
					return (0);
				}
			}

			bind_req = (struct T_bind_req *)mp->b_rptr;

			/*
			 * Check args and address requested.
			 */
			if ((error = chk_bind(mp)) != 0) {
				/*
				 * The requested address is taken, so
				 * arrange for another to be allocated.
				 */
				bind_req->ADDR_length = 0;
			}

			/*
			 * At this point make sure we can allocate
			 * the reply message.
			 */
			size = sizeof (struct T_bind_ack) +
	(bind_req->ADDR_length == 0 ? sizeof (gport) : bind_req->ADDR_length);
			if ((bp = allocb(size, BPRI_MED)) == (mblk_t *)NULL) {
				_RECOVER(q, mp, size);
				return (0);
			}

			if (bind_req->ADDR_length == 0) {
				/*
				 * We have to choose an address.
				 */
				do	{
					if (++gport == LODG_MAXADDR)
						gport = 0;
				} while (lodg_findlink((caddr_t)&gport,
					sizeof (gport)));
				LODGTLI_PRINTF(
					"lodg_wsrv-%d: Chose address %d\n",
					lodg-lodg_cnt, gport);
				addr = (caddr_t)&gport;
				addrlen = sizeof (gport);
			} else	{
				addr = (caddr_t)(mp->b_rptr +
						bind_req->ADDR_offset);
				addrlen = bind_req->ADDR_length;
			}

			/*
			 * Get address buffer.
			 */
			if ((lodg->laddr = (caddr_t)kmem_alloc(addrlen)) ==
							(char *)NULL) {
				_RECOVER(q, mp, addrlen);
				freemsg(bp);
				return (0);
			} else	{
				(void)bcopy(addr, lodg->laddr, addrlen);
				lodg->laddrlen = addrlen;

				/*
				 * Add it to the list of bound
				 * endpoints.
				 */
				lodg_addlink(lodg);

				lodg->state = TS_IDLE;
			}

			/*
			 * Send reply.
			 */
			bp->b_datap->db_type = M_PCPROTO;
			bp->b_wptr += size;
			bind_ack = (struct T_bind_ack *)bp->b_rptr;
			bind_ack->PRIM_type = T_BIND_ACK;
			bind_ack->CONIND_number = 0;
			bind_ack->ADDR_offset = sizeof (struct T_bind_ack);
			bind_ack->ADDR_length = addrlen;
			(void)bcopy(addr,
				(caddr_t)(bp->b_rptr + bind_ack->ADDR_offset),
				addrlen);
			qreply(q, bp);

			freemsg(mp);
			goto wgetnext;
		}

		case T_UNBIND_REQ: {
			register mblk_t		*bp;

			LODGTLI_PRINTF(
				"lodg_wsrv-%d: T_UNBIND_REQ q: %x, mp: %x\n",
				lodg-lodg_cnt, q, mp);
			/*
			 * Make sure we can allocate the reply message.
			 */
			size = max(sizeof (struct T_error_ack),
					sizeof (struct T_ok_ack));
			if ((bp = allocb(size, BPRI_MED)) == (mblk_t *)NULL) {
				_RECOVER(q, mp, size);
				return (0);
			}

			/*
			 * Check state.
			 */
			if (lodg->state != TS_IDLE) {
				(void)snd_ERRACK(RD(q), bp, T_UNBIND_REQ,
						TOUTSTATE, 0);
			} else	{
				lodg->state = TS_UNBND;
				(void)kmem_free(lodg->laddr, lodg->laddrlen);
				lodg->laddr = (char *)NULL;
				lodg->laddrlen = 0;

				/*
				 * Delete from the linked list.
				 */
				lodg_dellink(lodg);

				(void)snd_OKACK(RD(q), bp, T_UNBIND_REQ);
			}

			/*
			 * Send up an M_FLUSH.
			 */
			mp->b_datap->db_type = M_FLUSH;
			mp->b_wptr = mp->b_rptr = mp->b_datap->db_base;
			*mp->b_wptr++ = FLUSHRW;
			qreply(q, mp);

			goto wgetnext;
		}

		case T_OPTMGMT_REQ:
			LODGTLI_PRINTF(
				"lodg_wsrv-%d: T_OPTMGMT_REQ q: %x, mp: %x\n",
				lodg-lodg_cnt, q, mp);

			if (lodg_optmgmt(q, mp) < 0)
				return (0);

			goto wgetnext;

		case T_UNITDATA_REQ: {
			register char			*addr;
			register int			addrlen;
			register struct T_unitdata_req	*udata_req;
			register struct lodg_lodg	*olodg;

			LODGTLI_PRINTF(
			"lodg_wsrv-%d: T_UNITDATA_REQ q: %x, mp: %x\n",
			lodg-lodg_cnt, q, mp);

			if (lodg->flags & TT_BLOCKED) {
				/*
				 * Must have just been enabled.
				 */
				lodg->flags &= ~TT_BLOCKED;
				lodg->dest_lodg = (struct lodg_lodg *)NULL;
			}

			/*
			 * Handle directly
			 */
			udata_req = (struct T_unitdata_req *)mp->b_rptr;
			error = chk_udata(mp);
			if (error > 0) {
				if (snd_UDERR(RD(q), udata_req, error) < 0) {
				_RECOVER(q, mp, sizeof (struct T_uderror_ind));
					return (0);
				}
				goto wgetnext;
			}
			if (error == -1) {
				freemsg(mp);
				goto wgetnext;
			}

			/*
			 * Lookup destination.
			 */
			addr = (caddr_t)(mp->b_rptr + udata_req->DEST_offset);
			addrlen = udata_req->DEST_length;
			if ((olodg = lodg_findlink(addr, addrlen)) ==
				(struct lodg_lodg *)NULL) {
				if (snd_UDERR(RD(q), udata_req,
						ECONNREFUSED) < 0) {
				_RECOVER(q, mp, sizeof (struct T_uderror_ind));
					return (0);
				}
				goto wgetnext;
			}

			if (olodg->state != TS_IDLE) {
				if (snd_UDERR(RD(q), udata_req,
						ECONNREFUSED) < 0) {
				_RECOVER(q, mp, sizeof (struct T_uderror_ind));
					return (0);
				}
			}

			if (canput(olodg->rdq->q_next) == 0) {
				/*
				 * Facilitate this write sevice
				 * procedure to be enabled when
				 * flow control permits.
				 */
				add_backq(q, mp, olodg);
				return (0);
			} else	{
				if ((mp = mk_udata_ind(q, mp, olodg->rdq)) ==
						(mblk_t *)NULL)
					return (0);

				putnext(olodg->rdq, mp);
			}
			goto wgetnext;
		}

		default:
			break;
		}			/* PRIM_type */
	default:
		break;
	}				/* bp->b_datap->db_type */
	freemsg(mp);
	goto wgetnext;
}

/*
 * Check T_BIND_REQ for correctness - validate address requested.
 *	Returns:
 *		0 	On success.
 *		> 0	On failure (T_ERROR_ACK to be sent).
 */
static int
chk_bind(mp)
	register mblk_t			*mp;
{
	register struct T_bind_req	*bind_req;
	register char			*addr;
	register int			addrlen;
	register int			error;

	bind_req = (struct T_bind_req *)mp->b_rptr;

	error = 0;
	if (MSGBLEN(mp) < sizeof (*bind_req) ||
		MSGBLEN(mp) < (bind_req->ADDR_offset + bind_req->ADDR_length))
		error = TBADADDR;
	else
	if (bind_req->ADDR_offset < 0 || bind_req->ADDR_length > LODG_ADDRSZ)
		error = TBADADDR;
	else	{
		/*
		 * Validate address requested.
		 */
		addrlen = bind_req->ADDR_length;
		addr = (caddr_t)(mp->b_rptr + bind_req->ADDR_offset);
		if (addrlen > 0) {
			if (lodg_findlink(addr, addrlen) !=
					(struct lodg_lodg *)NULL)
				error = TNOADDR;
		}
	}
	return (error);
}

/*
 * Check T_UNITDATA_REQ for correctness.
 *	Returns:
 *		0	on sucess
 *		> 0	on error generate T_UDERROR_IND
 *		-1	on error don't generate T_UDERROR_IND.
 */
static int
chk_udata(mp)
	register mblk_t			*mp;
{
	register struct T_unitdata_req	*udata_req;
	register int			addroffset;
	register int			addrlen;
	register int			optoffset;
	register int			optlen;
	register int			msgsz;
	int				error;

	msgsz = mp->b_wptr - mp->b_rptr;
	if (msgsz < sizeof (*udata_req))
		return (-1);

	udata_req = (struct T_unitdata_req *)mp->b_rptr;
	addrlen = udata_req->DEST_length;
	addroffset = udata_req->DEST_offset;
	optlen = udata_req->OPT_length;
	optoffset = udata_req->OPT_offset;

	error = 0;
	if (((addrlen > 0) && (addrlen + addroffset) > msgsz) ||
		((optlen > 0) && (optlen + optoffset) > msgsz))
		error = -1;
	else
	if (addrlen == 0)
		error = TNOADDR;
	else
	if (addrlen > LODG_ADDRSZ)
		error = TBADADDR;
	else
	if (optlen > LODG_OPTSZ)
		error = TBADOPT;
	return (error);
}

static struct lodg_lodg *
lodg_findlink(addr, len)
	register char 			*addr;
	register int			len;
{
	register struct lodg_lodg	*lodg;


	for (lodg = lodg_listhead; lodg != (struct lodg_lodg *)NULL;
				lodg = lodg->lodg_link.next) {
		if (len == lodg->laddrlen &&
			bcmp(addr, (caddr_t)lodg->laddr, len) == 0) {
			return (lodg);
		}
	}
	return ((struct lodg_lodg *)NULL);
}

static void
lodg_addlink(lodg)
	register struct lodg_lodg	*lodg;
{
	lodg->lodg_link.next = lodg_listhead;
	lodg->lodg_link.prev = (struct lodg_lodg *)NULL;
	if (lodg_listhead)
		lodg_listhead->lodg_link.prev = lodg;
	lodg_listhead = lodg;

}

static void
lodg_dellink(lodg)
	register struct lodg_lodg	*lodg;
{
	register struct lodg_lodg	*olodg;

	if ((olodg = lodg->lodg_link.next) != (struct lodg_lodg *)NULL)
		olodg->lodg_link.prev = lodg->lodg_link.prev;

	if ((olodg = lodg->lodg_link.prev) != (struct lodg_lodg *)NULL)
		olodg->lodg_link.next = lodg->lodg_link.next;
	else	lodg_listhead = lodg->lodg_link.next;
}

/*
 * Send error acknowledgment.
 */
static int
snd_ERRACK(q, mp, prim, terr, serr)
	register queue_t		*q;
	register mblk_t			*mp;
	register int			serr;
	register int			prim;
{
	register struct T_error_ack	*tea;


	if (MBLKLEN(mp) < sizeof (*tea))
		return (-1);

	mp->b_rptr = mp->b_wptr = mp->b_datap->db_base;
	mp->b_wptr += sizeof (struct T_error_ack);
	mp->b_datap->db_type = M_PCPROTO;

	tea = (struct T_error_ack *)mp->b_rptr;
	tea->ERROR_prim = prim;
	tea->PRIM_type = T_ERROR_ACK;
	tea->TLI_error = terr;
	tea->UNIX_error = serr;

	putnext(q, mp);
	return (0);
}

/*
 * Send OK acknowledgment.
 */
static int
snd_OKACK(q, mp, prim)
	register queue_t		*q;
	register mblk_t			*mp;
	register int			prim;
{
	register struct T_ok_ack	*ok_ack;

	if (MBLKLEN(mp) < sizeof (*ok_ack))
		return (-1);

	mp->b_rptr = mp->b_wptr = mp->b_datap->db_base;
	mp->b_datap->db_type = M_PCPROTO;
	mp->b_wptr += sizeof (struct T_ok_ack);

	ok_ack = (struct T_ok_ack *)mp->b_rptr;
	ok_ack->CORRECT_prim = prim;
	ok_ack->PRIM_type = T_OK_ACK;

	putnext(q, mp);
	return (0);
}

static int
snd_UDERR(q, udata_req, err)
	register queue_t		*q;
	register struct T_unitdata_req	*udata_req;
	register int			err;
{
	register struct T_uderror_ind	*uderr_ind;
	register mblk_t			*bp;
	int				size;

	LODGTLI_PRINTF("snd_UDERR: err %d\n", err);

	size = sizeof (*uderr_ind);
	if (udata_req->DEST_length > 0)
		size += udata_req->DEST_length;
	if (udata_req->OPT_length > 0)
		size += udata_req->OPT_length;
	if ((bp = allocb(size, BPRI_MED)) == (mblk_t *)NULL)
		return (-1);

	bp->b_datap->db_type = M_PROTO;
	bp->b_wptr += size;

	uderr_ind = (struct T_uderror_ind *)bp->b_rptr;
	uderr_ind->PRIM_type = T_UDERROR_IND;
	uderr_ind->DEST_length = udata_req->DEST_length;
	if (uderr_ind->DEST_length > 0) {
		uderr_ind->DEST_offset = sizeof (*uderr_ind);
		(void)bcopy((caddr_t)udata_req + udata_req->DEST_offset,
				(caddr_t)(bp->b_rptr + uderr_ind->DEST_offset),
				(size_t)udata_req->DEST_length);
	}
	uderr_ind->OPT_length = udata_req->OPT_length;
	if (uderr_ind->OPT_length > 0) {
		uderr_ind->OPT_offset = sizeof (*uderr_ind) +
					uderr_ind->DEST_offset;
		(void)bcopy((caddr_t)udata_req + udata_req->OPT_offset,
				(caddr_t)(bp->b_rptr + uderr_ind->OPT_offset),
				(size_t)udata_req->OPT_length);
	}
	uderr_ind->ERROR_type = err;

	putnext(q, bp);
	return (0);
}

static void
add_backq(q, mp, olodg)
	register struct lodg_lodg	*olodg;
	register queue_t		*q;
	register mblk_t			*mp;
{
	register struct backwq		*mycell;
	register struct lodg_lodg	*lodg;

	lodg = (struct lodg_lodg *)q->q_ptr;

	/*
	 * Allocate a cell.
	 */
	if ((mycell = (struct backwq *)kmem_alloc(sizeof (*mycell))) ==
						(struct backwq *)NULL) {
		_RECOVER(q, mp, sizeof (*mycell));
		return;
	}
	mycell->wrq = q;

	/*
	 * The list is ordered the wrong way, but what the
	 * hell!
	 */
	mycell->next = olodg->oqptr;
	olodg->oqptr = mycell;

	lodg->flags |= TT_BLOCKED;
	lodg->dest_lodg = olodg;
}

static void
del_backq(q, lodg)
	register queue_t		*q;
	register struct lodg_lodg	*lodg;
{
	register struct backwq		*wrq_link;
	register struct backwq		*owrq_link;

	LODGTLI_PRINTF("del_backq: q: %x\n", q);

	wrq_link = owrq_link = lodg->oqptr;
	while (wrq_link && wrq_link->wrq != q) {
		owrq_link = wrq_link;
		wrq_link = wrq_link->next;
	}

	if (wrq_link == (struct backwq *)NULL) {
		/*
		 * Not found.
		 */
		LODGTLI_PRINTF("del_backq: q: %x Not found\n", q);
	} else	{
		LODGTLI_PRINTF("del_backq: q: %x Found\n", q);
		owrq_link->next = wrq_link->next;
		(void)kmem_free((caddr_t)wrq_link, sizeof (struct backwq));
	}
}

static mblk_t *
mk_udata_ind(q, mp, oq)
	register queue_t	*q;
	register queue_t	*oq;
	register mblk_t		*mp;

{
	struct lodg_lodg	*lodg;
	struct lodg_lodg	*olodg;
	mblk_t			*bp;
	struct T_unitdata_ind	*udata_ind;
	int			size;
	int			optlen;

	LODGTLI_PRINTF("mk_udata_ind: Entered\n");

	lodg = (struct lodg_lodg *)q->q_ptr;
	olodg = (struct lodg_lodg *)oq->q_ptr;

	optlen = 0;
	if (olodg->idflag != 0) {
		LODGTLI_PRINTF("mk_udata_ind: idflag %x\n", olodg->idflag);
		if (olodg->idflag & OPT_UID)
			optlen += sizeof (opthdr_t) + sizeof (struct opt_uid);

		if (olodg->idflag & OPT_GID)
			optlen += sizeof (opthdr_t) + sizeof (struct opt_gid);

		if (olodg->idflag & OPT_RUID)
			optlen += sizeof (opthdr_t) + sizeof (struct opt_ruid);

		if (olodg->idflag & OPT_RGID)
			optlen += sizeof (opthdr_t) + sizeof (struct opt_rgid);
	}

	LODGTLI_PRINTF("mk_udata_ind: optlen %d\n", optlen);

	size = sizeof (*udata_ind) + lodg->laddrlen + optlen;
	if ((bp = allocb(size, BPRI_MED)) == (mblk_t *)NULL) {
		_RECOVER(q, mp, size);
		return ((mblk_t *)NULL);
	}
	udata_ind = (struct T_unitdata_ind *)bp->b_rptr;
	udata_ind->PRIM_type = T_UNITDATA_IND;

	udata_ind->SRC_length = lodg->laddrlen;
	udata_ind->SRC_offset = sizeof (*udata_ind);

	LODGTLI_PRINTF("mk_udata_ind: laddr %d, laddrlen %d, optlen %d\n",
		*(int *)lodg->laddr, lodg->laddrlen, optlen);

	(void)bcopy(lodg->laddr, (caddr_t)(bp->b_rptr + udata_ind->SRC_offset),
		lodg->laddrlen);

	LODGTLI_PRINTF("mk_udata_ind: idflag %d\n", olodg->idflag);

	if (olodg->idflag == 0) {
		udata_ind->OPT_length = 0;
		udata_ind->OPT_offset = 0;
	} else	{
		udata_ind->OPT_offset = udata_ind->SRC_offset +
			udata_ind->SRC_length;
		udata_ind->OPT_length = optlen;

		lodg_fill_opt(lodg, olodg, (char *)(bp->b_rptr +
				udata_ind->OPT_offset));
	}

	bp->b_datap->db_type = M_PROTO;
	bp->b_wptr += size;
	bp->b_cont = mp->b_cont;

	mp->b_cont = NULL;
	freeb(mp);
	return (bp);
}

static int
lodg_optmgmt(q, mp)
	register queue_t	*q;
	register mblk_t		*mp;
{
	register struct lodg_lodg	*lodg;
	int				size;
	struct T_optmgmt_req		*optmgmt_req;
	struct T_optmgmt_ack		*optmgmt_ack;

	lodg = (struct lodg_lodg *)q->q_ptr;
	optmgmt_req = (struct T_optmgmt_req *)mp->b_rptr;

	/*
	 * Validate the request
	 */
	size = MSGBLEN(mp);
	if (size < sizeof (*optmgmt_req)) {
		if (snd_ERRACK(RD(q), mp, T_BIND_REQ, TSYSERR, EINVAL) < 0) {
			_RECOVER(q, mp, sizeof (struct T_error_ack));
			return (-1);
		}
		return (0);
	}

	switch (optmgmt_req->MGMT_flags) {
	default:
		if (snd_ERRACK(RD(q), mp, T_OPTMGMT_REQ, TBADFLAG, 0) < 0) {
			_RECOVER(q, mp, sizeof (struct T_error_ack));
			return (-1);
		}
		return (0);

	case T_CHECK: {
		int	fail;

		/*
		 * Check option requested.
		 */
		LODGTLI_PRINTF("lodg_optmgmt: T_CHECK\n");

		if ((fail = option_check(mp)) < 0) {
			if (snd_ERRACK(RD(q), mp, T_OPTMGMT_REQ, TBADOPT,
								0) < 0) {
				_RECOVER(q, mp, sizeof (struct T_error_ack));
				return (-1);
			}
			return (0);
		}

		/*
		 * In this case, the size of the ack
		 * is the same as the size of the req.
		 */
		mp->b_datap->db_type = M_PCPROTO;
		optmgmt_ack = (struct T_optmgmt_ack *)mp->b_rptr;
		optmgmt_ack->PRIM_type = T_OPTMGMT_ACK;
		if (fail == 0)
			optmgmt_ack->MGMT_flags |= T_SUCCESS;
		else	optmgmt_ack->MGMT_flags |= T_FAILURE;

		(void)qreply(q, mp);

		return (0);
	}

	case T_DEFAULT:
		LODGTLI_PRINTF("lodg_optmgmt: T_DEFAULT\n");

		return (do_default_opts(q, mp));

	case T_NEGOTIATE: {
		union opt_req	*opt_req;
		opthdr_t	*opthdr;
		int		fail;

		/*
		 * Negotiate option.
		 */
		LODGTLI_PRINTF("lodg_optmgmt: T_NEGOTIATE\n");

		if (optmgmt_req->OPT_length == 0)
			return (do_default_opts(q, mp));

		fail = option_check(mp);
		if (fail < 0 || fail == 1) {
			if (snd_ERRACK(RD(q), mp, T_OPTMGMT_REQ, TBADOPT,
								0) < 0) {
				_RECOVER(q, mp, sizeof (struct T_error_ack));
				return (-1);
			}
			return (0);
		}

		mp->b_datap->db_type = M_PCPROTO;
		optmgmt_ack = (struct T_optmgmt_ack *)mp->b_rptr;
		optmgmt_ack->PRIM_type = T_OPTMGMT_ACK;

		opthdr = (opthdr_t *)(mp->b_rptr + optmgmt_req->OPT_offset);
		opt_req = (union opt_req *)((caddr_t)opthdr +
						opthdr->opt_offset);

		switch (opt_req->type) {
			case SETID:
				LODGTLI_PRINTF("lodg_optmgmt: SETID\n");
				lodg->idflag |= opt_req->setid.id;
				break;
	
			case GETID:
				LODGTLI_PRINTF("lodg_optmgmt: GETID\n");
				opt_req->getid.id = lodg->idflag;
				break;
	
			case OPT_UID:
				LODGTLI_PRINTF("lodg_optmgmt: OPT_UID\n");
				opt_req->uid.uid = lodg->cr_uid;
				break;
	
			case OPT_GID:
				LODGTLI_PRINTF("lodg_optmgmt: OPT_GID\n");
				opt_req->gid.gid = lodg->cr_gid;
				break;
	
			case OPT_RUID:
				LODGTLI_PRINTF("lodg_optmgmt: OPT_RUID\n");
				opt_req->ruid.ruid = lodg->cr_ruid;
				break;
	
			case OPT_RGID:
				LODGTLI_PRINTF("lodg_optmgmt: OPT_RGID\n");
				opt_req->rgid.rgid = lodg->cr_rgid;
				break;
		}
		qreply(q, mp);
		return (0);
	}
	}
}

/*
 * Insert options.
 */
static void
lodg_fill_opt(lodg, olodg, opt_buf)
	struct lodg_lodg		*lodg;
	struct lodg_lodg		*olodg;
	char				*opt_buf;
{
	opthdr_t			*opthdr;
	opthdr_t			*oopthdr;

	LODGTLI_PRINTF("lodg_fill_opt: Entered\n");
	if (olodg->idflag & OPT_UID) {
		struct opt_uid		*uid;

		LODGTLI_PRINTF("lodg_fill_opt: UID requested\n");
		opthdr = (opthdr_t *)opt_buf;
		opthdr->opt_offset = sizeof (opthdr_t);
		opthdr->next_hdr_offset = sizeof (opthdr_t) + sizeof (*uid);

		uid = (struct opt_uid *)(opt_buf + opthdr->opt_offset);
		uid->type = OPT_UID;
		uid->uid = lodg->cr_uid;

		oopthdr = opthdr;
	}

	if (olodg->idflag & OPT_GID) {
		struct opt_gid		*gid;

		LODGTLI_PRINTF("lodg_fill_opt: GID requested\n");
		opthdr = (opthdr_t *)(opt_buf + oopthdr->next_hdr_offset);
		opthdr->opt_offset = oopthdr->next_hdr_offset +
					sizeof (*opthdr);
		opthdr->next_hdr_offset = opthdr->opt_offset + sizeof (*gid);

		gid = (struct opt_gid *)(opt_buf + opthdr->opt_offset);
		gid->type = OPT_GID;
		gid->gid = lodg->cr_gid;

		oopthdr = opthdr;
	}

	if (olodg->idflag & OPT_RUID) {
		struct opt_ruid		*ruid;

		LODGTLI_PRINTF("lodg_fill_opt: RUID requested\n");
		opthdr = (opthdr_t *)(opt_buf + oopthdr->next_hdr_offset);
		opthdr->opt_offset = oopthdr->next_hdr_offset +
					sizeof (*opthdr);
		opthdr->next_hdr_offset = opthdr->opt_offset + sizeof (*ruid);

		ruid = (struct opt_ruid *)(opt_buf + opthdr->opt_offset);
		ruid->type = OPT_RUID;
		ruid->ruid = lodg->cr_ruid;

		oopthdr = opthdr;
	}

	if (olodg->idflag & OPT_RGID) {
		struct opt_rgid		*rgid;

		LODGTLI_PRINTF("lodg_fill_opt: RGID requested\n");
		opthdr = (opthdr_t *)(opt_buf + oopthdr->next_hdr_offset);
		opthdr->opt_offset = oopthdr->next_hdr_offset +
					sizeof (*opthdr);
		opthdr->next_hdr_offset = opthdr->opt_offset + sizeof (*rgid);

		rgid = (struct opt_rgid *)(opt_buf + opthdr->opt_offset);
		rgid->type = OPT_RGID;
		rgid->rgid = lodg->cr_rgid;

		oopthdr = opthdr;
	}

	/*
	 * Terminate the list.
	 */
	oopthdr->next_hdr_offset = 0;
}

/*
 * Return:
 *	-1	if bad format,
 *	 1	if format good, but unrecognized type or bad value.
 *	 0 	if all good.
 */
static int
option_check(mp)
	register mblk_t	*mp;
{
	struct T_optmgmt_req	*optmgmt_req;
	int			opt_len;
	int			opt_offset;
	int			size;
	opthdr_t		*opthdr;
	union opt_req		*opt;

	optmgmt_req = (struct T_optmgmt_req *)mp->b_rptr;

	LODGTLI_PRINTF("option_check: Entered\n");
	size = MSGBLEN(mp);
	opt_len = optmgmt_req->OPT_length;
	opt_offset = optmgmt_req->OPT_offset;

	if (opt_len < 0 || opt_len > LODG_OPTSZ ||  opt_offset < 0 ||
				(opt_offset + opt_len) > size) {
		LODGTLI_PRINTF("option_check: 1\n");
		return (-1);
	}

	opthdr = (opthdr_t *)(mp->b_rptr + opt_offset);

	/*
	 * Check the option header for alignment.
	 */
	if ((int)opthdr % NBPW) {
		LODGTLI_PRINTF("option_check: 2\n");
		return (-1);
	}

	if (opt_len <= sizeof (opthdr_t) || opt_len < (opthdr->opt_offset +
					sizeof (long))) {
		LODGTLI_PRINTF("option_check: 3\n");
		return (-1);
	}

	opt = (union opt_req *)((caddr_t)opthdr + opthdr->opt_offset);

	/*
	 * Check the option for alignment.
	 */
	if ((int)opt % NBPW) {
		LODGTLI_PRINTF("option_check: 4\n");
		return (-1);
	}

	switch (opt->type) {
	case SETID:
		LODGTLI_PRINTF("option_check: SETID specified\n");
		if (opt_len < (opthdr->opt_offset + sizeof (struct opt_setid))) {
			LODGTLI_PRINTF("option_check: 5\n");
			return (-1);
		}

		if ((opt->setid.id & ~(OPT_UID|OPT_GID|OPT_RUID|OPT_RGID)) != 0) {
			LODGTLI_PRINTF("lodg_optmgmt: T_NEGOTIATE\n");
			return (1);
		}
		break;

	case GETID:
		LODGTLI_PRINTF("option_check: GETID specified\n");
		if (opt_len < (opthdr->opt_offset + sizeof (struct opt_getid)))
			return (-1);
		break;

	case OPT_UID:
		LODGTLI_PRINTF("option_check: UID specified\n");
		if (opt_len < (opthdr->opt_offset + sizeof (struct opt_uid)))
			return (-1);
		break;

	case OPT_GID:
		LODGTLI_PRINTF("option_check: GID specified\n");
		if (opt_len < (opthdr->opt_offset + sizeof (struct opt_gid)))
			return (-1);
		break;

	case OPT_RUID:
		LODGTLI_PRINTF("option_check: RUID specified\n");
		if (opt_len < (opthdr->opt_offset + sizeof (struct opt_ruid)))
			return (-1);
		break;

	case OPT_RGID:
		LODGTLI_PRINTF("option_check: RGID specified\n");
		if (opt_len < (opthdr->opt_offset + sizeof (struct opt_rgid)))
			return (-1);
		break;

	default:
		LODGTLI_PRINTF("option_check: Invalid option\n");
		return (1);
	}
	return (0);
}

static int
do_default_opts(q, mp)
	register queue_t	*q;
	register mblk_t		*mp;
{
	struct opt_noop		*optnop;
	opthdr_t		*opthdr;
	int			size;
	struct T_optmgmt_ack	*optmgmt_ack;
	mblk_t			*bp;

	/*
	 * Return default option
	 */
	LODGTLI_PRINTF("do_default_opts: Entered\n");
	size = sizeof (*optmgmt_ack) + sizeof (opthdr_t) +
			sizeof (struct opt_noop);

	if ((bp = allocb(size, BPRI_HI)) == NULL) {
		_RECOVER(q, mp, size);
		return (-1);
	}

	bp->b_datap->db_type = M_PCPROTO;
	bp->b_wptr = bp->b_rptr + size;
	optmgmt_ack = (struct T_optmgmt_ack *)bp->b_rptr;
	optmgmt_ack->PRIM_type = T_OPTMGMT_ACK;
	optmgmt_ack->MGMT_flags = T_DEFAULT;
	optmgmt_ack->OPT_length = sizeof (opthdr_t) +
				sizeof (struct opt_noop);
	optmgmt_ack->OPT_offset = sizeof (*optmgmt_ack);

	opthdr = (opthdr_t *)(bp->b_rptr + optmgmt_ack->OPT_offset);
	opthdr->opt_offset = sizeof (opthdr_t);
	opthdr->next_hdr_offset = 0;

	optnop = (struct opt_noop *)((caddr_t)opthdr + opthdr->opt_offset);
	optnop->type = NOOP;	/* default opt */

	freemsg(mp);
	qreply(q, bp);

	return (0);
}
#endif	NLODGTLI > 0
