#ifndef lint
static  char sccsid[] = "@(#)udp_tli.c 1.10 90/08/29 SMI";
#endif

/*
 * Copyright (c) 1988 by Sun Microsystems, Inc.
 */

/*
 * Streams interface to the socket-based UDP implementation.
 * Compatible with the AT&T Transport Level Interface.
 *
 * General notes:
 * 1)	This module's notion of an address is a struct sockaddr_in.
 *	However, at one point, AT&T defined a "short" address format
 *	as the std TLI addressing format known as sockaddr_in:
 *		struct sockaddr_in == (struct sockaddr_in) - (char zero[8])
 *	which might exist in some streams TCP implementations. Both
 *	formats are supported here.
 *
 * 2)   UDPTLI_PRINTF debug statements are switched by the global
 *	variable udptli_debug.
 */

#include "udptli.h"

#if	NUDPTLI > 0

#define	TLIDEBUG

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/debug.h>
#include <sys/user.h>
#include <sys/file.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/wait.h>
#include <sys/vnode.h>
#include <sys/protosw.h>

#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_pcb.h>
#include <netinet/ip_var.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>

#include <nettli/tihdr.h>
#include <nettli/tiuser.h>
#include <nettli/udp_tli.h>
#include <nettli/udp_tlivar.h>

#ifdef DYNAMIC
/******************************************************************************//*
 * Wrapper code to facilitate making this driver a loadabe
 * module.
 */
#include        "sys/conf.h"
#include        "sun/autoconf.h"
#include        "sun/vddrv.h"

extern int              nodev();
extern struct streamtab udptli_info;
extern int              udptli_ttpcount;
extern int		udptli_debug;

struct cdevsw udptli_cdev = {
        nodev, nodev, nodev, nodev,
        nodev, nodev, nodev, 0,
        &udptli_info, 0
};
 
struct vdldrv vd = {
        VDMAGIC_PSEUDO,
        "udptlidrv",
#ifndef sun4c
	NULL,
	NULL,
	NULL,
	0,
	0,
#else sun4c
	NULL,
#endif sun4c
        NULL,
        &udptli_cdev,
        0,
        0
};
 
udptli_init(function_code, vdp, vdi, vds)
        unsigned int    function_code;
        struct vddrv    *vdp;
        addr_t          vdi;
        struct  vdstat  *vds;
{
        switch (function_code) {
                case VDLOAD:
                        printf("udptli driver loaded\n");
                        vdp->vdd_vdtab = (struct vdlinkage *)&vd;
                        return (0);

                case VDUNLOAD:
                        printf("udptli driver unloaded\n");
                        if (udptli_ttpcount == 0)
                                return (0);
                        else    return (-1);

                case VDSTAT:
                        UDPTLI_PRINTF(
                        "udptli: VDSTAT: udptli_ttpcount %d\n",
				udptli_ttpcount);
                        return (0);
        }
}

/******************************************************************************/
#endif DYNAMIC

static void	udptli_t_info_req();
static void	udptli_t_bind_req();
static void	udptli_t_unbind_req();
static void	udptli_t_optmgmt_req();
static void	udptli_t_unitdata_req();
static void	udptli_ioctl();

/*
 * Stereotyped streams glue.
 */
int	udptli_open();
int	udptli_close();
int	udptli_wput();
int	udptli_wsrv();

struct module_info	udptli_minfo = {
	0,			/* module id XXX: define real value */
	"udptli",
	0,
	INFPSZ,			/* max msg size */
	6144,			/* high water: 6 x RFS's */
	STRLOW			/* low water XXX: make reasonable */
};

struct qinit	udptli_rinit = {
	NULL,
	NULL,
	udptli_open,
	udptli_close,
	NULL,
	&udptli_minfo,
	NULL
};

/*
 * Write-side queue info.
 *
 * Although the driver has the equivalent of a write-side service procedure,
 * it's not visible to the streams framework.  The auxiliary process is
 * responsible for processing entries on the write queue.
 *
 * udptli_wsrv() does not really exist. The write-service routine
 * is handled in the auxiliary process. The existence of the
 * entry point declaration is to fake the streams code such that
 * flow control will be checked by canput().
 */

struct qinit	udptli_winit = {
	udptli_wput,
	udptli_wsrv,
	udptli_open,
	udptli_close,
	NULL,
	&udptli_minfo,
	NULL
};

struct streamtab	udptli_info = {
	&udptli_rinit,
	&udptli_winit,
	NULL,
	NULL
};

/*
 * Global variables:
 *
 * 1) Per-device instance state information.
 * This private data structure plays the key role in this
 * UDPTLI implementation.
 *
 * 3) Counter of total opened ttp
 *
 * 4) Head pointer to the pending ttp-queue
 *
 */
struct	tt_softc	tt_udpsoftc[NUDPTLI];
int			udptli_ttpcount = 0;	/* opened ttp counter */
int			udptli_debug = 0;
/* head ptr to pending ttp list */
struct	tt_softc	*udptli_ttp = (struct tt_softc *)NULL;
/* aux proc ptr */
struct	proc		*udptli_auxprocp = (struct proc *)NULL;


/* ARGSUSED */
int
udptli_open(q, dev, oflag, sflag)
	struct queue	*q;
	dev_t		dev;
	int		oflag;
	int		sflag;
{
	register int			unit;
	register struct tt_softc	*ttp;
	struct tt_sockinfo		*tsp;
	struct tt_sockinfo		*udptli_mksockinfo();
	struct socket			*so;
	int 				udptli_wakeup();
	extern int 			udptli_main();

	UDPTLI_PRINTF(
		"udptli_open:q: %x, WR(q): %x, dev: %x, oflag: %x, sflag: %x\n",
		q, WR(q), dev, oflag, sflag);

	/*
	 * Get our minor device number, verifying while we're at it
	 * that nobody's trying to open us as a module (as opposed
	 * to a device).
	 */
	if (sflag == CLONEOPEN) {
		/*
		 * Find an unused minor device instance.
		 * XXX -- Skip returning instance 0, until
		 * strclose() code is fixed
		 */
		for (unit = 1 /* XXX -- 0 */; unit < NUDPTLI; unit++)
			if (!(tt_udpsoftc[unit].tt_flags &
					(TT_OPEN | TT_CLOSE)))
				break;
	} else if (sflag)
		return (OPENFAIL);
	else	unit = minor(dev);

	if (unit < 0 || unit >= NUDPTLI) {
		u.u_error = ENXIO;
		return (OPENFAIL);
	}

	ttp = &tt_udpsoftc[unit];
	UDPTLI_PRINTF("udptli_open: ttp: %x tt_udpsoftc: %x\n",
				ttp, tt_udpsoftc);
	if (ttp->tt_flags & TT_OPEN) {
		u.u_error = EBUSY;
		return (OPENFAIL);
	}

	/*
	 * Pre-allocate space for the last-resort M_ERROR
	 * message, forcing the open to fail if the allocation
	 * fails.
	 */
	if (!(ttp->tt_merror = allocb(sizeof (char), BPRI_HI)))
		return (OPENFAIL);
	/*
	 * Grab space for an error acknowledgement.
	 * Failure here is acceptable.
	 */
	ttp->tt_errack = allocb(sizeof (struct T_error_ack), BPRI_HI);

	/*
	 * Opening the device corresponds to performing a
	 * socket call.  Grab a socket for our internal use.
	 * Set the socket options: SO_REUSEADDR & SO_KEEPALIVE
	 */
	if (u.u_error = socreate(AF_INET, &so, SOCK_DGRAM, 0))
		return (OPENFAIL);
	so->so_state |= SS_NBIO;

	/*
	 * Criteria for setting socket buffer sizes:
	 * size < SM_MAX * MCLBYTES / (2 * MSIZE + MCLBYTES)
	 * taken from sbreserve(), but not checked here to
	 * improve performance.
	 */
	so->so_rcv.sb_hiwat = TT_RECVSIZE;
	so->so_rcv.sb_mbmax = MIN(TT_RECVSIZE * 2, SB_MAX);
	so->so_snd.sb_hiwat = TT_SENDSIZE;
	so->so_snd.sb_mbmax = MIN(TT_SENDSIZE * 2, SB_MAX);
	/*
	 * Setup the special socket wakeup hooks here
	 */
	tsp = udptli_mksockinfo(ttp);
	if (!tsp)
		goto release;
	so->so_wupalt = &tsp->ts_sowakeup;
	ttp->tt_so = so;

	/*
	 * Set up cross-links between device and queues.
	 */
	ttp->tt_rq = q;
	q->q_ptr = (char *)ttp;
	WR(q)->q_ptr = (char *)ttp;

	/*
	 * Perform tli automaton state transition to the initial state.
	 */
	if (ttp->tt_state != 0) {
		/*
		 * This check counts on 0 == 0.
		 */
		panic("udptli_open: bogus tli state at open\n");
	}
	ttp->tt_state = TS_UNBND;

	ttp->tt_flags = TT_OPEN;

	ttp->tt_event = 0;

	/*
	 * Start up the auxiliary process if this is the
	 * first instance.
	 */
	if (!udptli_ttpcount++) {
		udptli_ttp = (struct tt_softc *)NULL;
		kern_proc((void (*)())udptli_main, (int)NULL);
		/*
		 * parent returns here -- child cannot, eventually exits
		 */
		UDPTLI_PRINTF("udptli_open: parent returning\n");
	}


	/*
	 * Increment the open ttp count.
	 */
	return (unit);

release:
	ttp->tt_flags &= ~TT_OPEN;	/* mark it not used */
	if (ttp->tt_so) {		/* release socket */
		(void)soclose(ttp->tt_so);
		ttp->tt_so = (struct socket *)NULL;
	}
	/*
	 * Free pre-allocated resources
	 */
	if (ttp->tt_merror) {
		freeb(ttp->tt_merror);
		ttp->tt_merror = (mblk_t *)NULL;
	}
	if (ttp->tt_errack) {
		freeb(ttp->tt_errack);
		ttp->tt_errack = (mblk_t *)NULL;
	}
	ttp->tt_rq = (queue_t *)NULL;		/* break the queue */
	q->q_ptr = (caddr_t)NULL;
	WR(q)->q_ptr = (caddr_t)NULL;
	ttp->tt_state = 0;	/* set tli transition state */
	return (OPENFAIL);
}


/* Release various resources when starting auxiliary process */
udptli_memfree()
{
	int 			i;
	extern	struct vnode	*rootdir;  /* root vnode set at system init */

	VN_RELE(u.u_cdir);
	u.u_cdir = rootdir;
	VN_HOLD(u.u_cdir);
	if (u.u_rdir) {
		VN_RELE(u.u_rdir);
	}
	u.u_rdir = (struct vnode *)NULL;
	/* Close all open files, including the partially open udptli device */
	for (i = 0; i < u.u_lastfile + 1; i++)
		if (u.u_ofile[i] != (struct file *)NULL) {
			closef(u.u_ofile[i]);
			u.u_ofile[i] = (struct file *)NULL;
			u.u_pofile[i] = 0;
		}
}


/*
 * When the TT_CLOSE flag is set, it means the ttp is in the
 * process of closing/clean-up. All requests will be ignored.
 *
 * When the flag (ttp->tt_flag = 0) is 0, that means it is
 * no longer possible to have any event outstanding and
 * there should be no ttp pending (i.e. on the pending queue).
 */
udptli_close(q)
	register queue_t	*q;
{
	register struct tt_softc	*ttp = (struct tt_softc *)q->q_ptr;
	register int			error = 0;
	register int			s;
	extern int			udptli_timeout();

	UDPTLI_PRINTF("udptli_close-%d: q: %x, WR(q): %x \n",
		ttp - tt_udpsoftc, q, WR(q));

	/*
	 * Flush outstanding traffic on its way
	 * downstream to us.
	 */
	flushq(WR(q), FLUSHALL);

	/*
	 * If this ttp has had a timeout scheduled then
	 * cancel it.
	 */
	if (ttp->tt_flags & TT_TIMER)
		untimeout(udptli_timeout, ttp);

	ttp->tt_flags = TT_CLOSE;

	/*
	 * Close the socket associated with this device instance.
	 */
	if (ttp->tt_so) {
		error = soclose(ttp->tt_so);
		ttp->tt_so = (struct socket *)NULL;
	}

	/*
	 * Perform tli automaton state transition.
	 *
	 */
	ttp->tt_state = 0;

	/*
	 * Free unused pre-allocated resources.
	 */
	if (ttp->tt_merror) {
		freeb(ttp->tt_merror);
		ttp->tt_merror = (mblk_t *)NULL;
	}
	if (ttp->tt_errack) {
		freeb(ttp->tt_errack);
		ttp->tt_errack = (mblk_t *)NULL;
	}


	/*
	 * Clear the ttp->tt_flags to indicate this ttp is
	 * not in use any more.
	 * Decrement the count on opened ttp, wake up
	 * aux proc if this is the last one
	 */
	UDPTLI_PRINTF("udptli_close: open count %d\n", udptli_ttpcount);
	if (--udptli_ttpcount == 0) {
		struct	wait4_args	args;

		/*
		 * If we get here, we are ready to shutdown the
		 * aux proc.
		 */
		(void)wakeup((caddr_t)&udptli_ttp);

		/*
		 * Wait for the aux proc to exit
		 * If the aux_proc is not our child, then
		 * the parent does not wait. The ttp->tt_flags
		 * is cleared by the aux_proc before
		 * it exits, so we are ok here.
		 */
		args.pid = udptli_auxprocp->p_pid;
		args.status = (union wait *)NULL;
		args.options = 0;
		args.rusage = (struct rusage *)NULL;

		wait4(&args);

	}

	/*
	 * More cleanup: remove any outstanding
	 * event on the pending queue.
	 *
	 * Do the search only if the ttp has been put on the queue.
	 *
	 * If there is only 1 event on the queue, just remove it.
	 * Otherwise, march down the singlely-linked list to
	 * find the match.
	 */
	s = splnet();
	if ((ttp->tt_event & TTE_ONQUEUE) &&
			(udptli_ttp != (struct tt_softc *)NULL)) {
		register struct tt_softc	*tp;	/* current ptr */
		register struct tt_softc	*btp;	/* previous ptr */

		if (udptli_ttp->tt_next == (struct tt_softc *)NULL) {
			udptli_ttp = (struct tt_softc *)NULL;
		} else {
			for (btp=tp=udptli_ttp; tp; btp=tp, tp=tp->tt_next) {
				if (tp == ttp) {
					if (btp == tp)
						udptli_ttp =
							 udptli_ttp->tt_next;
					else	btp->tt_next = tp->tt_next;
					break;
				}
			}
		}
	}
	(void)splx(s);
	/*
	 * Break association between queue and driver instance.
	 */
	UDPTLI_PRINTF("ttp-%d: %x, ttp->tt_rq: %x closed\n",
		ttp - tt_udpsoftc, ttp, ttp->tt_rq);
	ttp->tt_rq = (queue_t *)NULL;
	ttp->tt_flags = 0;

	if (error)
		u.u_error = error;
}
/*
 * Wake up routine to be used by the main process
 * to wake up the aux proc so it will process messages
 * queued on the WR(q).
 *
 */
/*ARGSUSED*/
int
udptli_wakeup(so, info, warg)
	struct	socket		*so;
	caddr_t			info,
				warg;
{
	register struct	tt_sockinfo	*tsp = (struct tt_sockinfo *)warg;
	register struct	tt_softc	*ttp = tsp->ts_ttp;
	register int			s;

	UDPTLI_PRINTF("udptli_wakeup: ttp-%d : %x socket: %x info: %d\n",
		ttp - tt_udpsoftc, ttp, so, info);
	UDPTLI_PRINTF("udptli_wakeup: ttp-%d : tsp: %x \n",
		ttp - tt_udpsoftc, tsp);

	/*
	 * If info == 0, then it is not a real wakeup, but
	 * an indication that we are called from sofree.
	 * So we free the tt_sockinfo memory, and also
	 * check to see if the freeing socket is the same
	 * kept by ttp. If yes, then set ttp->tt_so = NULL
	 * because when the aux proc wakes up, we can check
	 * to see if the socket has gone away.
	 */
	if (info == 0) {
		UDPTLI_PRINTF("udptli_wakeup: info = 0\n");
		if (so == tsp->ts_ttp->tt_so)
			tsp->ts_ttp->tt_so = (struct socket *)NULL;
		so->so_wupalt = NULL;
		kmem_free((caddr_t)tsp, sizeof *tsp);
	} else if (ttp->tt_flags & TT_CLOSE || ttp->tt_flags == 0) {
		UDPTLI_PRINTF("udptli_wakeup: tt_flags & TT_CLOSE\n");
		return;
	} else {
		/*
		 * Check if read is available, if so
		 * calls udptli_Ercv() directly. TT_DIRECT_READ
		 * is a C-macro defined in nettli/udp_tlivar.h.
		 *
		 * If this ttp is already on the queue pending
		 * for processing, then no insertion of ttp is needed.
		 */
		UDPTLI_PRINTF("udptli_wakeup: Check read\n");
		TT_DIRECT_READ(ttp, so);

		s = splnet();
		if (!(ttp->tt_event & TTE_ONQUEUE)) {
			ttp->tt_event |= TTE_EVENT;
			udptli_insq(ttp);
		}
		(void)splx(s);
		(void)wakeup((caddr_t)&udptli_ttp);
	}
	return;
}
/*
 * Insert to the tail of the udptli_ttp queue.
 * When an ttp is inserted in the queue, the ttp->tt_event's
 * TTE_ONQUEUE flag is set. It is cleared when it is removed
 * from the list.
 * Both udptli_insq() and udptli_remq() are called at splnet
 * level.
 */
udptli_insq(ttp)
register struct tt_softc	*ttp;
{
	register struct tt_softc	*tp = udptli_ttp;

	UDPTLI_PRINTF("udptli_insq: ttp-%d : %x\n", ttp - tt_udpsoftc, ttp);
	if (tp) {
		while (tp->tt_next)
			tp = tp->tt_next;
		tp->tt_next = ttp;
	} else {			/* first time */
		udptli_ttp = ttp;
	}
	ttp->tt_event |= TTE_ONQUEUE;	/* mark it's on the queue */
	ttp->tt_next = (struct tt_softc *)NULL;	/* mark as the last in queue */
}

/*
 * Remove from the head of the udptli_ttp queue
 */
struct tt_softc *
udptli_remq(tp)
	register struct tt_softc	*tp;
{
	if (tp == (struct tt_softc *)NULL) {
		UDPTLI_PRINTF("udptli_remq: returning NULL\n");
		return ((struct tt_softc *)NULL);
	} else {
		udptli_ttp = tp->tt_next;
		tp->tt_event &= ~TTE_ONQUEUE;
		UDPTLI_PRINTF("udptli_remq: returning ttp-%d : %x\n",
			tp - tt_udpsoftc, tp);
		return (tp);
	}
}
/*
 * Write-side put procedure.
 *
 * Streams-specific messages are processed immediately.  All other
 * (tli-related) messages are put onto the write queue for processing
 * by the auxiliary process.
 *
 * Stated differently, we maintain the following invariant on the write
 * queue: the write queue contains only messages that require action
 * by the auxiliary process and none relating only to the streams framework.
 */
udptli_wput (q, mp)
	queue_t	*q;
	mblk_t	*mp;
{
	register struct tt_softc	*ttp = (struct tt_softc *)q->q_ptr;
	register int			s;

	/*
	 * If we've issued an M_ERROR on this stream,
	 * we can accept no new messages.
	 */
	if (ttp->tt_flags & TT_ERROR) {
		freemsg(mp);
		return;
	}

	switch (mp->b_datap->db_type) {

	case M_FLUSH:
		UDPTLI_PRINTF("udptli_wput-%d: M_FLUSH q: %x, mp: %x\n",
			ttp - tt_udpsoftc, q, mp);
		if (*mp->b_rptr & FLUSHW) {

			flushq(q, FLUSHALL);
			/*
			 * Besides flushing the stream write queue,
			 * we must flush our associated socket's write
			 * queue as well.
			 *
			 */
			*mp->b_rptr &= ~FLUSHW;
		}
		if (*mp->b_rptr & FLUSHR) {
			/*
			 * N.B.: FLUSHW has been turned off by the
			 * time we get here.  (This is necessary to
			 * prevent a FLUSHW from circling back through
			 * the stream head to hit us a second time.)
			 */
			qreply(q, mp);
		} else	freemsg(mp);

		return;

	case M_PCPROTO:
		/*
		 * Priority request: validity check the message type,
		 * putting it onto the write queue if it's a legitimate
		 * tli message.
		 */
		UDPTLI_PRINTF("udptli_wput-%d: M_PCPROTO q: %x, mp: %x\n",
			ttp - tt_udpsoftc, q, mp);
		switch (((union T_primitives *)mp->b_rptr)->type) {

		case T_INFO_REQ:
			putq(q, mp);
			s = splnet();
			if (!(ttp->tt_event & TTE_ONQUEUE)) {
				ttp->tt_event |= TTE_EVENT;
				udptli_insq(ttp);
			}
			(void)splx(s);
			(void)wakeup((caddr_t)&udptli_ttp);
			return;

		default:
			/* Unknown message type. */
			break;
		}

		break;

	case M_PROTO:
		/*
		 * In-band request: validity check the message type,
		 * putting it onto the write queue if it's a legitimate
		 * tli message.
		 */
		UDPTLI_PRINTF("udptli_wput-%d: M_PROTO q: %x, mp: %x\n",
			ttp -tt_udpsoftc, q, mp);
		switch (((union T_primitives *)mp->b_rptr)->type) {

		case T_OPTMGMT_REQ:
		case T_BIND_REQ:
		case T_UNBIND_REQ:
		case T_UNITDATA_REQ:
			putq(q, mp);
			s = splnet();
			if (!(ttp->tt_event & TTE_ONQUEUE)) {
				ttp->tt_event |= TTE_EVENT;
				udptli_insq(ttp);
			}
			(void)splx(s);
			(void)wakeup((caddr_t)&udptli_ttp);
			return;

		default:
			/* Unknown message type. */
			break;
		}

		break;

	case M_IOCTL:
		/*
		 * We don't support any ioctl's, so pass
		 * this down.
		 */
		putq(q, mp);
		s = splnet();
		if (!(ttp->tt_event & TTE_ONQUEUE)) {
			ttp->tt_event |= TTE_EVENT;
			udptli_insq(ttp);
		}
		(void)splx(s);
		(void)wakeup((caddr_t)&udptli_ttp);
		return;

	default:
		/* Unknown message type. */
		break;
	}

	/*
	 * We should probably generate an error message
	 * here rather than dropping the message.
	 */
	freemsg(mp);
}


/*
 * N.B.: udptli_wsrv appears in udp_tliaux.c, not here, since the
 * auxiliary process invokes it, not the streams framework code.
 * This is just a null procedure but the existence of the
 * function call must exist otherwise flow control routines
 * such as canput() will not work.
 */
int
udptli_wsrv()
{
}


/*
 * Process tli primitive originated by the transport user.  Return
 * zero iff there were insufficient resources to process the primitive.
 *
 * Since we have no guarantees about the size of the request message,
 * we can't count on it being large enough to contain the corresponding
 * reply.  So in each processing of tli event, we allocate the
 * buffer for error msg so we don't have to risk the
 * possibility of allocation failure.  The canputback argument is useful
 * in this regard, being nonzero when it's allowable to push the incoming
 * message back onto the write queue, thereby avoiding the necessity of
 * having to deal with the failure.
 */
int
udptli_usrreq(ttp)
	struct	tt_softc	*ttp;
{
	/*
	 * The following variables are used in almost all cases
	 * of the switch below, so are declared here.
	 */
	register queue_t		*q;
	register mblk_t			*mp;
	long				type;
	struct tt_sockinfo		*udptli_mksockinfo();

	q = OTHERQ(ttp->tt_rq);

	/*
	 * Normal T_UNITDATA_REQ messages are processed in
	 * udptli_sendudata()  called from udptli_wput
	 * before it gets to this udptli_usrreq. The provision to
	 * handle T_UNITDATA_REQ messages here is for a msg
	 * that got put back on the queue due to a mclgetx() failure.
	 *
	 */
	while ((mp = getq(q))) {
		switch (mp->b_datap->db_type) {
		case M_PCPROTO:
		case M_PROTO:
			type = ((union T_primitives *)mp->b_rptr)->type;
			switch (type) {
				case T_INFO_REQ:
					(void)udptli_t_info_req(q, mp, ttp,
						type);
					break;

				case T_BIND_REQ:
					(void)udptli_t_bind_req(q, mp, ttp,
						type);
					break;

				case T_UNBIND_REQ:
					(void)udptli_t_unbind_req(q, mp, ttp,
						type);
					break;

				case T_OPTMGMT_REQ:
					(void)udptli_t_optmgmt_req(q, mp, ttp,
						type);
					break;

				case T_UNITDATA_REQ:
					(void)udptli_t_unitdata_req(q, mp, ttp,
						type);
					break;

				default:
					(void)udptli_error(q, mp, type,
						TBADDATA, 0);

			}	/* end of switch */
			break;

		case M_IOCTL:
			(void)udptli_ioctl(q, mp, ttp);
			break;

		default:
			UDPTLI_PRINTF("udp_tlireq: bad mp type %x\n",
				mp->b_datap->db_type);
			freemsg(mp);
			break;
		}
	}		/* end of while */
	return (0);
}

static void
udptli_ioctl(q, mp, ttp)
	queue_t *q;
	mblk_t  *mp;
	struct  tt_softc *ttp;
{
	struct iocblk	*iocmp;
	int		cmd;
	int		error;
	struct socket	*so;
	char		*data;

	UDPTLI_PRINTF("udptli-%d: ttp: %x tt_so: %x M_IOCTL\n",
		ttp - tt_udpsoftc, ttp, ttp->tt_so);

	so = ttp->tt_so;

	iocmp = (struct iocblk *)mp->b_rptr;
	cmd = iocmp->ioc_cmd;
	data = (char *)mp->b_cont->b_rptr;

#define	cmdbyte(x)	(((x) >> 8) & 0xff)
	switch (cmdbyte(cmd)) {
	case 'i':
		error = ifioctl(so, cmd, data);
		break;

	case 'r':
		error = rtioctl(cmd, data);
		break;

	default:
		error = ((*so->so_proto->pr_usrreq)(so, PRU_CONTROL,
		(struct mbuf *)cmd, (struct mbuf *)data, (struct mbuf *)0));
	}

	if (error != 0) {
		/*
		 * Package the error into an M_IOCNAK
		 * and send it upstream.
		 */
		mp->b_datap->db_type = M_IOCNAK;
		iocmp->ioc_error = error;
		iocmp->ioc_rval = 0;

		qreply(q, mp);

	} else	{
		/*
		 * Package the answer into an M_IOCACK
		 * and send it upstream.
		 */
		mp->b_datap->db_type = M_IOCACK;
		iocmp->ioc_error = 0;
		iocmp->ioc_rval = 0;

		qreply(q, mp);
	}
}

static void
udptli_t_unitdata_req(q, mp, ttp, type)
	queue_t *q;
	mblk_t  *mp;
	struct  tt_softc *ttp;
	long    type;
{
	register struct T_unitdata_req  *rp;
	register mblk_t			*mrp;   /* reply message */
	int				error = 0;

	UDPTLI_PRINTF("udptli-%d: ttp: %x tt_so: %x T_UNITDATA_REQ\n",
		ttp - tt_udpsoftc, ttp, ttp->tt_so);

	rp = (struct T_unitdata_req *)mp->b_rptr;

	/*
	 * Verify that we're in the proper state for the request.
	 */
	if (!(ttp->tt_state & TS_IDLE)) {
		(void)udptli_error(q, mp, type, TOUTSTATE, 0);
		return;
	}

	mrp = allocb(sizeof (union T_primitives), BPRI_HI);

	if (!mrp) {
		(void)udptli_error(q, mp, type, TSYSERR, ENOBUFS);
		return;
	}

	if (rp->DEST_length != sizeof (struct sockaddr_in)) {
		freemsg(mp);
		(void)tcptli_error(q, mrp, type, TBADADDR, 0);
		return;
	}

	/*
	 * Options not supported.
	 */
	if (rp->OPT_length > 0) {
		freemsg(mp);
		(void)udptli_error(q, mrp, type, TNOTSUPPORT, 0);
		return;
	}

	if (!mp->b_cont) {
		freemsg(mp);
		(void)udptli_error(q, mrp, type, TBADDATA, 0);
		return;
	}

	if ((error = udptli_sendudata(ttp, OTHERQ(ttp->tt_rq), mp)) != 0) {
		UDPTLI_PRINTF("udptli_sendudata rtnd %d\n", error);
		if (error == ENOBUFS) {
			putq(q, mp);
			udptli_schedwakeup(ttp, 10);
			freeb(mrp);
		} else {
			(void)udptli_error(ttp->tt_rq, mrp,
				(long)0, 0, EPROTO);
		}
	} else {
		freeb(mrp);
	}
	return;
}

/*
 * This routine processes
 * T_UNITDATA_REQ msgs.
 */
udptli_sendudata(ttp, q, mp)
	struct	tt_softc	*ttp;
	queue_t			*q;
	mblk_t			*mp;
{
	register mblk_t 	*mbp;
	register struct	mbuf	*top;
	register struct	mbuf	*prev;
	register struct	mbuf	*cur;
	struct	mbuf		*mclgetx();
	extern	int		freeb();
	int			s;
	int			error;
	struct	socket		*so;
	struct mbuf		*nam;
	struct sockaddr_in	sa;
	struct T_unitdata_req	*t_udata;
	int			offset;

	UDPTLI_PRINTF("udptli_mdata: ttp: %x, q: %x \n",
		ttp - tt_udpsoftc, q);

	error = 0;
	so = ttp->tt_so;
	top = prev = cur = (struct mbuf *)NULL;

	/* Stuff the address into an mbuf. */
	t_udata = (struct T_unitdata_req *)mp->b_rptr;
	offset = t_udata->DEST_offset;
	(void)bzero((caddr_t)sa.sin_zero, 8);
	(void)bcopy((caddr_t)(mp->b_rptr + offset),
			(caddr_t)&sa, sizeof (struct sockaddr_in));
	error = udptli_sockargs(&nam, (caddr_t)&sa,
		sizeof (struct sockaddr_in), MT_SONAME);
	UDPTLI_PRINTF("udptli_sockargs rtnd %d\n", error);
	if (error)
		return (error);

	mbp = mp->b_cont;
	mp->b_cont = NULL;
	freeb(mp);
	mp = mbp;

	/*
	 * Stuff into mbuf(s).
	 * Send the data and put the msg back on the queue
	 * then return.
	 */
	if (mp) {
		for (mbp = mp; mbp; mbp = mbp->b_cont) {
			cur = mclgetx(freeb, (int)mbp, (caddr_t)mbp->b_rptr,
				(int)(mbp->b_wptr - mbp->b_rptr), M_DONTWAIT);
			if (!cur) {		/* allocation failed */
				m_freem(top);	/* free the mbuf chain */
				return (ENOBUFS);
			}
			if (!top)
				top = cur;
			if (prev)
				prev->m_next = cur;
			prev = cur;
			cur->m_next = 0;
		}
	}
	/*
	 * If no more msg in the queue, ==> this is
	 * is the last (or only) M_DATA msg, go ahead and send it
	 */
	UDPTLI_PRINTF("udptli_mdata: Sending M_DATA msg mp: %x\n", mp);

	/*
	 * This routine may be called when the streams queues are run from
	 * interrupt level so we can't afford to sleep here. We will sleep
	 * if we try to lock the socket and it's already locked !!
	 * In that case just return error and the i/o will be scheduled
	 * later to be run out the process context provided by the auxproc.
	 */
	if (so->so_rcv.sb_flags & SB_LOCK) {
		if (top)
			m_freem(top);   /* free the mbuf chain */
		return (ENOBUFS);
	}
	sblock(so, &so->so_snd);
	s = splnet();
	error = (*so->so_proto->pr_usrreq)(so, PRU_SEND, top, (caddr_t)nam,
		(struct mbuf *)0);
	(void)splx(s);
	sbunlock(so, &so->so_snd);

	m_freem(nam);
	return (error);
}


static void
udptli_t_info_req(q, mp, ttp, type)
	queue_t	*q;
	mblk_t	*mp;
	struct	tt_softc *ttp;
	long	type;
{
	register struct T_info_ack	*ap;
	register mblk_t			*mrp;	/* reply message */

	UDPTLI_PRINTF("udptli-%d: ttp: %x tt_so: %x T_INFO_REQ\n",
		ttp - tt_udpsoftc, ttp, ttp->tt_so);

	mrp = allocb(sizeof (union T_primitives), BPRI_HI);
	if (!mrp) {
		udptli_error(q, mp, type, TSYSERR, ENOBUFS);
		return;
	}
	freemsg(mp);
	mrp->b_datap->db_type = M_PCPROTO;
	mrp->b_wptr += sizeof (struct T_info_ack);
	ap = (struct T_info_ack *)mrp->b_rptr;
	ap->PRIM_type = T_INFO_ACK;

	/*
	 *
	 * Questionable choices:
	 *   TSDU_size:
	 *	If this concept is supported at all,
	 *	it probably maps to the udp segment size.
	 *   ETSDU_size:
	 *	Really this is a bit stating that there's
	 *	urgent data coming further along.
	 *   OPT_size:
	 *	It's still unknown which of the possible
	 *	protocol options make sense for udp and
	 *	how we can make them correspond to those
	 *	of other streams udp implementations.
	 *   SERV_type:
	 *	It remains to be seen whether udp supports
	 *	orderly disconnect in the ISO sense.
	 */
	ap->TSDU_size = strmsgsz;
	ap->ETSDU_size = 1;
	ap->CDATA_size = -2;			/* not supported */
	ap->DDATA_size = -2;			/* not supported */
	ap->ADDR_size = sizeof (struct sockaddr_in);
	ap->OPT_size = sizeof (struct tt_soopt); /* TBD */
	ap->TIDU_size = strmsgsz;
	ap->SERV_type = T_CLTS;
	ap->CURRENT_state = ((struct tt_softc *)q->q_ptr)->tt_state;

	qreply(q, mrp);
	return;
}
/*
 * This primitive roughly corresponds to paired 4.3 bind/listen calls.
 */
static void
udptli_t_bind_req(q, mp, ttp, type)
	queue_t	*q;
	mblk_t	*mp;
	struct	tt_softc *ttp;
	long	type;
{
	register struct T_bind_req	*rp;
	register struct T_bind_ack	*rap;
	register mblk_t			*mrp;	/* reply message */
	register int			error;
	int				offset;
	struct mbuf			*nam;

	UDPTLI_PRINTF("udptli-%d: ttp: %x tt_so: %x T_BIND_REQ\n",
		ttp - tt_udpsoftc, ttp, ttp->tt_so);

	nam = (struct mbuf *)NULL;

	/*
	 * Verify that the request is allowable.
	 */
	if (ttp->tt_state != TS_UNBND) {
		udptli_error(q, mp, type, TOUTSTATE, 0);
		return;
	}

	rp = (struct T_bind_req *)mp->b_rptr;
	offset = rp->ADDR_offset;

	/*
	 * Alloc a reply buffer
	 */
	mrp = allocb(MAX(sizeof (union T_primitives),
		sizeof (struct T_bind_ack) + sizeof (struct sockaddr_in)),
		BPRI_HI);
	if (!mrp) {
		udptli_error(q, mp, type, TSYSERR, ENOBUFS);
		return;
	}

	/*
	 * Stuff the address to bind to into an mbuf.  (If one
	 * wasn't supplied, we're supposed to pick one, which
	 * in_pcbbind will do during the call below to sobind.)
	 */
	if (rp->ADDR_length > 0) {
		struct	sockaddr_in	sa;

		if (rp->ADDR_length != sizeof (struct sockaddr_in)) {
			freemsg(mp);
			udptli_error(q, mrp, type, TBADADDR, 0);
			return;
		}
		(void)bzero((caddr_t)sa.sin_zero, 8);
		(void)bcopy((caddr_t)(mp->b_rptr + offset), (caddr_t)&sa,
			sizeof (struct sockaddr_in));
		error = udptli_sockargs(&nam, (caddr_t)&sa,
			sizeof (struct sockaddr_in), MT_SONAME);
		if (error) {
			freemsg(mp);
			udptli_error(q, mrp, type, TSYSERR, error);
			return;
		}
	}

	/*
	 * Bind the socket, translating unix errors into tli errors.
	 * Also save the port and address in ttp struct for
	 * future reference (a backup) in the case the inpcb
	 * is blown away.
	 */
	error = sobind(ttp->tt_so, nam);
	if (nam)
		m_freem(nam);
	if (error) {
		int	te;

		freemsg(mp);
		if (error == EADDRINUSE || error == EADDRNOTAVAIL)
			te = TNOADDR;
		else if (error == EACCES)
			te = TACCES;
		else
			te = TSYSERR;
		udptli_error(q, mrp, type, te, te == TSYSERR ? error : 0);
		return;
	}


	/*
	 * Generate acknowledgement message by digging resulting
	 * values out of socket and in_pcb structures.
	 */
	offset = sizeof (struct T_bind_ack);
	mrp->b_datap->db_type = M_PCPROTO;
	mrp->b_wptr += offset + sizeof (struct sockaddr_in);
	rap = (struct T_bind_ack *)mrp->b_rptr;
	rap->PRIM_type = T_BIND_ACK;

	rap->ADDR_length = sizeof (struct sockaddr_in);
	rap->ADDR_offset = offset;
	rap->CONIND_number = 0;
	freemsg(mp);
	/*
	 * Copy family, port, and address into mblk to be
	 * sent up to the Transport User.
	 */
	{
		struct sockaddr_in		sa;
		struct inpcb		*inp = sotoinpcb(ttp->tt_so);

		bzero((caddr_t)&sa, sizeof sa);
		sa.sin_family = AF_INET;
		sa.sin_port = inp->inp_lport;
		sa.sin_addr = inp->inp_laddr;
		bcopy((caddr_t)&sa, (caddr_t)(mrp->b_rptr + offset),
			sizeof (sa));
	}
	qreply(q, mrp);

	/*
	 * Perform tli automaton state transition.
	 */
	ttp->tt_state = TS_IDLE;

	return;
}
/*
 *
 */
static void
udptli_t_unbind_req(q, mp, ttp, type)
	queue_t	*q;
	mblk_t	*mp;
	struct	tt_softc *ttp;
	long	type;
{
	register mblk_t			*mrp;	/* reply message */
	register struct T_ok_ack	*rap;
	mblk_t				*mfp;
	struct	tt_sockinfo		*tsp;
	struct	tt_sockinfo		*udptli_mksockinfo();
	struct socket 			*so;
	int				error;
	int				fatalerror;

	UDPTLI_PRINTF("udptli-%d: ttp: %x tt_so: %x T_UNBIND_REQ\n",
		ttp - tt_udpsoftc, ttp, ttp->tt_so);

	fatalerror = 0;
	mfp = (mblk_t *)NULL;

	/*
	 * Verify that the request is allowable.
	 */
	if (ttp->tt_state != TS_IDLE) {
		udptli_error(q, mp, type, TOUTSTATE, 0);
		return;
	}

	/*
	 * Must issue M_FLUSH for both read and write queues
	 * before sending back the ack.
	 */
	mfp = allocb(sizeof (struct T_ok_ack), BPRI_HI);
	if (!mfp) {
		udptli_error(q, mp, type, TSYSERR, ENOBUFS);
		return;
	}
	mfp->b_datap->db_type = M_FLUSH;
	*(mfp->b_rptr) |= FLUSHRW;
	qreply(q, mfp);

	/*
	 * allocate a reply msg buffer
	 */
	mrp = allocb(sizeof (union T_primitives), BPRI_HI);
	if (!mrp) {
		udptli_error(q, mp, type, TSYSERR, ENOBUFS);
		return;
	}
	/*
	 * There's some difficulty with leaving the socket in a state
	 * such that it can later be reused.  In_pcbbind refuses to
	 * allow rebinding once already bound.  The only
	 * way is to do this is to issue an soclose
	 * followed by an socreate.
	 */
	if (ttp->tt_so) {
		error = soclose(ttp->tt_so);
		ttp->tt_so = (struct socket *)NULL;
	}
	if (error) {
		freemsg(mp);
		udptli_error(q, mrp, type, TSYSERR, error);
		return;
	}
	error = socreate(AF_INET, &so, SOCK_DGRAM, 0);
	if (error) {
		fatalerror = 1;
		goto reply;
	}

	/*
	 * Set the socket options and send/recv socket buffer.
	 * Criteria for setting socket buffer sizes:
	 * size < SM_MAX * MCLBYTES / (2 * MSIZE + MCLBYTES)
	 * taken from sbreserve(), but not checked here to
	 * imporve performance.
	 */

	so->so_rcv.sb_hiwat = TT_RECVSIZE;
	so->so_rcv.sb_mbmax = MIN(TT_RECVSIZE * 2, SB_MAX);
	so->so_snd.sb_hiwat = TT_SENDSIZE;
	so->so_snd.sb_mbmax = MIN(TT_SENDSIZE * 2, SB_MAX);
	/*
	 * Setup the special socket wakeup hooks here
	 */
	tsp = udptli_mksockinfo(ttp);
	if (tsp == (struct tt_sockinfo *)NULL) {
		fatalerror = 1;
		goto reply;
	}
	so->so_wupalt = &tsp->ts_sowakeup;

	so->so_state |= SS_NBIO;
	ttp->tt_so = so;

reply:
	/*
	 * Acknowledge success.
	 */
	mrp->b_datap->db_type = M_PCPROTO;
	mrp->b_wptr += sizeof (struct T_ok_ack);
	rap = (struct T_ok_ack *)mrp->b_rptr;
	rap->PRIM_type = T_OK_ACK;

	rap->CORRECT_prim = T_UNBIND_REQ;
	qreply(q, mrp);

	/*
	 * Perform tli automaton state transition.
	 */
	ttp->tt_state = TS_UNBND;

	if (fatalerror)
		udptli_error(q, mp, type, NULL, error);
	else
		freemsg(mp);

	return;
}
static void
udptli_t_optmgmt_req(q, mp, ttp, type)
	queue_t			*q;
	mblk_t			*mp;
	struct	tt_softc	*ttp;
	long			type;
{
	register struct T_optmgmt_req	*rp;
	register struct socket		*so = ttp->tt_so;
	struct	T_optmgmt_ack 		*rap;
	int				offset;

	UDPTLI_PRINTF("udptli-%d: ttp: %x tt_so: %x T_OPTMGMT_REQ\n",
		ttp - tt_udpsoftc, ttp, ttp->tt_so);
	rp = (struct T_optmgmt_req *)mp->b_rptr;
	offset = rp->OPT_offset;

	/*
	 * Verify that the request is allowable; options
	 * may be negotiated only from the bound state.
	 */
	if (ttp->tt_state != TS_IDLE) {
		udptli_error(q, mp, type, TOUTSTATE, 0);
		return;
	}
	switch (rp->MGMT_flags) {

	case T_DEFAULT: {
		register mblk_t	*mrp;	/* reply message */
		struct	tt_soopt	*optp;

		mrp = allocb(sizeof (struct T_optmgmt_ack)
			+ sizeof (struct tt_soopt), BPRI_HI);
		if (!mrp) {
			udptli_error(q, mp, type, TSYSERR, ENOBUFS);
			return;
		}

		freemsg(mp);
		mrp->b_datap->db_type = M_PCPROTO;
		mrp->b_wptr = mrp->b_rptr + sizeof (struct T_optmgmt_ack)
				+ sizeof (struct tt_soopt);
		rap = (struct T_optmgmt_ack *)mrp->b_rptr;
		rap->MGMT_flags = T_DEFAULT;
		rap->PRIM_type = T_OPTMGMT_ACK;
		rap->OPT_length = sizeof (struct tt_soopt);
		rap->OPT_offset = sizeof (struct T_optmgmt_ack);
		optp = (struct tt_soopt *)(mrp->b_rptr
			+ sizeof (struct T_optmgmt_ack));
		optp->tts_reuseaddr = TTS_DFLT_REUSEADDR;
		optp->tts_broadcast = TTS_DFLT_BROADCAST;
		optp->tts_sendsize = TTS_BUFSIZE;
		optp->tts_recvsize = TTS_BUFSIZE;
		qreply(q, mrp);
		return;

	}
	case T_CHECK: {
		if ((rp->OPT_length != sizeof (struct tt_soopt)) ||
			(rp->OPT_offset < sizeof (struct T_optmgmt_req))) {

			udptli_error(q, mp, type, TBADOPT, 0);
			return;
		}

		rap = (struct T_optmgmt_ack *)mp->b_rptr;
		rap->PRIM_type = T_OPTMGMT_ACK;
		rap->MGMT_flags |= T_SUCCESS;
		mp->b_datap->db_type = M_PCPROTO;
		qreply(q, mp);
		return;
	}
	case T_NEGOTIATE: {
		struct	tt_soopt	opt,
					*optp;
		struct	mbuf		*m = NULL;
		int			error;


		if ((rp->OPT_length != sizeof (struct tt_soopt)) ||
			(rp->OPT_offset < sizeof (struct T_optmgmt_req))) {

			udptli_error(q, mp, type, TBADOPT, 0);
			return;
		}

		bcopy((char *)(mp->b_rptr + offset), (char *)&opt,
			sizeof (struct tt_soopt));

		if (opt.tts_reuseaddr) {
			error = sosetopt(so, SOL_SOCKET, SO_REUSEADDR,
				(struct mbuf *)0);
			if (error < 0) {
				udptli_error(q, mp, type,
					TSYSERR, error);
				return;
			}
		}
		if (opt.tts_broadcast) {
			error = sosetopt(so, SOL_SOCKET, SO_BROADCAST,
				(struct mbuf *)0);
			if (error < 0) {
				udptli_error(q, mp, type, TSYSERR, error);
				return;
			}
		}
		if (opt.tts_sendsize > 0) {
			/*
			 * mbuf is freed at the end of sosetopt()
			 */
			m = m_get(M_WAIT, MT_SOOPTS);
			if (m == (struct mbuf *)NULL) {
				udptli_error(q, mp, type, TSYSERR, ENOBUFS);
				return;
			}
			bcopy((caddr_t)&opt.tts_sendsize,
				mtod(m, caddr_t), sizeof (int));
			m->m_len = sizeof (int);
			error = sosetopt(so, SOL_SOCKET, SO_SNDBUF, m);
			if (error < 0) {
				udptli_error(q, mp, type, TSYSERR, error);
				return;
			}
		}
		if (opt.tts_recvsize > 0) {
			m = m_get(M_WAIT, MT_SOOPTS);
			if (m == (struct mbuf *)NULL) {
				udptli_error(q, mp, type, TSYSERR, ENOBUFS);
				return;
			}
			bcopy((caddr_t)&opt.tts_recvsize,
				mtod(m, caddr_t), sizeof (int));
			m->m_len = sizeof (int);
			error = sosetopt(so, SOL_SOCKET, SO_RCVBUF, m);
			if (error < 0) {
				udptli_error(q, mp, type, TSYSERR, error);
				return;
			}
		}

		mp->b_datap->db_type = M_PCPROTO;
		rap = (struct T_optmgmt_ack *)mp->b_rptr;
		rap->PRIM_type = T_OPTMGMT_ACK;
		rap->MGMT_flags = T_NEGOTIATE;
		rap->OPT_length = sizeof (struct tt_soopt);
		rap->OPT_offset = sizeof (struct T_optmgmt_ack);
		optp = (struct tt_soopt *)(mp->b_rptr
			+ sizeof (struct T_optmgmt_ack));
		optp->tts_reuseaddr = opt.tts_reuseaddr;
		optp->tts_sendsize = so->so_snd.sb_hiwat;
		optp->tts_recvsize = so->so_rcv.sb_hiwat;
		optp->tts_broadcast = opt.tts_broadcast;

		qreply(q, mp);
		return;
	}
	default:
		udptli_error(q, mp, type, TBADFLAG, 0);
		return;
	}
}
#endif	NUDPTLI > 0
