#ifndef lint
static  char sccsid[] = "@(#)udp_tliaux.c 1.3 90/03/28 SMI";
#endif

/*
 * Copyright (c) 1988 by Sun Microsystems, Inc.
 */


/*
 * The auxiliary process half of the streams interface
 * to the socket-based UDP implementation.
 */

#define	TLIDEBUG

#include "udptli.h"

#if	NUDPTLI > 0

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/debug.h>
#include <sys/user.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/uio.h>

#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_pcb.h>
#include <netinet/udp.h>
#include <netinet/ip_var.h>
#include <netinet/udp_var.h>

#include <nettli/tihdr.h>
#include <nettli/tiuser.h>
#include <nettli/udp_tli.h>
#include <nettli/udp_tlivar.h>


/*
 * Round up to machine alignment boundary.
 *
 * XXX:	belongs in <machine/param.h>.
 */
#ifndef	align
#define	align(x)	(((x) + 3) & ~3)
#endif	align


int		udptli_auxproc_running = 0;
extern int	udptli_debug;

/*
 * Auxiliary process main loop.  This consists of waiting until
 * the stream bottom code requests some processing, then doing
 * whatever can be done.
 */
udptli_main()
{
	register int			s;
	register struct tt_softc	*ttp;
	extern struct proc		*udptli_auxprocp;
	extern struct tt_softc		*udptli_ttp;
	extern int			udptli_ttpcount;
	extern int			udptli_auxproc_running;
	struct tt_softc			*udptli_remq();

	UDPTLI_PRINTF("udptli_auxproc: Up and running ...\n");
	/*
	 * Initialization of global veriables
	 */
	udptli_ttp = (struct tt_softc *)NULL;
	udptli_auxprocp = u.u_procp;
	udptli_memfree();
	bcopy((caddr_t)"udptli_auxproc", (caddr_t)u.u_comm, (size_t)15);
	(void)setjmp((caddr_t)&u.u_qsave);

	for (;;) {
		/*
		 * Wait for the next processing request.
		 *
		 */

		s = splnet();
		while ((ttp = udptli_remq(udptli_ttp)) ==
			(struct tt_softc *)NULL && udptli_ttpcount != 0) {
			udptli_auxproc_running = 0;
			(void)sleep((caddr_t)&udptli_ttp, PZERO + 1);
		}
		udptli_auxproc_running = 1;
		(void)splx(s);

		UDPTLI_PRINTF("udptli_auxproc: checking queues ...\n");

		/*
		 * If no more open ttp, terminate this aux proc
		 */
		if (ttp == (struct tt_softc *)NULL && udptli_ttpcount == 0) {
			UDPTLI_PRINTF("udptli_auxproc: Exiting ...\n");
			exit(0);
		}
		/*
		 * If the ttp already marked TT_CLOSE, then no
		 * processing is done.
		 */
		if (ttp->tt_flags & TT_CLOSE)
			continue;


		/*
		 * If this wakeup was done as a result of sofree
		 * then no processing is done.
		 */

		/*
		 * Process the read side first ...
		 *
		 * If the read queue is full, do nothing.
		 *
		 * Otherwise check each possible event to see whether
		 * it's occurred, doing the appropriate processing for
		 * each that has.
		 *
		 * The code below is complicated by the need to observe
		 * tli state discipline.  Each event is legal only from
		 * certain tli states.  Moreover, at a given instant, the
		 * set of possible events is a function of the socket (and
		 * underlying UDP) states.
		 *
		 **********
		 *  NOTE  *
		 **********
		 * The read side is performed first because read events
		 * are asynchronous. In the event that a socket
		 * is blown away unexpectedly, the repair will be
		 * done before any other operation (which has no knowledge
		 * of possible damage to the ttp struct) is attempted.
		 */
		udptli_rdevent(ttp);

		/*
		 * Process the write side.
		 *
		 * udptli_usrreq takes the entire queue and does
		 * getq within it.
		 *
		 */
		(void)udptli_usrreq(ttp);
	}
}


/*
 * Process receive events.
 *
 */
udptli_rdevent (ttp)
	struct tt_softc	*ttp;
{
	int			processedevent;

	/*
	 * Search for new events until there are none or we can't
	 * handle any more for the moment.  The latter circumstance
	 * amounts to running out of resources -- mblks, room in the
	 * read queue, etc.  In this case we'll break from the loop.
	 */
	do {
		processedevent = 0;

		processedevent |= udptli_Ercv(ttp);

	} while (processedevent);
}


/*
 * Process a pending "rcv" event, returning 1 if there
 * is one and it was processed successfully and 0 otherwise.
 *
 * This event corresponds to the local events occurring when a
 * foreign peer sends us data.
 *
 */
int upcount = 0;
int
udptli_Ercv(ttp)
	register struct tt_softc	*ttp;
{
	struct socket			*so = ttp->tt_so;
	register struct	T_unitdata_ind	*udip;
	struct	iovec			iov[TT_MAXUIO];
	struct	uio			uio;
	struct	mbuf			*nam;
	struct	mbuf			*m;
	mblk_t				*mp;
	int				s;
	int				error;

	extern struct tt_softc		tt_udpsoftc[];

	if ((ttp->tt_state & T_IDLE) == 0)
		return (0);

	/*
	 * Need to mark the SB_WAIT flag in the socket-recv
	 * buffer. The socket will never get woken up if
	 * this flag is not set. Look at sbwakeup() for
	 * details.
	 */
	s = splnet();
	so->so_rcv.sb_flags |= SB_WAIT;
	(void)splx(s);
	/*
	 *
	 *	Basic outline:
	 *	while (socket nonempty && canput) {
	 *		transcribe as big a chunk as we can.
	 *	}
	 *
	 *	Data detected by:
	 *		so->so_rcv.sb_cc > 0
	 */

	if (!canput(ttp->tt_rq)) {
		udptli_schedwakeup(ttp, 100);
		return (0);
	}
	/*
	 * We could take soreceive and let its uiomove to
	 * copy data to mblk's. If we use m_cpytoc(), then
	 * we need to free mbufs. So by using soreceive(),
	 * we don't have to worry about freeing mbufs.
	 */

	if (so->so_rcv.sb_cc != 0) {
		int	size;		/* amount of data in socket queue */
		caddr_t	addr;
		mblk_t	*smp;
		mblk_t	*tmp;

		UDPTLI_PRINTF("Ercv ... [udptli-%d] ttp: %x tt_so %x\n",
			ttp - tt_udpsoftc, ttp, so);

		smp = tmp = (mblk_t *)NULL;

		uio.uio_iov = iov;
		uio.uio_iovcnt = 0;
		uio.uio_offset = 0;
		uio.uio_segflg = UIO_SYSSPACE;
		uio.uio_fmode = 0;
		uio.uio_resid = 0;

		size = sizeof (struct T_unitdata_ind) +
				sizeof (struct sockaddr_in);
		if ((mp = allocb(size, BPRI_MED)) == (mblk_t *)NULL) {
			return (0);
		}
		mp->b_datap->db_type = M_PROTO;
		mp->b_wptr += size;
		udip = (struct T_unitdata_ind *)mp->b_rptr;
		udip->PRIM_type = T_UNITDATA_IND;
		udip->SRC_offset = sizeof (struct T_unitdata_ind);
		udip->SRC_length = sizeof (struct sockaddr_in);
		udip->OPT_offset = 0;
		udip->OPT_length = 0;

		/*
		 * calculate size of receive buffer required.
		 */
		size = 0;
		for (m = so->so_rcv.sb_mb; m; m = m->m_next) {
			if (m->m_type == MT_RIGHTS || m->m_type == MT_SONAME)
				continue;
			size += m->m_len;
		}
		uio.uio_resid = size;
		for (smp = mp; size > 0; smp = smp->b_cont) {
			tmp = allocb(MIN(size, TT_BUFSIZE), BPRI_MED);
			/*
			 * Allocb fails, if got some buffers, then
			 * do the sorecive() on what are available
			 * If not, just quit.
			 */
			if (!tmp) {
				UDPTLI_PRINTF("udptli_Ercv: allocb failed\n");
				freeb(mp);
				return (0);
			}
			uio.uio_iov->iov_base = (caddr_t)tmp->b_rptr;
			tmp->b_wptr += MIN(size, TT_BUFSIZE);
			uio.uio_iov->iov_len = MIN(size, TT_BUFSIZE);
			smp->b_cont = tmp;
			size -= uio.uio_iov->iov_len;
			uio.uio_iov++;
			uio.uio_iovcnt++;
			if (uio.uio_iovcnt >= TT_MAXUIO)
				break;
		}
		if (size > 0)
			udptli_schedwakeup(ttp, 100);

		UDPTLI_PRINTF("udptli_Ercv: size = %d, uio_resid = %d\n",
				size, uio.uio_resid);
		uio.uio_resid -= size;
		if (uio.uio_resid == 0) {
			freemsg(mp);
			return (0);
		}
		uio.uio_iov = iov;	/* reset to iov[0] */
		/*
		 * start the soreceive
		 */
		UDPTLI_PRINTF("udptli_Ercv: Receiving %d bytes\n",
				uio.uio_resid);
		error = soreceive(so, &nam, &uio, 0, (struct mbuf **)0);
		UDPTLI_PRINTF("udptli_Ercv: uio_resid = %d after soreceive\n",
				uio.uio_resid);
		if (error) {
			UDPTLI_PRINTF(
		"udptli_Ercv: unexpected error from soreceive\n");
			freemsg(mp);
			return (0);
		}

		/*
		 * Copy address into T_UNITDATA_IND header.
		 */
		addr = (caddr_t)(mp->b_rptr + udip->SRC_offset);
		(void)bcopy(mtod(nam, caddr_t), addr, nam->m_len);
		m_freem(nam);

		putnext(ttp->tt_rq, mp);
		return (1);
	} else	return (0);
}


#endif	NUDPTLI > 0
