#ifndef lint
static  char sccsid[] = "@(#)udp_tlisubr.c 1.1 89/12/14;from 1.9 88/10/13 Copyr 1988 Sun Micro";
#endif

#define	TLIDEBUG

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/debug.h>
#include <sys/user.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/uio.h>
#include <sys/protosw.h>

#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_pcb.h>

#include <nettli/tihdr.h>
#include <nettli/tiuser.h>
#include <nettli/udp_tli.h>
#include <nettli/udp_tlivar.h>


/*
 * Copyright (c) 1988 by Sun Microsystems, Inc.
 */

/*
 * If tlierr == NULL, then issues M_ERROR msg
 */
void
udptli_error(q, mrp, prim, tlierr, unixerr)
	queue_t	*q;
	mblk_t	*mrp;
	long	prim;
	int	tlierr;
	int	unixerr;
{
	register struct T_error_ack	*ap;
	register struct tt_softc	*ttp;

	ttp = (struct tt_softc *)q->q_ptr;

	/*
	 * If NULL is passed in the mrp field, then use
	 * the reserve tt_errack buffer allocated
	 * earlier during open. Buffer tt_merror is not
	 * used unless tt_errack does not exist (for whatever
	 * reason).
	 */
	if (!mrp) {
		if (!ttp->tt_errack) {
			mrp = ttp->tt_merror;
			ttp->tt_merror = (mblk_t *)NULL;
		} else {
			mrp = ttp->tt_errack;
			ttp->tt_errack = (mblk_t *)NULL;
		}
	}
	if (tlierr == 0) {
		mrp->b_datap->db_type = M_ERROR;
		*(mrp->b_wptr) = (char) unixerr;
		mrp->b_wptr++;
	} else {
		mrp->b_datap->db_type = M_PCPROTO;
		mrp->b_wptr = mrp->b_rptr + sizeof (struct T_error_ack);
		ap = (struct T_error_ack *)mrp->b_rptr;
		ap->PRIM_type = T_ERROR_ACK;

		ap->ERROR_prim = prim;
		ap->TLI_error = tlierr;
		ap->UNIX_error = unixerr;
	}
	/*
	 * Always issue on the read queue.
	 */
	putnext(ttp->tt_rq, mrp);
}

/*
 * udptli_sockargs() is the same as sockargs(),
 * except this one copies data within the
 * kernel address space
 */
udptli_sockargs(aname, name, namelen, type)
	struct mbuf	**aname;
	caddr_t		name;
	int 		namelen;
	int		type;
{
	register struct mbuf *m;

	if (namelen > MLEN)
		return (EINVAL);
	m = m_get(M_WAIT, type);
	if (m == (struct mbuf *)NULL)
		return (ENOBUFS);
	m->m_len = namelen;
	bcopy(name, mtod(m, caddr_t), (size_t)namelen);
	*aname = m;
	return (0);
}

/*
 * Routine to create a new tt_sockinfo block. The kmem_free
 * is done in sofree() i.e. when the associated socket is
 * closed.
 */
struct tt_sockinfo *
udptli_mksockinfo(ttp)
	struct	tt_softc	*ttp;
{
	register struct tt_sockinfo *tsp;
	int	udptli_wakeup();

	tsp = (struct tt_sockinfo *)kmem_alloc(sizeof (struct tt_sockinfo));
	if (tsp == (struct tt_sockinfo *)NULL) {
		(void)printf("udptli_mksockinfo: kmem_alloc failed\n");
		return ((struct tt_sockinfo *)NULL);
	}
	tsp->ts_sowakeup.wup_func = udptli_wakeup;
	tsp->ts_sowakeup.wup_arg = (caddr_t) tsp;
	tsp->ts_ttp = ttp;
	return (tsp);
}
/*
 * When a service was attempted but failed, requiring a wakeup
 * at a later time. The scheme here is to check to see if a
 * timer is set already (possibly by another service failure) and
 * then issues a timer to call udptli_timeout() [see below], which
 * will clear the TT_TIMER flag when the timer goes off, thus
 * the TT_TIMER flag is set/clear only by these two routines.
 */
udptli_schedwakeup(ttp, hz)
	struct	tt_softc	*ttp;
	int			hz;	/* clock ticks */
{
	int	udptli_timeout();

	UDPTLI_PRINTF("udptli_schedwakeup: for ttp = %x\n", ttp);
	if (!(ttp->tt_flags & TT_TIMER)) {
		ttp->tt_flags |= TT_TIMER;
		timeout((int (*)()) udptli_timeout, (caddr_t)ttp, hz);
	}
	return;
}
/*
 * When timer goes off, the TT_TIMER flag is clears
 * and a real wakeup is issued
 */
udptli_timeout(ttp)
	struct	tt_softc	*ttp;
{
	register int		s;
	extern struct tt_softc	*udptli_ttp;

	ttp->tt_flags &= ~TT_TIMER;
	ttp->tt_event |= TTE_EVENT;
	UDPTLI_PRINTF("udptli_timer: ttp: %x\n", ttp);
	s = splnet();
	if (!(ttp->tt_event & TTE_ONQUEUE)) {
		udptli_insq(ttp);
	}
	(void) splx(s);
	(void) wakeup((caddr_t)&udptli_ttp);
	return;
}
