/*   "@(#)netdir.h	1.5 91/01/14  Copyright 1989 Sun Microsystems, Inc." */

/*
 * netdir.h
 *
 * This is the include file that defines various structures and
 * constants used by the netdir routines.
 */

#ifndef _netdir_h
#define _netdir_h

struct nd_addrlist {
	int 		n_cnt;		/* number of netbufs */
	struct netbuf 	*n_addrs;	/* the netbufs */
};

struct nd_hostservlist {
	int			h_cnt;		/* number of nd_hostservs */
	struct nd_hostserv	*h_hostservs;	/* the entries */
};

struct nd_hostserv {
	char		*h_host;	/* the host name */
	char		*h_serv;	/* the service name */
};

struct nd_mergearg {	/* parameter struct for option ND_MERGEADDR */
	char *s_uaddr;			/* input: server's univeral address */
	char *c_uaddr;			/* input: client's univeral address */
	char *m_uaddr;		/* output: merged univeral address */
};

#ifdef __cplusplus
extern "C" int netdir_getbyname (struct netconfig *, struct nd_hostserv *, struct nd_addrlist **);
extern "C" int netdir_getbyaddr (struct netconfig *, struct nd_hostservlist **, struct netbuf *);
extern "C" int netdir_options (struct netconfig *, int, int, char *);
extern "C" void netdir_free (char *, int);
extern "C" struct netbuf *uaddr2taddr (struct netconfig *, char *);
extern "C" char *taddr2uaddr (struct netconfig *, struct netbuf *);
extern "C" void netdir_perror (char *);
extern "C" char *netdir_sperror ();
extern "C" struct nd_addrlist *_netdir_getbyname (struct netconfig *, struct nd_hostserv *);
extern "C" struct nd_hostservlist *_netdir_getbyaddr (struct netconfig *, struct netbuf *);
extern "C" int _netdir_options (struct netconfig *, int, int, char *);
extern "C" struct netbuf *_uaddr2taddr (struct netconfig *, char *);
extern "C" char *_taddr2uaddr (struct netconfig *, struct netbuf *);
#elif (__STDC__ || c_plusplus)
int netdir_getbyname (struct netconfig *, struct nd_hostserv *, struct nd_addrlist **);
int netdir_getbyaddr (struct netconfig *, struct nd_hostservlist **, struct netbuf *);
int netdir_options (struct netconfig *, int, int, char *);
void netdir_free (char *, int);
struct netbuf *uaddr2taddr (struct netconfig *, char *);
char *taddr2uaddr (struct netconfig *, struct netbuf *);
void netdir_perror (char *);
struct nd_addrlist *_netdir_getbyname (struct netconfig *, struct nd_hostserv *);
struct nd_hostservlist *_netdir_getbyaddr (struct netconfig *, struct netbuf *);
int _netdir_options (struct netconfig *, int, int, char *);
struct netbuf *_uaddr2taddr (struct netconfig *, char *);
char *_taddr2uaddr (struct netconfig *, struct netbuf *);
#else
int netdir_getbyname ();
int netdir_getbyaddr ();
int netdir_options ();
void netdir_free ();
struct netbuf *uaddr2taddr ();
char *taddr2uaddr ();
void netdir_perror ();
char *netdir_sperror ();
struct nd_addrlist *_netdir_getbyname ();
struct nd_hostservlist *_netdir_getbyaddr ();
int _netdir_options ();
struct netbuf *_uaddr2taddr ();
char *_taddr2uaddr ();
#endif

#ifdef __STDC__
char *netdir_sperror (void);
#endif

#ifdef c_plusplus
char *netdir_sperror ();
#endif

/*
 * These are all objects that can be freed by netdir_free
 */
#define ND_HOSTSERV	0
#define ND_HOSTSERVLIST	1
#define ND_ADDR		2
#define ND_ADDRLIST	3

/*
 * These are the options for netdir_options
 */
#define ND_SET_BROADCAST	1
#define ND_SET_RESERVEDPORT	2
#define ND_CHECK_RESERVEDPORT	3
#define ND_MERGEADDR		4

/*
 * These are the various errors that can be encountered while attempting
 * to translate names to addresses. Note that none of them (except maybe
 * no memory) are truely fatal unless the ntoa deamon is on its last attempt
 * to translate the name. 
 *
 * Negative errors terminate the search resolution process, positive errors
 * are treated as warnings.
 */
#define ND_TRY_AGAIN	-5	/* Non-Authoritive Host not found, or SERVERFAIL */
#define ND_NO_RECOVERY	-4	/* Non recoverable errors, FORMERR, REFUSED, NOTIMP */
#define ND_NO_DATA	-3 /* Valid name, no data record of requested type */
#define ND_NO_ADDRESS	ND_NO_DATA	/* no address, look for MX record */
#define ND_BADARG	-2	/* Bad arguments passed 	*/
#define ND_NOMEM 	-1	/* No virtual memory left	*/
#define ND_OK		0	/* Translation successful	*/
#define ND_NOHOST	1	/* Hostname was not resolvable	*/
#define ND_NOSERV	2	/* Service was unknown		*/
#define ND_NOSYM	3	/* Couldn't resolve symbol	*/
#define ND_OPEN		4	/* File couldn't be opened	*/
#define ND_ACCESS	5	/* File is not accessable	*/
#define ND_UKNWN	6	/* Unknown object to be freed	*/
#define ND_NOCTRL	7       /* Unknown option passed to netdir_options */
#define ND_FAILCTRL	8       /* Option failed in netdir_options */
#define ND_SYSTEM	9       /* Other System error           */

/*
 * The following special case host names are used to give the underlying
 * transport provider a clue as to the intent of the request.
 */

#define HOST_SELF	"\\1"	/* The generic bind address for this tp */
#define HOST_ANY	"\\2"	/* A "don't care" option for the host   */
#define HOST_BROADCAST	"\\3"	/* The broadcast address for this tp    */

#endif /* !_netdir_h */
