/*	@(#)netconfig.h	1.9 91/01/14 SMI	*/

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */

#ifndef __NETCONFIG__
#define __NETCONFIG__

#define NETPATH		"NETPATH"
#define NETCONFIG	"/etc/netconfig"
#define MAXNETCONFIGLINE    1000


struct netconfig {
    char	    *nc_netid;	/* Network ID, including NULL terminator */
    unsigned long   nc_semantics;   /* Semantics */
    unsigned long   nc_flag;	    /* Flags */
    char	    *nc_protofmly;  /* Protocol family */
    char	    *nc_proto;	    /* Protocol name */
    char	    *nc_device;	    /* Full pathname of network device */
    unsigned long   nc_nlookups;    /* Number of directory lookup libraries */
    char	    **nc_lookups;   /* Full pathnames of the directory */
				    /*	lookups themselves. */
    unsigned long   nc_unused[8];   /* Reserved for future expansion */
};

typedef struct {
    struct netconfig **nc_head;
    struct netconfig **nc_curr;
} NCONF_HANDLE;

/* nc_semantics */
#define NC_TPI_CLTS	1	    /* Datagram */
#define NC_TPI_COTS	2	    /* Virtual Circuit */
#define NC_TPI_COTS_ORD	3

/* semantics as strings */
#define NC_TPI_CLTS_S	    "tpi_clts"
#define	NC_TPI_COTS_S	    "tpi_cots"
#define	NC_TPI_COTS_ORD_S   "tpi_cots_ord"


/* nc_flag */
#define NC_NOFLAG	0x0	    /* ??? */
#define NC_VISIBLE	0x1	    /* is visible */
#define NC_BROADCAST	0x2	    /* supports broadcast */
#define NC_TRANSPORT	0x4	    /* is a transport */
#define NC_SESSION	0x8	    /* is a session layer */
#define NC_PRESENTATION	0x10	    /* is a presentation layer */
#define NC_APPLICATION	0x20	    /* is an application layer */

/* flags as characters */
#define	NC_NOFLAG_C	'-'
#define	NC_VISIBLE_C	'v'
#define	NC_BROADCAST_C	'b'
    /* these are guesses: */
#define NC_TRANSPORT_C	't'
#define NC_SESSION_C	's'
#define NC_PRESENTATION_C 'p'
#define NC_APPLICATION_C 'a'


/* nc_protofmly */
#define NC_NOPROTOFMLY	"-"
#define NC_LOOPBACK	"loopback"
#define NC_INET		"inet"
#define NC_IMPLINK	"implink"
#define NC_PUP		"pup"
#define NC_CHAOS	"chaos"
#define NC_NS		"ns"
#define NC_NBS		"nbs"
#define NC_ECMA		"ecma"
#define NC_DATAKIT	"datakit"
#define NC_CCITT	"ccitt"
#define NC_SNA		"sna"
#define NC_DECNET	"decnet"
#define NC_DLI		"dli"
#define NC_LAT		"lat"
#define NC_HYLINK	"hylink"
#define NC_APPLETALK	"appletalk"
#define NC_NIT		"nit"
#define NC_IEEE802	"ieee802"
#define NC_OSI		"osi"
#define NC_X25		"x25"
#define NC_OSINET	"osinet"
#define NC_GOSIP	"gosip"

/* nc_proto */
#define NC_NOPROTO	"-"
#define NC_TCP		"tcp"
#define NC_UDP		"udp"
#define NC_ICMP		"icmp"


/* lookup libraries */
#define NC_NOLOOKUP	"-"


#ifdef __cplusplus
extern "C" void *setnetconfig ();
extern "C" struct netconfig *getnetconfig (void *);
extern "C" int endnetconfig (void *);
extern "C" struct netconfig *getnetconfigent (char *);
extern "C" int freenetconfigent (struct netconfig *);
extern "C" void *setnetpath ();
extern "C" struct netconfig *getnetpath (void *);
extern "C" int endnetpath (void *);
#elif c_plusplus
extern void *setnetconfig ();
extern struct netconfig *getnetconfig (void *);
extern int endnetconfig (void *);
extern struct netconfig *getnetconfigent (char *);
extern int freenetconfigent (struct netconfig *);
extern void *setnetpath ();
extern struct netconfig *getnetpath (void *);
extern int endnetpath (void *);
#elif __STDC__
extern void *setnetconfig (void);
extern struct netconfig *getnetconfig (void *);
extern int endnetconfig (void *);
extern struct netconfig *getnetconfigent (char *);
extern int freenetconfigent (struct netconfig *);
extern void *setnetpath (void);
extern struct netconfig *getnetpath (void *);
extern int endnetpath (void *);
#else
extern void *setnetconfig ();
extern struct netconfig *getnetconfig ();
extern int endnetconfig ();
extern struct netconfig *getnetconfigent ();
extern int freenetconfigent ();
extern void *setnetpath ();
extern struct netconfig *getnetpath ();
extern int endnetpath ();
#endif

#endif /* ndef __NETCONFIG__ */
