/*
 * Basic console driver for apa8 (6153) display adapter in the IBM RT PC.
 * This driver takes minimal advantage of the hardware support provided 
 * by the apa8 to do bit manipulation.  A significant speed advantage can
 * be obtained if the driver is modified to use the hardware support provided
 * by the card.
 *
 * $Source: /usr/src/sys.44l/rt.build/dev/RCS/apa8_driver.c,v $
 * $Revision: 1.1 $ $Date: 1994/10/23 20:07:48 $
 * $Author: md $ $State: Exp $
 *
 * $Log: apa8_driver.c,v $
 * Revision 1.1  1994/10/23  20:07:48  md
 * Initial revision
 *
 *
 */

#include "cons.h"
#include "apa8_font.h"

#undef	APA8_DEBUG

/*
 * Debugging support
 */
#ifdef	APA8_DEBUG
#define	DPRINTF(cond, stmt) {if (apa8_debug & cond) {stmt;}}

#define	APA8_DEBUG_BLIT		0x0001	/* bit blitting code */
#define	APA8_DEBUG_EMUL		0x0002	/* adm3a emulator */
#define	APA8_DEBUG_SCROLL	0x0004	/* scrolling routines */

int apa8_debug=0;
#else
#define	DPRINTF(cond, stmt)
#endif

static int apa8_row;	/* current row the cursor is at */
static int apa8_col;	/* current column the cursor is at */
static int apa8_cursor;	/* is cursor currently visible */

/*
 * Move a single line in the visible display area of the apa8.  This
 * routine is optimised for speed.
 */
static void
apa8_screen_line_move(to, from)
unsigned short *to, *from;
{
	*to=*from; to+=2; from+=2;	/* word  1 */
	*to=*from; to+=2; from+=2;	/* word  2 */
	*to=*from; to+=2; from+=2;	/* word  3 */
	*to=*from; to+=2; from+=2;	/* word  4 */
	*to=*from; to+=2; from+=2;	/* word  5 */
	*to=*from; to+=2; from+=2;	/* word  6 */
	*to=*from; to+=2; from+=2;	/* word  7 */
	*to=*from; to+=2; from+=2;	/* word  8 */
	*to=*from; to+=2; from+=2;	/* word  9 */
	*to=*from; to+=2; from+=2;	/* word 10 */
	*to=*from; to+=2; from+=2;	/* word 11 */
	*to=*from; to+=2; from+=2;	/* word 12 */
	*to=*from; to+=2; from+=2;	/* word 13 */
	*to=*from; to+=2; from+=2;	/* word 14 */
	*to=*from; to+=2; from+=2;	/* word 15 */
	*to=*from; to+=2; from+=2;	/* word 16 */
	*to=*from; to+=2; from+=2;	/* word 17 */
	*to=*from; to+=2; from+=2;	/* word 18 */
	*to=*from; to+=2; from+=2;	/* word 19 */
	*to=*from; to+=2; from+=2;	/* word 20 */
	*to=*from; to+=2; from+=2;	/* word 21 */
	*to=*from; to+=2; from+=2;	/* word 22 */
	*to=*from; to+=2; from+=2;	/* word 23 */
	*to=*from; to+=2; from+=2;	/* word 24 */
	*to=*from; to+=2; from+=2;	/* word 25 */
	*to=*from; to+=2; from+=2;	/* word 26 */
	*to=*from; to+=2; from+=2;	/* word 27 */
	*to=*from; to+=2; from+=2;	/* word 28 */
	*to=*from; to+=2; from+=2;	/* word 29 */
	*to=*from; to+=2; from+=2;	/* word 30 */
	*to=*from; to+=2; from+=2;	/* word 31 */
	*to=*from; to+=2; from+=2;	/* word 32 */
	*to=*from; to+=2; from+=2;	/* word 33 */
	*to=*from; to+=2; from+=2;	/* word 34 */
	*to=*from; to+=2; from+=2;	/* word 35 */
	*to=*from; to+=2; from+=2;	/* word 36 */
	*to=*from; to+=2; from+=2;	/* word 37 */
	*to=*from; to+=2; from+=2;	/* word 38 */
	*to=*from; to+=2; from+=2;	/* word 39 */
	*to=*from; to+=2; from+=2;	/* word 40 */
	*to=*from; to+=2; from+=2;	/* word 41 */
	*to=*from; to+=2; from+=2;	/* word 42 */
	*to=*from; to+=2; from+=2;	/* word 43 */
	*to=*from; to+=2; from+=2;	/* word 44 */
	*to=*from; 			/* word 45 */
}

/*
 * Fill a single line in the visible display area of the apa8 with zeroes.
 * This routine is optimised for speed.
 */
static void
apa8_screen_line_fill(where)
unsigned short *where;
{
	*where=0x0000; where+=2;	/* word  1 */
	*where=0x0000; where+=2;	/* word  2 */
	*where=0x0000; where+=2;	/* word  3 */
	*where=0x0000; where+=2;	/* word  4 */
	*where=0x0000; where+=2;	/* word  5 */
	*where=0x0000; where+=2;	/* word  6 */
	*where=0x0000; where+=2;	/* word  7 */
	*where=0x0000; where+=2;	/* word  8 */
	*where=0x0000; where+=2;	/* word  9 */
	*where=0x0000; where+=2;	/* word 10 */
	*where=0x0000; where+=2;	/* word 11 */
	*where=0x0000; where+=2;	/* word 12 */
	*where=0x0000; where+=2;	/* word 13 */
	*where=0x0000; where+=2;	/* word 14 */
	*where=0x0000; where+=2;	/* word 15 */
	*where=0x0000; where+=2;	/* word 16 */
	*where=0x0000; where+=2;	/* word 17 */
	*where=0x0000; where+=2;	/* word 18 */
	*where=0x0000; where+=2;	/* word 19 */
	*where=0x0000; where+=2;	/* word 20 */
	*where=0x0000; where+=2;	/* word 21 */
	*where=0x0000; where+=2;	/* word 22 */
	*where=0x0000; where+=2;	/* word 23 */
	*where=0x0000; where+=2;	/* word 24 */
	*where=0x0000; where+=2;	/* word 25 */
	*where=0x0000; where+=2;	/* word 26 */
	*where=0x0000; where+=2;	/* word 27 */
	*where=0x0000; where+=2;	/* word 28 */
	*where=0x0000; where+=2;	/* word 29 */
	*where=0x0000; where+=2;	/* word 30 */
	*where=0x0000; where+=2;	/* word 31 */
	*where=0x0000; where+=2;	/* word 32 */
	*where=0x0000; where+=2;	/* word 33 */
	*where=0x0000; where+=2;	/* word 34 */
	*where=0x0000; where+=2;	/* word 35 */
	*where=0x0000; where+=2;	/* word 36 */
	*where=0x0000; where+=2;	/* word 37 */
	*where=0x0000; where+=2;	/* word 38 */
	*where=0x0000; where+=2;	/* word 39 */
	*where=0x0000; where+=2;	/* word 40 */
	*where=0x0000; where+=2;	/* word 41 */
	*where=0x0000; where+=2;	/* word 42 */
	*where=0x0000; where+=2;	/* word 43 */
	*where=0x0000; where+=2;	/* word 44 */
	*where=0x0000; 			/* word 45 */
}

/*
 * Clear the apa8 visible bitmap
 */
static void
apa8_clear_screen()
{
	int lines=APA8_Y_BITS;
	unsigned short *p=(unsigned short *)APA8_DISPLAY;

	/* just pass data to/from apa8 directly (via register A) */
	SET_APA8_DATAMASK(0xFF00);	/* data mask A */
	SET_APA8_WRITEMASK(0x0000);	/* Write everything through */
	SET_APA8_CONTROL_REG(APA8CR_SYNC | APA8CR_VIDEO | APA8CR_MM_SWR |
			     APA8CR_X | APA8CR_LU_PA);
	while(lines--) {
	    apa8_screen_line_fill(p);	/* fill each line */
	    p+=APA8_X_BYTES;
	}
}

/*
 * Calculate the upper left corner location of the bounding box for the
 * character located at the specified row and column.  The location is
 * specified as a word address plus a bit offset within the word.  The
 * location is relative to address zero.
 */
static void
apa8_bounding_box(row, col, addr, offset)
int row, col;
int *addr, *offset;
{
	int bits;

	bits = (row*APA8_CHAR_MATRIX_Y*(APA8_X_BITS+APA8_X_BITS_HIDDEN)) +
	       (col*APA8_CHAR_MATRIX_X);
	*addr=(bits/8)&0xFFFE;	/* addr must be on an even boundry & < 64K */
	*offset=bits-(*addr*8);
}

/*
 * Copy the bitmap specified by "data" to the bounding box whose upper left
 * corner is located at the offset specified by "addr" and "offset" (see
 * description of apa8_bounding_box() for location details).  A cursor
 * is added/removed if "cursor" is non-zero.  Specifying "data" as NULL
 * will suppress loading a bitmap.
 *
 * NOTE: although this routine is somewhat generic, it still assumes that
 *	 the bitmap is no wider than 8 bits (ie. a char).  No apa8 specific
 *	 hardware support is used so that we can hopefully use this with
 *	 other adapters such as the apa16 in the future.
 */
static void
apa8_char_blit(addr, offset, data, cursor)
int addr;
int offset;
const unsigned char* data;
int cursor;
{
	int i;
	register unsigned short *scr_addr;
	unsigned short mask1, xmask1;
	unsigned short mask2, xmask2;

	/* just pass data to/from apa8 directly (via register A) */
	SET_APA8_DATAMASK(0xFF00);	/* data mask A */
	SET_APA8_WRITEMASK(0x0000);	/* Write everything through */
	SET_APA8_CONTROL_REG(APA8CR_SYNC | APA8CR_VIDEO | APA8CR_MM_SWR |
			     APA8CR_X | APA8CR_LU_PA);
	if (offset <= (16-APA8_CHAR_MATRIX_X)) { /* can we do it in one word? */
	    int shift_amount = 16-offset-APA8_CHAR_MATRIX_X+1;

	    mask1 = 0xFF << shift_amount;
	    xmask1 = mask1^0xFFFF;
	    DPRINTF(APA8_DEBUG_BLIT,
	        printf("addr=%x  offset=%x  shift_amount=%d  xmask1=%x\n", 
	        addr, offset, shift_amount, xmask1));
	    for (i=0; i<APA8_CHAR_MATRIX_Y; i++) {
		scr_addr = (unsigned short *)APA8_DISPLAY+addr +
			   (i*((APA8_X_BITS+APA8_X_BITS_HIDDEN)/8));
		if (data) {
		    *scr_addr &= xmask1;		/* all bits off first */
		    *scr_addr |= (*data++)<<shift_amount; /* add in data */
		}
		if (cursor)
		    *scr_addr ^= mask1;		/* reverse the bits */
		DPRINTF(APA8_DEBUG_BLIT, printf("addr = %x\n", scr_addr));
	    }
	} else {	/* character spans two words */
	    int bits_word1 = 16-offset;
	    int shift_amount_word2 = 16-(APA8_CHAR_MATRIX_X-bits_word1)+1;

	    DPRINTF(APA8_DEBUG_BLIT, printf("*** spanning ***\n"));
	    DPRINTF(APA8_DEBUG_BLIT,
		printf("addr=%x+%x bits_word1=%d shift_amount_word2=%d\n",
			addr, offset, bits_word1,shift_amount_word2));
	    mask1 = (1 << bits_word1)-1;
	    xmask1 = mask1^0xFFFF;
	    mask2 = 0xFF << shift_amount_word2;
	    xmask2 = mask2^0xFFFF;
	    DPRINTF(APA8_DEBUG_BLIT,
		printf("xmask1=%x  xmask2=%x\n", xmask1, xmask2));
	    for (i=0; i<APA8_CHAR_MATRIX_Y; i++) {
		scr_addr = (unsigned short *)APA8_DISPLAY+addr +
			   (i*((APA8_X_BITS+APA8_X_BITS_HIDDEN)/8));
		if (data) {
		    *scr_addr &= xmask1;		/* all bits off first */
		    *scr_addr |= (*data)>>(8-bits_word1); /* add in data */
		}
		if (cursor)
		    *scr_addr ^= mask1;		/* reverse the bits */
		scr_addr+=2;
		if (data) {
		    *scr_addr &= xmask2;		/* all bits off first */
		    *scr_addr |= (unsigned short)((*data++)<<shift_amount_word2);
		}
		if (cursor)
		    *scr_addr ^= mask2;		/* reverse the bits */
	    }
	}
}

/*
 * Scroll the screen up the specified number of rows.
 */
static void
apa8_scroll_up(rows)
int rows;
{
	unsigned short *from, *to;
	int lines=APA8_Y_BITS-rows*APA8_CHAR_MATRIX_Y;

	/* just pass data to/from apa8 directly (via register A) */
	SET_APA8_DATAMASK(0xFF00);	/* data mask A */
	SET_APA8_WRITEMASK(0x0000);	/* Write everything through */
	SET_APA8_CONTROL_REG(APA8CR_SYNC | APA8CR_VIDEO | APA8CR_MM_SWR |
			     APA8CR_X | APA8CR_LU_PA);
	from = (unsigned short *)(APA8_DISPLAY)+
		(rows*APA8_CHAR_MATRIX_Y*APA8_X_BYTES);
	to = (unsigned short *)APA8_DISPLAY;
	DPRINTF(APA8_DEBUG_SCROLL,
		printf("apa8_scroll_up: from=%x  to=%x  lines=%d\n",
		from, to, lines));
	while(lines--) {
	    DPRINTF(APA8_DEBUG_SCROLL,
		printf("apa8_scroll_up: lines left=%d  from=%x  to=%x\n",
		lines, from, to));
	    apa8_screen_line_move(to, from);
	    to+=APA8_X_BYTES;
	    from+=APA8_X_BYTES;
	}
	lines=rows*APA8_CHAR_MATRIX_Y;
	while(lines--) {
	    apa8_screen_line_fill(to);
	    to+=APA8_X_BYTES;
	}
}

/*
 * Move the cursor to the right taking into account scrolling the screen
 * if needed.
 */
static void
apa8_cursor_right()
{
	if (++apa8_col >= 80) {
	    apa8_col=0;
	    if (++apa8_row >= 32) {
		apa8_row=31;
		apa8_scroll_up(1);	/* scroll up one row */
	    }
	}
}

/*
 * Move the cursor to the left stopping at the top left of the screen.
 */
static void
apa8_cursor_left()
{
	if (--apa8_col < 0) {
	    apa8_col=79;
	    if (--apa8_row < 0)
		apa8_row=apa8_col=0;
	}
}

/*
 * Turn the cursor on at the current position if it's off.
 */
static void
apa8_cursor_on()
{
	int box_addr, box_offset;

	if (!apa8_cursor) {
	    apa8_bounding_box(apa8_row, apa8_col, &box_addr, &box_offset);
	    apa8_char_blit(box_addr, box_offset, 0, 1);
	    apa8_cursor^=1;
	}
}

/*
 * Turn the cursor off at the current position if it's on.
 */
static void
apa8_cursor_off()
{
	int box_addr, box_offset;

	if (apa8_cursor) {
	    apa8_bounding_box(apa8_row, apa8_col, &box_addr, &box_offset);
	    apa8_char_blit(box_addr, box_offset, 0, 1);
	    apa8_cursor^=1;
	}
}

/*
 * Initialise the apa8 display subsystem
 */
void
apa8_init()
{
	apa8_row=0;	/* home cursor */
	apa8_col=0;
	apa8_cursor=0;	/* turn it off */
	apa8_clear_screen();
	apa8_cursor_on();
	SET_APA8_DATAMASK(0xFF00);	/* data mask A */
	SET_APA8_WRITEMASK(0x0000);	/* Write everything through */
	SET_APA8_CONTROL_REG(APA8CR_SYNC | APA8CR_VIDEO | APA8CR_MM_SWR |
			     APA8CR_X | APA8CR_LU_PA);
}

/*
 * Probe for the precence of an apa8 display adapter
 */
int
apa8_probe()
{
	apa8_init();	/* need to set up registers properly first */
	*(unsigned short *)(APA8_DISPLAY+((0xFFFF-1)<<1))=0xAAAA;
	return((*(unsigned short *)(APA8_DISPLAY+((0xFFFF-1)<<1)) == 0xAAAA)
		? 1 : 0);
}

#define	ADM3A_UP	0x0b	/* ^K */
#define	ADM3A_DOWN	0x0a	/* ^J */
#define	ADM3A_RIGHT	0x0c	/* ^L */
#define	ADM3A_LEFT	0x08	/* ^H */
#define	ADM3A_CLEAR	0x1a	/* ^Z */
#define	ADM3A_HOME	0x1e	/* ^^ */
#define	ADM3A_POS	0x1b	/* esc (lead in only) */
#define	ADM3A_CR	0x0d	/* carriage return */
#define	ADM3A_TAB	0x09	/* ^H */

/*
 * Put a single character on the apa8 display at the current row and column.
 */
void
apa8_putchar(c)
const char c;
{
	int box_addr, box_offset;
	static int positioning=0;

	if (positioning) {
	    switch(positioning) {
		case 1:		/* next char must be an '=' */
		    if (c != '=') {
			break;		/* abort positioning sequence */
		    } else {
			apa8_cursor_off();
			positioning++;	/* need one more char */
			return;
		    }
		case 2:
		    if ((apa8_row=c-' ') < 0)
			apa8_row=0;
		    if (apa8_row >= 32)
			apa8_row=31;
		    positioning++;	/* need one more char */
		    return;
		case 3:
		    if ((apa8_col=c-' ') < 0)
			apa8_col=0;
		    if (apa8_col >= 80)
			apa8_col=79;
		    apa8_cursor_on();
		    break;
		default:	/* this should NOT happen, bail out */
		    break;
	    }
	    positioning=0;
	    return;
	}
	switch (c) {
	    case ADM3A_UP:
		DPRINTF(APA8_DEBUG_EMUL,
			printf("amd3a emulator: up\n"));
		apa8_cursor_off();
		if (--apa8_row < 0)
		    apa8_row=0;
		apa8_cursor_on();
		break;
	    case ADM3A_DOWN:
		DPRINTF(APA8_DEBUG_EMUL,
			printf("amd3a emulator: down\n"));
		apa8_cursor_off();
		if (++apa8_row >= 32) {
		    apa8_row=31;
		    apa8_scroll_up(1);	/* scroll up one row */
		}
		apa8_cursor_on();
		break;
	    case ADM3A_RIGHT:
		DPRINTF(APA8_DEBUG_EMUL,
			printf("amd3a emulator: right\n"));
		apa8_cursor_off();
		apa8_cursor_right();
		apa8_cursor_on();
		break;
	    case ADM3A_LEFT:
		DPRINTF(APA8_DEBUG_EMUL,
			printf("amd3a emulator: left\n"));
		apa8_cursor_off();
		apa8_cursor_left();
		apa8_cursor_on();
		break;
	    case ADM3A_CLEAR:
		DPRINTF(APA8_DEBUG_EMUL,
			printf("amd3a emulator: clearing screen\n"));
		apa8_row=0;	/* home cursor */
		apa8_col=0;
		apa8_cursor=0;	/* turn it off */
		apa8_clear_screen();
		apa8_cursor_on();
		break;
	    case ADM3A_HOME:
		DPRINTF(APA8_DEBUG_EMUL,
			printf("amd3a emulator: homing cursor\n"));
		apa8_cursor_off();
		apa8_row=apa8_col=0;
		apa8_cursor_on();
		break;
	    case ADM3A_POS:
		DPRINTF(APA8_DEBUG_EMUL,
			printf("amd3a emulator: start positioning cursor\n"));
		positioning=1;
		break;
	    case ADM3A_CR:
		DPRINTF(APA8_DEBUG_EMUL,
			printf("amd3a emulator: carriage return\n"));
		apa8_cursor_off();
		apa8_col=0;
		apa8_cursor_on();
		break;
	    case ADM3A_TAB:
		DPRINTF(APA8_DEBUG_EMUL,
			printf("amd3a emulator: tab\n"));
		DPRINTF(APA8_DEBUG_EMUL,
			printf("amd3a emulator: start col=%d\n", apa8_col));
		apa8_cursor_off();
		if (apa8_col == 79)
		    apa8_cursor_right();
		apa8_col=((apa8_col+1+8)/8)*8-1;
		DPRINTF(APA8_DEBUG_EMUL,
			printf("amd3a emulator: end col=%d\n", apa8_col));
		apa8_cursor_on();
		break;
	    default:	/* normal chars go through here */
		apa8_cursor_off();
		apa8_bounding_box(apa8_row, apa8_col, &box_addr, &box_offset);
				/* blit char but don't supply a cursor */
		apa8_char_blit(box_addr, box_offset, 
			       apa8_bitmaps+*(apa8_fontindex+c), 0);
		apa8_cursor_right();
		apa8_cursor_on();
		break;
	}
}
