/*
 * Copyright (c) 1988 University of Utah.
 * Copyright (c) 1982, 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * from: Utah $Hdr: cpu.h 1.16 91/03/25$
 *
 *	@(#)cpu.h	8.4 (Berkeley) 1/5/94
 */

#ifndef	_CPU_MACHINE_
#define	_CPU_MACHINE_

#include <sys/cdefs.h>

/*
 * Exported definitions unique to ROMP cpu support.
 */

/*
 * defines for CPU types
 */

#define CPU_OLD		0	/* original RT processor, not supported */
#define CPU_APC		1	/* standard RT APC processor */
#define	CPU_EAPC	2	/* Enhanced APC, aka the MOD135 processor */

/*
 * ROMP mmu definitions
 */

/* structure of a single HAT/IPT entry */

__BEGIN_C_ONLY
struct hatipt {
	unsigned char  key	 :2;	/* protection bits */
					/* 00 - private read/write */
					/* 01 - global read, private write */
					/* 10 - global read/write */
					/* 11 - global read-only */
	unsigned char  reserved1 :1;
	unsigned short sid	 :12;	/* segment id */
	unsigned int   vpage	 :17;	/* virtual page index (2K page) */
	unsigned char  hat_empty :1;	/* hat pointer valid? */
	unsigned char  reserved2 :2;
	unsigned short hat_ptr   :13;	/* hat pointer */
	unsigned char  ipt_last	 :1;	/* ipt pointer valid? */
	unsigned char  reserved3 :2;
	unsigned short ipt_ptr	 :13;	/* ipt pointer */
	unsigned char  reserved4 :7;
	unsigned char  write_en	 :1;	/* write enable for special segments */
	unsigned short tid	 :8;	/* TID for special segments */
	unsigned short locks	 :16;	/* lock bits for special segments */
	unsigned short reserved5 :16;	/* what a waste! */
	unsigned short rev_ptr   :16;	/* is this really true? */
};

struct hatipt_coarse {		/* same as above but not at detailed */
	unsigned long  key_addrtag;
	unsigned short hat_ptr;
	unsigned short ipt_ptr;
	unsigned char  w;
	unsigned char  tid;
	unsigned short lockbits;
	unsigned short reserved;
	unsigned short rev_ptr;
};

extern struct hatipt_coarse *ROMP_HATIPT; /* pointer to actual HAT/IPT table */
__END_C_ONLY

#define HATIPTSIZE	16		/* size of one HAT IPT entry */

/* offset of various words in the hatipt */

#define IPTADDRTAG	0		/* offset of IPT ADDR TAG field */
#define IPTHATPTR	4		/* offset of IPT HAT PTR field */
#define IPTIPTPTR	6		/* offset of IPT IPT PTR field */
#define IPTLOCK		8		/* offset of IPT LOCK WORD field */

/* the hat link valid bit */

#define	IPT_HATLINK		0x8000
#define IPT_HATLINK_MASK	0x7fff

#define ROMP_ENDCHAIN(ptr) ((ptr)&0x8000)  /* true if at end of ipt chain */

/* virtual page index field */

#define ROMP_VPAGE_BITS 17
#define ROMP_VPAGE_MASK ((1<<ROMP_VPAGE_BITS)-1)

/* address and tag field */

#define ROMP_ADDRTAG_MASK 0x1FFFFFFF
#define ROMP_SEGSHIFT 24

/* available types of protections in the key bits */

#define ROMP_KEY_SHIFT 30
#define ROMP_KEY_BITS 0xC0000000
#define ROMP_KEY_KW   0  /* kernel write, user none */
#define ROMP_KEY_URKW 1  /* kernel write, user read */
#define ROMP_KEY_UW   2  /* kernel write, user write */
#define ROMP_KEY_URKR 3  /* kernel read, user read */

/* breakdown of bits in the hardware segment registers */

#define ROMP_SID_MASK	(0xFFF<<ROMP_VPAGE_BITS)
#define ROMP_SID_SYSTEM	0x200		/* segment used by the kernel */
#define ROMP_SID_UNUSED	0x0
#define ROMP_SEG_PRESENT 0x10000	/* bit indicating segment is present */
#define ROMP_SEG_P	0x10000		/* 1=segment present */
#define ROMP_SEG_C	0x08000		/* 1=no cpu access */
#define ROMP_SEG_I	0x04000		/* 1=no I/O access */
#define ROMP_SEG_SID	0x03ffc		/* mask for Segment ID */
#define ROMP_SEG_S	0x00002		/* 1=special, 0=normal */
#define ROMP_SEG_K	0x00001		/* Protect key bit, 1=public rules */
#define ROMP_SID_SHIFT	2		/* shift to get SID */

#define ROMP_NSEGREGS	16		/* 16 segment registers */
#define ROMP_SEGREGSTEP	1
#define	ROMP_NSEGS	4096		/* total number of virtual segments */
#define	SYS_SEG		0x0E		/* segment kernel lives in */
#define	NUSR_SEGS	SYS_SEG		/* Hacky, but correct */

#define SYSBASE_ADDR	0xe0000000	/* virt addr of system origin */
#define SYSBASE		((u_long)SYSBASE_ADDR)

#define ROMP_UNDEF_PTR	0x97FF	/* undefined pointer value for debugging */

/* Addresses in the I/O space which control the ROMP's mmu */

#define ROMPBASE	0x810000
#define ROMP_SEGREG0	ROMPBASE	/* segment register 0 */
#define ROMP_IOBR	ROMPBASE+0x10	/* I/O base address register */
#define ROMP_SER	ROMPBASE+0x11	/* Memory exception register */
#define ROMP_SEAR	ROMPBASE+0x12	/* Memory exception addr register */
#define ROMP_TRAR	ROMPBASE+0x13	/* Translated real addr register */
#define ROMP_TIDR	ROMPBASE+0x14	/* Transaction ID register */
#define ROMP_TCR	ROMPBASE+0x15	/* Translation control register */
#define ROMP_RAM	ROMPBASE+0x16	/* RAM specification register */
#define	ROMP_ROM	ROMPBASE+0x17	/* ROM specification register */
#define ROMP_INV_TLB	ROMPBASE+0x80	/* Invalidate entire TLB */
#define ROMP_INV_SEG	ROMPBASE+0x81	/* Invalidate TLB for a segment */
#define ROMP_INV_ADDR	ROMPBASE+0x82	/* invalidate TLB for an address */
#define ROMP_VTOP	ROMPBASE+0x83	/* Load real address */
#define ROMP_REFBITS	ROMPBASE+0x1000	/* page reference bits */
#define ROMP_MODBITS	ROMPBASE+0x1000	/* page modify bits */

/* bits in the translation control register */

#define TCR_S	0x00000100	/* 4K page bit in TCR */
#define TCR_V	0x00008000	/* Virtual = Real mode bit in TCR */
#define	TCR_I	0x00001000	/* TLIPT bit in TCR */

/* macros for extracting/adding segment values to an address */

#define atoseg(a)	((unsigned)(a)>>28)
#define	segtoa(s)	((unsigned)(s)<<28)
#define segoffset(a)	((a) & 0x0fffffff)

/* macros for fiddling with the hardware segment registers */

#define	get_rosetta_seg(seg_number)	(ior(ROMP_SEGREG0 + (seg_number)))
#define	set_rosetta_seg(seg_number,val)	(iow(ROMP_SEGREG0 + (seg_number),(val)))
#define	sid_to_seg(sid)			(((sid)<<2) | ROMP_SEG_PRESENT)

/* macros for converting between an address and a page number */

#define	a_to_rosetta_page(addr) ((addr) >> rose_page_shift)
#define vm_to_rosetta_page(addr) ((addr) << (PAGE_SHIFT - rose_page_shift))

/* macros for fiddling with the tlb */

#define	tlb_invalidate_seg(seg_number)	(iow(ROMP_INV_SEG,(seg_number)<<28))
#define	tlb_invalidate_addr(address)	(iow(ROMP_INV_ADDR,(address)))
#define tlb_invalidate_all()		(iow(ROMP_INV_TLB,1))

/* macros for doing virtual <-> physical address translation */

#define	rosetta_write_lrar(addr)	(iow(ROMP_VTOP,(addr)))
#define	rosetta_read_trar()		(ior(ROMP_TRAR))

/* macros for setting/clearing modify and reference bits */

#define get_mod_bit(rpage)	(ior(ROMP_MODBITS+(rpage))&1)
#define set_mod_bit(rpage,v)	(iow(ROMP_MODBITS+(rpage),\
				(ior(ROMP_MODBITS+(rpage))&2)|(v)))

#define get_ref_bit(rpage)	((ior(ROMP_REFBITS+(rpage))>>1)&1)
#define set_ref_bit(rpage,v)	(iow(ROMP_REFBITS+(rpage),\
				(ior(ROMP_REFBITS+(rpage))&1)|((v)<<1)))

#define set_refmod_bits(rpage,ref,mod)	iow(ROMP_REFBITS+(rpage),\
					    ((ref)<<1)|(mod))

/* macros for getting and setting HAT and IPT pointers */

#define get_hatptr(ipte)	((ipte)->hat_ptr & 0x9FFF)
#define set_hatptr(ipte,v)	((ipte)->hat_ptr = (v))

#define get_iptptr(ipte)	((ipte)->ipt_ptr & 0x9FFF)
#define set_iptptr(ipte,v)	((ipte)->ipt_ptr = (v))

__BEGIN_C_ONLY
extern int ROMP_HASHMASK;	/* defined by lohatipt.s */
__END_C_ONLY

#define ROMP_HASH(seg,vpage)	(((seg) ^ (vpage)) & ROMP_HASHMASK)

#endif /* _CPU_MACHINE_ */
