#include <stdio.h>

struct hatipt {
	unsigned char  key	 :2;	/* protection bits */
					/* 00 - private read/write */
					/* 01 - global read, private write */
					/* 10 - global read/write */
					/* 11 - global read-only */
	unsigned char  reserved1 :1;
	unsigned short sid	 :12;	/* segment id */
	unsigned int   vpage	 :17;	/* virtual page index (2K page) */
	unsigned char  hat_empty :1;	/* hat pointer valid? */
	unsigned char  reserved2 :2;
	unsigned short hat_ptr   :13;	/* hat pointer */
	unsigned char  ipt_last	 :1;	/* ipt pointer valid? */
	unsigned char  reserved3 :2;
	unsigned short ipt_ptr	 :13;	/* ipt pointer */
	unsigned char  reserved4 :7;
	unsigned char  write_en	 :1;	/* write enable for special segments */
	unsigned short tid	 :8;	/* TID for special segments */
	unsigned short locks	 :16;	/* lock bits for special segments */
	unsigned short reserved5 :16;	/* what a waste! */
	unsigned short rev_ptr   :16;	/* what a waste! */
  };


int coremap[8192];

dump_hatipt(hatipt_table, nentries, map)
struct hatipt *hatipt_table;
int nentries, map;
{
	struct hatipt *p=hatipt_table;
	int max_llsize=0, min_llsize=1024, s;
	int pages_used=0;
	int hat=0, hat_used=0, hat_free=0;
	int i;

	printf("processing %d entries\n", nentries);
	bzero(coremap, sizeof(coremap));
	while(nentries--) {
	    if (!p->hat_empty) {
		hat_used++;
		if (!map)
		    printf("[0x%03x] ptr=0x%04x\n", hat, p->hat_ptr);
		if ((s=dump_ipt_pointers(hatipt_table, p->hat_ptr, hat, map))) {
		    max_llsize = s > max_llsize ? s : max_llsize;
		    min_llsize = s < min_llsize ? s : min_llsize;
		}
		pages_used+=s;
	    } else
		hat_free++;
	    p++;
	    hat++;
	}
	printf("\nHAT/IPT Statistics:\n");
	printf("    Total HAT entries:      %d\n", hat_used+hat_free);
	printf("                 Used:      %d\n", hat_used);
	printf("                 Free:      %d\n", hat_free);
	printf("    Maximum IPT chain size: %d\n", max_llsize);
	printf("    Minimum IPT chain size: %d\n", min_llsize);
	printf("    Active pages:           %d\n", pages_used);
	printf("\nCoremap (in pages):\n");
	for (i=0; i<8192; i++) {
	    if (i%16 == 0)
		printf(" ");
	    if (i%64 == 0)
		printf("\n[0x%04x] ", i);
	    if (coremap[i] > 1)
		printf(" ");	/* umm, something is *severely* wrong! */
	    else
		printf("%1d", coremap[i] ? 1 : 0);

	}
	printf("\n");
}

dump_ipt_pointers(hatipt_table, iptr, hat, map)
struct hatipt *hatipt_table;
int iptr, hat, map;
{
	int p=iptr;
	int size=0;
	const char *key_prot[] = { "G:none P:r/w  ",
				   "G:read P:write",
				   "G:r/w  P:---- ",
				   "G:read P:---- " };

	if (iptr > 8192) {
	    fprintf(stderr, "***WARNING: initial ipt pointer from hat is too large!\n");
	    return(0);
	}
	for (;;) {
	    if (((hatipt_table+p)->sid^(hatipt_table+p)->vpage&0x1fff) != hat) {
	    	fprintf(stderr, "***WARNING: hatipt entry at 0x%04x is on the wrong hash chain!\n", p);
		fprintf(stderr, "            hat=%x  sid=%x  vpage=%x\n", hat, 
			(hatipt_table+p)->sid, (hatipt_table+p)->vpage);
		return(0);
	    }
	    size++;
	    if (!map)
		printf("  [0x%04x] 0x%03x/%05x  key=%s we=%x next=0x%04x rev=0x%04x\n",
		p, (hatipt_table+p)->sid, (hatipt_table+p)->vpage,
		key_prot[(hatipt_table+p)->key], (hatipt_table+p)->write_en, 
		(hatipt_table+p)->ipt_ptr, (hatipt_table+p)->rev_ptr);
	    coremap[p]++;
	    if ((hatipt_table+p)->ipt_last)
		break;
	    p=(hatipt_table+p)->ipt_ptr;
	    if (size >= 128) {
		fprintf(stderr, "\n***WARNING: loop in IPT pointer list!\n");
		size=0;
		break;
	    }
	}
	return(size);
}
