/*
 * Declaration of external interfaces to the various display adapters
 * available on the RT PC.
 *
 * $Source: /sys/rt/dev/RCS/apa8_cons.h,v $
 * $Revision: 1.1 $ $Date: 1994/10/23 20:13:09 $
 * $Author: md $ $State: Exp $
 *
 * $Log: apa8_cons.h,v $
 * Revision 1.1  1994/10/23  20:13:09  md
 * Initial revision
 *
 *
 */

#ifndef	_RT_CONS_H_
#define	_RT_CONS_H_

#include <sys/cdefs.h>

/*
 * Definitions for the apa8 (6153) and apa8c (6154) displays.
 */

#define APA8_X_BITS		720	/* screen dimensions in pixels */
#define APA8_Y_BITS		512
#define APA8_X_BITS_HIDDEN	304	/* extra pixels on X axis */
#define APA8_X_WORDS		64	/* number of words in one scan line */
#define APA8_X_BYTES		128	/* number of bytes in one scan line */

#define	APA8_CHAR_MATRIX_X	9	/* use a 9x16 character matrix */
#define	APA8_CHAR_MATRIX_Y	16

			/* I/O addresses used by the apa8 & apa8c cards */
#define APA8_STATUS	0xF0000150	/* RAS status register */
#define APA8_BLK_RELOAD	0xF0000152	/* Block transfer reload */
#define APA8_CONTROL	0xF0000160	/* Control register (16 bits) */
#define APA8_DATAMASK	0xF0000162	/* Data mask registers (16 bits) */

			/* I/O addresses used only by the apa8 card */
#define APA8_DISPLAY	0xF4D00000	/* Display memory for the apa8 */

			/* I/O addresses used only by the apa8c card */
#define APA8C_DISPLAY	0xF4D20000	/* Display memory for the apa8c */
#define APA8C_CPLANE	0xF0000154	/* Colour plane select register */
#define APA8C_FG_BG	0xF0000156	/* Foreground/Background register */
#define APA8C_VLOOKUP	0xF0000158	/* Video look-up table register */

			/* Breakdown of apa8 control register */
#define APA8CR_ROT_MASK	0x0007		/* 0-2: rotate count */

#define APA8CR_LU_MASK	0x0038		/* 3-5: logic unit function */
#define APA8CR_LU_OR	0x0028
#define APA8CR_LU_NOR	0x0038
#define APA8CR_LU_PA	0x0010
#define APA8CR_LU_PB	0x0000
#define APA8CR_LU_PNA	0x0030
#define APA8CR_LU_PNB	0x0020
					/* 6-7: reserved */
#define APA8CR_MM_MASK	0x0300		/* 8-9: memory mode */
#define APA8CR_MM_SWR	0x0000
#define APA8CR_MM_OWR	0x0100
#define APA8CR_MM_ADWR	0x0200
#define APA8CR_MM_AURW	0x0300

#define APA8CR_DEC	0x0400		/* 10: address counter mode (inc/dec) */
#define APA8CR_INC	0x0000

#define APA8CR_X	0x0800		/* 11: address counter stepping (x/y) */
#define APA8CR_Y	0x0000

#define APA8CR_BLOCK_TX	0x1000		/* 12:  block transfer enabled */

#define APA8CR_INT	0x2000		/* 13:  interrupt enable */

#define APA8CR_SYNC	0x4000		/* 14:  sync enable */

#define APA8CR_VIDEO	0x8000		/* 15:  video enable */

/*
 * Macro to write to an unused location (and undisplayable) so that
 * the writemask register in the apa8 is updated.  The value on the
 * bitmap is never used.
 */
#define	APA8_WRMASK_LOC (*(unsigned short *)(APA8_DISPLAY + ((0xFFFF-1)<<1)))
#define SET_APA8_WRITEMASK(value) {					 \
	SET_APA8_CONTROL_REG(APA8CR_SYNC | APA8CR_VIDEO | APA8CR_MM_OWR| \
			     APA8CR_X | APA8CR_LU_OR);			 \
	APA8_WRMASK_LOC = (unsigned short) value;			 \
}

			/*Set the apa8 control register */
#define SET_APA8_CONTROL_REG(value)					\
	*(unsigned short *) APA8_CONTROL = (unsigned short)(value)

			/*Set the apa8 data mask */
#define SET_APA8_DATAMASK(mask) 					\
	*(unsigned short *)APA8_DATAMASK = (unsigned short)mask

__BEGIN_DECLS
int	apa8_probe __P((void));
void	apa8_init __P((void));
void	apa8_putchar __P((const char));
int	apa8_getchar __P((void));
__END_DECLS

#endif	/* _RT_CONS_H_ */
