/* 
 * Copyright (c) 1987 Carnegie-Mellon University
 * Copyright (c) 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)pmap.h	8.1 (Berkeley) 6/10/93
 */

#ifndef	_PMAP_MACHINE_
#define	_PMAP_MACHINE_

#include <sys/cdefs.h>
#include <machine/cpu.h>

#ifndef	__asmcode
			/* XXX use Mach version of linked lists for now */
#include "rt/rt/mach_queue.h"
#endif

__BEGIN_C_ONLY
/*
 * WARNING: load_context in ca/loutil.s relies on the fact that sidtab is
 * 	    an array of shorts.  If you change this, you must also change
 *	    load_context or all hell will break loose.  It also relies on 
 *	    sidtab being the first thing in this structure.
 */
struct pmap {
	short			sidtab[NUSR_SEGS];	/* seg id's */
	int			ref_count;	/* reference count */
	simple_lock_data_t	lock;		/* lock on map */
	struct pmap_statistics	stats;		/* map statistics */
	queue_head_t		ppage_list;	/* list of physical pages */
	int			ppage_count;	/* count of known ppages */
	queue_head_t		cache_entries;	/* shm cache entries */
	int			cache_count;	/* count of cache entries */
};

typedef struct pmap_list_entry {
    	queue_chain_t	pmap_pages;		/* chain entry for the list */
	struct pmap	*pmap;			/* pmap that owns this page */
	boolean_t	in_use;			/* debugging */
    	vm_offset_t	vaddr;			/* virtual address */
	vm_offset_t	paddr;			/* physical page */
} *pmap_list_entry_t;

/*
 * locache.s depends on this structure being 32 bytes long.
 *
 * If you change this, you must also change locache.s
 */
typedef struct pmap_cache_entry {
    unsigned int 	addrtag;	/* sid/vpage key */
    unsigned short 	ppage;		/* physical page index */
    unsigned short 	prot;		/* protection on the page */
    struct pmap_cache_entry *next;	/* the next v->p mapping for this page*/
    struct pmap_cache_entry *prev;	/* the prev v->p mapping for this page*/
    queue_chain_t 	cache_entries;	/* chain of cache entries in pmap */
    struct pmap		*pmap;		/* the owner of this cache entry */
    int 		reserved1;	/* padding to make it 32 bytes 	*/
} *pmap_cache_entry_t;

typedef struct pmap		*pmap_t;

extern	struct pmap		kernel_pmap_store;

extern	struct pmap_list_entry	*pmap_list_entries;

extern	struct pmap_cache_entry	*alias;
extern	struct pmap_cache_entry	*headers;
extern	int 			cache_size;
__END_C_ONLY

#define	PMAP_NULL	((pmap_t) 0)

#define kernel_pmap	(&kernel_pmap_store)

#define	PMAP_ACTIVATE(pmap, th, cpu) \
{\
	extern pmap_t active_pmap; \
	active_pmap = pmap; \
}

#define PMAP_DEACTIVATE(pmap,th,cpu) \
{\
	extern pmap_t active_pmap; \
	active_pmap = PMAP_NULL; \
}

#define CACHE_NULL ((struct pmap_cache_entry *) 0)

#define	ENTRY_FREE		0x1000	

#define	CACHE_STATISTICS	1

#define	CACHE_ENTRY_SHIFT	5

#endif /* _PMAP_MACHINE_ */
