/* 
 * Mach Operating System
 * Copyright (c) 1989, 1988, 1987 Carnegie-Mellon University
 * All rights reserved.  
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 */
/*
 * HISTORY
 * $Log:	pmap.h,v $
 * Revision 2.11  90/03/09  00:36:43  af
 * 	Added null pmap_attribute() macro.
 * 	[90/03/09            af]
 * 
 * Revision 2.10  89/10/11  22:27:41  dlb
 * 	Put macro, zalloc and lock stuff under #ifdef KERNEL
 * 	conditionals so that the file would compile outside
 * 	of the kernel where it is included by adb.
 * 	[89/10/11            mrt]
 * 
 * Revision 2.9  89/10/11  13:50:47  dlb
 * 	Define PMAP_CONTEXT macro.
 * 	[89/10/02  21:53:15  dlb]
 * 
 * Revision 2.8  89/09/05  20:41:14  jsb
 * 	Added pmap_phys_to_frame definition.
 * 	[89/09/05  17:09:31  jsb]
 * 
 * Revision 2.7  89/03/09  19:42:11  rpd
 * 	More cleanup.
 * 
 * Revision 2.6  89/02/25  15:08:50  gm0w
 * 	Changes for cleanup.
 * 
 * Revision 2.5  89/01/15  21:33:27  rpd
 * 	Use decl_simple_lock_data.
 * 	[89/01/15  14:51:07  rpd]
 * 
 * Revision 2.4  88/10/27  10:41:02  rpd
 * 	Removed include of romp_shared_seg.h.
 * 
 * Revision 2.3.1.1  88/10/26  14:25:56  rpd
 * 	Removed include of romp_shared_seg.h.
 * 
 * Revision 2.3  88/10/10  22:23:35  sanzi
 * 	Move declaration of pmap_resident_count() to within the 
 * 	#ifndef ASSEMBLER condition. 
 * 
 * Revision 2.2  88/10/06  13:57:15  sanzi
 * 	Fix stupid automatic generated log message in previous revision.
 * 	Make pmap_resident_count() a function declaration instead of 
 * 	a macro.
 * 	Fix include file references.
 * 	Delete ROMP_SHARED_SEG conditionals.
 * 
 * Revision 2.1.1.5  88/10/04  09:43:58  sanzi
 * 	Fix stupid automatic generated log message in previous revision.
 * 
 * Revision 2.1.1.4  88/10/04  09:41:41  sanzi
 * 	Make pmap_resident_count() a function declaration instead of
 *	a macro.
 *
 * Revision 2.1.1.3  88/10/04  09:38:56  sanzi
 * 	Fix include file references.
 * 
 * 14-Dec-87  Michael Young (mwyoung) at Carnegie-Mellon University
 *	Added pmap_phys_address() declaration.
 *
 * 15-Jul-87  Michael Young (mwyoung) at Carnegie-Mellon University
 *	Added pmap_resident_count() macro.
 *
 * 22-Jun-87  Bill Bolosky (bolosky) at Carnegie-Mellon University
 *	ROMP_SHARED_SEG: Added ss table and vmmap.
 *
 * 19-Mar-87  Avadis Tevanian (avie) at Carnegie-Mellon University
 *	Export kernel_pmap as all good pmap modules should.
 *
 *  8-Apr-86  Bill Bolosky (bolosky) at Carnegie-Mellon University
 *	Created.
 *
 */
/*
 *	File:	ca/pmap.h
 *
 *	RT Version:
 *	Copyright (c) 1986, 1987, William J. Bolosky
 *
 *	Romp/Rosetta machine-dependent structures for the pmap module.
 */

#ifndef	_PMAP_MACHINE_
#define _PMAP_MACHINE_	1

/* If you change this then you must also change the line in mkmakefile.c
 * in the source for config which gives the load address to the loader and
 * the sed script in make.ws which converts kernel virtual addresses into
 * real addresses while building the debugger symbol table.  You must also
 * chnage the .set for SYS_ORG in locore.h.  And change 
 * VM_MIN/MAX_{KERNEL,}ADDRESS in vm_param.h (sigh).
 */
#define SYS_SEG	0xe
#define SYSBASE_ADDR	0xe0000000	  /* virt addr of system origin */
#define SYSBASE  ((u_long)SYSBASE_ADDR)
#define NUSR_SEGS	SYS_SEG	/* Hacky, but correct */
/*BJB#define IOSAVEL		19 * 4 */
#define IOSAVEL		21 * 4  /* BJB-for debugging interrupts */

#ifndef	ASSEMBLER
#include <mach/vm_param.h>
#include <ca/vmparam.h>
#include <mach/vm_statistics.h>

#define KERNSTACK       (ENDOFP1-sizeof(struct user)) /* top of kernel stack */
#define uareabase(u)    ((caddr_t)(u)+sizeof(struct user)-UPAGES*NBPG)

#ifdef	KERNEL
#include <kern/macro_help.h>
#include <kern/zalloc.h>
#include <kern/lock.h>

/*
 * WARNING: load_context in ca/loutil.s relies on the fact that sidtab is
 * 	    an array of shorts.  If you change this, you must also change
 *	    load_context or all hell will break loose.  It also relies on 
 *	    sidtab being the first thing in this structure.
 */
struct pmap {
	short			sidtab[NUSR_SEGS];	/* seg id's */
	int			ref_count;	/* reference count */
	decl_simple_lock_data(,	lock)		/* lock on map */
	struct pmap_statistics	stats;		/* map statistics */
	queue_head_t		ppage_list;	/* list of physical pages */
	int			ppage_count;	/* count of known ppages */
	queue_head_t		cache_entries;	/* shm cache entries */
	int			cache_count;	/* count of cache entries */
};

typedef struct pmap_list_entry {
    	queue_chain_t	pmap_pages;		/* chain entry for the list */
	struct pmap	*pmap;			/* pmap that owns this page */
	boolean_t	in_use;			/* debugging */
    	vm_offset_t	vaddr;			/* virtual address */
	vm_offset_t	paddr;			/* physical page */
} *pmap_list_entry_t;

typedef struct pmap	*pmap_t;

#define PMAP_NULL	((pmap_t) 0)
extern	loadpt;
extern	pmap_t	kernel_pmap;
extern	int pmap_resident_count(/* pmap_t pmap */);

vm_offset_t	pmap_phys_address();
int		pmap_phys_to_frame();

#define PMAP_ACTIVATE(pmap, th, cpu)	\
MACRO_BEGIN				\
	extern pmap_t active_pmap;	\
	active_pmap = pmap;		\
MACRO_END

#define PMAP_DEACTIVATE(pmap,th,cpu)	\
MACRO_BEGIN				\
	extern pmap_t active_pmap;	\
	active_pmap = PMAP_NULL;	\
MACRO_END

#define PMAP_CONTEXT(pmap, thread)
#define pmap_attribute(pmap,addr,siz,attr,val)	0

#endif	KERNEL
#endif	ASSEMBLER
#endif	_PMAP_MACHINE_


