/*
 * Copyright (c) 1988 University of Utah.
 * Copyright (c) 1982, 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * from: Utah $Hdr: clock.c 1.18 91/01/21$
 *
 *	@(#)clock.c	8.2 (Berkeley) 1/12/94
 */

#include <sys/param.h>
#include <sys/kernel.h>

#include <machine/psl.h>
#include <machine/cpu.h>

#include "clock.h"

#ifdef GPROF
#include <sys/gmon.h>
#endif


			/* number of timer counts to a short timer int */
static unsigned timer_speed;

/*
 * Machine-dependent clock routines.
 *
 */

#define ROMP_HZ		1024 /* number of decrements per second for scr 7 */
#define COUNTS_PER_SEC	(1000000 / CLK_RESOLUTION)

/*
 * Set up the real-time and statistics clocks.  Leave stathz 0 only if
 * no alternative timer is available.
 *
 */
cpu_initclocks()
{
			/* pointer to MC146818 clock chip */
	volatile struct clkreg *rt_clock = (struct clkreg *)CLKBASE;

	if (COUNTS_PER_SEC*hz % 1000000) {
		printf("cannot get %d Hz clock; using 64 Hz\n", hz);
		hz = 64;
	}
	/*
	 * Clock has several counters, so we can always use separate
	 * statclock.
	 */
	if (stathz == 0)		/* XXX should be set in param.c */
		stathz = hz;
	else if (COUNTS_PER_SEC*hz % 1000000) {
		printf("cannot get %d Hz statclock; using 64 Hz\n", stathz);
		stathz = 64;
	}
	if (profhz == 0)		/* XXX should be set in param.c */
		profhz = stathz * 5;
	else if (profhz < stathz || COUNTS_PER_SEC % profhz) {
		printf("cannot get %d Hz profclock; using %d Hz\n",
		    profhz, stathz);
		profhz = stathz;
	}

	tick = 1000000/hz;	/* ticks per second */

	timer_speed = ROMP_HZ / hz;

				/* initialise the clock chip */
	rt_clock->reg_a = CLK_SPEED;
	rt_clock->reg_b = CLK_SQUARE | CLK_BINARY | CLK_24HR | CLK_DST;

	mtsr(SCR_TIMER, timer_speed);		/* start with a short tick */
	mtsr(SCR_TIMER_RESET, timer_speed);	/* 1st tick is a guess */
	mtsr(SCR_TIMER_STATUS, TS_ENABLE | TS_ILEVEL);

	delay_refine();			/* refine the value used in delay() */
}

/*
 * We assume newhz is either stathz or profhz, and that neither will
 * change after being set up above.  Could recalculate intervals here
 * but that would be a drag.
 */
void
setstatclockrate(newhz)
	int newhz;
{
	/* XXX  need to fill in details -md */
}


/*
 * Return the best possible estimate of the current time.
 */
microtime(tvp)
	struct timeval *tvp;
{
	/* XXX  need to fill in details -md */
}


/*
 * Based on the current time, restore the time of day hardware after a 
 * time change.  Used when the current clock is preposterous, when the 
 * time is changed, and when rebooting.  Do nothing if the time is not 
 * yet known, e.g., when crashing during autoconfig.
 */
resettodr()
{
	if (!time.tv_sec)	/* no time set yet */
	    return;

	/* XXX  need to fill in details -md */
}

/*
 * Anything below here is specific to the RT's implementation and is not
 * used by the 4.4 kernel.
 */

		/* number of loop iterations to get 1ms of delay */
#define DEFAULT_DELAY_COUNT 1000	
int ms_delay_loop = DEFAULT_DELAY_COUNT;

/* 
 * Provide a delay of 'n' milliseconds (in real time, NOT virtual time).
 */

delay(n)
	int n;
{
	int i;

	while (--n >= 0)
	    for (i = ms_delay_loop; --i >= 0;)
		*(char *)DELAY_ADDR = 0xff;
}

/*
 * Adjust the loop counter "ms_delay_loop" used in delay() to provide an
 * exact 1ms delay.
 */
#define REFINE_TIME	1000	/* number of ms to sample for (1 second) */

delay_refine()
{
	struct timeval tv=time;		/* current time */
	long elapsed;			/* actual elapsed time in ms */

	delay(REFINE_TIME);		/* wait the sampling time */

	elapsed = (time.tv_sec-tv.tv_sec)*1000 + (time.tv_usec-tv.tv_usec)/1000;
	if (elapsed <= 0) {
	    printf("Clock doesn't appear to be running (t=%d)\n", elapsed);
	} else {
		ms_delay_loop = (ms_delay_loop * REFINE_TIME) / elapsed;
	}
}

/* 
 * Set the number of times we should loop in the delay() routine to a value
 * which is close to what the real delay_refine() routine would come up
 * with.
 */
approx_delay_refine(cpu_type)
{
	switch(cpu_type) {
		case CPU_ROMPC:
			ms_delay_loop=600;
		case CPU_MOD135:
			ms_delay_loop=800;
		default:
			ms_delay_loop=400;
	}
}
