/*
 * Copyright (c) 1988 University of Utah.
 * Copyright (c) 1982, 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * from: Utah $Hdr: clockreg.h 1.14 91/01/18$
 *
 *	@(#)clockreg.h	8.2 (Berkeley) 1/12/94
 */

/*
 * RT "real time clock" (MC146818) registers
 */

struct clkreg {
	u_char	sec;		/* seconds		*/
	u_char	sec_alarm;	/* seconds alarm	*/
	u_char	min;		/* minutes		*/
	u_char	min_alarm;	/* minutes alarm	*/
	u_char	hrs;		/* hours		*/
	u_char	hrs_alarm;	/* hours alarm		*/
	u_char	day_of_week;	/* day of the week	*/
	u_char	day_of_month;	/* day of the month	*/
	u_char	month;		/* month		*/
	u_char	year;		/* year			*/
	u_char	reg_a;		/* register A (R/W)	*/
	u_char	reg_b;		/* register B (R/W)	*/
	u_char	reg_c;		/* register C (R/O)	*/
	u_char	reg_d;		/* register D (R/O)	*/
};

/* base/offsets for register access (for locore.s) */
#define	CLKBASE		0xf0008800
#define	CLK_SEC		CLKBASE
#define	CLK_ASEC	CLKBASE+1
#define	CLK_MIN		CLKBASE+2
#define	CLK_AMIN	CLKBASE+3
#define	CLK_HRS		CLKBASE+4
#define	CLK_AHRS	CLKBASE+5
#define	CLK_DOW		CLKBASE+6
#define	CLK_DOM		CLKBASE+7
#define	CLK_MON		CLKBASE+8
#define	CLK_YR		CLKBASE+9
#define	CLK_REG_A	CLKBASE+10
#define	CLK_REG_B	CLKBASE+11
#define	CLK_REG_C	CLKBASE+12
#define	CLK_REG_D	CLKBASE+13

				/* speed to which to default the clock to */
#define CLK_SPEED	(CLK_DIV_1+CLK_RATE_2+CLK_RATE1) 

/*
 * MC146818 register A breakdown
 */

#define CLK_UPDATE	0x80	/* update currently in progress */
#define CLK_DIV_2	0x40	/* select divisor 2 */
#define CLK_DIV_1	0x20	/* select divisor 1 */
#define CLK_DIV_0	0x10	/* select divisor 0 */
#define CLK_RATE_3	0x08	/* select rate 3 */
#define CLK_RATE_2	0x04	/* select rate 2 */
#define CLK_RATE_1	0x02	/* select rate 1 */
#define CLK_RATE_0	0x01	/* select rate 0 */

/*
 * MC146818 register B breakdown
 */

#define CLK_INIT	0x80	/* 0 - Update normal, 1 - Halt and init */
#define CLK_PERIODIC	0x40	/* enable periodic interrupts */
#define CLK_ALARM	0x20	/* enable alarm interrupt */
#define CLK_UPDATE_END	0x10	/* enable update-ended interrupt */
#define CLK_SQUARE	0x08	/* enable square-wave signal */
#define CLK_BINARY	0x04	/* data mode, 0 - BCD, 1 - binary */
#define CLK_24HR	0x02	/* 0 - 12 hour mode, 1 - 24 hour mode */
#define CLK_DST		0x01	/* enable daylight savings time */

#define CLK_RESOLUTION	64	/* 64Hz */

#define	leapyear(year)		((year) % 4 == 0)
#define	days_in_year(a) 	(leapyear(a) ? 366 : 365)
#define	range_test(n, l, h)	if ((n) < (l) || (n) > (h)) return(0)
