#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/user.h>

#include <vm/vm.h>
#include <vm/vm_kern.h>
#include <vm/vm_page.h>

#include <machine/cpu.h>

/*
 * The following need to be defined in locore.s
 */
struct hatipt_coarse *ROMP_HATIPT;		/* pointer to page table */
unsigned char	*is_in;				/* bitmap for page map */
int	physmem;				/* number of ram pages */
int	ROMP_HASHMASK=8192-1;			/* number of pages minus 1 */
int	holestart, holeEnd;
long	start;

/*
 * The following are found in vm/vm_page.c and are used by the macros in
 * vm/vm_param.h so we just define them here with reasonable values to keep
 * everyone happy.
 */
vm_page_t	vm_page_array;
long		first_page;
int		page_shift=11;	/* 2K pages */
vm_size_t	page_mask=2048-1;

/*
 * Simulation of hardware registers found in the ROMP
 */

unsigned int romp_regs[0x3000];

main()
{
	int i;
	pmap_t		pmap, pmap2;
	vm_offset_t	vaddr, paddr;
	vm_offset_t	vstart, vend;
	vm_prot_t	prot;
	boolean_t	wired;

				/* 8192 entries / 8 bits char */
	is_in=(unsigned char *)malloc(1024); 

				/* 16M worth of entries @ 2K/page */
	ROMP_HATIPT=(struct hatipt_coarse *)malloc(8193*sizeof(struct hatipt));
	if (!ROMP_HATIPT) {
	    printf("Unable to allocate memory for ROMP_HATIPT\n");
	    exit(1);
	}
					/* align to a 16 byte boundry */
	ROMP_HATIPT=(struct hatipt_coarse *)((((int)ROMP_HATIPT+15)>>4)<<4);
	physmem=8192; /* number of pages */
	printf("ROMP_HATIPT located at vaddr 0x%08x\n", ROMP_HATIPT);
	printf("physmem is 0x%04x pages\n", physmem);
	/* mmu_init(ROMP_HATIPT); */
	/* foo(ROMP_HATIPT, 8193*sizeof(struct hatipt)); */
	romp_init(ROMP_HATIPT);

	/*
	set_rosetta_seg(2,0x1020);
	dump_seg_regs();
	*/

	printf("HATIPT after pmap_bootstrap:\n");
	dump_hatipt(ROMP_HATIPT, 8192, 1);

#ifdef	MAPIN_MAPOUT_TEST
	printf("HATIPT after mapin(0x00300, 0x0000000, 0, 128, 0x500):\n");
	/* map phys addr 0x00180000 (page 0x300) to 
	   vaddr 0x500 0000000 for 128 pages */
	/*     ppage      vpage   prot  count  sid */
	mapin(0x00300, 0x0000000,  0,    128, 0x500);
	dump_hatipt(ROMP_HATIPT, 8192, 1);

	printf("HATIPT after mapout(0x300, 128):\n");
	mapout(0x300, 128);
	dump_hatipt(ROMP_HATIPT, 8192, 1);
#endif

	pmap=pmap_create(0);	/* create a new pmap */
	printf("new pmap handle is 0x%x\n", pmap);
	pmap_activate(pmap, 0);	/* activate it now */
	dump_pmap(pmap);

		/* add in a bunch of pages to the pmap */
	for (i=0; i<10; i++) {
	    vaddr=(segtoa(1) | 0x00000000) + i*PAGE_SIZE;
	    paddr=0x00100000 + i*PAGE_SIZE; /* page 0x200+ */
	    prot=VM_PROT_DEFAULT;
	    wired=0;
	    pmap_enter(pmap, vaddr, paddr, prot, wired);
	}
		/* remove some pages right in the middle */
	vstart=(segtoa(1) | 0x00000000) + 2*PAGE_SIZE;
	vend=(segtoa(1) | 0x00000000) + 3*PAGE_SIZE;
	pmap_remove(pmap, vstart, vend);
	/* dump_hatipt(ROMP_HATIPT, 8192, 1); */
	dump_pmap(pmap);

		/* add in more pages in a different segment */
	vaddr=segtoa(2) | 0x02000000;
	paddr=0x00200000; /* page 0x400 */
	pmap_enter(pmap, vaddr, paddr, prot, wired);
	dump_pmap(pmap);
	dump_hatipt(ROMP_HATIPT, 8192, 1); /* */
	dump_page_trans(0x400);

	pmap2=pmap_create(0);	/* create another pmap and activate it */
	pmap_activate(pmap2, 0);

		/* make this pmap overlap the previous one */
	vaddr=segtoa(1) | 0x02000000;
	paddr=0x00200000; /* page 400 */
	pmap_enter(pmap2, vaddr, paddr, prot, wired);
	printf("Shared pmap view:\n");
	printf("   >>>> pmap 1:\n");
	dump_pmap(pmap);
	printf("   >>>> pmap 2:\n");
	dump_pmap(pmap2);
	dump_page_trans(0x400);

	printf("   >>>> Current address translations:\n");
	printf("         pmap 1 page 0x400 ppage=%x\n",
		pmap_extract(pmap,  segtoa(2) | 0x02000000));
	printf("         pmap 2 page 0x400 ppage=%x\n",
		pmap_extract(pmap2, segtoa(1) | 0x02000000));
}

dump_seg_regs()
{
	int i,sr;

	for (i=0; i<16; i++) {
	    sr=get_rosetta_seg(i);
	    printf("SR[%x]: valid=%d cpu=%d i/o=%d id=0x%03x spec=%d key=%d\n",
		i,
		sr&ROMP_SEG_P ? 1 : 0,
		sr&ROMP_SEG_C ? 1 : 0,
		sr&ROMP_SEG_I ? 1 : 0,
		(sr&ROMP_SEG_SID) >> ROMP_SID_SHIFT,
		sr&ROMP_SEG_S ? 1 : 0,
		sr&ROMP_SEG_K ? 1 : 0);
	}
}

dump_page_trans(page)
int page;
{
	struct hatipt *ipte;

	ipte = (struct hatipt *)&ROMP_HATIPT[page];
	printf("Page 0x%x: sid=%x  vpage=%x\n", page, ipte->sid, ipte->vpage);
}


/*
 * Fake support routines follow:
 */
void
panic(s)
const char *s;
{
	printf("panic: %s\n", s);
	exit(1);
}

_spl1() {};
splx() {};

/* 
 * the following are in the kernel 
 */
void lock_init(x,y)
lock_t x;
boolean_t y;
{};

void
lock_read(x)
lock_t x;
{};

void
lock_write(x)
lock_t x;
{};

void
lock_done(x)
lock_t x;
{};

/*
 * The following (real versions) are in loutil.s
 */

iow(a,v)
int a,v;
{
	if (a >= ROMPBASE && a < ROMPBASE+0x3000) {
	    romp_regs[a-ROMPBASE]=v;
	    /* printf("romp seg reg %d=%x\n", a-ROMPBASE, v); */
	} else {
	    printf("iow:write to unknown location: 0x%x\n", a);
	}
};

ior(a)
int a;
{
	if (a >= ROMPBASE && a < ROMPBASE+0x3000) {
	    return(romp_regs[a-ROMPBASE]);
	} else {
	    printf("ior:read from unknown location: 0x%x\n", a);
	}
};

foo(p, len)
char *p;
int len;
{
	for (;len; len--, p++) {
	    if (len%16 == 0)
		printf("\n");
	    printf("%02x ", *p);
	}
	exit(1);
}
