/*
 * Copyright (c) 1988, 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * from: Utah $Hdr: in_cksum.c 1.1 90/07/09$
 *
 *	@(#)in_cksum.c	8.1 (Berkeley) 6/10/93
 */

/*
 * in_cksum - checksum routine for the Internet Protocol family.
 */

#include <sys/param.h>
#include <sys/mbuf.h>

/*
 * Checksum routine for Internet Protocol family headers.
 *
 * This routine is very heavily used in the network code and should be modified
 * for each CPU to be as fast as possible.
 */

in_cksum(m, len)
	register struct mbuf *m;
	register int    len;
{
	u_short	*w;
	int	sum = 0;
	int	mlen = 0;
	int	scklen;
	int	tmpsum;

	for (;;) {
		/*
		 * Each trip around loop adds in
		 * word from one mbuf segment.
		 */
		w = mtod(m, u_short *);
		if (mlen == -1) {
			sum += *(u_char *) w;
			if (sum >= 0xFFFF)
				sum = (sum + 1) & 0xFFFF;
			w = (u_short *) ((char *) w + 1);
			mlen = m->m_len - 1;
			len--;
		} else
			mlen = m->m_len;
		m = m->m_next;
		if (len < mlen)
			mlen = len;
		len -= mlen;
		if ((int) w & 01) {
			if (mlen >= 2) {
				sum += ((*(char *) w) & 0xFF) << 8;
				if (sum >= 0xFFFF)
					sum = (sum + 1) & 0xFFFF;
				w = (u_short *)((unsigned long)w + 1);
				scklen = (mlen - 2) & 0xfffffffe;
				tmpsum = scksum(w,scklen);
				sum += ((tmpsum & 0xff) << 8) | ((tmpsum & 0xff00) >> 8);
					if (sum >= 0xFFFF)
					sum = (sum + 1) & 0xFFFF;
				w = (u_short *)((unsigned long)w + scklen);
				sum += (*(char *) w) & 0xFF;
				if (sum >= 0xFFFF)
					sum = (sum + 1) & 0xFFFF;
				if (mlen & 01)
					mlen = -1;
				else
					mlen = -2;
				w = (u_short *)((unsigned long)w + 1);
			} else
				mlen -= 2;
		 } else {
			if (mlen >= 2) {
				scklen = mlen & 0xfffffffe;
				sum += scksum(w,scklen);
				if (sum >= 0xFFFF)
					sum = (sum + 1) & 0xFFFF;
				if (mlen & 01)
					mlen = -1;
				else
					mlen = -2;
				w = (u_short *)((unsigned long)w + scklen);
			} else
				mlen -= 2;
		 }
		if (mlen == -1) {
			sum += *(u_char *) w << 8;
			if (sum >= 0xFFFF)
				sum = (sum + 1) & 0xFFFF;
		}
		if (len == 0)
			break;
		/*
		 * Locate the next block with some data. If there is a word
		 * split across a boundary we will wrap to the top with mlen
		 * == -1 and then add it in shifted appropriately. 
		 */
		for (;;) {
			if (m == 0) {
				printf("cksum: out of data\n");
				goto done;
			}
			if (m->m_len)
				break;
			m = m->m_next;
		}
	}
done:
	if (sum)
		sum ^= 0xFFFF;
	return sum;
}
