
#include <machine/cpu.h>
#include "rt/rt/locore.h"
#include "rt/rt/postcb.h"

#define SYS_ADDRTAG ROMP_SID_SYSTEM<<ROMP_VPAGE_BITS

	.globl	.oVncs
	.set	.oVncs, 0

	.text
zeroaddr:	

	# Setup interrupt vectors
        .org zeroaddr + 0x100

#define INTVEC_INIT(label,iar,ics) \
label:  .long   0xB0B0B0B0             /* Old IAR */\
        .short  0xB0B0                 /* Old ICS (SCR 14) */\
        .short  0xB0B0                 /* Old CS  (SCR 15) */\
        .long   iar			/* New IAR */\
        .short  ics			/* New ICS */\
        .short  0x0000                  /* Reserved */

INTVEC_INIT(LEV0,lev0intr-zeroaddr,0x0000)
INTVEC_INIT(LEV1,lev1intr-zeroaddr,0x0000)
INTVEC_INIT(LEV2,lev2intr-zeroaddr,0x0000)
INTVEC_INIT(LEV3,lev3intr-zeroaddr,0x0000)
INTVEC_INIT(LEV4,lev4intr-zeroaddr,0x0000)
INTVEC_INIT(LEV5,lev5intr-zeroaddr,0x0000)
INTVEC_INIT(LEV6,lev6intr-zeroaddr,0x0000)
INTVEC_INIT(MCHK,mchkintr-zeroaddr,0x0000)
INTVEC_INIT(PCHK,pchkintr-zeroaddr,0x0000)
INTVEC_INIT(SVC,svcintr-zeroaddr,0x0000)


	# Move past the POSTCB
	.org zeroaddr + POSTCB_END

	# this lands at 0x1000.
	.globl _start
_start:
	.using	zeroaddr,r0	# tell the assembler to use r0 as base
	cal16	r15,start	# address of start, ignoring segment number
	balr	r15,r15		# transfer control to real 'start'

	.align	2

	# global variables, etc
	.globl _loboothowto
	.globl _lobootdev
	.globl _holestart
	.globl _holeEnd
	.globl _physmem
	.globl hatipt
	.globl _ROMP_HATIPT
	.globl _ROMP_HASHMASK
	.globl	_is_in
	.globl	_cpu

_loboothowto:	.long	0	# XXX
_lobootdev:	.long	0	# XXX
_holestart:	.long	0	# memory hole start
_holeEnd:	.long	0	# memory hole end
_physmem:	.long	0	# number of physical pages
hatipt:         .long   0       # pointer to page table (real)
_ROMP_HATIPT:	.long	0	# pointer to page table (virtual)
_ROMP_HASHMASK:	.long	0	# number of pages minus 1
_is_in:		.long	0	# bitvector for physical pages
_cpu:		.long	0	# cpu identifier

	.align	2

	# real execution begins here 
	.globl	start
ASENTRY(start):
	st	r2,_loboothowto		# save first arg for init
	st	r3,_lobootdev		# second arg is bootdev

#	cau	r1,(KERNSTACK)>>16(r0)	# setup kernel stack
#	oil	r1,r1,(KERNSTACK)&0xffff

	# Disable interrupt handling.
	cal	r4,0x180(r0)		# set interrupt & checkstop mask
	mts	scr_ics,r4		# set ICS.

	# now initialize the hardware.
	get 	r0,$_init_hdwr;
	bali	r15,_.init_hdwr;

	wait				# we dont know how to do anything yet

	.align	2

#############################################################################
# hardware bootstrap
#
ENTRY(init_hdwr)

	# determine what processor we are running on.
	# 3 choices, original RT ROMP, APC, and EAPC -- aka mod135.
	# NOTE: the original RT ROMP processor is not supported.

	# XXX:  TODO (?) 
	#	do a read from 0xf0008c80 (memory configuration register).
	#	Check bit 7.  If set to 1, original processor, so DIE.
	#	Else if 0, its APC or EAPC.
	# XXX:  ASSUME we are on an APC-stype processor.
	#	Besides, I have no way to test this code.
#	get	r4,$0xf0008c80		# memory configuration register
#	lcs	r4,0(r4)		# read it.
#	nilz	r4,r4,0x40		# if 0 then OLD cpu, if 1 then *APC
#	jeq	cpu_is_apc_type
#	wait				# OLD cpu, just die.
#cpu_is_apc_type:

	#
	# initialize IOIM 1
	# tech ref, appendix B-12.
	#
	get	r4,$0x800400	# address of IOIM1_BASE
	cal16	r5,0xa090(r0)	# reset default values.
        iow	r5,0(r4)	# write it.
	#
	# disable parity (tech ref appendix B says to do so!)
	#
	get	r4,$0x900008	# address of IOIM1+IOIM_GCR
        ior     r5,0(r4)	# read it.
	oil	r6,r5,0x8	# OR in IOIM_NOPARITY
        iow	r6,0(r4)	# write it.

	# Now that IOIM 1 is initialized, continue determining which
	# CPU we are running on.
	# Definitely on an APC style processor, so check if its an EAPC.
	# check bit 20 in misc chip control register.  If 1, this
	# is an APC, if 0 then EAPC.

	get	r4,$0x900000		# get base of IOIM1
	ior	r4,0xd(r4)		# get chip configuration register
	nilz	r4,r4,0x800		# if 1 then APC, if 0 then EAPC
	jeq	cpu_is_mod_eapc		# branch if EAPC
	lis	r0,CPU_APC
	st	r0,_cpu			# store cpu model information
	j	cpu_check_done		# done with CPU model checks
cpu_is_mod_eapc:
	lis	r0,CPU_EAPC		# else its a EAPC
	st	r0,_cpu			# store cpu model number

	# definitely on an EAPC processor.  Read General Status Register
	# twice to see if we need to set a bit in General Control Register.
	# Hardware ref appendix C.

	get	r6,$0x900000		# get base of IOIM1
	ior	r4,0xb(r6)		# get GSR to clear it
	ior	r4,0xb(r6)		# get GSR to see if bit 20 stuck at 1
	nilz	r4,r4,0x800		# is bit 20 set?
	jeq	cpu_check_done		# no, new ZIC
	ior	r4,0x8(r6)		# yes bit 20 set, old ZIC, apply fix
	setbl	r4,6			# turn on bit 22
	iow	r4,0x8(r6)		# write to GCR
cpu_check_done:

	#
	# Do a POR to all boards via CRRA and CRRB
	# tech ref p5-59
	#
	cal	r4,0xff(r0)		# 0xff -- reset all bits
	get     r5,$0xf0008c40		# component reset register A
	stcs    r4,0(r5)		# issue POR
	get     r5,$0xf0008c60		# component reset register B
	stcs    r4,0(r5)		# issue POR
	#
	# initialize IOIM 2 (NOT present on EAPC card)
	#
	load	r2,_cpu
	cis	r2,CPU_EAPC
	jeq	ioim2_skip
	get	r4,$0x800100	# address of IOIM2_base
	cal16	r5,0xb091(r0)	# make it different from IOIM1
        iow	r5,0(r4)	# write it.

ioim2_skip:
        br      r15		# return to caller

	.align	2

/* 
#############################################################################
# interrupt handlers
# Level 0
# -- keyboard attention keystroke
#
# Level 1
# -- Real Time Clock
#
# Level 2
# -- IOCC errors || MMU Program Check
#
# Level 3
# -- 8259 fed IRQ's
# 8259     I/O chan        I/O Adapter
# Level    interrupt lev
#   0         0            8237 Terminal Count (sys board)
#   1        10            IRQ 10
#   2         9            IRQ 9
#   3         3            IRQ 3
#   4         4            IRQ 4
#   5         1            Keyboard Controller (sys board)
#   6         2            8530 Serial port (sys board)
#   7         7            IRQ 7
#
# Level 4
# -- 8259 fed IRQ's
# 8259     I/O chan        I/O Adapter
# Level    interrupt lev
#   0         8            Reserved (sys board)
#   1        11            IRQ 10
#   2        14            IRQ 14         
#   3        12            IRQ 12
#   4         6            IRQ 6
#   5         5            IRQ 5
#   6        15            IRQ 15
#   7        13            Serial Port Ex Ct1 IRPT (sys board)
#
# Level 5
# -- unused
#
# Level 6
# -- software interrupts
*/

ASENTRY(lev0intr):
	br	r15
ASENTRY(lev1intr):
	br	r15
ASENTRY(lev2intr):
	br	r15
ASENTRY(lev3intr):
	br	r15
ASENTRY(lev4intr):
	br	r15
ASENTRY(lev5intr):
	br	r15
ASENTRY(lev6intr):
	br	r15
ASENTRY(mchkintr):
	br	r15
ASENTRY(pchkintr):
	br	r15
ASENTRY(svcintr):
	br	r15


ENTRY(spln)
	mfs   scr_ics,r4		# get the current status
	ni    r3,r4,-8			# get all but priority in r3
	o     r3,r2			# overlay priority bits
	nilz  r2,r4,7			# get only old priority bits
	brx   r15			# return and...
	mts   scr_ics,r3		# set new status

ENTRY(splx)
	nilz  r2,r2,7 			# get only low 3 bits
	mfs   scr_ics,r3		# get the current status
	ni    r3,r3,-8			# get all but priority in r3
	o     r2,r3			# overlay priority bits
	brx   r15			# return and...
	mts   scr_ics,r2		# set new status

	.align	2

############################################################################
# mmu related routines
#
ENTRY(mmu_init)
	stm	r14,-44(sp)
	cal	sp,-44(sp)

 #  Loop through HATIPT mapping system page i to real page i.

	get	r0,$(0x01ffffff)	# Protection bits:
			# spare(7),write(1)=1,tid(8)=ff,lockbits(16)=ffff	
	get	r12,$(SYS_ADDRTAG)	# part of addrtag for seg id
	lis	r9,0			# page number
	load	r11,_rose_page_shift	# which page size are we using?
#ifdef	SIMULATOR
	load	r3,_ROMP_HATIPT		# point to HATIPT at end of memory
#else
	load	r3,hatipt		# point to HATIPT at end of memory
#endif
	load	r4,_ROMP_HASHMASK	# hashing significant bits
	ai	r2,r4,1			# r2 = number of HATIPT entries
	nilz	r4,r4,ROMP_SID_SYSTEM	# partial result for hash calculations
	cal16	r5,ROMP_UNDEF_PTR(r0)	# Invalid IPTPTR value for debug
	load	r7,_physmem		# pointer to end of valid memory
	get	r8,$-1			# invalid HATPTR

hatiptloop:
	mr	r10,r9			# copy of page number
	cis	r11,0xc			# 4K pages?
	jne	0f
	sli	r10,1			# double page # for addrtag
0:
	o	r10,r12			# add in segid part
	oiu	r10,r10,0x4000		# use key 01->user read, kernel write
	sts	r10,IPTADDRTAG(r3)	# set ADDRTAG field to sys seg addr
	sths	r8,IPTHATPTR(r3)	# in case the hatptr is invalid here
	mr	r6,r9			# copy page number for mungeing
	x	r6,r4			# "unhash" to get page pointer
	c	r6,r7			# beyond end of installed memory?
	jnl	hatinvalid		# yes, leave hat ptr invalid

	sths	r6,IPTHATPTR(r3)	# set HATPTR field to this page
hatinvalid:
	sths	r5,IPTIPTPTR(r3)	# invalid IPTPTR
	sts	r0,IPTLOCK(r3)		# set lock bits and flags
	inc	r9,1			# next page
	sis	r2,1			# decrement counter
	bnzx	hatiptloop		# loop if not zero
	ai	r3,r3,HATIPTSIZE	# ...while pointing to next entry

return:
	lm	r14,0(sp)
	brx	r15
	cal	sp,44(sp)


##########################################################################
# utility routines

# XXX: should inline these....
ENTRY(ior)
	ior   r2,0(r2)			# read data from port
	br   r15

ENTRY(iow)
	iow   r3,0(r2)			# write data to port
	br   r15
