
/* definitions for the Power On Self Test Control Block */

/*
 * memory between these two addresses must be preserved for all time
 * or else the machine will have to be physically powered off to reboot.
 */
#define POSTCB_START	0x800
#define POSTCB_END	0x1000

#define POSTCB_NVRAM	0xc00	/*nvram indicator. Contains 0 if CRC is valid*/
#define POSTCB_IPLDEV	0xc04	/* boot device id (0xd0->0xd3, 0xf0->0xf3) */
#define POSTCB_KBDID	0xc08	/* keyboard id */
#define POSTCB_MACHID	0xc0c	/* machine type (0 == 6150, else 6151) */
#define POSTCB_SOFTIPL	0xc10	/* soft ipl indicator */
#define POSTCB_RAMSIZE	0xc14	/* size of contiguous memory starting at 0 */
#define POSTCB_RAMSPEC	0xc3c	/* misc info */
#define   POSTCB_PCMONO_RESET_MASK	0xff	/* ??? */
#define   POSTCB_PCMONO_PRESENT		0x800	/* pcmono is present */
#define   POSTCB_SLOTC_BAD		0x1000	/* memory in slot C is bad */
#define   POSTCB_SLOTD_BAD		0x2000	/* memory in slot D is bad */
#define POSTCB_SLOTID_1	0xc84	/* list of adapters in slot XX */
#define POSTCB_SLOTID_2	0xc8c
#define POSTCB_SLOTID_3	0xc94
#define POSTCB_SLOTID_4	0xc9c
#define POSTCB_SLOTID_5	0xca4
#define POSTCB_SLOTID_6	0xcac
#define POSTCB_SLOTID_7	0xcb4
#define POSTCB_SLOTID_8	0xcbc
#define   POSTCB_DISK_ID 0x52	/* id for generic disk (mfm, ESDI, IDE) */
#define   POSTCB_FLOP_ID 0x53	/* id for floppy device */
#define   POSTCB_EESDI_ID 0x86	/* id for IBM EESDI */
#define POSTCB_BADRAM_MAP 0xcc4	/* 128k bit map of defects (4 words) */
#define POSTCB_BADRAM_LOW 0xcd4	/* first byte with correctable ECC errors */
#define POSTCB_BADRAM_HIGH 0xcd4 /* last byte with correctable ECC errors */
#define POSTCB_BADRAM_INFO 0xcdc
#define   POSTCB_RAM_SLOTC_BAD	0x01	/* memory in slot C contains errors */
#define   POSTCB_RAM_SLOTD_BAD	0x02	/* memory in slot D contains errors */
#define   POSTCB_RAM_SLOTA_BAD	0x04	/* memory in slot A contains errors */
#define POSTCB_RAM_INFO	0xcdf	/* same as memory configuration register */
#define   POSTCB_OLDCPU		0x80
#define   POSTCB_SLOWREFRESH	0x08
#define   POSTCB_RAM_SLOTC_MASK 0x07
#define   POSTCB_RAM_SLOTD_MASK 0x70
#define   POSTCB_RAM_4MB	0x0
#define   POSTCB_RAM_1MB	0x1
#define   POSTCB_RAM_8MB	0x4
#define   POSTCB_RAM_2MB	0x5
#define   POSTCB_RAM_512	0x6
#define   POSTCB_RAM_NONE	0x7
#define POSTCB_ROM_DATE	0xce4	/* date ROM was linked */
#define POSTCB_POSTSTART 0xd00	/* address of first POST block */


#ifndef LOCORE

/* a post-structure that will will find an array of starting at 0xd00 */
/* NOTE: the BOOT ROM only fills in info for the disk/diskette adapter(s) */

struct post {
	u_char	post_fill[4];		/* unused */
	u_int	post_error;		/* error code word */
#define POST_ERROR_BAD		0x00000001	/* adapter is bad */
#define POST_ERROR_NOADAPT	0x00000002	/* no adapter present */
#define POST_ERROR_INFO_MASK	0xfffffffc	/* error info */
	u_char	post_adaptype;		/* adapter type (uses ID's above) */
	u_char	post_slot;		/* slot number */
#define POST_PRIMARY_MASK	0x0f
#define POST_SECONDARY_MASK	0xf0
#define POST_SLOT_NONE		0x0f	/* no adapter found */
	u_short	post_primary_addr;	/* primary adapter address */
	u_char	post_interrupt;		/* interrupt levels */
#define POST_INTR_NONE		0x08	/* interrupts not used */
	u_char	post_dma;		/* dma levels */
#define POST_DMA_NONE		0x0f	/* dma not used */
	u_short	post_secondary_addr;	/* secondary adapter address */
} ;

#define	NUM_POSTS	48

struct	postcb {
	u_int	postcb_numbadblock;	/* number of bad blocks of memory (?)*/
	u_int	postcb_badblock[255];	/* bad blocks of memory (?) */
	u_int   postcb_nvram;		/* 0xc00: is nvram valid? */
	u_int	postcb_ipldev;		/* boot device */
	u_int	postcb_kbdid;		/* keyboard id */
	u_int	postcb_machid;		/* machine id */
	u_int	postcb_softipl;		/* soft ipl indicator */
	u_int	postcb_ramsize;		/* size of contiguous memory */
	u_int	postcb_romentry;	/* rom entry point */
	u_int	postcb_resv1;
	u_int	postcb_START;		/* stack pointer save */
	u_int	postcb_XPROL;		/* checks ROM stack requirements */
	u_int	postcb_XPROL0;		/* address of XPROL if not 0 */
	u_int	postcb_romstack;	/* rom stack start for soft ipl */
	u_int	postcb_prior_iar;
	u_int	postcb_prior_ics;
	u_int	postcb_level1_ret;
	u_int	postcb_ramspec;		/* 0xc3c: misc info */
	u_int	postcb_flopparm[7];
	u_int	postcb_flopdig[4];
	u_int	postcb_diskread[6];
	u_int	postcb_slotid[16];
	u_int	postcb_badram_map[4];	/* 0xcc4: */
	u_int	postcb_badram_low;	/* first bad ram address */
	u_int	postcb_badram_high;	/* last bad ram address */
	u_char	postcb_badram_info;
	u_char	postcb_resv2[2];
	u_char	postcb_ram_info;	/* memory config register */
	u_int	postcb_expandrom;
	u_int	postcb_rom_date[2];
	u_int	postcb_romvestige;
	u_int	postcb_resv3;
	u_int	postcb_ram16[3];
	struct post postcb_posts[NUM_POSTS];	/* 0xd00: POST blocks */
};
#endif
