/* 
 * Copyright (c) 1986, 1987
 *	Carnegie-Mellon University.   All rights reserved.
 *
 * This code is derived from software contributed by Carnegie-Mellon 
 * University from the Mach Operating System.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)romp_init.c	2.3 (CMU) 12/19/88
 */

#include <sys/param.h>
#include <sys/systm.h>

#include <vm/vm.h>
#include <vm/vm_kern.h>
#include <vm/vm_page.h>

#include <machine/cpu.h>

vm_size_t	mem_size;
vm_offset_t	first_addr;
vm_offset_t	last_addr;

vm_offset_t	avail_start, avail_end;
vm_offset_t	virtual_avail, virtual_end;

extern int	rose_page_size;
extern int	rose_page_shift;

romp_init(p)
struct hatipt *p;
{
	extern	char	end;
	extern	int	physmem;	/* Set by lohatipt */
	extern	caddr_t	is_in;

	mmu_init(p);	/* initialise the mmu and map pages 1:1 */

	iow(ROMP_TCR,ior(ROMP_TCR) | TCR_I);
		/* Debugging - TLIPT */
	/* INIT_INTR();	/* XXX cnatch needs interrupts...so we initialize now */
#if	ROMP_APC
#if	ROMP_135
	if (cpu_model == CPU_ROMPC || cpu_model == CPU_MOD135)
#endif	ROMP_135
	if (cpu_model == CPU_ROMPC)
		init_ioims(cpu_model);
#endif	ROMP_APC
	/* cnatch(0);	XXX Add our own console attach routine here */
#if	ROMP_APC
	check_mem_config();
#endif	ROMP_APC

#ifdef	SIMULATOR
	cnt.v_page_size = DEFAULT_PAGE_SIZE;	/* 4K */
#else
	/* Initialize CCR */
	* (char *) CCR = CCR_DEFAULT;

	vm_set_page_size();
#endif

	mem_size = physmem * rose_page_size;
#ifdef	SIMULATOR
			/* approx truth for 16M machine */
	first_addr  = 0x00100000;	/* 1M kernel */
	first_addr  = round_page(first_addr);
	last_addr   = 0x00FDD000;
	last_addr   = trunc_page(last_addr);
#else
	first_addr  = (vm_offset_t)(&end) & 0x0fffffff; /* Make it physical */
	first_addr  = round_page(first_addr);
	last_addr   = (((int)ROMP_HATIPT) & 0x0fffffff) - 
			(((sizeof (struct msgbuf) + 2047) >> 11) << 11) - 
			((((mem_size / (rose_page_size * 8)) + 2047) >> 11)
				<< 11);
			/* Assumption1: hat/ipt is at top of memory */
			/* Assumption2: machdep will allocate msgbuf here */
			/* Leave space for the is_in table here, too. */
	last_addr   = trunc_page(last_addr);
	/*Plunk the is_in table right above available memory. */
	/*Make it virtual.*/
	is_in = (caddr_t)((int)last_addr + (int)SYSBASE_ADDR); 
#endif

	avail_start = first_addr;
	avail_end   = last_addr;
	printf("RT/PC Mach, avail_start = 0x%x, avail_end = 0x%x.\n",
		avail_start, avail_end);
	pmap_bootstrap(SYSBASE_ADDR, &avail_start, &avail_end, &virtual_avail,
		&virtual_end);

}


#ifdef	ignore_for_now

#if	ROMP_APC

int ioim1_config;		/* IOIM 1 configuration */
int ioim2_config;		/* IOIM 2 configuration */
int ioim2;			/* non-zero if IOIM2 present */
label_t ioim_label;

#define MEM_2_CYCLE 0x80	/* On if 2 Cycle Memory */
/*
 * initialize both IOIM1 and IOIM2 (IOIM2 requires complete setup
 * IOIM1 has already been initialized by the ROS)
 */
init_ioims(cpu_model)
int cpu_model;
{
	extern int memconfig;	/* memory configuration register (locore) */
	int mem_cycle = (memconfig & MEM_2_CYCLE) != 0 ? 0 : IOIM_2_CYCLE;

	iow(IOIM1_BASE, (BUS1 << 8) + (IOIM1 >> 16));
	ioim1_config = ior(IOIM1 + IOIM_MCC);	/* get configuration */
	/* for now disable parity (chips don't work at 100ns) */
	iow(IOIM1+IOIM_GCR, ior(IOIM1+IOIM_GCR) | IOIM_NOPARITY | mem_cycle);
	/* Turn off user access to FPA, but leave IOCC controlled alone */
	iow(IOIM1+IOIM_SUB_PROB, SEG_0+SEG_4);

#if	ROMP_135
	if (cpu_model != CPU_MOD135)
#endif	ROMP_135
	{
		extern int (*_trap)();
		extern int ioim_trap();
		register int (*oldtrap)() = _trap;
		_trap = ioim_trap;
		if (setjmp(&ioim_label)) {
			_trap = oldtrap;
			return;
		}
		/* initialize IOIM2 for 68881 operation */
		iow(IOIM2_BASE, (BUS2 << 8) + (IOIM2 >> 16));
		ioim2_config = ior(IOIM2 + IOIM_MCC);	/* get configuration */
		iow(IOIM2+IOIM_MCC, IOIM2_MCC);		/* set TAG and ARB */
		iow(IOIM2+IOIM_GCR, IOIM_NOPARITY | IOIM_LOOPMODE);
		iow(IOIM2+IOIM_MISC_0, IOIM2_MISC_0);	/* XXX obsolete hardware */
		iow(IOIM2+IOIM_MISC_1, IOIM2_MISC_1);
		iow(IOIM2+IOIM_SUB_SEG_1, IOIM2_VALUE1);
		iow(IOIM2+IOIM_SUB_PROB, IOIM2_PROB_DISABLE);	/* */
/*		iow(IOIM2+IOIM_SUB_PROB, IOIM2_PROB_ENABLE);	/* for testing */
		iow(IOIM2+IOIM_SUB_SEG_3, IOIM2_VALUE3);
		iow(IOIM2+IOIM_SUB_SEG_4, IOIM2_VALUE4);
		iow(IOIM2+IOIM_SUB_SEG_5, IOIM2_VALUE5);
		_trap = oldtrap;
		ioim2 = IOIM2;
	}
}

ioim_trap()
{
	longjmp(&ioim_label);
}

check_mem_config()
{
 extern int memconfig;	/* memory configuration register (locore) */
 int mem_cycle = (memconfig & MEM_2_CYCLE);
 int x;

 if (cpu_model == CPU_ROMPC)
	{
	if (mem_cycle) {
		printf("Warning: slow memory cards are not supported on APC\n");
		printf("	machine checks are likely\n");
	}
	/* following tests if APC has an 8meg card in first slot - which 
	 * means that either the cards should be swapped so that the 8meg
	 * card is in the second slot, or that the two 8meg cards are present
	 * which is wasteful as the first could be replaced by a 4meg card.
	 */
	if ((memconfig & 0xf07) == 0x804)
		printf("Warning: 4MB of 8MB memory card unusable - %s\n",
			((memconfig & 0x070) == 0x040) ? 
			"4MB card recommended" : "exchange card order");
	/*
	 * test to see if the reply register exists
	 * if it does then the value will change when the
	 * base register is read.
	 */
	if ((ioim1_config & 0x0f00) != 0x0900)
		printf("IOIM1 EC level different than expected\n");
	x = ior(IOIM1+IOIM_REPLY);
	(void) ior(IOIM1+IOIM_BASE);
	if (ior(IOIM1+IOIM_REPLY) == x)
		printf("Warning: back-level IOIM - no reply register\n");
	}
}
#endif	ROMP_APC

#endif

/*
 * Initialise the mmu by looping through the HATIPT and mapping system page
 * i to real page i.
 */
mmu_init(p)
struct hatipt *p;
{
	int page;

	for (page=0; page<physmem; page++, p++) {


		/* all pages are owned by the kernel */
	    p->key = ROMP_KEY_URKW;
	    p->sid = ROMP_SID_SYSTEM;

				/* map pages 1:1 in system segment */
					/* double page value for 4K pages */
	    p->vpage     = rose_page_shift == 11 ? page : page<<1;
	    p->hat_empty = 0;
				/* make sure we don't go overboard */
	    p->hat_ptr = ROMP_HASH(ROMP_SID_SYSTEM, page) > physmem ?
			    ROMP_UNDEF_PTR :
			    ROMP_HASH(ROMP_SID_SYSTEM, page);

				/* there is only ever one page on a chain */
	    p->ipt_last = 1;
	    p->ipt_ptr  = ROMP_UNDEF_PTR;

				/* enable to kernel to write to all pages */
	    p->write_en = 1;

	   			/* tid's are unused, set them all to 1's */
	    p->tid   = 0xff;
	    p->locks = 0xffff;

				/* initialise unused fields to 0's */
	    p->rev_ptr = 0;
	    p->reserved1=p->reserved2=p->reserved3=p->reserved4=p->reserved5 =0;
	}
}
