/* 
 * Copyright (c) 1987, 1988 
 *	Carnegie-Mellon University.   All rights reserved.
 *
 * This code is derived from software contributed by Carnegie-Mellon 
 * University from the Mach Operating System.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)scksum.s	1.00 (CMU) 03/10/89
 */

	#
	# Fast routines for Internet checksum
 	#

	.globl	.oVncs
	.set	.oVncs, 0

	#
	# Simple checksum of a block of data.
	#
	# Callable from C as
	# 	int scksum(start,len)
	#	char *start;	/* start of block to checksum */
	#	int  len;	/* number of bytes in block */
	#
	# Notes: the start address must be aligned on a 16-bit boundary.
	# 	 the number of bytes must be even.
	#
	# On entry:
	#	r2 = start
	#	r3 = len
	#
	# On exit:
	#	r2 = checksum
	#
  	.data
	.globl	_scksum
	.globl	_.scksum
_scksum:
	.long	_.scksum
  	.long	0
  	.long	0

  	.text
_.scksum:

	#
	# Check for zero length
	#
	cis	r3,0
	beq	return

	st	r4,-4(r1)
	#
	# Take care of a possible short word at the head, to get to
	# a 32-bit boundary.
	#
	nilz	r4,r2,2		# check starting address
	bz	after_head
	lhs	r4,0(r2)	# get first short word
	ais	r2,2
	sis	r3,2
after_head:
	s	r5,r5		# clear r5
	a	r5,r5		# clear carry

	#
	# Iterate over all long words.
	#
	cis	r3,4	
	bl	after_long	# if no long words, skip
loop_long:
	ls	r5,0(r2)	# get one long word
	dec	r3,4		# one less word to do
	ae	r4,r5		# add it to checksum
	cis	r3,4
	bhex	loop_long
		inc	r2,4	# update address

	#
	# Take care of a possible short word at the tail.
	#
after_long:
	nilz	r3,r3,2		# r3 = number of remaining bytes
	bz	after_tail
	lhs	r5,0(r2)	# get last short word
	ae	r4,r5

after_tail:
	#
	# Fold to 16 bits and fix all the carries.
	#
	aei	r4,r4,0		# add the last carry
    	aei	r4,r4,0		# add the very last carry
	nilz	r2,r4,0xffff	# r2 = lower bits of r4
	sri16	r4,0		# r4 = upper bits of r4
	a	r2,r4
  	nilz	r4,r2,0xffff
  	sri16	r2,0
  	a	r2,r4
  #	nilz	r4,r2,0xffff
  #	sri16	r2,0
  #	a	r2,r4
	nilz	r2,r2,0xffff

	l	r4,-4(r1)	# restore r4
return:
	br	r15		# return

	.long	0xdf02df00	# trace table
