/*-
 * Copyright (c) 1982, 1986, 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	from: @(#)boot.c	8.1 (Berkeley) 6/10/93
 *	     $Id: boot.c,v 1.3 1994/11/03 08:14:01 roger Exp $
 */

#ifndef lint
static char rcsid[] = "$Id: boot.c,v 1.3 1994/11/03 08:14:01 roger Exp $";
#endif /* not lint */

#include <sys/param.h>
#include <sys/reboot.h>
#include <a.out.h>
#include "stand.h"
#include "samachdep.h"

/*
 * Boot program... bits in `howto' determine whether boot stops to
 * ask for system name.	 Boot device is derived from ROM provided
 * information.
 */

extern	unsigned opendev;
extern	int noconsole;
extern	int testing;

char *ssym, *esym;

char *name;
char *names[] = {
	"/vmunix", "/netbsd", "/onetbsd", "/netbsd.old",
};
#define NUMNAMES	(sizeof(names)/sizeof(char *))

static int bdev, bctlr, bunit, bpart;

main()
{
	int currname = 0;
	int io;

	printf("\n>> NetBSD BOOT RT-%s CPU [%s]\n",
	       getmachineid(), "$Revision:$");

	bdev  = B_TYPE(bootdev);
	bctlr = B_CONTROLLER(bootdev);
	bunit = B_UNIT(bootdev);
	bpart = B_PARTITION(bootdev);

	for (;;) {
		set_twiddle(0);	
		name = names[currname++];
		if (currname == NUMNAMES)
		    currname = 0;

		if (!noconsole) {
		    howto = 0;
		    getbootdev(&howto);
		}
		else
		    printf(": %s\n", name);

		io = open(name, 0);
		if (io >= 0) {
			copyunix(howto, opendev, io);
			close(io);
		}
		else
		    printf("boot: %s\n", strerror(errno));
	}
}

#include <rt/postcb.h>

/* #define BASEMASK     0x0ff80000 */
#define BASEMASK        0x0fff0000
#define NOSEGMASK	0x0fffffff

/*ARGSUSED*/
copyunix(howto, devtype, io)
	register int howto;	/* boot flags */
	register u_int devtype;	/* boot device */
	register int io;
{
	struct exec x;
	int i;
	register char *load;	/* load addr for unix */
	register char *addr;
	int dev, ctlr, unit, part;

	/* XXX use devtype? */
	dev = B_TYPE(opendev);
	ctlr = B_CONTROLLER(opendev);
	unit = B_UNIT(opendev);
	part = B_PARTITION(opendev);
	
	i = read(io, (char *)&x, sizeof(x));
	if (i != sizeof(x) || N_BADMAG(x)) {
		printf("Bad format\n");
		return;
	}

	load = addr = (char *)(x.a_entry & BASEMASK);
	printf("Booting %s%d%c:%s @ 0x%x\n",
	    devsw[dev].dv_name, unit + (8*ctlr), 'a'+part, name, addr);

        /* check if the target will overwrite the this program. */
        if ((((int)x.a_entry) & BASEMASK) == (BASEMASK & (int)main)) {
                (int)addr ^= 0x80000;    /* hope it's relocatable! */
                x.a_entry = (x.a_entry & ~BASEMASK) + (int)addr;
        }

	/* Text */
	set_twiddle(1);		/* ok to twiddle now */
	reset_twiddle();
	printf("%d", x.a_text);
	if (N_GETMAGIC(x) == ZMAGIC && lseek(io, NBPG, SEEK_SET) == -1)
		goto shread;

/* XXX: start */
        /*
         * protect the POST by splitting up any read the crosses the POST
         * into three reads (refusing to read data into the POST.
         */
        if ((int)addr <= POSTCB_START && x.a_text > POSTCB_END) {
                /* text will span the POST */
                register int len = POSTCB_START - (int)addr;
                register char *p;
                char buff[POSTCB_END - POSTCB_START]; /* data for POST */

                if (read(io, (char *)addr, len) != len)
                        goto shread;   /* read up to POST */
/* lseek start... */
                if (read(io, buff, POSTCB_END - POSTCB_START) != POSTCB_END - POSTCB_START)
                        goto shread;   /* read POST */
                for (len = 0, p = buff; p < buff + POSTCB_END - POSTCB_START; ++ p)
                        if (*p)
                                ++len;
/* ...end */

                len = x.a_text - (POSTCB_END - (int)addr);
                if (read(io, (char *)POSTCB_END, len) != len)
                        goto shread;
        } else if ((int)addr + x.a_text < POSTCB_END) {
                printf("text/data inside POST\n");
                led_display(0x96);
        }
        else                           /* fall thru to the normal case */
/* XXX: end */

	if (read(io, (char *)addr, x.a_text) != x.a_text)
		goto shread;

	addr += x.a_text;
	if (N_GETMAGIC(x) == NMAGIC || N_GETMAGIC(x) == ZMAGIC)
		while ((int)addr & CLOFSET)
			*addr++ = 0;
	/* Data */
	reset_twiddle();
	printf("+%d", x.a_data);
	if (read(io, addr, x.a_data) != x.a_data)
		goto shread;
	addr += x.a_data;

	/* Bss */
	reset_twiddle();
	printf("+%d", x.a_bss);
	for (i = 0; i < x.a_bss; i++)
		*addr++ = 0;

	/* Symbols */
	ssym = addr;
#ifdef not_yet
	if (x.a_syms != 0) {
		bcopy(&x.a_syms, addr, sizeof(x.a_syms));
		addr += sizeof(x.a_syms);
		printf(" [%d", x.a_syms);
		if (read(io, addr, x.a_syms) != x.a_syms)
			goto shread;
		addr += x.a_syms;

		/* read size of string table */
		if (read(io, &i, sizeof(int)) != sizeof(int))
			goto shread;

		/* read strings */
		reset_twiddle();
		printf("+%d]", i);
		bcopy(&i, addr, sizeof(int));
		if (i) {
			i -= sizeof(int);
			addr += sizeof(int);
			if (read(io, addr, i) != i)
			    goto shread;
			reset_twiddle();
			addr += i;
		}
	}
#endif /* not_yet */

#define	round_to_size(x,t) \
	(((int)(x) + sizeof(t) - 1) & ~(sizeof(t) - 1))
	esym = (char *)round_to_size(addr - load,int);
#undef round_to_size

	/* and note the end address of all this	*/
	printf(" total=0x%x", addr);

	x.a_entry &= NOSEGMASK;
	printf(" start 0x%x\n", x.a_entry);

#ifdef DEBUG
	printf("ssym=0x%x esym=0x%x\n", ssym, esym);
	printf("\n\nReturn to boot...\n");
	if (getchar() == 'n')
	    return;
#endif

	(*((int (*)()) &x.a_entry))(howto,devtype);

	exit(0);
shread:
	printf("Short read\n");
	return;
}

char line[100];

getbootdev(howto)
     int *howto;
{
	char c, *ptr = line;

	printf("Boot: [[[%s%d%c:]%s][-s][-a][-d]] :- ",
	    devsw[bdev].dv_name, bunit + (8 * bctlr), 'a'+bpart, name);

	if (tgets(line)) {
		while (c = *ptr) {
			while (c == ' ')
				c = *++ptr;
			if (!c)
				return;
			if (c == '-')
				while ((c = *++ptr) && c != ' ')
					switch (c) {
					case 'a':
						*howto |= RB_ASKNAME;
						continue;
					case 's':
						*howto |= RB_SINGLE;
						continue;
					case 'd':
						*howto |= RB_KDB;
						continue;
					case 'b':
						*howto |= RB_HALT;
						continue;
					}
			else {
				name = ptr;
				while ((c = *++ptr) && c != ' ');
				if (c)
					*ptr++ = 0;
			}
		}
	} else
		printf("\n");
}

