
#include <sys/param.h>
#include <dev/isareg.h>
#include <rt/postcb.h>
#include <machine/pio.h>

#include "stand.h"

/* #define FDDEBUG */
#define NFDC		1	/* number of floppy disk controllers, max 2 */
#define	NFD		1	/* number of floppy units supported, max 2 */

struct	fd_softc {		/* controller info */
	u_int	fc_port;
	int	fc_ctlr;	/* my controller number */
	int	fc_unit;	/* current unit that is open */
	int	fc_part;	/* current part that is open */
} fd_softc[NFDC];

fdopen(f, ctlr, unit, part)
	struct open_file *f;
	int ctlr, unit, part;
{
	register struct fd_softc *fs;
	
#ifdef FDDEBUG
printf ("fdopen(%d,%d,%d)\n", ctlr, unit, part);
#endif
	if (ctlr >= NFDC)
		return (EADAPT);
	if (unit >= NFD)
		return (ECTLR);
	fs = &fd_softc[ctlr];
	fs->fc_part = part;
	fs->fc_unit = unit;
	fs->fc_ctlr = ctlr;

	/* init the controller */
	if (fdinit(ctlr, unit) == 0)
		return (ENXIO);

	f->f_devdata = (void *)fs;
	return (0);
}

fdstrategy(fs, func, dblk, size, buf, rsize)
	register struct fd_softc *fs;
	int func;
	daddr_t dblk;
	u_int size;
	char *buf;
	u_int *rsize;
{
	u_int nblk = size / 512;
        long    cylin, head, sector;
	int	secpercyl;
	char stat;

	if (size == 0)
		return(0);

#ifdef FDDEBUG
	printf("fdstrategy: dblk=%d, size=%d, nblk=%d\n", dblk, size, nblk);
#endif

        /* Calculate data for output.           */
	secpercyl = 15*2;		/* XXX: 1.2meg disk */
        cylin = dblk / secpercyl;
        head = (dblk % secpercyl) / 15;
        sector = dblk % 15;
	sector ++;

	twiddle();

	if (rom_flopread(cylin,head,sector,nblk,buf)) {
		printf ("fdstrategy: EIO\n");
		return(EIO);
	}

	*rsize = size;

	return(0);
}

/* initialize a controller */
int
fdinit(ctlr, unit)
	int ctlr, unit;
{
	register struct fd_softc *fs;
	int ret, fdc;

#ifdef FDDEBUG
printf ("fdinit(%d,%d)\n", ctlr, unit);
#endif

	fs = &fd_softc[ctlr];
	fdc = fs->fc_port = (ctlr ? IO_FD2 : IO_FD1);

	ret = rom_floprestore(RT_IO_BASE + fdc + 0x2,unit);
#ifdef FDDEBUG
printf ("fdinit: rom_floprestore returned %d\n", ret);
#endif
	return (!ret);
}

rom_floprestore(adap,drive)
u_int	adap;		/* register 2 */
u_int	drive;		/* register 3 */
{
	struct postcb *postcb = (struct postcb *) POSTCB_START;
	u_int entry, addr2addr();
	static u_int fnaddr = 0;

	if (!fnaddr) {
	    entry = postcb -> postcb_romentry;
	    fnaddr = addr2addr(entry + 0x24);
	}

#ifdef FDDEBUG
	printf ("rom_floprestore(0x%x,%d)\n", adap, drive);
#endif

	/* call a ROM routine.  
	 * NOTE: r1 gets destroyed, so we save it.  r14
	 * must point to a decremented stack.  So for "ease", we
	 * point it at the ROM stack already setup for us.
	 */
	asm volatile("   cas r2,%0,r0" : : "r" (adap));
	asm volatile("   cas r3,%0,r0" : : "r" (drive));
	asm volatile("   cas r6,r1,r0" : : : "r6" );
	asm volatile("   cas r7,r14,r0" : : : "r7" );
	asm volatile("   ls r14,%0" : : "m" (postcb->postcb_START));
	asm volatile("   balr r15,%0" : : "r" (fnaddr));
	asm volatile("   cas r14,r7,r0" : : : "r7" );
	asm volatile("   cas r1,r6,r0" : : : "r6" );

	/* return value in r2 */
}

rom_flopread(cyl,trk,sec,num,dest)
u_int	cyl;	
u_int	trk;	
u_int	sec;
u_int	num;
u_int	dest;
{
	struct postcb *postcb = (struct postcb *) POSTCB_START;
	u_int entry, addr2addr();
	static u_int fnaddr = 0;

	if (!fnaddr) {
	    entry = postcb -> postcb_romentry;
	    fnaddr = addr2addr(entry + 0x08);
	}

#ifdef FDDEBUG
	printf ("rom_flopread(%x,%x,%x,%x,%x)\n", cyl, trk, sec, num, dest);
#endif

	/* call a ROM routine.  
	 * NOTE: r1 gets destroyed, so we save it.  r14
	 * must point to a decremented stack.  So for "ease", we
	 * point it at the ROM stack already setup for us.
	 */
	asm volatile("   ls r9,%0" : : "m" (postcb->postcb_START) : "r9" );
	asm volatile("   cas r2,%0,r0" : : "r" (cyl) : "r2" );
	asm volatile("   cas r3,%0,r0" : : "r" (trk) : "r3" );
	asm volatile("   cas r4,%0,r0" : : "r" (sec) : "r4" );
	asm volatile("   cas r5,%0,r0" : : "r" (num) : "r5" );
	asm volatile("   cas r6,%0,r0" : : "r" (dest) : "r6" );
	asm volatile("   cas r7,r1,r0" : : : "r7" );
	asm volatile("   cas r8,r14,r0" : : : "r8" );
	asm volatile("   cas r14,r9,r0" : : : "r9", "r14" );
	asm volatile("   balr r15,%0" : : "r" (fnaddr));
	asm volatile("   cas r14,r8,r0" : : : "r8" );
	asm volatile("   cas r1,r7,r0" : : : "r7" );

	/* return value in r2 */
}
