
#include <sys/param.h>
#include <sys/disklabel.h>
#include "hdcompat.h"

#include "stand.h"

/* #define HDDEBUG */

char *
hdgetlabel (sc, strategy, lp)
	caddr_t	sc;			/* driver's soft-c info */
	int	(* strategy)();		/* driver's strategy routine */
	register struct disklabel *lp;	/* disklabel to fill-up */
{
	int i, err, part, maxpart;
	char *msg, io_buf[DEV_BSIZE];
	config_rec *cf;
	minidisk_dir *md;
	mini_dirent *mdr;

	if (err = (* strategy)(sc, F_READ, CONFIG_REC_LSN, DEV_BSIZE,
		       io_buf, &i) < 0) {
	    msg = "hdgetlabel: strategy error";
	    return(msg);
	}

	cf = (config_rec *) io_buf;
	if (cf -> id != CONFIG_MAGIC) {
	    msg = "hdgetlabel: no CONFIG record";
	    return(msg);
	}

	lp->d_magic = DISKMAGIC;
	lp->d_magic2 = DISKMAGIC;
	lp->d_rpm = 3600;
	lp->d_interleave = 1;
	lp->d_type = DTYPE_ESDI;
	lp->d_secsize = DEV_BSIZE;
	lp->d_nsectors = cf->last_sector;
	lp->d_ntracks = cf->last_head + 1;
	lp->d_ncylinders = cf->CE_cyl + 1;
	lp->d_secpercyl = lp->d_nsectors * lp->d_ntracks;
	lp->d_secperunit = lp->d_secpercyl * lp->d_ncylinders;

#ifdef HDDEBUG
	printf ("nsect=%d, ntrack=%d, ncyl=%d, secpercyl=%d, secperunit=%d\n",
		lp->d_nsectors, lp->d_ntracks, lp->d_ncylinders, 
		lp->d_secpercyl,lp->d_secperunit);
#endif

	if (err = (* strategy)(sc, F_READ, MD_DIR_LSN, DEV_BSIZE,
		       io_buf, &i) < 0) {
	    msg = "hdgetlabel: strategy error";
	    return(msg);
	}

	md = (minidisk_dir *) io_buf;
#ifdef HDDEBUG
	printf ("nmini=%d, bad_lsn=%d, bad_size=%d, first=%d\n",
		md->nmini, md->bad_lsn, md->bad_size, md->first);
#endif

	/* XXX: we *always* have a 'c' partition that defines the entire disk*/
	lp->d_partitions[2].p_size = lp->d_secperunit;
	lp->d_partitions[2].p_offset = 0;
	maxpart = 3;

	/*
	 * loop thru the minidisk entries, and make up a
	 * disklabel partition table.
	 */
	mdr = &md->entry[md->first];
	for (i = 0; i < md->nmini; i++, mdr = &md->entry[mdr->next]) {
#ifdef HDDEBUG
	    printf("%d: name=%c%c%c%c, iodn=%d, type=0x%x, start=%d, size=%d\n",
		    i, mdr->name[0],mdr->name[1],mdr->name[2],mdr->name[3],
		    mdr->iodn, mdr->type, mdr->lsn_start, mdr->nsectors);
#endif
	    if (!mdr->iodn)
		continue;		/* XXX: free space */
	    part = mdr->name[3] - 'a';	/* XXX: aka: hd0a */
	    if (part > MAXPARTITIONS) {
		continue;		/* XXX: unknown disk name */
	    }

	    if (part >= maxpart)
		maxpart = part + 1;
	    lp->d_partitions[part].p_size = mdr->nsectors;
	    lp->d_partitions[part].p_offset = mdr->lsn_start;

	}
	lp->d_npartitions = maxpart;

	msg = (char *)0;
	return(msg);
}
