

#define CRRB		0x8C60	/* should be elsewhere */
#define KB_CRRB_MASK	0xFB

/* read operations */
#define READ_8255_PA	0x8404
#define READ_8255_PB	0x8405
#define READ_8255_PC	0x8406
#define   PA_OUTPUT_EMPTY 0x80	/* PA output buffer is empty */
#define	  PA_INPUT_FULL 0x20	/* PA input buffer is full */
#define   AIIR_MASK	0x7	/* mask of interrupt bits in PC register */
#define   INT_RECV_KB	0x1	/* byte received from keyboard */
#define	  INT_RECV_UART	0x2	/* byte received from UART */
#define	  INT_RET_BYTE	0x3	/* returning byte requested by system */
#define   INT_BLOCK_RDY 0x4	/* block transfer ready */
#define   INT_8051_TEST 0x6	/* 8051 self test completed */
#define	  INT_8051_ERR  0x7	/* 8051 detected an error condition */

/* write operations */
#define WRITE_8255_PA	0x8400
#define PA_CMD(cmd,data)	((cmd) + ((data) <<8))
#define   WR_EXTENDED	0x00	/* select extended command set */
#define   WR_KEYBOARD	0x01	/* write to keyboard */
#define	  WR_SPEAKER	0x02	/* write to speaker */
#define   WR_UART_CTRL	0x03	/* write UART control (no response) */
#define   WR_UART_QURY	0x04	/* write UART query (response expected) */
#define   WR_UART_BAUD	0x05	/* set UART baud rate */
#define   WR_UART_FRAME	0x06	/* initialize UART framing */
#define   WR_SPK_DUR	0x07	/* set speaker duration */
#define WRITE_8255_CONF	0x8407

/* write responses */
#define KBW_ACCEPT	0x00	/* keyboard command accepted */
#define KBW_BUSY	0x41	/* adapter is busy with previous transmission*/
#define KBW_LOCKED	0x42	/* keylock switch is on */
#define KBW_DISABLED	0x43	/* keyboard interface is disabled */
#define KBW_INVALID	0x44	/* data byte is invalid */

/* keyboard commands (chpt 9) */
#define KBC_RESET	0xff	/* perform power-on check and report */
#define KBC_ALLMAKE	0xf9	/* set all keys to make */
#define KBC_ALLMAKEBRK	0xf8	/* set all keys to make/break */
#define KBC_ALLREPEAT	0xf7	/* set all keys to repeat */
#define KBC_DEFAULT	0xf6	/* set all keys to default (repeat-make/break)*/
#define KBC_DISABLE	0xf5	/* disable scanning */
#define KBC_ENABLE	0xf4	/* enable scanning */
#define KBC_READ_ID	0xf2
#define KBC_SET_LED	0xed	/* set LED indicator */
#define   KBC_LED_NUMLOCK	0x20	/* set numlock led */
#define   KBC_LED_CAPSLOCK	0x40	/* set capslock led */
#define   KBC_LED_SCROLLLOCK	0x80	/* set scrolllock led */

/* keyboard responses */
#define KBR_BAT		0xaa	/* successful completion of self test */
#define KBR_ID1		0xbf	/* id byte 1 */
#define KBR_ID2		0xb0	/* id byte 2 */

/* specific keyboard key sequences */
#define KEY_BREAK	0xf0
#define KEY_CAPSLOCK	0x14
#define KEY_SHIFT_LEFT	0x12
#define KEY_SHIFT_RIGHT	0x59
#define KEY_CONTROL	0x11
#define KEY_ALT_LEFT	0x19
#define KEY_ALT_RIGHT	0x39
#define KEY_ACTION	0x58
#define KEY_NUMLOCK	0x76

#define KEY_MOD_SHIFTED		0x01
#define KEY_MOD_CAPSLOCK	0x02
#define KEY_MOD_NUMLOCK		0x04
#define KEY_MOD_CONTROL		0x08
#define KEY_MOD_ALT		0x10
#define KEY_MOD_ACTION		0x20
