/*
 * Copyright (c) 1988 University of Utah.
 * Copyright (c) 1982, 1986, 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * from: Utah $Hdr: machdep.c 1.10 92/06/18
 * from: @(#)machdep.c	8.1 (Berkeley) 6/10/93
 *
 * $Id: machdep.c,v 1.2 1994/11/03 07:08:19 roger Exp $
 */

#include <sys/param.h>
#include <sys/reboot.h>
#include <machine/pio.h>
#include <rt/postcb.h>
#include "samachdep.h"

#include "cpu.h"

char *
getmachineid()
{
	struct postcb *postcb = (struct postcb *) POSTCB_START;
	char *cp;
	static char *id = "6150/???  ";
	extern int cpu;

	if (postcb -> postcb_machid)
	    id[3] = '1';
	cp = id+5;	/* past the '/' */
	switch (cpu) {
		case CPU_EAPC: 
			*cp++ = 'E'; 
			/* FALL-THRU */
		case CPU_APC: 
			*cp++ = 'A'; *cp++ = 'P'; *cp++ = 'C'; *cp = '\0';
			break;
	}
	return id;
}

void
insw(port,dest,len)
int port;
short *dest;
int len;
{ 
	while(len--) 
		*dest++ = inw (port);
}

void
outsw(port,dest,len) 
int port;
short *dest;
int len;
{ 
	while(len--) 
		outw ((port), dest++);
}

/*
 * Convert RT ROS info to a valid bootdev layout:
 *      TYPE is looked up in devsw table. (major numbers must match!!!)
 *      PARTITION is set to 0 ('a')
 *      UNIT comes from ipldev
 *      CONTROLLER comes from ipldev
 *      ADAPTER not used on the RT.
 */
u_long
rostobdev ()
{
	struct postcb *postcb = (struct postcb *) POSTCB_START;
	char *devname;
	u_long bdev;
	int dev, unit, ctlr = 0;

	switch (postcb -> postcb_ipldev >> 4) {
	    case 0xd: devname = "wd"; break;
	    case 0xf: devname = "fd"; break;
	    default: devname = "??"; break;
	}
	if ((dev = devlookup(devname, 2)) < 0) {
#if 0
	    printf ("could not determine the boot device\n");
#endif
	    return bdev;
	}

	/* XXX: the following might not be true for booting off of
	 *	the second EESDI card....
	 */
	unit = atoi (postcb -> postcb_ipldev & 0x0f);
	if (unit >= 2) {
	    ctlr ++;
	    unit -= 2;
	}

	bdev = MAKEBOOTDEV(dev, 0, ctlr, unit, 0);

#if 0
	printf ("ipldev=%x, dev=%x, ctlr=%x, unit=%x, bdev set to 0x%x\n", 
		postcb -> postcb_ipldev, dev, ctlr, unit, bdev);
#endif

	return bdev;
}

u_int
addr2addr(addr)
u_int addr;
{
	u_int low, high;

	high = INW(addr);
	low = INW(addr + 2);
	return ((high << 16) + low);
}

#if 0
rombeep()
{
	struct postcb *postcb = (struct postcb *) POSTCB_START;
	u_int entry, fnaddr, addr2addr();

	entry = postcb -> postcb_romentry;
	fnaddr = addr2addr(entry + 0x14);

	/* call a ROM routine.  
	 * NOTE: r1 gets destroyed, so we save it.  r14
	 * must point to a decremented stack.  So for "ease", we
	 * point it at the ROM stack already setup for us.
	 */
	asm volatile("   cas r6,r1,r0" : : : "r6" );
	asm volatile("   cas r7,r14,r0" : : : "r7" );
	asm volatile("   ls r14,%0" : : "m" (postcb->postcb_START));
	asm volatile("   balr r15,%0" : : "r" (fnaddr));
	asm volatile("   cas r14,r7,r0" : : : "r7" );
	asm volatile("   cas r1,r6,r0" : : : "r6" );
}

romled(val,len)
u_int	val;		/* register 2 */
u_int	len;		/* register 3 */
{
	struct postcb *postcb = (struct postcb *) POSTCB_START;
	u_int entry, fnaddr, addr2addr();

	entry = postcb -> postcb_romentry;
	fnaddr = addr2addr(entry + 0x10);

	/* call a ROM routine.  
	 * NOTE: r1 gets destroyed, so we save it.  r14
	 * must point to a decremented stack.  So for "ease", we
	 * point it at the ROM stack already setup for us.
	 */
	asm volatile("   cas r2,%0,r0" : : "r" (val));
	asm volatile("   cas r3,%0,r0" : : "r" (len));
	asm volatile("   cas r6,r1,r0" : : : "r6" );
	asm volatile("   cas r7,r14,r0" : : : "r7" );
	asm volatile("   ls r14,%0" : : "m" (postcb->postcb_START));
	asm volatile("   balr r15,%0" : : "r" (fnaddr));
	asm volatile("   cas r14,r7,r0" : : : "r7" );
	asm volatile("   cas r1,r6,r0" : : : "r6" );
}
#endif
