/*-
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * William Jolitz.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	from: @(#)wd.c	7.3 (Berkeley) 5/4/91
 *	$Id: wd.c,v 1.3 1994/11/03 07:08:19 roger Exp $
 */

/*  device driver for winchester disk  */

#include <sys/param.h>
#include <sys/dkbad.h>
#include <sys/disklabel.h>
#include <rt/dev/isareg.h>
#include "wdreg.h"
#include <machine/pio.h>

#include "stand.h"

#define WDCNDELAY	100000	/* delay = 100us; so 10s for a controller state change */
#define WDCDELAY	100

#define SMALL
#define NWDC		1	/* number of hard disk controllers, max 2 */
#define	NWD		1	/* number of hard disk units supported, max 3 */
#define	RETRIES		5	/* number of retries before giving up */


/* transform a large disklabel into something much smaller */
struct	wdminilabel {
	u_char	wds_flags;	/* drive was initialized, etc */
#define WDS_LABEL	0x1
	u_int	wds_secsize;
	u_int	wds_secpercyl;
	u_int	wds_secperunit;
	u_int	wds_nsectors;
	u_int	wds_ncylinders;
	u_int	wds_ntracks;
	u_short	npart;
	u_long	offset[MAXPARTITIONS];
};

struct	wd_softc {		/* controller info */
	u_int	wc_port;
	int	wc_ctlr;	/* my controller number */
	int	wc_unit;	/* current unit that is open */
	int	wc_part;	/* current part that is open */
	char	wc_flags;	/* controller was initialized, etc */
#define WD_ALIVE	0x1
#define WD_EESDI	0x2
	u_char	wc_error;	/* copy of error register */
	u_char	wc_status;	/* copy of status register */
	struct	wdminilabel wc_pinfo[NWD];	/* slave info */
} wd_softc[NWDC];

int noretries, wdquiet;
/* #define WDDEBUG */

struct disklabel wdlabel;

/*
 * Record for the bad block forwarding code.
 * This is initialized to be empty until the bad-sector table
 * is read from the disk.
 */

struct	dkbad	dkbad[NWD];

char io_buf[DEV_BSIZE];


static int wdsetctlr __P((struct wd_softc *));

/* ST506 spec says that if READY or SEEKCMPLT go off, then the read or write
   command is aborted. */
#define	wait_for_drq(d)		wdcwait(d, WDCS_READY | WDCS_SEEKCMPLT | WDCS_DRQ)
#define	wait_for_ready(d)	wdcwait(d, WDCS_READY | WDCS_SEEKCMPLT)
#define	wait_for_unbusy(d)	wdcwait(d, 0)


wdgetinfo(ws)
	register struct wd_softc *ws;
{
	register int unit = ws->wc_unit;
	register struct wdminilabel *pi = &ws->wc_pinfo[unit];
	register struct disklabel *lp = &wdlabel;
	char *msg, *getdisklabel();
	int wdstrategy(), i, err, part = ws->wc_part;

	bzero((caddr_t)lp, sizeof *lp);

	ws->wc_part = 2;	/* XXX: always use 'c' partition */
	if (err = wdstrategy(ws, F_READ, LABELSECTOR, DEV_BSIZE,
		       io_buf, &i) < 0) {
	    printf("wdgetinfo: wdstrategy error %d\n", err);
	    ws->wc_part = part;
	    return(0);
	}
	
	msg = getdisklabel(io_buf, lp);
	if (msg) {
		printf("wd(%d,%d,%d): WARNING: %s, ",
		       ws->wc_ctlr, ws->wc_unit, ws->wc_part, msg);
#ifdef COMPAT_NOLABEL
		/*
		 * call the compatibility routines to try to make
		 * up a disklabel for us from "other" info stored
		 * on the drive.
		 */
		msg = (char *) hdgetlabel (ws, wdstrategy, lp);
		if (msg) {
			printf("\nWARNING: %s, ", msg);
#endif
			printf("defining `c' partition as entire disk\n");
			pi->npart = 3;
			pi->offset[0] = pi->offset[1] = -1;
			pi->offset[2] = 0;
			return 0;	/* DONT set disk parameters */
#ifdef COMPAT_NOLABEL
		}
		printf("using hdcompat disklabel\n");
#endif
	}
#ifdef HDDEBUG
	printf ("found %d partitions\n", lp->d_npartitions);
	for (i = 0; i < lp->d_npartitions; i++) {
	    if (lp->d_partitions[i].p_size == 0)
		continue;
	    printf ("%c size=%d offset=%d\n", 'a' + i,
		lp->d_partitions[i].p_size,
		lp->d_partitions[i].p_offset);
	}
#endif

	/* setup the minilabel */
	ws->wc_part = part;		/* restore it */
	pi->wds_flags |= WDS_LABEL;
	pi->wds_secsize = lp->d_secsize;
	pi->wds_secpercyl = lp->d_secpercyl;
	pi->wds_secperunit = lp->d_secperunit;
	pi->wds_nsectors = lp->d_nsectors;
	pi->wds_ncylinders = lp->d_ncylinders;
	pi->wds_ntracks = lp->d_ntracks;
	pi->npart = lp->d_npartitions;
	for (i = 0; i < pi->npart; i++) 
		pi->offset[i] = lp->d_partitions[i].p_size == 0 ?
			-1 : lp->d_partitions[i].p_offset;

	return(1);
}

wdopen(f, ctlr, unit, part)
	struct open_file *f;
	int ctlr, unit, part;
{
	register struct wd_softc *ws;
	register struct wdminilabel *pi;
	
#ifdef WDDEBUG
printf ("wdopen(%d,%d,%d)\n", ctlr, unit, part);
#endif
	if (ctlr >= NWDC)
		return (EADAPT);
	if (unit >= NWD)
		return (ECTLR);
	ws = &wd_softc[ctlr];
	pi = &ws->wc_pinfo[unit];
	ws->wc_part = part;
	ws->wc_unit = unit;
	ws->wc_ctlr = ctlr;

	/* init the controller */
	if (ws->wc_flags == 0 && wdinit(ctlr, unit) == 0)
		return (ENXIO);

	/* init the drive */
	if (pi->wds_flags == 0 && wdsinit(ws) == 0)
		return (ws->wc_error);

	/* check partition number */
	if (part >= ws->wc_pinfo[unit].npart 
	    || ws->wc_pinfo[unit].offset[part] == -1)
		return (EPART);
	f->f_devdata = (void *)ws;
	return (0);
}

wdstrategy(ws, func, dblk, size, buf, rsize)
	register struct wd_softc *ws;
	int func;
	daddr_t dblk;
	u_int size;
	char *buf;
	u_int *rsize;
{
	register int ctlr = ws->wc_ctlr;
	register int unit = ws->wc_unit;
	daddr_t blk = dblk + ws->wc_pinfo[unit].offset[ws->wc_part];
	u_int nblk = size / ws->wc_pinfo[unit].wds_secsize;
	int nblks;
	char stat;

	if (size == 0)
		return(0);

	twiddle();
        while (nblk > 0) {
                if ((nblks = wdio(ws, func, unit, blk, buf, nblk)) < 0) {
			printf ("wdstrategy: EIO\n");
                        return(EIO);
		}
		nblk -= nblks;
		blk += nblks;
                buf += nblks * ws->wc_pinfo[unit].wds_secsize;
        }
	*rsize = size;

	return(0);
}

/* 
 * Routine to do a one-sector I/O operation, and wait for it
 * to complete.
 */
int
wdio(ws, func, unit, blknm, addr, nblks)
	register struct wd_softc *ws;
	int func;
	int unit;
	daddr_t blknm;
        short *addr;
	int nblks;
{
	register struct wdminilabel *pi = &ws->wc_pinfo[unit];
	register wdc = ws->wc_port;
	struct bt_bad *bt_ptr;
        int    i;
	int retries = 0;
        long    cylin, head, sector;
        u_char opcode, erro;

	opcode = (func == F_WRITE) ? WDCC_WRITE : WDCC_READ;
	if (nblks > 64)
	    nblks = 64;

        /* Calculate data for output.           */
        cylin = blknm / pi->wds_secpercyl;
        head = (blknm % pi->wds_secpercyl) / pi->wds_nsectors;
        sector = blknm % pi->wds_nsectors;

#ifdef DO_BBT
	/* 
	 * See if the current block is in the bad block list.
	 */
	if (blknm > BBSIZE/DEV_BSIZE)	/* should be BBSIZE */
	    for (bt_ptr = dkbad[unit].bt_bad; bt_ptr->bt_cyl != -1; bt_ptr++) {
		if (bt_ptr->bt_cyl > cylin)
			/* Sorted list, and we passed our cylinder. quit. */
			break;
		if (bt_ptr->bt_cyl == cylin &&
			bt_ptr->bt_trksec == (head << 8) + sector) {
			/*
			 * Found bad block.  Calculate new block addr.
			 * This starts at the end of the disk (skip the
			 * last track which is used for the bad block list),
			 * and works backwards to the front of the disk.
			 */
#ifdef WDDEBUG
			    printf("--- badblock code -> Old = %d; ",
				blknm);
#endif
			    printf("--- badblock code -> Old = %d; ",
				blknm);
			blknm = pi->wds_secperunit - pi->wds_nsectors
				- (bt_ptr - dkbad[unit].bt_bad) - 1;
			cylin = blknm / pi->wds_secpercyl;
			head = (blknm % pi->wds_secpercyl) / pi->wds_nsectors;
			sector = blknm % pi->wds_nsectors;
#ifdef WDDEBUG
			    printf("new = %d\n", blknm);
#endif
			break;
		}
	}
#endif

        sector += 1;
retry:
	wait_for_unbusy(ws);

#ifdef WDDEBUG
	printf("sec %d sdh %x cylin %d\n", sector,
		WDSD_IBM | (unit<<4) | (head & 0xf), cylin);
#endif

	wdcommand(ws, cylin, head & 0xf, sector, nblks, opcode);

	for (i = 0; i < nblks; i++) {
	    if (wait_for_unbusy(ws) && opcode == WDCC_READ) 
		goto error;
	    if (wait_for_drq(ws)) 
		goto error;

	    if (opcode == WDCC_READ) 
		insw(wdc+wd_data,addr,256);
	    else
		outsw(wdc+wd_data,addr,256);

#if 0	    /* not true on EESDI...!!! */
	    /* Check data request (should be done).         */
	    if (inb(wdc+wd_status) & WDCS_DRQ) 
		goto error;
#endif

	    if (opcode == WDCC_WRITE && wait_for_unbusy(ws))
		goto error;

	    if (inb(wdc+wd_status) & WDCS_ERR) 
		goto error;

	    addr += 512 / sizeof(*addr);
	}

#ifdef WDDEBUG
printf("addr %x\n",addr);
#endif
        return (nblks);
error:
	erro = inb(wdc+wd_error);
	if (++retries < RETRIES) {
		goto retry;
	}
	if (!wdquiet)
	    printf("wd%d: hard %s error: sector %d status %b error %b\n", unit,
		opcode == WDCC_READ? "read" : "write", blknm, 
		inb(wdc+wd_status), WDCS_BITS, erro, WDERR_BITS);
	return (-1);
}

/* initialize a controller */
int
wdinit(ctlr, unit)
	int ctlr, unit;
{
	register struct wd_softc *ws;
	register struct wdminilabel *pi;
	register wdc;
	struct dkbad *db;
	int i;

#ifdef WDDEBUG
printf ("wdinit(%d,%d)\n", ctlr, unit);
#endif

	/* do some basic setup to boot-strap ourselves */
	/* NOTE: we purposely side-step setting any drive
	 *	parameters until we read the disklabel off of
	 *	the drive.  This keeps the EESDI card happy,
	 *	and should work fine with generic ESDI/IDE.
	 */
	ws = &wd_softc[ctlr];
	wdc = ws->wc_port = (ctlr ? IO_WD2 : IO_WD1);
	for (i = 0; i < NWD; i++) {
	    pi = &ws->wc_pinfo[unit];
	    bzero((caddr_t)pi, sizeof *pi);	/* no partition table yet */
	    pi->wds_secsize = DEV_BSIZE;
	    pi->wds_nsectors = 17;	
	    pi->wds_ntracks = 2;
	    pi->wds_secpercyl = pi->wds_nsectors * pi->wds_ntracks;
	}

	/* reset controller */
	outb(wdc+wd_ctlr,WDCTL_RST|WDCTL_IDS);
	DELAY(1000);
	outb(wdc+wd_ctlr,WDCTL_IDS);

	/* if stuck-on-busy, issue a DIAGNOSE command */
	/* XXX: the EESDI card doesn't reset until issued a DIAGNOSE */
	if (wait_for_unbusy (ws)) {
	    char c;

	    printf ("wdinit: issuing a DIAGNOSE\n");
	    outb(wdc+wd_command,WDCC_DIAGNOSE);
	    c = inb(wdc+wd_error);
	    delay (1000000);
	    if ((c & 0x1f) != 0x01) {
		printf ("wdinit: diagnose failed: 0x%x:0x%x\n", c, c&0x1f);
		return 0;	/* bad controller ? */
	    }
	    else {
		printf ("wdinit: diagnose OK\n");
	    }
	}
	outb(wdc+wd_ctlr,WDCTL_4BIT);

	ws->wc_flags |= WD_ALIVE;

	/* determine if this is an EESDI card */

	return 1;
}

/* initialize a slave (ie: harddrive) */
int
wdsinit(ws)
	register struct wd_softc *ws;
{
	register int unit = ws->wc_unit;
	register struct wdminilabel *pi = &ws->wc_pinfo[unit];
	register wdc = ws->wc_port;
	int i, errcnt = 0;

#ifdef WDDEBUG
printf ("wdsinit(%d)\n", unit);
#endif
	/* set SDH, step rate, do restore to recalibrate drive */
tryagainrecal:
	outb(wdc+wd_sdh, WDSD_IBM | (unit << 4));
	outb(wdc+wd_command, WDCC_RESTORE | WD_STEP);

	if ((i = wait_for_ready(ws)) == WDCS_ERR) {
		printf("wd%d: recal status %b error %b\n",
			unit, ws->wc_status, WDCS_BITS, 
			ws->wc_error, WDERR_BITS);
		if (++errcnt < 10)
			goto tryagainrecal;
		return 0 ;
	}

	if (wdgetinfo(ws) == 0) {	/* get disklabel info from drive */
		ws->wc_error = ERDLAB;
		return 0;
	}
	if (wdsetctlr(ws) == 0)	{	/* set drive parameters *.
		ws->wc_error = ENXIO;	/* ??? */
		return 0;
	}

#ifdef DO_BBT
	dkbad[unit].bt_bad[0].bt_cyl = -1;

	if (dd->d_flags & D_BADSECT) {
	/*
	 * Read bad sector table into memory.
	 */
	i = 0;
	do {
		int blknm = dd->d_secperunit - dd->d_nsectors + i;
		errcnt = wdio(F_READ, unit, blknm, io_buf);
	} while (errcnt && (i += 2) < 10 && i < dd->d_nsectors);
	db = (struct dkbad *)(io_buf);
#define DKBAD_MAGIC 0x4321
	if (errcnt == 0 && db->bt_mbz == 0 && db->bt_flag == DKBAD_MAGIC)
		dkbad[unit] = *db;
	else {
		if (!wdquiet)
			printf("wd%d: error in bad-sector file\n", unit);
		dkbad[unit].bt_bad[0].bt_cyl = -1;
	}
	}
#endif

	return(1);
}


/*
 * Send a command and wait uninterruptibly until controller is finished.
 * Return -1 if controller busy for too long, otherwise return non-zero if
 * error.  Intended for brief controller commands at critical points.
 * Assumes interrupts are blocked.
 */
static int
wdcommand(wd, cylin, head, sector, count, cmd)
        struct wd_softc *wd;
        int cylin, head, sector, count;
        int cmd;
{
        int stat;
        u_short iobase;

        /* Select drive. */
        iobase = wd->wc_port;
        outb(iobase+wd_sdh, WDSD_IBM | (wd->wc_unit << 4) | head);
        if (cmd == WDCC_DIAGNOSE || cmd == WDCC_IDC)
                stat = wait_for_unbusy(wd);
        else
                stat = wdcwait(wd, WDCS_READY);
        if (stat < 0)
                return 0;

        /* Load parameters. */
/*         outb(iobase+wd_precomp, wd->sc_label.d_precompcyl / 4); */
        outb(iobase+wd_cyl_lo, cylin & 0xff);
        outb(iobase+wd_cyl_hi, cylin >> 8);
        outb(iobase+wd_sector, sector);
        outb(iobase+wd_seccnt, count);

        /* Send command, await results. */
        outb(iobase+wd_command, cmd);

        return 1;
}

/*
 * Issue IDC to drive to tell it just what geometry it is to be.
 */
static int
wdsetctlr(wd)
        struct wd_softc *wd;
{
	register int unit = wd->wc_unit;
	register struct wdminilabel *pi = &wd->wc_pinfo[unit];

#ifdef WDDEBUG
        printf("wd(%d,%d) C%dH%dS%d\n", wd->wc_ctlr, wd->wc_unit,
            pi->wds_ncylinders, pi->wds_ntracks, pi->wds_nsectors);
#endif

        if (wdcommand(wd, pi->wds_ncylinders, pi->wds_ntracks - 1,
            0, pi->wds_nsectors, WDCC_IDC) == 0) {
                printf("wdsetctlr: wdcommand failed\n");
                return 0;
        }

        return 1;
}

int
wdcwait(wd, mask)
	struct wd_softc *wd;
	int mask;
{
	u_short iobase = wd->wc_port;
	int timeout = 0;
	u_char status;
	extern int cold;

	for (;;) {
		wd->wc_status = status = inb(iobase+wd_status);
		if ((status & WDCS_BUSY) == 0 && (status & mask) == mask)
			break;
		if (++timeout > WDCNDELAY)
			return -1;
		delay(WDCDELAY);
	}
	if (status & WDCS_ERR) {
		wd->wc_error = inb(iobase+wd_error);
		return WDCS_ERR;
	}
	return 0;
}
