/*
 * P_R_P_Q_# (C) COPYRIGHT IBM CORPORATION 1989
 * LICENSED MATERIALS - PROPERTY OF IBM
 * REFER TO COPYRIGHT INSTRUCTIONS FORM NUMBER G120-2083
 */

/*
 * Windgate (Pegasus) Ethernet driver for PS/2
 *
 * (c) Copyright 1987 Locus Computing Corporation
 *
 * this driver supports the NICPS2 adapter from Ungermann Bass
 * which is based on the Intel 82586 chip. See the Intel Microcomunications
 * handbook for hardware details.
 */

#ifdef AIX
#ifndef SCCSID
#include	<sccs.h>
#endif /* ! SCCSID */
SCCSID(@(#)pegasus.c	1.32	AIX)	/* Modified 8/18/88 21:31:22 */
/*SCCSID(@(#)pegasus.c	1.31	LCC)	/* Modified: 18:17:37 8/4/88 */
#endif AIX

#define kludge_flush()	/* do nothing */
#define inb(port) IN(port)
#define outb(port,value) OUT(port,value)
#define swaph(value) ((((value)>>8)&0xff)|(((value)&0xff)<<8))
#define PGIRQ 3

#include	"pg.h"
#include "../machine/pte.h"

#include "param.h"
#include "systm.h"
#include "mbuf.h"
#include "buf.h"
#include "protosw.h"
#include "socket.h"
#include "vmmac.h"
#include "ioctl.h"
#include "errno.h"

#include "../net/if.h"
#include "../net/netisr.h"
#include "../net/route.h"

#ifdef INET
#include "../netinet/in.h"
#include "../netinet/in_systm.h"
#include "../netinet/in_var.h"
#include "../netinet/ip.h"
#include "../netinet/if_ether.h"
#endif INET

#ifdef NS
#include "../netns/ns.h"
#include "../netns/ns_if.h"
#endif NS

#include "../machine/io.h"
#include "if_unreg.h"
#include "../machineio/ioccvar.h"
#include "time.h"
#include "kernel.h"
#include "../machine/debug.h"
#include "../ca_atr/pcif.h"



#define		PGCARDID	0xEFF5		/* POS register ID */
/* bits in the CSR registers */
#define		INTPNDG		0x80
#define		CHINTENB	0x08
#define		TMINTENB	0x04
#define		LOOPBACK	0x02
#define		RESET		0x01


/* Bits in the Transmit Control Blocks */
#define 	BBIT		0x4000		/* Busy executing command */
#define		DONEBIT		0x8000		/* command completed */
#define		OKBIT		0x2000		/* error-free completion */
#define		ABORTED		0x1000		/* command aborted */
#define		NOCARRIER	0x0400		/* no carrier sense */
#define		XMIT_BAD	0x0200		/* transmit failed */
#define		XMIT_BAD_DMA	0x0100		/* transmit failed (DMA) */
#define		XMIT_DEFER	0x0080		/* transmit failed (traffic) */
#define		XMIT_HEART	0x0040		/* transmit failed (heartbeat) */
#define		XMIT_MAX_COL_16	0x0020		/* more than 16 collisions */
#define		XMIT_NCOL(status) ((((status)&XMIT_MAX_COL_16)>>1)|((status)&0xf))	/* number of colisions */

/* Bits in the System Control Block */
#define		CX		0x8000		/* the CU finished a command */
#define		FR		0x4000		/* the RU got a frame */
#define		CNA		0x2000		/* the CU left the ACT state */
#define		RNR		0x1000		/* the RU left the RDY state */
#define		NO_RESOURCES	0x0020		/* we got an illegal packet */
/* command bits. */
#define		CUCSTART	0x100
#define		RUCSTART	0x10
#define		RUCRESUME	0x20

/* Bits in the Receive Buffer Descriptor (rbd) */
#define ACT_MASK	0x3fff	/* received count */
#define	EOFBIT		0x8000	/* End of a Frame */

/* Bits in the Frame descriptor command */
#define	FDSUSPEND	0x4000	/* Tell receive unit to suspend */
#define	FDEOL		0x8000

 /* States of Chip */
#define		CRESET		0x0001		/* we just reset the chip */
#define		CREADY		0x0002		/* turned on rcvr */
#define		CBROKEN		0x8000		/* something's screwed */

#define		FREE	0
#define		BUSY	1

#define		RBUFFS	15	/* number of rcvr buffers */

#ifndef	NPG
#define		NPG	2	/* max number of pegasus boards */
#endif

#define NPEGS NPG

extern struct tty *notty();

struct	pg_softc {
	struct	arpcom es_ac;		/* common Ethernet structures */
#define es_if	es_ac.ac_if		/* network-visable interfaces */
#define es_addr	es_ac.ac_enaddr	/* hardware ethernet addresses */
} pg_softc[NPEGS];


unsigned short pgstate[NPEGS];

unsigned short tbuffstate[2][NPEGS];
typedef int spl_t;

/*
 * Ethernet software status.
 */
struct	{
	long	if_collisions;
	long	if_ipackets;
	long	if_ierrors;
	long	if_opackets;
	long	if_oerrors;
	long	if_buf0;
	long	if_buf1;
	long	if_nobuf;
} pg_if[NPEGS];


struct iroot {
	unsigned short sysbus;
	unsigned short not_used1;
	unsigned short not_used2;
	unsigned short iscp_low;
	unsigned short iscp_high;
};

struct iscp {
	unsigned short busy;
	unsigned short scb_offset;
	unsigned short scb_base_low;
	unsigned short scb_base_high;
};

struct scb {
	unsigned short status;
	unsigned short command;
	unsigned short cbl_offset;
	unsigned short rfa_offset;
	unsigned short crcerrs;
	unsigned short alnerrs;
	unsigned short rscerrs;
	unsigned short ovrnerrs;
};

struct setup {
	unsigned short status;
	unsigned short command;
	unsigned short link_offset;
	unsigned short addr0;
	unsigned short addr1;
	unsigned short addr2;
};

struct tcb {
	unsigned short status;
	unsigned short command;
	unsigned short link_offset;
	unsigned short tbd_offset;
	unsigned char dest[6];
	unsigned short type;
};

struct romaddr {
	unsigned short bytea;
	unsigned short byteb;
	unsigned short bytec;
	unsigned short byted;
	unsigned short bytee;
	unsigned short bytef;
};

struct fd {
	unsigned short status;
	unsigned short command;
	unsigned short link_offset;
	unsigned short rbd_offset;
	unsigned short dest_addr[3];
	unsigned short source_addr[3];
	unsigned short type;
};

struct rbd {
	unsigned short status;
	unsigned short nxt_rbd_offset;
	unsigned short buff_addr_l;
	unsigned short buff_addr_h;
	unsigned short size;
};

struct tbd {
	unsigned short status;		/* eof & act count */
	unsigned short nxt_tbd_offset;
	unsigned short buff_addr_l;
	unsigned short buff_addr_h;
};

struct buff {
	struct ether_header hdr;
	char data[1518];
};


struct iroot *initptr;
struct iscp *iscpptr;
struct scb *scbptr[NPEGS];
struct setup *setupptr;
struct romaddr *romaddrptr;
struct fd *fdptr[RBUFFS][NPEGS];
struct tcb *tcbptr[2][NPEGS];
struct rbd *rbdptr[RBUFFS][NPEGS];
struct tbd *tbdptr[2][NPEGS];
struct buff *rbufptr[RBUFFS][NPEGS];
struct buff *tbufptr[2][NPEGS];

int	rec_nm_buf[NPEGS];	/* next receive buffer to fill */

extern struct devdata devdata[];
unsigned int rom_addr = 0xc0000,
	ram_addr = 0xd0000;
unsigned int ram_seg;
unsigned short pgcsr[NPEGS];
unsigned short pgalt = 0;

int pgintr();
#define ncprintf printf

int 
pgprobe(), pgattach();

extern int pcif_512_fw;
#define PTOKV(addr) (set_512_window(addr) + pcif_512_fw);

caddr_t         pgstd[] = {(caddr_t) 0x00001550, (caddr_t) 0x00001554,
			   (caddr_t) 0x00001558, (caddr_t) 0x0000155c, 0};

struct iocc_device *pginfo[NPG];

int 
pginit(), pgioctl(), pgoutput(), pgreset(), pgwatch();
int pgsuspend();

struct iocc_driver pgdriver =
{pgprobe, 0, pgattach, 0, pgstd, "un", pginfo,
 /*     int   csr chanrel flags 		suspend */
 0, 0, pgintr, 0,    0,     0,   0 };


#define Offset(off)	((ram_addr + off) & 0x0ffff)

int pgdebug = 0;
int pgnoint = 0;
int nopg = 0;

#define UB_GSFTINT      0x20    /* Generate software interrupt */
#ifdef notdef
#endif


/*
 *  pgprobe - This adapter can only be at interrupt 3
 */
pgprobe(p)
	register caddr_t p;
{
	int x = inb(p+1);
	int y = inb(p+3);
	printf("pgprobe called: port=%x\n", p);
	if (nopg)
		return(PROBE_BAD);
	if (pgnoint)
		return(PROBE_NOINT);	/* didn't attempt hardware int */
	outb(p,UB_GSFTINT);
	PROBE_DELAY(200000);
	outb(p, 0);
	outb(p+1,0xFF);			/* CA */
	return (PROBE_OK);		/* attempted hardware interrupt */
}

int pg_ipc_output(), pg_ipc_ioctl();

/*
 *  pgattach - make the interface available to the network software
 *  (if the auto-configuration software determines that the interface
 *  exists).  The system will initialize the interface when it is
 *  ready to accept packets.
 */
pgattach(iod)
	register struct iocc_device *iod;
{
	int unit = iod->iod_unit;
	register struct pg_softc *us = &pg_softc[unit];
	register struct ifnet *ifp = &us->es_if;
	int	old_window = get_512_window();
	char	*p = (char *) iod->iod_addr;
	int	x = inb(p+1);
	int	y = inb(p+3);
	int	ram = ((x << 8) | ((y & 0x3) << 6)) << 8;
	int	rom;

	pgcsr[unit] = (short) iod->iod_addr;
	p = (char *) set_512_window(ram) + pcif_512_fw;
	rom = p[0x29] << 12;		/* get the rom addr */
	printf("pgattach called: port=%x ca=%x pos5=%x ram=%x rom=%x\n", pgcsr[unit], x, y,
		ram,rom);
	if (rom >= 0xc0000 && rom < 0xdffff)
		rom_addr = rom;		/* use it if it looks valid */
	ram_addr = ram;
	ram_seg = ram_addr >> 16;	/* high order bits */
	ifp->if_unit = unit;
	ifp->if_name = "un";
	ifp->if_mtu = ETHERMTU;

	set_512_window(old_window);
	ifp->if_init = pginit;
	ifp->if_ioctl = pg_ipc_ioctl;
	ifp->if_output = pg_ipc_output;
	ifp->if_reset = pgreset;
	ifp->if_flags = IFF_BROADCAST;
	if_attach(ifp);
	DEBUGF(pgdebug, printf("pg%d: attached\n", unit));
}

pginit(unit)
int unit;
{
	spl_t ospl;
	int i; 
	int ii;
	unsigned iv = 0;	/* subscript of board we are init-ing */
	int	old_window = get_512_window();
	register struct pg_softc *us = &pg_softc[unit];
	register struct ifnet *ifp = &us->es_if;

	if (ifp->if_addrlist == (struct ifaddr *) 0) {
		/* no address */
		return;
	}
	if (ifp->if_flags & IFF_RUNNING)
		return;		/* already running! */

	DEBUGF(pgdebug, printf("pg%d: pginit ram=%x rom=%x\n", unit,ram_addr,rom_addr));
	ifp->if_flags |= IFF_RUNNING;

	/* 							 */
	/* System		D7FDE	STATUS			 */
	/* Control		D7FE0	COMMAND			 */
	/* Block		D7FE2	CBL OFFSET		 */
	/* 			D7FE4	RFA OFFSET		 */
	/* 			D7FE6	CRCERRS			 */
	/* 			D7FE8	ALNERRS			 */
	/* 			D7FEA	RSCERRS			 */
	/* 			D7FEC	OVRNERRS		 */
	/* 							 */
	/* Intermediate		D7FEE	BUSY = 1 (reset)	 */
	/* System		D7FF0	SCB OFFSET = 0x7FDE	 */
	/* Configuration	D7FF2	SCB BASE (lo) = 0x0000	 */
	/* Pointer		D7FF4	SCB BASE (hi) = 0x000D 	 */
	/* 							 */
	/* Initialization	D7FF6	SYSBUS = 0 (16 bit bus)	 */
	/* Root			D7FF8	not used		 */
	/* 			D7FFA	not used		 */
	/* 			D7FFC	ISCP ADDR (lo) = 0x7FEE	 */
	/* 			D7FFE	ISCP ADDR (hi) = 0x000D	 */
	/* 							 */
	set_pcvec_map(PGIRQ, 0);	/* disable int forwarding */
	initptr = (struct iroot *) PTOKV(ram_addr + 0x7FF6);
	initptr->sysbus = swaph(0);		/* 16 bit bus */
	initptr->not_used1 = swaph(0);
	initptr->not_used2 = swaph(0);
	initptr->iscp_low = swaph(Offset(0x7FEE));
	initptr->iscp_high = swaph(ram_seg);
	iscpptr = (struct iscp *) PTOKV(ram_addr + 0x7FEE);
	iscpptr->busy = swaph(0x01);		/* init in progress */
	iscpptr->scb_offset = swaph(Offset(0x7FDE));
	iscpptr->scb_base_low = swaph(Offset(0x0));
	iscpptr->scb_base_high = swaph(ram_seg);
	scbptr[iv] = (struct scb *) PTOKV(ram_addr + 0x7FDE);
	scbptr[iv]->status = 0;
	scbptr[iv]->command = 0;
	scbptr[iv]->cbl_offset = 0;
	scbptr[iv]->rfa_offset = 0;
	scbptr[iv]->crcerrs = 0;
	scbptr[iv]->alnerrs = 0;
	scbptr[iv]->rscerrs = 0;
	scbptr[iv]->ovrnerrs = 0;
	outb(pgcsr[iv],(CHINTENB | RESET));
	DELAY(10);
	outb(pgcsr[iv], CHINTENB);
	outb(pgcsr[iv]+1,0xFF);			/* CA */
	while (!(inb(pgcsr[iv]) & INTPNDG));
	DEBUGF(pgdebug, printf("pg%d: pginit reset done\n", unit));
	pgstate[iv] = CRESET;
	rec_nm_buf[iv] = 0;			/* which rcvr buffer we use */
	romaddrptr = (struct romaddr *) PTOKV(rom_addr + 0x10);
	setupptr = (struct setup *) PTOKV(ram_addr + 0x7FD2);
	setupptr->status = 0;
	setupptr->command = swaph(0x8001);    /* EL + SETUP */
	setupptr->link_offset = 0;
	setupptr->addr0 = ((romaddrptr->byteb) >> 8) | romaddrptr->bytea;
	setupptr->addr1 = ((romaddrptr->byted) >> 8) | romaddrptr->bytec;
	setupptr->addr2 = ((romaddrptr->bytef) >> 8) | romaddrptr->bytee;

#define LAN_ADDR_SIZE ETH_ADDR_SIZE
	bcopy((char *) &setupptr->addr0, pg_softc[iv].es_addr, LAN_ADDR_SIZE);

	printf("pg%d: ethernet address ", iv);
	pgprintethaddr(pg_softc[iv].es_addr);
	printf("\n");

	ospl = splimp();
	while (scbptr[iv]->command != 0);
	scbptr[iv]->command = swaph(CUCSTART);
	scbptr[iv]->cbl_offset = swaph(Offset(0x7FD2));
	outb(pgcsr[iv]+1,0xFF);			/* CA */
	splx(ospl);
	DEBUGF(pgdebug, printf("pg%d: pginit cucstart done\n", unit));
	for (i=0; !(setupptr->status & swaph(OKBIT));++i)
		if (i > 1000) {
			printf("pg%d: timed out waiting for OKBIT\n",iv);
			break;
		} else
			delay(1);
	fdptr[0][iv] = (struct fd *) PTOKV(ram_addr + 0x7FBC);
	for (i = 1; i < RBUFFS; i++)
		fdptr[i][iv] = fdptr[i-1][iv] - 1;
	rbdptr[0][iv] = ((struct rbd *) fdptr[RBUFFS-1][iv]) - 1;
	for (i = 1; i < RBUFFS; i++)
		rbdptr[i][iv] = rbdptr[i-1][iv] - 1;
	tcbptr[0][iv] = ((struct tcb *) rbdptr[RBUFFS-1][iv]) - 1;
	tcbptr[1][iv] = tcbptr[0][iv] - 1;
	tbdptr[0][iv] = ((struct tbd *) tcbptr[1][iv]) - 1;
	tbdptr[1][iv] = tbdptr[0][iv] - 1;
	scbptr[iv]->rfa_offset = swaph((unsigned short) fdptr[0][iv]);
	for (i=0 ; i < RBUFFS; i++)
		fdptr[i][iv]->link_offset = swaph((ushort) fdptr[((i+1)%RBUFFS)][iv]);
	for (i = 0; i < RBUFFS; i++) {
		fdptr[i][iv]->rbd_offset = swaph((ushort) rbdptr[i][iv]);
		fdptr[i][iv]->status = 0;
		fdptr[i][iv]->command = 0;
		fdptr[i][iv]->type = 0;
		for (ii = 0; ii < 3; ii++) {
			fdptr[i][iv]->dest_addr[ii] = 0;
			fdptr[i][iv]->source_addr[ii] = 0;
		}
	}
	fdptr[RBUFFS-1][iv]->command |= swaph(FDSUSPEND);	/* Overrun protection */
	rbufptr[0][iv] = (struct buff *) PTOKV(ram_addr);
	for( i = 1; i < RBUFFS; i++) rbufptr[i][iv] = rbufptr[i-1][iv] + 1;
	for (i = 0; i < RBUFFS; i++) {
		rbdptr[i][iv]->buff_addr_l =swaph((ushort) &(rbufptr[i][iv]->data[0]));
		rbdptr[i][iv]->buff_addr_h = swaph(ram_seg);
		rbdptr[i][iv]->status = 0;
		rbdptr[i][iv]->nxt_rbd_offset = 0;
		rbdptr[i][iv]->size = swaph(0x85EE);	/* 1518 (last RBD) */
	}
	tbufptr[0][iv] = rbufptr[RBUFFS-1][iv] + 1;
	tbufptr[1][iv] = tbufptr[0][iv] + 1;
	tbuffstate[0][iv] = FREE;	/* mark xmit buffers as free */
	tbuffstate[1][iv] = FREE;
	tcbptr[0][iv]->status = 0;
	tcbptr[1][iv]->status = 0;

	DEBUGF(pgdebug, printf("pg%d: pginit calling ostart\n", unit));
	/* Attach this device to the outside world! */

	pg_ipc_ostart(&pg_softc[iv].es_if);

done:
	set_pcvec_map(PGIRQ, 1);	/* enable int forwarding */
	set_512_window(old_window);
	DEBUGF(pgdebug, printf("pg%d: pginit done\n", unit));
}

pgintr(iv)
int iv;		/* board we are going to service */
{
	unsigned short build_comm;
	int	old_window = get_512_window();
	struct scb *scb;
	register struct ifnet *ifp = &pg_softc[iv].es_if;

	kludge_flush();
	if (! (inb(pgcsr[iv]) & INTPNDG)) {
		DEBUGF(pgdebug&SHOW_INTR, printf("pg%d: pgintr - no int\n",iv));
		return(1);		/* No work for us */
	}

	(void) set_512_window(ram_addr);	/* set to proper 512k window */
	DEBUGF(pgdebug&SHOW_INTR, printf("pg%d: pgintr\n",iv));
	scb = scbptr[iv];	/* get scb pointer */
	build_comm = 0;
	if (pgstate[iv] & CRESET) {
		ncprintf("pgintr: CRESET was set (starting receiver)\n");
		ncprintf("pgintr: CRESET board %d\n",iv);
		pgstate[iv] = CREADY;
		while (scb->command != 0)
			;
		scb->command = swaph(CX | CNA | RUCSTART);	/* start rcvr */
		outb(pgcsr[iv]+1, 0xff);		/* CA */
		while (scb->command != 0)
			;
		goto out;
	}

	if (scb->status & swaph(FR)) {     /* rcvd a frame */
		build_comm |= FR;
		pgrint(iv);
	}
	if (scb->status & swaph(CX)) {     /* xmitted a frame */
		build_comm |= CX;
		if ((tbuffstate[0][iv] == BUSY) && (tcbptr[0][iv]->status & swaph(DONEBIT))) {
			int status = swaph(tcbptr[0][iv]->status);
			tbuffstate[0][iv] = FREE;
			ifp->if_collisions += XMIT_NCOL(status);
			if (status&OKBIT)
				ifp->if_opackets++;
			else
				ifp->if_oerrors++;
			tcbptr[0][iv]->status = 0;
		}
		if ((tbuffstate[1][iv] == BUSY) && (tcbptr[1][iv]->status & swaph(DONEBIT))) {
			int status = swaph(tcbptr[1][iv]->status);
			ifp->if_collisions += XMIT_NCOL(status);
			if (status&OKBIT)
				ifp->if_opackets++;
			else
				ifp->if_oerrors++;
			tbuffstate[1][iv] = FREE;
			tcbptr[1][iv]->status = 0;
		}
	}
	if (scb->status & swaph(CNA)) {
		build_comm |= CNA;
	}
	if (scb->status & swaph(NO_RESOURCES)) {
	       build_comm |= (RUCSTART|RNR);
	       scb->rfa_offset=swaph((unsigned short)fdptr[rec_nm_buf[iv]][iv]);
	       ncprintf("pgintr: restart after non-802.3 packet\n");
	}
	if ((scb->status & swaph(RNR)) && 
			(!(scb->status & swaph(NO_RESOURCES)))) {
		/* The receive unit ran out of frames... restart it, since
		 * pgrint must have read in all the packets.
		 */
		ncprintf("pgintr: restarting receiver (ran out of bufs)\n");
		build_comm |= (RUCRESUME|RNR);	/* resume receiving */
	}
	while (scb->command != 0)
		;
	scb->command = swaph(build_comm);
	outb(pgcsr[iv]+1,0xFF);		/* CA */
	if (build_comm & CX)
		pg_ipc_ostart(&pg_softc[iv].es_if);
out:
	DEBUGF(pgdebug&SHOW_INTR, printf("pg%d: pgintr done crc=%d ovrn=%d aln=%d rsc=%d\n",iv,swaph(scb->crcerrs),swaph(scb->ovrnerrs),swaph(scb->alnerrs),swaph(scb->rscerrs)));
	if (scb->crcerrs || scb->alnerrs)
	{	/* got some input errors - count them */
		ifp->if_ierrors += swaph(scb->crcerrs) + swaph(scb->alnerrs);
		scb->crcerrs = 0;
		scb->alnerrs = 0;
	}
	set_512_window(old_window);
	return(0);
}


/*
 * Pegasus receive interrupt.
 */
pgrint(iv)
int iv;		/* which board */
{
	spl_t s;
	int stat;
	int len, len_or_type;
	int prev_nm_buf[NPEGS];
	register struct ether_header *ehp;

	DEBUGF(pgdebug&SHOW_INTR, printf("pg%d: pgrint\n",iv));
	for (; len = swaph(rbdptr[rec_nm_buf[iv]][iv]->status);
				rec_nm_buf[iv]=(rec_nm_buf[iv]+1)%RBUFFS) {
		stat = swaph(fdptr[rec_nm_buf[iv]][iv]->status);
		if (!(stat & DONEBIT)) {
			ncprintf("pgrint: stat %x, not complete\n", stat);
			break;
		}
		/* NEEDSWORK: receiver should be shut down */
		if (! (pg_softc[iv].es_if.if_flags & IFF_UP))
			goto discard;
		if (!(stat & (OKBIT|BBIT))) {
			/* Discard bad (not "ok") frames. */
			ncprintf("pgintr: got not OK packet %x\n", stat);
			goto discard;
		}
		/*
		 * copy in the header to prefix the packet data to
		 * make access simple.
		 */
		bcopy(&(fdptr[rec_nm_buf[iv]][iv]->dest_addr[0]),
			rbufptr[rec_nm_buf[iv]][iv], sizeof(*ehp));
		/* clear flag bits in "len": its now length of data portion */
		len &= ACT_MASK;

		/*
		 * At this point we have the frame and the header
		 * If we want the whole frame we use rbufptr[x]
		 * If we want just the data rbufptr[x]->data[0]
		 * the packet type is in fdptr[x]->type
		 * as well as being in the header
		 * the size of the frame is in rbdptr[x]->status
		 * where x = rec_nm_buf
		 */

		/*
		 * Process the packet
		 */
		s = splimp();
		ehp = (struct ether_header *) rbufptr[rec_nm_buf[iv]][iv];
		DEBUGF(pgdebug&SHOW_DATA, pgprintpacket("pgrint: ",ehp,"\n"));
#define eth_type ether_type
		len_or_type = ntohs(* (u_short *) &ehp->eth_type);
		if (len_or_type > ETHERMTU) {	/* KLUDGE!!! */
			struct mbuf *m = NULL;
			struct mbuf *pgget();

			switch (len_or_type) {
#define LANTYPE_IP ETHERTYPE_IP
			case LANTYPE_IP:
				m = pgget((char *) (ehp + 1), len,
					  &pg_softc[iv].es_if);
				if (m == NULL)
					break;
				if (IF_QFULL(&ipintrq)) {
					IF_DROP(&ipintrq);
					m_freem(m);
					break;
				}
				IF_ENQUEUE(&ipintrq, m);
				schednetisr(NETISR_IP);
				break;
#define LANTYPE_ARP ETHERTYPE_ARP
			case LANTYPE_ARP:
				m = pgget((char *) (ehp + 1), len,
					  &pg_softc[iv].es_if);
				if (m == NULL)
					break;
				arpinput(&pg_softc[iv].es_ac, m);
				break;
			default:
				DEBUGF(pgdebug, printf("pg%d: unknown packet type %x\n",iv,len_or_type));

				break;
			}
		}
		else 
			DEBUGF(pgdebug, printf("pg%d: unknown len or type %x\n",iv,len_or_type));
#ifdef notdef
		{
			struct mbuf *m = NULL;
			struct mbuf *pgget();
#define eth_802_header ether_header
			struct eth_802_header *eh;

			eh = (struct eth_802_header *) ehp;
			len += sizeof(struct ether_header);
			if (len < sizeof(*eh)) {
				printf("pgrint: short packet %d\n", len);
				splx(s);
				goto discard;
			}
#ifdef TEST_CONTROL
			switch (eh->eth_llc.llc_ctrl & ~CTRL_P_OR_F) {
			case TEST_CONTROL:
				m = pgget((char *) eh, len,&pg_softc[iv].es_if);
				if (m == NULL) {
					splx(s);
					goto discard;
				}
				lan_test_recv(&pg_softc[iv].es_if, m);
				goto discard;
			case XID_CONTROL:
				m = pgget((char *) eh, len,&pg_softc[iv].es_if);
				if (m == NULL) {
					splx(s);
					goto discard;
				}
				lan_xid_recv(&pg_softc[iv].es_if, m);
				goto discard;
			case UI_CONTROL:
				break;
			default:
				printf("pgrint: bad LLC control field %d\n",
					eh->eth_llc.llc_ctrl);
				splx(s);
				goto discard;
			}
#endif
			switch (ntohs(* (u_short *) eh->eth_snap.snap_type)) {
#define eth_snap ether_snap
			case LANTYPE_IP:
				m = pgget((char *) (eh + 1), len - sizeof(*eh),
					  &pg_softc[iv].es_if);
				if (m == NULL)
					break;
				if (IF_QFULL(&ipintrq)) {
					IF_DROP(&ipintrq);
					m_freem(m);
					break;
				}
				IF_ENQUEUE(&ipintrq, m);
				schednetisr(NETISR_IP);
				break;
			case LANTYPE_ARP:
				m = pgget((char *) (eh + 1), len - sizeof(*eh),
					  &pg_softc[iv].es_if);
				if (m == NULL)
					break;
				arpinput(&pg_softc[iv].es_if, m);
				break;
			default:
				break;
			}
		}
#endif notdef
		splx(s);
discard:
		/* So this driver knows we've read this frame. */
		rbdptr[rec_nm_buf[iv]][iv]->status = 0;
		/* Now advance the "SUSPEND" indicator from previous buffer to
		 * the receive buffer we just read in.
		 */
		prev_nm_buf[iv] = rec_nm_buf[iv]? rec_nm_buf[iv] - 1: RBUFFS-1;
		fdptr[prev_nm_buf[iv]][iv]->command &= ~swaph(FDSUSPEND);
		fdptr[rec_nm_buf[iv]][iv]->command |= swaph(FDSUSPEND);
	}
}

/*
 * pgxmit -- transmit a buffer -- before calling, you must:
 *	find a free buffer by checking tbuffstate[bufnum] ( 0 = free )
 *	copy your entire eth packet into tbufptr[bufnum]->{hdr,data}.
 */
pgxmit(bufnum, len, iv)
u_int bufnum;
u_int len;	/* Length of entire packet, including header */
u_int iv;	/* which board */
{
	register struct tcb *tcbp;
	register struct tbd *tbdp;
	register struct buff *tbp;

	DEBUGF(pgdebug&SHOW_IO, printf("pg%d: pgxmit bufnum=%d len=%d\n",iv,bufnum,len));
	kludge_flush();
	if (bufnum > 1)
		panic("pgxmit");
	tcbp = tcbptr[bufnum][iv];
	tbp = tbufptr[bufnum][iv];
	tbdp = tbdptr[bufnum][iv];

	/* Copy dest & type to where card expects to find them. */
#define eth_dhost ether_dhost
#define eth_shost ether_shost
	bcopy(tbp->hdr.eth_dhost, tcbp->dest, ETH_ADDR_SIZE);
/*	bcopy(&tbp->hdr.eth_type, tcbp->type, 2);	*/
	tcbp->type = (tbp->hdr.eth_type);	/* length ?? */
	DEBUGF(pgdebug&SHOW_IO, pgprintpacket("pgxmit: ",&tbp->hdr,"\n"));

	tcbp->command = swaph(0xA004);	/* EL & I & XMT */
	tcbp->link_offset = 0;
	tcbp->tbd_offset = swaph((unsigned short) tbdp);
/*	tbdp->status = swaph((len - sizeof(struct ether_header)) | EOFBIT); */
	tbdp->status = swaph((len) | EOFBIT);
	tbdp->nxt_tbd_offset = 0;
	tbdp->buff_addr_l = swaph((unsigned short) tbp->data);
	tbdp->buff_addr_h = swaph(ram_seg);
	while (scbptr[iv]->command != 0);
	scbptr[iv]->command = swaph(CUCSTART);
	scbptr[iv]->cbl_offset = swaph((unsigned short) tcbp);
	outb(pgcsr[iv]+1,0xFF);	/* CA */
}

struct mbuf *
pgget(addr, totlen, ifp)
register u_char *addr;
register int totlen;
struct ifnet *ifp;
{
	register int len;
	register struct mbuf *m;
	struct mbuf *top = NULL, **mp = &top;
	u_char *mcp;

	++ifp->if_ipackets;

	DEBUGF(pgdebug&SHOW_IO, {
		int i;
		printf("pg%d: pgget addr=%x",ifp->if_unit,addr);
		if(pgdebug&SHOW_DATA)
			for (i=0; i<16; ++i)
				printf(" %x",addr[i]);
		printf("\n");});

	while (totlen > 0) {
		MGET(m, M_DONTWAIT, MT_DATA);
		if (m == NULL)
			goto bad;
		len = totlen;
		if (ifp != NULL)
			len += sizeof(ifp);
#ifdef notdef
		/* Should we try to use a cluster for efficiency? */
		if (len >= mincluster) {
			MCLGET(m);
			if (m->m_len == CLBYTES)
				m->m_len = len = MIN(CLBYTES, len);
			else
				m->m_len = len = MIN(MLEN, len);
		}
		else
#endif
		{
			m->m_len = len = MIN(MLEN, len);
			m->m_off = MMINOFF;
		}
		mcp = mtod(m, u_char *);
		if (ifp != NULL) {
			* (mtod(m, struct ifnet **)) = ifp;
			mcp += sizeof(ifp);
			len -= sizeof(ifp);
			ifp = NULL;
		}
		bcopy(addr, mcp, len);
		addr += len;
		*mp = m;
		mp = &m->m_next;
		totlen -= len;
	}
	return top;
bad:
	m_freem(top);
	DEBUGF(pgdebug&SHOW_IO, printf("pg%d: pgget BAD\n",ifp->if_unit));
	return NULL;
}

pg_ipc_output(ifp, m0, dst)
struct ifnet *ifp;
struct mbuf *m0;
struct sockaddr *dst;
{
	int error;
	struct mbuf *m = m0, *mcopy = NULL;
	register struct ether_header *ec;
	struct lan_arp *ah;
	char edst[LAN_ADDR_SIZE];
	int usetrailers;	/* N.B. this is currently ignored */
	struct in_addr idst;
	short type;
	spl_t s;
	extern struct ifnet loif;
	int	old_window = get_512_window();
	int iv;

	(void) set_512_window(ram_addr);	/* set to proper 512k window */
	iv = ifp->if_unit;
	DEBUGF(pgdebug&SHOW_IO, printf("pg%d: pgoutput\n",iv));

	if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)) {
		error = ENETDOWN;
		goto bad;
	}

	switch (dst->sa_family) {
#ifdef INET
	case AF_INET:
		idst = ((struct sockaddr_in *) dst)->sin_addr;
		if (! arpresolve(&pg_softc[iv].es_ac,m,&idst,edst,&usetrailers)) {
			set_512_window(old_window);
			return 0;		/* not resolved */
		}
		if (! bcmp(edst, etherbroadcastaddr, ETH_ADDR_SIZE))
			mcopy = m_copy(m, 0, M_COPYALL);
		type = LANTYPE_IP;
		break;
#endif INET
#ifdef AF_ARP
	case AF_ARP:
		ah = mtod(m, struct lan_arp *);
		bcopy(ah->arp_tha, edst, LAN_ADDR_SIZE);
		type = LANTYPE_ARP;
		break;
#endif
	case AF_UNSPEC:
		ec = (struct ether_header *) dst->sa_data;
		bcopy((caddr_t) ec->eth_dhost, (caddr_t) edst, sizeof(edst));
#ifdef notdef
		type = NO_SNAP;
#else
		type = ec->ether_type;
#endif
		break;
	default:
		ncprintf("pg_ipc_output: cannot handle address family 0x%x\n",
			dst->sa_family);
		error = EAFNOSUPPORT;
		goto bad;
	}
#ifdef IFF_IEEE
	if (!(ifp->if_flags & IFF_IEEE)) /* Xerox ethernet header */
#endif
	{
		if ((m->m_off > MMAXOFF)
		||  ((MMINOFF + sizeof(struct ether_header)) > m->m_off)) {
			m = m_get(M_DONTWAIT, MT_HEADER);
			if (m == NULL) {
				error = ENOBUFS;
				goto bad;
			}
			m->m_next = m0;
			m->m_off = MMINOFF;
			m->m_len = sizeof(struct ether_header);
		}
		else {
			m->m_off -= sizeof(struct ether_header);
			m->m_len += sizeof(struct ether_header);
		}
		ec = mtod(m, struct ether_header *);
		bcopy(edst, (caddr_t) ec->eth_dhost, LAN_ADDR_SIZE);
		bcopy(pg_softc[iv].es_addr, (caddr_t) ec->eth_shost, LAN_ADDR_SIZE);
		ec->eth_type = htons(type);
	}
#ifdef IFF_IEEE
	else
	{		/* IEEE802.3 header */
		register struct eth_mac_header *mh;
		register int header_size;

		header_size = sizeof(struct lan_llc_header)
				+ sizeof(struct lan_snap_header);

		switch (type) {
		case LANTYPE_IP:
		case LANTYPE_ARP:
			if ((m->m_off > MMAXOFF)
			||  ((MMINOFF + header_size) > m->m_off)) {
				m = m_get(M_DONTWAIT, MT_HEADER);
				if (m == NULL) {
					error = ENOBUFS;
					goto bad;
				}
				m->m_next = m0;
				m->m_off = MMINOFF;
				m->m_len = header_size;
				m0 = m;
			}
			else {
				m->m_off -= header_size;
				m->m_len += header_size;
			}
			fill_llc_header(mtod(m, struct lan_llc_header *), type);
			/* FALLSTHROUGH */
		case NO_SNAP:
			if ((m->m_off > MMAXOFF)
			||  ((MMINOFF + sizeof(*mh)) > m->m_off)) {
				m = m_get(M_DONTWAIT, MT_HEADER);
				if (m == NULL) {
					error = ENOBUFS;
					goto bad;
				}
				m->m_next = m0;
				m->m_off = MMINOFF;
				m->m_len = sizeof(*mh);
			}
			else {
				m->m_off -= sizeof(*mh);
				m->m_len += sizeof(*mh);
			}
			/* Fill in ethernet MAC header */
			mh = mtod(m, struct eth_mac_header *);
			bcopy(edst, (caddr_t) mh->eth_dhost, LAN_ADDR_SIZE);
			bcopy(pg_softc[iv].es_addr, (caddr_t) mh->eth_shost,
				LAN_ADDR_SIZE);
			/* mh->eth_len = XXX; */	/* Set later */
		}
	}
#endif

	s = splimp();
	if (IF_QFULL(&ifp->if_snd)) {
		IF_DROP(&ifp->if_snd);
		error = ENOBUFS;
		m0 = m;
		splx(s);
		goto bad;
	}
	IF_ENQUEUE(&ifp->if_snd, m);
	pg_ipc_ostart(ifp);
	splx(s);
	set_512_window(old_window);
	return (mcopy ? looutput(&loif, mcopy, dst) : NULL);
bad:
	m_freem(m0);
	if (mcopy)
		m_freem(mcopy);
	set_512_window(old_window);
	return error;
}

pg_ipc_ostart(ifp)
register struct ifnet *ifp;
{
	int bufnum;
	struct mbuf *m, *m0;
	register char *bufp;
	register unsigned short length;
	spl_t s;
	int iv;

	s = splimp();

	iv = ifp->if_unit;
	DEBUGF(pgdebug&SHOW_IO, printf("pg%d: pgostart\n",iv));

#define IF_EMPTYQUEUE(queue) ((queue)->ifq_head == 0)
	while (! IF_EMPTYQUEUE(&ifp->if_snd)) {
		/* find a free buffer to put a packet in */
		if (tbuffstate[0][iv] == FREE) {
			tbuffstate[0][iv] = BUSY;
			bufnum = 0;
			++pg_if[iv].if_buf0;
		}
		else if (tbuffstate[1][iv] == FREE) {
			tbuffstate[1][iv] = BUSY;
			bufnum = 1;
			++pg_if[iv].if_buf1;
		}
		else {
			++pg_if[iv].if_nobuf;
			break;
		}
		IF_DEQUEUE(&ifp->if_snd, m);

		bufp = (char *) tbufptr[bufnum][iv];
		length = 0;
		for (m0 = m; m0 != NULL; m0 = m0->m_next) {
			bcopy(mtod(m0, char *), bufp, m0->m_len);
			bufp += m0->m_len;
			length += m0->m_len;
		}
#ifdef IFF_IEEE
		/* For IEEE802.3 we need to fill in the length */
		if (ifp->if_flags & IFF_IEEE) {
			bufp = (char *) tbufptr[bufnum][iv];
			((struct eth_mac_header *) bufp)->eth_len =
				htons(length - sizeof(struct eth_mac_header));
		}
#endif IFF_IEEE
		length = MAX(length, ETHERMIN);
		m_freem(m);

		pgxmit(bufnum, length, iv); 
	}
	splx(s);
	DEBUGF(pgdebug&SHOW_IO, printf("pg%d: pgostart done\n",iv));
	return;
}

pg_ipc_ioctl(ifp, cmd, data)
struct ifnet *ifp;
int cmd;
caddr_t data;
{
	register struct ifaddr *ifa = (struct ifaddr *) data;
	int error = 0;
	spl_t s;

	DEBUGF(pgdebug&SHOW_IO, printf("pg%d: pgioctl cmd=%x\n",ifp->if_unit,cmd));
	s = splimp();
	switch (cmd) {
	case SIOCSIFADDR:
		ifp->if_flags |= IFF_UP;

		switch (ifa->ifa_addr.sa_family) {
#ifdef INET
		case AF_INET:
			pginit(ifp->if_unit);	/* before arpwhohas */
			((struct arpcom *)ifp)->ac_ipaddr=IA_SIN(ifa)->sin_addr;
			arpwhohas((struct arpcom *)ifp, &IA_SIN(ifa)->sin_addr);
			break;
#endif INET
		default:
			pginit(ifp->if_unit);
			break;

		}
		break;
        case SIOCSIFFLAGS:
                if ((ifp->if_flags & IFF_UP) == 0 && ifp->if_flags &
                    IFF_RUNNING) {
                        pgzap(ifp->if_unit);
                        ifp->if_flags &= ~IFF_RUNNING;
                } else if (ifp->if_flags & IFF_UP && (ifp->if_flags &
                                                      IFF_RUNNING) == 0)
                        pginit(ifp->if_unit);
                break;
	default:
		error = EINVAL;
		break;
	}
	splx(s);
	DEBUGF(pgdebug&SHOW_IO, printf("pg%d: pgioctl done\n",ifp->if_unit));
	return error;
}

pgreset(unit)
        register unsigned int unit;
{
        register struct iocc_device *iod;
	if (unit < NPG && (iod = pginfo[unit]) != 0 && iod->iod_alive != 0) {
                if (pg_softc[unit].es_if.if_flags & IFF_RUNNING) {
                        pg_softc[unit].es_if.if_flags &= ~IFF_RUNNING;
                        DEBUGF(pgdebug, printf("pg%d: reset\n", unit));
                        pgzap(unit);
                }

	}
}


static pgzap(unit)
        register unsigned int unit;
{
	DEBUGF(pgdebug, printf("pg%d: pgzap\n", unit));

	outb(pgcsr[unit],(RESET));
	DELAY(10);
	outb(pgcsr[unit], 0);
	outb(pgcsr[unit]+1,0xFF);			/* CA */
	delay(1);
	DEBUGF(pgdebug, printf("pg%d: pgzap done\n", unit));
}

/*
 *  pgprintethaddr - print an ethernet address
 */
static pgprintethaddr(p)
	register char  *p;
{
	register int    i;

	for (i = 0; i < ETH_ADDR_SIZE; i++) {
		if (i != 0)
			printf(":");
		printf("%x", *p++);
	}
}


#ifdef DEBUG
static char *ethertype(type)
{
	if (type == ETHERTYPE_IP)
		return("ip");
	if (type == ETHERTYPE_ARP)
		return("arp");
	return("?");
}

static pgprintpacket(prefix,eh,postfix)
	register struct ether_header *eh;
	char *prefix, *postfix;
{
	char            cbuf[6];
	printf("%sfrom = ",prefix);
	bcopy(eh->ether_shost, cbuf, sizeof(cbuf));
	pgprintethaddr(cbuf);
	printf("  to = ");
	bcopy(eh->ether_dhost, cbuf, sizeof(cbuf));
	pgprintethaddr(cbuf);
	printf(" type=%s(%x) %s",ethertype(eh->ether_type),eh->ether_type,postfix);
}
#endif
