/*
 * if_adxvar.h - Definitions for the ADAX PC-SDMA driver.
 */

/*
 * The  board and register interface.
 */
struct adx_device {
    volatile u_char chA_data;	/* channel A Data Register */
    volatile u_char chB_data;	/* channel B Data Register */
    volatile u_char chA_csr;	/* Channel A Command/Status register */
    volatile u_char chB_csr;	/* Channel B Command/Status register */
    volatile u_char ctr_0;	/* COUNTER/TIMER Counter 0 */
    volatile u_char ctr_1;	/* COUNTER/TIMER Counter 1 */
    volatile u_char ctr_2;	/* COUNTER/TIMER Counter 2 */
    volatile u_char ctr_csr;	/* COUNTER/TIMER Command/Status reg */
};


/*
 * Software status associated with each device.
 */
struct adx_softc {
    struct ifnet *adx_if;	/* Interface pointer */
    int adx_oactive;		/* Output in progress */
    int adx_flags;		/* Flags */
    int adx_tx_dma_chan;	/* Tx dma channel to use */
    int adx_rx_dma_chan;	/* Rx dma channel to use */
    char *adx_tx_buffer;	/* Tx buffer */
    char *adx_rx_buffer0;	/* Rx buffer #0 */
    char *adx_rx_buffer1;	/* Rx buffer #1 */
    int adx_which_rx_buffer;	/* Which is current Rx buffer */
    int adx_dcd_tmo;		/* Timeout for DCD to assert */
    int adx_dtr_tmo;		/* DTR toggle period */
    int adx_tx_retries;		/* Number of retries on transmit */
    int adx_tx_len;		/* Length of current transmit frame */
};

#define ADX_STARTED	0x01	/* Device started */
#define ADX_DCD		0x02	/* DCD active from DCE */
#define ADX_SYNC	0x04	/* Receiver SYNCronized */
#define ADX_CTS		0x08	/* CTS active from DCE */
#define ADX_ABORT	0x10	/* DCE aborting */
#define ADX_DCDWARN	0x20	/* Warned about DCD */
#define ADX_INTERNALCLK	0x40	/* Use internal 8254 clock */

#define ADX_MTU		1506	/* Adax MTU including header and trailer */

#define ADX_SLOP	20	/* Amount of extra buffer for dma operations */

#define ADX_DFLT_DCD_TMO 30000	/* Timeout for DCD to assert */
#define ADX_DFLT_DTR_TMO 500	/* DTR toggle period */

#define ADX_MAX_TX_RETRIES 5	/* Max number of transmit retries */

/*
 * Convert number of milliseconds to hz.  Hz is the number of clock ticks
 * desired before a timeout is called.
 * Shifting by 10 is the same as dividing by 1024 which is "close" to
 * dividing by 1000.
 */
#define mstohz(ms)	(((ms) * hz) >> 10)
