
#ifndef VCAIO_H
#define VCAIO_H

#define	VCAMajor 25			/* Corresponds to major number of all VCAs. */

/* minor and major macros defined in types.h */
#define VcaCard(dev) ((minor((dev))>>1)&0x03)
#define VcaChannel(dev) (minor((dev))&0x01)
#define SpecialMark(dev) ((minor((dev))>>7)&0x01)
#define SpecialMux(dev) ((minor((dev))>>3)&0x03)
#define SpecialRec(dev) ((minor((dev))>>5)&0x01)

#define	DebugRead   1<<0
#define	DebugWrite  1<<1
#define	DebugIoctl  1<<2
#define	DebugIntr   1<<3
#define	DebugLoad   1<<4
#define	DebugSigs   1<<5
#define	DebugAudio  1<<6
#define	DebugIO	    DebugRead | DebugWrite
#define	DebugAny    DebugRead | DebugWrite | DebugIoctl | DebugIntr | DebugSigs | DebugAudio
#define	DebugAll    DebugAny

#if 0	/* Select what works best */
#define	vcabufsize 0x2000   /* 16k bytes, 8k unsigned shorts (1 second of audio). */
#define	vcabufsize 0x1000   /* 8k bytes, 4k unsigned shorts (1/2 second of audio). */
#define	vcabufsize 0x800   /* 4k bytes, 2k unsigned shorts (1/4 second of audio). */
#define	vcabufsize 0x400   /* 2k bytes, 1k unsigned shorts (1/8 second of audio). */
#define	vcabufsize 0x200   /* 1k bytes, 0.5k unsigned shorts (1/16 second of audio). */
#endif
#define	vcabufsize 0x2000   /* 16k bytes, 8k unsigned shorts (1 second of audio). */
#define vcabytebufsize vcabufsize*2

#ifdef KERNEL

#define splvca() splhigh()

#ifdef VCADebug
#include <stdio.h>
char debug[NVCA*2];
long DebugCount[NVCA*2];
#define	DEBUGprintf(d,m,x)	{if ((debug[VcaCard(d)+VcaChannel(d)]&(m))!=0) printf x;}
#else
#define DEBUGprintf(d,m,x)
#endif /* VCADebug */

struct ADCArgs {
    short start, end, size, overrun;
};

struct DACArgs {
    short start, end, ReadyToWrite;
};

struct VCA_Arguments {
    struct ADCArgs ADC1;
    struct ADCArgs ADC2;
    struct DACArgs DAC1;
    struct DACArgs DAC2;
};

#endif /* KERNEL */

typedef unsigned short vcalocation;
typedef unsigned short psp_program;

typedef struct { char type; vcalocation loc; vcalocation data; } loadpsp;

#define	Phone	0
#define	Tele	1
#define Mic	2
#define	ACWrap	3

#define ADC_CH2 1
#define ADC_CH1 2
#define DAC_CH2 4
#define DAC_CH1 8

/* IO Control register bit definitions: */
#define	VCA_DISABLE_ALL			0x00
#define	VCA_ENA_PSP			0x01
#define VCA_SELECT_INST_MEM_HIGH	0x02
#define VCA_SELECT_INST_MEM		0x04
#define VCA_INT_PSP			0x08
					/* Bits 4-5 reserved. */
#define VCA_ENA_INT			0x40
#define VCA_INT_STATUS			0x80

typedef struct {
    unsigned int INT_STATUS:		1;
    unsigned int ENA_INT:		1;
    unsigned int reserved:		2;
    unsigned int INT_PSP:		1;
    unsigned int SEL_INST_MEM:		1;
    unsigned int SEL_INST_MEM_HIGH:	1;
    unsigned int ENA_PSP:		1;
} VCA_IOCR;

typedef struct {
    unsigned int CH2_MUX:		2;
    unsigned int CH1_MUX:		2;
    unsigned int DCWRAP:		1;
    unsigned int TELE2_OFFHOOK:		1;
    unsigned int TELE1_OFFHOOK:		1;
    unsigned int BYPASS:		1;
    unsigned int CH2_GAIN:		2;
    unsigned int CH1_GAIN:		2;
    unsigned int CH1_DAC_ENABLE:	1;
    unsigned int CH2_DAC_ENABLE:	1;
    unsigned int CH1_ADC_ENABLE:	1;
    unsigned int CH2_ADC_ENABLE:	1;
} VCA_SC2;

#define	VCA_SC2_BYPASS		0x0100
#define	VCA_SC2_TELE1_OFFHOOK	0x0200
#define	VCA_SC2_TELE2_OFFHOOK	0x0400
#define	VCA_SC2_DCWRAP		0x0800

#define	VCA_WRITE_IN_PROGRESS_P	    0
#define	VCA_SET_RECORDING	    1
#define	VCA_SET_MULTIPLEXER	    2
#define	VCA_SET_OFF_HOOK	    3
#define	VCA_SET_RECEIVE_GAIN	    4
#define	VCA_SET_PLAYING		    5
#define	VCA_DEBUGING_ON		    6
#define	VCA_DEBUGING_OFF	    7
#define	VCA_WAKEUP		    8
#define	VCA_INTERRUPT_PSP	    9
#define	VCA_ENABLE_PSP_INTERRUPTS   10
#define	VCA_DISABLE_PSP_INTERRUPTS  11
#define	VCA_ENABLE_PSP		    12
#define	VCA_DISABLE_PSP		    13
#define	VCA_LOAD		    14
#define	VCA_SET_RW_PID		    15
#define	VCA_WRITE_PROGRAM_ID	    16
#define	VCA_READ_PROGRAM_ID	    17
#define	VCA_SET_BYPASS		    18
#define	VCA_SET_DCWRAP		    19
#define	VCA_SET_SAMPLE_RATE	    20
#define	VCA_GET_SC2		    21
#define	VCA_BLOCKING_READS	    22
#define	VCA_CROSS_CONNECT	    23
#define	VCA_RESET_CHANNEL	    24
#define	VCA_SET_RI_PID		    25
#define	VCA_DISCONNECT_CHANNEL	    26
#define	VCA_GENERATE_HOST_INTERRUPT 27
#define	VCA_CRADLE_ON_HOOK_P	    28
#define	VCA_SET_VCA2LAN		    29
#define	VCA_LAN_DEBUG_ON	    30
#define	VCA_LAN_DEBUG_OFF	    31
#define	VCA_DUMP_VCA2LAN_TIMESTAMPS 32
#define	VCA_RETURN_DEVICE_NUMBER    33
#define	VCA_SET_PACKET_PRIORITY	    34
#define	VCA_RESIZE_RECORD_BUFFERS   35
#define	VCA_DROP_PACKETS_FROM_LAN   36
#define	VCA_SET_VCA2LAN_PACKET_TOTAL_SIZE    37
#define	VCA_SEND_TOKEN_RING_PACKET  38
#define	VCA_SET_QUEUE_PRIORITY	    39

#define	VCA_PROGRAM_ID_OFFSET	    0x00
#define	VCA_PROGRAM_ID_SIZE	    0x10

#define	VCA_SHADOW_OF_SC2_OFFSET    0x10
#define	VCA_NUMBER_OF_INTERRUPTS    0x12
#define	VCA_SHADOW_OF_SS1_OFFSET    0x1f

#define	VCA_FUNCTION_REQUEST_DEBUG_OFFSET 0x11

#define VCA_DAC1_ENALBED_OFFSET 0x2f
#define VCA_DAC2_ENALBED_OFFSET 0x3f

#define	VCA_INTERRUPT_ARGUMENTS_OFFSET 0x50

#define	VCA_PH1_RI_OFFSET	    0x5e
#define	VCA_PH2_RI_OFFSET	    0x5f

#define	VCA_SET_RECORDING_OFFSET    0x60
#define	VCA_SET_PLAYING_OFFSET	    0x62
#define	VCA_SET_MULTIPLEXER_OFFSET  0x64
#define	VCA_SET_OFF_HOOK_OFFSET	    0x66
#define	VCA_SET_RECEIVE_GAIN_OFFSET 0x68

#define	VCA_SET_BYPASS_OFFSET	    0x6a
#define	VCA_SET_DCWRAP_OFFSET	    0x6b
#define	VCA_SET_SAMPLE_RATE_OFFSET  0x6c
#define	VCA_GENERATE_HOST_INTERRUPT_OFFSET  0x6d
#define	VCA_REINITIALIZE_BUFFERS_OFFSET	    0x6e

#define VCA_MIN_VALUE_DROP_PACKETS_FROM_LAN 0
#define	VCA_DO_NOT_DROP		0
#define	VCA_DROP_WITH_COPY	1
#define	VCA_DROP_WITHOUT_COPY	2
#define	VCA_SEND_USING_IOMEM	3
#define VCA_MAX_VALUE_DROP_PACKETS_FROM_LAN 3

#endif /* VCAIO_H */
