/*	@(#)if_tl.c	7.1 (Berkeley) 10/10/86 */

/*
 * Terminal Line Interface for Fuzzball (see RFC 891)
 *
 * October 10, 1986
 *
 * Robert B. Hoffman, hoffman@pitt
 * Computer Science Department, University of Pittsburgh
 * Sean McLinden, sean@cadre.dsl.pittsburgh.edu
 * Decision Systems Laboratory, University of Pittsburgh
 *
 * Based in part on the SL handler from Rick Adams (rick@seismo),
 * Chris Torek (chris@mimsy.umd.edu), and Kirk Smith, and on the
 * 4.2bsd TL handler from Russell Yount (ry@cadre).
 *
 * N.B.: this belongs in netinet, not net, the way it stands now.
 * Should have a link-layer type designation, but wouldn't be
 * backwards-compatible.
 *
 */

/* $Header: if_tl.c,v 1.12 85/12/20 21:54:55 chris Exp $ */
/* from if_tl.c,v 1.11 84/10/04 12:54:47 rick Exp */

#include "tl.h"
#if NTL > 0

#include "param.h"
#include "mbuf.h"
#include "buf.h"
#include "dk.h"
#include "socket.h"
#include "ioctl.h"
#include "file.h"
#include "tty.h"
#include "errno.h"

#include "if.h"
#include "netisr.h"
#include "route.h"
#if INET
#include "../netinet/in.h"
#include "../netinet/in_systm.h"
#include "../netinet/in_var.h"
#include "../netinet/ip.h"
#endif

#ifdef vax
#include "../vax/mtpr.h"
#endif vax

/*
 * N.B.: TLMTU is now a hard limit on input packet size.
 * TLMTU must be <= CLBYTES - sizeof(struct ifnet *).
 */
#define	TLMTU	1006		/* max size of one transmission */
#define	TLIP_HIWAT	1000	/* don't start a new packet if HIWAT on queue */
#define	CLISTRESERVE	1000	/* Can't let clists get too low */

struct tl_softc {
	struct	ifnet tc_if;	/* network-visible interface */
	short	tc_istate;	/* current state of receiver */
	short	tc_flags;	/* see below */
	short	tc_ilen;	/* length of input-packet-so-far */
	struct	tty *tc_ttyp;	/* pointer to tty structure */
	char	*tc_mp;		/* pointer to next available buf char */
	char	*tc_buf;	/* input buffer */
} tl_softc[NTL];

/* states */
#define TC_RESYNC_IDLE	0	/* out of sync */
#define TC_NORMAL_DATA	1	/* receiving data */
#define TC_NORMAL_IDLE	2	/* idling */
#define TC_MARKED_DATA	3	/* mark in data */
#define TC_MARKED_IDLE	4	/* mark while idling */

/* flags */
#define	TC_OACTIVE	0x0001	/* output tty is active */

/* characters */
#define ASCII_DLE	0020
#define ASCII_STX	0002
#define ASCII_ETX	0003
#define ASCII_DEL	0177
#define ASCII_CAN	0030

#define t_tc T_LINEP

int tloutput(), tlioctl(), ttrstrt();

/*
 * Called from boot code to establish tl interfaces.
 */
tlattach()
{
	register struct tl_softc *tc;
	register int i = 0;

	for (tc = tl_softc; i < NTL; tc++) {
		tc->tc_if.if_name = "tl";
		tc->tc_if.if_unit = i++;
		tc->tc_if.if_mtu = TLMTU;
		tc->tc_if.if_flags = IFF_POINTOPOINT;
		tc->tc_if.if_ioctl = tlioctl;
		tc->tc_if.if_output = tloutput;
		tc->tc_if.if_snd.ifq_maxlen = IFQ_MAXLEN;
		if_attach(&tc->tc_if);
	}
}

/*
 * Line specific open routine.
 * Attach the given tty to the first available tl unit.
 */
/* ARGSUSED */
tlopen(dev, tp)
	dev_t dev;
	register struct tty *tp;
{
	register struct tl_softc *tc;
	register int ntl;

	if (!suser())
		return (EPERM);
	if (tp->t_line == TLIPDISC)
		return (EBUSY);

	for (ntl = 0, tc = tl_softc; ntl < NTL; ntl++, tc++)
		if (tc->tc_ttyp == NULL) {
			tc->tc_flags = 0;
			tc->tc_ilen = 0;
			if (tlinit(tc) == 0)
				return (ENOBUFS);
			tp->t_tc = (caddr_t)tc;
			tc->tc_ttyp = tp;
			ttyflush(tp, FREAD | FWRITE);
			tc->tc_istate = TC_NORMAL_IDLE;
			return (0);
		}

	return (ENXIO);
}

/*
 * Line specific close routine.
 * Detach the tty from the tl unit.
 * Mimics part of ttyclose().
 */
tlclose(tp)
	struct tty *tp;
{
	register struct tl_softc *tc;
	int s;

	ttywflush(tp);
	tp->t_line = 0;
	s = splimp();		/* paranoid; splnet probably ok */
	tc = (struct tl_softc *)tp->t_tc;
	if (tc != NULL) {
		if_down(&tc->tc_if);
		tc->tc_ttyp = NULL;
		tp->t_tc = NULL;
		MCLFREE((struct mbuf *)tc->tc_buf);
		tc->tc_buf = 0;
	}
	splx(s);
}

/*
 * Line specific (tty) ioctl routine.
 * Provide a way to get the tl unit number.
 */
/* ARGSUSED */
tltioctl(tp, cmd, data, flag)
	struct tty *tp;
	caddr_t data;
{

	if (cmd == TIOCGETD) {
		*(int *)data = ((struct tl_softc *)tp->t_tc)->tc_if.if_unit;
		return (0);
	}
	return (-1);
}

/*
 * Queue a packet.  Start transmission if not active.
 */
tloutput(ifp, m, dst)
	register struct ifnet *ifp;
	register struct mbuf *m;
	struct sockaddr *dst;
{
	register struct tl_softc *tc;
	int s;

	/*
	 * `Cannot happen' (see tlioctl).  Someday we will extend
	 * the line protocol to support other address families.
	 */
	if (dst->sa_family != AF_INET) {
		printf("tl%d: af%d not supported\n", ifp->if_unit,
			dst->sa_family);
		m_freem(m);
		return (EAFNOSUPPORT);
	}

	tc = &tl_softc[ifp->if_unit];
	if (tc->tc_ttyp == NULL) {
		m_freem(m);
		return (ENETDOWN);	/* sort of */
	}
	if ((tc->tc_ttyp->t_state & TS_CARR_ON) == 0) {
		m_freem(m);
		return (EHOSTUNREACH);
	}
	s = splimp();
	if (IF_QFULL(&ifp->if_snd)) {
		IF_DROP(&ifp->if_snd);
		splx(s);
		m_freem(m);
		tc->tc_if.if_oerrors++;
		return (ENOBUFS);
	}
	IF_ENQUEUE(&ifp->if_snd, m);
	if ((tc->tc_flags & TC_OACTIVE) == 0) {
		splx(s);
		tlstart(tc->tc_ttyp);
	} else
		splx(s);
	return (0);
}

/*
 * Start output on interface.  Get another datagram
 * to send from the interface queue and map it to
 * the interface before starting output.
 */
tlstart(tp)
	register struct tty *tp;
{
	register struct tl_softc *tc = (struct tl_softc *)tp->t_tc;
	register struct mbuf *m;
	register int len;
	register u_char *cp;
	int n, s;
	struct mbuf *m2;
	extern int cfreecount;

	for (;;) {
		/*
		 * If there is more in the output queue, just send it now.
		 * We are being called in lieu of ttstart and must do what
		 * it would.
		 */
		if (tp->t_outq.c_cc > 0)
			ttstart(tp);
		if (tp->t_outq.c_cc > TLIP_HIWAT)
			return;

		/*
		 * This happens briefly when the line shuts down.
		 */
		if (tc == NULL)
			return;

		/*
		 * If system is getting low on clists
		 * and we have something running already, stop here.
		 */
		if (cfreecount < CLISTRESERVE + TLMTU &&
		    tc->tc_flags & TC_OACTIVE)
			return;

		/*
		 * Get a packet and send it to the interface.
		 */
		s = splimp();
		IF_DEQUEUE(&tc->tc_if.if_snd, m);
		if (m == NULL) {
			if (tp->t_outq.c_cc == 0)
				tc->tc_flags &= ~TC_OACTIVE;
			splx(s);
			return;
		}
		tc->tc_flags |= TC_OACTIVE;
		splx(s);

		/*
		 * The DLE-STX will start up a new packet, and thus
		 * will flush any accumulated garbage.
		 */
		if ( putc(ASCII_DLE, &tp->t_outq) || putc(ASCII_STX, &tp->t_outq))
			goto full;

		while (m) {
			cp = mtod(m, u_char *);
			len = m->m_len;
			while (len > 0) {
				/*
				 * Find out how many bytes in the string we can
				 * handle without doing something special.
				 */
				n = len - locc(ASCII_DLE, len, cp);
				if (n) {
					/*
					 * Put n characters at once
					 * into the tty output queue.
					 */
					if (b_to_q((char *)cp, n, &tp->t_outq))
						break;
					len -= n;
					cp += n;
				}
				/*
				 * If there are characters left in the mbuf,
				 * the first one must be special..
				 * Put it out in a different form.
				 */
				if (len) {
					if (putc(ASCII_DLE, &tp->t_outq))
						break;
					if (putc(*cp, &tp->t_outq)) {
						(void) unputc(&tp->t_outq);
						break;
					}
					cp++;
					len--;
				}
			}
			MFREE(m, m2);
			m = m2;
		}

		if (putc(ASCII_DLE, &tp->t_outq) || putc(ASCII_ETX, &tp->t_outq)) {
full:
			/*
			 * Not enough room.  Remove a char to make room
			 * and end the packet normally.
			 * If you get many collisions (more than one or two
			 * a day) you probably do not have enough clists
			 * and you should increase "nclist" in param.c.
			 */
			(void) unputc(&tp->t_outq);	/* make some space */
			(void) unputc(&tp->t_outq);
			(void) putc(ASCII_DLE, &tp->t_outq);	/* end the packet */
			(void) putc(ASCII_CAN, &tp->t_outq);	/* maybe force error? */
			tc->tc_if.if_collisions++;
		} else
			tc->tc_if.if_opackets++;
	}
}

tlinit(tc)
	register struct tl_softc *tc;
{
	struct mbuf *p;

	if (tc->tc_buf == (char *) 0) {
		MCLALLOC(p, 1);
		if (p) {
			tc->tc_buf = (char *)p;
			tc->tc_mp = tc->tc_buf + sizeof(struct ifnet *);
		} else {
			printf("tl%d: can't allocate buffer\n", tc - tl_softc);
			tc->tc_if.if_flags &= ~IFF_UP;
			return (0);
		}
	}
	return (1);
}

/*
 * Copy data buffer to mbuf chain; add ifnet pointer ifp.
 */
struct mbuf *
tl_btom(tc, len, ifp)
	struct tl_softc *tc;
	register int len;
	struct ifnet *ifp;
{
	register caddr_t cp;
	register struct mbuf *m, **mp;
	register unsigned count;
	struct mbuf *top = NULL;

	cp = tc->tc_buf + sizeof(struct ifnet *);
	mp = &top;
	while (len > 0) {
		MGET(m, M_DONTWAIT, MT_DATA);
		if ((*mp = m) == NULL) {
			m_freem(top);
			return (NULL);
		}
		if (ifp)
			m->m_off += sizeof(ifp);
		/*
		 * If we have at least NBPG bytes,
		 * allocate a new page.  Swap the current buffer page
		 * with the new one.  We depend on having a space
		 * left at the beginning of the buffer
		 * for the interface pointer.
		 */
		if (len >= NBPG) {
			MCLGET(m);
			if (m->m_len == CLBYTES) {
				cp = mtod(m, char *);
				m->m_off = (int)tc->tc_buf - (int)m;
				tc->tc_buf = cp;
				if (ifp) {
					m->m_off += sizeof(ifp);
					count = MIN(len,
					    CLBYTES - sizeof(struct ifnet *));
				} else
					count = MIN(len, CLBYTES);
				goto nocopy;
			}
		}
		if (ifp)
			count = MIN(len, MLEN - sizeof(ifp));
		else
			count = MIN(len, MLEN);
		bcopy(cp, mtod(m, caddr_t), count);
nocopy:
		m->m_len = count;
		if (ifp) {
			m->m_off -= sizeof(ifp);
			m->m_len += sizeof(ifp);
			*mtod(m, struct ifnet **) = ifp;
			ifp = NULL;
		}
		cp += count;
		len -= count;
		mp = &m->m_next;
	}
	return (top);
}

/*
 * tty interface receiver interrupt.
 */
tlinput(c, tp)
	register int c;
	register struct tty *tp;
{
	register struct tl_softc *tc;
	register struct mbuf *m;
	int s;

	tk_nin++;
	tc = (struct tl_softc *)tp->t_tc;
	if (tc == NULL)
		return;

	c &= 0xff;

	switch (tc->tc_istate) {

	case TC_RESYNC_IDLE:
		switch (c) {
		case ASCII_DLE:
			tc->tc_istate = TC_MARKED_IDLE;
			return;

		case ASCII_DEL:
			tc->tc_istate = TC_NORMAL_IDLE;
			return;

		default:
			return;
		}

	case TC_NORMAL_IDLE:
		switch (c) {
		case ASCII_DLE:
			tc->tc_istate = TC_MARKED_IDLE;
			return;

		case ASCII_DEL:
			return;

		default:
			tc->tc_if.if_ierrors++;
			tc->tc_istate = TC_RESYNC_IDLE;
			return;
		}

	case TC_MARKED_IDLE:
		switch (c) {

		case ASCII_STX:
			tc->tc_istate = TC_NORMAL_DATA;
			return;

		default:
			tc->tc_if.if_ierrors++;
			tc->tc_istate = TC_RESYNC_IDLE;
			return;
		}

	case TC_MARKED_DATA:
		switch (c) {
		case ASCII_DLE:
			tc->tc_istate = TC_NORMAL_DATA;
			break;

		case ASCII_ETX:
			if (tc->tc_ilen == 0)	/* ignore */
				return;
			m = tl_btom(tc, tc->tc_ilen, &tc->tc_if);
			if (m == NULL) {
				tc->tc_if.if_ierrors++;
				return;
			}
			tc->tc_mp = tc->tc_buf + sizeof(struct ifnet *);
			tc->tc_ilen = 0;
			tc->tc_if.if_ipackets++;
			s = splimp();
			if (IF_QFULL(&ipintrq)) {
				IF_DROP(&ipintrq);
				tc->tc_if.if_ierrors++;
				m_freem(m);
			} else {
				IF_ENQUEUE(&ipintrq, m);
				schednetisr(NETISR_IP);
			}
			splx(s);
			tc->tc_istate = TC_NORMAL_IDLE;
			return;

		default:
			tc->tc_if.if_ierrors++;
			tc->tc_istate = TC_RESYNC_IDLE;
			return;
		}
		break;

	case TC_NORMAL_DATA:
		switch (c) {
		case ASCII_DLE:
			tc->tc_istate = TC_MARKED_DATA;
			return;

		default:
			break;
		}
		break;
	}

	if (++tc->tc_ilen > TLMTU) {
		tc->tc_if.if_ierrors++;
		tc->tc_mp = tc->tc_buf + sizeof(struct ifnet *);
		tc->tc_ilen = 0;
		return;
	}
	*tc->tc_mp++ = c;
}

/*
 * Process an ioctl request.
 */
tlioctl(ifp, cmd, data)
	register struct ifnet *ifp;
	int cmd;
	caddr_t data;
{
	register struct ifaddr *ifa = (struct ifaddr *)data;
	int s = splimp(), error = 0;

	switch (cmd) {

	case SIOCSIFADDR:
		if (ifa->ifa_addr.sa_family == AF_INET)
			ifp->if_flags |= IFF_UP;
		else
			error = EAFNOSUPPORT;
		break;

	case SIOCSIFDSTADDR:
		if (ifa->ifa_addr.sa_family != AF_INET)
			error = EAFNOSUPPORT;
		break;

	default:
		error = EINVAL;
	}
	splx(s);
	return (error);
}
#endif
