
#include "vca.h"
#if NVCA > 0

/* Voice Communications Adapter (PN 6294771) driver for the IBM RT PC. */

#ifdef IOMEM		/* Defined in configuartion file as an option */
#define IOMEM_ADDR 0xf4100000
int iomem_exists;
int iomem_alloc;
#endif /* IOMEM */

#include "../h/param.h"			/* Defines priorities. */
#include "../h/errno.h"
#include "../h/uio.h"
#include "../h/ioctl.h"
#include "../h/tty.h"
#include "../machine/pte.h"
#include "../machine/io.h"
#include "../machine/scr.h"
#include "../h/buf.h"
#include "../machineio/ioccvar.h"
#include "../machine/debug.h"
#include "../h/time.h"
#include "../h/kernel.h"		/* This defines the variable time. */
#include "../h/dir.h"			/* Needed because user.h does not include this! */
#include "../h/user.h"			/* includes "../h/time.h" */
#include "../h/file.h"
#ifdef VFS
#include "../h/vnode.h"
#else  /* VFS */
#include "../h/inode.h"
#endif /* VFS */

#if 0
#define TOKEN_RING_MODIFICATIONS
#endif

#ifdef TOKEN_RING_MODIFICATIONS
/* We are building a kernel with the token ring driver modifications */

extern struct mbuf *lan_get_token_ring_header();    /* Defined in if_lan.c */
extern int tr_arpresolve();			    /* Defined in netinet/if_ether.c */
#define MAX_TOKEN_RING_HEADER_LENGTH \
sizeof(struct mac_hdr) + sizeof(u_short)*(1+LAN_MAX_BRIDGE) + \
sizeof(struct llc_hdr) + sizeof(struct snap_hdr)

/* The following includes specifically for the vca-to-lan shortcut */
/* Start of list */
#include "../h/socket.h"
#include "../h/socketvar.h"
#include "../h/mbuf.h"
 
#include "../net/if.h"
#include "../net/route.h"
#include "../netinet/tcp.h"
#include "../netinet/tcp_timer.h"
#include "../netinet/tcp_var.h"
#include "../netinet/in.h"
#include "../netinet/in_systm.h"
#include "../netinet/in_var.h"
#include "../netinet/in_pcb.h"
#include "../netinet/ip_var.h"
#include "../netinet/if_ether.h"
#include "../netinet/if_lan.h"

#include "../machineif/if_lanreg.h"
#include "../machineif/if_landma.h"
#include "../caif/if_lanvar.h"
#include "../caif/if_lanio.h"

/* End of list */

#endif /* TOKEN_RING_MODIFICATIONS */

#include "../h/proc.h"

#define VCADebug
#ifdef VCADebug
static unsigned short CompletedHostFunctionRequests;
#endif /* VCADebug */

#include "vcaio.h"
#include "vca_rw.h"
#include "vca_probe.h"

int vcaprobe(), vcaattach(), vcaopen(), vcaclose(), vcaread(), vcawrite(), vcaioctl(), vcaintr(), vcaselect();

caddr_t vcastd[] = {
    (caddr_t)(IO_BASE + 0x021f),		/* Memory at DA000 */
    (caddr_t)(IO_BASE + 0x221f),		/* Memory at D8000 */
    (caddr_t)(IO_BASE + 0x421f),		/* Memory at DC000 */
    (caddr_t)(IO_BASE + 0x621f),		/* Memory at DE000 */
    (caddr_t)0 };

caddr_t vcamem[] = {
    (caddr_t)(MEM_BASE + 0x0da000),
    (caddr_t)(MEM_BASE + 0x0d8000),
    (caddr_t)(MEM_BASE + 0x0dc000),
    (caddr_t)(MEM_BASE + 0x0de000) };

struct iocc_device *vcadinfo[NVCA];

struct iocc_driver vcadriver = {
	vcaprobe, 0, vcaattach, 0 /* dgo */,
	vcastd, "vca", vcadinfo, 0, 0, vcaintr
};

/* IO Delay Register: Reference pages 5-12 and 5-74 in IBM RT PC Technical Reference Volumne 1. */
#define IODelayRegister 0xf00080e0
#define OneMicroSecondDelay *((char *)IODelayRegister) = 0

struct VCAChannel {
    dev_t dev;
    short open;
    short RWpid;
    short RIpid;
    short BlockingReads;
    short ADC_Bit;
    int DAC_Enabled_Offset;
#ifdef TOKEN_RING_MODIFICATIONS
    int vca2lan;
    int drop_packets;
    int packet_total_size;
    int vca2lan_device;
    struct ifnet *ifnet;
    struct { int len, queue, priority; char data[MAX_TOKEN_RING_HEADER_LENGTH]; } trhdr;
    int vca2lan_packet_number;
    int lan2vca_packet_number;
#endif /* TOKEN_RING_MODIFICATIONS */
    int fp, disconnect;
    struct VCAChannel *connected, *from;
    int in_start, in_end;
    int out_start, out_end;
    vcalocation output[vcabufsize];
    vcalocation input[vcabufsize];
};

struct VCACard {
    char *iocr;
    vcalocation *basemem;
    struct VCAChannel channel[2];
};

struct VCACard vcacards[NVCA];

#define	VcaFunctionRequest(card, chan, offset, value)	\
{   vcalocation	*loc = &((card)->basemem[(offset)+(chan)]);		\
    *(card)->iocr |= VCA_ENA_INT;					\
    *loc = 0x8000 | (value);						\
    if ((*(card)->iocr & VCA_INT_PSP) == 0 && *loc != 0)		\
	    *(card)->iocr |= VCA_INT_PSP; }

int vcaStartPlaying(dev)
dev_t dev;
{
    DEBUGprintf(dev, DebugIoctl,
		 ("     Start Playing %d\n", VcaCard(dev), VcaChannel(dev)));
    VcaFunctionRequest(&(vcacards[VcaCard(dev)]), VcaChannel(dev),
			VCA_SET_PLAYING_OFFSET, 1);
}

void vcaLoadPSP(iocr, basemem, program, length, checksum)
char *iocr;
vcalocation *basemem;
loadpsp *program;
vcalocation *checksum;
{
    int x, loc;
    *iocr = VCA_DISABLE_ALL;
    for (x=0; x<length; x++) {
	if (checksum != 0) {
	    checksum[x%VCA_PROGRAM_ID_SIZE] ^= program[x].loc;
	    checksum[x%VCA_PROGRAM_ID_SIZE] ^= program[x].data;
	}
	loc = program[x].loc;
	if (program[x].type == 'i') {
	    if (loc >= 0x4000) {
		/* Correct for instruction mapped into high 4k of instruction memory. */
		*iocr = (VCA_SELECT_INST_MEM_HIGH | VCA_SELECT_INST_MEM);
		loc -= 0x4000;
	    } else
		*iocr = VCA_SELECT_INST_MEM;
	} else
	    *iocr = VCA_DISABLE_ALL;
	if (program[x].type != 'x' ||
	    loc<VCA_PROGRAM_ID_OFFSET ||
	    loc>=VCA_PROGRAM_ID_OFFSET+VCA_PROGRAM_ID_SIZE) {
	    if (loc < 0 || loc > 0x4000)
		printf("VCA Location Error: %d 0x%x\n", x, loc);
	    else
		basemem[loc] = program[x].data;
	}
    }
    *iocr = VCA_DISABLE_ALL;
}

vcaprobe(iocr)
char *iocr;
{
    static FirstProbe = 1;

    int dev = (0xffff & (unsigned long)iocr)>>13;
    vcalocation *basemem = (vcalocation *)vcamem[dev];
    char test;
    int delay;

    DEBUGprintf(dev<<1, DebugAny,
		 ("vcaProbe Called: 0x%x, Base Address: 0x%x\n", iocr, basemem));

    if (FirstProbe == 1) {
	int x;
	for (x=0; x<NVCA; x++) {
	    debug[x] = 0;
	    vcacards[x].iocr = NULL;
	    vcacards[x].basemem = NULL;
	}
	FirstProbe = 0;
#ifdef IOMEM
	if (! iomem_exists) {
	  *((u_short *) IOMEM_ADDR) = 0x5aa5;
	  if (*((u_short *) IOMEM_ADDR) == 0x5aa5) {
	    /* IO memory exist at IOMEM_ADDR. Use this for the transmit/receive buffers. */
	    int x;
	    u_char *mem = (u_char *)IOMEM_ADDR;
	    
	    iomem_exists = 1;
	    printf("Using Existing IO Memory for vca data buffers\n");
	    for (x = 0; x < 512 * 1024; x++)
	      mem[x] = 0;
	  }
	}
#endif /* IOMEM */
    }

    /* Select memory, disable processor, disable interrupts in both directions. */
    *iocr = (char)VCA_DISABLE_ALL;
    *basemem = (test = *basemem ^ (char)0xa5);
    DEBUGprintf(dev<<1, DebugAny,
		 ("Memory Write/Read %s.\n", (*basemem == test) ? "ok" : "failed"));
    DEBUGprintf(dev<<1, DebugAny, ("   Value written: %x\n", (vcalocation)test));
    if (test != *basemem) {
	printf("VCA memory read/write test failed\n");
	return PROBE_BAD;
    }

    /* Load probe program */
    vcaLoadPSP(iocr, basemem, vca_probe_asm, sizeof(vca_probe_asm)/sizeof(loadpsp), 0);
    /* Clear Program ID. */
    bzero(((char *)basemem)+VCA_PROGRAM_ID_OFFSET, VCA_PROGRAM_ID_SIZE*2);

    DEBUGprintf(dev<<1, DebugAny,
		 ("Value at location 0 before PSP is Enabled: 0x%x\n", *basemem));
    *iocr = (char)VCA_ENA_PSP;
    /* Delay while PSP has chance to initialize. */
    for (delay=0; delay<1000; delay++) OneMicroSecondDelay;
    DEBUGprintf(dev<<1, DebugAny,
		 ("PSP%s initialized memory location zero to zero.\n",
		  (*basemem == 0) ? " correctly" : " failed to"));
    DEBUGprintf(dev<<1, DebugAny, ("Value at location 0: 0x%x\n", *basemem));
    if (*basemem != 0) {
	printf("VCA Enable PSP test failed\n");
	return PROBE_BAD;
    }

    /* Test Host to PSP interrupt. */
    *iocr = (char)(VCA_INT_PSP | VCA_ENA_PSP);
    /* Delay while PSP has chance to service interrupt. */
    for (delay=0; delay<1000 && (*iocr & VCA_INT_PSP!=0); delay++)
	OneMicroSecondDelay;
    /* Delay to finish interrupt handler routine. */
    for (delay=0; delay<1000; delay++) OneMicroSecondDelay;
    DEBUGprintf(dev<<1, DebugAny,
		 ("Host to PSP interrupt functioned %scorrectly.\n",
		  (*basemem == 1 && (*iocr & VCA_INT_PSP)==0) ? "" : "in"));
    DEBUGprintf(dev<<1, DebugAny, ("Value at location 0: 0x%x\n", *basemem));
    if (*basemem != 1 || (*iocr & VCA_INT_PSP) != 0) {
	printf("VCA Host to PSP interrupt test failed\n");
	return PROBE_BAD;
    }

    /* Test PSP to Host interrupt. */
    *iocr = (char)VCA_SELECT_INST_MEM;	/* Select PSP's Instruction memory Low. */
    *(basemem+1) = 0x100;			/* Modify interrupt vector. */
    *iocr = (char)(VCA_ENA_PSP | VCA_ENA_INT);
    for (delay=0; delay<1000; delay++)
	OneMicroSecondDelay;
    *iocr = (char)VCA_DISABLE_ALL;

    return PROBE_OK;
}

vcaattach(iod)
struct iocc_device *iod;
{
    int channel;
    int dev = (0xffff & (unsigned long)iod->iod_addr)>>13;
    struct VCACard *card = &(vcacards[dev]);
    card->iocr = iod->iod_addr;
    card->basemem = (vcalocation *)vcamem[dev];
    DEBUGprintf(dev<<1, DebugAny, ("vcaAttach Called: %d (0x%x, 0x%x)\n",
			    dev, card->iocr, card->basemem));
    for (channel=0; channel<2; channel++)
	card->channel[channel].open = 0;
    card->channel[0].ADC_Bit = ADC_CH1;
    card->channel[0].DAC_Enabled_Offset = VCA_DAC1_ENALBED_OFFSET;
    card->channel[1].ADC_Bit = ADC_CH2;
    card->channel[1].DAC_Enabled_Offset = VCA_DAC2_ENALBED_OFFSET;
    return 0;
}

resetChannel(dev)
dev_t dev;
{
    struct VCACard *card = &(vcacards[VcaCard(dev)]);
    struct VCAChannel *channel = &(card->channel[VcaChannel(dev)]);

    if (*card->iocr & VCA_ENA_PSP) {
	VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_PLAYING_OFFSET, 0);
	closeChannel(dev);
    }

    /* If this card does not have the other channel active at this time then load the standard interface program code onto this card. */
    if (card->channel[(VcaChannel(dev)==0) ? 1:0].open == 0) {
	vcalocation checksum[VCA_PROGRAM_ID_SIZE];
	vcaLoadPSP(card->iocr, card->basemem, vca_rw_asm,
		   sizeof(vca_rw_asm)/sizeof(loadpsp), checksum);
#ifdef VCADebug
	if (debug[VcaCard(dev)+VcaChannel(dev)]) {
	    int x;
	    printf("%d %d: Checksum:", VcaCard(dev), VcaChannel(dev));
	    for (x=0;x<VCA_PROGRAM_ID_SIZE;x++)
		printf(" %x", 0xffff&checksum[x]);
	    printf("\n");
	}
#endif /* VCADebug */
	bcopy(checksum, &(card->basemem[VCA_PROGRAM_ID_OFFSET]),
	      VCA_PROGRAM_ID_SIZE*sizeof(vcalocation));
    }

    *card->iocr |= VCA_ENA_PSP | VCA_ENA_INT;

    /* Just in case one resets this channel to wakeup */
    wakeup(channel->input);
    wakeup(channel->output);

    if (SpecialMark(dev)) {
	switch (SpecialMux(dev)) {
	    case Phone:
		/* Set mux to phone and take phone off hook. */
		DEBUGprintf(dev, DebugAny,
			    ("%d %d: VCA Special: Tele\n", VcaCard(dev), VcaChannel(dev)));
		VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_MULTIPLEXER_OFFSET, Phone);
		VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_OFF_HOOK_OFFSET, 1);
		break;
	    case Tele:
		/* Set bypass and set mux to telephone. */
		DEBUGprintf(dev, DebugAny,
			    ("%d %d: VCA Special: Tele\n", VcaCard(dev), VcaChannel(dev)));
		VcaFunctionRequest(card, 0, VCA_SET_BYPASS_OFFSET, 1);
		VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_MULTIPLEXER_OFFSET, Tele);
		break;
	    case Mic:
		/* Set mux to Microphone/Speaker. */
		DEBUGprintf(dev, DebugAny,
			    ("%d %d: VCA Special: Mic/Spk\n", VcaCard(dev), VcaChannel(dev)));
		VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_MULTIPLEXER_OFFSET, Mic);
		break;
	    case ACWrap:
		/* Set mux to AC Wrap. */
		DEBUGprintf(dev, DebugAny,
			    ("%d %d: VCA Special: AC Wrap\n", VcaCard(dev), VcaChannel(dev)));
		VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_MULTIPLEXER_OFFSET, ACWrap);
		break;
	    default:
		printf("%d %d: Unknown Special VCA device\n", VcaCard(dev), VcaChannel(dev));
		break; }
	if (SpecialRec(dev))
	    VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_RECORDING_OFFSET, 1);
    }

    return 0;
}

int lan_drop_audio_packets_before_copy;
vcaopen(dev, flag)
dev_t dev;
int flag;
{
    struct VCACard *card;
    struct VCAChannel *channel;
    int attached = 0;
    int x;

    DEBUGprintf(dev, DebugAny, ("VCA open called: dev: (%d, %d (%d, %d)) flag: %d\n",
		  major(dev), minor(dev), VcaCard(dev), VcaChannel(dev), flag));
    if (VcaCard(dev) > (sizeof(vcadinfo)/sizeof(struct iocc_device *) - 1)) {
	DEBUGprintf(dev, DebugAny, ("No such device\n"));
	return ENXIO;
    }
    card = &(vcacards[VcaCard(dev)]);
    for (x=0; x<NVCA; x++)
	if (vcastd[x] == card->iocr && (vcalocation *)(vcamem[x]) == card->basemem) {
	    attached = 1;
	    break; }
    if (attached == 0) {
	DEBUGprintf(dev, DebugAny, ("Device not attached\n"));
	return ENXIO;
    }

    channel = &(card->channel[VcaChannel(dev)]);

#ifdef VCADebug
    CompletedHostFunctionRequests = 0;
#endif /* VCADebug */

    channel->dev = dev;
    channel->BlockingReads = 1;
    channel->RWpid = 0;
    channel->RIpid = 0;
    channel->in_start = channel->in_end = 0;
    channel->out_start = channel->out_end = 0;
    channel->open++;
#ifdef TOKEN_RING_MODIFICATIONS
    channel->vca2lan = -1;
    channel->drop_packets = VCA_DO_NOT_DROP;
    channel->packet_total_size = -1;
    channel->vca2lan_device = -1;
    channel->ifnet = NULL;
    channel->trhdr.len = 0;
    channel->trhdr.queue = 0;
    channel->trhdr.priority = 0;
    channel->vca2lan_packet_number = 0;
    channel->lan2vca_packet_number = 0;
#endif /* TOKEN_RING_MODIFICATIONS */
    channel->fp = -1;
    channel->disconnect = 0;
    channel->connected = NULL;
    channel->from = NULL;

    lan_drop_audio_packets_before_copy = 0;

    resetChannel(dev);
    return 0;
}

closeChannel(dev)
dev_t dev;
{
    struct VCACard *card = &(vcacards[VcaCard(dev)]);
    struct VCAChannel *channel = &(card->channel[VcaChannel(dev)]);
    vcalocation *sc2 = &(card->basemem[VCA_SHADOW_OF_SC2_OFFSET]);
    int loop;

    /* if this channel is connected from another vca channel then turn recording off on the connecting channel and disconnect. */
    if (channel->from != NULL) {
	VcaFunctionRequest(&(vcacards[VcaCard(channel->from->dev)]),
			   VcaChannel(channel->from->dev), VCA_SET_RECORDING_OFFSET, 0);
	channel->from->connected = NULL;
	channel->from = NULL;
    }

    /* If recording is on for this channel then turn recording off and loop until vca acknowledges. */
    if (*sc2&channel->ADC_Bit != 0) {
	VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_RECORDING_OFFSET, 0);
	loop = 0;
	while (*sc2&channel->ADC_Bit != 0 &&
	       loop<10000) {
	    loop++;
	    OneMicroSecondDelay;
	}
	DEBUGprintf(dev, DebugAny, ("     %d %d: Recording Turned off\n",
		    VcaCard(dev), VcaChannel(dev)));
    }

    /* If this channel is connected to another vca channel then disconnect. */
    if (channel->connected != NULL) {
#ifdef VCADebug
	dev_t connected = channel->connected->dev;
#endif /* VCADebug */
	channel->connected->from = NULL;
	channel->connected = NULL;
	DEBUGprintf(dev, DebugAny,
		    ("     %d %d: Disconnect from (%d, %d)\n",
		     VcaCard(dev), VcaChannel(dev), VcaCard(connected), VcaChannel(connected)));
    }

    /* If write in progress on this channel then allow write to complete. */
    {
	short test;
	char hold_debug = debug[VcaCard(dev)+VcaChannel(dev)];
	debug[VcaCard(dev)+VcaChannel(dev)] = 0;
	vcaioctl(dev, _IOR(\0, VCA_WRITE_IN_PROGRESS_P, test), &test, 0);
	if (test == 1) {
	    loop = 0;
	    do {
		loop++;
		OneMicroSecondDelay;
		vcaioctl(dev, _IOR(\0, VCA_WRITE_IN_PROGRESS_P, test), &test, 0);
	    } while (test == 1 && loop<100000);
	    debug[VcaCard(dev)+VcaChannel(dev)] = hold_debug;
	    if (test == 1) {
		int priority = splvca();
		channel->out_start = channel->out_end = 0;
		splx(priority);
		DEBUGprintf(dev, DebugAny,
			    ("     %d %d: Write terminated\n",
			     VcaCard(dev), VcaChannel(dev)));
	    } else {
		DEBUGprintf(dev, DebugAny,
			    ("     %d %d: Write allowed to complete\n",
			     VcaCard(dev), VcaChannel(dev)));
	    }
	} else {
	    debug[VcaCard(dev)+VcaChannel(dev)] = hold_debug;
	    DEBUGprintf(dev, DebugAny,
			("     %d %d: No write in progress\n",
			 VcaCard(dev), VcaChannel(dev)));
	}
    }

    /* If this channel is channel zero and the Bypass bit is set then clear the Bypass bit and loop until vca acknowledges. */
    if (VcaChannel(dev) == 0 && ((*sc2)&VCA_SC2_BYPASS) != 0) {
	VcaFunctionRequest(card, 0, VCA_SET_BYPASS_OFFSET, 0);
	loop = 0;
	while (*sc2&VCA_SC2_BYPASS != 0 &&
	       loop<10000) {
	    loop++;
	    OneMicroSecondDelay;
	}
	DEBUGprintf(dev, DebugAny, ("     %d %d: Cleared Bypass bit\n",
				    VcaCard(dev), VcaChannel(dev)));
    }

    /* If the telephone for this channel is off hook then hang up the phone and loop until vca acknowledges. */
    {
	int shift = (VcaChannel(dev) == 0) ? 9:10;
	if (((*sc2>>shift)&1) != 0) {
	    VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_OFF_HOOK_OFFSET, 0);
	    loop = 0;
	    while (((*sc2>>shift)&1) != 0 &&
		   loop<10000) {
		loop++;
		OneMicroSecondDelay;
	    }
	    DEBUGprintf(dev, DebugAny, ("     %d %d: Phone line hung up\n",
					VcaCard(dev), VcaChannel(dev)));
	}
    }

    /* If DC Wrap is set then clear DC Wrap and loop until vca acknowledges. */
    if (*sc2&VCA_SC2_DCWRAP != 0) {
	VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_DCWRAP_OFFSET, 0);
	loop = 0;
	while (*sc2&VCA_SC2_DCWRAP != 0 &&
	       loop<10000) {
	    loop++;
	    OneMicroSecondDelay;
	}
	DEBUGprintf(dev, DebugAny, ("     %d %d: Cleared DC Wrap bit\n",
				    VcaCard(dev), VcaChannel(dev)));
    }

    /* If the Multiplexer for this channel is not ACWrap then set the Mux to ACWrap and loop until vca acknowledges. */
    if (((*sc2>>((VcaChannel(dev)==0) ? 12 : 14))&3) != ACWrap) {
	VcaFunctionRequest(card, VcaChannel(dev),
					 VCA_SET_MULTIPLEXER_OFFSET, ACWrap);
	loop = 0;
	while (((*sc2>>((VcaChannel(dev)==0) ? 12 : 14))&3) != ACWrap &&
	       loop<10000) {
	    loop++;
	    OneMicroSecondDelay;
	}
	DEBUGprintf(dev, DebugAny, ("     %d %d: Multiplexer reset to AC Wrap\n",
				    VcaCard(dev), VcaChannel(dev)));
    }

    return 0;
}

vcaclose(dev, flag)
dev_t dev;
int flag;
{
    struct VCACard *card = &(vcacards[VcaCard(dev)]);
    struct VCAChannel *channel = &(card->channel[VcaChannel(dev)]);

    DEBUGprintf(dev, DebugAny, ("VCA close called(%d): dev: (%d, %d, %d) flag: %d\n",
				 (int)u.u_procp->p_pid, major(dev),
				 VcaCard(dev), VcaChannel(dev), flag));
    channel->open--;
    if (channel->RWpid == (int)u.u_procp->p_pid)
	channel->RWpid = 0;
    if (channel->RIpid == (int)u.u_procp->p_pid)
	channel->RIpid = 0;
    if (channel->open == 0)
	closeChannel(dev);

    DEBUGprintf(dev, DebugAny,
		 ("     %d %d: Close complete: ", VcaCard(dev), VcaChannel(dev)));

    if (card->channel[0].open == 0 && card->channel[1].open == 0) {
	/* Disable VCA. */
	*card->iocr &= (0xff ^ VCA_ENA_PSP);
	DEBUGprintf(dev, DebugAny, ("card disabled\n"));
#ifdef VCADebug
	/* Turn off Debuging. */
	DebugCount[VcaCard(dev)+VcaChannel(dev)] = 0;
	debug[VcaCard(dev)+VcaChannel(dev)] = 0;
#endif /* VCADebug */
    } else {
	if (card->channel->open != 0) {
	    DEBUGprintf(dev, DebugAny, ("Channel still open by another process\n"));
	} else { 
	    DEBUGprintf(dev, DebugAny, ("Other channel still active\n"));
	}
    }

    return 0;
}

vcaread(dev, uio)
dev_t dev;
struct uio *uio;
{
    struct VCACard *card = &(vcacards[VcaCard(dev)]);
    struct VCAChannel *channel = &(card->channel[VcaChannel(dev)]);
    int in_start, in_end;
    int count, diff;
    char *inbuf;
    int priority;

    DEBUGprintf(dev, DebugRead,
		 ("%d %d: VCA Read Called\n", VcaCard(dev), VcaChannel(dev)));
    count = uio->uio_resid;
    DEBUGprintf(dev, DebugRead, ("     %d %d: Number of bytes requested: %d\n",
				  VcaCard(dev), VcaChannel(dev), count));
    /* Only tranfer whole samples. */
    count -= count%sizeof(vcalocation);
    if (count == 0) return 0;

    inbuf = (char *)channel->input;

    /* If not currently recording then start recording. */
    if ((card->basemem[VCA_SHADOW_OF_SC2_OFFSET] & channel->ADC_Bit) ==0) {
	channel->in_start = channel->in_end = 0;
	VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_RECORDING_OFFSET, 1);
	DEBUGprintf(dev, DebugRead, ("     %d %d: Start recording",
				     VcaCard(dev), VcaChannel(dev)));
    }

    /* Test for data available. Must be done with interrupts disabled because otherewise we might miss the wakeup on which we might sleep. */
    priority = splvca();
    if (channel->in_start == channel->in_end) {
	if (channel->BlockingReads != 0) {
	    DEBUGprintf(dev, DebugRead, ("     %d %d: Sleep on 0x%x\n",
					 VcaCard(dev), VcaChannel(dev), channel->input));
	    sleep(inbuf, PUSER);
	} else {
	    splx(priority);
	    return 0;
	}
    }

    /* There is a posibility that an interrupt may change the values of in_start and in_end during the following processing. This would only be a problem if an overrun occurred. But why not be paranoid? */
    priority = splvca();

    /* At this point there is data in the buffer. */
    /* All calculations are done in bytes not samples with the exception of updating the in_start pointer. */
    in_start = channel->in_start*sizeof(vcalocation);
    in_end = channel->in_end*sizeof(vcalocation);

    /* Determine if the number of bytes requested is less than is available. */
    diff = in_end - in_start;
    /* If either of the two cases below are true than more has been requested then is available.  Adjust count to be exactly the number of bytes*/
    if (diff > 0) {
	if (count > diff)
	    /* Start Pointer before End Pointer: Buffer is not looped. */
	    count = diff;
    } else if (count > vcabytebufsize + diff)
	/* Start Pointer after End Pointer: Buffer is looped. */
	count = vcabytebufsize + diff;

    /* Transfer into user space 'count' number of bytes. */
    if ((diff = vcabytebufsize - in_start) < count) {
	/* Count bytes wraps around the ring buffer. Transfer in two pieces. */
	if (uiomove(&(inbuf[in_start]), diff, UIO_READ, uio)) {
	    DEBUGprintf(dev, DebugRead, ("     %d %d: Read A uiomove Error\n",
					 VcaCard(dev), VcaChannel(dev)));
	    return(EFAULT);
	}
	/* From the test above, count will not be requesting more of the input buffer then is available. */
	count -= diff;
	if (uiomove(inbuf, count, UIO_READ, uio)) {
	    DEBUGprintf(dev, DebugRead, ("     %d %d: Read B uiomove Error\n",
					 VcaCard(dev), VcaChannel(dev)));
	    return(EFAULT);
	}
	/* Update the IN pointer. */
	channel->in_start = count/sizeof(vcalocation);
	DEBUGprintf(dev, DebugRead,
		    ("     %d %d: Read Trans1: (0x%x, 0x%x)\n     %d %d: Read Trans2: (0x%x, 0x%x)\n",
				VcaCard(dev), VcaChannel(dev), &(inbuf[in_start]), diff,
				VcaCard(dev), VcaChannel(dev), inbuf, count));
    } else {
	if (uiomove(&(inbuf[in_start]), count, UIO_READ, uio)) {
	    DEBUGprintf(dev, DebugRead, ("     %d %d: Read uiomove Error\n",
					 VcaCard(dev), VcaChannel(dev)));
	    return(EFAULT);
	}
	/* Update the IN pointer. */
	if ((channel->in_start += count/sizeof(vcalocation)) == vcabufsize)
	    channel->in_start = 0;
	DEBUGprintf(dev, DebugRead,
		    ("     %d %d: Read Trans: (0x%x, 0x%x)\n",
		     VcaCard(dev), VcaChannel(dev), &(inbuf[in_start]), count));
    }

    splx(priority);
    return 0;
}

vcawrite(dev, uio)
dev_t dev;
struct uio *uio;
{
    struct VCACard *card = &(vcacards[VcaCard(dev)]);
    struct VCAChannel *channel = &(card->channel[VcaChannel(dev)]);
    int count, move, priority;

    DEBUGprintf(dev, DebugWrite, ("%d %d: VCA Write Called\n", VcaCard(dev), VcaChannel(dev)));
    count = uio->uio_resid;
    DEBUGprintf(dev, DebugWrite, ("     %d %d: Number of bytes to write: 0x%x\n",
		  VcaCard(dev), VcaChannel(dev), count));
    count -= count%sizeof(vcalocation);
    if (count == 0)
	return 0;

    /* Previous write may still be in progress. Test must be done with interrupts disabled. */
    priority = splvca();
    if (channel->out_start != channel->out_end) {
	int diff;
	if (card->basemem[channel->DAC_Enabled_Offset] == 0)
	    vcaStartPlaying(dev);
	/* Previous write still in progress. */
	diff = (vcabufsize - channel->out_end)*sizeof(vcalocation);
	DEBUGprintf(dev, DebugWrite,
		    ("     %d %d: Previous: (0x%x, 0x%x), bytes to end: 0x%x\n",
		     VcaCard(dev), VcaChannel(dev),
		     channel->out_start, channel->out_end, diff));
	if (diff > 0) {
	    /* Room in kernal buffer for more User samples. */
	    if (count > diff) {
		/* User has given more samples then there is room left in buffer. Move User samples to fill the kernal buffer and sleep on the buffer waiting for the samples to be output. */
		if (uiomove(&(channel->output[channel->out_end]), diff, UIO_WRITE, uio)) {
		    DEBUGprintf(dev, DebugWrite,
				("     %d %d: Write A uiomove Error 0x%x remaining\n",
				 VcaCard(dev), VcaChannel(dev), count))
		    return(EFAULT);
		}
		channel->out_end += diff/sizeof(vcalocation);
		/* If the process wants an interrupt when there is no more data in the kernel buffer then return else sleep. */
		if (channel->RWpid != 0) {
		    splx(priority);
		    return 0;
		} else {
		    sleep(channel->output, PUSER);
		    /* Update the count of the bytes left to transfer. */
		    count -= diff;
		}
	    } else {
		/* User has given fewer samples then there is room left in buffer. Add to the kernal buffer the User samples, adjust the end pointer, reset the priority level of the process and return. */
		if (uiomove(&(channel->output[channel->out_end]), count, UIO_WRITE, uio)) {
		    DEBUGprintf(dev, DebugWrite,
				("     %d %d: Write B uiomove Error 0x%x remaining\n",
				 VcaCard(dev), VcaChannel(dev), count))
		    return(EFAULT);
		}
		channel->out_end += count/sizeof(vcalocation);
		splx(priority);
		return 0;
	    }
	} else {
	    /* No room in kernal buffer for more User samples. If the process wants an interrupt when there is no more data in the kernel buffer then return else sleep. */
	    if (channel->RWpid != 0) {
		splx(priority);
		return 0;
	    } else
		sleep(channel->output, PUSER);
	}
    } else {
	DEBUGprintf(dev, DebugWrite,
		    ("     %d %d: Previous write not in progress\n",
		     VcaCard(dev), VcaChannel(dev)));
	/* Previous write not in progress. Reset the priority level. */
	splx(priority);
    }

    for (;;) {
	/* One should try to leave this routine with a kernal buffer as full as possible so that the user process has the maximum time possible to compute new samples before the VCA needs more. The following if statement is a stab at doing this. Namely, if the count is smaller than the kernal buffer size then move the whole thing. Else first move the piece which would be left and then full kernal buffers after that. */
	if ((move = count) > vcabytebufsize)
	    if ((move = count%vcabytebufsize) == 0)
		move = vcabytebufsize;
	if (uiomove(channel->output, move, UIO_WRITE, uio)) {
	    DEBUGprintf(dev, DebugWrite,
			("     %d %d: Write uiomove Error 0x%x remaining\n",
			 VcaCard(dev), VcaChannel(dev), count))
	      return(EFAULT);
	}
	count -= move;
	channel->out_start = 0;
	channel->out_end = move/sizeof(vcalocation);

#ifdef VCADebug
	if (debug[VcaCard(dev)+VcaChannel(dev)]&DebugWrite) {
	    printf("     %d %d: Moved 0x%x bytes", VcaCard(dev), VcaChannel(dev), move);
	    if (count>0)
		printf(" 0x%x remaining\n", count);
	    else
		printf("\n");
	}
#endif /* VCADebug */

	/* Kick start the writing of samples on the vca board */
	VcaFunctionRequest(card, 0, VCA_GENERATE_HOST_INTERRUPT_OFFSET, 0);

	if (count>0 && channel->RWpid==0) {
	    DEBUGprintf(dev, DebugWrite, ("     %d %d: going to sleep",
					  VcaCard(dev), VcaChannel(dev)));
	    sleep(channel->output, PUSER);
	    DEBUGprintf(dev, DebugWrite, ("     %d %d: woken up",
					  VcaCard(dev), VcaChannel(dev)));
	} else
	    break;
    }

    return 0;
}

#ifdef TOKEN_RING_MODIFICATIONS
int lan2vca(m)
struct mbuf *m;
{
    int priority;
    struct mbuf *m0 = m;
    int len=0, mbufs=0;
    int dev, packet_number, lan2vca_packet_number;
    struct VCACard *card;
    struct VCAChannel *channel;
    char *data = mtod(m, char *);

    dev = ((short *)data)[0];
    bcopy(&data[sizeof(short)], (char *)&packet_number, sizeof(int));
    m->m_off += sizeof(short) + sizeof(int);
    m->m_len -= sizeof(short) + sizeof(int);
    card = &(vcacards[VcaCard(dev)]);
    channel = &(card->channel[VcaChannel(dev)]);

    /* XXX Reset count if packet number is zero */
    priority = splvca();
    if (packet_number == 0)
	channel->lan2vca_packet_number = 0;
    lan2vca_packet_number = channel->lan2vca_packet_number;
    if (channel->lan2vca_packet_number == 0x7fffffff)
	channel->lan2vca_packet_number = 0;
    else
	channel->lan2vca_packet_number++;
    splx(priority);

    if (major(dev) == VCAMajor) {
	if (channel->drop_packets < VCA_DROP_WITHOUT_COPY) {
	    int end = channel->out_end;
	    char *outbuf = (char *)&(channel->output[end]); 
	    for (m0=m; m0; m0=m0->m_next) {
		char *data = mtod(m0, char *);
		bcopy(data, outbuf, m0->m_len);
		outbuf += m0->m_len;
		len += m0->m_len;
		mbufs++;
	    }
	    /* Make sure that the vca has been attached and opened otherwise act as /dev/null */
	    if (channel->drop_packets == VCA_DO_NOT_DROP &&
		card->basemem != NULL &&
		channel->open != 0) {
		/* Update pointers into output buffer and make sure channel is active */
		if (channel->out_end == channel->out_start) {
		    channel->out_start = end;
		    channel->out_end = end + len/sizeof(vcalocation);
		} else {
		    /* Crittical code: Cannot let an interrupt change these pointers while we are trying to change the pointers */
		    priority = splvca();
		    channel->out_end = end + len/sizeof(vcalocation);
		    /* If, while this data copy happened, the card interrupted and the lower half of the kernel reset the output buffer pointers to zero then reset the output start pointer to the beginning of the just copied data. */
		    if (channel->out_start == 0 && end != 0)
			channel->out_start = end;
		    splx(priority);
		}
		/* Kick the card */
		VcaFunctionRequest(card, 0, VCA_GENERATE_HOST_INTERRUPT_OFFSET, 0);
	    }
	}
    } else {
	/* Must be testing the lan. Drop the data on the floor */
	for (m0=m; m0; m0=m0->m_next) {
	    len += m0->m_len;
	    mbufs++;
	}
    }

    m_freem(m);

    /* Test for packet received out of order */
    if (lan2vca_packet_number != packet_number) {
#if 0
#ifdef VCADebug
	if (debug[VcaCard(dev)+VcaChannel(dev)]&DebugAny)
	    printf("Packet 0x%x (0x%x) received out of order\n",
		   packet_number, lan2vca_packet_number);
#endif /* VCADebug */
#endif /* 0 */
	lan2vca_packet_number = packet_number;
    }

#if 0
    DEBUGprintf(dev, DebugIntr|DebugAudio,
		 ("dev 0x%x, Count 0x%x, Packet 0x%x, size 0x%x\n", dev,
		  lan2vca_packet_number, packet_number, len/sizeof(vcalocation)));
#endif /* 0 */
    return 0;
}
#endif /* TOKEN_RING_MODIFICATIONS */

#ifdef TOKEN_RING_MODIFICATIONS
int vca_setup_trhdr(channel, route)
struct VCAChannel *channel;
struct route *route;
{
    struct ifnet *ifp = route->ro_rt->rt_ifp;
    struct sockaddr *dst = &(route->ro_dst);
    extern struct lan_softc lan_softc[];
    u_char edst[LAN_L_ADDR];
    struct in_addr idst = ((struct sockaddr_in *)dst)->sin_addr;
    register struct lan_softc *lns = &lan_softc[ifp->if_unit];
    int usetrailers; /* not used */
    u_short rcf = 0; 	
    u_short rseg[LAN_MAX_BRIDGE];
    struct mbuf *m;
    int loop = 0;

    while (! tr_arpresolve(&lns->lns_ac, (struct mbuf *)0,
			   &idst, edst, &usetrailers, &rcf, rseg)) {
	OneMicroSecondDelay;
	if (++loop >= 1000000)
	    return -1;
    }

    m = lan_get_token_ring_header(lns, edst, rcf, rseg, LAN_UI_CMD, LAN_AUDIO);
    if (m->m_len > MAX_TOKEN_RING_HEADER_LENGTH)
	panic("token ring header too long");
    channel->trhdr.len = m->m_len;
    bcopy(mtod(m, char *), channel->trhdr.data, m->m_len);
    m_freem(m);
    if (channel->trhdr.priority != 0)
	channel->trhdr.data[0] = (char)((channel->trhdr.priority & 7)<<5);
    channel->ifnet = ifnet;
    return 0;
}
#endif /* TOKEN_RING_MODIFICATIONS */

vcaioctl(dev, cmd, addr, flag)
dev_t dev;
caddr_t addr;
{
    struct VCACard *card = &(vcacards[VcaCard(dev)]);
    struct VCAChannel *channel = &(card->channel[VcaChannel(dev)]);

    if ((cmd & 0xff)!=VCA_DEBUGING_OFF)
	DEBUGprintf(dev, DebugIoctl, ("%d %d: VCA IoCtl (0x%x) ",
				      VcaCard(dev), VcaChannel(dev), cmd));

    switch (cmd & 0xff) {
	case VCA_WRITE_IN_PROGRESS_P:
	    if (channel->out_start != channel->out_end ||
		card->basemem[channel->DAC_Enabled_Offset] != 0) {
		if (channel->out_start != channel->out_end &&
		    card->basemem[channel->DAC_Enabled_Offset] == 0)
		    vcaStartPlaying(dev);
		((short *)addr)[0] =  1;
	    } else
		((short *)addr)[0] = 0;
	    DEBUGprintf(dev, DebugIoctl, ("Write %sin Progress\n",
					  ((short *)addr)[0]==0 ? "not ":""));
	    break;
	case VCA_SET_RECORDING:
	    channel->in_start = channel->in_end = 0;
	    VcaFunctionRequest(card, VcaChannel(dev),
			       VCA_SET_RECORDING_OFFSET, ((short *)addr)[0]);
	    DEBUGprintf(dev, DebugIoctl, ("Set Recording %s\n",
					  ((short *)addr)[0]==0 ? "Off":"On"));
	    break;
	case VCA_SET_MULTIPLEXER:
	    VcaFunctionRequest(card, VcaChannel(dev),
			       VCA_SET_MULTIPLEXER_OFFSET, ((short *)addr)[0]);
#ifdef VCADebug
	    if (debug[VcaCard(dev)+VcaChannel(dev)]&DebugIoctl) {
		printf("Set Multiplexer to ");
		switch (((short *)addr)[0]) {
		    case Phone:
			printf("Phone\n");
			break;
		    case Tele:
			printf("Tele\n");
			break;
		    case Mic:
			printf("Mic/Speaker\n");
			break;
		    case ACWrap:
			printf("ACWrap\n");
			break;
		    default:
			printf("Invalid - %d\n", ((short *)addr)[0]);
		}
	    }
#endif /* VCADebug */
	    break;
	case VCA_SET_OFF_HOOK:
	    VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_OFF_HOOK_OFFSET,
			       ((short *)addr)[0]);
	    DEBUGprintf(dev, DebugIoctl,
			("Set %s Hook\n", ((short *)addr)[0]==0 ? "On":"Off"));
	    break;
	case VCA_SET_RECEIVE_GAIN:
	    VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_RECEIVE_GAIN_OFFSET,
			       ((short *)addr)[0]);
	    DEBUGprintf(dev, DebugIoctl,
			("Set Receive Gain 0x%x\n", ((short *)addr)[0]));
	    break;
	case VCA_SET_PLAYING:
	    if (((short *)addr)[0]) {
		vcaStartPlaying(dev);
	    } else {
		VcaFunctionRequest(card, VcaChannel(dev), VCA_SET_PLAYING_OFFSET, 0);
		((short *)addr)[0] = channel->out_end - channel->out_start;
		channel->out_start = channel->out_end = 0;
		DEBUGprintf(dev, DebugIoctl, ("Stop Playing\n"));
	    }
	    break;
	case VCA_DEBUGING_ON:
#ifdef VCADebug
	    DebugCount[VcaCard(dev)+VcaChannel(dev)] = 0;
	    debug[VcaCard(dev)+VcaChannel(dev)] = *addr;
	    printf("%d %d: VCA Debugging turned On (%x)\n",
		   VcaCard(dev), VcaChannel(dev), debug[VcaCard(dev)+VcaChannel(dev)]);
#else /* VCADebug */
	    printf("Kernel was not compiled with VCA debuging turned on\n");
#endif /* VCADebug */
	    break;
	case VCA_DEBUGING_OFF:
#ifdef VCADebug
	    debug[VcaCard(dev)+VcaChannel(dev)] = 0;
#endif /* VCADebug */
	    break;
	case VCA_DISCONNECT_CHANNEL:
	    channel->disconnect = 1;
	    /* Drop into wakeup */
	case VCA_WAKEUP:
	    /* Wakeup on all buffers. */
	    wakeup(channel->input);
	    wakeup(channel->output);
	    DEBUGprintf(dev, DebugIO|DebugIoctl,
			("Wakeup on: 0x%x, 0x%x\n",
			 channel->input, channel->output));
	    break;
	case VCA_INTERRUPT_PSP:
	    *card->iocr |= VCA_INT_PSP;
	    DEBUGprintf(dev, DebugIoctl, ("Interrupt PSP\n"));
	    break;
	case VCA_ENABLE_PSP_INTERRUPTS:
	    *card->iocr |= VCA_ENA_INT;
	    DEBUGprintf(dev, DebugIoctl, ("Enable PSP Interrupts\n"));
	    break;
	case VCA_DISABLE_PSP_INTERRUPTS:
	    if (*card->iocr & VCA_ENA_INT)
		*addr = 1;
	    else
		*addr = 0;
	    *card->iocr &= (0xff ^ VCA_ENA_INT);
	    DEBUGprintf(dev, DebugIoctl, ("Disable PSP Interrupts\n"));
	    break;
	case VCA_ENABLE_PSP:
	    *card->iocr |= VCA_ENA_PSP;
	    DEBUGprintf(dev, DebugIoctl, ("Enable PSP\n"));
	    break;
	case VCA_DISABLE_PSP:
	    if (*card->iocr & VCA_ENA_PSP)
		*addr = 1;
	    else
		*addr = 0;
	    *card->iocr &= (0xff ^ VCA_ENA_PSP);
	    DEBUGprintf(dev, DebugIoctl, ("Disable PSP\n"));
	    break;
	case VCA_LOAD: {
	    char save = *card->iocr;
	    DEBUGprintf(dev, DebugIoctl, ("Load Program/Data Memory\n"));
	    *card->iocr = VCA_DISABLE_ALL;
	    DEBUGprintf(dev, DebugLoad,
			("LoadIntoPSP: (%c, 0x%x, 0x%x)\n",
			 ((loadpsp *)addr)->type, ((loadpsp *)addr)->loc,
			 ((loadpsp *)addr)->data));
	    if (((loadpsp *)addr)->type == 'i') {
		if (((loadpsp *)addr)->loc > 0x4000) {
		    /* Correct for instruction mapped into high 4k of instruction memory. */
		    *card->iocr = (VCA_SELECT_INST_MEM_HIGH | VCA_SELECT_INST_MEM);
		    ((loadpsp *)addr)->loc -= 0x4000;
		} else
		    *card->iocr = VCA_SELECT_INST_MEM;
	    }
	    *(card->basemem+((loadpsp *)addr)->loc) = ((loadpsp *)addr)->data;
	    *card->iocr = save;
	    break;
	    }
	case VCA_SET_RW_PID:
	    channel->RWpid = ((short *)addr)[0];
	    DEBUGprintf(dev, DebugIoctl, ("RW PID: %d\n", ((short *)addr)[0]));
	    break;
	case VCA_SET_RI_PID:
	    channel->RIpid = ((short *)addr)[0];
	    DEBUGprintf(dev, DebugIoctl, ("RI PID: %d\n", ((short *)addr)[0]));
	    break;
	case VCA_WRITE_PROGRAM_ID:
	    bcopy(addr, card->basemem+VCA_PROGRAM_ID_OFFSET, VCA_PROGRAM_ID_SIZE*2);
#ifdef VCADebug
	    if (debug[VcaCard(dev)+VcaChannel(dev)]&DebugIoctl) {
		int x;
		printf("Write Program ID: '");
		for (x=0; x<VCA_PROGRAM_ID_SIZE*2; x++) printf("%c", addr[x]);
		printf("'\n");
	    }
#endif /* VCADebug */
	    break;
	case VCA_READ_PROGRAM_ID:
	    bcopy(card->basemem+VCA_PROGRAM_ID_OFFSET, addr, VCA_PROGRAM_ID_SIZE*2);
#ifdef VCADebug
	    if (debug[VcaCard(dev)+VcaChannel(dev)]&DebugIoctl) {
		int x;
		printf("Read Program ID: '");
		for (x=0; x<VCA_PROGRAM_ID_SIZE*2; x++) printf("%c", addr[x]);
		printf("'\n");
	    }
#endif /* VCADebug */
	    break;
	case VCA_SET_BYPASS:
	    VcaFunctionRequest(card, 0, VCA_SET_BYPASS_OFFSET, ((short *)addr)[0]);
	    DEBUGprintf(dev, DebugIoctl,
			("%s ByPass\n", (((short *)addr)[0])==0 ? "Clear":"Set"));
	    break;
	case VCA_SET_DCWRAP:
	    VcaFunctionRequest(card, 0, VCA_SET_DCWRAP_OFFSET, ((short *)addr)[0]);
	    DEBUGprintf(dev, DebugIoctl,
			("%s DCWrap\n", (((short *)addr)[0])==0 ? "Clear":"Set"));
	    break;
	case VCA_SET_SAMPLE_RATE:
	    VcaFunctionRequest(card, 0, VCA_SET_SAMPLE_RATE_OFFSET, ((short *)addr)[0]);
	    DEBUGprintf(dev, DebugIoctl,
			("Set Sample Rate: 0x%x\n", ((short *)addr)[0]));
	    break;
	case VCA_GET_SC2:
	    ((short *)addr)[0] = card->basemem[VCA_SHADOW_OF_SC2_OFFSET];
	    DEBUGprintf(dev, DebugIoctl, ("SC2: 0x%x\n", ((short *)addr)[0]));
	    break;
	case VCA_BLOCKING_READS:
	    channel->BlockingReads = ((short *)addr)[0];
	    DEBUGprintf(dev, DebugIoctl, ("%s Blocking Reads\n",
					  ((short *)addr)[0]==1 ? "Set":"Clear"));
	    break;
	case VCA_CROSS_CONNECT: {
	    dev_t dev2;
	    if (*((short *)addr) > u.u_lastfile) {
		DEBUGprintf(dev, DebugIoctl,
			    ("Invalid File Descriptor - arg: %d u.u_lastfile: %d\n",
			     *((short *)addr), u.u_lastfile));
		break;
	    }
#ifdef VFS
	    dev2 = ((struct vnode *)(u.u_ofile[*((short *)addr)]->f_data))->v_rdev;
#else /* VFS */
	    dev2 = ((struct inode *)(u.u_ofile[*((short *)addr)]->f_data))->i_rdev;
#endif /* VFS */

	    if (major(dev2) != VCAMajor) {
		if (channel->fp != -1) {
		    printf("VCA Error: Already connected to File descriptor %d\n",
			   channel->fp);
		} else {
		    struct file *fp;
		    channel->fp = *((short *)addr);
		    channel->connected = NULL;
		    DEBUGprintf(dev, DebugIoctl,
				("Cross Connect to File Descriptor %d\n",
				 channel->fp));

		    /* Taken from sys/sys_generic rwuio() */
		    GETF(fp, channel->fp);
#if 0
		    /* XXX Should do this test */
		    if ((fp->f_flag&(rw==UIO_READ ? FREAD : FWRITE)) == 0) {
			u.u_error = EBADF;
			return;
		    }
#endif /* 0 */
		    for (;;) {
			int signal, in_start, diff, result;
			struct uio uio;
			struct iovec iov;

			uio.uio_iov = &iov;
			uio.uio_iovcnt = 1;

			signal = sleep(channel->input, PUSER | PCATCH);
			in_start = channel->in_start;
			diff = channel->in_end-in_start;
			channel->in_start = channel->in_end = 0;

			if (diff != 0) {
			    /* Taken from sys/sys_generic.c write()*/
			    iov.iov_base = (char *)&(channel->input[in_start]);
			    iov.iov_len = diff*sizeof(vcalocation);
			    /* Taken from sys/sys_generic.c rwuio()*/
#ifdef VICE
			    struct inode *tInode;
#endif /* VICE */
			    int count;
			    uio.uio_resid = iov.iov_len;
			    uio.uio_segflg = UIO_SYSSPACE;
			    count = uio.uio_resid;
			    uio.uio_offset = fp->f_offset;
			    if (setjmp(&u.u_qsave)) {
				if (uio.uio_resid == count) {
				    if ((u.u_sigintr & sigmask(u.u_procp->p_cursig)) != 0)
					u.u_error = EINTR;
				    else
					u.u_eosys = RESTARTSYS;
				}
			    } else
				u.u_error = (*fp->f_ops->fo_rw)(fp, UIO_WRITE, &uio);
#ifdef VICE
			    if (rw == UIO_WRITE && (fp->f_flag & FFILL)) {
				tInode = (struct inode *) fp->f_data;
				if (tInode ->i_flag & IFILLWAIT) {
				    tInode->i_flag &= ~IFILLWAIT;
				    wakeup(&tInode->i_flag);
				}
			    }
#endif /* VICE */
			    result = count - uio.uio_resid;
#ifdef AFS
			    /* special hack for making read work like readdir */
			    if (fp->f_type == DTYPE_VNODE) {
				afs_AdvanceFD(fp, uio.uio_offset, result);
			    } else
				fp->f_offset += result;
#else /* AFS */
			    fp->f_offset += result;
#endif /* AFS */
			    /* XXX need a test for result non-zero */
			    if (result == -1)
				printf("     %d %d -> FP(%d) Swap (0x%x, 0x%x) Error: %d\n",
				       VcaCard(dev), VcaChannel(dev),
				       channel->fp, &(channel->input[in_start]), diff, u.u_error);
#ifdef VCADebug
			    else if (debug[VcaCard(dev)+VcaChannel(dev)]&DebugIoctl)
				printf("     %d %d -> FP(%d) Swap (0x%x, 0x%x) 0x%x\n",
				       VcaCard(dev), VcaChannel(dev),
				       channel->fp, &(channel->input[in_start]),
				       diff, result);
#endif /* VCADebug */
			}
			if (signal || channel->disconnect) {
			    channel->disconnect = 0;
			    channel->fp = -1;
			    return 0; }
		    }
		}
	    } else {
		if (channel->connected != NULL)
		    printf("Error connected to (%d, %d)\n",
			   VcaCard(channel->connected->dev),
			   VcaChannel(channel->connected->dev));
		else {
		    DEBUGprintf(dev, DebugIoctl, ("Cross Connect to (%d, %d)\n",
					     VcaCard(dev2), VcaChannel(dev2)));
		    channel->connected = &(vcacards[VcaCard(dev2)].channel[VcaChannel(dev2)]);
		    channel->connected->from = channel;
		}
	    }
	    break;
	    }
	case VCA_RESET_CHANNEL:
	    DEBUGprintf(dev, DebugIoctl, ("Reset channel\n"));
	    resetChannel(dev);
	    break;
	case VCA_GENERATE_HOST_INTERRUPT:
	    DEBUGprintf(dev, DebugIoctl, ("Generate host interrupt\n"));
	    VcaFunctionRequest(card, 0, VCA_GENERATE_HOST_INTERRUPT_OFFSET, 0);
	    break;
	case VCA_CRADLE_ON_HOOK_P: {
	    int loop = 0;
	    short vca_int = card->basemem[VCA_NUMBER_OF_INTERRUPTS];
	    DEBUGprintf(dev, DebugIoctl, ("Test Cradle on Hook\n"));
	    /* Generate an interrupt so the VCA updates the shadow copy of SS1 */
	    *card->iocr |= VCA_INT_PSP;
	    while (vca_int == card->basemem[VCA_NUMBER_OF_INTERRUPTS] &&
		   loop<100) {
		loop++;
		OneMicroSecondDelay;
	    }
	    ((short *)addr)[0] = (card->basemem[VCA_SHADOW_OF_SS1_OFFSET] & 0x0080) ? 1:0;
	    break;
	    }
	case VCA_SET_VCA2LAN:
#ifdef TOKEN_RING_MODIFICATIONS
	    {
	    struct a { int socket, device; } *a = (struct a *)addr;
	    struct socket *socket;
	    struct inpcb *inp;
	    DEBUGprintf(dev, DebugIoctl, ("Set Vca2Lan %d\n", a->socket));
	    channel->vca2lan = -1;
	    if (a->socket < 0)
		break;
	    if (a->socket > u.u_lastfile) {
		DEBUGprintf(dev, DebugIoctl,
			    ("Invalid File Descriptor - arg: %d u.u_lastfile: %d\n",
			     a->socket, u.u_lastfile));
		break;
	    }
	    socket = (struct socket *)(u.u_ofile[a->socket]->f_data);
	    inp = sotoinpcb(socket);
	    if (inp) {
		channel->vca2lan = vca_setup_trhdr(channel, &(inp->inp_route));
		channel->vca2lan_device = a->device;
	    }
	    }
#else /* TOKEN_RING_MODIFICATIONS */
	    printf("vcaioclt: VCA_SET_VCA2LAN failure: Kernel built without Token Ring Driver Modifications");
#endif /* TOKEN_RING_MODIFICATIONS */
	    break;
#ifdef DEBUG
	case VCA_LAN_DEBUG_ON:
	    DEBUGprintf(dev, DebugIoctl, ("Lan Debug On\n"));
	    lan_set_debug(*((short *)addr));
	    break;
	case VCA_LAN_DEBUG_OFF:
	    DEBUGprintf(dev, DebugIoctl, ("Lan Debug Off\n"));
	    lan_set_debug(0);
	    break;
#else /* DEBUG */
	case VCA_LAN_DEBUG_ON:
	case VCA_LAN_DEBUG_OFF:
	    printf("Kernel was not compiled with VCA LAN debuging enabled\n");
	    break;
#endif /* DEBUG */
	case VCA_DUMP_VCA2LAN_TIMESTAMPS:
#ifdef TOKEN_RING_MODIFICATIONS
	    {
	    struct a { int len; char *buf; } *a = (struct a *)addr;
	    int len;
	    DEBUGprintf(dev, DebugIoctl, ("Dump vca2lan timestamps\n"));
	    len = channel->vca2lan_packet_number*sizeof(struct timeval)*2;
	    if (len > vcabytebufsize)
		len = vcabytebufsize - vcabytebufsize%(2*sizeof(struct timeval));
	    if (a->len > len)
		a->len = len;
	    copyout((char *)channel->input, a->buf, a->len);
	    DEBUGprintf(dev, DebugIoctl, ("Total number of vca2lan calls made: %d\n",
					  channel->vca2lan_packet_number));
	    }
#else /* TOKEN_RING_MODIFICATIONS */
	    printf("vcaioclt: VCA_DUMP_VCA2LAN_TIMESTAMPS failure: Kernel built without Token Ring Driver Modifications");
#endif /* TOKEN_RING_MODIFICATIONS */
	    break;
	case VCA_RETURN_DEVICE_NUMBER:
	    *((short *)addr) = (short)makedev(VCAMajor, channel->dev);
	    DEBUGprintf(dev, DebugIoctl, ("Return device number: 0x%x\n", *((short *)addr)));
	    break;
	case VCA_SET_PACKET_PRIORITY:
#ifdef TOKEN_RING_MODIFICATIONS
	    channel->trhdr.priority = *((int *)addr);
	    DEBUGprintf(dev, DebugIoctl, ("Set Packet Priority: %d\n",
					  channel->trhdr.priority));
#else /* TOKEN_RING_MODIFICATIONS */
	    printf("vcaioclt: VCA_SET_PACKET_PRIORITY failure: Kernel built without Token Ring Driver Modifications");
#endif /* TOKEN_RING_MODIFICATIONS */
	    break;
	case VCA_SET_QUEUE_PRIORITY:
#ifdef TOKEN_RING_MODIFICATIONS
	    channel->trhdr.queue = *((int *)addr);
	    if (channel->trhdr.queue < 0)
		channel->trhdr.queue = 0;
	    if (channel->trhdr.queue >= NUMBER_OF_SEND_QUEUES)
		channel->trhdr.queue = NUMBER_OF_SEND_QUEUES - 1;
	    DEBUGprintf(dev, DebugIoctl, ("Set Queue Priority: %d\n",
					  channel->trhdr.priority));
#else /* TOKEN_RING_MODIFICATIONS */
	    printf("vcaioclt: VCA_SET_QUEUE_PRIORITY failure: Kernel built without Token Ring Driver Modifications");
#endif /* TOKEN_RING_MODIFICATIONS */
	    break;
	case VCA_RESIZE_RECORD_BUFFERS: {
	    int x, newsize = *((int *)addr);
	    vcalocation *loc = &(card->basemem[0x77]);
	    DEBUGprintf(dev, DebugIoctl, ("Resize Record Buffers: 0x%x\n", newsize));
	    /* Starting at 0x77, recording buffer start locations are laid out. At 0x7b the first playback buffer is started. At location 0x7f the last playback buffer's end is stored. The start of the next buffer in the list is the end of the last buffer in the list. The list of buffers is as follows (everything is double buffered): record buffers for channel 1, record buffers for channel 2, playback buffers for channel 1, playback buffers for channel 2 */
#ifdef DOCUMENTATION_ONLY
	    char loc77[] = "start of channel 1 record buffer A";
	    char loc78[] = "start of channel 1 record buffer B, end of channel 1 record buffer A";
	    char loc79[] = "start of channel 2 record buffer A, end of channel 1 record buffer B";
	    char loc7a[] = "start of channel 2 record buffer B, end of channel 2 record buffer A";
	    char loc7b[] = "start of channel 1 playback buffer A, end of channel 2 record buffer B";
	    char loc7c[] = "start of channel 1 playback buffer B, end of channel 1 playback buffer A";
	    char loc7d[] = "start of channel 2 playback buffer A, end of channel 1 playback buffer B";
	    char loc7e[] = "start of channel 2 playback buffer B, end of channel 2 playback buffer A";
	    char loc7f[] = "end of channel 2 playback buffer B";
#endif /* DOCUMENTATION_ONLY */
	    if (newsize%16 != 0)
		newsize -= newsize%16;
	    if (newsize < 16 || newsize > 0x80)
		break;
	    for (x=0; x<4; x++)
		loc[x+1] = loc[x] + newsize;
	    VcaFunctionRequest(card, 0, VCA_REINITIALIZE_BUFFERS_OFFSET, 0);
	    break;
	    }
	case VCA_DROP_PACKETS_FROM_LAN:
#ifdef TOKEN_RING_MODIFICATIONS
	    channel->drop_packets = *((int *)addr);
	    if (channel->drop_packets < 0
		|| channel->drop_packets > VCA_MAX_VALUE_DROP_PACKETS_FROM_LAN)
		channel->drop_packets = VCA_MAX_VALUE_DROP_PACKETS_FROM_LAN;
	    if (channel->drop_packets == VCA_MAX_VALUE_DROP_PACKETS_FROM_LAN)
		lan_drop_audio_packets_before_copy = 1;
	    DEBUGprintf(dev, DebugIoctl, ("Drop packets from lan: %d\n",
					  channel->drop_packets));
#else /* TOKEN_RING_MODIFICATIONS */
	    printf("vcaioclt: VCA_DROP_PACKETS_FROM_LAN failure: Kernel built without Token Ring Driver Modifications");
#endif /* TOKEN_RING_MODIFICATIONS */
	    break;
	case VCA_SET_VCA2LAN_PACKET_TOTAL_SIZE:
#ifdef TOKEN_RING_MODIFICATIONS
	    channel->packet_total_size = *((int *)addr);
	    DEBUGprintf(dev, DebugIoctl, ("Set packets total size: 0x%x\n",
					  channel->packet_total_size));
#else /* TOKEN_RING_MODIFICATIONS */
	    printf("vcaioclt: VCA_SET_VCA2LAN_PACKET_TOTAL_SIZE failure: Kernel built without Token Ring Driver Modifications");
#endif /* TOKEN_RING_MODIFICATIONS */
	    break;
	case VCA_SEND_TOKEN_RING_PACKET:
#ifdef TOKEN_RING_MODIFICATIONS
	    DEBUGprintf(dev, DebugIoctl, ("Send Token Ring Packet"));
	    if (channel->vca2lan >= 0) {
		struct a { int len; char *buf; } *a = (struct a *)addr;
		struct mbuf *m;
		DEBUGprintf(dev, DebugIoctl,
			    ("Sending packet of data length %d", a->len));
		MGET(m, M_DONTWAIT, MT_DATA);
		if (m != 0) {
		    char *data = mtod(m, char *);
		    /* Copy token ring header to front of packet to send */
		    bcopy(channel->trhdr.data, data, channel->trhdr.len);
		    data += channel->trhdr.len;
		    m->m_len = channel->trhdr.len;
		    /* Copy user data into packet */
		    copyin(a->buf, data, a->len);
		    DEBUGprintf(dev, DebugIoctl,
				("First 4 data bytes: 0x%x:%x:%x:%x\n",
				 data[0], data[1], data[2], data[3]));
		    m->m_len += a->len;
		    lan_output_llc_work(channel->ifnet, channel->trhdr.queue, m);
		}
	    }
#else /* TOKEN_RING_MODIFICATIONS */
	    printf("vcaioclt: VCA_SEND_TOKEN_RING_PACKET failure: Kernel built without Token Ring Driver Modifications");
#endif /* TOKEN_RING_MODIFICATIONS */
	    break;
	default:
	    printf("vca: Unknown command %d\n", (int)(cmd & 0xff));
    }
    return 0;
}

#if 0 /* This is a random thought process. I think I was trying to determine the value of the constants for the number of bytes per page so that I could use the same number as the size of the io memory page size to use -- pasieka */
/* CLBYTES is defined in param.h as: */
#define	CLBYTES		(CLSIZE*NBPG)
/* NBPG and CLSIZE are defined in machparam.h as: */
#define        NBPG    2048            /* bytes/page  */
#define        CLSIZE          1
#endif /* 0 */

#ifdef TOKEN_RING_MODIFICATIONS
vca2lan(card, channel, Args, packet_number)
struct VCACard *card;
struct VCAChannel *channel;
struct ADCArgs *Args;
int packet_number;
{
    int result = 0;
    int bytesize, sample_len;
    char *start;
    struct mbuf *m;
#ifdef VCADebug
    int mbufs = 0;
#endif /* VCADebug */

    if (channel->packet_total_size > 0)
	start = (char *)&channel->input;
    else
	start = (char *)&(card->basemem[Args->start]);

    MGET(m, M_DONTWAIT, MT_DATA);
    if (m != 0) {
	char *data = mtod(m, char *);
	sample_len = Args->size;
	int mbuflen;
#ifdef VCADebug
	mbufs = 1;
#endif /* VCADebug */
	/* Copy token ring header to front of packet to send */
	bcopy(channel->trhdr.data, data, channel->trhdr.len);
	data += channel->trhdr.len;
	m->m_len = channel->trhdr.len;
	/* Put VCA device number and packet count into mbuf */
	((short	*)data)[0] = (short)channel->vca2lan_device;
	bcopy((char *)&packet_number, &data[sizeof(short)], sizeof(int));
	data += sizeof(short) + sizeof(int);
	m->m_len += sizeof(short) + sizeof(int);
	if (channel->packet_total_size > 0) {
	    bytesize = channel->packet_total_size - m->m_len;
	    if (bytesize < 0)
		bytesize = 0;
	    else if (bytesize > 2048 - m->m_len)
		bytesize = 2048 - m->m_len;
	} else
	    bytesize = sample_len*sizeof(vcalocation);
#ifdef IOMEM
	if (iomem_exists && channel->drop_packets == VCA_SEND_USING_IOMEM) {
	    struct mbuf *m_iomem = (struct mbuf *)(IOMEM_ADDR + (iomem_alloc)*2048);
	    m_iomem->m_next = 0;
	    m_iomem->m_off = MMINOFF;
	    m_iomem->m_len = bytesize;
	    m_iomem->m_type = MT_DATA;
	    m->m_next = m_iomem;
	} else
#endif /* IOMEM */
	{
	    /* Copy data into mbuf */
	    if (bytesize <= (mbuflen = MLEN - m->m_len)) {
		/* Data fits into same mbuf as header */
		if (channel->drop_packets < VCA_DROP_WITHOUT_COPY)
		    bcopy(start, data, bytesize);
		m->m_len += bytesize;
	    } else {
		/* Header plus data fills more than one mbuf */
		struct mbuf *m0 = NULL, *m_lnk = m;
		/* Copy to fill first mbuf */
		if (channel->drop_packets < VCA_DROP_WITHOUT_COPY)
		    bcopy(start, data, mbuflen);
		start += mbuflen;
		bytesize -= mbuflen;
		m->m_len += mbuflen;
		do {
		    MGET(m0, M_DONTWAIT, MT_DATA);
		    if (m0 == 0) {
			m_freem(m);
			m = NULL;
			break;
		    }
#ifdef VCADebug
		    mbufs++;
#endif /* VCADebug */
		    m_lnk->m_next = m0;
		    m_lnk = m0;
		    if (bytesize <= MLEN) {
			if (channel->drop_packets < VCA_DROP_WITHOUT_COPY)
			    bcopy(start, mtod(m0, char *), bytesize);
			m0->m_len = bytesize;
			break;
		    } else {
			if (channel->drop_packets < VCA_DROP_WITHOUT_COPY)
			    bcopy(start, mtod(m0, char *), MLEN);
			start += MLEN;
			bytesize -= MLEN;
			m0->m_len = MLEN;
			continue;
		    }
		} while (bytesize > 0);
	    }
	}
    } else
	printf("vca2lan: No more Mbufs!\n");

    /* XXX Need to deal better with this. */
#ifdef CARD_OVERRUN
    if (Args->overrun > 0)
	printf("     %d %d Card Overrun of %d samples\n",
	       VcaCard(channel->dev), VcaChannel(channel->dev), Args->overrun);
#endif /* CARD_OVERRUN */
    /* Release VCA buffer. */
    Args->size = 0;

#define SpecialMemoryAddressOnVcaCard() \
(0xfff&((long)start+bytesize-sample_len*sizeof(vcalocation)))/sizeof(vcalocation)

    if (m == 0) {
	DEBUGprintf(channel->dev, DebugIntr|DebugAudio,
		    ("     %d %d: MBuf for data (start 0x%x len 0x%x) not available\n",
		     VcaCard(channel->dev), VcaChannel(channel->dev),
		     SpecialMemoryAddressOnVcaCard(), sample_len)); }
    else {
	result = lan_output_llc_work(channel->ifnet, channel->trhdr.queue, m);
	DEBUGprintf(channel->dev, DebugIntr|DebugAudio,
		    ("     %d %d: Sending %d MBuf%s (start 0x%x len 0x%x) as packet 0x%x%s\n",
		     VcaCard(channel->dev), VcaChannel(channel->dev),
		     mbufs, mbufs>1 ? "s":"", 
		     SpecialMemoryAddressOnVcaCard(), sample_len,
		     packet_number, result ? " *** FAILED ***":"")); }
    return 0;
}
#endif /* TOKEN_RING_MODIFICATIONS */

#ifdef TOKEN_RING_MODIFICATIONS
#define	vcaReadIn(card,	enable_bit, channel, args)			\
    if ((enables & enable_bit) != 0 && args->size != 0) {		\
	if (channel->vca2lan >=	0) {					\
	    /* XXX timestamp before and	after the call to vca2lan */	\
	    int	priority, packet_number, stamp;				\
	    priority = splvca();					\
	    packet_number = channel->vca2lan_packet_number;		\
	    if (channel->vca2lan_packet_number == 0x7fffffff)		\
		channel->vca2lan_packet_number = 0;			\
	    else							\
		channel->vca2lan_packet_number++;			\
	    splx(priority);						\
	    stamp = packet_number*sizeof(struct	timeval)*2;		\
	    if (stamp+2*sizeof(struct timeval) > vcabytebufsize)	\
		stamp =	(int)channel->input;				\
	    else							\
		stamp += (int)channel->input;				\
	    microtime((struct timeval *)stamp);				\
	    stamp += sizeof(struct timeval);				\
	    vca2lan(card, channel, args, packet_number);		\
	    microtime((struct timeval *)stamp);				\
	} else /* XXX should special case channel->connected also */	\
	    vcareadin(card, channel, args);				\
	if (channel->connected != NULL)					\
	    vcaConnectedCopy(channel, channel->connected);		\
    }
#else /* TOKEN_RING_MODIFICATIONS */
#define	vcaReadIn(card,	enable_bit, channel, args)			\
    if ((enables & enable_bit) != 0 && args->size != 0) {		\
	    vcareadin(card, channel, args);				\
	if (channel->connected != NULL)					\
	    vcaConnectedCopy(channel, channel->connected);		\
    }

#endif /* TOKEN_RING_MODIFICATIONS */

vcareadin(card, channel, Args)
struct VCACard *card;
struct VCAChannel *channel;
struct ADCArgs *Args;
{
    vcalocation *start = &(card->basemem[Args->start]);
    int p_start, p_end;
    int end, p_overrun;

    /* There is no issue of synchonization between the VCA's DMA and the processor. The processor is three orders of magnitued faster than the VCA's DMA. On the other hand, the VCA's PSP is within an order of magnitude as fast as the processor. The input is double buffered on the VCA card to avoid the VCA overwriting what the processor is copying out to the kernel buffer. Of course it is possible that the processor gets so far behind that there is an overrun on the card. This is dealt with on the card. */
    /* Determine the start and end of where to place samples in the processor ring buffer. */
    DEBUGprintf(channel->dev, DebugIntr|DebugRead,
		 ("     %d %d: Samples: (0x%x, 0x%x)\n",
		  VcaCard(channel->dev), VcaChannel(channel->dev),
		  (0xfff&(long)start)/sizeof(vcalocation), Args->size));
    p_start = channel->in_end;
    p_end = p_start + Args->size;
    DEBUGprintf(channel->dev, DebugIntr|DebugRead,
		 ("     %d %d: Start: 0x%x\n",
		  VcaCard(channel->dev), VcaChannel(channel->dev), p_start));
    if (p_end > vcabufsize) {
	/* Read in bytes to to fill the processor buffer to the end. */
	DEBUGprintf(channel->dev, DebugIntr|DebugRead,
		    ("     %d %d: Fill to the end\n",
		     VcaCard(channel->dev), VcaChannel(channel->dev)));
	bcopy(start, channel->input+p_start, (vcabufsize-p_start)*sizeof(vcalocation));
	start += (vcabufsize-p_start);
	p_start = 0;
	p_end = p_end - vcabufsize;
    }
    /* Read in bytes into processor buffer. */
    DEBUGprintf(channel->dev, DebugIntr|DebugRead,
		 ("     %d %d: Fill to: 0x%x\n",
		  VcaCard(channel->dev), VcaChannel(channel->dev), p_end));
    bcopy(start, channel->input+p_start, (p_end-p_start)*sizeof(vcalocation));

    /* XXX Need to deal better with this. */
    if (Args->overrun > 0)
	printf("     %d %d Card Overrun of %d samples\n",
	       VcaCard(channel->dev), VcaChannel(channel->dev), Args->overrun);
    /* Release VCA buffer. */
    Args->size = 0;

    /* If you are changing this, use a piece of paper to figure what the original is doing first. Determine if the number of samples received plus the nuber of samples already in the buffer is larger than the size of the buffer. All numbers are in samples not bytes. */
    if ((end = channel->in_end)<channel->in_start)
	/* The buffer is wrapped. Adjust the end pointer so that the relative distance between start and end is preserved but end will now be larger than start. */
	end += vcabufsize;
    /* The number of samples in the buffer is the end plus the number just received minus the start. If this number minus the size of the buffer is still positive then an overrun condition exists. */
    p_overrun = ((end + Args->size) - channel->in_start) - vcabufsize;
    if (p_overrun > 0) {
	DEBUGprintf(channel->dev, DebugIntr|DebugRead,
		    ("     %d %d: Overrun of %d bytes\n",
		     VcaCard(channel->dev), VcaChannel(channel->dev),
		     p_overrun*sizeof(vcalocation)));
	channel->in_start = p_end+1;
    }
    channel->in_end = p_end;
    if (channel->in_end == vcabufsize)
	channel->in_end = 0;
    /* Wakeup any processes waiting for input. */
    DEBUGprintf(channel->dev, DebugIntr|DebugRead,
		 ("     %d %d Wakeup on 0x%x\n",
		  VcaCard(channel->dev), VcaChannel(channel->dev), channel->input));
    wakeup(channel->input);
    if (channel->RWpid != 0 && p_overrun+vcabufsize>500) {
	/* Signal the user process that more then 500 samples are available. */
	struct proc *p = pfind(channel->RWpid);
	if (p) {
	    DEBUGprintf(channel->dev, DebugSigs,
			("%d %d: Data available\n",
			 VcaCard(channel->dev), VcaChannel(channel->dev)));
	    psignal(p, SIGIO);
	}
    }

    return 0;
}

vcaConnectedCopy(in, out)
struct VCAChannel *in, *out;
{
    int diff = in->in_end-in->in_start;
    if (out->out_end+diff>vcabufsize)
	printf("%d %d: vcaConnectedCopy Swap Error -> %d %d\n",
	       VcaCard(in->dev), VcaChannel(in->dev), VcaCard(out->dev), VcaChannel(out->dev));
    else {
	DEBUGprintf(in->dev, DebugIntr,
		    ("     %d %d -> %d %d Swap (0x%x, 0x%x): 0x%x\n",
		     VcaCard(in->dev), VcaChannel(in->dev), VcaCard(out->dev), VcaChannel(out->dev),
		     &(in->input[in->in_start]), &(out->output[out->out_end]), diff));
	bcopy(&(in->input[in->in_start]), &(out->output[out->out_end]),
	      diff*sizeof(vcalocation));
	out->out_end += diff;
	in->in_start = in->in_end = 0;
    }

    return 0;
}

/* Move the data into the card's buffer if the current buffer is free and there is data to move and the size of the card's buffer is non-zero. */
#define	vcaWriteOut(card, channel, Args)		    \
{   if ((Args)->ReadyToWrite ==	0 &&			    \
	 (channel)->out_start != (channel)->out_end &&	    \
	 (Args)->end - (Args)->start > 0)		    \
	vcawriteout((card), (channel), (Args));	}

vcawriteout(card, channel, Args)
struct VCACard *card;
struct VCAChannel *channel;
struct DACArgs *Args;
{
    short size = Args->end - Args->start;

    DEBUGprintf(channel->dev, DebugIntr|DebugWrite,
		 ("     %d %d: vStart, vSize: (0x%x, 0x%x)",
		  VcaCard(channel->dev), VcaChannel(channel->dev), Args->start, size));
    if (size > channel->out_end - channel->out_start) {
	size = channel->out_end - channel->out_start;
	DEBUGprintf(channel->dev, DebugIntr|DebugWrite, (" New size: 0x%x\n", size));
    } else
	DEBUGprintf(channel->dev, DebugIntr|DebugWrite, ("\n"));

    bcopy(&(channel->output[channel->out_start]),
          &(card->basemem[Args->start]), size*sizeof(vcalocation));

    channel->out_start += size;
    DEBUGprintf(channel->dev, DebugIntr|DebugWrite,
		 ("     %d %d: pStart, pEnd: (0x%x, 0x%x) ",
		  VcaCard(channel->dev), VcaChannel(channel->dev),
		  channel->out_start, channel->out_end));

    Args->ReadyToWrite = size;

    if (channel->out_start == channel->out_end) {
	if (channel->RWpid != 0) {
	    struct proc *p = pfind(channel->RWpid);
	    if (p) {
		DEBUGprintf(channel->dev, DebugSigs,
			    ("     %d %d: Write kernel buffer empty\n",
			     VcaCard(channel->dev), VcaChannel(channel->dev)));
		psignal(p, SIGIO);
	    }
	}
	wakeup(channel->output);
	channel->out_start = channel->out_end = 0;
	DEBUGprintf(channel->dev, DebugIntr|DebugWrite, ("Done\n"));
    } else {
	DEBUGprintf(channel->dev, DebugIntr|DebugWrite, ("Another\n"));
    }

    /* Inform VCA data available for DMA if the channel is not active. */
    if (card->basemem[channel->DAC_Enabled_Offset] == 0)
	vcaStartPlaying(channel->dev);

    return 0;
}

vcaintr()
{
    static int in_service = 0;
    int dev;
    int priority;
#ifdef VCADebug
    int device;
#endif /* VCADebug */
    struct VCACard *card;
    struct VCA_Arguments *Args;
    int enables;
    struct proc *p;

    /* Determine the vca which interrupted. */
    for (dev=0; dev<NVCA; dev++) {
	if (vcacards[dev].iocr != NULL) {
	    if ((*(vcacards[dev].iocr) & (char)0x80) != 0) {
#ifdef VCADebug
		device = dev<<1;
		if ((debug[VcaCard(device)] & DebugIntr) || debug[VcaCard(device)+1] & DebugIntr)
		    printf("VCA%d Interrupt: ", dev);
#endif /* VCADebug */
		break;
	    }
	}
    }
    if (dev == NVCA)
	return INT_NOT_MINE;

    priority = splvca();
    if (++in_service > 1) {
	splx(priority);
	return INT_SERVICED;
    }
    splx(priority);

    card = &(vcacards[dev]);

    again:
    do {
	Args = (struct VCA_Arguments *)&(card->basemem[VCA_INTERRUPT_ARGUMENTS_OFFSET]);
	enables = card->basemem[VCA_SHADOW_OF_SC2_OFFSET];

#ifdef VCADebug
	if ((debug[VcaCard(device)] & DebugIntr) || debug[VcaCard(device)+1] & DebugIntr) {
	    printf("%x\n", enables);
	    if (card->basemem[VCA_FUNCTION_REQUEST_DEBUG_OFFSET] != CompletedHostFunctionRequests) {
		CompletedHostFunctionRequests = card->basemem[VCA_FUNCTION_REQUEST_DEBUG_OFFSET];
		printf("     Function request %d completed\n",
		       CompletedHostFunctionRequests);
	    }
	}
#endif /* VCADebug */

	vcaReadIn(card, ADC_CH1, (&(card->channel[0])), (&(Args->ADC1)));
	vcaReadIn(card, ADC_CH2, (&(card->channel[1])), (&(Args->ADC2)));

	vcaWriteOut(card, &(card->channel[0]), &(Args->DAC1));
	vcaWriteOut(card, &(card->channel[1]), &(Args->DAC2));

	if (card->basemem[VCA_PH1_RI_OFFSET] == 1) {
	    DEBUGprintf(device, DebugSigs,
			("     %d 0: Leading edge of Ring Indicator for phone line 1\n",
			 dev));
	    card->basemem[VCA_PH1_RI_OFFSET] = 0;
	    if (card->channel[0].RIpid && (p = pfind(card->channel[0].RIpid)))
		psignal(p, SIGALRM);
	} else if (card->basemem[VCA_PH1_RI_OFFSET] == 3) {
	    DEBUGprintf(device, DebugSigs,
			("     %d 0: Trailing edge of Ring Indicator for phone line 1\n",
			 dev));
	    card->basemem[VCA_PH1_RI_OFFSET] = 0;
	    if (card->channel[0].RIpid && (p = pfind(card->channel[0].RIpid)))
		psignal(p, SIGALRM);
	}
	if (card->basemem[VCA_PH2_RI_OFFSET] == 1) {
	    DEBUGprintf(device|1, DebugSigs,
			("     %d 1: Leading edge of Ring Indicator for phone line 2\n",
			 dev));
	    card->basemem[VCA_PH2_RI_OFFSET] = 0;
	    if (card->channel[1].RIpid && (p = pfind(card->channel[1].RIpid)))
		psignal(p, SIGALRM);
	} else if (card->basemem[VCA_PH2_RI_OFFSET] == 3) {
	    DEBUGprintf(device|1, DebugSigs,
			("     %d 1: Trailing edge of Ring Indicator for phone line 2\n",
			 dev));
	    card->basemem[VCA_PH2_RI_OFFSET] = 0;
	    if (card->channel[1].RIpid && (p = pfind(card->channel[1].RIpid)))
		psignal(p, SIGALRM);
	}

    } while ((*((char *)vcastd[dev]) & (char)0x80) != 0);

    priority = splvca();
    if (--in_service > 0) {
	splx(priority);
	goto again;
    }
    splx(priority);

    return INT_SERVICED;
}

int vcaselect(dev, rw)
dev_t dev;
int rw;
{
    struct VCAChannel *channel = &(vcacards[VcaCard(dev)].channel[VcaChannel(dev)]);
    int result = 1;
    switch (rw) {
	case FREAD:
	    if (channel->in_start == channel->in_end)
		result = 0;
	    DEBUGprintf(channel->dev, DebugAny,
			("vcaselect read: %s\n", result ? "True":"False"));
	    break;

	case FWRITE:
	    if ((vcabufsize - channel->out_end)*sizeof(vcalocation) <= 0)
		result = 0;
	    DEBUGprintf(channel->dev, DebugAny,
			("vcaselect write: %s\n", result ? "True":"False"));
	    break;
    }
    return(result);
}

#endif	/* NVCA > 0 */
