/*-
 * Copyright (c) 1990 William Jolitz.
 * Copyright (c) 1991 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)npx.c	7.2 (Berkeley) 5/12/91
 */
static char rcsid[] = "$Header: /usr/bill/working/sys/i386/isa/RCS/npx.c,v 1.2 92/01/21 14:34:27 william Exp $";
#include "npx.h"
#if NNPX > 0

#include "param.h"
#include "systm.h"
#include "conf.h"
#include "file.h"
#include "proc.h"
#include "machine/cpu.h"
#include "machine/pcb.h"
#include "machine/trap.h"
#include "ioctl.h"
#include "machine/specialreg.h"
#include "i386/isa/isa_device.h"
#include "i386/isa/isa.h"

/*
 * 387 and 287 Numeric Coprocessor Extension (NPX) Driver.
 */

#ifdef	__GNUC__

#define	disable_intr()		__asm("cli")
#define	enable_intr()		__asm("sti")
#define	fldcw(addr)		__asm("fldcw %0" : : "m" (*addr))
#define	fnclex()		__asm("fnclex")
#define	fninit()		__asm("fninit")
#define	fnsave(addr)		__asm("fnsave %0" : "=m" (*addr) : "0" (*addr))
#define	fnstcw(addr)		__asm("fnstcw %0" : "=m" (*addr) : "0" (*addr))
#define	fnstsw(addr)		__asm("fnstsw %0" : "=m" (*addr) : "0" (*addr))
#define	fp_divide_by_0()	__asm("fldz; fld1; fdiv %st,%st(1); fwait")
#define	frstor(addr)		__asm("frstor %0" : : "m" (*addr))
#define	start_emulating()	__asm("smsw %%ax; orb %0,%%al; lmsw %%ax" \
				      : : "n" (CR0_TS) : "ax")
#define	stop_emulating()	__asm("clts")

#else	/* not __GNUC__ */

void	disable_intr	__P((void));
void	enable_intr	__P((void));
void	fldcw		__P((caddr_t addr));
void	fnclex		__P((void));
void	fninit		__P((void));
void	fnsave		__P((caddr_t addr));
void	fnstcw		__P((caddr_t addr));
void	fnstsw		__P((caddr_t addr));
void	fp_divide_by_0	__P((void));
void	frstor		__P((caddr_t addr));
void	start_emulating	__P((void));
void	stop_emulating	__P((void));

#endif	/* __GNUC__ */

int	npxprobe(), npxattach(), npxintr();
struct	isa_driver npxdriver = {
	npxprobe, npxattach, "npx",
};

struct proc *npxproc;		/* process who owns device, otherwise zero */
static int npxexists;
unsigned npx0mask;
static unsigned npx_emulator;
static unsigned npx_ex16;
static unsigned npx_irq13;
static unsigned npx_traps_while_probing;
static volatile unsigned probing_npx_trap;

/*
 * Probe routine - look device, otherwise set emulator bit
 */
npxprobe(dvp)
	struct isa_device *dvp;
{
	int control;
	int status;

#ifdef lint
	npxintr();
#endif
	/*
	 * Prepare to trap all ESC (i.e., NPX) instructions and all WAIT
	 * instructions.  We must set the CR0_MP bit and use the CR0_TS
	 * bit to control the trap, because setting the CR0_EM bit does
	 * not cause WAIT instructions to trap.  It's important to trap
	 * WAIT instructions - otherwise the "wait" variants of no-wait
	 * control instructions would degenerate to the "no-wait" variants
	 * after FP context switches but work correctly otherwise.  It's
	 * particularly important to trap WAITs when there is no NPX -
	 * otherwise the "wait" variants would always degenerate.
	 */
	load_cr0(rcr0() | CR0_MP | CR0_NE);

	stop_emulating();
	fninit();		/* initialize device, if it exists */

	/* check for a proper status of zero */
	status = 0x5a5a;	
	fnstsw(&status);

	if ((status & 0xb8ff) == 0) {

		/* good, now check for a proper control word */
		control = 0x5a5a;	
		fnstcw(&control);
		if ((control & 0x1f3f) == 0x033f) {
			/*
			 * We have an npx, now divide by 0 to see if exception
			 * 16 works.
			 */
			control &= ~(1 << 2);	/* enable divide by 0 trap */
			fldcw(&control);
			probing_npx_trap = 1;
			fp_divide_by_0();
			probing_npx_trap = 0;

#if 1 /* change to #if 0 to test IRQ13 exception reporting on 486's */
			if (npx_traps_while_probing == 1) {
				/* good, exception 16 works */
				npx_ex16 = 1;
				dvp->id_irq = 0;	/* zap the interrupt */
				return (1);
			}
#endif

			/* bad, we are stuck with IRQ13 */
			if (isa_irq_pending(dvp)) {
				npx_irq13 = 1;
				npx0mask = dvp->id_irq;	/* npxattach too late */
				return (1);
			}

			/*
			 * IRQ13 is broken, but a fast and broken npx setup
			 * is preferable to our slow and broken emulator, and
			 * it would work reasonably well if we adjusted the
			 * default control word to mask all exceptions (masking
			 * is necessary to get the IEEE default fixups).
			 */
			dvp->id_irq = 0;
			return (1);
		}
	}

	/*
	 * Probe failed, but we want to get to npxattach to initialize the
	 * emulator and say that it has been installed.  XXX - handle
	 * devices that aren't really devices better.
	 */
	npx_emulator = 1;
	dvp->id_irq = 0;
	return (1);
}

/*
 * Attach routine - announce which it is, and wire into system
 */
npxattach(dvp)
	struct isa_device *dvp;
{
	if (npx_ex16)
		printf(" <Errors reported via Exception 16>");
	else if (npx_irq13)
		printf(" <Errors reported via IRQ 13>");
	else if (npx_emulator)
		printf(" <387 Emulator>");
	else
		printf(" <Error reporting broken (no IRQ13)>");
	if (!npx_emulator)
		npxexists++;
	npxinit(__INITIAL_NPXCW__);
}

/*
 * Initialize floating point unit.
 */
npxinit(control)
	unsigned short control;
{

	if (npxexists == 0)
		return;
	if (npxproc != 0) {
		printf("npxinit: npxproc = %x, curproc = %x\n",
		       npxproc, curproc);
		panic("npxinit clobbering in-use state");
	}
	stop_emulating();
	disable_intr();		/* avoid bogus IRQ13 */
	fninit();
	outb(0xf0, 0);		/* clear BUSY# latch to discard any IRQ13 */
	enable_intr();
	fldcw(&control);
	if (curpcb) {
		fnsave(&curpcb->pcb_savefpu);
		curpcb->pcb_flags |= FP_NEEDSRESTORE;
	}
	start_emulating();
	outb(0xf1, 0);		/* reset coproc. (XXX - too little too late) */
}

/*
 * Free coprocessor (if we have it).
 */
void
npxexit(p)
	struct proc *p;
{

	if (p == npxproc) {
		start_emulating();
		npxproc = 0;
	}
}

/*
 * Record the FPU state and reinitialize it all except for the control word.
 * Then generate a SIGFPE.
 *
 * Reinitializing the state allows naive SIGFPE handlers to longjmp without
 * doing any fixups.
 *
 * XXX - there is currently no way to pass the full error state to signal
 * handlers, and if this is a nested interrupt there is no way to pass even
 * a status code!  So there is no way to have a non-naive SIGFPE handler.  At
 * best a handler could do an fninit followed by an fldcw of a static value.
 * fnclex would be of little use because it would leave junk on the FPU stack.
 * Returning from the handler would be even less safe than usual because
 * IRQ13 exception handling makes exceptions even less precise than usual.
 */
npxintr(frame) struct intrframe frame; {
	int code;
	struct save87 npx_barf;

	if (probing_npx_trap) {
		++npx_traps_while_probing;
		fnclex();
		return;
	}
	if (npxproc == 0 || npxexists == 0) {
		printf("npxintr: npxproc = %x, curproc = %x, npxexists = %x\n",
		       npxproc, curproc, npxexists);
		panic("npxintr from nowhere");
	}
	if (npxproc != curproc) {
		printf("npxintr: npxproc = %x, curproc = %x, npxexists = %x\n",
		       npxproc, curproc, npxexists);
		panic("npxintr from non-current process");
	}

	/* save state (does implied fninit) */
	fnsave(&npx_barf);

	/*
	 * (Sanity) check that the pcb flags specify the state in the FPU and
	 * not the state in the pcb.
	 */
	if ((curpcb->pcb_flags & (FP_WASUSED | FP_NEEDSSAVE | FP_NEEDSRESTORE))
	    != (FP_WASUSED | FP_NEEDSSAVE))
		printf("npxintr: strange FP flags %x in pcb\n",
		       curpcb->pcb_flags
		       & (FP_WASUSED | FP_NEEDSSAVE | FP_NEEDSRESTORE));

	/* restore control word (was clobbered by fnsave) */
	fldcw(&npx_barf.sv_env.en_cw);

	outb(0xf0, 0);		/* clear BUSY# latch */

	/* pass exception to process */
	if (ISPL(frame.if_cs) == SEL_UPL) {
		/*
		 * Interrupt is essentially a trap, so we can afford to call
		 * the SIGFPE handler (if any) as soon as the interrupt
		 * returns.
		 *
		 * XXX - little or nothing is gained from this, and plenty is
		 * lost - the interrupt frame has to contain the trap frame
		 * (this is otherwise only necessary for the rescheduling trap
		 * in doreti, and the frame for that could easily be set up
		 * just before it is used).
		 */
		curproc->p_regs = (int *)&frame.if_es;
		curpcb->pcb_flags |= FM_TRAP;	/* used by sendsig */
#ifdef notyet
		/*
		 * Encode the appropriate code for detailed information on
		 * this exception.
		 */
		code = XXX_ENCODE(npx_barf.sv_env.en_sw);
#else
		code = 0;	/* XXX */
#endif
		trapsignal(curproc, SIGFPE, code);
		curpcb->pcb_flags &= ~FM_TRAP;
	} else {
		/*
		 * Nested interrupt.  These losers occur when:
		 *	o an IRQ13 is bogusly generated at a bogus time, e.g.:
		 *		o immediately after an fnsave or frstor of an
		 *		  error state.
		 *		o a couple of 386 instructions after
		 *		  "fstpl _memvar" causes a stack overflow.
		 *	  These are especially nasty when combined with a
		 *	  trace trap.
		 *	o an IRQ13 occurs at the same time as another higher-
		 *	  priority interrupt.
		 *
		 * Treat them like a true async interrupt.
		 */
		psignal(npxproc, SIGFPE);
	}
}

/*
 * Implement device not available (DNA) exception
 */
npxdna() {

	if (!npxexists)
		return (0);

	stop_emulating();

	if (npxproc != 0 && npxproc != curproc) {
		/*
		 * It would be better to switch FP context here (only), but
		 * I'm not sure how to get at the pcb for a swapped npxproc.
		 */
		printf("npxdna: npxproc = %x, curproc = %x\n",
		       npxproc, curproc);
		panic("npxdna");
	}

	/* record new context early in case frstor causes an IRQ13 */
	curpcb->pcb_flags |= FP_WASUSED | FP_NEEDSSAVE;
	npxproc = curproc;

    	if (curpcb->pcb_flags & FP_NEEDSRESTORE) {
		curpcb->pcb_flags &= ~FP_NEEDSRESTORE;
		/*
		 * The following frstor may cause an IRQ13 when the state being
		 * restored has a pending error.  The error will appear to have
		 * been triggered by the current (npx) user instruction even
		 * when that instruction is a no-wait instruction that should
		 * not trigger an error (e.g., fnclex).  On at least one 486
		 * system all of the no-wait instructions are broken the same
		 * as frstor, so our treatment does not amplify the breakage.
		 * On at least one 386/Cyrix 387 system, fnclex works correctly
		 * while frstor and fnsave are broken, so our treatment breaks
		 * fnclex if it is the first FPU instruction after a context
		 * switch.
		 */
		frstor(&curpcb->pcb_savefpu);
	}
	return (1);
}
#endif
