/*-
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)crt0.c	5.7 (Berkeley) 7/3/91";
#endif /* LIBC_SCCS and not lint */

/*
 *	C start up routine.
 *	Robert Henry, UCB, 20 Oct 81
 *
 *	We make the following (true) assumption:
 *	1) The only register variable that we can trust is the frame pointer,
 *	ebp, which points to the base of the kernel calling frame.
 */

extern	char	**environ;
extern	int	errno;
extern	int	end, *minbrk asm("minbrk"), *curbrk asm("curbrk");

asm(".text");
asm(".long 0xc000c000");

extern	unsigned char	etext;
extern	unsigned char	eprol asm ("eprol");
extern			start() asm("start");
extern			mcount() asm ("mcount");

static void _load_sharable_libraries(const char *);

start()
{
	struct kframe {
		int	kargc;
		char	*kargv[1];	/* size depends on kargc */
		char	kargstr[1];	/* size varies */
		char	kenvstr[1];	/* size varies */
	};
	/*
	 *	ALL REGISTER VARIABLES!!!
	 */
	register struct kframe *kfp;
	register char **targv;
	register char **argv;
	extern void _mcleanup();

	/* just above the saved frame pointer */
	asm ("lea 4(%%ebp), %0" : "=r" (kfp) );
	for (argv = targv = &kfp->kargv[0]; *targv++; /* void */)
		/* void */ ;
	if (targv >= (char **)(*argv))
		--targv;

	_load_sharable_libraries(*argv);
	environ = targv;
	minbrk = curbrk = &end;

asm("eprol:");
	exit(main(kfp->kargc, argv, environ));
}

/*
 * null mcount and moncontrol,
 * just in case some routine is compiled for profiling
 */
moncontrol(val)
	int val;
{

}

mcount(void) { }

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/file.h>
#include "shlib.h"

static const char *progname;

extern struct SharedLibraryTable _shared_library_table[];

asm(".text");
asm("__exit: ; lea 1, %eax ; .byte 0x9a ; .long 0; .word 7; ret");
asm("2: movl $-1,%eax; ret");
asm("_read: ; lea 3,%eax; .byte 0x9a ; .long 0; .word 7; jb 2b; ret");
asm("_write: ; lea 4,%eax; .byte 0x9a ; .long 0; .word 7; jb 2b; ret");
asm("_open: ; lea 5,%eax; .byte 0x9a ; .long 0; .word 7; jb 2b; ret");
asm("_close: ; lea 6,%eax; .byte 0x9a ; .long 0; .word 7; jb 2b; ret");
asm("_mmap: ; lea 71,%eax; .byte 0x9a ; .long 0; .word 7; jb 2b; ret");

static void print_string(const char *message) {
   const char *p = message;
   int len = 0;

   while (*p++) len++;
   write(2, message, len);
}

static void print_message_and_abort(const char *msg, const char *library) {

   print_string(progname);
   print_string("[");
   print_string(library);
   print_string("]: ");
   print_string(msg);
   print_string(" Aborted.\n");
   _exit(1);
}

void _load_sharable_libraries(const char *program_name) {
   struct SharedLibraryTable *p = _shared_library_table;

   progname = program_name;
   for (; p->name; p += 1) {
      int            fd = open(p->name, O_RDONLY);
      ShareLibHeader share;

      if (fd < 0 || read(fd, &share, sizeof share) != sizeof share)
        print_message_and_abort("Can't open shared library.", p->name);

      if (share.header.magic != SHLIB_MAGIC)
        print_message_and_abort("Not a shared library.", p->name);

      if (share.header.version != p->version)
        print_message_and_abort("Version mismatch of shared library.", p->name);

      if (mmap(share.header.text_addr, share.header.text_size,
	       PROT_READ | PROT_EXEC, MAP_FILE | MAP_FIXED, fd,
	       sizeof share) != share.header.text_addr)
          print_message_and_abort("Can't map text segment.", p->name);

      if (mmap(share.header.data_addr, share.header.data_size,
	       PROT_READ | PROT_WRITE, MAP_FILE | MAP_FIXED | MAP_PRIVATE, fd,
	       share.header.text_size + sizeof share) != share.header.data_addr)
          print_message_and_abort("Can't map data segment.", p->name);

      close(fd);
   }
}
