/*
 *  This program is free software; you can do with it as you see fit,
 *  as long as you leave this header intact.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *
 *	Soren Schmidt 		Email:	sos@kmd-ac.dk
 *	Tritonvej 36		UUCP:	...uunet!dkuug!kmd-ac!sos
 *	DK9210 Aalborg SO	Phone:  +45 9814 8076
 */

#ifndef	_CONSOLE_H_
#define	_CONSOLE_H_

#define KDGKBMODE 	_IO('K',  6)
#define KDSKBMODE 	_IO('K',  7)
#define KDMKTONE	_IO('K',  8)
#define KDGETMODE	_IO('K',  9)
#define KDSETMODE	_IO('K', 10)
#define KDENABIO	_IO('K', 60)
#define KDDISABIO	_IO('K', 61)
#define KDGKBTYPE	_IO('K', 64)
#define KDGETLED	_IO('K', 65)
#define KDSETLED	_IO('K', 66) 
#define KDSETRAD	_IO('K', 67)

#define GETFKEY		_IOWR('k', 0, fkeyarg_t)
#define SETFKEY		_IOWR('k', 1, fkeyarg_t)
#define GIO_SCRNMAP	_IOR('k', 2, scrmap_t)
#define PIO_SCRNMAP	_IOW('k', 3, scrmap_t)
#define GIO_KEYMAP 	_IOR('k', 6, keymap_t)
#define PIO_KEYMAP 	_IOW('k', 7, keymap_t)

#define CONS_BLANKTIME	_IOW('c', 4, long)
#define PIO_FONT8x8	_IOW('c', 64, fnt8_t)
#define GIO_FONT8x8	_IOR('c', 65, fnt8_t)
#define PIO_FONT8x14	_IOW('c', 66, fnt14_t)
#define GIO_FONT8x14	_IOR('c', 67, fnt14_t)
#define PIO_FONT8x16	_IOW('c', 68, fnt16_t)
#define GIO_FONT8x16	_IOR('c', 69, fnt16_t)
#define CONS_80x25TEXT	_IO('c', 102)
#define CONS_80x50TEXT	_IO('c', 103)

/* compatibility to old pccons & X386 */
#define CONSOLE_X_MODE_ON	_IO('t', 121)
#define CONSOLE_X_MODE_OFF	_IO('t', 122)

#define KD_TEXT		0
#define KD_GRAPHICS	1
#define K_RAW		0
#define K_XLATE		1
#define KB_84		1
#define KB_101		2
#define KB_OTHER	3
#define LED_SCR		4
#define LED_NUM		2
#define LED_CAP		1

/* possible flag values */
#define	FLAG_LOCK_O	0
#define	FLAG_LOCK_C	1
#define FLAG_LOCK_N	2

#define NUM_KEYS	128
#define NUM_STATES	8

struct keymap {
	u_short	n_keys;
	struct key_t {
		u_char map[NUM_STATES];
		u_char spcl;
		u_char flgs;
	} key[NUM_KEYS];
};

#define MAXFK		16

struct fkeytab {
	u_char	str[MAXFK];
	u_char	len;
};

struct fkeyarg {
	u_short	keynum;
	char	keydef[MAXFK];
	char	flen;
};

typedef struct keymap keymap_t;
typedef struct fkeytab fkeytab_t;
typedef struct fkeyarg fkeyarg_t;
typedef struct {char scrmap[256];} scrmap_t;
typedef struct {char fnt8x8[8*256];} fnt8_t;
typedef struct {char fnt8x14[14*256];} fnt14_t;
typedef struct {char fnt8x16[16*256];} fnt16_t;

#define F(x)		((x)+F_FN-1)
#define	S(x)		((x)+F_SCR-1)
#define NOP		0x00
#define LSH		0x02
#define RSH		0x03
#define CLK		0x04
#define NLK		0x05
#define SLK		0x06
#define LALT		0x07
#define LCTR		0x09
#define RCTR		0x7b
#define RALT		0x7c

#define F_SCR		11	/* switch to first screen */
#define L_SCR		26	/* switch to last screen */
#define F_FN		27	/* first function key */
#define L_FN		122	/* last function key */
#define FKEY		0x200	/* funtion key marker */

#define	KBSTAT		0x64	/* kbd status port */
#define	KBREADY		0x02	/* kbd char ready */
#define	KBDATA		0x60	/* kbd data port */
#define	KBSTATP		0x61	/* kbd status */

#endif
