/*-
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * William Jolitz and Don Ahn.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)pccons.c	5.11 (Berkeley) 5/21/91
 */
/*
 * Heavily modified by sos@kmd-ac.dk to provide:
 *
 * 	virtual screens 
 *	ioctl's for allowing IO operations to the video controller
 *	  "     for switching to graphics (unknown) modes
 *	  "	support for 25 & 50 line text modes
 *	  "     support for raw kbd scancode mode
 *	  "	support for setting/getting keymap table
 *	  "	support for setting/getting kbd LED status
 *	  "	support for setting kbd delay & repeat rate
 *	  "	support for sound (noise) generation 
 *	  "	support for loading fonts 
 *	  "	support for reprogramming function keys
 */

#include "param.h"
#include "conf.h"
#include "ioctl.h"
#include "proc.h"
#include "user.h"
#include "tty.h"
#include "uio.h"
#include "callout.h"
#include "systm.h"
#include "kernel.h"
#include "syslog.h"
#include "errno.h"
#include "console.h"
#include "malloc.h"
#include "i386/isa/icu.h"
#include "i386/isa/isa_device.h"
#include "i386/include/pc/display.h"
#include "i386/i386/cons.h"
#include "machine/psl.h"
#include "machine/frame.h"
#include "sc.h"
#include "ddb.h"
#include "iso8859.font"
#include "kbdtables.h"

#if NSC > 0
#ifndef NCONS
#define NCONS 12
#endif

#define MAX_ESC_PAR 	2
#define BLANK_TIMEOUT	300
#define TEXT80x25	1
#define TEXT80x50	2
#define UNKNOWN_MODE	0x0001
#define KBD_RAW_MODE	0x0002

#define	COL		80
#define	ROW		25
#define	CHR		2
#define	MONO_BASE	0x3B4
#define	MONO_BUF	0xfe0B0000
#define	CGA_BASE	0x3D4
#define	CGA_BUF		0xfe0B8000
#define	VGA_BUF		0xfe0A0000
#define	IOPHYSMEM	0xA0000

typedef struct scr_stat {
	u_short *crt_base;
	u_short *crtat;
	int esc;
	int n_par;
	int last_par;
	int par[MAX_ESC_PAR];
	int posx;
	int posy;
	int max_posx;
	int max_posy;
	int attr;
	int std_attr;
	int rev_attr;
	u_short status;
	u_short mode;
	u_short *scr;
	} scr_stat;

static	scr_stat	cons_scr_stat[NCONS];
static	scr_stat	*cur_scr_stat = &cons_scr_stat[0];

u_short			*Crtat = (u_short *)MONO_BUF;
static 	u_short	 	*crtat = 0;
static	u_int		addr_6845 = MONO_BASE;
static 	u_char		shfts, ctls, alts, caps, num, stp, scroll;
static 	const u_int 	n_fkey_tab = sizeof (fkey_tab) / sizeof (*fkey_tab);
static	int 		cur_cursor_pos = -1;
static	int 		in_putc, nx_scr;
static	long		blank_time = 0;		/* screen saver timout value */
static  scrmap_t	scr_map;

struct	tty 		pccons[NCONS];
struct	tty 		*cur_pccons = &pccons[0];

extern	int hz;
extern	u_int sc0mask;
extern	struct timeval time;

#define	CSF_ACTIVE	0x1	/* timeout active */
#define	CSF_POLLING	0x2	/* polling for input */

struct	pcconsoftc {
	char	cs_flags;
	char	cs_lastc;	/* last char sent */
	int	cs_timo;	/* timeouts since interrupt */
	u_long	cs_wedgecnt;	/* times restarted */
} pcconsoftc;


/* special characters */
#define bs	8
#define lf	10	
#define cr	13	
#define cntlc	3	
#define del	0177	
#define cntld	4

/* function prototypes */
int pcprobe(struct isa_device *dev);
int pcattach(struct isa_device *dev);
int pcopen(dev_t dev, int flag, int mode, struct proc *p);
int pcclose(dev_t dev, int flag, int mode, struct proc *p);
int pcread(dev_t dev, struct uio *uio, int flag);
int pcwrite(dev_t dev, struct uio *uio, int flag);
int pcparam(struct tty *tp, struct termios *t);
int pcioctl(dev_t dev, int cmd, caddr_t data, int flag);
int pcxint(dev_t dev);
int pcstart(struct tty *tp);
int pccnprobe(struct consdev *cp);
int pccninit(struct consdev *cp);
int pccnputc(dev_t dev, char c);
int pccngetc(dev_t dev);
int scintr(dev_t dev, int irq, int cpl);
void scrn_saver(int test);
static struct tty *get_pccons(dev_t dev);
static scr_stat *get_scr_stat(dev_t dev);
static void cursor_shape(int start, int end);
static void cursor_pos(void);
static void clear_screen(scr_stat *scp);
static void switch_scr(u_int next_scr);
static void move_crsr(scr_stat *scp, u_int x, u_int y);
static void move_up(u_short *s, u_short *d, u_int len);
static void move_down(u_short *s, u_short *d, u_int len);
static void scan_esc(scr_stat *scp, u_char c);
static void ansi_put(scr_stat *scp, u_char c);
void consinit(void);
static void sput(u_char c, u_char ca);
static u_char *get_fstr(u_int c, u_int *len);
static update_led(void);
void reset_cpu(void);
u_int sgetc(int noblock);
int pg(char *p, int q, int r, int s, int t, int u, int v, int w, int x, int y, int z);
static int getchar(void);
int pcmmap(dev_t dev, int offset, int nprot);
static void set_mode(int mode);
static void set_border(int color);
static load_font(int segment, int size, char* font);
static change_winsize(struct tty *tp, int x, int y);

struct	isa_driver scdriver = {
	pcprobe, pcattach, "sc",
};


int pcprobe(struct isa_device *dev)
{
	u_char c;
	int again = 0;

	/* Enable interrupts and keyboard controller */
	while (inb(KBSTAT)&KBREADY); outb(KBSTAT,0x60);
	while (inb(KBSTAT)&KBREADY); outb(KBDATA,0x4D);

	/* Start keyboard stuff RESET */
	while (inb(KBSTAT)&KBREADY);			/* wait input ready */
	outb(KBDATA,0xFF);				/* RESET */
	while((c=inb(KBDATA))!=0xFA) {
		if ((c == 0xFE) ||  (c == 0xFF)) {
			if(!again)
				printf("KEYBOARD disconnected: RECONNECT \n");
			while (inb(KBSTAT)&KBREADY);	/* wait input ready */
			outb(KBDATA,0xFF);		/* RESET */
			again = 1;
		}
	}
	/* pick up keyboard reset return code */
	while((c=inb(KBDATA))!=0xAA);
	sc0mask = dev->id_irq;
	return 1;
}


int pcattach(struct isa_device *dev)
{
	scr_stat *scp;
	int i;

	if (Crtat == (u_short*)MONO_BUF) 
		printf("VGA mono");
	else	
		printf("VGA color");

	if (NCONS > 1) 
		printf(" <%d virtual consoles>", NCONS);

	load_font(0, 16, font_8x16);
	load_font(1, 8, font_8x8);
	for (i = 0; i < NCONS; i++) {
		scp = &cons_scr_stat[i];
		scp->mode = TEXT80x25;
		scp->scr = (u_short *)malloc(COL * ROW * 2, M_DEVBUF, M_NOWAIT);
		scp->std_attr = (FG_LIGHTGREY | BG_BLACK) << 8;
		scp->rev_attr = (FG_BLACK | BG_LIGHTGREY) << 8;
		scp->attr = scp->std_attr;
		scp->max_posx = COL;
		scp->max_posy = ROW;
		if (i > 0) {
			scp->crt_base = scp->crtat = scp->scr;
			fillw (scp->attr | ' ', scp->scr, COL * ROW);
		}
	}
	cursor_shape(0, 16);
	cursor_pos();
}


static struct tty *get_pccons (dev_t dev)
{
	int i = minor (dev);

	if (i >= NCONS)
		return (NULL);
	return (&pccons[i]);
}


static scr_stat *get_scr_stat (dev_t dev)
{
	int i = minor (dev);

	if (i >= NCONS)
		return (NULL);
	return (&cons_scr_stat[i]);
}


pcopen(dev_t dev, int flag, int mode, struct proc *p)
{
	struct tty *tp = get_pccons(dev);

	if (!tp)
		return (ENXIO);
	tp->t_oproc = pcstart;
	tp->t_param = pcparam;
	tp->t_dev = dev;
	if ((tp->t_state & TS_ISOPEN) == 0) {
		tp->t_state |= TS_WOPEN;
		ttychars(tp);
		tp->t_iflag = TTYDEF_IFLAG;
		tp->t_oflag = TTYDEF_OFLAG;
		tp->t_cflag = TTYDEF_CFLAG;
		tp->t_lflag = TTYDEF_LFLAG;
		tp->t_ispeed = tp->t_ospeed = TTYDEF_SPEED;
		pcparam(tp, &tp->t_termios);
		ttsetwater(tp);
	} else if (tp->t_state&TS_XCLUDE && p->p_ucred->cr_uid != 0)
		return (EBUSY);
	tp->t_state |= TS_CARR_ON;
	return ((*linesw[tp->t_line].l_open)(dev, tp));
}


pcclose(dev_t dev, int flag, int mode, struct proc *p)
{
	struct tty *tp = get_pccons(dev);

	if (!tp)
		return (ENXIO);
	(*linesw[tp->t_line].l_close)(tp, flag);
	ttyclose(tp);
	return(0);
}


pcread(dev_t dev, struct uio *uio, int flag)
{
	struct tty *tp;

	tp = get_pccons (dev);
	if (!tp)
		return (ENXIO);
	return ((*linesw[tp->t_line].l_read)(tp, uio, flag));
}


pcwrite(dev_t dev, struct uio *uio, int flag)
{
	struct tty *tp;

	tp = get_pccons (dev);
	if (!tp)
		return (ENXIO);
	return ((*linesw[tp->t_line].l_write)(tp, uio, flag));
}


/*
 * Got a console receive interrupt -
 * the console processor wants to give us a character.
 * Catch the character, and see who it goes to.
 */
scintr(dev_t dev, int irq, int cpl)
{
	int c, len;
	u_char *cp;

	scrn_saver(0);

	c = sgetc(1);
	if (c & 0x100)
		return;
	if (pcconsoftc.cs_flags & CSF_POLLING)
		return;
	if (c < 0x100)
		(*linesw[cur_pccons->t_line].l_rint)(c & 0xFF, cur_pccons);
	else if (cp = get_fstr ((u_int)c, (u_int *)&len)) {
		while (len-- >  0)
			(*linesw[cur_pccons->t_line].l_rint)(*cp++ & 0xFF, cur_pccons);
	}
}


/*
 * Set line parameters
 */
pcparam(struct tty *tp, struct termios *t)
{
	register int cflag = t->c_cflag;
	/* and copy to tty */
	tp->t_ispeed = t->c_ispeed;
	tp->t_ospeed = t->c_ospeed;
	tp->t_cflag = cflag;

	return(0);
}


pcioctl(dev_t dev, int cmd, caddr_t data, int flag)
{
	register error;
	u_char byte;
	struct tty *tp;
	struct syscframe *fp;
	scr_stat *scp; 

	tp = get_pccons (dev);
	if (!tp)
		return (ENXIO);
	scp = get_scr_stat (tp->t_dev);

	switch (cmd) {	/* process console hardware related ioctl's */
	case CONS_BLANKTIME:
		blank_time = *(int*)data;
		return 0;

	case CONS_80x25TEXT:	/* set 80x25 text mode */
		set_mode(TEXT80x25);
		scp->mode = TEXT80x25;
		scp->max_posy = 25;
		change_winsize(tp, scp->max_posx, scp->max_posy); 
		clear_screen(scp);
		free(scp->scr, M_DEVBUF); 
		scp->scr = (u_short *)malloc( scp->max_posx * scp->max_posy * 2, M_DEVBUF, M_NOWAIT);
		return 0;

	case CONS_80x50TEXT:	/* set 80x50 text mode */
		set_mode(TEXT80x50);
		scp->mode = TEXT80x50;
		scp->max_posy = 50;
		change_winsize(tp, scp->max_posx, scp->max_posy); 
		clear_screen(scp);
		free(scp->scr, M_DEVBUF); 
		scp->scr = (u_short *)malloc( scp->max_posx * scp->max_posy * 2, M_DEVBUF, M_NOWAIT);
		return 0;

	case KDENABIO:		/* allow io operations */
	 	fp = (struct syscframe *)curproc->p_regs;
	 	fp->sf_eflags |= PSL_IOPL;
		return 0; 

	case KDDISABIO:		/* disallow io operations (default) */
	 	fp = (struct syscframe *)curproc->p_regs;
	 	fp->sf_eflags &= ~PSL_IOPL;
	 	return 0;

        case KDSETMODE:		/* set current mode of this (virtual) tty */
		switch (*data) {
		case KD_TEXT:		/* switch to TEXT (known) mode */
			/* restore fonts ! */
			load_font(0, 16, font_8x16);
			load_font(1, 8, font_8x8);
			scp->status &= ~UNKNOWN_MODE;
			set_mode(scp->mode);
			clear_screen(scp);
			return 0;
		case KD_GRAPHICS:	/* switch to GRAPHICS (unknown) mode */
			scp->status |= UNKNOWN_MODE;
			return 0;
		default:
			return -1;
		}

	case KDGETMODE:		/* get current mode of this (virtual) tty */
		*data = (scp->status & UNKNOWN_MODE) ? KD_GRAPHICS : KD_TEXT;
		return 0;

	case KDSETRAD:		/* set keyboard repeat & delay rates */
		while (inb(KBSTAT) & KBREADY) ;
		outb(KBDATA, 0xF3);		/* set delay & rate command */
		while (inb(KBSTAT) & KBREADY) ;
		outb(KBDATA, ( *data & 0x7f) );	/* set delay & rate */
		return 0;

	case KDSKBMODE:		/* set keyboard mode */
		switch (*data) {
		case K_RAW:	/* switch to RAW scancode mode */
			scp->status |= KBD_RAW_MODE;
			return 0;
		case K_XLATE:	/* switch to XLT ascii mode */
			scp->status &= ~KBD_RAW_MODE;
			return 0;
		default:
			return -1;
		}

	case KDGKBMODE:		/* get keyboard mode */
		*data = (scp->status & KBD_RAW_MODE) ? K_RAW : K_XLATE;
		return 0;

	case KDMKTONE:		/* sound the bell */
		sysbeep (0x31b, hz/4);
		return 0;

	case KDSETLED:		/* set keyboard LED status */
		scroll = (*data & LED_SCR) >> 2;
		num    = (*data & LED_NUM) >> 1;
		caps   = (*data & LED_CAP);
		update_led();
		return 0;

	case KDGETLED:		/* get keyboard LED status */
		*data = scroll<<2 | num<<1 | caps;
		return 0;

	case GETFKEY:		/* get functionkey string */
		if (*(u_short*)data < n_fkey_tab) {
		 	fkeyarg_t *ptr = (fkeyarg_t*)data;
			bcopy(&fkey_tab[ptr->keynum].str, ptr->keydef, fkey_tab[ptr->keynum].len);
			ptr->flen = fkey_tab[ptr->keynum].len;
			return 0;
		}
		else
			return -1;

	case SETFKEY:		/* set functionkey string */
		if (*(u_short*)data < n_fkey_tab) {
		 	fkeyarg_t *ptr = (fkeyarg_t*)data;
			bcopy(ptr->keydef, &fkey_tab[ptr->keynum].str, ptr->flen);
			fkey_tab[ptr->keynum].len = ptr->flen;
			return 0;
		}
		else
			return -1;

	case GIO_SCRNMAP: 	/* get output translation table */
		bcopy(&scr_map, data, sizeof(scrmap_t));
		return 0;

	case PIO_SCRNMAP:	/* set output translation table */
		bcopy(data, &scr_map, sizeof(scrmap_t));
		return 0;

	case GIO_KEYMAP: 	/* get keyboard translation table */
		bcopy(&key_map, data, sizeof(key_map));
		return 0;

	case PIO_KEYMAP:	/* set keyboard translation table */
		bcopy(data, &key_map, sizeof(key_map));
		return 0;

	case PIO_FONT8x8:	/* set 8x8 dot font */
		bcopy(data, &font_8x8, sizeof(fnt8_t));
		load_font(1, 8, font_8x16);
		return 0;

	case GIO_FONT8x8:	/* get 8x8 dot font */
		bcopy(&font_8x8, data, sizeof(fnt8_t));
		return 0;

	case PIO_FONT8x14:	/* set 8x14 dot font */
		bcopy(data, &font_8x14, sizeof(fnt14_t));
		load_font(2, 14, font_8x14);
		return 0;

	case GIO_FONT8x14:	/* get 8x14 dot font */
		bcopy(&font_8x14, data, sizeof(fnt14_t));
		return 0;

	case PIO_FONT8x16:	/* set 8x16 dot font */
		bcopy(data, &font_8x16, sizeof(fnt16_t));
		load_font(0, 16, font_8x16);
		return 0;

	case GIO_FONT8x16:	/* get 8x16 dot font */
		bcopy(&font_8x16, data, sizeof(fnt16_t));
		return 0;

	case CONSOLE_X_MODE_ON:	/* just to be compatible */
	 	fp = (struct syscframe *)curproc->p_regs;
	 	fp->sf_eflags |= PSL_IOPL;
		scp->status |= UNKNOWN_MODE;
		scp->status |= KBD_RAW_MODE;
		return 0;

	case CONSOLE_X_MODE_OFF:/* just to be compatible */
	 	fp = (struct syscframe *)curproc->p_regs;
	 	fp->sf_eflags &= ~PSL_IOPL;
		load_font(0, 16, font_8x16);
		load_font(1, 8, font_8x8);
		scp->status &= ~UNKNOWN_MODE;
		set_mode(scp->mode);
		clear_screen(scp);
		scp->status &= ~KBD_RAW_MODE;
		return 0;

	default:
		break;
	}	

	error = (*linesw[tp->t_line].l_ioctl)(tp, cmd, data, flag);
	if (error >= 0)
		return (error);
	error = ttioctl(tp, cmd, data, flag);
	if (error >= 0)
		return (error);
	return (ENOTTY);
}


/*
 * Got a console transmission interrupt -
 * the console processor wants another character.
 */
pcxint(dev_t dev)
{
	register struct tty *tp;
	register int unit;

	pccons[0].t_state &= ~TS_BUSY;
	pcconsoftc.cs_timo = 0;
	if (pccons[0].t_line)
		(*linesw[pccons[0].t_line].l_start)(&pccons[0]);
	else
		pcstart(&pccons[0]);
}


pcstart(struct tty *tp)
{
	register c, s;
	scr_stat *scp = get_scr_stat (tp->t_dev);

	s = spltty ();
	if (!(tp->t_state & (TS_TIMEOUT|TS_BUSY|TS_TTSTOP)))
		for (;;) {
			if (RB_LEN(&tp->t_out) <= tp->t_lowat) {
				if (tp->t_state & TS_ASLEEP) {
					tp->t_state &= ~TS_ASLEEP;
					wakeup ((caddr_t)&tp->t_out);
				}
				if (tp->t_wsel) {
					selwakeup (tp->t_wsel, tp->t_state & TS_WCOLL);
					tp->t_wsel = 0;
					tp->t_state &= ~TS_WCOLL;
				}
			}
			if (RB_LEN(&tp->t_out) == 0)
				break;
			c = getc (&tp->t_out);
			splx (s);
			ansi_put (scp, c);
			s = spltty ();
		}
	splx(s);
}


pccnprobe(struct consdev *cp)
{
	int maj;

	/* locate the major number */
	for (maj = 0; maj < nchrdev; maj++)
		if (cdevsw[maj].d_open == pcopen)
			break;

	/* initialize required fields */
	cp->cn_dev = makedev(maj, 0);
	cp->cn_tp = &pccons[0];
	cp->cn_pri = CN_INTERNAL;
}


pccninit(struct consdev *cp)
{
}


pccnputc(dev_t dev, char c)
{
	int pos;

	if (cur_scr_stat->status & UNKNOWN_MODE) 
		return;

	if (c == '\n')
		sput('\r', FG_LIGHTGREY | BG_BLACK);
	sput(c, FG_LIGHTGREY | BG_BLACK);

 	pos = cur_scr_stat->crtat - cur_scr_stat->crt_base;
	if (pos != cur_cursor_pos) {
		cur_cursor_pos = pos;
		outb(addr_6845,14);
		outb(addr_6845+1,pos >> 8);
		outb(addr_6845,15);
		outb(addr_6845+1,pos&0xff);
	}
}


pccngetc(dev_t dev)
{
	register int c, s;

	s = spltty();		/* block scintr while we poll */
	c = sgetc(0);
	if (c == '\r') c = '\n';
	splx(s);
	return (c);
}


void scrn_saver(int test)
{
	u_char val;
	static int blanked = 0;
	static long time_stamp;

	if (test && blank_time) {
		if (time.tv_sec > time_stamp + blank_time) {
			blanked = 1;
  			outb(0x3C4, 0x01); val = inb(0x3C5); outb(0x3C4, 0x01);
  			outb(0x3C5, val | 0x20);
		}
	}
	else {
		time_stamp = time.tv_sec;
		if (blanked) {
			blanked = 0;
  			outb(0x3C4, 0x01); val = inb(0x3C5); outb(0x3C4, 0x01);
  			outb(0x3C5, val & 0xDF);
		}
	}
}


static void cursor_shape(int start, int end)
{
	outb(addr_6845, 10);
	outb(addr_6845+1, start & 0xFF);
	outb(addr_6845, 11);
	outb(addr_6845+1, end & 0xFF);
}


static void cursor_pos ()
{
	int pos = cur_scr_stat->crtat - cur_scr_stat->crt_base;

	if (cur_scr_stat->status & UNKNOWN_MODE) 
		return;
	scrn_saver(1);
	if (pos != cur_cursor_pos) {
		cur_cursor_pos = pos;
		outb(addr_6845,14);
		outb(addr_6845+1,pos >> 8);
		outb(addr_6845,15);
		outb(addr_6845+1,pos&0xff);
	}
	timeout(cursor_pos,0,hz/20);
}


static void clear_screen(scr_stat *scp)
{
		move_crsr (scp, 0, 0);
		fillw (scp->attr | ' ', scp->crt_base,
		       scp->max_posx * scp->max_posy);
		cursor_shape(0, 16);
		cursor_pos();
}


static void switch_scr (u_int next_scr)
{
	scr_stat *scp, *oscp;

	if (next_scr >= NCONS || cur_scr_stat->status & UNKNOWN_MODE) {
		sysbeep (0x31b, hz/4);
		return;
	}
	oscp = cur_scr_stat;
	scp = &cons_scr_stat[next_scr];
	if (scp == oscp)
		return;
	if (in_putc) {
		nx_scr = next_scr + 1;
		return;
	}
	bcopy (Crtat, oscp->scr, oscp->max_posx * oscp->max_posy * 2);
	oscp->crt_base = oscp->scr;
	move_crsr (oscp, oscp->posx, oscp->posy); /* fix oscp->crtat */
	cur_scr_stat = scp;
	cur_pccons = &pccons[next_scr];
	scp->crt_base = Crtat;
	if (oscp->mode != scp->mode)
		set_mode(scp->mode);
	move_crsr (scp, scp->posx, scp->posy);
	bcopy (scp->scr, Crtat, scp->max_posx * scp->max_posy * 2);
	nx_scr = 0;
}


static void move_crsr (scr_stat *scp, u_int x, u_int y)
{
	if (x < 0 || y < 0 || x >= scp->max_posx || y >= scp->max_posy)
		return;
	scp->posx = x;
	scp->posy = y;
	scp->crtat = scp->crt_base + scp->posy * scp->max_posx + scp->posx;
}


static void move_up (u_short *s, u_short *d, u_int len)
{
	s += len;
	d += len;
	while (len-- > 0)
		*--d = *--s;
}


static void move_down (u_short *s, u_short *d, u_int len)
{
	while (len-- > 0)
		*d++ = *s++;
}


static void scan_esc (scr_stat *scp, u_char c)
{
	static u_char ansi_col[16] = {0, 4, 2, 6, 1, 5, 3, 7, 8, 12, 10, 14, 9, 13, 11, 15};
	int i, n, m;
	u_short *src, *dst, count;

	if (scp->esc == 1) {
		switch (c) {
		case 'c':	/* Clear screen & home */
			clear_screen(scp);
			break;
		case '[':
			/* Start ESC [ sequence */
			scp->esc = 2;
			scp->last_par = -1;
			for (i = scp->n_par; i < MAX_ESC_PAR; i++)
				scp->par[i] = 1;
			scp->n_par = 0;
			return;
		case 'M':
			if (scp->posy > 0)
				move_crsr (scp, scp->posx, scp->posy - 1);
			else {
				move_up (scp->crt_base, scp->crt_base + scp->max_posx,
					(scp->max_posy - 1) * scp->max_posx);
				fillw (scp->attr | ' ', scp->crt_base, scp->max_posx);
			}
			break;
		}
	}
	else if (scp->esc == 2) {
		if (c >= '0' && c <= '9') {
			if (scp->n_par < MAX_ESC_PAR) {
				if (scp->last_par != scp->n_par) {
					scp->last_par = scp->n_par;
					scp->par[scp->n_par] = 0;
				}
				else
					scp->par[scp->n_par] *= 10;
				scp->par[scp->n_par] += c - '0';
				return;
			}
		}
		scp->n_par = scp->last_par + 1;
		switch (c) {
		case ';':
			if (scp->n_par < MAX_ESC_PAR)
				return;
			break;
		case 'x':
			if (scp->n_par == 0)
				n = 0;
			else
				n = scp->par[0];
			switch (n) {
			case 0: /* reset attributes */
				scp->attr = scp->std_attr = FG_LIGHTGREY << 8;
				set_border(scp->attr>>12);
				break;
			case 1: /* set ansi background */
				scp->attr = scp->std_attr =  
					(scp->std_attr & 0x0F00) 
					| (ansi_col[(scp->par[1]) & 0x0F] << 12);
				set_border(scp->attr>>12);
				break;
			case 2: /* set ansi foreground */
				scp->attr = scp->std_attr =  
					(scp->std_attr & 0xF000) 
					| (ansi_col[(scp->par[1]) & 0x0F] << 8);
				break;
			case 3: /* set ansi attribute directly */
				scp->attr = scp->std_attr =
					(scp->par[1] & 0xFF) << 8;
				set_border(scp->attr>>12);
				break;
			case 5: /* set ansi reverse video background */
				scp->rev_attr = 
					(scp->rev_attr & 0x0F00) 
					| (ansi_col[(scp->par[1]) & 0x0F] << 12);
				break;
			case 6: /* set ansi reverse video foreground */
				scp->rev_attr = 
					(scp->rev_attr & 0xF000) 
					| (ansi_col[(scp->par[1]) & 0x0F] << 8);
				break;
			case 7: /* set ansi reverse video attribute directly */
				scp->rev_attr = (scp->par[1] & 0xFF) << 8;
				break;
			}
			break;
		case 'm': /* change mode */
			if (scp->n_par == 0)
				n = 0;
			else
				n = scp->par[0];
			switch (n) {
			case 0:	/* back to normal */
				scp->attr = scp->std_attr;
				break;
			case 1:	/* highlight (bold) */
				scp->attr &= 0xFF00;
				scp->attr |= 0x0800;
				break;
			case 4: /* highlight (underline) */
				scp->attr &= 0x0F00;
				scp->attr |= 0x0800;
				break;
			case 5: /* blink */
				scp->attr &= 0xFF00;
				scp->attr |= 0x8000;
				break;
			case 7: /* reverse video */
				scp->attr = scp->rev_attr; 
				break;
			case 30: case 31: case 32: case 33: /* set fg color */
			case 34: case 35: case 36: case 37:
				scp->attr = (scp->attr & 0xF0FF)
					    | (ansi_col[(n - 30) & 7] << 8);
				break;
			case 40: case 41: case 42: case 43: /* set bg color */
			case 44: case 45: case 46: case 47:
				scp->attr = (scp->attr & 0x0FFF)
					    | (ansi_col[(n - 40) & 7] << 12);
				set_border(scp->attr>>12);
				break;
			}
			break;
		case 'A': /* up n rows */
			move_crsr (scp, scp->posx, scp->posy - scp->par[0]);
			break;
		case 'B': /* down n rows */
			move_crsr (scp, scp->posx, scp->posy + scp->par[0]);
			break;
		case 'C': /* right n columns */
			move_crsr (scp, scp->posx + scp->par[0], scp->posy);
			break;
		case 'D': /* left n columns */
			move_crsr (scp, scp->posx - scp->par[0], scp->posy);
			break;
		case 'f': /* System V consoles .. */
		case 'H': /* Cursor move */
			move_crsr (scp, scp->par[1] - 1, scp->par[0] - 1);
			break;
		case 'J': /* Clear all or part of display */
			if (scp->n_par == 0)
				n = 0;
			else
				n = scp->par[0];
			switch (n) {
			case 0: /* clear form cursor to end of display */
				fillw (scp->attr | ' ', scp->crtat,
				       scp->crt_base + scp->max_posx * scp->max_posy - scp->crtat);
				break;
			case 1: /* clear from beginning of display to cursor */
				fillw (scp->attr | ' ', scp->crt_base,
				       scp->crtat - scp->crt_base);
				break;
			case 2: /* clear entire display */
				clear_screen(scp);
				break;
			}
			break;
		case 'K': /* Clear all or part of line */
			if (scp->n_par == 0)
				n = 0;
			else
				n = scp->par[0];
			switch (n) {
			case 0: /* clear form cursor to end of line */
				fillw (scp->attr | ' ', scp->crtat,
				       scp->max_posx - scp->posx);
				break;
			case 1: /* clear from beginning of line to cursor */
				fillw (scp->attr | ' ', 
				       scp->crtat - (scp->max_posx - scp->posx),
				       (scp->max_posx - scp->posx) + 1);
				break;
			case 2: /* clear entire line */
				fillw (scp->attr | ' ', 
				       scp->crtat - (scp->max_posx - scp->posx),
				       scp->max_posx);
				break;
			}
			break;
		case 'L':	/* Insert n lines */
			n = scp->par[0];
			if (n < 1)
				n = 1;
			if (n > scp->max_posy - scp->posy)
				n = scp->max_posy - scp->posy;
			src = scp->crt_base + scp->posy * scp->max_posx;
			dst = src + n * scp->max_posx;
			count = scp->max_posy - (scp->posy + n);
			move_up (src, dst, count * scp->max_posx);
			fillw (scp->attr | ' ', src, n * scp->max_posx);
			break;
		case 'M':	/* Delete n lines */
			n = scp->par[0];
			if (n < 1)
				n = 1;
			if (n > scp->max_posy - scp->posy)
				n = scp->max_posy - scp->posy;
			dst = scp->crt_base + scp->posy * scp->max_posx;
			src = dst + n * scp->max_posx;
			count = scp->max_posy - (scp->posy + n);
			move_down (src, dst, count * scp->max_posx);
			src = dst + count * scp->max_posx;
			fillw (scp->attr | ' ', src, n * scp->max_posx);
			break;
		case 'P':	/* Delete n chars */
			n = scp->par[0];
			if (n < 1)
				n = 1;
			if (n > scp->max_posx - scp->posx)
				n = scp->max_posx - scp->posx;
			dst = scp->crtat;
			src = dst + n;
			count = scp->max_posx - (scp->posx + n);
			move_down (src, dst, count);
			src = dst + count;
			fillw (scp->attr | ' ', src, n);
			break;
		case '@':	/* Insert n chars */
			n = scp->par[0];
			if (n < 1)
				n = 1;
			if (n > scp->max_posx - scp->posx)
				n = scp->max_posx - scp->posx;
			src = scp->crtat;
			dst = src + n;
			count = scp->max_posx - (scp->posx + n);
			move_up (src, dst, count);
			fillw (scp->attr | ' ', src, n);
			break;
		case 'S':	/* scroll up n lines */
			n = scp->par[0];
			if (n < 1) 
				n = 1;
			bcopy (scp->crt_base + (scp->max_posx * n),
			       scp->crt_base, 
			       scp->max_posx * (scp->max_posy - n) * CHR);
			fillw (scp->attr | ' ',
			       scp->crt_base + scp->max_posx * (scp->max_posy - 1), 
			       scp->max_posx);
			break;
		case 'T':
			n = scp->par[0];
			if (n < 1) 
				n = 1;
			bcopy (scp->crt_base, 
			       scp->crt_base + (scp->max_posx * n),
			       scp->max_posx * (scp->max_posy - n) * CHR);
			fillw (scp->attr | ' ', scp->crt_base,  scp->max_posx);
			break;
		}
	}
	scp->esc = 0;
}


#define wrtchar(scp, c) ( *scp->crtat = (c), scp->crtat++, scp->posx++ )

static void ansi_put (scr_stat *scp, u_char c)
{
	int s;

	if (scp == cur_scr_stat )
		scrn_saver(0);
	if (scp->status & UNKNOWN_MODE) {
		sgetc(1);
		return;
	}
	in_putc++;
	if (scp->esc)
		scan_esc (scp, c);
	else switch(c) {
	case 0x1B:
		scp->esc = 1;
		scp->n_par = 0;
		break;

	case '\t':
		do {
			wrtchar (scp, scp->attr | ' ');
		} while (scp->posx % 8);
		break;

	case '\b':      /* non-destructive backspace */
		if (scp->crtat > scp->crt_base) {
			scp->crtat--;
			if (scp->posx > 0)
				scp->posx--;
			else {
				scp->posx += scp->max_posx - 1;
				scp->posy--;
			}
		}
		break;

	case '\r':
		move_crsr (scp, 0, scp->posy);
		break;

	case '\n':
		scp->crtat += scp->max_posx;
		scp->posy++;
		break;

	default:
		if (c == 7)
			sysbeep (0x31b, hz/4);
		/* Print only printables */
		else /*if (c >= ' ') */ {
			wrtchar (scp, scp->attr | c);
			if (scp->posx >= scp->max_posx) {
				scp->posx = 0;
				scp->posy++;
			}
			break;
		}
	}
	if (scp->crtat >= scp->crt_base + scp->max_posy * scp->max_posx) {
		/* scroll check */
		do
			sgetc(1);
		while (scroll);
		bcopy (scp->crt_base + scp->max_posx, scp->crt_base,
			scp->max_posx * (scp->max_posy - 1) * CHR);
		fillw (scp->attr | ' ',
			scp->crt_base + scp->max_posx * (scp->max_posy - 1), scp->max_posx);
		scp->crtat -= scp->max_posx;
		scp->posy--;
	}
	s = splx (sc0mask);
	in_putc--;
	if (nx_scr)
		switch_scr (nx_scr - 1);
	splx (s);
}


void consinit (void)
{
	u_short *cp = Crtat + (CGA_BUF-MONO_BUF)/CHR, was;
	scr_stat *scp;
	unsigned cursorat;
	int i;

	/* Crtat initialized to point to MONO buffer   */
	/* if not present change to CGA_BUF offset     */
	/* ONLY ADD the difference since locore.s adds */
	/* in the remapped offset at the right time    */

	was = *cp;
	*cp = (u_short) 0xA55A;
	if (*cp != 0xA55A) {
		addr_6845 = MONO_BASE;
	} else {
		*cp = was;
		addr_6845 = CGA_BASE;
		Crtat = Crtat + (CGA_BUF-MONO_BUF)/CHR;
	}
	/* Extract cursor location */
	outb(addr_6845,14);
	cursorat = inb(addr_6845+1)<<8 ;
	outb(addr_6845,15);
	cursorat |= inb(addr_6845+1);

	crtat = Crtat + cursorat;
	fillw((0x0f<<8)|' ', crtat, COL*ROW-cursorat);

	cons_scr_stat[0].crtat = crtat;
	cons_scr_stat[0].crt_base = Crtat;
	cons_scr_stat[0].posx = cursorat % COL;
	cons_scr_stat[0].posy = cursorat / COL;
	cons_scr_stat[0].std_attr = (FG_LIGHTGREY | BG_BLACK) << 8;
	cons_scr_stat[0].rev_attr = (FG_BLACK | BG_LIGHTGREY) << 8;
	cons_scr_stat[0].attr = scp->std_attr;
	cons_scr_stat[0].max_posx = COL;
	cons_scr_stat[0].max_posy = ROW;
}


static void sput (u_char c, u_char ca)
{
	int i;
	scr_stat *scp = &cons_scr_stat[0];

	if (crtat == 0)
		consinit ();
	i = scp->attr;
	scp->attr = ca << 8;
	ansi_put (scp, c);
	scp->attr = i;
}


static u_char *get_fstr (u_int c, u_int *len)
{
	u_int i;

	if (!(c & FKEY))
		return (NULL);
	i = (c & 0xFF) - F_FN;
	if (i > n_fkey_tab)
		return (NULL);
	*len = fkey_tab[i].len;
	return (fkey_tab[i].str);
}


static update_led()
{
	while (inb(KBSTAT)&KBREADY);		/* wait input ready */
	outb (KBDATA,0xED);			/* LED Command */
	while (inb(KBSTAT)&KBREADY);		/* wait input ready */
	outb (KBDATA, scroll | 2*num | 4*caps);
}


volatile void reset_cpu (void)
{
	for (;;) {
		while (inb(KBSTAT)&KBREADY);	/* wait input ready */
		outb(KBSTAT,0xFE);		/* Reset Command */
		DELAY(4000000);
		while (inb(KBSTAT)&KBREADY);	/* wait input ready */
		outb(KBSTAT,0xFF);		/* Keyboard Reset Command */
	}
}


/*
 * sgetc(noblock) : get a character from the keyboard. 
 * If noblock = 0 wait until a key is gotten.  Otherwise return a 0x100 (256).
 */
u_int sgetc (int noblock)
{
	u_char dt;
	u_int state, action;
	struct key_t *key;
	static u_int esc_flag;

loop:
	esc_flag = 0;
next_code:
	/* First see if there is something in the keyboard port */
	if (inb (KBSTAT) & 1)
		dt = inb (KBDATA);
	else if (noblock)
		return (0x100);
	else
		goto loop;

	/* Check for cntl-alt-del */
	if ((dt == 83) && ctls && alts)
		cpu_reset();

#if NDDB > 0
	/* Check for cntl-alt-esc */
	if ((dt == 1) && ctls && alts)
		Debugger();
#endif
	if (cur_scr_stat->status & KBD_RAW_MODE)
		return dt;

	if ((dt & ~1) == 0xE0) {
		esc_flag = dt - 0xE0 + 1;
		goto next_code;
	}

	if ((dt & 0x7F) >= key_map.n_keys)
		goto loop;
	state = (shfts ? 1 : 0 ) | (2 * (ctls ? 1 : 0)) | (4 * (alts ? 1 : 0));
	key = &key_map.key[dt & 0x7F];
	if (((key->flgs & FLAG_LOCK_C) && caps)
	    || ((key->flgs & FLAG_LOCK_N) && num))
		state ^= 1;

	/* Check for make/break */
	action = key->map[state];
	if (dt & 0x80) {
		/* break */
		if (key->spcl & 0x80) {
			switch (action) {
			case LSH:
				shfts &= ~1;
				break;
			case RSH:
				shfts &= ~2;
				break;
			case LCTR:
				ctls &= ~1;
				break;
			case RCTR:
				ctls &= ~2;
				break;
			case LALT:
				alts &= ~1;
				break;
			case RALT:
				alts &= ~2;
				break;
			}
		}
	} else {
		/* make */
		if (key->spcl & (0x80>>state)) {
			switch (action) {
			/* LOCKING KEYS */
			case NLK:
				num ^= 1;
				update_led ();
				break;
			case CLK:
				caps ^= 1;
				update_led ();
				break;
			case SLK:
				scroll ^= 1;
				update_led ();
				break;
	
			/* NON-LOCKING KEYS */
			case LSH:
				shfts |= 1;
				break;
			case RSH:
				shfts |= 2;
				break;
			case LCTR:
				ctls |= 1;
				break;
			case RCTR:
				ctls |= 2;
				break;
			case LALT:
				alts |= 1;
				break;
			case RALT:
				alts |= 2;
				break;
			default:
				if (action >= F_SCR && action <= L_SCR) {
					switch_scr (action - F_SCR);
					break;
				}
				if (action >= F_FN && action <= L_FN) {
					return (action | FKEY);
				}
				return (action);
			}
		}
		else  return (action);
	}
	goto loop;
}


pg(p,q,r,s,t,u,v,w,x,y,z) char *p; {
	switch_scr (0);
	printf(p,q,r,s,t,u,v,w,x,y,z);
	printf("\n");
	return(getchar());
}


static int getchar()
{
	register char thechar;
	register delay;
	int x;

	pcconsoftc.cs_flags |= CSF_POLLING;
	x=splhigh();
	sput('>', FG_RED | BG_BLACK);
	thechar = (char) sgetc(0);
	pcconsoftc.cs_flags &= ~CSF_POLLING;
	splx(x);
	switch (thechar) {
	default: 
		if (thechar >= ' ')
			sput(thechar, FG_RED | BG_BLACK);
		return(thechar);
	case cr:
	case lf: 
		sput(cr, FG_RED | BG_BLACK);
		sput(lf, FG_RED | BG_BLACK);
		return(lf);
	case bs:
	case del:
		sput(bs, FG_RED | BG_BLACK);
		sput(' ', FG_RED | BG_BLACK);
		sput(bs, FG_RED | BG_BLACK);
		return(thechar);
	case cntld:
		sput('^', FG_RED | BG_BLACK) ;
	 	sput('D', FG_RED | BG_BLACK) ; 
		sput('\r', FG_RED | BG_BLACK) ; 
		sput('\n', FG_RED | BG_BLACK) ;
		return(0);
	}
}


int pcmmap(dev_t dev, int offset, int nprot)
{
	if (offset > 0x20000)
		return -1;
	return i386_btop((0xa0000 + offset));
}


static void set_mode(int mode)
{
	u_char byte;

	switch(mode) {
	case TEXT80x25:
		outb(addr_6845, 9); byte = inb(addr_6845+1);
		outb(addr_6845, 9); outb(addr_6845+1, byte | 0x0F);
    		outb(0x3C4, 0x03); outb(0x3C5, 0x00);	/* select font 0 */
		break;
	case TEXT80x50:
		outb(addr_6845, 9); byte = inb(addr_6845+1);
		outb(addr_6845, 9); outb(addr_6845+1, (byte & 0xF0) | 0x07);
    		outb(0x3C4, 0x03); outb(0x3C5, 0x05);	/* select font 1 */
		break;
	}
}


static void set_border(int color)
{
  inb(0x3DA); 				/* reset flip-flop */
  outb(0x3C0,0x11); outb(0x3C0, color); 
  inb(0x3DA); 				/* reset flip-flop */
  outb(0x3C0, 0x20);			/* enable Palette */
}


static load_font(int segment, int size, char* font)
{
  	int ch, line;
    	extern char *atdevbase;
	inb(addr_6845 + 6);
	outb(0x3C0, 0x30); outb(0x3C0, 0x01);
	outb(0x3C4, 0x02); outb(0x3C5, 0x04);
	outb(0x3C4, 0x04); outb(0x3C5, 0x06);
	outb(0x3CE, 0x04); outb(0x3CF, 0x02);
	outb(0x3CE, 0x05); outb(0x3CF, 0x00);
	outb(0x3CE, 0x06); outb(0x3CF, 0x05);
    	for (ch=0; ch < 256; ch++) 
		for (line=0; line < size; line++) 
			*((char *)atdevbase+(segment*0x4000)+(ch*32)+line) = 
				font[(ch*size)+line];	
	inb(addr_6845 + 6);
	outb(0x3C0, 0x30); outb(0x3C0, 0x0C);
	outb(0x3C4, 0x02); outb(0x3C5, 0x03);
	outb(0x3C4, 0x04); outb(0x3C5, 0x02);
	outb(0x3CE, 0x04); outb(0x3CF, 0x00);
	outb(0x3CE, 0x05); outb(0x3CF, 0x10);
	outb(0x3CE, 0x06); outb(0x3CF, 0x0E);
}


static change_winsize(struct tty *tp, int x, int y)
{
	if (tp->t_winsize.ws_col != x || tp->t_winsize.ws_row != y) {
		tp->t_winsize.ws_col = x;
		tp->t_winsize.ws_row = y;
		pgsignal(tp->t_pgrp, SIGWINCH, 1);
	}
}

#endif
