/*-
 * Copyright (c) 1992 Eugene W. Stark
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Eugene W. Stark.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 5. This software may not be distributed for profit in either source or
 *    binary form without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EUGENE W. STARK (THE AUTHOR) ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Xten - Control X-10 devices from shell command level.
 * Eugene W. Stark (stark@cs.sunysb.edu)
 * November 25, 1992
 */

#include <stdio.h>
#include <time.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>

char *X10housenames[] = {
  "A", "B", "C", "D", "E", "F", "G", "H",
  "I", "J", "K", "L", "M", "N", "O", "P",
  NULL
};

char *X10cmdnames[] = {
  "1", "2", "3", "4", "5", "6", "7", "8",
  "9", "10", "11", "12", "13", "14", "15", "16",
  "AllUnitsOff", "AllLightsOn", "On", "Off", "Dim", "Bright", "AllLightsOff",
  "ExtendedCode", "HailRequest", "HailAcknowledge", "PreSetDim0", "PreSetDim1",
  "ExtendedData", "StatusOn", "StatusOff", "StatusRequest",
  NULL
};

/*
 * Flags for first byte of received packet
 */

#define TW_RCV_LOCAL	1  /* The packet arrived during a local transmission */
#define TW_RCV_ERROR	2  /* An invalid/corrupted packet was received */

main(argc, argv)
int argc;
char *argv[];
{
  int fd, sel;
  fd_set fs;
  unsigned char xpkt[3], rpkt[3];
  char *cp, *cp1;
  int h, k, c;
  struct timeval tv;
  time_t tod;
  char *tw523 = "/dev/tw0";

  if(argc < 3) {
    fprintf(stderr, "Usage: %s house key[:cnt] [ key[:cnt] ... ]\n", argv[0]);
    exit(1);
  }
  if((fd = open(tw523, O_RDWR)) < 0) {
    fprintf(stderr, "Can't open %s\n", tw523);
    exit(1);
  }
  if((h = find(argv[1], X10housenames)) < 0) {
    fprintf(stderr, "Invalid house code: %s\n", argv[1]);
    exit(1);
  }
  xpkt[0] = h;
  argv++;
  argv++;
  while(argc >= 3) {
    cp = argv[0];
    while(*cp != '\0' && *cp != ':') cp++;
    if(*cp == ':') c = atoi(cp+1);
    else c = 2;
    *cp = '\0';
    if((k = find(argv[0], X10cmdnames)) < 0) {
      fprintf(stderr, "Invalid key/unit code: %s\n", argv[0]);
      exit(1);
    }
    xpkt[1] = k;
    xpkt[2] = c;
    if(write(fd, xpkt, 3) != 3) {
      fprintf(stderr, "Transmission error\n");
      exit(1);
    }
    while(1) {
      FD_ZERO(&fs);
      FD_SET(fd, &fs);
      tv.tv_sec = 0;
      tv.tv_usec = 250000;
      sel = select(FD_SETSIZE, &fs, 0, 0, &tv);
      if(sel == 0 || !FD_ISSET(fd, &fs)) break;
      tod = time(NULL);
      cp = cp1 = ctime(&tod);
      while(*cp1 != '\n') cp1++;
      *cp1 = '\0';
      printf("%s:  ", cp);
      if(read(fd, rpkt, 3) < 3) {
	printf("Error reading from TW523\n");
      } else {
	printf("%s %s ", X10housenames[rpkt[1]],
	                           X10cmdnames[rpkt[2]]);
	switch(rpkt[0] & (TW_RCV_LOCAL | TW_RCV_ERROR)) {
	case 0:
	  printf("(remote,ok)");
	  break;
	case TW_RCV_LOCAL:
	  printf("(local,ok)");
	  break;
	case TW_RCV_ERROR:
	  printf("(remote,error)");
	  break;
	case TW_RCV_LOCAL|TW_RCV_ERROR:
	  printf("(local,error)");
	  break;
	}
	printf("\n");
      }
    }
    argc--;
    argv++;
  }
}

find(s, tab)
char *s;
char *tab[];
{
	int i;

	for(i = 0; tab[i] != NULL; i++) {
	  if(strcmp(s, tab[i]) == 0) return(i);
	}
	return(-1);
}
