static char sccsid[] = "@(#)87  1.25.2.4  R2/sysx/scsi/pscsidd.c, sysxscsi, bos320, 9308320 12/21/92 10:30:07";
/************************************************************************/
/* COMPONENT_NAME: (SYSXSCSI) IBM SCSI Adapter Driver                   */
/*                                                                      */
/* FUNCTIONS:                                                           */
/* psc_config           psc_open                psc_readblk             */
/* psc_close            psc_chip_register_init  psc_issue_abort         */
/* psc_fail_open        psc_config_adapter      psc_issue_BDR           */
/* psc_logerr           psc_read_reg            psc_adp_str_init        */
/* psc_write_reg        psc_read_POS            psc_ioctl               */
/* psc_write_POS        psc_strategy            psc_alloc_STA           */
/* psc_start            psc_start_dev           psc_free_STA            */
/* psc_stop_dev         psc_inquiry             psc_alloc_TCW           */
/* psc_start_unit       psc_test_unit_rdy       psc_free_TCW            */
/* psc_iodone           psc_build_command       psc_dumpwrt             */
/* psc_enq_active       psc_allocate_resources  psc_deq_active          */
/* psc_enq_wait         psc_free_resources      psc_deq_wait            */
/* psc_dump             psc_command_watchdog    psc_reset_iowait_jump   */
/* psc_dumpstrt         psc_suspend_watchdog    psc_update_dataptr      */
/* psc_dump_dev         psc_find_devinfo        psc_issue_abort_script  */
/* psc_fail_cmd         psc_fail_free_resources psc_issue_bdr_script    */
/* psc_intr             psc_clear_interrupts    psc_diagnostic          */
/* psc_issue_cmd        psc_issue_abort_bdr     psc_run_diagnostics     */
/* psc_check_wait_queue psc_scsi_parity_error   psc_loop_test_diagnotics*/
/* psc_cleanup_error    psc_script_init         psc_register_test       */
/* psc_cleanup_reset    psc_prep_main_script    psc_pos_register_test   */
/* psc_epow             psc_enq_wait_resources  psc_delay               */
/* psc_handle_extended_messages                                         */
/* psc_deq_wait_resources                                               */
/* psc_diag_reset_scsi_bus                                              */
/* psc_patch_async_switch_int                                           */
/* psc_verify_neg_answer						*/
/* psc_command_reset_scsi_bus                                           */
/* ORIGINS: 27                                                          */
/*                                                                      */
/* (C) COPYRIGHT International Business Machines Corp. 1989, 1990       */
/* All Rights Reserved                                                  */
/* Licensed Materials - Property of IBM                                 */
/*                                                                      */
/* US Government Users Restricted Rights - Use, duplication or          */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.    */
/*                                                                      */
/************************************************************************/
/*                                                                      */
/* COMPONENT:   SYSXSCSI                                                */
/*                                                                      */
/* NAME:        pscsidd.c                                               */
/*                                                                      */
/* FUNCTION:    IBM SCSI Adapter Driver Source File                     */
/*                                                                      */
/*      This adapter driver is the interface between a SCSI device      */
/*      driver and the actual SCSI adapter.  It executes commands       */
/*      from multiple drivers which contain generic SCSI device         */
/*      commands, and manages the execution of those commands.          */
/*      Several ioctls are defined to provide for system management     */
/*      and adapter diagnostic functions.                               */
/*                                                                      */
/* STYLE:                                                               */
/*                                                                      */
/*      To format this file for proper style, use the indent command    */
/*      with the following options:                                     */
/*                                                                      */
/*      -bap -ncdb -nce -cli0.5 -di8 -nfc1 -i4 -l78 -nsc -nbbb -lp      */
/*      -c4 -nei -nip                                                   */
/*                                                                      */
/*      Following formatting with the indent command, comment lines     */
/*      longer than 80 columns will need to be manually reformatted.    */
/*      To search for lines longer than 80 columns, use:                */
/*                                                                      */
/*      cat <file> | untab | fgrep -v sccsid | awk "length >79"         */
/*                                                                      */
/*      The indent command may need to be run multiple times.  Make     */
/*      sure that the final source can be indented again and produce    */
/*      the identical file.                                             */
/*                                                                      */
/************************************************************************/

/* INCLUDED SYSTEM FILES */
#include <sys/types.h>
#include <sys/sysmacros.h>
#include <sys/syspest.h>
#include <sys/dma.h>
#include <sys/sysdma.h>
#include <sys/ioacc.h>
#include <sys/intr.h>
#include <sys/malloc.h>
#include <sys/buf.h>
#include <sys/uio.h>
#include <sys/user.h>
#include <sys/file.h>
#include <sys/pin.h>
#include <sys/sleep.h>
#include <sys/ioctl.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/except.h>
#include <sys/param.h>
#include <sys/lockl.h>
#include <sys/priv.h>
#include <sys/watchdog.h>
#include <sys/device.h>
#include <sys/devinfo.h>
#include <sys/dump.h>
#include <sys/xmem.h>
#include <sys/time.h>
#include <sys/errids.h>
#include <sys/ddtrace.h>
#include <sys/trchkid.h>
#include <sys/trcmacros.h>
#include <sys/adspace.h>
#include <sys/scsi.h>
#include <sys/pscsidd.h>
/* END OF INCLUDED SYSTEM FILES  */

/************************************************************************/
/*                                                                      */
/* NAME:        psc_config                                              */
/*                                                                      */
/* FUNCTION:    Adapter Driver Configuration Routine                    */
/*                                                                      */
/*      For the INIT option, this routine allocates and initializes     */
/*      data structures required for processing user requests to the    */
/*      adapter.  If the TERM option is specified, this routine will    */
/*      delete a previously defined device and free the structures      */
/*      associated with it.  If the QVPD option is specified, this      */
/*      routine will return the adapter vital product data.             */
/*                                                                      */
/*      During device define time, this routine allocates and           */
/*      initializes data structures for processing of requests from     */
/*      the device driver heads to the scsi chip device driver.  The    */
/*      code is set up to handle only a single device instance.         */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can only be called by a process.                   */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      adapter_def - adapter unique data structure                     */
/*      adap_ddi - adapter dependent information structure              */
/*      uio     - user i/o area struct                                  */
/*      devsw   - kernel entry point struct                             */
/*                                                                      */
/* INPUTS:                                                              */
/*      devno   - device major/minor number                             */
/*      op      - operation code (INIT, TERM, or QVPD)                  */
/*      uiop    - pointer to uio structure for data for the             */
/*                specified operation code                              */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:  The following are the return values:      */
/*      EFAULT  - return from uiomove                                   */
/*      EBUSY   - on terminate, means device still opened               */
/*      ENOMEM  - memory space unavailable for required allocation      */
/*      EINVAL  - invalid config parameter passed                       */
/*      EIO     - bad operation, or permanent I/O error                 */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      lockl           unlockl                                         */
/*      devswadd        devswdel                                        */
/*      bcopy           bzero                                           */
/*      xmalloc         xmfree                                          */
/*      uiomove                                                         */
/*                                                                      */
/************************************************************************/
int
psc_config(
	   dev_t devno,
	   int op,
	   struct uio * uiop)
{
    struct devsw psc_dsw;
    struct adap_ddi local_ddi;
    int     ret_code, rc, i, data;
    uchar   vpd[VPD_SIZE];
    extern int nodev();

    DEBUG_0 ("Entering psc_config routine.\n")

    ret_code = 0;       /* default return code */
    rc = lockl(&(psc_lock), LOCK_SHORT);	/* serialize this */
    if (rc == LOCK_SUCC)
    {	/* lock succeded */
	switch (op)
	{	/* begin op switch */
	case CFG_INIT:
	    if (adp_str_inited)
	    {	/* pointer already initialized */
		ret_code = EIO;
		break;
	    }
	    else
	    {	/* new uninitialized pointer */
		/* set up entry points to the driver    */
		/* for the device switch table          */
		psc_dsw.d_open = (int (*) ()) psc_open;
		psc_dsw.d_close = (int (*) ()) psc_close;
		psc_dsw.d_read = nodev;
		psc_dsw.d_write = nodev;
		psc_dsw.d_ioctl = (int (*) ()) psc_ioctl;
		psc_dsw.d_strategy = (int (*) ()) psc_strategy;
		psc_dsw.d_ttys = 0;
		psc_dsw.d_select = nodev;
		psc_dsw.d_config = (int (*) ()) psc_config;
		psc_dsw.d_print = nodev;
		psc_dsw.d_dump = (int (*) ()) psc_dump;
		psc_dsw.d_mpx = nodev;
		psc_dsw.d_revoke = nodev;
		psc_dsw.d_dsdptr = 0;
		psc_dsw.d_selptr = 0;
		psc_dsw.d_opts = 0;

		rc = devswadd(devno, &psc_dsw);
		if (rc != 0)
		{	/* failed to add to dev switch table */
		    ret_code = EIO;
		    break;
		}

	    }	/* new uninitialized pointer */

	    /* move adapter configuration data (from uio  */
	    /* space) into local area (kernel space).     */
	    rc = uiomove((caddr_t) (&local_ddi),
			 (int) sizeof(struct adap_ddi),
			 UIO_WRITE, (struct uio *) uiop);
	    if (rc != 0)
	    {	/* copy failed */
		(void) devswdel(devno);	/* clean up */
		ret_code = EIO;
		break;
	    }

	    /* */
	    /* do any data validation here.  tcw area must  */
	    /* start on a tcw boundary, and must be twice   */
	    /* MAXREQUEST in length interrupt priority must */
	    /* be CLASS2, bus type must be Micro Channel,   */
	    /* and base address must be on a 4K boundary    */
	    /* */
	    if ((local_ddi.tcw_length < (2 * MAXREQUEST)) ||
		(local_ddi.tcw_start_addr & (PAGESIZE - 1)) ||
		(local_ddi.int_prior != INTCLASS2) ||
		(local_ddi.bus_type != BUS_MICRO_CHANNEL) ||
		!((local_ddi.base_addr == 0x0080)))
	    {	/* problem with ddi data */
		(void) devswdel(devno);	/* clean up */
		ret_code = EINVAL;
		break;
	    }

	    /* copy local ddi to global, static adapter    */
	    /* structure                                   */
	    bcopy(&local_ddi, &adp_str.ddi,
		  sizeof(struct adap_ddi));
	    /* do any required processing on the            */
	    /* configuration data                           */
	    adp_str.ddi.card_scsi_id &= 0x07;

	    /* if the tcw area is greater than 1MB plus the */
	    /* number of 4K tcws used, then scale up the    */
	    if (adp_str.ddi.tcw_length >
		(0x100000 + (TCW_TABLE_SIZE * PAGESIZE)))
		adp_str.max_request = (adp_str.ddi.tcw_length -
		     (0x100000 + (TCW_TABLE_SIZE * PAGESIZE))) +
		     MAXREQUEST;
	    else
		adp_str.max_request = MAXREQUEST;

	    /* This table reflects the min xfer period value */
	    /* returned by the device during negotiate	 */
	    adp_str.xfer_max[0] = 0x34;
	    adp_str.xfer_max[1] = 0x3E;
	    adp_str.xfer_max[2] = 0x48;
	    adp_str.xfer_max[3] = 0x53;
	    adp_str.xfer_max[4] = 0x5D;
	    adp_str.xfer_max[5] = 0x68;
	    adp_str.xfer_max[6] = 0x72;
	    adp_str.xfer_max[7] = 0xFF;

#ifdef PSC_TRACE
	    adp_str.current_trace_line = 1;
	    adp_str.trace_ptr = (struct psc_trace_struct *)
	    xmalloc(sizeof(struct psc_trace_struct), 4, pinned_heap);
#endif
	    TRACE_1 ("**PSCSTART*** ", 0)
	    TRACE_1 ("in config adp ", 0)
	    rc = psc_config_adapter();
	    TRACE_1 ("out config adp", 0)
	    if (rc != 0)
	    {	/* unsuccessful adapter config */
		(void) devswdel(devno); /* clean up */
		ret_code = EIO;
		break;
	    }

	    adp_str.devno = devno;
	    adp_str.opened = FALSE;
	    adp_str_inited = TRUE;
	    break;	/* CFG_INIT case */

	    /* handle request to terminate an adapter here      */
	case CFG_TERM:
	    if (!adp_str_inited)
	    {	/* device already deleted */
		ret_code = 0;
		break;
	    }

	    if (adp_str.opened)
	    {	/* error, someone else has it opened */
		ret_code = EBUSY;
		break;
	    }

	    /* Disable the chip */
	    (void) psc_write_POS(POS2, 0x00);

	    (void) devswdel(devno);	/* clean up */
	    adp_str_inited = FALSE;

	    break;	/* CFG_TERM case */

	    /* handle query for adapter VPD here                */
	case CFG_QVPD:
	    if (!adp_str_inited)
	    {	/* device structure not initialized */
		ret_code = ENXIO;
		break;
	    }

	    for (i = 0; i < VPD_SIZE; i++)
	    {	/* get VPD data one character at a time */
		if (psc_write_POS((uint) POS6, (int) (i + 1)) != 0)
		{
		    ret_code = EIO;
		    break;
		}
		data = psc_read_POS(POS3);
		if (data == -1)
		{
		    ret_code = EIO;
		    break;
		}
		vpd[i] = (uchar) data;	/* load VPD data */
	    }	/* get VPD data one character at a time */

	    /* leave POS6 in known state */
	    (void) psc_write_POS((uint) POS6, (int) 0x01);
	    /* move the VPD to the caller's uio structure */
	    rc = uiomove((caddr_t) vpd, (int) VPD_SIZE,
			 UIO_READ, (struct uio *) uiop);

	    if (rc != 0)
	    {	/* unsuccessful copy */
		ret_code = EIO;
	    }

	    break;	/* CFG_QVPD case */

	    /* handle invalid config parameter here */
	default:
	    ret_code = EINVAL;

	}
    }
    else
    {	/* lock failed */
	ret_code = EIO;	/* error--kernel service call failed */
	return (ret_code);
    }

    DEBUG_0 ("Leaving psc_config routine.\n")
    unlockl(&(psc_lock));
    return (ret_code);
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_open                                                */
/*                                                                      */
/* FUNCTION:    Adapter Driver Open Routine                             */
/*                                                                      */
/*      This routine opens the scsi chip and makes it ready.  It        */
/*      allocates adapter specific structures and initializes           */
/*      appropriate fields in them.  The adapter is marked as           */
/*      opened.                                                         */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      adapter_def - adapter unique data structure (one per adapter)   */
/*                                                                      */
/* INPUTS:                                                              */
/*      devno   - device major/minor number                             */
/*      devflag - unused                                                */
/*      chan    - unused                                                */
/*      ext     - extended data; this is 0 for normal use, or           */
/*                a value of SC_DIAGNOSTIC selects diagnostic mode      */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:  The following are the return values:      */
/*      EIO     - kernel service failed or invalid operation            */
/*      EPERM   - authority error                                       */
/*      EACCES  - illegal operation due to current mode (diag vs norm)  */
/*      ENOMEM  - not enough memory available                           */
/*          ENODEV      - device is not defined                         */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      lockl           unlockl                                         */
/*      pincode         unpincode                                       */
/*      privcheck       e_sleep                                         */
/*      i_init          i_clear                                         */
/*      xmalloc         xmfree                                          */
/*      dmp_add         dmp_del                                         */
/*      d_init          d_clear                                         */
/*      d_mask          d_unmask                                        */
/*      d_master        d_complete                                      */
/*      i_disable       i_enable                                        */
/*      w_init          w_start                                         */
/*                                                                      */
/************************************************************************/
int
psc_open(
	 dev_t devno,
	 ulong devflag,
	 int chan,
	 int ext)
{
    struct adapter_def *ap;
    int     rc, ret_code;
    int     dev, i;
    int     remaining_tcws;
    int     undo_level;
    uint    dma_addr;

    DEBUG_0 ("Entering psc_open routine.\n")

    ret_code = 0;
    undo_level = 0;

    DDHKWD5(HKWD_DD_SCSIDD, DD_ENTRY_OPEN, ret_code, devno, devflag,
	    chan, ext, 0);

    rc = lockl(&(psc_lock), LOCK_SHORT);	/* serialize this */
    if (rc == LOCK_SUCC)
    {	/* no problem with lock word */

	if (ext & SC_DIAGNOSTIC)
	    adp_str.open_mode = DIAG_MODE;
	else
	    adp_str.open_mode = NORMAL_MODE;

	if (!adp_str_inited)
	{
	    ret_code = EIO;
	    psc_fail_open(undo_level, ret_code, devno);
	    return (ret_code);
	}
	if ((ext & SC_DIAGNOSTIC) && adp_str.opened)
	{	/* cannot open in DIAG if already opened */
	    ret_code = EACCES;
	    psc_fail_open(undo_level, ret_code, devno);
	    return (ret_code);
	}
	if (adp_str.opened && (adp_str.open_mode == DIAG_MODE))
	{	/* cannot open if already opened in DIAG MODE */
	    ret_code = EACCES;
	    psc_fail_open(undo_level, ret_code, devno);
	    return (ret_code);
	}
	if ((ext & SC_DIAGNOSTIC) && (privcheck(RAS_CONFIG) != 0))
	{   /* trying to open in DIAG MODE without RAS_CONFIG authority */
	    ret_code = EPERM;
	    psc_fail_open(undo_level, ret_code, devno);
	    return (ret_code);
	}
	if ((privcheck(DEV_CONFIG) != 0) && (!(devflag & DKERNEL)))
	{	/* must be normal open */
	    ret_code = EPERM;
	    psc_fail_open(undo_level, ret_code, devno);
	    return (ret_code);
	}
	if (!adp_str.opened)
	{	/* not opened yet */
	    adp_str.errlog_enable = FALSE;

	    rc = pincode(psc_intr);
	    if (rc != 0)
	    {	/* pin failed */
		ret_code = EIO;
		psc_fail_open(undo_level, ret_code, devno);
		return (ret_code);
	    }
	    undo_level++;

	    if (adp_str.open_mode == NORMAL_MODE)
	    {
		/* init SIOP watchdog timer struct */
		adp_str.adap_watchdog.dog.next = NULL;
		adp_str.adap_watchdog.dog.prev = NULL;
		adp_str.adap_watchdog.dog.func = psc_command_watchdog;
		adp_str.adap_watchdog.dog.count = 0;
		adp_str.adap_watchdog.dog.restart = SIOPWAIT;
		adp_str.adap_watchdog.timer_id = PSC_SIOP_TMR;

		/* init reset watchdog timer struct */
		adp_str.reset_watchdog.dog.next = NULL;
		adp_str.reset_watchdog.dog.prev = NULL;
		adp_str.reset_watchdog.dog.func = psc_command_watchdog;
		adp_str.reset_watchdog.dog.count = 0;
		adp_str.reset_watchdog.dog.restart = RESETWAIT;
		adp_str.reset_watchdog.timer_id = PSC_RESET_TMR;

		/* init restart watchdog timer struct */
		adp_str.restart_watchdog.dog.next = NULL;
		adp_str.restart_watchdog.dog.prev = NULL;
		adp_str.restart_watchdog.dog.func = psc_command_watchdog;
		adp_str.restart_watchdog.dog.count = 0;
		adp_str.restart_watchdog.dog.restart = 0;
		adp_str.restart_watchdog.timer_id = PSC_RESTART_TMR;

		/* initialize the system timers */
		w_init(&adp_str.adap_watchdog.dog);
		w_init(&adp_str.reset_watchdog.dog);
		w_init(&adp_str.restart_watchdog.dog);
		undo_level++;

		/* initialize the adapter structure */
		psc_adp_str_init();
		rc = i_init(&(adp_str.intr_struct));
		if (rc != 0)
		{	/* i_init of scsi chip interrupt handler */
		    ret_code = EIO;
		    psc_fail_open(undo_level, ret_code, devno);
		    return (ret_code);
		}
		undo_level++;

		INIT_EPOW(&epow_struct, (int (*) ()) psc_epow,
			  adp_str.ddi.bus_id);
		rc = i_init(&epow_struct);
		if (rc != 0)
		{	/* i_init of epow structure failed */
		    ret_code = EIO;
		    psc_fail_open(undo_level, ret_code, devno);
		    return (ret_code);
		}
		undo_level++;

		rc = psc_write_reg((uint) DCNTL,
				   (uchar) DCNTL_SIZE, 0x01);
		if (rc != 0)
		{	/* error in chip reset */
		    ret_code = EIO;
		    psc_fail_open(undo_level, ret_code, devno);
		    return (ret_code);
		}	/* error in chip reset */
		delay(1 * HZ / 1000);
		rc |= psc_write_reg((uint) DCNTL,
				    (char) DCNTL_SIZE, 0x00);
		if (rc != 0)
		{	/* error in chip reset */
		    ret_code = EIO;
		    psc_fail_open(undo_level, ret_code, devno);
		    return (ret_code);
		}	/* error in chip reset */

		rc = psc_chip_register_init();
		if (rc != 0)
		{	/* error in chip setup */
		    ret_code = EIO;
		    psc_fail_open(undo_level, ret_code, devno);
		    return (ret_code);
		}

		/* allocate and set up the component dump table entry */
		psc_cdt = (struct psc_cdt_table *) xmalloc(
					(uint) sizeof(struct psc_cdt_table),
						     (uint) 2, pinned_heap);
		if (psc_cdt == NULL)
		{	/* error in dump table memory allocation */
		    ret_code = ENOMEM;
		    psc_fail_open(undo_level, ret_code, devno);
		    return (ret_code);
		}
		undo_level++;

		/* initialize the storage for the dump table */
		bzero((char *) psc_cdt, sizeof(struct psc_cdt_table));
		rc = dmp_add(psc_cdt_func);
		if (rc != 0)
		{
		    ret_code = ENOMEM;
		    psc_fail_open(undo_level, ret_code,
				  devno);
		    return (ret_code);
		}
		undo_level++;

		/* set up STA tables and malloc the work area   */
		/* set up index for STA in the TCW static table */
		adp_str.STA_tcw = 0;

		/* malloc the 4K work area for the STAs */
		adp_str.STA[0].sta_ptr = (char *) xmalloc((uint)
				     PAGESIZE, (uint) PGSHIFT, pinned_heap);
		if (adp_str.STA[0].sta_ptr == NULL)
		{	/* error in malloc for STA */
		    ret_code = ENOMEM;
		    psc_fail_open(undo_level, ret_code, devno);
		    return (ret_code);
		}
		adp_str.TCW_alloc_table[adp_str.STA_tcw] = TCW_RESERVED;
		DEBUG_1 ("STA Table@ = %x\n", adp_str.STA[0].sta_ptr)

		for     (i = 0; i < NUM_STA; i++)
		{	/* initialize the STA management table */
		    adp_str.STA[i].sta_ptr = adp_str.STA[0].sta_ptr +
			(i * ST_SIZE);
		    adp_str.STA[i].in_use = STA_UNUSED;
		}
		adp_str.next_STA_req = 0;
		undo_level++;

		/* set up the 4K Transfer indices into the static */
		/* TCW status table                               */
		adp_str.begin_4K_TCW = adp_str.STA_tcw + 2;
		adp_str.ending_4K_TCW = TCW_TABLE_SIZE - 1;
		adp_str.next_4K_req = adp_str.begin_4K_TCW;

		/* set up SCRIPTS tables and malloc the work area */
		/* malloc the initial 4K work area for the SCRIPTS */
		adp_str.SCRIPTS[0].script_ptr = (ulong *) xmalloc((uint)
				     PAGESIZE, (uint) PGSHIFT, pinned_heap);
		if (adp_str.SCRIPTS[0].script_ptr == NULL)
		{	/* error in malloc for scripts */
		    ret_code = ENOMEM;
		    psc_fail_open(undo_level, ret_code, devno);
		    return (ret_code);
		}
		DEBUG_1 ("SCRIPTS Table@ = %x\n",
			         adp_str.SCRIPTS[0].script_ptr)
		/* set up the area for scripts and copy. */
		/* set up index for SCRIPTS in the TCW static table */
		        dma_addr = DMA_ADDR(adp_str.ddi.tcw_start_addr,
					            adp_str.STA_tcw + 1);
		for (i = 0; i < INSTRUCTIONS * 2; i++)
		{
		    *(adp_str.SCRIPTS[0].script_ptr + i) =
			PSC_SCRIPT[i];
		}
		adp_str.SCRIPTS[0].dma_ptr = (ulong *) dma_addr;
		DEBUG_1 ("psc_open: SCRIPTS@=%x\n", adp_str.SCRIPTS)
		        adp_str.SCRIPTS[0].TCW_index = adp_str.STA_tcw + 1;
		adp_str.SCRIPTS[0].in_use = SCR_UNUSED;
		adp_str.TCW_alloc_table[adp_str.STA_tcw + 1] = TCW_RESERVED;
		adp_str.num_scripts_created = 1;
		undo_level++;

		/* Now we have to calculate the "left over space"   */
		/* for the Large Transfer Area.  We take the space  */
		/* taken up by the STA, 4K Transfer Area, and the   */
		/* SCRIPTS and find what is left from the original  */
		/* area given to us in our dds.                     */

		remaining_tcws = (adp_str.ddi.tcw_length -
				  TCW_TABLE_SIZE * PAGESIZE) / LARGESIZE;
		adp_str.large_tcw_start_addr = adp_str.ddi.tcw_start_addr +
		    (TCW_TABLE_SIZE * PAGESIZE);
		adp_str.large_TCW_table = (char *) xmalloc((uint)
				     remaining_tcws, (uint) 4, pinned_heap);
		if (adp_str.large_TCW_table == NULL)
		{	/* table malloc failed */
		    ret_code = ENOMEM;
		    psc_fail_open(undo_level, ret_code, devno);
		    return (ret_code);
		}
		for (i = 0; i < remaining_tcws; i++)
		    adp_str.large_TCW_table[i] = TCW_UNUSED;
		/* initialize the table to unused */
		adp_str.large_req_begin = 0;
		adp_str.next_large_req = 0;
		adp_str.large_req_end = remaining_tcws - 1;
		adp_str.errlog_enable = TRUE;
		adp_str.epow_state = 0;
		undo_level++;

		DEBUG_2 ("dma_lvl=%x, busid=%x\n",
			         adp_str.ddi.dma_lvl, adp_str.ddi.bus_id)
		        adp_str.channel_id = d_init(adp_str.ddi.dma_lvl,
				              DMA_INIT, adp_str.ddi.bus_id);
		if (adp_str.channel_id == DMA_FAIL)
		{	/* failed to init the dma channel */
		    ret_code = EIO;
		    psc_fail_open(undo_level, ret_code, devno);
		    return (ret_code);
		}

		d_unmask(adp_str.channel_id);
		/* lock in the dma areas for STA and SCRIPTS */
		dma_addr = DMA_ADDR(adp_str.ddi.tcw_start_addr,
				    adp_str.STA_tcw);
		DEBUG_2 ("tcw_start_address = %x, dma_addr = %x\n",
			         adp_str.ddi.tcw_start_addr, dma_addr)
		DEBUG_1 ("buffer = %x\n", adp_str.STA[0].sta_ptr)

		        d_master(adp_str.channel_id, DMA_NOHIDE, adp_str.
				         STA[0].sta_ptr, (size_t) PAGESIZE,
			              &adp_str.xmem_STA, (char *) dma_addr);

		/* execute d_master on the SCRIPTS area */
		dma_addr = (uint) adp_str.SCRIPTS[0].dma_ptr;
		DEBUG_2 ("tcw_start_address = %x, dma_addr = %x\n",
			         adp_str.ddi.tcw_start_addr, dma_addr)
		DEBUG_1 ("buffer = %x\n", adp_str.SCRIPTS[0].script_ptr)

		        d_master(adp_str.channel_id, DMA_NOHIDE,
		           adp_str.SCRIPTS[0].script_ptr, (size_t) PAGESIZE,
			             &adp_str.xmem_SCR, (caddr_t) dma_addr);

		DEBUG_1 ("TCW @ = %x\n", adp_str.TCW_alloc_table)
	    }	/* opened in normal mode */
	}	/* not opened yet */
    }	/* no problem with lock word */
    else
    {
	ret_code = ENOLCK;
	DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_OPEN, ret_code, devno);
	return (ret_code);
    }
    DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_OPEN, ret_code, devno);
    adp_str.opened = TRUE;
    unlockl(&(psc_lock));
    DEBUG_0 ("Leaving psc_open routine.\n")
    return (ret_code);
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_close                                               */
/*                                                                      */
/* FUNCTION:    Adapter Driver Close Routine                            */
/*                                                                      */
/*      This routine closes the scsi chip instance and releases any     */
/*      resources (as well as unpins the code) that were setup at open  */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      adapter_def - adapter unique data structure (one per adapter)   */
/*                                                                      */
/* INPUTS:                                                              */
/*      devno   - device major/minor number                             */
/*      chan    - 0; unused                                             */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:  The following are the return values:      */
/*      0       - successful                                            */
/*      EIO     - kernel service failed or invalid operation            */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      lockl           unlockl                                         */
/*      i_disable       i_enable                                        */
/*      e_sleep         w_clear                                         */
/*      d_mask          d_clear                                         */
/*      unpincode       xmfree                                          */
/*      i_clear                                                         */
/*                                                                      */
/************************************************************************/
int
psc_close(
	  dev_t devno,
	  int chan)
{
    int     ret_code, rc;
    int     i_hash, i;
    int     flag, old_pri;
    ulong   *save_page;


    DEBUG_0 ("Entering psc_close routine.\n")
            DDHKWD1(HKWD_DD_SCSIDD, DD_ENTRY_CLOSE, ret_code, devno);

    ret_code = 0;	/* default to good completion */
    rc = lockl(&(psc_lock), LOCK_SHORT);	/* serialize this */
    if (rc != LOCK_SUCC)
    {	/* lock kernel call failed */
	ret_code = EIO;
	DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_CLOSE, ret_code, devno);
	return (ret_code);
    }

    if (!adp_str.opened)
    {	/* adapter never opened */
	ret_code = EIO;
	DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_CLOSE, ret_code, devno);
	unlockl(&(psc_lock));
	return (ret_code);
    }

    /* Loop through the device hash table and stop devices */
    /* that are still started.                            */
    for (i = 0; i < MAX_DEVICES; i++)
    {
	if (adp_str.device_queue_hash[i] != NULL)
	{	/* If there's a device still started */
	    psc_stop_dev(i);	/* stops and frees resources */
	}
    }
    if (adp_str.open_mode == NORMAL_MODE)
    {	/* normal open */
	d_mask(adp_str.channel_id);
	d_clear(adp_str.channel_id);	/* free DMA channel */
	(void) xmfree((void *) adp_str.STA[0].sta_ptr,
		      pinned_heap);
	(void) xmfree((void *) adp_str.SCRIPTS[0].script_ptr,
		      pinned_heap);
	(void) xmfree((void *) adp_str.large_TCW_table,
		      pinned_heap);
	/* Disable dma and SCSI interrupts from the chip */
	(void) psc_write_reg((uint) SIEN, (uchar) SIEN_SIZE, 0x00);
	(void) psc_write_reg((uint) DIEN, (uchar) DIEN_SIZE, 0x00);

	i_clear(&(adp_str.intr_struct));
	/* clear and free the cdt_table */
	(void) dmp_del(psc_cdt_func);
	(void) xmfree((void *) psc_cdt, pinned_heap);
	i_clear(&epow_struct);
	/* clear all system timers */
	w_clear(&adp_str.adap_watchdog.dog);
	w_clear(&adp_str.reset_watchdog.dog);
	w_clear(&adp_str.restart_watchdog.dog);
    }

    (void) unpincode(psc_intr);
    adp_str.opened = FALSE;

    DEBUG_0 ("Leaving psc_close routine.\n")
    unlockl(&(psc_lock));
    DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_CLOSE, ret_code, devno);
    return (ret_code);

}  /* end psc_close */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_chip_register_init                                          */
/*                                                                        */
/* FUNCTION:  Initializes adapter chip registers.                         */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can only be called by a process.                     */
/*                                                                        */
/* NOTES:  This routine is called to initialize the adapter chip reg-     */
/*         isters to before command processing is to begin.               */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      adapter_def - scsi chip unique data structure                     */
/*                                                                        */
/* INPUTS:                                                                */
/*      None                                                              */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      None                                                              */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_chip_register_init()
{

    int     rc;

    rc = psc_write_reg((uint) SCNTL0, (char) SCNTL0_SIZE,
		       (uchar) SCNTL0_INIT);
    if (rc != 0)
	return (EIO);

    rc = psc_write_reg((uint) SCNTL1, (char) SCNTL1_SIZE,
		       (char) SCNTL1_INIT);
    if (rc != 0)
	return (EIO);

    rc = psc_write_reg((uint) SIEN, (char) SIEN_SIZE, (uchar) SIEN_MASK);
    if (rc != 0)
	return (EIO);

    rc = psc_write_reg((uint) SCID, (char) SCID_SIZE,
		       (uchar) (0x01 << adp_str.ddi.card_scsi_id));
    if (rc != 0)
	return (EIO);

    rc = psc_write_reg((uint) SXFER, (char) SXFER_SIZE, (uchar) SYNC_VAL);
    if (rc != 0)
	return (EIO);

    rc = psc_write_reg((uint) DMODE, (char) DMODE_SIZE,
		       (uchar) DMODE_INIT);
    if (rc != 0)
	return (EIO);

    rc = psc_write_reg((uint) DIEN, (char) DIEN_SIZE, (uchar) DIEN_MASK);
    if (rc != 0)
	return (EIO);

    rc = psc_write_reg((uint) DCNTL, (char) DCNTL_SIZE,
		       (uchar) DCNTL_INIT);
    if (rc != 0)
	return (EIO);

    return (PSC_NO_ERR);
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_chip_register_reset                                         */
/*                                                                        */
/* FUNCTION:  Resets SCSI chip.                                           */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can only be called by the interrupt handler.         */
/*                                                                        */
/* NOTES:  This routine is called to reset the adapter chip after a       */
/*         catastrophic error has occurred.                               */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      adapter_def - scsi chip unique data structure                     */
/*                                                                        */
/* INPUTS:                                                                */
/*      None                                                              */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      None                                                              */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*      d_mask          d_clear                                           */
/*      d_master                                                          */
/*                                                                        */
/**************************************************************************/
void
psc_chip_register_reset(int reset_flag)

{

    int     i, dma_addr;
    caddr_t iocc_addr;

    DEBUG_0 ("Entering psc_chip_register_reset routine.\n")
    TRACE_1 ("in chipR", 0)
    if (reset_flag & PSC_RESET_CHIP)
    {
	/* Reset SCSI in Component reset register   */
	iocc_addr = IOCC_ATT(adp_str.ddi.bus_id, 0);
	DEBUG_1 ("COMP_MASK1=%x\n", component_reset_val)
	(void) BUS_PUTLX((iocc_addr + PSC_COMP_RESET),
		          PSC_COMP_OFF);
	(void) BUS_PUTLX((iocc_addr + PSC_COMP_RESET),
	                  PSC_COMP_ON);
	IOCC_DET(iocc_addr);

	/* Reset the DMA channel for the chip */
	d_mask(adp_str.channel_id);
	d_clear(adp_str.channel_id);    /* free DMA channel */

	/* reset the POS registers for the chip */
	psc_config_adapter();

	/* Disable dma and SCSI interrupts from the chip */
	(void) psc_write_reg(SIEN, SIEN_SIZE, 0x00);
	(void) psc_write_reg(DIEN, DIEN_SIZE, 0x00);

	/* Assert a chip reset by writing to the DCNTL register */
	(void) psc_write_reg(DCNTL, DCNTL_SIZE, 0x01);
	psc_delay(1);
	(void) psc_write_reg(DCNTL, DCNTL_SIZE, 0x00);

	/* Call routine to re-initialize all chip registers  */
	(void) psc_chip_register_init();

	/* re-initialize DMA channel */
	adp_str.channel_id = d_init(adp_str.ddi.dma_lvl,
				    DMA_INIT, adp_str.ddi.bus_id);
	d_unmask(adp_str.channel_id);
    }

    if (reset_flag & PSC_RESET_DMA)
    {
	/* lock in the dma areas for STA and SCRIPTS */
	dma_addr = DMA_ADDR(adp_str.ddi.tcw_start_addr,
			    adp_str.STA_tcw);
	d_master(adp_str.channel_id, DMA_NOHIDE, adp_str.
		 STA[0].sta_ptr, (size_t) PAGESIZE,
		 &adp_str.xmem_STA, (char *) dma_addr);

	dma_addr = DMA_ADDR(adp_str.ddi.tcw_start_addr,
			    adp_str.STA_tcw + 1);
	d_master(adp_str.channel_id, DMA_NOHIDE,
		 adp_str.SCRIPTS[0].script_ptr, (size_t) PAGESIZE,
		 &adp_str.xmem_SCR, (char *) dma_addr);
	for (i = 0; i < MAX_SCRIPTS; i++)
	{
	    if (adp_str.SCRIPTS[i].script_ptr != NULL)
	    {
		dma_addr = DMA_ADDR(adp_str.ddi.tcw_start_addr,
				    adp_str.SCRIPTS[i].TCW_index);
		d_master(adp_str.channel_id, DMA_NOHIDE,
			 adp_str.SCRIPTS[i].script_ptr,
			 (size_t) PAGESIZE,
			 &adp_str.xmem_SCR, (char *) dma_addr);
	    }
	}
    }
    TRACE_1 ("outchipR", 0)
    DEBUG_0 ("Leaving psc_chip_register_reset routine.\n")
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_fail_open                                                   */
/*                                                                        */
/* FUNCTION:  Cleans up during failed "open" processing.                  */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can only be called by a process.                     */
/*                                                                        */
/* NOTES:  This routine is called to perform processing when a failure    */
/*         of the open occurs. This entails freeing storage, unpinning    */
/*         the code, etc.                                                 */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      adapter_def - scsi chip unique data structure                     */
/*                                                                        */
/* INPUTS:                                                                */
/*      undo_level  -  value of amount of cleanup required.               */
/*      ret_code    -  return value used for trace.                       */
/*      devno       -  device major/minor number.                         */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      None                                                              */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*      unpincode       xmfree                                            */
/*      dmp_del         i_clear                                           */
/*      w_clear         unlockl                                           */
/*                                                                        */
/**************************************************************************/
void
psc_fail_open(
	      int undo_level,
	      int ret_code,
	      dev_t devno)
{

    DEBUG_0 ("Entering psc_fail_open routine.\n")

    switch  (undo_level)
    {	/* begin switch */
    case 9:
	/* free dynamic table for Large Transfer Area */
	(void) xmfree((void *) adp_str.large_TCW_table,
		      pinned_heap);
    case 8:
	/* free memory taken by SCRIPTS */
	(void) xmfree((void *) adp_str.SCRIPTS[0].script_ptr,
		      pinned_heap);
    case 7:
	/* free memory taken by STA */
	(void) xmfree((void *) adp_str.STA[0].sta_ptr,
		      pinned_heap);
    case 6:
	/* clear the cdt_table */
	(void) dmp_del(psc_cdt_func);
    case 5:
	/* free the cdt_table */
	(void) xmfree((void *) psc_cdt, pinned_heap);
    case 4:
	/* clear the EPOW structure */
	i_clear(&epow_struct);
    case 3:
	/* clear out the scsi chip interrupt handler */
	i_clear(&(adp_str.intr_struct));
    case 2:
	/* clear all system timers */
	w_clear(&adp_str.adap_watchdog.dog);
	w_clear(&adp_str.reset_watchdog.dog);
	w_clear(&adp_str.restart_watchdog.dog);
    case 1:
	/* unpin the device driver */
	(void) unpincode(psc_intr);
    default:
	break;

    }	/* end switch */

    DEBUG_0 ("Leaving psc_fail_open routine.\n")
    DEBUG_0 ("Leaving psc_open through psc_fail_open (with errors)\n")
    DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_OPEN, ret_code, devno);
    unlockl(&(psc_lock));
    return;
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_config_adapter                                      */
/*                                                                      */
/* FUNCTION:    Adapter Configuration Routine                           */
/*                                                                      */
/*      This internal routine performs actions required to make         */
/*      the driver ready for the first open to the adapter.             */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine must be called from the process level only.        */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      Global_adapter_ddi                                              */
/*                                                                      */
/* INPUTS:                                                              */
/*      None                                                            */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:  The following are the return values:      */
/*      return code = 0 for successful return                           */
/*                  = EIO for unsuccessful operation                    */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      delay                                                           */
/*                                                                      */
/************************************************************************/
int
psc_config_adapter()
{
    int     ret_code;
    int     IPL_timer_count;
    int     val0, val1;
    uchar   int_level, data;

    DEBUG_0 ("Entering psc_config_adapter routine.\n")
    ret_code = 0;       /* default to no error */

    /* check here to make sure planer board is ready */
    /* for POS register set-up.                      */
    IPL_timer_count = 0;

    /* loop waiting for either card ready or timeout */

    while (TRUE)
    {	/* loop as long as it takes */
	val0 = psc_read_POS((uint) POS0);
	val1 = psc_read_POS((uint) POS1);
	if ((val0 == -1) || (val1 == -1))
	{	/* permanent I/O error */
	    ret_code = EIO;
	    break;
	}
	else
	{	/* no I/O error */
	    if (((uchar) val0 == POS0_VAL) && ((uchar) val1 == POS1_VAL))
		/* POS0 AND POS1 have both been set correctly */
		break;
	    /* card ready for POS reg loading */
	    else
	    {	/* planer IPL is still going */
		if (IPL_timer_count >= IPL_MAX_SECS)
		{	/* timed-out waiting for planer diag */
		    psc_logerr(ERRID_SCSI_ERR1,
			       UNKNOWN_ADAPTER_ERROR, 10, 0,
			       NULL, FALSE);
		    ret_code = EIO;
		    break;
		}
		/* */
		/* delay, then come back and check POS0 and   */
		/* POS1 again.                                */
		/* */
		psc_delay(1);
		IPL_timer_count++;
	    }	/* planer IPL is still going */
	}       /* no I/O error */
    }	/* loop as long as it takes */
    if (ret_code == 0)
    {	/* no errors so far, continue to initialize POS registers */
	/* Set up interrupt value for the chip */
	if (adp_str.ddi.int_lvl == 5)
	    int_level = 0x80;
	else
	    int_level = 0;
	if (psc_write_POS((uint) POS2, (int_level | 0x03)))
	    ret_code = EIO;
    }	/* no errors so far, continue to initialize POS registers */
    if (ret_code == 0)
    {
	data = psc_read_POS((uint) POS4);
	if (data == REG_FAIL)
	    ret_code = EIO;
	else
	    if (psc_write_POS((uint) POS4, ((data & 0xf0) |
					    adp_str.ddi.dma_lvl)))
		ret_code = EIO;
    }
    /* Clear any pending interrupts that may have been */
    /* left over from ipl.                             */
    if (psc_write_reg(ISTAT, ISTAT_SIZE, 0x00) == REG_FAIL)
	ret_code = EIO;
    else
	if (psc_read_reg(DSTAT, DSTAT_SIZE) == REG_FAIL)
	    ret_code = EIO;
	else
	    if (psc_read_reg(SSTAT0, SSTAT0_SIZE) == REG_FAIL)
		ret_code = EIO;

    DEBUG_0 ("Leaving psc_config_adapter routine.\n")
    return (ret_code);

}  /* end psc_config_adapter */

/************************************************************************/
/*                                                                      */
/* NAME:        psc_logerr                                              */
/*                                                                      */
/* FUNCTION:    Adapter Driver Error Logging Routine                    */
/*                                                                      */
/*      This routine provides a common point through which all driver   */
/*      error logging passes.                                           */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This is an internal routine which can be called from any        */
/*      other driver routine.                                           */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      adapter_def - adapter unique data structure (one per adapter)   */
/*      error_log_def - adapter driver error logging information        */
/*                      structure                                       */
/*                                                                      */
/* INPUTS:                                                              */
/*      errid   - the error log unique id for this entry                */
/*      add_hw_status - additional hardware status for this entry       */
/*      errnum  - a unique value which identifies which error is        */
/*                causing this call to log the error                    */
/*      data1   - additional, error dependent data                      */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:  The following are the return values:      */
/*      none                                                            */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      bcopy           bzero                                           */
/*      errsave                                                         */
/*                                                                      */
/************************************************************************/
void
psc_logerr(
	   int errid,
	   int add_hw_status,
	   int errnum,
	   int data1,
	   struct dev_info * dev_ptr,
	   uchar read_regs)
{
    struct error_log_def log;
    caddr_t iocc_addr;
    int     cmd_length, cmd_filler;
    int     script_size;
    int     dsp;

    DEBUG_0 ("Entering psc_logerr routine.\n")
    if (adp_str.errlog_enable)
    {	/* set up info and log */
	bzero(&log, sizeof(struct error_log_def));
	log.errhead.error_id = (uint) errid;
	bcopy(&adp_str.ddi.resource_name[0], &log.errhead.
	      resource_name[0], 8);
	log.data.diag_validity = 0;
	log.data.diag_stat = 0;
	if (add_hw_status != 0)
	{
	    log.data.ahs_validity = 0x01;
	    log.data.ahs = (uchar) add_hw_status;
	}
	log.data.un.card2.errnum = (uint) errnum;

	/* see if additional data is needed due to error id */
	if (((uint) errid == (uint) ERRID_SCSI_ERR7) ||
	    ((uint) errid == (uint) ERRID_SCSI_ERR8))
	    log.data.sys_rc = (uint) data1;

	/* indicate which register caused this error logging to be called */
	if ((((uint) errid == (uint) ERRID_SCSI_ERR1) ||
	     ((uint) errid == (uint) ERRID_SCSI_ERR2)) &&
	    (data1 != 0))
	    log.data.un.card2.reg_validity = data1;

	/* see if additional data is needed due to add_hw_status value */
	if ((uint) add_hw_status & DMA_ERROR)
	    log.data.sys_rc = (uint) data1;
	/* Get POS register data, directly read the regs   */
	/* here to avoid another call to error log         */
	/* routine due to error in read functions.         */
	iocc_addr = IOCC_ATT(adp_str.ddi.bus_id, 0);
	BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) +
		   POS0), &log.data.un.card2.pos0_val);
	BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) +
		   POS1), &log.data.un.card2.pos1_val);
	BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) +
		   POS2), &log.data.un.card2.pos2_val);
	BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) +
		   POS6), &log.data.un.card2.pos6_val);
	BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) +
		   POS3), &log.data.un.card2.pos3_val);
	BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) +
		   POS4), &log.data.un.card2.pos4_val);
	BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) +
		   POS5), &log.data.un.card2.pos5_val);
	IOCC_DET(iocc_addr);

	if (read_regs)
	{	/* on an interrupt level, read these regs */
	    iocc_addr = BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
	    BUS_GETCX(iocc_addr + ISTAT, &log.data.un.card2.istat_val);
	    BUS_GETCX(iocc_addr + SSTAT1, &log.data.un.card2.sstat1_val);
	    BUS_GETCX(iocc_addr + SSTAT2, &log.data.un.card2.sstat2_val);
	    BUS_GETCX(iocc_addr + SDID, &log.data.un.card2.sdid_val);
	    BUS_GETCX(iocc_addr + SCID, &log.data.un.card2.scid_val);
	    BUS_GETLX(iocc_addr + DBC, &log.data.un.card2.dbc_val);
	    BUS_GETLX(iocc_addr + DSP, &log.data.un.card2.dsp_val);
	    BUS_GETLX(iocc_addr + DSPS, &log.data.un.card2.dsps_val);
	    BUSIO_DET(iocc_addr);

	    log.data.un.card2.dbc_val = word_reverse(
					   log.data.un.card2.dbc_val);
	    log.data.un.card2.dsp_val = word_reverse(
					   log.data.un.card2.dsp_val);
	    log.data.un.card2.dsps_val = word_reverse(
					   log.data.un.card2.dsps_val);
	    script_size = INSTRUCTIONS * 8;
	    dsp = log.data.un.card2.dsps_val;
	    if ((dev_ptr != NULL) &&
		(dev_ptr->active != NULL))
	    {	/* found valid ptr to the job the MOST */
		/* likely caused us to go here */
		cmd_length = (int) dev_ptr->active->scsi_command.scsi_length;
		log.data.un.card2.scsi_cmd[0] = dev_ptr->active->scsi_command.
		    scsi_cmd.scsi_op_code;
		log.data.un.card2.scsi_cmd[1] = dev_ptr->active->scsi_command.
		    scsi_cmd.lun;
		for (cmd_filler = 2; cmd_filler < cmd_length; cmd_filler++)
		    log.data.un.card2.scsi_cmd[cmd_filler] =
			dev_ptr->active->scsi_command.scsi_cmd.
			scsi_bytes[cmd_filler - 2];
		log.data.un.card2.target_id = dev_ptr->scsi_id;
		log.data.un.card2.queue_state = dev_ptr->queue_state;
		log.data.un.card2.cmd_activity_state =
					       dev_ptr->cmd_activity_state;
		log.data.un.card2.resource_state = dev_ptr->resource_state;
	    }
	}

	/* log the error here */
	errsave(&log, sizeof(struct error_log_def));
    }
    DEBUG_0 ("Leaving psc_logerr routine.\n")

}  /* end psc_logerr */

/************************************************************************/
/*                                                                      */
/* NAME:        psc_read_reg                                            */
/*                                                                      */
/* FUNCTION:    Access the specified register.                          */
/*                                                                      */
/*      This routine reads from a selected adapter                      */
/*      register and returns the data, performing                       */
/*      appropriate error detection and recovery.                       */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This is an internal routine which can be called at any          */
/*      point to read a single register; 8,24 or 32 bit                 */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      adapter_def - adapter unique data structure (one per adapter)   */
/*                                                                      */
/* INPUTS:                                                              */
/*      offset  - offset to the selected register                       */
/*      reg_size - size of register in bytes                            */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:  The following are the return values:      */
/*      return code = 32-bits of data if good or recovered error        */
/*                  = 32-bits of -1 if permanent I/O error encountered  */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      none                                                            */
/*                                                                      */
/************************************************************************/
int
psc_read_reg(
	     uint offset,
	     char reg_size)
{
    int     ret_code, pio_rc, retry_count;
    int     val1, old_level;
    char    valc1;
    caddr_t iocc_addr;

    retry_count = 0;
    DEBUG_1 ("Entering psc_read_reg routine.REG=0x%x\n", offset);
    /* disable our interupts while reading registers */
    /* to prevent our interrupts killing the read    */
    old_level = i_disable(INTMAX);
    iocc_addr = BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
    if (reg_size == 1)
	pio_rc = BUS_GETCX((iocc_addr + offset), &valc1);
    else
	pio_rc = BUS_GETLX((iocc_addr + offset), &val1);
    while ((retry_count < MAX_POS_RETRIES) && (pio_rc == -1))
    {
	if ((offset == ISTAT) ||	/* Don't want to retry these */
	    (offset == DSTAT) ||	/* registers because rereading */
	    (offset == SSTAT0) ||	/* them corrupts the value */
	    (offset == SSTAT1) ||	/* orginally in them.  We will */
	    (offset == SSTAT2))	/* just log the error */
	    break;
	else
	    retry_count++;
	if (reg_size == 1)
	    pio_rc = BUS_GETCX((iocc_addr + offset), &valc1);
	else
	    pio_rc = BUS_GETLX((iocc_addr + offset), &val1);
    }

    if (pio_rc != 0)
    {	/* log as unrecoverable I/O bus problem */
	if (pio_rc == PIO_EXCP_DPRTY)
	    psc_logerr(ERRID_SCSI_ERR1, PIO_RD_DATA_ERR, 15,
		       offset, NULL, FALSE);
	else
	    psc_logerr(ERRID_SCSI_ERR1, PIO_RD_OTHR_ERR, 15,
		       offset, NULL, FALSE);
	ret_code = REG_FAIL;
    }
    else
    {
	if (retry_count > 0)
	    psc_logerr(ERRID_SCSI_ERR2, PIO_RD_OTHR_ERR, 20,
		       offset, NULL, FALSE);
	if (reg_size == 1)
	    ret_code = valc1;
	else
	    if (reg_size == 4)
		ret_code = val1;
	    else	/* reg size = 3 */
		ret_code = (val1 >> 8) & 0x00ffffff;
    }
    DEBUG_0 ("Leaving psc_read_reg routine.\n")
            BUSIO_DET(iocc_addr);
    i_enable(old_level);
    return (ret_code);
}  /* end psc_read_reg */

/************************************************************************/
/*                                                                      */
/* NAME:        psc_write_reg                                           */
/*                                                                      */
/* FUNCTION:    Store passed data in specified register.                */
/*                                                                      */
/*      This routine loads a selected adapter register with the         */
/*      passed data value, performing appropriate error detection       */
/*      and recovery.                                                   */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This is an internal routine which can be called at any          */
/*      point to load a single register.                                */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      adapter_def - adapter unique data structure (one per adapter)   */
/*                                                                      */
/* INPUTS:                                                              */
/*      offset  - offset to the selected POS reg                        */
/*      reg_size - size of register in bytes                            */
/*      data    - value to be loaded                                    */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:  The following are the return values:      */
/*      return code =  0 if good data or recovered error                */
/*                  = -1 if permanent I/O error encountered             */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      none                                                            */
/*                                                                      */
/************************************************************************/
int
psc_write_reg(
	      uint offset,
	      char reg_size,
	      int data)
{
    int     ret_code, pio_rc, retry_count;
    int     old_level;
    caddr_t iocc_addr;

    ret_code = 0;
    retry_count = 0;
    if (offset == SXFER)
    {
	TRACE_1 ("w SXFER ", data)
    }
    DEBUG_2 ("Entering psc_write_reg routine.REG=0x%x 0x%x\n", offset, data)
    /* disable our interupts while writing registers */
    /* to prevent our interrupts killing the read    */
            old_level = i_disable(INTMAX);
    iocc_addr = BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
    if (reg_size == 1)
	pio_rc = BUS_PUTCX((iocc_addr + offset), (char) data);
    else
	pio_rc = BUS_PUTLX((iocc_addr + offset), data);
    while ((retry_count < MAX_POS_RETRIES) && (pio_rc == -1))
    {
	retry_count++;
	if (reg_size == 1)
	    pio_rc = BUS_PUTCX((iocc_addr + offset), (char) data);
	else
	    pio_rc = BUS_PUTLX((iocc_addr + offset), data);
    }
    if (pio_rc != 0)
    {	/* log as unrecoverable I/O bus problem */
	if (pio_rc == PIO_EXCP_DPRTY)
	    psc_logerr(ERRID_SCSI_ERR1, PIO_WR_DATA_ERR,
		       25, offset, NULL, FALSE);
	else
	    psc_logerr(ERRID_SCSI_ERR1, PIO_WR_OTHR_ERR,
		       25, offset, NULL, FALSE);
	ret_code = REG_FAIL;
    }
    else
	if (retry_count > 1)
	{
	    psc_logerr(ERRID_SCSI_ERR2, PIO_WR_OTHR_ERR, 30,
		       offset, NULL, FALSE);
	}
    DEBUG_0 ("Leaving psc_write_reg routine.\n")
    BUSIO_DET(iocc_addr);
    i_enable(old_level);
    return (ret_code);

}  /* end psc_write_reg */

/************************************************************************/
/*                                                                      */
/* NAME:        psc_read_POS                                            */
/*                                                                      */
/* FUNCTION:    Access the specified register.                          */
/*                                                                      */
/*      This routine reads from a selected adapter POS                  */
/*      register and returns the data, performing                       */
/*      appropriate error detection and recovery.                       */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This is an internal routine which can be called at any          */
/*      point to read a single register 8 bit POS                       */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      adapter_def - adapter unique data structure (one per adapter)   */
/*                                                                      */
/* INPUTS:                                                              */
/*      offset  - offset to the selected register                       */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:  The following are the return values:      */
/*      return code = 32-bits of data if good or recovered error        */
/*                  = 32-bits of -1 if permanent I/O error encountered  */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      none                                                            */
/*                                                                      */
/************************************************************************/
int
psc_read_POS(
	     uint offset)
{
    int     reg_size, count, ret_code, err_flag, rc;
    int     old_level;
    char    val1, val2;
    caddr_t iocc_addr;

    ret_code = 0;
    reg_size = 1;
    count = 0;
    val1 = 0x1;
    val2 = 0x0;

    DEBUG_1 ("Entering psc_read_POS routine.REG=0x%x\n", offset);

    iocc_addr = IOCC_ATT(adp_str.ddi.bus_id, 0);
    old_level = i_disable(INTMAX);

    while ((count < MAX_POS_RETRIES) && ((val1 != val2) || (rc != 0)))
    {	/* bus miscompare, retry loop */
	rc = BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) + offset),
		       &val1);
	rc |= BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) + offset),
			&val2);
	count++;
    }

    if (count >= MAX_POS_RETRIES)
    {	/* log as unrecoverable I/O bus problem */
	psc_logerr(ERRID_SCSI_ERR1, PIO_RD_DATA_ERR, 35,
		   offset, NULL, FALSE);
	ret_code = -1;
    }
    else
	if (count > 1)
	    /* if we went into the while loop on an error, count will be
	       inc'd */
	{	/* log as recovered I/O bus problem */
	    psc_logerr(ERRID_SCSI_ERR2, PIO_RD_DATA_ERR, 40,
		       offset, NULL, FALSE);
	    ret_code = val1;	/* no error in bus access */
	    DEBUG_1 ("In read_POS, count = 0x%x\n", count);
	}
	else
	    ret_code = val1;	/* no error in bus access */

    DEBUG_0 ("Leaving psc_read_POS routine.\n")
    i_enable(old_level);
    IOCC_DET(iocc_addr);
    return (ret_code);
}  /* end psc_read_POS */

/************************************************************************/
/*                                                                      */
/* NAME:        psc_write_POS                                           */
/*                                                                      */
/* FUNCTION:    Store passed data in specified POS register.            */
/*                                                                      */
/*      This routine loads a selected adapter POS reg with the          */
/*      passed data value, performing appropriate error detection       */
/*      and recovery.                                                   */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This is an internal routine which can be called at any          */
/*      point to load a single POS register.                            */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      adapter_def - adapter unique data structure (one per adapter)   */
/*                                                                      */
/* INPUTS:                                                              */
/*      offset  - offset to the selected POS reg                        */
/*      data    - value to be loaded                                    */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:  The following are the return values:      */
/*      return code =  0 if good data or recovered error                */
/*                  = -1 if permanent I/O error encountered             */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      none                                                            */
/*                                                                      */
/************************************************************************/
int
psc_write_POS(
	      uint offset,
	      int data)

{
    int     count, count1, ret_code, err_flag, rc, i, j;
    int     temp, end_of_reg, old_level;
    char    data_byte[5], val1;
    char    reg_size;
    caddr_t iocc_addr;

    ret_code = 0;
    count = 0;
    val1 = 0x0;
    reg_size = 1;

    DEBUG_2 ("Entering psc_write_POS routine.REG=0x%x 0x%x\n", offset, data)
    iocc_addr = IOCC_ATT(adp_str.ddi.bus_id, 0);

    old_level = i_disable(INTMAX);
    /* strip out each byte so we can load the register */
    /* byte by byte (32 bits = 4 bytes)                */
    j = 4;
    for (i = 1; i < 5; i++)
    {
	temp = data << ((j - 1) * 8);
	temp &= 0xFF000000;
	temp = temp >> 24;
	data_byte[i] = (char) (temp & 0xFF);
	j--;
    }

    /* load register  */
    BUS_PUTCX((iocc_addr + (adp_str.ddi.slot << 16) + offset),
	      data_byte[1]);

    /* left justify the value so that we can compare it */
    /* to the value we get from BUS_GETSTRX             */
    rc = BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) + offset),
		   &val1);
    /************************************************************/
    /* There has been problems with BUS_PUT not putting values  */
    /* correctly on the first try, so lets compare what was     */
    /* written to what the data is. If it is incorrect then try */
    /* a few more times to put it correctly                     */
    /************************************************************/
    while ((count < MAX_POS_RETRIES) && ((val1 != data_byte[1]) ||
					 (rc != 0)))
    {	/* bus miscompare, retry loop */
	/* load register */
	BUS_PUTCX((iocc_addr + (adp_str.ddi.slot << 16) + offset),
		  data_byte[1]);
	rc = BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) + offset),
		       &val1);
	count++;
    }

    if (count >= MAX_POS_RETRIES)
    {	/* log as unrecoverable I/O bus problem */
	psc_logerr(ERRID_SCSI_ERR1, PIO_WR_DATA_ERR, 45,
		   offset, NULL, FALSE);
	ret_code = -1;
    }
    else
	if (count > 1)
	    /* if we went into the while loop on an error, count will be
	       inc'd */
	{	/* log as recovered I/O bus problem */
	    psc_logerr(ERRID_SCSI_ERR2, PIO_WR_DATA_ERR, 50,
		       offset, NULL, FALSE);
	    DEBUG_1 ("In write_POS, count = 0x%x\n", count);
	}

    DEBUG_0 ("Leaving psc_write_POS routine.\n")
    i_enable(old_level);
    IOCC_DET(iocc_addr);
    return (ret_code);

}  /* end psc_write_POS */

/************************************************************************/
/*                                                                      */
/* NAME:        psc_strategy                                            */
/*                                                                      */
/* FUNCTION: Scsi Chip Device Driver Strategy Routine                   */
/*                                                                      */
/*      This routine will accept commands from the device heads         */
/*      and queue them up to the appropriate device structure for       */
/*      execution.  The command, in the form of scbufs, contains        */
/*      a bufstruct at the beginning of the structure and pertinent     */
/*      data appended after cannot exceed the maximum transfer size     */
/*      allowed.  Note that the av_forw field of the bufstruct MUST     */
/*      be NULL as chained commands are not supported.                  */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can only be called by a process or interrupt       */
/*      handler.  It can page fault only if called under a process      */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      adapter_def - scsi chip unique data structure                   */
/*      scbuf - input command structure.                                */
/*                                                                      */
/* INPUTS:                                                              */
/*      bp - command buffer pointer                                     */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:  The following are the return values:      */
/*      0 - For good completion                                         */
/*      ERRNO - value otherwise                                         */
/*                                                                      */
/* ERROR DESCRIPTION                                                    */
/*      EIO - Adapter not defined or initialized                        */
/*          - Adapter not opened                                        */
/*          - Device not started                                        */
/*          - Device is being stopped                                   */
/*      ENXIO - Device queue is currently awaiting a resume cmd         */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      i_disable, i_enable                                             */
/*                                                                      */
/************************************************************************/
int
psc_strategy(
	     struct sc_buf * bp)
{
    int     ret_code;
    dev_t   devno;
    int     dev_index;
    int     new_pri;
    struct dev_info *dev_ptr;

    DEBUG_0 ("Entering psc_strategy routine \n")
    TRACE_1 ("in strat", bp)

    ret_code = 0;
    devno = bp->bufstruct.b_dev;
    DDHKWD5(HKWD_DD_SCSIDD, DD_ENTRY_STRATEGY, ret_code, devno, bp,
	    bp->bufstruct.b_flags, bp->bufstruct.b_blkno,
	    bp->bufstruct.b_bcount);

    if ((adp_str_inited) && (adp_str.opened))
    {
	/* passed the init and the open test */
	/* get index into device table for this device */
	dev_index = INDEX(bp->scsi_command.scsi_id,
			  (bp->scsi_command.scsi_cmd.lun) >> 5);
	dev_ptr = adp_str.device_queue_hash[dev_index];

	/* Disable interrupt to keep this single threaded. */
	new_pri = i_disable(adp_str.ddi.int_prior);

	/* miscellaneous validation of request */
	if ((dev_ptr != NULL) &&
	    (bp->bufstruct.b_bcount <= adp_str.max_request)
	    && (dev_ptr->queue_state != STOPPING)
	    && (bp->resvd1 == 0x0))
	{
	    /* passed check of device structure info */

	    /* The HALTED flag is set when the adapter driver */
	    /* encounters an error.  Do not continue the pro- */
	    /* cessing if the caller has not set RESUME flag. */
	    if ((dev_ptr->queue_state != HALTED) ||
		((dev_ptr->queue_state == HALTED)
		 && (bp->flags & SC_RESUME)))
	    {
		/* passed queue_state and RESUME tests */

		/* if we get here, device is not halted anymore */
		/* set normal state */
		dev_ptr->queue_state = ACTIVE;
		bp->bufstruct.av_forw = NULL;	/* only accept one cmd */
		bp->bufstruct.b_work = NULL;
		/* Initialize the following flag to the "no */
		/* error" state.                            */
		bp->bufstruct.b_error = 0;
		bp->bufstruct.b_flags &= ~B_ERROR;
		bp->bufstruct.b_resid = 0;
		bp->status_validity = 0;
		bp->general_card_status = 0;
		bp->scsi_status = 0;

		/* Queue this command to the device's pending      */
		/* queue for processing by psc_start.              */
		if (dev_ptr->head_pend == NULL)
		{	/* queue is empty  */
		    dev_ptr->head_pend = bp;
		    dev_ptr->tail_pend = bp;
		    if ((!(dev_ptr->restart_in_prog)) &&
			(!(dev_ptr->flags & CHK_SCSI_ABDR)) &&
			(dev_ptr->active == NULL))
			psc_start(dev_ptr);
		}
		else
		{	/* pending queue not empty */
		    /* point last cmd's av_forw at the new request */
		    dev_ptr->tail_pend->bufstruct.av_forw = (struct buf *) bp;
		    dev_ptr->tail_pend = bp;
		}
	    }
	    else
	    {	/* check queue_state and RESUME flag failed */
		bp->bufstruct.b_error = ENXIO;
		bp->bufstruct.b_flags |= B_ERROR;
		bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
		bp->status_validity = 0;
		iodone((struct buf *) bp);
		ret_code = ENXIO;
	    }
	}
	else
	{	/* device structure validation failed */
	    bp->bufstruct.b_error = EIO;
	    bp->bufstruct.b_flags |= B_ERROR;
	    bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
	    bp->status_validity = SC_ADAPTER_ERROR;
	    bp->general_card_status = SC_ADAPTER_SFW_FAILURE;
	    iodone((struct buf *) bp);
	    ret_code = EIO;
	}
    }
    else
    {	/* scsi chip has not been initialized or opened */
	bp->bufstruct.b_error = EIO;
	bp->bufstruct.b_flags |= B_ERROR;
	bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
	bp->status_validity = SC_ADAPTER_ERROR;
	bp->general_card_status = SC_ADAPTER_SFW_FAILURE;
	iodone((struct buf *) bp);
	DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_STRATEGY, EIO, devno);
	TRACE_1 ("out strat", EIO)
	return (EIO);
    }

    DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_STRATEGY, ret_code, devno);
    DEBUG_0 ("Leaving psc_strategy routine \n")
    TRACE_1 ("out strat", 0)
    i_enable(new_pri);
    return (ret_code);
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_start                                               */
/*                                                                      */
/* FUNCTION: Scsi Chip Device Driver Start Command Routine              */
/*                                                                      */
/*      This routine is called from the strategy routine when a command */
/*      becomes available for a device that does not currently have a   */
/*      command in the process of executing.  Resources are allocated,  */
/*      QUEUE, and an attempt is made to suspend the chip to get this   */
/*      command running.                                                */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can only be called by a process or interrupt       */
/*      handler.  It can page fault only if called under a process      */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      adapter_def - scsi chip unique data structure                   */
/*                                                                      */
/* INPUTS:                                                              */
/*      dev_index - index to device information                         */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      NONE                                                            */
/*                                                                      */
/* ERROR DESCRIPTION                                                    */
/*      NONE                                                            */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      w_start                                                         */
/*                                                                      */
/************************************************************************/
void
psc_start(
	  struct dev_info * dev_ptr)
{
    struct sc_buf *bp;
    int     ret_code, i, dev_index;
    uchar   istat_val, chk_disconnect;
    caddr_t iocc_addr;
    register int rc;

    DEBUG_0 ("Entering psc_start routine.\n")
    TRACE_1 ("in strt ", dev_ptr)
    bp = dev_ptr->head_pend;
    DDHKWD5(HKWD_DD_SCSIDD, DD_ENTRY_BSTART, ret_code, adp_str.devno,
	    bp, bp->bufstruct.b_flags, bp->bufstruct.b_blkno,
	    bp->bufstruct.b_bcount);

    if (adp_str.epow_state != EPOW_PENDING)
    {	/* non-EPOW state */
	/* Attempt to allocate the resource that are  */
	/* needed. i.e. TCW's, small xfer area, etc.  */
	ret_code = psc_alloc_resources(dev_ptr);
	if (ret_code == PSC_NO_ERR)
	{  /* Resources were allocated successfully */
	    if (adp_str.DEVICE_ACTIVE_head == NULL)
	    {	/* There's no commands currently active */
		if (!(dev_ptr->negotiate_flag))
		{	/* Negotiation not required, Assume SYNC */
		    /* Call routine to start command at the chip */
		    if (dev_ptr->async_device)
		    {
			rc = psc_write_reg(SXFER, SXFER_SIZE, 0x00);
			if (rc == REG_FAIL)
			{
			    bp->bufstruct.b_error = EIO;
			    bp->bufstruct.b_flags |= B_ERROR;
			    bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
			    bp->status_validity = SC_ADAPTER_ERROR;
			    bp->general_card_status = SC_ADAPTER_HDW_FAILURE;
			    (void) psc_free_resources(dev_ptr, FALSE);
			    iodone((struct buf *) bp);
			    dev_ptr->active = NULL;
			    /* Fail the rest of the pending queue. */
			    if (dev_ptr->head_pend != NULL)
				psc_fail_cmd(dev_ptr, ENXIO, 1);
			    DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_BSTART,
				    EIO, adp_str.devno);
			    return;
			}
		    }
		    else
			if ((dev_ptr->agreed_xfer !=
			     DEFAULT_MIN_PHASE) ||
			    (dev_ptr->agreed_req_ack <
			     DEFAULT_BYTE_BUF))
			{
			    rc = psc_write_reg(SXFER, SXFER_SIZE,
					       dev_ptr->special_xfer_val);
			    if (rc == REG_FAIL)
			    {
				bp->bufstruct.b_error = EIO;
				bp->bufstruct.b_flags |= B_ERROR;
				bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
				bp->status_validity = SC_ADAPTER_ERROR;
				bp->general_card_status =
						    SC_ADAPTER_HDW_FAILURE;
				(void) psc_free_resources(dev_ptr, FALSE);
				iodone((struct buf *) bp);
				dev_ptr->active = NULL;
				/* Fail the rest of the pending queue. */
				if (dev_ptr->head_pend != NULL)
				    psc_fail_cmd(dev_ptr, ENXIO, 1);
				DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_BSTART,
					EIO, adp_str.devno);
				return;
			    }
			}
		    chk_disconnect = bp->scsi_command.flags & SC_NODISC;
		    if ((dev_ptr->disconnect_flag ^ chk_disconnect) != 0)
			psc_set_disconnect(dev_ptr, chk_disconnect);
		    psc_prep_main_script(adp_str.SCRIPTS[0].script_ptr,
			adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].script_ptr,
					 dev_ptr, adp_str.SCRIPTS[0].dma_ptr,
					 dev_ptr->script_dma_addr);
		    ISSUE_MAIN_TO_DEVICE(dev_ptr->script_dma_addr);
		    dev_ptr->cmd_activity_state = CMD_IN_PROGRESS;
		}
		else
		    if (bp->scsi_command.flags & SC_ASYNC)
		    {	/* Negotiation not required, known ASYNC */
			/* Call routine to set asynch in script.     */
			rc = psc_write_reg(SXFER, SXFER_SIZE, 0x00);
			if (rc == REG_FAIL)
			{
			    bp->bufstruct.b_error = EIO;
			    bp->bufstruct.b_flags |= B_ERROR;
			    bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
			    bp->status_validity = SC_ADAPTER_ERROR;
			    bp->general_card_status = SC_ADAPTER_HDW_FAILURE;
			    (void) psc_free_resources(dev_ptr, FALSE);
			    iodone((struct buf *) bp);
			    dev_ptr->active = NULL;
			    /* Fail the rest of the pending queue. */
			    if (dev_ptr->head_pend != NULL)
				psc_fail_cmd(dev_ptr, ENXIO, 1);
			    DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_BSTART,
				    EIO, adp_str.devno);
			    return;
			}
			/* Call routine to start command at the chip */
			dev_index = INDEX(dev_ptr->scsi_id,
					  dev_ptr->lun_id);
			psc_patch_async_switch_int(adp_str.SCRIPTS
			   [dev_ptr->cmd_script_ptr].script_ptr, dev_index);
			chk_disconnect = bp->scsi_command.flags & SC_NODISC;
			if ((dev_ptr->disconnect_flag ^ chk_disconnect) != 0)
			    psc_set_disconnect(dev_ptr, chk_disconnect);
			psc_prep_main_script(adp_str.SCRIPTS[0].script_ptr,
			adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].script_ptr,
					dev_ptr, adp_str.SCRIPTS[0].dma_ptr,
		                        dev_ptr->script_dma_addr);
			ISSUE_MAIN_TO_DEVICE(dev_ptr->script_dma_addr);
			dev_ptr->cmd_activity_state = CMD_IN_PROGRESS;
			dev_ptr->async_device = TRUE;
			dev_ptr->negotiate_flag = FALSE;
		    }
		    else	/* Need to issue negotiate command */
		    {
			/* Call routine to start negotiate at the chip */
			chk_disconnect = bp->scsi_command.flags & SC_NODISC;
			if ((dev_ptr->disconnect_flag ^ chk_disconnect) != 0)
			    psc_set_disconnect(dev_ptr, chk_disconnect);
			ISSUE_NEGOTIATE_TO_DEVICE(dev_ptr->script_dma_addr);
			dev_ptr->cmd_activity_state = NEGOTIATE_IN_PROGRESS;
		    }
		/* Start the watchdog timer for this command */
		dev_ptr->dev_watchdog.dog.restart = (ulong)bp->timeout_value+1;
		w_start(&(dev_ptr->dev_watchdog.dog));

		/* Add to the DEVICE ACTIVE QUEUE */
		psc_enq_active(dev_ptr);
		DEBUG_0 ("Issuing command\n")
	    }
	    else
		if (adp_str.DEVICE_WAITING_head == NULL)
		{
		    /* Add to the DEVICE WAITING QUEUE */
		    /* A read of the istat register indicates that */
		    /* the scsi chip is currently not busy on the  */
		    /* scsi bus, so try to signal the chip that a  */
		    /* command is waiting.                         */
		    istat_val = -1;
		    iocc_addr =
			BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
		    psc_patch_iowait_int_on();
		    for (i = 0; i < 3; i++)
		    {
			rc = BUS_GETCX(iocc_addr + ISTAT, &istat_val);
			if (!(istat_val & CONN_INT_TEST))
			{	/* The register read is successful */
			    rc |= BUS_PUTCX(iocc_addr + ISTAT, 0x80);
			    if (rc != REG_FAIL)
			    {
				/* Start a timer for the chip suspend */
				w_start(&(adp_str.adap_watchdog.dog));
				break;
			    }
			    else
			    {
				/* Failed during a register write */
				istat_val = rc;
			    }
			}
			if (istat_val != REG_FAIL)
			{	/* Can't issue abort so  */
			    break;
			}
		    }
		    BUSIO_DET(iocc_addr);
		    if (istat_val == REG_FAIL)
		    {
			bp->bufstruct.b_error = EIO;
			bp->bufstruct.b_flags |= B_ERROR;
			bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
			bp->status_validity = SC_ADAPTER_ERROR;
			bp->general_card_status = SC_ADAPTER_HDW_FAILURE;
			(void) psc_free_resources(dev_ptr, FALSE);
			iodone((struct buf *) bp);
			dev_ptr->active = NULL;
			/* Fail the rest of the pending queue. */
			if (dev_ptr->head_pend != NULL)
			    psc_fail_cmd(dev_ptr, ENXIO, 1);
			DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_BSTART,
				EIO, adp_str.devno);
			return;
		    }
		    else
		    {
			psc_enq_wait(dev_ptr);
		    }
		}
		else
		{	/* There's other device waiting, so add to queue */
		    /* Add to the DEVICE WAITING QUEUE */
		    psc_enq_wait(dev_ptr);
		    DEBUG_0 ("Other device waiting\n")
		}
	}
	else
	{   /* Resources could not be allocated */
	    /* Add to the DEVICE WAITING FOR RESOURCES QUEUE */
	    psc_enq_wait_resources(dev_ptr);
	    DEBUG_0 ("Could not allocate resources\n")
	}
    }
    else
    {	/* An EPOW is occuring */
	DEBUG_0 ("EPOW!!!\n")
    }
    TRACE_1 ("out strt", dev_ptr)
    DEBUG_0 ("Leaving psc_start routine.\n")
    DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_BSTART, 0, adp_str.devno);
    return;
}

/**************************************************************************/
/*                                                                        */
/* NAME:        psc_start_dev                                             */
/*                                                                        */
/* FUNCTION:  Allocates resources and starts a device                     */
/*                                                                        */
/*      This routine initializes the device queue to prepare for          */
/*      command processing.                                               */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can only be called by a process or interrupt         */
/*      handler.  It can page fault only if called under a process        */
/*      and the stack is not pinned.                                      */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      adapter_def - scsi chip unique data structure                     */
/*                                                                        */
/* INPUTS:                                                                */
/*      dev_index - index to device information                           */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      0 - for good completion, ERRNO value otherwise.                   */
/*                                                                        */
/* ERROR DESCRIPTION                                                      */
/*      EINVAL - device already opened                                    */
/*      EACCES - adapter not opened in normal mode                        */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*      w_start                                                           */
/*                                                                        */
/**************************************************************************/
int
psc_start_dev(
	      int dev_index)
{

    int     i, old_level;
    struct dev_info *dev_ptr;
    int     scr_index, TCW_index;
    uint    dma_addr;

    DEBUG_0 ("Entering psc_start_dev routine.\n")

    if (adp_str.open_mode == DIAG_MODE)
    {  /* adapter opened in diagnostic mode */
	return (EACCES);
    }
    if (adp_str.device_queue_hash[dev_index] != NULL)
    {	/* device queue structure already allocated */
	return (EINVAL);
    }
    if (adp_str.ddi.card_scsi_id == SID(dev_index))
    {	/* Device SCSI ID matches adapter SCSI ID.  */
	return (EINVAL);
    }

    /* device queue structure doesn't exist yet */
    adp_str.device_queue_hash[dev_index] = xmalloc(
			    (uint) sizeof(struct dev_info), 4, pinned_heap);
    if (adp_str.device_queue_hash[dev_index] == NULL)
    {	/* malloc failed */
	return (ENOMEM);
    }
    dev_ptr = adp_str.device_queue_hash[dev_index];
    bzero(dev_ptr, (int) sizeof(struct dev_info));

    /* Search through the device queue area for a free script */
    /* instance.                                              */
    if (adp_str.SCRIPTS[0].in_use == SCR_UNUSED)
    {	/* A scripts sequence has been found */
	scr_index = 0;	/* used to address scripts */
    }
    else
	/* If no free script was found, then it's necessary  */
	/* to malloc another page and create additional      */
	/* script instances.                                 */
    {
	for (scr_index = 1; scr_index < MAX_SCRIPTS; scr_index++)
	{
	    if (adp_str.SCRIPTS[scr_index].script_ptr == NULL)
		break;
	}

	/* malloc an additional 4K work area for the SCRIPTS */
	adp_str.SCRIPTS[scr_index].script_ptr =
	    (ulong *) xmalloc((uint) PAGESIZE, (uint) PGSHIFT, pinned_heap);
	if (adp_str.SCRIPTS[scr_index].script_ptr == NULL)
	{	/* Unable to malloc script area. */
	    adp_str.device_queue_hash[dev_index] = NULL;
	    (void) xmfree(dev_ptr, pinned_heap);
	    return (ENOMEM);
	}
	/* find a tcw from the 4K xfer area to use */
	old_level = i_disable(adp_str.ddi.int_prior);
	for (i = adp_str.begin_4K_TCW; i < NUM_4K_TCWS; i++)
	{
	    if (adp_str.TCW_alloc_table[i] == TCW_UNUSED)
	    {
		adp_str.TCW_alloc_table[i] = TCW_RESERVED;
		TCW_index = i;
		break;
	    }
	}
	for (i = 0; i < NUM_4K_TCWS; i++)
	{
	    if (adp_str.TCW_alloc_table[i] != TCW_RESERVED)
	    {
		adp_str.begin_4K_TCW = i;
		if (adp_str.begin_4K_TCW > adp_str.next_4K_req)
		    adp_str.next_4K_req = adp_str.begin_4K_TCW;
		break;
	    }
	}
	i_enable(old_level);
	/* set up the area for scripts and copy. */
	/* set up index for SCRIPTS in the TCW static table */
	dma_addr = DMA_ADDR(adp_str.ddi.tcw_start_addr, TCW_index);
	for (i = 0; i < INSTRUCTIONS * 2; i++)
	{
	    *(adp_str.SCRIPTS[scr_index].script_ptr + i) =
		PSC_SCRIPT[i];
	}
	adp_str.SCRIPTS[scr_index].dma_ptr = dma_addr;
	DEBUG_1 ("psc_open: SCRIPTS@=%x\n", adp_str.SCRIPTS)
	        adp_str.SCRIPTS[scr_index].TCW_index = TCW_index;
	d_master(adp_str.channel_id, DMA_NOHIDE,
		 adp_str.SCRIPTS[scr_index].script_ptr,
		 (size_t) PAGESIZE,
		 &adp_str.xmem_SCR, (char *) dma_addr);
	adp_str.num_scripts_created++;
    }
    dev_ptr->script_dma_addr = adp_str.SCRIPTS[scr_index].dma_ptr;
    dev_ptr->cmd_script_ptr = scr_index;
    adp_str.SCRIPTS[scr_index].in_use = SCR_IN_USE;

    /* Call the setup routine for this script entry */
    psc_script_init((uint *) (adp_str.SCRIPTS[0].script_ptr),
		(uint *) (adp_str.SCRIPTS[scr_index].script_ptr), dev_index,
		    adp_str.SCRIPTS[0].dma_ptr, dev_ptr->script_dma_addr);

    /* Initialize device queue flags.             */
    dev_ptr->DEVICE_ACTIVE_fwd = NULL;
    dev_ptr->DEVICE_ACTIVE_bkwd = NULL;
    dev_ptr->DEVICE_WAITING_fwd = NULL;
    dev_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd = NULL;
    dev_ptr->ABORT_BDR_fwd = NULL;
    dev_ptr->ABORT_BDR_bkwd = NULL;
    dev_ptr->head_pend = NULL;
    dev_ptr->tail_pend = NULL;
    dev_ptr->active = NULL;
    dev_ptr->scsi_id = SID(dev_index);
    dev_ptr->lun_id = LUN(dev_index);
    dev_ptr->negotiate_flag = TRUE;
    dev_ptr->async_device = FALSE;
    dev_ptr->restart_in_prog = FALSE;
    dev_ptr->ioctl_wakeup = FALSE;
    dev_ptr->ioctl_event = EVENT_NULL;
    dev_ptr->stop_event = EVENT_NULL;
    dev_ptr->agreed_xfer = DEFAULT_MIN_PHASE;
    dev_ptr->agreed_req_ack = DEFAULT_BYTE_BUF;
    dev_ptr->flags = RETRY_ERROR;
    dev_ptr->disconnect_flag = 0;

    /* init watchdog timer struct                 */
    dev_ptr->dev_watchdog.dog.next = NULL;
    dev_ptr->dev_watchdog.dog.prev = NULL;
    dev_ptr->dev_watchdog.dog.func = psc_command_watchdog;
    dev_ptr->dev_watchdog.dog.count = 0;
    dev_ptr->dev_watchdog.dog.restart = 0;
    dev_ptr->dev_watchdog.timer_id = PSC_COMMAND_TMR;
    w_init(&(dev_ptr->dev_watchdog.dog));

    /* init command state flags                    */
    dev_ptr->queue_state = ACTIVE;
    dev_ptr->opened = TRUE;
    /* init the max_disconnect to 0 */
    dev_ptr->max_disconnect = 0;

    DEBUG_1 ("Device pointer address = %x\n", dev_ptr)
    DEBUG_0 ("Leaving psc_start_dev routine.\n")
    return (0);
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_stop_dev                                                    */
/*                                                                        */
/* FUNCTION:  Stops a device and deallocates resources.                   */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine is called to stop a device.  Any further command  */
/*         processing for the device will be halted from this point on.   */
/*                                                                        */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      adapter_def - scsi chip unique data structure                     */
/*                                                                        */
/* INPUTS:                                                                */
/*      dev_index - index to device information                           */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      0 - for good completion, ERRNO value otherwise.                   */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*    EINVAL - device already opened.                                     */
/*    EACCES - Adapter not opened in normal mode.                         */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED: i_enable, i_disable, e_sleep, w_clear      */
/*                                                                        */
/**************************************************************************/
int
psc_stop_dev(
	     int dev_index)
{
    int     i, old_level, TCW_index;
    struct dev_info *dev_ptr;
    struct sc_buf *bp;

    DEBUG_0 ("Entering psc_stop_dev routine.\n")

    if (adp_str.open_mode == DIAG_MODE)
    {	/* adapter opened in diagnostic mode */
	return (EACCES);
    }
    if (adp_str.device_queue_hash[dev_index] == NULL)
    {	/* device queue structure already allocated */
	return (EINVAL);
    }
    /* Obtain device structure pointer and disable */
    /* interrupts for processing.                 */
    dev_ptr = adp_str.device_queue_hash[dev_index];
    old_level = i_disable(adp_str.ddi.int_prior);
    dev_ptr->queue_state = STOPPING;

    /* Search both the active and pending queue of */
    /* this device structure to determine if there */
    /* are any commands outstanding.              */
    while ((dev_ptr->head_pend != NULL) || (dev_ptr->active != NULL))
    {
	e_sleep(&dev_ptr->stop_event, EVENT_SHORT);
    }
    i_enable(old_level);	/* let interrupts in */

    /* Free the device's script entry for other   */
    /* use.                                       */
    /* script 0 is freed when the device is closed */
    if (dev_ptr->cmd_script_ptr != 0)
    {
	old_level = i_disable(adp_str.ddi.int_prior);
	TCW_index = adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].TCW_index;
	adp_str.TCW_alloc_table[TCW_index] = TCW_UNUSED;
	for (i = 0; i < NUM_4K_TCWS; i++)
	{
	    if (adp_str.TCW_alloc_table[i] != TCW_RESERVED)
	    {
		adp_str.begin_4K_TCW = i;
		if (adp_str.begin_4K_TCW > adp_str.next_4K_req)
		    adp_str.next_4K_req = adp_str.begin_4K_TCW;
		break;
	    }
	}
	i_enable(old_level);	/* let interrupts in */
	adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].in_use = SCR_UNUSED;
	adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].dma_ptr = 0;
	adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].TCW_index = 0;
	(void) xmfree(adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].script_ptr,
		      pinned_heap);
	adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].script_ptr = NULL;
	adp_str.num_scripts_created--;
    }
    else
    {
	adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].in_use = SCR_UNUSED;
    }
    /* Free the device's watchdog timer entry.   */
    w_clear(&(dev_ptr->dev_watchdog.dog));
    /* Free the device information table and clear */
    /* the hash table entry.                      */
    adp_str.device_queue_hash[dev_index] = NULL;
    (void) xmfree(dev_ptr, pinned_heap);
    DEBUG_0 ("Leaving psc_stop_dev routine.\n")
    return (0);
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_inquiry                                                     */
/*                                                                        */
/* FUNCTION:  Issues a SCSI inquiry command to a device.                  */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine is called to issue an inquiry to a device.        */
/*         The calling process is responsible for NOT calling this rou-   */
/*         tine if the SCIOSTART failed.  Such a failure would indicate   */
/*         that another process has this device open and interference     */
/*         could cause improper error reporting.                          */
/*                                                                        */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      adapter_def - scsi chip unique data structure                     */
/*                                                                        */
/* INPUTS:                                                                */
/*      devno   - passed device major/minor number                        */
/*      arg     - passed argument value                                   */
/*      devflag - device flag                                             */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      0 - for good completion, ERRNO value otherwise.                   */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*      EINVAL - Device already opened.                                   */
/*      EACCES - Adapter not opened in normal mode.                       */
/*      ENOMEM - Could not allocate an scbuf for this command.            */
/*      ENODEV - Device could not be selected.                            */
/*      ETIMEDOUT - The inquiry command timed out.                        */
/*      EIO    - Error returned from psc_strategy.                        */
/*      EIO    - No data returned from inquiry command.                   */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:  iowait, xmfree, copyin, copyout, bcopy    */
/*                                                                        */
/**************************************************************************/
int
psc_inquiry(
	    dev_t devno,
	    int arg,
	    ulong devflag)
{
    int     ret_code, dev_index, inquiry_length;
    struct sc_buf *bp;
    struct sc_inquiry sc_inq;

    DEBUG_0 ("Entering psc_inquiry routine.\n")
    ret_code = 0;
    /* Copy in the arg structure. If the buffer resides */
    /* user space, use copyin, else use bcopy.          */
    if (!(devflag & DKERNEL))
    {
	ret_code = copyin((char *) arg, &sc_inq,
			  sizeof(struct sc_inquiry));
	if (ret_code != 0)
	{
	    DEBUG_0 ("Leaving psc_inquiry routine.\n")
	    return (EFAULT);
	}
    }
    else
    {	/* buffer is in kernel space */
	bcopy((char *) arg, &sc_inq, sizeof(struct sc_inquiry));
    }
    if (adp_str.open_mode == DIAG_MODE)
    {	/* adapter opened in diagnostic mode */
	DEBUG_0 ("Leaving psc_inquiry routine.\n")
	return (EACCES);
    }
    /* Obtain device structure from hash on the scsi id */
    /* and lun id.                                      */
    dev_index = INDEX(sc_inq.scsi_id, sc_inq.lun_id);

    if (adp_str.device_queue_hash[dev_index] == NULL)
    {	/* device queue structure already allocated */
	DEBUG_0 ("Leaving psc_inquiry routine.\n")
	return (EINVAL);
    }
    bp = psc_build_command();
    if (bp == NULL)
    {
	DEBUG_0 ("Leaving psc_inquiry routine.\n")
	return (ENOMEM);        /* could'nt allocate command */
    }

    bp->scsi_command.scsi_id = sc_inq.scsi_id;
    bp->scsi_command.scsi_length = 6;
    bp->scsi_command.scsi_cmd.scsi_op_code = SCSI_INQUIRY;
    if (sc_inq.get_extended) 
    {
        bp->scsi_command.scsi_cmd.lun = (sc_inq.lun_id << 5) | 0x01;
        bp->scsi_command.scsi_cmd.scsi_bytes[0] = sc_inq.code_page_num;
    }
    else
    {
        bp->scsi_command.scsi_cmd.lun = sc_inq.lun_id << 5;
        bp->scsi_command.scsi_cmd.scsi_bytes[0] = 0;
    }
    bp->scsi_command.scsi_cmd.scsi_bytes[1] = 0;
    /* Always set for the maximum amt of inquiry data */
    bp->scsi_command.scsi_cmd.scsi_bytes[2] = 255;
    bp->scsi_command.scsi_cmd.scsi_bytes[3] = 0;

    /* Set the no disconnect flag and the synch/asynch flag  */
    /* for proper data tranfer to occur.                     */
    bp->scsi_command.flags = (sc_inq.flags & SC_ASYNC) | SC_NODISC;

    bp->bufstruct.b_bcount = 255;
    bp->bufstruct.b_flags |= B_READ;
    bp->bufstruct.b_dev = devno;

    /* Set resume flag in case caller is retrying this operation. */
    /* This assumes the inquiry is only running single-threaded  */
    /* to this device.                                           */
    bp->flags = SC_RESUME;

    /* Call our strategy routine to issue the inquiry command. */
    if (psc_strategy(bp))
    {	/* an error was returned */
	(void) xmfree(bp, pinned_heap);	/* release buffer */
	return (EIO);
	DEBUG_0 ("Leaving psc_inquiry routine.\n")
    }

    iowait((struct buf *) bp);  /* Wait for commmand completion */

    /* The return value from the operation is examined to deter- */
    /* what type of error occurred.  Since the calling applica-  */
    /* requires an ERRNO, this value is interpreted here.        */
    if (bp->bufstruct.b_flags & B_ERROR)
    {	/* an error occurred */
	if (bp->status_validity & SC_ADAPTER_ERROR)
	{	/* if adapter error */
	    switch (bp->general_card_status)
	    {
	    case SC_CMD_TIMEOUT:
		psc_logerr(ERRID_SCSI_ERR2, COMMAND_TIMEOUT,
			   55, 0, NULL, FALSE);
		ret_code = ETIMEDOUT;
		break;
	    case SC_NO_DEVICE_RESPONSE:
		ret_code = ENODEV;
		break;
	    case SC_SCSI_BUS_FAULT:
		ret_code = ENOCONNECT;
		break;
	    default:
		ret_code = EIO;
		break;
	    }
	}
	else
	{
	    ret_code = EIO;
	}
    }

    /* if no other errors, and yet no data came back, then fail */
    if ((ret_code == 0) &&
	(bp->bufstruct.b_resid == bp->bufstruct.b_bcount))
    {
	ret_code = EIO;
    }
    /* give the caller the lesser of what he asked for, or */
    /* the actual transfer length                          */
    if (ret_code == 0)
    {
	inquiry_length = bp->bufstruct.b_bcount - bp->bufstruct.b_resid;
	if (inquiry_length > sc_inq.inquiry_len)
	    inquiry_length = sc_inq.inquiry_len;
	/* Copy out the inquiry data. If the buffer resides */
	/* user space, use copyin, else use bcopy.          */
	if (!(devflag & DKERNEL))
	{
	    ret_code = copyout(bp->bufstruct.b_un.b_addr,
			       sc_inq.inquiry_ptr, inquiry_length);
	    if (ret_code != 0)
	    {
		ret_code = EFAULT;
	    }
	}
	else
	{ /* buffer is in kernel space */
	    bcopy(bp->bufstruct.b_un.b_addr, sc_inq.inquiry_ptr,
		  inquiry_length);
	}
    }
    (void) xmfree(bp, pinned_heap);	/* release buffer */
    DEBUG_1 ("Leaving psc_inquiry routine. %d\n", ret_code);
    return (ret_code);
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_start_unit                                                  */
/*                                                                        */
/* FUNCTION:  Issues a SCSI start unit command to a device.               */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine is called to to issue a start unit command to a   */
/*         device.                                                        */
/*         The calling process is responsible for NOT calling this rou-   */
/*         tine if the SCIOSTART failed.  Such a failure would indicate   */
/*         that another process has this device open and interference     */
/*         could cause improper error reporting.                          */
/*                                                                        */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      adapter_def - scsi chip unique data structure                     */
/*                                                                        */
/* INPUTS:                                                                */
/*      devno   - passed device major/minor number                        */
/*      arg     - passed argument value                                   */
/*      devflag - device flag                                             */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      0 - for good completion, ERRNO value otherwise.                   */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*    EINVAL - Device already opened.                                     */
/*    EACCES - Adapter not opened in normal mode.                         */
/*    ENOMEM - Could not allocate an scbuf for this command.              */
/*    ENODEV - Device could not be selected.                              */
/*    ETIMEDOUT - The start unit command timed out.                       */
/*    EIO    - Error returned from psc_strategy.                          */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:  iowait, xmfree, copyin, bcopy             */
/*                                                                        */
/**************************************************************************/
int
psc_start_unit(
	       dev_t devno,
	       int arg,
	       ulong devflag)
{
    int     ret_code, dev_index;
    struct sc_buf *bp;
    struct sc_startunit sc_stun;

    DEBUG_0 ("Entering psc_start_unit.\n")
            ret_code = 0;
    /* Copy in the arg structure. If the buffer resides */
    /* user space, use copyin, else use bcopy.          */
    if (!(devflag & DKERNEL))
    {
	ret_code = copyin((char *) arg, &sc_stun,
			  sizeof(struct sc_startunit));
	if (ret_code != 0)
	{
	    DEBUG_0 ("EFAULT: Leaving psc_start_unit routine.\n")
	    return (EFAULT);
	}
    }
    else
    {	/* buffer is in kernel space */
	bcopy((char *) arg, &sc_stun, sizeof(struct sc_startunit));
    }
    if (adp_str.open_mode == DIAG_MODE)
    {	/* adapter opened in diagnostic mode */
	DEBUG_0 ("Leaving psc_start_unit routine. EACCES\n")
	return (EACCES);
    }
    /* Obtain device structure from hash on the scsi id */
    /* and lun id.                                      */
    dev_index = INDEX(sc_stun.scsi_id, sc_stun.lun_id);

    if (adp_str.device_queue_hash[dev_index] == NULL)
    {	/* device queue structure not allocated */
	DEBUG_0 ("Leaving psc_start_unit .EINVAL\n")
	return (EINVAL);
    }
    bp = psc_build_command();
    if (bp == NULL)
    {
	DEBUG_0 ("Leaving psc_start_unit ENOMEM.\n")
	return (ENOMEM);        /* could'nt allocate command */
    }

    bp->scsi_command.scsi_id = sc_stun.scsi_id;
    bp->scsi_command.scsi_length = 6;
    bp->scsi_command.scsi_cmd.scsi_op_code = SCSI_START_STOP_UNIT;
    /* The immediate bit for this command is set depending */
    /* on the value of the immediate flag.                 */
    bp->scsi_command.scsi_cmd.lun = (sc_stun.lun_id << 5) |
	(sc_stun.immed_flag ? 0x01 : 0);
    bp->scsi_command.scsi_cmd.scsi_bytes[0] = 0;
    bp->scsi_command.scsi_cmd.scsi_bytes[1] = 0;
    /* Set the command start flag */
    bp->scsi_command.scsi_cmd.scsi_bytes[2] =
	(sc_stun.start_flag ? 0x01 : 0);
    bp->scsi_command.scsi_cmd.scsi_bytes[3] = 0;

    /* Set the no disconnect flag and the synch/asynch flag  */
    /* for proper data tranfer to occur.                     */
    bp->scsi_command.flags = (sc_stun.flags & SC_ASYNC) | SC_NODISC;

    bp->bufstruct.b_bcount = 0;
    bp->bufstruct.b_dev = devno;
    bp->timeout_value = sc_stun.timeout_value;	/* set timeout value */

    /* Set resume flag in case caller is retrying this operation. */
    /* This assumes the inquiry is only running single-threaded  */
    /* to this device.                                           */
    bp->flags = SC_RESUME;

    /* Call our strategy routine to issue the start unit cmd. */
    if (psc_strategy(bp))
    {	/* an error was returned */
	(void) xmfree(bp, pinned_heap);	/* release buffer */
	return (EIO);
	DEBUG_0 ("Leaving psc_start_unit .\n")
    }

    iowait((struct buf *) bp);  /* Wait for commmand completion */

    /* The return value from the operation is examined to deter- */
    /* what type of error occurred.  Since the calling applica-  */
    /* requires an ERRNO, this value is interpreted here.        */
    if (bp->bufstruct.b_flags & B_ERROR)
    {	/* an error occurred */
	if (bp->status_validity & SC_ADAPTER_ERROR)
	{	/* if adapter error */
	    switch (bp->general_card_status)
	    {
	    case SC_CMD_TIMEOUT:
		psc_logerr(ERRID_SCSI_ERR2, COMMAND_TIMEOUT,
			   60, 0, NULL, FALSE);
		ret_code = ETIMEDOUT;
		break;
	    case SC_NO_DEVICE_RESPONSE:
		ret_code = ENODEV;
		break;
	    case SC_SCSI_BUS_FAULT:
		ret_code = ENOCONNECT;
		break;
	    default:
		ret_code = EIO;
		break;
	    }
	}
	else
	{
	    ret_code = EIO;
	}
    }

    (void) xmfree(bp, pinned_heap);	/* release buffer */
    DEBUG_0 ("Leaving psc_start_unit .\n")
    return (ret_code);
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_test_unit_rdy                                               */
/*                                                                        */
/* FUNCTION:  Issues a SCSI test unit ready command to a device.          */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine is called to to issue a test unit ready command   */
/*         to a device.                                                   */
/*         The calling process is responsible for NOT calling this rou-   */
/*         tine if the SCIOSTART failed.  Such a failure would indicate   */
/*         that another process has this device open and interference     */
/*         could cause improper error reporting.                          */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      adapter_def - scsi chip unique data structure                     */
/*                                                                        */
/* INPUTS:                                                                */
/*      devno   - passed device major/minor number                        */
/*      arg     - passed argument value                                   */
/*      devflag - device flag                                             */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      0 - for good completion, ERRNO value otherwise.                   */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*    EINVAL - Device already opened.                                     */
/*    EACCES - Adapter not opened in normal mode.                         */
/*    ENOMEM - Could not allocate an scbuf for this command.              */
/*    ENODEV - Device could not be selected.                              */
/*    ETIMEDOUT - The start unit command timed out.                       */
/*    EIO    - Error returned from psc_strategy.                          */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:  iowait, xmfree, copyin, copyout, bcopy    */
/*                                                                        */
/**************************************************************************/
int
psc_test_unit_rdy(
		  dev_t devno,
		  int arg,
		  ulong devflag)
{
    int     ret_code, ret_code2, dev_index;
    struct sc_buf *bp;
    struct sc_ready sc_rdy;

    DEBUG_0 ("Entering psc_test_unit_rdy routine.\n")

    ret_code = 0;
    /* Copy in the arg structure. If the buffer resides */
    /* user space, use copyin, else use bcopy.          */
    if (!(devflag & DKERNEL))
    {
	ret_code = copyin((char *) arg, &sc_rdy,
			  sizeof(struct sc_ready));
	if (ret_code != 0)
	{
	    DEBUG_0 ("Leaving psc_test_unit_rdy routine. EFAULT\n")
	    return (EFAULT);
	}
    }
    else
    { /* buffer is in kernel space */
	bcopy((char *) arg, &sc_rdy, sizeof(struct sc_ready));
    }
    if (adp_str.open_mode == DIAG_MODE)
    {	/* adapter opened in diagnostic mode */
	DEBUG_0 ("Leaving psc_test_unit_rdy routine. EACCES\n")
	return (EACCES);
    }
    /* Obtain device structure from hash on the scsi id */
    /* and lun id.                                      */
    dev_index = INDEX(sc_rdy.scsi_id, sc_rdy.lun_id);
    if (adp_str.device_queue_hash[dev_index] == NULL)
    {	/* device queue structure not allocated */
	DEBUG_0 ("Leaving psc_test_unit_rdy routine. EINVAL\n")
	return (EINVAL);
    }
    bp = psc_build_command();
    if (bp == NULL)
    {
	DEBUG_0 ("Leaving psc_test_unit_rdy routine. ENOMEM\n")
	return (ENOMEM);        /* could'nt allocate command */
    }

    bp->scsi_command.scsi_id = sc_rdy.scsi_id;
    bp->scsi_command.scsi_length = 6;
    bp->scsi_command.scsi_cmd.scsi_op_code = SCSI_TEST_UNIT_READY;
    bp->scsi_command.scsi_cmd.lun = (sc_rdy.lun_id << 5);
    bp->scsi_command.scsi_cmd.scsi_bytes[0] = 0;
    bp->scsi_command.scsi_cmd.scsi_bytes[1] = 0;
    bp->scsi_command.scsi_cmd.scsi_bytes[2] = 0;
    bp->scsi_command.scsi_cmd.scsi_bytes[3] = 0;

    /* There is no need to set the no disconnect flag for the */
    /* test unit ready command so just set the synch/asynch  */
    /* flag for the sc_rdy structure.                        */
    bp->scsi_command.flags = sc_rdy.flags & SC_ASYNC;
    bp->bufstruct.b_bcount = 0;
    bp->bufstruct.b_dev = devno;

    /* Initialize default status to zero.                    */
    sc_rdy.status_validity = 0;
    sc_rdy.scsi_status = 0;

    /* Set resume flag in case caller is retrying this operation. */
    /* This assumes the inquiry is only running single-threaded  */
    /* to this device.                                           */
    bp->flags = SC_RESUME;

    /* Call our strategy routine to issue the start unit cmd. */
    if (psc_strategy(bp))
    {	/* an error was returned */
	(void) xmfree(bp, pinned_heap);	/* release buffer */
	return (EIO);
	DEBUG_0 ("Leaving psc_test_unit_rdy strategy error EIO.\n")
    }

    iowait((struct buf *) bp);  /* Wait for commmand completion */

    /* The return value from the operation is examined to deter- */
    /* what type of error occurred.  Since the calling applica-  */
    /* requires an ERRNO, this value is interpreted here.        */
    if (bp->bufstruct.b_flags & B_ERROR)
    {	/* an error occurred */
	if (bp->status_validity & SC_ADAPTER_ERROR)
	{	/* if adapter error */
	    switch (bp->general_card_status)
	    {
	    case SC_CMD_TIMEOUT:
		psc_logerr(ERRID_SCSI_ERR2, COMMAND_TIMEOUT,
			   65, 0, NULL, FALSE);
		ret_code = ETIMEDOUT;
		break;
	    case SC_NO_DEVICE_RESPONSE:
		ret_code = ENODEV;
		break;
	    case SC_SCSI_BUS_FAULT:
		ret_code = ENOCONNECT;
		break;
	    default:
		ret_code = EIO;
		break;
	    }
	}
	else
	    if (bp->status_validity & SC_SCSI_ERROR)
	    {	/* if a scsi status error */
		sc_rdy.status_validity = SC_SCSI_ERROR;
		sc_rdy.scsi_status = bp->status_validity;
		ret_code = EIO;
	    }
	    else
	    {	/* if general error (fall through case) */
		ret_code = EIO;
	    }
    }

    /* Copy out the device status to the st_ready structure      */
    /* passed in by the calling application.                     */
    if (!(devflag & DKERNEL))
    {
	ret_code2 = copyout(&sc_rdy, (char *) arg,
			    sizeof(struct sc_ready));
	if (ret_code2 != 0)
	{
	    ret_code = EFAULT;
	}
    }
    else
    {	/* buffer is in kernel space */
	bcopy((char *) arg, &sc_rdy, sizeof(struct sc_ready));
    }
    (void) xmfree(bp, pinned_heap);	/* release buffer */
    DEBUG_0 ("Leaving psc_test_unit_rdy routine.\n")
    return (ret_code);
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_readblk                                                     */
/*                                                                        */
/* FUNCTION:  Issues a SCSI read command to a device.                     */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine is called to read command to a device.            */
/*         The calling process is responsible for NOT calling this rou-   */
/*         tine if the SCIOSTART failed.  Such a failure would indicate   */
/*         that another process has this device open and interference     */
/*         could cause improper error reporting.                          */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      adapter_def - scsi chip unique data structure                     */
/*                                                                        */
/* INPUTS:                                                                */
/*      devno   - passed device major/minor number                        */
/*      arg     - passed argument value                                   */
/*      devflag - device flag                                             */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      0 - for good completion, ERRNO value otherwise.                   */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*    EINVAL - Device already opened.                                     */
/*    EACCES - Adapter not opened in normal mode.                         */
/*    ENOMEM - Could not allocate an scbuf for this command.              */
/*    ENODEV - Device could not be selected.                              */
/*    ETIMEDOUT - The inquiry command timed out.                          */
/*    EIO    - Error returned from psc_strategy.                          */
/*    EIO    - No data returned from read command.                        */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:  iowait, xmfree, copyin, copyout, bcopy    */
/*                                                                        */
/**************************************************************************/
int
psc_readblk(
	    dev_t devno,
	    int arg,
	    ulong devflag)
{
    int     ret_code, dev_index;
    struct sc_buf *bp;
    struct sc_readblk sc_rd;

    DEBUG_0 ("Entering psc_readblk routine.\n")
    ret_code = 0;
    /* Copy in the arg structure. If the buffer resides */
    /* user space, use copyin, else use bcopy.          */
    if (!(devflag & DKERNEL))
    {
	ret_code = copyin((char *) arg, &sc_rd,
			  sizeof(struct sc_readblk));
	if (ret_code != 0)
	{
	    DEBUG_0 ("Leaving psc_readblk routine.\n")
	    return (EFAULT);
	}
    }
    else
    {	/* buffer is in kernel space */
	bcopy((char *) arg, &sc_rd, sizeof(struct sc_readblk));
    }
    if (adp_str.open_mode == DIAG_MODE)
    {	/* adapter opened in diagnostic mode */
	DEBUG_0 ("Leaving psc_readblk routine.\n")
        return (EACCES);
    }
    /* Obtain device structure from hash on the scsi id */
    /* and lun id.                                      */
    dev_index = INDEX(sc_rd.scsi_id, sc_rd.lun_id);

    if (adp_str.device_queue_hash[dev_index] == NULL)
    {	/* device queue structure already allocated */
	DEBUG_0 ("Leaving psc_readblk routine.\n")
	return (EINVAL);
    }
    if ((int) sc_rd.blklen > PAGESIZE)
    {	/* The tranfer length is too long to be allowed. */
	DEBUG_0 ("Leaving psc_readblk routine.\n")
	return (EINVAL);
    }
    bp = psc_build_command();
    if (bp == NULL)
    {
	DEBUG_0 ("Leaving psc_readblk routine.\n")
	return (ENOMEM);        /* could'nt allocate command */
    }

    /* Xmalloc a page to be used for the data transfer.     */
    bp->bufstruct.b_un.b_addr = xmalloc((uint) PAGESIZE,
					(uint) PGSHIFT,
					pinned_heap);

    if (bp->bufstruct.b_un.b_addr == NULL)
    {	/* Unable to get required storage */
	(void) xmfree(bp, pinned_heap);	/* release sc_buf */
	DEBUG_0 ("Leaving psc_readblk routine.\n")
	return (ENOMEM);
    }

    bp->scsi_command.scsi_id = sc_rd.scsi_id;
    bp->scsi_command.scsi_length = 6;
    bp->scsi_command.scsi_cmd.scsi_op_code = SCSI_READ;
    /* Set up the byte count for this command */
    bp->scsi_command.scsi_cmd.lun = (sc_rd.lun_id << 5) |
	((sc_rd.blkno >> 16) & 0x1f);
    bp->scsi_command.scsi_cmd.scsi_bytes[0] = (sc_rd.blkno >> 8) & 0xff;
    bp->scsi_command.scsi_cmd.scsi_bytes[1] = sc_rd.blkno & 0xff;
    bp->scsi_command.scsi_cmd.scsi_bytes[2] = 1;	/* single block */
    bp->scsi_command.scsi_cmd.scsi_bytes[3] = 0;

    /* Set the no disconnect flag and the synch/asynch flag  */
    /* for proper data tranfer to occur.                     */
    bp->scsi_command.flags = (sc_rd.flags & SC_ASYNC) | SC_NODISC;
    /* put no disconnect flag back in after fix */

    bp->bufstruct.b_bcount = (unsigned) sc_rd.blklen;
    bp->bufstruct.b_flags |= B_READ;
    bp->bufstruct.b_dev = devno;
    bp->timeout_value = sc_rd.timeout_value;

    /* Set resume flag in case caller is retrying this operation. */
    /* This assumes the readblk is only running single-threaded  */
    /* to this device.                                           */
    bp->flags = SC_RESUME;

    /* Call our strategy routine to issue the readblk command. */
    if (psc_strategy(bp))
    {	/* an error was returned */
	(void) xmfree(bp->bufstruct.b_un.b_addr, pinned_heap);
	(void) xmfree(bp, pinned_heap);	/* release sc_buf */
	return (EIO);
	DEBUG_0 ("Leaving psc_readblk routine.\n")
    }

    iowait((struct buf *) bp);	/* Wait for commmand completion */

    /* The return value from the operation is examined to deter- */
    /* what type of error occurred.  Since the calling applica-  */
    /* requires an ERRNO, this value is interpreted here.        */
    if (bp->bufstruct.b_flags & B_ERROR)
    {	/* an error occurred */
	if (bp->status_validity & SC_ADAPTER_ERROR)
	{	/* if adapter error */
	    switch (bp->general_card_status)
	    {
	    case SC_CMD_TIMEOUT:
		psc_logerr(ERRID_SCSI_ERR2, COMMAND_TIMEOUT,
			   70, 0, NULL, FALSE);
		ret_code = ETIMEDOUT;
		break;
	    case SC_NO_DEVICE_RESPONSE:
		ret_code = ENODEV;
		break;
	    case SC_SCSI_BUS_FAULT:
		ret_code = ENOCONNECT;
		break;
	    default:
		ret_code = EIO;
		break;
	    }
	}
	else
	{
	    ret_code = EIO;
	}
    }

    /* if no other errors, and yet no data came back, then fail */
    if ((ret_code == 0) &&
	(bp->bufstruct.b_resid == bp->bufstruct.b_bcount))
    {
	ret_code = EIO;
    }
    /* No errors, so give the calling routine the data     */
    if (ret_code == 0)
    {
	/* Copy out the readblk data. If the buffer resides */
	/* user space, use copyin, else use bcopy.          */
	if (!(devflag & DKERNEL))
	{
	    ret_code = copyout(bp->bufstruct.b_un.b_addr,
			       sc_rd.data_ptr, sc_rd.blklen);
	    if (ret_code != 0)
	    {
		ret_code = EFAULT;
	    }
	}
	else
	{	/* buffer is in kernel space */
	    bcopy(bp->bufstruct.b_un.b_addr, sc_rd.data_ptr,
		  sc_rd.blklen);
	}
    }

    (void) xmfree(bp->bufstruct.b_un.b_addr, pinned_heap);
    (void) xmfree(bp, pinned_heap);	/* release sc_buf */
    DEBUG_0 ("Leaving psc_readblk routine.\n")
    return (ret_code);
}

/**************************************************************************/
/*                                                                        */
/* NAME:    psc_issue_abort                                               */
/*                                                                        */
/* FUNCTION:  Issues a SCSI abort to a device.                            */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine causes an abort command to be issued to a SCSI    */
/*         device.  Note that this action will also halt the device queue.*/
/*         The calling process is responsible for NOT calling this rou-   */
/*         tine if the SCIOSTART failed.  Such a failure would indicate   */
/*         that another process has this device open and interference     */
/*         could cause improper error reporting.                          */
/*                                                                        */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    scsi chip structure - scsi chip information.                        */
/*                                                                        */
/* INPUTS:                                                                */
/*    dev_index - index to device information                             */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*    0 - for good completion, ERRNO value otherwise                      */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*    EINVAL    - device already opened.                                  */
/*    EACCES    - Adapter not opened in normal mode.                      */
/*    ETIMEDOUT - Command timed out.                                      */
/*    ENOCONNECT- No connection.                                          */
/*    EIO       - Adapter error.                                          */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED: i_disable, i_enable, e_sleep               */
/*                                                                        */
/**************************************************************************/
int
psc_issue_abort(
		int dev_index)
{
    struct dev_info *dev_ptr;
    struct dev_info *next_ptr;
    struct sc_buf *bp;
    int     old_pri1;
    int     i, ret_code;
    uchar   istat_val;
    caddr_t iocc_addr;
    register int rc;

    DEBUG_0 ("Entering psc_issue_abort routine\n")
    if (adp_str.open_mode == DIAG_MODE)
    {
	DEBUG_0 ("Leaving psc_issue_abort routine with EACCES.\n")
	return (EACCES);
    }

    /* Obtain device structure pointer and disable */
    /* interrupts for processing.                 */
    dev_ptr = adp_str.device_queue_hash[dev_index];
    old_pri1 = i_disable(adp_str.ddi.int_prior);

    /* if Q state is set to STOPPING or HALTED return EIO */
    if ((dev_ptr = adp_str.device_queue_hash[dev_index]) == NULL)
    {
	DEBUG_0 ("Leaving psc_issue_abort routine with EINVAL.\n")
	i_enable(old_pri1);     /* re-enable */
	return (EINVAL);
    }

    if ((dev_ptr->queue_state & STOPPING) ||
	(dev_ptr->queue_state & HALTED))
    {
	i_enable(old_pri1);	/* re-enable */
	return (EIO);
    }

    if ((dev_ptr->cmd_activity_state == ABORT_IN_PROGRESS) ||
	(dev_ptr->cmd_activity_state == BDR_IN_PROGRESS) ||
	(dev_ptr->ABORT_BDR_bkwd != NULL) ||
	(adp_str.ABORT_BDR_head == dev_ptr))

    {
	i_enable(old_pri1);	/* re-enable */
	return (EIO);
    }

    /************************************************************/
    /*           DEVICE_WAITING_FOR_RESOURCES QUEUE             */
    /************************************************************/

    if ((dev_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd != NULL) ||
	(adp_str.DEVICE_WAITING_FOR_RESOURCES_tail == dev_ptr))
    {
	dev_ptr->queue_state = HALTED;
	if (adp_str.DEVICE_WAITING_FOR_RESOURCES_head == dev_ptr)
	{
	    adp_str.DEVICE_WAITING_FOR_RESOURCES_head =
		    dev_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd;
	    if (adp_str.DEVICE_WAITING_FOR_RESOURCES_tail == dev_ptr)
		adp_str.DEVICE_WAITING_FOR_RESOURCES_tail = NULL;
	}
	else
	{
	    next_ptr = adp_str.DEVICE_WAITING_FOR_RESOURCES_head;
	    while (next_ptr != NULL)
	    {
		if (next_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd == dev_ptr)
		{
		    next_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd =
			dev_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd;
		    if (adp_str.DEVICE_WAITING_FOR_RESOURCES_tail ==
			dev_ptr)
			adp_str.DEVICE_WAITING_FOR_RESOURCES_tail =
			   next_ptr;
		    break;
		}
		next_ptr = next_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd;
	    }
	}
	/* setup status in sc_buf to indicate error for each  */
	/* pending command                                    */
	if (dev_ptr->head_pend != NULL)
	{
	    psc_fail_cmd(dev_ptr, ENXIO, 1);
	    dev_ptr->head_pend = NULL;
	}

	i_enable(old_pri1);	/* re-enable */
	return (PSC_NO_ERR);	/* no error */
    }

    /************************************************************/
    /*              DEVICE_WAITING QUEUE                        */
    /************************************************************/

    if ((dev_ptr->DEVICE_WAITING_fwd != NULL) ||
	(adp_str.DEVICE_WAITING_tail == dev_ptr))
    {
	dev_ptr->queue_state = HALTED;
	if (adp_str.DEVICE_WAITING_head == dev_ptr)
	{
	    adp_str.DEVICE_WAITING_head =
		    dev_ptr->DEVICE_WAITING_fwd;
	    if (adp_str.DEVICE_WAITING_tail == dev_ptr)
		adp_str.DEVICE_WAITING_tail = NULL;
	}
	else
	{
	    next_ptr = adp_str.DEVICE_WAITING_head;
	    while (next_ptr != NULL)
	    {
		if (next_ptr->DEVICE_WAITING_fwd == dev_ptr)
		{
		    next_ptr->DEVICE_WAITING_fwd = dev_ptr->DEVICE_WAITING_fwd;
		    if (adp_str.DEVICE_WAITING_tail == dev_ptr)
			adp_str.DEVICE_WAITING_tail = next_ptr;
		    break;
		}
		next_ptr = next_ptr->DEVICE_WAITING_fwd;
	    }
	}
	/* release resources for command */
	(void) psc_free_resources(dev_ptr, FALSE);
	/* setup status in sc_buf to indicate error on the  */
	/* currently executing command                      */
	bp = dev_ptr->active;
	bp->status_validity = 0;
	bp->bufstruct.b_resid = bp->bufstruct.b_bcount;

	/* setup status in sc_buf to indicate error for each */
	/* pending command                                   */
	/* iodone active and pending cmds */
	psc_fail_cmd(dev_ptr, ENXIO, 2);
	i_enable(old_pri1);	/* re-enable */
	return (PSC_NO_ERR);
    }

    /************************************************************/
    /*              DEVICE_ACTIVE QUEUE                         */
    /************************************************************/

    if ((dev_ptr->DEVICE_ACTIVE_bkwd != NULL) ||
	(adp_str.DEVICE_ACTIVE_head == dev_ptr))
    {
	dev_ptr->flags |= SCSI_ABORT;
	dev_ptr->ioctl_wakeup = TRUE;
	psc_enq_abort_bdr(dev_ptr);

	/* if dev is not connected                               */
	/* to the scsi bus, then SUSPEND processing by chip      */
	/* (use a timer to keep an eye on things)                */

	/* see if chip is waiting for reselection */
	istat_val = -1;
	iocc_addr =
	    BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
	psc_patch_iowait_int_on();
	for (i = 0; i < 3; i++)
	{
	    rc = BUS_GETCX(iocc_addr + ISTAT, &istat_val);
	    if (!(istat_val & CONN_INT_TEST))
	    {	/* The register read is successful */
		rc |= BUS_PUTCX(iocc_addr + ISTAT, 0x80);
		if (rc != REG_FAIL)
		{
		    /* Start a timer for the chip suspend */
		    w_start(&(adp_str.adap_watchdog.dog));
		    break;
		}
		else
		{
		    /* Failed during a register write */
		    istat_val = rc;
		}
	    }
	    if (istat_val != REG_FAIL)
	    {	/* Can't issue abort so  */
		break;
	    }
	}
	BUSIO_DET(iocc_addr);
	dev_ptr->ioctl_wakeup = TRUE;
	while (dev_ptr->ioctl_wakeup)
	    e_sleep(&dev_ptr->ioctl_event, EVENT_SHORT);
	ret_code = dev_ptr->ioctl_errno;

    }
    DEBUG_1 ("Leaving psc_issue_abort routine. rc=0x%x\n", ret_code);
    i_enable(old_pri1);	/* re-enable */
    return (ret_code);	/* no error */
}  /* psc_issue_abort  */

/**************************************************************************/
/*                                                                        */
/* NAME:    psc_issue_BDR                                                 */
/*                                                                        */
/* FUNCTION:  Issues a bus device reset to a scsi device.                 */
/*                                                                        */
/*      This performs actions necessary to send a SCSI bus                */
/*      device reset mesg to the scsi controller.                         */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine causes an reset command to be issued to a SCSI    */
/*         device.  Note that this action will also halt the device queue.*/
/*         The calling process is responsible for NOT calling this rou-   */
/*         tine if the SCIOSTART failed.  Such a failure would indicate   */
/*         that another process has this device open and interference     */
/*         could cause improper error reporting.                          */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    adapter_def - adapter unique data structure                         */
/*                                                                        */
/* INPUTS:                                                                */
/*    dev_index - index to device information                             */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*    0 - for good completion, ERRNO value otherwise                      */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*    EINVAL    - device already opened.                                  */
/*    EACCES    - Adapter not opened in normal mode.                      */
/*    ETIMEDOUT - Command timed out.                                      */
/*    ENOCONNECT- No connection.                                          */
/*    EIO       - Adapter  error.                                         */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*  i_disable     i_enable    e_sleep    w_start                          */
/*                                                                        */
/**************************************************************************/
int
psc_issue_BDR(
	      int dev_index)
{
    struct sc_buf *bp;
    struct dev_info *dev_ptr;
    struct dev_info *next_ptr;
    int     old_pri1;
    int     i, ret_code;
    uchar   istat_val;
    caddr_t iocc_addr;
    register int rc;


    DEBUG_0 ("Entering psc_issue_BDR routine.\n")
    if (adp_str.open_mode == DIAG_MODE)
    {
	DEBUG_0 ("Leaving psc_issue_BDR routine.\n")
	return (EACCES);
    }

    /* Obtain device structure pointer and disable */
    /* interrupts for processing.                 */
    dev_ptr = adp_str.device_queue_hash[dev_index];
    old_pri1 = i_disable(adp_str.ddi.int_prior);

    if ((dev_ptr = adp_str.device_queue_hash[dev_index]) == NULL)
    {
	i_enable(old_pri1);	/* re-enable */
	return (EINVAL);
    }

    /* if Q state is set to STOPPING or HALTED return EIO */
    if ((dev_ptr->queue_state & STOPPING) ||
	(dev_ptr->queue_state & HALTED))
    {
	i_enable(old_pri1);	/* re-enable */
	return (EIO);
    }

    /************************************************************/
    /*                  ABORT_BDR QUEUE                         */
    /************************************************************/

    if ((dev_ptr->cmd_activity_state == ABORT_IN_PROGRESS) ||
	(dev_ptr->cmd_activity_state == BDR_IN_PROGRESS) ||
	(dev_ptr->ABORT_BDR_bkwd != NULL) ||
	(adp_str.ABORT_BDR_head == dev_ptr))
    {
	i_enable(old_pri1);
	return (EIO);
    }

    /************************************************************/
    /*            DEVICE_WAITING_FOR_RESOURCES QUEUE            */
    /************************************************************/
    else if ((dev_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd != NULL) ||
	    (adp_str.DEVICE_WAITING_FOR_RESOURCES_tail == dev_ptr))
    {
	if (adp_str.DEVICE_WAITING_FOR_RESOURCES_head == dev_ptr)
	{
	    adp_str.DEVICE_WAITING_FOR_RESOURCES_head =
		    dev_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd;
	    if (adp_str.DEVICE_WAITING_FOR_RESOURCES_tail == dev_ptr)
		adp_str.DEVICE_WAITING_FOR_RESOURCES_tail = NULL;
	}
	else
	{
	    next_ptr = adp_str.DEVICE_WAITING_FOR_RESOURCES_head;
	    while (next_ptr != NULL)
	    {
		if (next_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd == dev_ptr)
		{
		    next_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd =
			dev_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd;
		    if (adp_str.DEVICE_WAITING_FOR_RESOURCES_tail == dev_ptr)
			adp_str.DEVICE_WAITING_FOR_RESOURCES_tail =
			   next_ptr;
		    break;
		}
		next_ptr = next_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd;
	    }
	}
    }
    /************************************************************/
    /*                   DEVICE_WAITING QUEUE                   */
    /************************************************************/
    else if ((dev_ptr->DEVICE_WAITING_fwd != NULL) ||
	    (adp_str.DEVICE_WAITING_tail == dev_ptr))
    {
	if (adp_str.DEVICE_WAITING_head == dev_ptr)
	{
	    adp_str.DEVICE_WAITING_head =
		    dev_ptr->DEVICE_WAITING_fwd;
	    if (adp_str.DEVICE_WAITING_tail == dev_ptr)
		adp_str.DEVICE_WAITING_tail = NULL;
	}
	else
	{
	    next_ptr = adp_str.DEVICE_WAITING_head;
	    while (next_ptr != NULL)
	    {
		if (next_ptr->DEVICE_WAITING_fwd == dev_ptr)
		{
		    next_ptr->DEVICE_WAITING_fwd = dev_ptr->DEVICE_WAITING_fwd;
		    if (adp_str.DEVICE_WAITING_tail == dev_ptr)
			adp_str.DEVICE_WAITING_tail = next_ptr;
		    break;
		}
		next_ptr = next_ptr->DEVICE_WAITING_fwd;
	    }
	}
	/* release resources for command */
	(void) psc_free_resources(dev_ptr, FALSE);
	if (dev_ptr->head_pend == NULL)
	{       /* queue is empty  */
	    dev_ptr->head_pend = dev_ptr->active;
	    dev_ptr->tail_pend = dev_ptr->active;
	}
	else
	{       /* pending queue not empty */
	    /* point first cmd's av_forw at the new request */
	    dev_ptr->active->bufstruct.av_forw =
					 (struct buf *) dev_ptr->head_pend;
	    dev_ptr->head_pend = dev_ptr->active;
	}
	dev_ptr->active = NULL;
    }
    dev_ptr->flags |= SCSI_BDR;
    dev_ptr->ioctl_wakeup = TRUE;
    if (adp_str.DEVICE_ACTIVE_head == NULL)
    {
	/* update the WAIT script */
	psc_reset_iowait_jump(adp_str.SCRIPTS[0].script_ptr, dev_index);
	/* start timer */
	dev_ptr->dev_watchdog.dog.restart = LONGWAIT;
	w_start(&dev_ptr->dev_watchdog.dog);
	/* issue the cmd */
	dev_ptr->cmd_activity_state = BDR_IN_PROGRESS;
	psc_enq_active(dev_ptr);
	psc_issue_bdr_script(adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].
			   script_ptr, dev_ptr->script_dma_addr, dev_index);
    }
    else
    {
	psc_enq_abort_bdr(dev_ptr);

	/* see if chip is waiting for reselection */
	istat_val = -1;
	iocc_addr =
	    BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
	psc_patch_iowait_int_on();
	for (i = 0; i < 3; i++)
	{
	    rc = BUS_GETCX(iocc_addr + ISTAT, &istat_val);
	    if (!(istat_val & CONN_INT_TEST))
	    {	/* The register read is successful */
		rc |= BUS_PUTCX(iocc_addr + ISTAT, 0x80);
		if (rc != REG_FAIL)
		{
		    /* Start a timer for the chip suspend */
		    w_start(&(adp_str.adap_watchdog.dog));
		    break;
		}
		else
		{
		    /* Failed during a register write */
		    istat_val = rc;
		}
	    }
	    if (istat_val != REG_FAIL)
	    {	/* Can't issue abort so  */
		break;
	    }
	}
	BUSIO_DET(iocc_addr);
    }
    dev_ptr->ioctl_wakeup = TRUE;
    while (dev_ptr->ioctl_wakeup)
	e_sleep(&dev_ptr->ioctl_event, EVENT_SHORT);
    ret_code = dev_ptr->ioctl_errno;
    i_enable(old_pri1);	/* re-enable */
    DEBUG_1 ("Leaving psc_issue_BDR routine. rc=0x%x\n", ret_code);
    return (ret_code);
}  /* psc_issue_BDR */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_adp_str_init                                                */
/*                                                                        */
/* FUNCTION:  Initializes adapter stucture variables.                     */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can only be called by a process.                     */
/*                                                                        */
/* NOTES:  This routine is called to initialize the adapter structure     */
/*         variables, arrays, etc.                                        */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      adapter_def - scsi chip unique data structure                     */
/*                                                                        */
/* INPUTS:                                                                */
/*      None                                                              */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      None                                                              */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
void
psc_adp_str_init()
{

    int     i;

    DEBUG_0 ("Entering psc_adp_str_init routine \n")

    /* init the adapter interrupt handler struct */
    adp_str.intr_struct.next = (struct intr *) NULL;
    adp_str.intr_struct.handler = psc_intr;
    adp_str.intr_struct.bus_type = adp_str.ddi.bus_type;
    adp_str.intr_struct.flags = 0;
    adp_str.intr_struct.level = adp_str.ddi.int_lvl;
    adp_str.intr_struct.priority = adp_str.ddi.int_prior;
    adp_str.intr_struct.bid = adp_str.ddi.bus_id;

    /* initialize other basic global variables to the adp_str */
    adp_str.channel_id = 0;
    adp_str.xmem_STA.aspace_id = XMEM_GLOBAL;
    adp_str.xmem_SCR.aspace_id = XMEM_GLOBAL;
    adp_str.iowait_inited = FALSE;
    adp_str.dump_inited = FALSE;
    adp_str.dump_started = FALSE;
    adp_str.dump_pri = 0;
    adp_str.num_4K_tcws_in_use = 0;

    /* initialize the various pointers in the structure */
    adp_str.DEVICE_ACTIVE_head = NULL;
    adp_str.DEVICE_ACTIVE_tail = NULL;
    adp_str.DEVICE_WAITING_head = NULL;
    adp_str.DEVICE_WAITING_tail = NULL;
    adp_str.DEVICE_WAITING_FOR_RESOURCES_head = NULL;
    adp_str.DEVICE_WAITING_FOR_RESOURCES_tail = NULL;
    adp_str.ABORT_BDR_head = NULL;
    adp_str.ABORT_BDR_tail = NULL;
    adp_str.large_TCW_table = NULL;

    for (i = 0; i < MAX_DEVICES; i++)
	adp_str.device_queue_hash[i] = NULL;
    for (i = 0; i < MAX_SCRIPTS; i++)
    {
	adp_str.SCRIPTS[i].script_ptr = NULL;
	adp_str.SCRIPTS[i].dma_ptr = 0;
	adp_str.SCRIPTS[i].TCW_index = 0;
	adp_str.SCRIPTS[i].in_use = SCR_UNUSED;
    }
    for (i = 0; i < TCW_TABLE_SIZE; i++)
	adp_str.TCW_alloc_table[i] = TCW_UNUSED;

    DEBUG_0 ("Leaving psc_adp_str_init routine \n")
    return;
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_ioctl                                                       */
/*                                                                        */
/* FUNCTION:  Scsi Chip Device Driver Ioctl Routine                       */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can only be called by a process.                     */
/*                                                                        */
/* NOTES:  This routine will accept commands to perform specific function */
/*         and diagnostic operations on the scsi chip.  Supported commands*/
/*         are:                                                           */
/*                                                                        */
/*         IOCINFO    -  Returns information about the tape device.       */
/*         SCIOSTART  -  Open a SCSI ID/LUN                               */
/*         SCIOSTOP   -  Close a SCSI ID/LUN.                             */
/*         SCIOINQU   -  Issues a SCSI Inquiry command to a device.       */
/*         SCIOSTUNIT -  Issues a SCSI start unit command to a device.    */
/*         SCIOTUR    -  Issues a SCSI test unit ready command to device. */
/*         SCIOREAD   -  Issues a SCSI 6-byte read command to device.     */
/*         SCIOHALT   -  Issues a abort to SCSI device.                   */
/*         SCIORESET  -  Issues a Bus Device Reset to a SCSI device.      */
/*         SCIODIAG   -  Run chip diagnostics.                            */
/*         SCIOTRAM   -  No operation.  Returns ENXIO.                    */
/*         SCIODNLD   -  No operation.  Returns ENXIO.                    */
/*         SCIORREG   -  Reads a particular register.                     */
/*         SCIOWREG   -  Writes a particular register.                    */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*                                                                        */
/* INPUTS:                                                                */
/*    devno  - major/minor number                                         */
/*    cmd    - code used to determine which operation to perform.         */
/*    arg    - address of a structure which contains values used in the   */
/*             'arg' operation.                                           */
/*    devflag - not used.                                                 */
/*    chan   - not used (for multiplexed devices).                        */
/*    ext    - not used.                                                  */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:  0 for good completion, ERRNO value otherwise*/
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:  lockl, unlockl, bcopy, copyout            */
/*                                                                        */
/**************************************************************************/
int
psc_ioctl(
	  dev_t devno,	/* major and minor device numbers */
	  int cmd,	/* operation to perform */
	  int arg,	/* pointer to the user structure */
	  ulong devflag,	/* not used */
	  int chan,	/* not used */
	  int ext)	/* not used */

{
    int     rc, ret_code;
    struct devinfo scinfo;
    DEBUG_0 ("Entering psc_ioctl routine.\n")
    DEBUG_6("devno=%x,cmd=%x,arg=%x,devflag=%x,chan=%x,ext=%x\n",
	     devno, cmd, arg, devflag, chan, ext)

    ret_code = 0;       /* default to no errors found  */
    DDHKWD5(HKWD_DD_SCSIDD, DD_ENTRY_IOCTL, ret_code, devno,
	    cmd, devflag, chan, ext);
    /* lock the global lock to serialize with open/close/config */
    rc = lockl(&(psc_lock), LOCK_SHORT);	/* serialize this */
    if (rc != LOCK_SUCC)
    {
	DEBUG_0 ("Leaving psc_ioctl routine.\n")
	DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_IOCTL, ret_code, devno);
	return (EIO);	/* error--kernel service call failed */
    }
    if ((!adp_str_inited) || (!adp_str.opened))
    {	/* scsi chip has not been inited, defined, or opened */
	DEBUG_0 ("Leaving psc_ioctl routine.\n")
	unlockl(&(psc_lock));
	DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_IOCTL, ret_code, devno);
	return (EIO);
    }

    switch (cmd)
    {	/* cmd switch */
    case IOCINFO:	/* get device information */
	scinfo.devtype = DD_BUS;
	scinfo.flags = 0;
	scinfo.devsubtype = DS_SCSI;
	scinfo.un.scsi.card_scsi_id = (char) adp_str.ddi.card_scsi_id;
	scinfo.un.scsi.max_transfer = adp_str.max_request;
	if (!(devflag & DKERNEL))
	{  /* for a user process */
	    rc = copyout(&scinfo, (char *) arg,
			 sizeof(struct devinfo));
	    if (rc != 0)
		ret_code = EFAULT;
	}	/* for a user process */
	else
	{  /* for a kernel process */
	    bcopy(&scinfo, (char *) arg, sizeof(struct devinfo));
	}  /* for a kernel process */
	break;
    case SCIOSTART:	/* start a device */
	/* psc_display_regs(); */
	ret_code = psc_start_dev(INDEX(arg >> 8, arg));
	break;
    case SCIOSTOP:	/* stop a device */
	ret_code = psc_stop_dev(INDEX(arg >> 8, arg));
	break;
    case SCIOHALT:	/* issue a SCSI abort cmd */
	ret_code = psc_issue_abort(INDEX(arg >> 8, arg));
	break;
    case SCIORESET:	/* issue a SCSI abort cmd */
	ret_code = psc_issue_BDR(INDEX(arg >> 8, arg));
	break;
    case SCIOINQU:	/* issue a SCSI device inquiry cmd */
	ret_code = psc_inquiry(devno, arg, devflag);
	break;
    case SCIOSTUNIT:	/* issue a SCSI device start unit */
	ret_code = psc_start_unit(devno, arg, devflag);
	break;
    case SCIOTUR:	/* issue  SCSI device test unit ready */
	ret_code = psc_test_unit_rdy(devno, arg, devflag);
	break;
    case SCIOREAD:	/* issue a SCSI read cmd (6-byte) */
	ret_code = psc_readblk(devno, arg, devflag);
	break;
    case SCIODIAG:	/* run adapter diagnostics command */
	ret_code = psc_diagnostic(arg, devflag);
	break;
    case SCIOTRAM:	/* no-op, no chip ram to test */
	ret_code = ENXIO;
	break;
    case SCIODNLD:	/* no-op, no microcode to download */
	ret_code = ENXIO;
	break;
    default:	/* catch unknown ioctls here */
	ret_code = EINVAL;
	break;
    }	/* cmd switch */

    unlockl(&(psc_lock));
    DEBUG_0 ("Leaving psc_ioctl routine.\n")
    DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_IOCTL, ret_code, devno);
    return (ret_code);
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_alloc_STA                                                   */
/*                                                                        */
/* FUNCTION:  Allocates a small transfer area for a small data xfer.      */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can only be called by a process.                     */
/*                                                                        */
/* NOTES:  This routine is called when allocating resources for a command.*/
/*         This routine obtains a transfer area of size 256 bytes from    */
/*         the memory page this dd has set aside for data transfers in-   */
/*         volving small amounts.                                         */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    adp_str structure - adapter information structure.                  */
/*                                                                        */
/* INPUTS:                                                                */
/*    dev_info structure - pointer to device information structure        */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*    TRUE  - Able to allocate STA.                                       */
/*    FALSE - Unable to allocate STA.                                     */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_alloc_STA(
	      struct dev_info * dev_ptr)
{

    int     i;

    DEBUG_0 ("Entering psc_alloc_STA routine.\n")
    for (i = 0; i < NUM_STA; i++)
    {	/* Loop through the STA area to find a free */
	if (adp_str.STA[i].in_use == STA_UNUSED)
	{  /* A free STA area has been found */
	    dev_ptr->STA_index = i;
	    dev_ptr->STA_addr = (uint) adp_str.STA[i].sta_ptr;
	    dev_ptr->resource_state = STA_RESOURCES_USED;
	    adp_str.STA[i].in_use = STA_IN_USE;
	    DEBUG_0 ("Leaving psc_alloc_STA routine.\n")
	    return (TRUE);
	}
    }
    DEBUG_0 ("Leaving psc_alloc_STA routine.\n")
    return (FALSE);     /* No STA's availible at this time */
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_free_STA                                                    */
/*                                                                        */
/* FUNCTION:  Frees a small transfer area.                                */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can only be called by a process.                     */
/*                                                                        */
/* NOTES:  This routine frees a small transfer area for later use.        */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    adp_str structure - adapter information structure.                  */
/*                                                                        */
/* INPUTS:                                                                */
/*    dev_info structure - pointer to device information structure        */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*    TRUE  - Able to allocate STA.                                       */
/*    FALSE - Unable to allocate STA.                                     */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
void
psc_free_STA(
	     struct dev_info * dev_ptr)
{
    DEBUG_0 ("Entering psc_free_STA routine.\n")
    adp_str.STA[dev_ptr->STA_index].in_use = STA_UNUSED;
    dev_ptr->STA_index = 0;
    dev_ptr->STA_addr = 0;
    DEBUG_0 ("Leaving psc_free_STA routine.\n")
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_alloc_TCW                                                   */
/*                                                                        */
/* FUNCTION:  Allocate TCWs for data transfer.                            */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can only be called by a process.                     */
/*                                                                        */
/* NOTES:  This routine is called when allocating resources for a command.*/
/*         Resources consist of either a single TCW allocated from the    */
/*         small TCW area or more TCWs which are allocated from the large */
/*         TCW area.                                                      */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    adp_str structure - adapter information structure.                  */
/*                                                                        */
/* INPUTS:                                                                */
/*    dev_info structure - pointer to device information structure        */
/*    sc_buf structure   - pointer to scsi command structure.             */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*    TRUE  - Able to allocate TCW(s).                                    */
/*    FALSE - Unable to allocate TCW(s).                                  */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_alloc_TCW(
	      struct sc_buf * bp,
	      struct dev_info * dev_ptr)
{

    uchar   found, wrap, end_search, alloc_val;
    int     baddr, num_TCW, num_large_TCW, t_ptr, i;

    DEBUG_0 ("Entering psc_alloc_TCW routine.\n")
    found = FALSE;
    /* Set up value to use as "in use" flag for debug */
    /* purposes.                                      */
    alloc_val = INDEX(bp->scsi_command.scsi_id,
		      (bp->scsi_command.scsi_cmd.lun) >> 5);
    /* Compute the number of TCWs required using the  */
    /* buffer address and byte count from the scsi    */
    /* command.                                       */
    baddr = (int) bp->bufstruct.b_un.b_addr;
    num_TCW = (((baddr & (PAGESIZE - 1)) + bp->bufstruct.b_bcount - 1) /
	       PAGESIZE) + 1;
    DEBUG_1 ("Number of TCWs = %d\n", num_TCW)
    if (num_TCW == 1)
    {	/* A TCW is allocated from the small TCW area */
	if ((adp_str.num_4K_tcws_in_use + adp_str.num_scripts_created)
	     >= (NUM_4K_TCWS - 1))
	    return (FALSE);
	/* The small TCW area is searched from start  */
	/* of the circular pointer.                   */
	TRACE_2 ("in TCWSM", dev_ptr, bp->bufstruct.b_bcount)
	t_ptr = adp_str.next_4K_req;
	do
	{
	    if (adp_str.TCW_alloc_table[t_ptr] == TCW_UNUSED)
	    {
		/* Setup a pointer to the TCW small    */
		/* xfer array in the dev_info structure */
		dev_ptr->dma_addr =
		    (uint) DMA_ADDR(adp_str.ddi.tcw_start_addr, t_ptr) +
		    (uint) (baddr & (PAGESIZE - 1));
		dev_ptr->TCW_index = t_ptr;
		dev_ptr->TCW_count = 1;
		/* Mark TCW as "in use" with alloc_val */
		adp_str.TCW_alloc_table[t_ptr] = alloc_val;
		found = TRUE;
	    }
	    t_ptr++;
	    if (t_ptr > adp_str.ending_4K_TCW)
		t_ptr = adp_str.begin_4K_TCW;
	} while (!(found) && (t_ptr != adp_str.next_4K_req));
	if (found)
	{
	    adp_str.next_4K_req = t_ptr;
	    DEBUG_0 ("Leaving psc_alloc_TCW routine.\n")
	    dev_ptr->resource_state = SMALL_TCW_RESOURCES_USED;
	    TRACE_2 ("out TCW ", dev_ptr->TCW_index,
		             dev_ptr->TCW_count)
	    adp_str.num_4K_tcws_in_use++;
	    return (TRUE);	/* TCW allocated */
	}
    }
    else
    {	/* The TCWs are allocated from the large TCW area */
	/* The large TCW area is searched from start- */
	/* at a circular pointer.  The number of TCWs */
	/* required are then searched for starting at */
	/* the first free large TCW.                  */

	/* Compute number of large TCWs required      */
	TRACE_2 ("in TCWLG", dev_ptr, bp->bufstruct.b_bcount)
	num_large_TCW = num_TCW / LARGE_TCW_SIZE;
	if (num_TCW % LARGE_TCW_SIZE)
	    num_large_TCW++;
	t_ptr = adp_str.next_large_req;
	wrap = FALSE;
	end_search = FALSE;
	do
	{
	    /* Is there enough room to search     */
	    if ((wrap) &&
		((t_ptr + num_large_TCW) > adp_str.next_large_req))
	    {	/* the search has wrapped around */
		end_search = TRUE;
	    }
	    else
		if ((t_ptr + num_large_TCW) > (adp_str.large_req_end + 1))
		{
		    t_ptr = adp_str.large_req_begin;
		    wrap = TRUE;
		    continue;
		}
	    found = TRUE;
	    /* Loop through the large TCW     */
	    /* table to determine if there is */
	    /* the required number of conti-  */
	    /* guous TCWs.                    */
	    for (i = t_ptr; i < (t_ptr + num_large_TCW); i++)
	    {
		if (adp_str.large_TCW_table[i] != TCW_UNUSED)
		{
		    found = FALSE;
		    t_ptr = i + 1;
		    break;
		}
	    }
	} while (!(found) && !(end_search));
	if (found)
	{
	    /* The TCWs were found and a pointer to   */
	    /* first is stored in the dev_info struct. */
	    dev_ptr->dma_addr =
		(uint) DMA_ADDR2(adp_str.large_tcw_start_addr, t_ptr) +
		(uint) (baddr & (PAGESIZE - 1));
	    dev_ptr->TCW_index = t_ptr;
	    dev_ptr->TCW_count = num_large_TCW;
	    /* Loop through large TCWs and mark in use */
	    for (i = 0; i < num_large_TCW; i++)
	    {
		adp_str.large_TCW_table[t_ptr + i] = alloc_val;
	    }
	    if ((t_ptr + num_large_TCW) > adp_str.large_req_end)
		adp_str.next_large_req = adp_str.large_req_begin;
	    else
		adp_str.next_large_req = t_ptr + num_large_TCW;
	    dev_ptr->resource_state = LARGE_TCW_RESOURCES_USED;
	    TRACE_2 ("out TCW ", dev_ptr->TCW_index,
		    dev_ptr->TCW_count)
	    return (TRUE);
	}
    }

    TRACE_1 ("NO TCWs!", dev_ptr)
    TRACE_1 ("out TCW ", 0)
    DEBUG_0 ("Leaving psc_alloc_TCW routine.\n")
    return (FALSE);     /* Not enough TCWs available */
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_free_TCW                                                    */
/*                                                                        */
/* FUNCTION:  Free TCWs after command completion.                         */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can only be called by a process.                     */
/*                                                                        */
/* NOTES:  This routine is called when freeing resources for a command.   */
/*         The used TCWs will be freed to the proper table.               */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    adp_str structure - adapter information structure.                  */
/*                                                                        */
/* INPUTS:                                                                */
/*    dev_info structure - pointer to device information structure        */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
void
psc_free_TCW(
	     struct dev_info * dev_ptr)
{

    int     i;

    DEBUG_0 ("Entering psc_free_TCW routine.\n")
    if (dev_ptr->resource_state == SMALL_TCW_RESOURCES_USED)
    {	/* A TCW is allocated from the small TCW area */
	/* Return this TCW to the small TCW area.      */
	adp_str.TCW_alloc_table[dev_ptr->TCW_index] = TCW_UNUSED;
	adp_str.num_4K_tcws_in_use--;
    }
    else
    {	/* The TCWs are allocated from the large TCW area */
	/* Return the TCWs to the large TCW area.     */
	for (i = 0; i < dev_ptr->TCW_count; i++)
	{
	    adp_str.large_TCW_table[dev_ptr->TCW_index + i] = TCW_UNUSED;
	}
    }
    /* Clear the TCW save area in the dev_info struct. */
    dev_ptr->TCW_index = 0;
    dev_ptr->TCW_count = 0;
    DEBUG_0 ("Leaving psc_free_TCW routine.\n")
    return;
}

/**************************************************************************/
/*                                                                        */
/* NAME:        psc_build_command                                         */
/*                                                                        */
/* FUNCTION:    Builds an internal command for ioctl routines.            */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine is called to initialize fields within the sc_buf  */
/*         structure that is allocated via this routine.  This routine is */
/*         is called by ioctl routines that issue a command via the       */
/*         psc_strategy routine.                                          */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      sc_buf  - input/output request struct used between the adapter    */
/*                driver and the calling SCSI device driver               */
/*                                                                        */
/* INPUTS:                                                                */
/*      none                                                              */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      returns a pointer to the sc_buf, or NULL, if it could not         */
/*      be allocated.                                                     */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*      none                                                              */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*      xmalloc         bzero                                             */
/*                                                                        */
/**************************************************************************/
struct sc_buf *
psc_build_command()
{
    struct sc_buf *bp;
    int     i;
    int    *ptr;

    DEBUG_0 ("Entering psc_build_command routine.\n")

    /* Allocate a sc_buf area for this command  */
    bp = (struct sc_buf *) xmalloc((uint) PAGESIZE,
				   (uint) PGSHIFT, pinned_heap);
    if (bp == NULL)
    {	/* xmalloc failed--return NULL pointer */
	DEBUG_0 ("Leaving psc_build_command routine.\n")
	return (NULL);
    }

    /* Clear the sc_buf structure to insure all */
    /* fields are initialized to zero.          */
    bzero(bp, sizeof(struct sc_buf));

    /* Initialize other fields of the sc_buf.   */
    bp->bufstruct.b_forw = NULL;
    bp->bufstruct.b_back = NULL;
    bp->bufstruct.av_forw = NULL;
    bp->bufstruct.av_back = NULL;
    bp->bufstruct.b_iodone = (void (*) ()) psc_iodone;
    bp->bufstruct.b_vp = NULL;
    bp->bufstruct.b_event = EVENT_NULL;
    bp->bufstruct.b_xmemd.aspace_id = XMEM_GLOBAL;
    bp->bufstruct.b_un.b_addr = (char *) bp + sizeof(struct sc_buf);

    /* Additional sc_buf initialization */
    bp->bp = NULL;	/* set for non-spanned command */
    bp->timeout_value = LONGWAIT;	/* set default timeout value */

    DEBUG_0 ("Leaving psc_build_command routine.\n")
    return (bp);
}

/**************************************************************************/
/*                                                                        */
/* NAME:        psc_allocate_resources                                    */
/*                                                                        */
/* FUNCTION:    Calls the appropriate routines to allocate resources.     */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine is called to the determine what resources are re- */
/*         quired for allocation to take place.  Upon completion,         */
/*         d_master will have been called, and the DMA address is saved   */
/*         along with other needed information in the device structure.   */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    adp_str structure - adapter information structure.                  */
/*                                                                        */
/* INPUTS:                                                                */
/*    dev_info structure - pointer to device information structure        */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      returns a pointer to the sc_buf, or NULL, if it could not         */
/*      be allocated.                                                     */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*    NO_ERR       -  Allocation of resources successful.                 */
/*    PSC_FAILED   -  Allocation of resources unsuccessful.               */
/*    PSC_COPY_ERROR - Error occurred during data copy.                   */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_alloc_resources(
		    struct dev_info * dev_ptr)
{
    struct sc_buf *bp;
    struct buf *next_bp;
    uint    local_dma_addr;
    int     ret_code, i;
    int    *ptr;

    DEBUG_0 ("Entering psc_allocate_resource routine.\n")
    TRACE_1 ("in allc ", dev_ptr)
    /* Obtain pointer to the command to process. */
    bp = dev_ptr->head_pend;

    if ((bp->bufstruct.b_bcount > ST_SIZE) ||
	((bp->bufstruct.b_bcount > 0) &&
	(bp->bufstruct.b_xmemd.aspace_id != XMEM_GLOBAL)))
    {
	/* Must be a normal data transfer if execution */
	/* reached here.                               */
	if (psc_alloc_TCW(bp, dev_ptr))
	{   /* The required TCW's are free */
	    /* Move command from the pending device queue */
	    /* to the device active queue.                */
	    dev_ptr->active = bp;
	    dev_ptr->head_pend = (struct sc_buf *) bp->bufstruct.av_forw;
	    if (dev_ptr->head_pend == NULL)
		dev_ptr->tail_pend = NULL;

	    /* Setup and call d_master for both spanned and */
	    /* unspanned commands.                          */
	    if (bp->bp == NULL)
	    {	/* Non-spanned command */
		d_master((int) adp_str.channel_id, DMA_TYPE |
			 ((bp->bufstruct.b_flags & B_READ) ? DMA_READ : 0) |
			 ((bp->bufstruct.b_flags & B_NOHIDE) ?
			  DMA_WRITE_ONLY : 0), bp->bufstruct.b_un.b_addr,
			 (size_t) bp->bufstruct.b_bcount,
			 &bp->bufstruct.b_xmemd,
			 (char *) dev_ptr->dma_addr);
	    }
	    else
	    {	/* must be a spanned command */
		next_bp = (struct buf *) bp->bp;
		local_dma_addr = dev_ptr->dma_addr;
		while (next_bp != NULL)
		{
		    d_master((int) adp_str.channel_id, DMA_TYPE |
			     ((bp->bufstruct.b_flags & B_READ) ?
			      DMA_READ : 0) |
			     ((next_bp->b_flags & B_NOHIDE) ?
			      DMA_WRITE_ONLY : 0),
			     next_bp->b_un.b_addr,
			     (size_t) next_bp->b_bcount,
			     &next_bp->b_xmemd,
			     (char *) local_dma_addr);

		    local_dma_addr += (uint) next_bp->b_bcount;
		    next_bp = (struct buf *) next_bp->av_forw;
		}
	    }
	    DEBUG_0 ("Leaving psc_allocate_resource routine.\n")
	    TRACE_1 ("out allc", PSC_NO_ERR)
	    return (PSC_NO_ERR);
	}
	else
	{   /* Not enough TCW's available */
	    DEBUG_0 ("Leaving psc_allocate_resource routine.\n")
	    TRACE_1 ("out allc", PSC_FAILED)
	    return (PSC_FAILED);
	}
    }
    if (bp->bufstruct.b_bcount == 0)
    {
	/* there is not data to transfer, thus no   */
	/* need to allocate resources.              */
	dev_ptr->resource_state = NO_RESOURCES_USED;
	/* Move command from the pending device queue */
	/* to the device active queue.                */
	dev_ptr->active = bp;
	dev_ptr->head_pend = (struct sc_buf *) bp->bufstruct.av_forw;
	if (dev_ptr->head_pend == NULL)
	    dev_ptr->tail_pend = NULL;
	DEBUG_0 ("Leaving psc_allocate_resource routine.\n")
	TRACE_1 ("out allc", PSC_NO_ERR)
	return (PSC_NO_ERR);
    }
    /* must allocate area from the small transfer area */
    if (psc_alloc_STA(dev_ptr))
    {	/* An STA area was free */
	/* Calculate the dma address of the STA and    */
	/* save in the device's info structure.        */
	dev_ptr->dma_addr =
	    (uint) (DMA_ADDR(adp_str.ddi.tcw_start_addr, 0)) +
	    ((uint) (adp_str.STA[dev_ptr->STA_index].sta_ptr) -
	     ((uint) (adp_str.STA[0].sta_ptr) & ~(PAGESIZE - 1)));

	/* Move command from the pending device queue */
	/* to the device active queue.                */
	dev_ptr->active = bp;
	dev_ptr->head_pend = (struct sc_buf *) bp->bufstruct.av_forw;
	if (dev_ptr->head_pend == NULL)
	    dev_ptr->tail_pend = NULL;

	/* For writes, data needs to be written to the */
	/* small transfer area, so that is done here.  */
	if (!(bp->bufstruct.b_flags & B_READ))
	{
	    /* The data is in kernal space so do a byte copy */
	    for (i = 0; i < bp->bufstruct.b_bcount; i++)
		*(adp_str.STA[dev_ptr->STA_index].sta_ptr + i) =
		    *(bp->bufstruct.b_un.b_addr + i);
	}
	DEBUG_0 ("Leaving psc_allocate_resource routine.\n")
	TRACE_1 ("out allc", PSC_NO_ERR)
	return (PSC_NO_ERR);
    }
    DEBUG_0 ("Leaving psc_allocate_resource routine.\n")
    TRACE_1 ("out allc", PSC_FAILED)
    return (PSC_FAILED);
}  /* psc_alloc_resources */

/**************************************************************************/
/*                                                                        */
/* NAME:        psc_free_resources                                        */
/*                                                                        */
/* FUNCTION:    Calls the appropriate routines to free all resources.     */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine is called to the determine what resources are to  */
/*         be freed.  Pointers and save locations are initialized         */
/*         to the unused state and a d_complete is performed for those    */
/*         commands that require it.                                      */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    adp_str structure - adapter information structure.                  */
/*                                                                        */
/* INPUTS:                                                                */
/*    dev_info structure - pointer to device information structure        */
/*    copy_required - flag to test for copy to a STA:                     */
/*                    TRUE - perform the copy to the user/kernel buffer.  */
/*                    FALSE - A previous error occurred so don't do copy. */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      returns a pointer to the sc_buf, or NULL, if it could not         */
/*      be allocated.                                                     */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*    NO_ERR         - Deallocation of resources successful.              */
/*    PSC_COPY_ERROR - Error occurred during data copy.                   */
/*    PSC_DMA_ERROR  - Error occurred during d_complete.                  */
/*    DMA_SYSTEM     - A DMA system error was detected.                   */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_free_resources(
		   struct dev_info * dev_ptr,
		   uchar copy_required)
{
    struct sc_buf *bp;
    struct buf *next_bp;
    uint    local_dma_addr;
    int     ret_code, ret_code2, i;

    DEBUG_0 ("Entering psc_free_resources routine.\n")
    TRACE_1 ("in free ", dev_ptr)
    ret_code = PSC_NO_ERR;      /* default to successful */

    /* Obtain pointer to the command to process. */
    bp = dev_ptr->active;

    switch (dev_ptr->resource_state)
    {
	/* There's no resources to free, so return. */
    case SMALL_TCW_RESOURCES_USED:
    case LARGE_TCW_RESOURCES_USED:
	/* process case of completion here.    */
	/* Setup and call d_complete for both spanned and */
	/* unspanned commands.                            */
	if (bp->bp == NULL)
	{	/* Non-spanned command */
	    ret_code = d_complete((int) adp_str.channel_id,
				  DMA_TYPE |
				  ((bp->bufstruct.b_flags & B_READ) ?
				   DMA_READ : 0) |
				  ((bp->bufstruct.b_flags & B_NOHIDE) ?
				   DMA_WRITE_ONLY : 0),
				  bp->bufstruct.b_un.b_addr,
				  (size_t) bp->bufstruct.b_bcount,
				  &bp->bufstruct.b_xmemd,
				  (char *) dev_ptr->dma_addr);
	}
	else
	{	/* must be a spanned command */
	    next_bp = (struct buf *) bp->bp;
	    local_dma_addr = dev_ptr->dma_addr;
	    while (next_bp != NULL)
	    {
		ret_code2 = d_complete((int) adp_str.channel_id,
				       DMA_TYPE |
				       ((bp->bufstruct.b_flags & B_READ) ?
					DMA_READ : 0) |
				       ((next_bp->b_flags & B_NOHIDE)
					? DMA_WRITE_ONLY : 0),
				       next_bp->b_un.b_addr,
				       (size_t) next_bp->b_bcount,
				       &next_bp->b_xmemd,
				       (char *) local_dma_addr);
		if (ret_code2 != 0)
		{   /* Was there a DMA type error?   */
		    if (ret_code2 == DMA_SYSTEM)
			/* if a system error has occurred */
			ret_code = DMA_SYSTEM;
		    else
			ret_code = PSC_DMA_ERROR;
		}
		local_dma_addr += (uint) next_bp->b_bcount;
		next_bp = (struct buf *) next_bp->av_forw;
	    }
	}
	psc_free_TCW(dev_ptr);
	dev_ptr->resource_state = 0;
	break;
    case STA_RESOURCES_USED:
	/* Call d_complete on the small xfer that took place */
	ret_code = d_complete(adp_str.channel_id, DMA_TYPE,
			      adp_str.STA[0].sta_ptr, (size_t) PAGESIZE,
			      &adp_str.xmem_STA,
			      DMA_ADDR(adp_str.ddi.tcw_start_addr,
				       adp_str.STA_tcw));
	if (ret_code != 0)
	{    /* If an error occurred */
	    if (ret_code != DMA_SYSTEM)
		ret_code = PSC_DMA_ERROR;
	    /* Attempt to do another d_master to clean */
	    /* up the error.                           */
	    d_master(adp_str.channel_id, DMA_TYPE,
		     adp_str.STA[0].sta_ptr, (size_t) PAGESIZE,
		     &adp_str.xmem_STA,
		     DMA_ADDR(adp_str.ddi.tcw_start_addr,
			      adp_str.STA_tcw));
		     (void) d_complete(adp_str.channel_id, DMA_TYPE,
				       adp_str.STA[0].sta_ptr,
				       (size_t) PAGESIZE, &adp_str.xmem_STA,
				       DMA_ADDR(adp_str.ddi.tcw_start_addr,
				       adp_str.STA_tcw));
	}
	else
	{  /* no error detected - good completion */
	    /* For reads, data needs to be read from the   */
	    /* small transfer area, so that is done here.  */
	    if ((bp->bufstruct.b_flags & B_READ) && (copy_required))
	    {
		if (bp->bufstruct.b_xmemd.aspace_id == XMEM_GLOBAL)
		{	/* The data is in kernal space so do a  */
		    /* byte copy.                           */
		    for (i = 0; i < bp->bufstruct.b_bcount; i++)
			*(bp->bufstruct.b_un.b_addr + i) =
			    *(adp_str.STA[dev_ptr->STA_index].sta_ptr + i);
		}
		else
		{
		    /* use xmemout to copy the data out to user space */
		    ret_code = xmemout(
				    adp_str.STA[dev_ptr->STA_index].sta_ptr,
				       bp->bufstruct.b_un.b_addr,
				       bp->bufstruct.b_bcount,
				       &bp->bufstruct.b_xmemd);
		    if (ret_code != XMEM_SUCC)
		    {
			psc_logerr(ERRID_SCSI_ERR1,
				   XMEM_COPY_ERROR, 80,
				   0, dev_ptr, FALSE);
			ret_code = PSC_COPY_ERROR;
		    }
		    else
		    {
			ret_code = PSC_NO_ERR;
		    }
		}
	    }
	}
	psc_free_STA(dev_ptr);
	dev_ptr->resource_state = 0;
	break;
    case NO_RESOURCES_USED:
    default:
	/* No action is required for these cases */
	dev_ptr->resource_state = 0;
	break;
    }
    DEBUG_0 ("Leaving psc_allocate_resource routine.\n")
    TRACE_1 ("out free", ret_code)
    return (ret_code);
}  /* psc_free_resources */

/************************************************************************/
/*                                                                      */
/* NAME:        psc_iodone                                              */
/*                                                                      */
/* FUNCTION:    Adapter Driver Iodone Routine                           */
/*                                                                      */
/*      This routine handles completion of commands initiated through   */
/*      the psc_ioctl routine.                                          */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine runs on the IODONE interrupt level, so it can      */
/*      be interrupted by the interrupt handler.                        */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      sc_buf  - input/output request struct used between the adapter  */
/*                driver and the calling SCSI device driver             */
/*                                                                      */
/* INPUTS:                                                              */
/*      bp      - pointer to the passed sc_buf                          */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:  none                                      */
/*                                                                      */
/* ERROR DESCRIPTION:  The following errno values may be returned:      */
/*      none                                                            */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      e_wakeup                                                        */
/*                                                                      */
/************************************************************************/
int
psc_iodone(struct sc_buf * bp)
{
    DEBUG_0 ("Entering psc_iodone routine.\n")
    TRACE_1 ("iodone  ", bp)
    e_wakeup(&bp->bufstruct.b_event);
    DEBUG_0 ("Leaving psc_iodone routine.\n")
}

/**************************************************************************/
/*                                                                        */
/* NAME: psc_enq_active, psc_enq_wait, psc_enq_wait_resources,            */
/*       psc_enq_abort_bdr                                                */
/*                                                                        */
/* FUNCTION:                                                              */
/*      Utility routines to handle queuing of device structures to each   */
/*      of the queues in use.                                             */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*                                                                        */
/* INPUTS:                                                                */
/*    dev_info structure - pointer to device information structure        */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:  none                                        */
/*                                                                        */
/* ERROR DESCRIPTION:  The following errno values may be returned:        */
/*      none                                                              */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*      none                                                              */
/*                                                                        */
/**************************************************************************/
void
psc_enq_active(
	       struct dev_info * dev_ptr)
{
    DEBUG_1 ("psc_enq_active dev_ptr 0x%x\n", dev_ptr)
    TRACE_1 ("in enqA ", dev_ptr)
    if (adp_str.DEVICE_ACTIVE_head == NULL)
    {
	dev_ptr->DEVICE_ACTIVE_fwd = NULL;
	dev_ptr->DEVICE_ACTIVE_bkwd = NULL;
	adp_str.DEVICE_ACTIVE_head = dev_ptr;
	adp_str.DEVICE_ACTIVE_tail = dev_ptr;
    }
    else
    {
	dev_ptr->DEVICE_ACTIVE_bkwd = adp_str.DEVICE_ACTIVE_tail;
	dev_ptr->DEVICE_ACTIVE_fwd = NULL;
	adp_str.DEVICE_ACTIVE_tail->DEVICE_ACTIVE_fwd = dev_ptr;
	adp_str.DEVICE_ACTIVE_tail = dev_ptr;
    }
    TRACE_1 ("out enqA", dev_ptr)
}

/************************************************************************/
void
psc_enq_wait(
	     struct dev_info * dev_ptr)
{
    DEBUG_1 ("psc_enq_wait dev_ptr 0x%x\n", dev_ptr)
    TRACE_1 ("in enqW ", dev_ptr)
    if (adp_str.DEVICE_WAITING_head == NULL)
    {
	 adp_str.DEVICE_WAITING_head = dev_ptr;
    }
    else
    {
	adp_str.DEVICE_WAITING_tail->DEVICE_WAITING_fwd = dev_ptr;
    }
    dev_ptr->DEVICE_WAITING_fwd = NULL;
    adp_str.DEVICE_WAITING_tail = dev_ptr;
    TRACE_1 ("out enqW", dev_ptr)
}

/************************************************************************/
void
psc_enq_wait_resources(
		       struct dev_info * dev_ptr)
{
    DEBUG_1 ("psc_enq_wait_resources dev_ptr 0x%x\n", dev_ptr)
    TRACE_1 ("in enqWR", dev_ptr)
    if (adp_str.DEVICE_WAITING_FOR_RESOURCES_head == NULL)
    {
	adp_str.DEVICE_WAITING_FOR_RESOURCES_head = dev_ptr;
    }
    else
    {
	adp_str.DEVICE_WAITING_FOR_RESOURCES_tail->
	    DEVICE_WAITING_FOR_RESOURCES_fwd = dev_ptr;
    }
    dev_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd = NULL;
    adp_str.DEVICE_WAITING_FOR_RESOURCES_tail = dev_ptr;
    TRACE_1 ("out eqWR", dev_ptr)
}

/************************************************************************/
void
psc_enq_abort_bdr(
		  struct dev_info * dev_ptr)
{
    DEBUG_1 ("psc_enq_abort_bdr dev_ptr 0x%x\n", dev_ptr)
    TRACE_1 ("in enqB ", dev_ptr)
    if (adp_str.ABORT_BDR_head == NULL)
    {
	dev_ptr->ABORT_BDR_fwd = NULL;
	dev_ptr->ABORT_BDR_bkwd = NULL;
	adp_str.ABORT_BDR_head = dev_ptr;
	adp_str.ABORT_BDR_tail = dev_ptr;
    }
    else
    {
	dev_ptr->ABORT_BDR_bkwd = adp_str.ABORT_BDR_tail;
	dev_ptr->ABORT_BDR_fwd = NULL;
	adp_str.ABORT_BDR_tail->ABORT_BDR_fwd = dev_ptr;
	adp_str.ABORT_BDR_tail = dev_ptr;
    }
    TRACE_1 ("out enqB", dev_ptr)
}

/**************************************************************************/
/*                                                                        */
/* NAME: psc_deq_active, psc_deq_wait, psc_deq_wait_resources,            */
/*       psc_deq_abort_bdr                                                */
/*                                                                        */
/* FUNCTION:                                                              */
/*      Utility routines to handle dequeueing device structures from      */
/*      each of the queues in use.                                        */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*                                                                        */
/* INPUTS:                                                                */
/*    dev_info structure - pointer to device information structure        */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:  none                                        */
/*                                                                        */
/* ERROR DESCRIPTION:  The following errno values may be returned:        */
/*      none                                                              */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*      none                                                              */
/*                                                                        */
/**************************************************************************/
void
psc_deq_active(
	       struct dev_info * dev_ptr)
{

    DEBUG_1 ("psc_deq_active dev_ptr 0x%x\n", dev_ptr)
    TRACE_1 ("in deqA ", dev_ptr)
    if (adp_str.DEVICE_ACTIVE_head == adp_str.DEVICE_ACTIVE_tail)
    {
	adp_str.DEVICE_ACTIVE_head = NULL;
	adp_str.DEVICE_ACTIVE_tail = NULL;
    }
    else
	if (adp_str.DEVICE_ACTIVE_head == dev_ptr)	/* first one */
	{
	    adp_str.DEVICE_ACTIVE_head = dev_ptr->DEVICE_ACTIVE_fwd;
	    dev_ptr->DEVICE_ACTIVE_fwd->DEVICE_ACTIVE_bkwd =
						dev_ptr->DEVICE_ACTIVE_bkwd;
	}
	else
	    if (adp_str.DEVICE_ACTIVE_tail == dev_ptr)	/* last one */
	    {
		adp_str.DEVICE_ACTIVE_tail = dev_ptr->DEVICE_ACTIVE_bkwd;
		dev_ptr->DEVICE_ACTIVE_bkwd->DEVICE_ACTIVE_fwd =
						 dev_ptr->DEVICE_ACTIVE_fwd;
	    }
	    else	/* in the middle */
	    {
		dev_ptr->DEVICE_ACTIVE_bkwd->DEVICE_ACTIVE_fwd =
						 dev_ptr->DEVICE_ACTIVE_fwd;
		dev_ptr->DEVICE_ACTIVE_fwd->DEVICE_ACTIVE_bkwd =
						dev_ptr->DEVICE_ACTIVE_bkwd;
	    }
    dev_ptr->DEVICE_ACTIVE_fwd = NULL;
    dev_ptr->DEVICE_ACTIVE_bkwd = NULL;
    dev_ptr->cmd_activity_state = 0;
    TRACE_1 ("out deqA", dev_ptr)
}

/************************************************************************/
void
psc_deq_wait(
	     struct dev_info * dev_ptr)
{

    DEBUG_1 ("psc_deq_wait dev_ptr 0x%x\n", dev_ptr)
    TRACE_1 ("in deqW ", dev_ptr)
    if (adp_str.DEVICE_WAITING_head == adp_str.DEVICE_WAITING_tail)
    {
	adp_str.DEVICE_WAITING_head = NULL;
	adp_str.DEVICE_WAITING_tail = NULL;
    }
    else
    {
	adp_str.DEVICE_WAITING_head = dev_ptr->DEVICE_WAITING_fwd;
    }
    dev_ptr->DEVICE_WAITING_fwd = NULL;
    TRACE_1 ("out deqW", dev_ptr)
}

/************************************************************************/
void
psc_deq_wait_resources(
		       struct dev_info * dev_ptr)
{

    DEBUG_1 ("psc_deq_wait_resources dev_ptr 0x%x\n", dev_ptr)
    TRACE_1 ("in deqWR", dev_ptr)
    if (adp_str.DEVICE_WAITING_FOR_RESOURCES_head ==
	adp_str.DEVICE_WAITING_FOR_RESOURCES_tail)
    {
	adp_str.DEVICE_WAITING_FOR_RESOURCES_head = NULL;
	adp_str.DEVICE_WAITING_FOR_RESOURCES_tail = NULL;
    }
    else
    {
	adp_str.DEVICE_WAITING_FOR_RESOURCES_head =
	    dev_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd;
    }
    dev_ptr->DEVICE_WAITING_FOR_RESOURCES_fwd = NULL;
    TRACE_1 ("out dqWR", dev_ptr)
}

/************************************************************************/
void
psc_deq_abort_bdr(
		  struct dev_info * dev_ptr)
{

    DEBUG_1 ("psc_deq_abort_bdr dev_ptr 0x%x\n", dev_ptr)
    TRACE_1 ("in deqB ", dev_ptr)
    if (adp_str.ABORT_BDR_head == adp_str.ABORT_BDR_tail)
    {
	adp_str.ABORT_BDR_head = NULL;
	adp_str.ABORT_BDR_tail = NULL;
    }
    else
	if (adp_str.ABORT_BDR_head == dev_ptr)	/* first one */
	{
	    adp_str.ABORT_BDR_head = dev_ptr->ABORT_BDR_fwd;
	    dev_ptr->ABORT_BDR_fwd->ABORT_BDR_bkwd = dev_ptr->ABORT_BDR_bkwd;
	}
	else
	    if (adp_str.ABORT_BDR_tail == dev_ptr)	/* last one */
	    {
		adp_str.ABORT_BDR_tail = dev_ptr->ABORT_BDR_bkwd;
		dev_ptr->ABORT_BDR_bkwd->ABORT_BDR_fwd =
						     dev_ptr->ABORT_BDR_fwd;
	    }
	    else	/* in the middle */
	    {
		dev_ptr->ABORT_BDR_bkwd->ABORT_BDR_fwd =
						     dev_ptr->ABORT_BDR_fwd;
		dev_ptr->ABORT_BDR_fwd->ABORT_BDR_bkwd =
						    dev_ptr->ABORT_BDR_bkwd;
	    }
    dev_ptr->ABORT_BDR_fwd = NULL;
    dev_ptr->ABORT_BDR_bkwd = NULL;
    TRACE_1 ("out deqB", dev_ptr)
}

/**************************************************************************/
/*                                                                        */
/* NAME: psc_dump                                                         */
/*                                                                        */
/* FUNCTION: Determine what type of dump operation is being sought        */
/*                                                                        */
/*      Allocate necessary segment registers and parse type of            */
/*      dump operation.  Call the specified routine.                      */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*                                                                        */
/*      This routine must be called in the interrupt environment.  It     */
/*      can not page fault and is pinned.                                 */
/*                                                                        */
/* (NOTES:) This routine handles the following operations :               */
/*      DUMPINIT   - initializes bus attached disk as dump device         */
/*      DUMPSTART  - prepares device for dump                             */
/*      DUMPQUERY  - returns the maximum and minimum number of bytes that */
/*                   can be transferred in a single DUMPWRITE command     */
/*      DUMPWRITE  - performs write to disk                               */
/*      DUMPEND    - cleans up device on end of dump                      */
/*      DUMPTERM   - terminates the bus attached disk as dump device      */
/*                                                                        */
/* (DATA STRUCTURES:) uio             - structure containing information  */
/*                                      about the data to transfer        */
/*                    adapter_def    - adapter info structure             */
/*                    dev_info       - device info structure              */
/*                    sc_buf         - i/o request struct between driver  */
/*                                     and device                         */
/*                    dmp_query       - queried transfer information is   */
/*                                      returned                          */
/*                                                                        */
/* INPUTS:                                                                */
/*      devno   - device major/minor number                               */
/*      uiop    - pointer to uio structure for data for the               */
/*                specified operation code                                */
/*      cmd     - Type of dump operation being requested                  */
/*      arg     - Pointer to dump query structure or buf_struct for cmd   */
/*      chan    - unused                                                  */
/*      ext     - unused                                                  */
/*                                                                        */
/* INTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/*      psc_dumpstrt                                                      */
/*      psc_dumpwrt                                                       */
/*      psc_dumpend                                                       */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/*      IOCC_ATT                                                          */
/*      IOCC_DET                                                          */
/*                                                                        */
/* (RECOVERY OPERATION:) If an error occurs, the proper errno is returned */
/*      and the caller is left responsible to recover from the error.     */
/*                                                                        */
/* RETURNS:                                                               */
/*      EBUSY     -  request sense inuse                                  */
/*      EINVAL    -  Invalid iovec argumenmt,or invalid cmd               */
/*      EIO       -  I/O error, quiesce or scsi writefailed.              */
/*      ENOMEM    -  Unable to allocate resources                         */
/*      ENXIO     -  Not inited as dump device                            */
/*      ETIMEDOUT - adapter not responding                                */
/*                                                                        */
/**************************************************************************/
int
psc_dump(
	 dev_t devno,
	 struct uio * uiop,
	 int cmd,
	 int arg,
	 int chan,
	 int ext)
{
    /****************************/
    /* Misc. Variables       */
    /****************************/
    struct sc_buf *scbuf_ptr;
    struct dmp_query *dmp_qry_ptr;
    struct dev_info *dev_ptr;
    int     dev_index;
    int     ret_code;
    int     rc, stack;
    /****************************/
    DEBUG_0 ("Entering psc_dump\n");
    TRACE_1 ("in psc_dump", 0);

    /* check to make sure adapter is inited and opened */
    ret_code = PSC_NO_ERR;
    if ((!adp_str_inited) || (!adp_str.opened))
	return (ENXIO);
    switch (cmd)
    {
    case DUMPINIT:
	adp_str.dump_inited = TRUE;
	adp_str.dump_started = FALSE;
	TRACE_1 ("DUMPINIT", &arg)
	break;
    case DUMPQUERY:
	dmp_qry_ptr = (struct dmp_query *) arg;
	dmp_qry_ptr->min_tsize = 512;
	dmp_qry_ptr->max_tsize = adp_str.max_request;
	TRACE_1 ("DUMPQUERY", &arg)
	break;
    case DUMPSTART:
	TRACE_1 ("DUMPSTART", &arg)
	break;
    case DUMPWRITE:
	i_disable(INTMAX);
	scbuf_ptr = (struct sc_buf *) arg;
	TRACE_1 ("DUMPWRITE", &arg)
	psc_dumpstrt(scbuf_ptr);
	while (adp_str.DEVICE_ACTIVE_head != NULL)
	{
	    ret_code = psc_dump_intr(adp_str.DEVICE_ACTIVE_head);
	}
	adp_str.dump_started = TRUE;
	if (ret_code == 0)
	    ret_code = psc_dumpwrt(scbuf_ptr);
	TRACE_1 ("rc_dmpwrt", ret_code)
	dev_index = INDEX(scbuf_ptr->scsi_command.scsi_id,
			  (scbuf_ptr->scsi_command.scsi_cmd.lun) >> 5);
	dev_ptr = adp_str.device_queue_hash[dev_index];
	dev_ptr->head_pend = scbuf_ptr;
	if (ret_code == 0)
	    ret_code = psc_dump_intr(dev_ptr);
	break;
    case DUMPEND:
	scbuf_ptr = (struct sc_buf *) arg;
	TRACE_1 ("DUMPEND", &arg)
	adp_str.dump_started = FALSE;
	/* enable interrupts */
	i_enable(INTMAX);
	break;
    case DUMPTERM:
	TRACE_1 ("DUMPTERM", &arg)
	if ((!adp_str.dump_inited) || (!adp_str.dump_started))
	    ret_code = EINVAL;
	else
	    adp_str.dump_started = FALSE;
	break;
    default:
	ret_code = EINVAL;
    }
    DEBUG_0 ("Leaving psc_dump\n")
    TRACE_1 ("out psc_dump", ret_code)
    return (ret_code);
}  /* end psc_dump */

/**************************************************************************/
/*                                                                        */
/* NAME: psc_dumpstrt                                                     */
/*                                                                        */
/* FUNCTION: Notify device of pending dump                                */
/*                                                                        */
/*        Prepare device for dump, fail existing commands on all devices  */
/*        including the dump device. Set up the device queue to except    */
/*        dump commands.                                                  */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*                                                                        */
/*      This routine must be called in the interrupt environment.  It     */
/*      can not page fault and is pinned                                  */
/*                                                                        */
/* (DATA STRUCTURES:)                                                     */
/*                    adapter_def    - adapter info structure             */
/*                    sc_buf    - controller info structure               */
/*                    dev_info    - device info structure                 */
/*                                                                        */
/* INPUTS:                                                                */
/*      bp       -  buf scruct for the pending dump command               */
/*                                                                        */
/* CALLED BY:                                                             */
/*      psc_dump                                                          */
/*                                                                        */
/* INTERNAL PROCEDURES CALLED:                                            */
/*      psc_free_resources                                                */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/* (RECOVERY OPERATION:) If an error occurs, the proper errno is returned */
/*      and the caller is left responsible to recover from the error.     */
/*                                                                        */
/* RETURNS:                                                               */
/*      EIO       -  Failed quiesce or scsi write                         */
/*      ENOMEM    -  Unable to allocate resources                         */
/*      ETIMEDOUT - adapter not responding                                */
/*                                                                        */
/**************************************************************************/
int
psc_dumpstrt(
	     struct sc_buf * bp)
{
    struct dev_info *dev_ptr;
    int dev_index, ret_code, loop, i, scsi_intr, dma_intr;
    int rc, save_isr, save_stat, dma_stat;

    DEBUG_0 ("Entering psc_dumpstrt\n")
    TRACE_1 ("in dumpstrt", bp)

    if (!adp_str.dump_inited)
	return (EINVAL);
    dev_index = INDEX(bp->scsi_command.scsi_id,
		      (bp->scsi_command.scsi_cmd.lun) >> 5);
    TRACE_1 ("dev_index :", dev_index)
    dev_ptr = adp_str.device_queue_hash[dev_index];
    dev_ptr->head_pend = bp;

    /* if dump dev already started return to caller */
    if (adp_str.dump_started)
	return (PSC_NO_ERR);

    /* set up the transfer rate for default transfer rate */
    if (psc_write_reg(SXFER, SXFER_SIZE, SYNC_VAL)
	== REG_FAIL)
    {
	ret_code = ENXIO;
    }
    /***********************************************************/
    /* Clear all abort_bdr queues. Then enter abort commands   */
    /* all active device queues.                               */
    /***********************************************************/
    adp_str.ABORT_BDR_head = NULL;
    adp_str.ABORT_BDR_tail = NULL;
    adp_str.DEVICE_WAITING_FOR_RESOURCES_head = NULL;
    dev_ptr = adp_str.DEVICE_WAITING_head;
    while (dev_ptr != NULL)
    {
	psc_deq_wait(dev_ptr);
	(void) psc_free_resources(dev_ptr, FALSE);
	if (dev_ptr->head_pend == NULL)
	{	/* queue is empty  */
	    dev_ptr->head_pend = dev_ptr->active;
	    dev_ptr->tail_pend = dev_ptr->active;
	}
	else
	{	/* pending queue not empty */
	    /* point first cmd's av_forw at the new request */
	    dev_ptr->active->bufstruct.av_forw =
					 (struct buf *) dev_ptr->head_pend;
	    dev_ptr->head_pend = dev_ptr->active;
	}
	dev_ptr->active = NULL;
	dev_ptr = adp_str.DEVICE_WAITING_head;
    }
    dev_ptr = adp_str.DEVICE_ACTIVE_head;
    while (dev_ptr != NULL)
    {
	/***************************************************/
	/* put an abort command on the ABORT_BDR queue for */
	/* this device	                                   */
	/***************************************************/
	dev_ptr->flags |= SCSI_ABORT;
	psc_enq_abort_bdr(dev_ptr);
	dev_ptr = dev_ptr->DEVICE_ACTIVE_fwd;
    }
    DEBUG_0 ("Leaving psc_dumpstrt\n")
    TRACE_1 ("out dumpstrt", ret_code)
    return (ret_code);
}  /* end psc_dumpstrt */

/**************************************************************************/
/*                                                                        */
/* NAME: psc_dump_intr                                                    */
/*                                                                        */
/* FUNCTION: Handle polling for interrupts for dump routines              */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*                                                                        */
/*      This routine may be call in the interrupt or process environment. */
/*      It can not page fault and is pinned.                              */
/*                                                                        */
/* (DATA STRUCTURES:)                                                     */
/*                    adapter_def    - adapter info structure             */
/*                    sc_buf    - controller info structure               */
/*                    dev_info    - device info structure                 */
/*                                                                        */
/* INPUTS:                                                                */
/*    dev_info structure - pointer to device information structure        */
/*                                                                        */
/*                                                                        */
/* CALLED BY:                                                             */
/*      psc_dumpwrt, psc_dumpstrt                                         */
/*                                                                        */
/* INTERNAL PROCEDURES CALLED:                                            */
/*      psc_free_resources          psc_read_reg       psc_delay          */
/*      psc_update_dataptr          psc_write_reg      psc_find_devinfo   */
/*      psc_verify_neg_answer       psc_deq_active     psc_fail_cmd       */
/*                                                                        */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*      word_reverse                                                      */
/*                                                                        */
/* RETURNS:                                                               */
/*    EIO - register failures                                             */
/*    ETIMEDOUT - command time outs                                       */
/*                                                                        */
/**************************************************************************/
int
psc_dump_intr(struct dev_info * dev_ptr)

{
    int dev_index, ret_code, loop, i, j, scsi_intr, dma_intr;
    int rc, save_isr, save_stat, dma_stat, save_interrupt;
    int start_new_job;   /* 1=start new job up */
    int multiple_intr, time_out_value, command_issued_flag;
    int istat_val;
    caddr_t iocc_addr;
    struct sc_buf *bp;

    DEBUG_0 ("Entering psc_dump_intr\n")
    TRACE_1 ("in dumpintr", 0)
    /* A read of the istat register indicates that */
    /* the scsi chip is currently not busy on the  */
    /* scsi bus, so try to signal the chip that a  */
    /* command is waiting.                         */
    istat_val = -1;
    iocc_addr =
        BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
    psc_patch_iowait_int_on();
    for (j = 0; j < 3; j++)
    {
        rc = BUS_GETCX(iocc_addr + ISTAT, &istat_val);
        if (!(istat_val & CONN_INT_TEST))
        {	/* The register read is successful */
	    rc |= BUS_PUTCX(iocc_addr + ISTAT, 0x80);
	    if (rc == REG_FAIL)
	    {	/* Failed during a register write */
	        istat_val = rc;
	    }
        }
        if (istat_val != REG_FAIL)
        {	/* Can't issue abort so  */
	    break;
        }
    }
    BUSIO_DET(iocc_addr);
    loop = TRUE;
    i = 0;
    time_out_value = dev_ptr->active->timeout_value;
    while ((loop) && (++i < (time_out_value * 1000)))
    {
	/* delay for 1000th of a second */
	psc_delay(1000);
	start_new_job = FALSE;
	/* check for interrupts */
	rc = psc_read_reg(SCSIRS, SCSIRS_SIZE);
	/* Micro channel check condition */
	if (rc == REG_FAIL)
	{
	    ret_code = EIO;
	    loop = FALSE;
	    break;
	}
	/* if not our interrupt go around again */
	save_isr = psc_read_reg(ISTAT, ISTAT_SIZE);
	if (!(save_isr & (SIP | DIP)))
	    continue;
	else
	    if (save_isr & SIP)	/* Planned SCSI interrupt */
	    {
		scsi_intr = psc_read_reg(SSTAT0, SSTAT0_SIZE);
		save_interrupt = scsi_intr;
		multiple_intr = FALSE;
		do
		{
		    switch (scsi_intr)
		    {
		    case PHASE_MIS:
			TRACE_1 ("duPHASE_MIS", 0)
			DEBUG_0 ("psc_dump: PHASE_MIS\n")
			if ((dev_ptr = psc_find_devinfo(0)) !=
			     NULL)
			{
			    TRACE_1 ("dphse ms", dev_ptr)
			    psc_update_dataptr(adp_str.
				           SCRIPTS[dev_ptr->cmd_script_ptr].
					   script_ptr, dev_ptr);
			    /* clear the DMA and SCSI FIFO */
			    if (psc_write_reg(DFIFO,
					      DFIFO_SIZE,
					      0x40) == REG_FAIL)
			    {
				loop = FALSE;
				multiple_intr = FALSE;
				ret_code = EIO;
				break;
			    }
			}
			else
			{
			    TRACE_1 ("no dev_ptr", 0)
			    DEBUG_0 ("PHS_MIS no dev_ptr\n")
			    psc_write_reg(DSP, DSP_SIZE,
					  word_reverse(adp_str.SCRIPTS[0].
						       dma_ptr));
			    start_new_job = FALSE;
			    multiple_intr = FALSE;
			    break;
			}
			start_new_job = FALSE;
			ISSUE_MAIN_AFTER_NEG(dev_ptr->script_dma_addr);
			multiple_intr = FALSE;
			break;

		    case SCSI_UNEXP_DISC:
			DEBUG_1 ("psc_dumpstrt: scsi unexp disc. 0x%x\n",
					 save_interrupt)
			TRACE_1 ("du_UNEXPDISC", 0)
			dev_ptr = psc_find_devinfo(0);
			if (dev_ptr != NULL)
			{
			    psc_deq_active(dev_ptr);
			}
		        if (dev_ptr->cmd_activity_state ==
			    ABORT_IN_PROGRESS)
			{
			    ret_code = PSC_NO_ERR;
			}
			else
			{
			    ret_code = ENXIO;
			}
			loop = FALSE;
			multiple_intr = FALSE;
			break;

		    default:
			TRACE_1 ("du_default", scsi_intr)
			if (scsi_intr & PHASE_MIS)
			{
			    scsi_intr = PHASE_MIS;
			    multiple_intr = TRUE;
			}
			break;
		    }	/* end switch */
		} while (multiple_intr);
	    }	/* End Planned SCSI interrupt */
	    else
		if (save_isr & DIP)	/* DMA interrupt */
		{
		    save_interrupt = psc_read_reg(DSTAT, DSTAT_SIZE);
		    dma_stat = word_reverse(psc_read_reg(DSPS, DSPS_SIZE));
		    dma_intr = 0x0000FFFF & dma_stat;
		    dev_index = ((dma_stat & 0xFFFF0000) >> 16);
		    dev_ptr = adp_str.device_queue_hash[dev_index];
		    switch (dma_intr)
		    {
		    case A_set_special_sync:
			DEBUG_0 ("psc_dump_intr: A_set_special_sync \n")
			TRACE_1 ("du_spc sync", dev_ptr)
			/* set xfer mode in chip to SYNC */
			if (psc_write_reg(SXFER, SXFER_SIZE,
				     dev_ptr->special_xfer_val) == REG_FAIL)
			{
			    ret_code = EIO;
			    loop = FALSE;
			    break;
			}
			/* and restart script */
			rc = psc_read_reg(DSP, DSP_SIZE);
			if (rc == REG_FAIL)
			{
			    ret_code = EIO;
			    loop = FALSE;
			    break;
			}
			if (psc_write_reg(DSP, DSP_SIZE, rc) == REG_FAIL)
			{
			    ret_code = EIO;
			    loop = FALSE;
			}
			break;

		    case A_change_to_async:
			DEBUG_0 ("psc_intr: A_change_to_async \n");
			TRACE_1 ("du_chg2async", 0);
			/* set xfer mode in chip to ASYNC */
			rc = psc_write_reg(SXFER, SXFER_SIZE, 0x00);
			if (rc == REG_FAIL)
			{
			    ret_code = EIO;
			    loop = FALSE;
			    break;
			}
			rc = psc_read_reg(DSP, DSP_SIZE);
			if (rc == REG_FAIL)
			{
			    ret_code = EIO;
			    loop = FALSE;
			    break;
			}
			if (psc_write_reg(DSP, DSP_SIZE, rc) == REG_FAIL)
			{
			    ret_code = EIO;
			    loop = FALSE;
			    break;
			}
			break;

		    case A_sync_neg_done:
			/* sync complete and successful */
			DEBUG_0 ("psc_intr: SYNC_NEG_DONE \n")
			TRACE_1 ("du_syncneg", dev_ptr)
			/* supports sync negotiation */
			dev_ptr->negotiate_flag = FALSE;
			/* sync complete and successful */
			dev_ptr->async_device = FALSE;
			if (psc_write_reg(SXFER, SXFER_SIZE, SYNC_VAL)
			    == REG_FAIL)
			{
			    ret_code = EIO;
			    loop = FALSE;
			    break;
			}
			/* supports synch negotiation */
			psc_verify_neg_answer(adp_str.SCRIPTS[dev_ptr->
						cmd_script_ptr].script_ptr,
					      dev_ptr,
					      dev_index);
			DEBUG_0 ("psc_dump: SYNC_NEG_DONE \n")
			/* setup script to issue the command    */
			/* already tagged to this device queue  */
			ret_code = PSC_NO_ERR;
			loop = TRUE;
			break;

		    case A_io_done:
		    case A_io_done_after_data:
			DEBUG_0 ("psc_dump: Current I/O done\n")
			TRACE_1 ("du_A_iodone", dev_ptr)
			/* reset completion flags */
			bp = dev_ptr->active;
			if (dma_intr == A_io_done_after_data)
			{
			    dev_ptr->active->bufstruct.b_resid = 0;
			    /* resid = 0 */
			}
			else
			    if (dev_ptr->flags & RESID_SUBTRACT)
			    {
				dev_ptr->active->bufstruct.b_resid -=
				    dev_ptr->bytes_moved;
				/* do subtraction */
			    }
			save_stat = GET_CMD_STATUS_BYTE(adp_str.
					   SCRIPTS[dev_ptr->cmd_script_ptr].
							script_ptr);
			DEBUG_1 ("STATUS BUFFER = 0x%x\n", save_stat)
			start_new_job = TRUE;
			if (save_stat == SC_GOOD_STATUS)
			{	/* scsi status is ok */
			    if ((bp != NULL) &&
				(!(bp->bufstruct.b_flags & B_ERROR)))
			    {
				TRACE_1 ("A_iodgds", dev_ptr)
				/* set scsi status in sc_buf */
				bp->bufstruct.b_error = 0;
				bp->status_validity = 0;
				bp->scsi_status = SC_GOOD_STATUS;
				/* free resources */
				rc = psc_free_resources(dev_ptr, TRUE);
			    }
			}
			else	/* status is not okay */
			{
			    if (adp_str.dump_started)
			    {
				loop = FALSE;
				ret_code = EIO;
				break;
			    }
			    if ((bp != NULL) &&
				(!(bp->bufstruct.b_flags &
				   B_ERROR)))
			    {
				TRACE_1 ("diodchk", dev_ptr);
				/* set scsi stat iuf */
				bp->scsi_status = save_stat;
				bp->status_validity = SC_SCSI_ERROR;
				bp->bufstruct.b_error = EIO;
				bp->bufstruct.b_flags |= B_ERROR;
				rc = psc_free_resources(dev_ptr, FALSE);
			    }
			    else
				if (bp != NULL)
				{	/* previous error */
				    TRACE_1 ("A_iodpe2", dev_ptr)
				    (void) psc_free_resources(dev_ptr, FALSE);
				}
			}	/* end else */
			if ((adp_str.ABORT_BDR_head == dev_ptr) ||
			    (dev_ptr->ABORT_BDR_bkwd != NULL))
			{
			    psc_deq_abort_bdr(dev_ptr);
			}
			dev_ptr->active = NULL;
			psc_deq_active(dev_ptr);
			/* finished this phase of xfer */
			loop = FALSE;
			ret_code = PSC_NO_ERR;
			break;
			/* tell controller to keep going */

		    case A_io_wait:
			TRACE_1 ("io wait ", dev_ptr)
			psc_patch_iowait_int_off();
			rc = psc_write_reg(DSP, DSP_SIZE,
				  word_reverse(adp_str.SCRIPTS[0].dma_ptr));
			start_new_job = FALSE;
			if (rc == REG_FAIL)
			{
			    ret_code = EIO;
			    loop = FALSE;
			}
			break;

		    case A_abort_io_complete:
		    case A_bdr_io_complete:
			TRACE_1 ("dbdr cmp", dev_ptr);
			dev_ptr->cmd_activity_state = 0;
			w_stop(&dev_ptr->dev_watchdog.dog);
			if (psc_write_reg(SXFER, SXFER_SIZE, SYNC_VAL) ==
			    REG_FAIL)
			{
			    ret_code = EIO;
			    loop = FALSE;
			    break;
			}
			/* remove device from active queue */
			/* if there is a command tagged to this */
			if ((bp = dev_ptr->active) != NULL)
			{
			    if (!(bp->bufstruct.b_flags & B_ERROR))
			    {	/* no previous error */
				bp->bufstruct.b_resid =
						     bp->bufstruct.b_bcount;
				bp->bufstruct.b_flags |= B_ERROR;
				bp->bufstruct.b_error = ENXIO;
			    }
			    rc = psc_free_resources(dev_ptr, FALSE);
			}
			dev_ptr->flags = RETRY_ERROR;
			dev_ptr->active = NULL;
			/* iodone pending commands and go issue a new one */
			if (dev_ptr->head_pend != NULL)
			    psc_fail_cmd(dev_ptr, ENXIO, 1);
			psc_deq_active(dev_ptr);
			start_new_job = TRUE;
			loop = FALSE;
			ret_code = PSC_NO_ERR;
			break;

		    case A_check_next_io:
		    case A_check_next_io_data:
			save_stat = GET_CMD_STATUS_BYTE(adp_str.
					   SCRIPTS[dev_ptr->cmd_script_ptr].
							script_ptr);
			/**********************************************/
			/* the device is about to disc. from the bus. */
			/* and a save DP has just been sent.  Move any */
			/* dataxferred into the buffer and adjust the */
			/* data pointer within our buffer to show     */
			/* buffer size has changed.                   */
			/**********************************************/
			if (dma_intr == A_check_next_io_data)
			{
			    dev_ptr->active->bufstruct.b_resid = 0;
			    /* resid = 0 */
			}
			else
			    if (dev_ptr->flags & RESID_SUBTRACT)
			    {
				dev_ptr->active->bufstruct.b_resid -=
				    dev_ptr->bytes_moved;
			    }
			/* Turn off resid subtract */
			dev_ptr->flags &= ~RESID_SUBTRACT;
			TRACE_1 ("dchk nxt", dev_ptr)
			if (psc_write_reg(SXFER, SXFER_SIZE, SYNC_VAL)
			    ==  REG_FAIL)
			{
			    TRACE_1 ("du_regfail", SXFER)
			    loop = FALSE;
			    ret_code = EIO;
			}
			else
			{
			    dev_ptr = NULL;
			    start_new_job = TRUE;
			}

			break;

		    case A_abort_select_failed:
			TRACE_1 ("d_abtselfail", 0)
			dev_ptr = adp_str.DEVICE_ACTIVE_head;
			while (dev_ptr != NULL)
			{
			    psc_restore_iowait_jump(
					      adp_str.SCRIPTS[0].script_ptr,
					 dev_ptr, dev_ptr->script_dma_addr);
			    dev_ptr = dev_ptr->DEVICE_ACTIVE_fwd;
			}
			psc_patch_iowait_int_off();
			psc_write_reg(DSP, DSP_SIZE, word_reverse(
					       adp_str.SCRIPTS[0].dma_ptr));
			start_new_job = FALSE;
			break;

		    default:
			TRACE_1 ("d_default2", dma_intr)
			start_new_job = TRUE;
			break;
		    }	/* End switch */
		    /******************************************/
		    /* ABORT CAUSED THIS INTERRUPT    */
		    /******************************************/
		    if (save_interrupt & DABRT)
		    {
			DEBUG_0 ("psc_dump_intr: Abort caused interrupt\n")
			TRACE_1 ("du_abrt int", 0)
			rc = psc_read_reg(ISTAT, ISTAT_SIZE);
			/* A read of the istat register verifies that */
			/* the scsi chip is currently not busy on the */
			/* scsi bus, so we can issue the new command. */
			if (!(rc & CONNECTED))
			{
			    if (psc_write_reg(SXFER, SXFER_SIZE,
					      SYNC_VAL) == REG_FAIL)
			    {
				loop = FALSE;
				ret_code = REG_FAIL;
			    }
			    else
			    {
				dev_ptr = NULL;
				start_new_job = TRUE;
			    }
			}
			else
			{
			    (void) psc_read_reg(DSTAT, DSTAT_SIZE);
			    /* restart iowait script */
			    psc_patch_iowait_int_off();
			    rc = psc_write_reg(DSP, DSP_SIZE,
				  word_reverse(adp_str.SCRIPTS[0].dma_ptr));
			    start_new_job = FALSE;
			    if (rc == REG_FAIL)
			    {
				loop = FALSE;
				ret_code = REG_FAIL;
			    }
			}
		    }
		}	/* End DMA interrupt */
	if (start_new_job)
	{
	    command_issued_flag = psc_issue_cmd();
	    if ((command_issued_flag == FALSE) &&
		(adp_str.DEVICE_ACTIVE_head != NULL))
	    {
		dev_ptr = adp_str.DEVICE_ACTIVE_head;
		while (dev_ptr != NULL)
		{
		    psc_restore_iowait_jump(adp_str.SCRIPTS[0].script_ptr,
					 dev_ptr, dev_ptr->script_dma_addr);
		    dev_ptr = dev_ptr->DEVICE_ACTIVE_fwd;
		}
		psc_patch_iowait_int_off();
		psc_write_reg(DSP, DSP_SIZE, word_reverse(
					       adp_str.SCRIPTS[0].dma_ptr));
	    }
	}
    }	/* while */
    /****************************************************************/
    /* if the command did not complete within the time specified in */
    /* the command, then free the resources and return an error     */
    /****************************************************************/
    if (i >= (time_out_value * 1000))
    {
	ret_code = ETIMEDOUT;
	TRACE_1 ("time_value", time_out_value)
	TRACE_1 ("loop_count", i)
    }
    if ((ret_code != 0) && (dev_ptr != NULL))
    {
	(void) psc_free_resources(dev_ptr, FALSE);
	dev_ptr->active = NULL;
	psc_deq_active(dev_ptr);
    }
    DEBUG_0 ("Leaving psc_dumpintr\n")
    TRACE_1 ("out dintr", ret_code)
    return (ret_code);
}  /* end psc_dump_intr */

/*************************************************************************/
/*                                                                       */
/* NAME: psc_dumpwrt                                                     */
/*                                                                       */
/* FUNCTION: Write to the dump device.                                   */
/*                                                                       */
/*          Issue as many scsi commands as neccessry to write uiop's     */
/*      data to device using the SCSI write command.  If any errors      */
/*      the fail and return with error.                                  */
/*      Actual division of labor is the dumpwrt monitors the looping     */
/*      and checks for timeouts and interrupts.  psc_dump_dev actually   */
/*      builds the command and sends it to the device.  Returning status */
/*      when it is complete.                                             */
/*                                                                       */
/* EXECUTION ENVIRONMENT:                                                */
/*                                                                       */
/*      This routine can  be called in the interrupt environment.  It    */
/*      can not page fault and is pinned.                                */
/*                                                                       */
/* (DATA STRUCTURES:)                                                    */
/*                    adapter_def    - adapter info structure            */
/*                    sc_buf    - controller info structure              */
/*                    dev_info    - device info structure                */
/*                                                                       */
/* INPUTS:                                                               */
/*      bp       -  buf scruct for the pending dump command              */
/*                                                                       */
/* CALLED BY:                                                            */
/*      psc_dump                                                         */
/*                                                                       */
/* INTERNAL PROCEDURES CALLED:                                           */
/*      psc_dump_dev     psc_dump_intr                                   */
/*                                                                       */
/* EXTERNAL PROCEDURES CALLED:                                           */
/*                                                                       */
/* (RECOVERY OPERATION:) If an error occurs, the proper errno is return- */
/*      ed and the caller is left responsible to recover from the error. */
/*                                                                       */
/* RETURNS:                                                              */
/*      EBUSY     -  request sense inuse                                 */
/*      EINVAL    -  Invalid iovec argument                              */
/*      EIO       -  I/O error                                           */
/*      ENOMEM    -  Unable to allocate resources                        */
/*      ENXIO     -  Not inited as dump device                           */
/*      ETIMEDOUT - timed out because no interrupt                       */
/*                                                                       */
/*************************************************************************/
int
psc_dumpwrt(
	    struct sc_buf * bp)
{
    /**********************/
    /* Misc. Variables    */
    /**********************/
    struct dev_info *dev_ptr;
    int     dev_index, ret_code, loop, i, scsi_intr, dma_intr;
    int     rc, save_isr, save_stat, dma_stat;
    /*********************************************/
    /* We will have to poll for interrupts, and  */
    /* handle them in the apropriate ways        */
    /*********************************************/
    DEBUG_0 ("Entering psc_dumpwrt\n")
    TRACE_1 ("in dumpwrt", 0)

    if ((!adp_str.dump_inited) || (!adp_str.dump_started))
    {
	(void) psc_read_reg(SCSIRS, SCSIRS_SIZE);
	return (EINVAL);
    }
    dev_index = INDEX(bp->scsi_command.scsi_id,
		      (bp->scsi_command.scsi_cmd.lun) >> 5);
    dev_ptr = adp_str.device_queue_hash[dev_index];
    if (dev_ptr == NULL)
    {
	(void) psc_read_reg(SCSIRS, SCSIRS_SIZE);
	return (ENXIO);
    }
    ret_code = psc_dump_dev(dev_ptr);	/* send a cmd to the disk */
    if (ret_code)
	return (ret_code);

    DEBUG_0 ("Leaving psc_dumpwrt\n")
    TRACE_1 ("out dumpwrt", 0)
    return (PSC_NO_ERR);
}  /* end psc_dumpwrt */

/***************************************************************************/
/*                                                                         */
/* NAME: psc_dump_dev                                                      */
/*                                                                         */
/* FUNCTION: Send scsi write cmd to the dump device                        */
/*                                                                         */
/*      This routine fills in cmd for the write scsi command for           */
/*      all data in one iovec. It also fills in the scsi command block for */
/*      scsi write.   And Patches the script and executes it.              */
/*                                                                         */
/* EXECUTION ENVIRONMENT:                                                  */
/*                                                                         */
/*      This routine can  be called in the interrupt environment.  It      */
/*      can not page fault and is pinned.                                  */
/*                                                                         */
/* (DATA STRUCTURES:)                                                      */
/*                    adapter_def    - adapter info structure              */
/*                    sc_buf    - controller info structure                */
/*                    dev_info    - device info structure                  */
/*                                                                         */
/* INPUTS:                                                                 */
/*                                                                         */
/*    dev_info structure - pointer to device information structure         */
/*                                                                         */
/* CALLED BY:                                                              */
/*      psc_dumpwrt                                                        */
/*                                                                         */
/* INTERNAL PROCEDURES CALLED:                                             */
/*                                                                         */
/* EXTERNAL PROCEDURES CALLED:                                             */
/*                                                                         */
/* (RECOVERY OPERATION:) If an error occurs, the proper errno is returned  */
/*      and the caller is left responsible to recover from the error.      */
/*                                                                         */
/*                                                                         */
/* RETURNS:                                                                */
/*      EINVAL -  Invalid iovec argument                                   */
/*      EIO    -  I/O error                                                */
/*      ENOMEM -  Unable to allocate resources                             */
/*                                                                         */
/***************************************************************************/
int
psc_dump_dev(
	     struct dev_info * dev_ptr)
{
    struct sc_buf *bp;

    DEBUG_0 ("Entering psc_dump_dev routine.\n")
    TRACE_1 ("in dump_dev", 0)
    if (psc_alloc_resources(dev_ptr) != PSC_NO_ERR)
	return (ENOMEM);
    /* assumes that negotiation is unnecessary at this point */
    if (!dev_ptr->negotiate_flag)
    {
	if (dev_ptr->async_device)
	{
	    if (psc_write_reg(SXFER, SXFER_SIZE, 0x00) == REG_FAIL)
	    {
		return (REG_FAIL);
	    }
	}
	else
	    if ((dev_ptr->agreed_xfer != DEFAULT_MIN_PHASE) ||
		(dev_ptr->agreed_req_ack < DEFAULT_BYTE_BUF))
	    {
		if (psc_write_reg(SXFER, SXFER_SIZE,
		     dev_ptr->special_xfer_val) == REG_FAIL)
		{
		    return (REG_FAIL);
		}
	    }
	else
	{
	    if (psc_write_reg(SXFER, SXFER_SIZE, SYNC_VAL) == REG_FAIL)
	    {
		return (REG_FAIL);
	    }
	}
	/* patch labels to script */
	psc_prep_main_script(adp_str.SCRIPTS[0].script_ptr,
			adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].script_ptr,
			dev_ptr, adp_str.SCRIPTS[0].dma_ptr,
			dev_ptr->script_dma_addr);
	/* start dump SCRIPT */
	ISSUE_MAIN_TO_DEVICE(dev_ptr->script_dma_addr);
	dev_ptr->cmd_activity_state = CMD_IN_PROGRESS;
	psc_enq_active(dev_ptr);
    }
    else
    {
	return (REG_FAIL);
    }
    DEBUG_0 ("Leaving psc_dump_dev routine.\n");
    TRACE_1 ("out dump_dev", 0);
    return (PSC_NO_ERR);
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_cdt_func                                            */
/*                                                                      */
/* FUNCTION:    Adapter Driver Component Dump Table Routine             */
/*                                                                      */
/*      This routine builds the driver dump table during a system dump. */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine runs on the interrupt level, so it cannot malloc   */
/*      or free memory.                                                 */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      cdt     - the structure used in the master dump table to keep   */
/*                track of component dump entries.                      */
/*                                                                      */
/* INPUTS:                                                              */
/*      arg     - =1 dump dd is starting to get dump table entries.     */
/*                =2 dump dd is finished getting the dump table entries.*/
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      Return code is a pointer to the struct cdt to be dumped for     */
/*      this component.                                                 */
/*                                                                      */
/* ERROR DESCRIPTION:  The following errno values may be returned:      */
/*      none                                                            */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      bcopy           bzero                                           */
/*      strcpy                                                          */
/*                                                                      */
/************************************************************************/
struct cdt *
psc_cdt_func(
	     int arg)
{
    int     i, d_index, size;
    struct adapter_def *ap;

    DEBUG_0 ("Entering psc_cdt_func routine.\n");

    if (arg == 1)
    {
	/* only build the dump table on the initial dump call */

	/* init the table */
	bzero((char *) psc_cdt, sizeof(struct psc_cdt_table));

	/* init the head struct */
	psc_cdt->psc_cdt_head._cdt_magic = DMP_MAGIC;
	strcpy(psc_cdt->psc_cdt_head._cdt_name, "pscsi");
	/* _cdt_len is filled in below */

	/* now begin filling in elements */
	psc_cdt->psc_entry[0].d_segval = 0;
	strcpy(psc_cdt->psc_entry[0].d_name, "adpstruc");
	psc_cdt->psc_entry[0].d_ptr = (char *) &adp_str;
	psc_cdt->psc_entry[0].d_len = (sizeof(struct adapter_def));
	size = sizeof(struct cdt_head) + sizeof(struct cdt_entry);

	d_index = 0;
	ap = &adp_str;
	psc_cdt->psc_entry[d_index].d_segval = 0;
	/* copy name to element:  s, d, l */
	bcopy((char *) ap->ddi.resource_name,
	      (char *) psc_cdt->psc_entry[d_index].d_name, 8);
	psc_cdt->psc_entry[d_index].d_ptr = (char *) ap;
	psc_cdt->psc_entry[d_index].d_len = (sizeof(struct adapter_def));

	/* fill in the actual table size */
	psc_cdt->psc_cdt_head._cdt_len = size;
    }
    return ((struct cdt *) psc_cdt);

}  /* end psc_cdt_func */

/**************************************************************************/
/*                                                                        */
/* NAME:        psc_fail_cmd                                              */
/*                                                                        */
/* FUNCTION:    Fail Active and Pending Commands Routine                  */
/*                                                                        */
/*      This internal routine is called by the interrupt handler          */
/*      to clear out pending and active commands for a particular         */
/*      device which has experienced some sort of failure which is        */
/*      not recoverable by the adapter driver.                            */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can only be called on priority levels greater        */
/*      than, or equal to that of the interrupt handler.                  */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      adapter_def - adapter unique data structure (one per adapter)     */
/*      sc_buf  - input/output request struct used between the adapter    */
/*                driver and the calling SCSI device driver               */
/*                                                                        */
/* INPUTS:                                                                */
/*      dev_info structure - pointer to device information structure      */
/*      err_type - type of errors which would cause case to be killed     */
/*      queue_type - which queue 0 = active 1= pending 2=both             */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:  The following are the return values:        */
/*      none                                                              */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*      i_disable       i_enable                                          */
/*      iodone          w_stop                                            */
/*                                                                        */
/**************************************************************************/
void
psc_fail_cmd(
	     struct dev_info * dev_ptr,
	     char err_type,
	     int queue_type)
{
    struct sc_buf *bp;
    int     old_pri;

    DEBUG_0 ("Entering psc_fail_cmd routine.\n")
    TRACE_1 ("in failC", dev_ptr)

    /* this assumes that for the failed command(s), the  */
    /* following are set previously:  sc_buf status,     */
    /* sc_buf resid value, and sc_buf b_error field.     */
    old_pri = i_disable(adp_str.ddi.int_prior);

    /* clean up active cmd queue */
    if ((queue_type != 1) &&	/* active queue or both */
	(dev_ptr->active != NULL))
    {
	bp = dev_ptr->active;
	/* set b_flags B_ERROR flag */
	bp->bufstruct.b_flags |= B_ERROR;
	bp->bufstruct.b_error = err_type;
	(void) psc_free_resources(dev_ptr, FALSE);
	iodone((struct buf *) bp);
	dev_ptr->flags = RETRY_ERROR;
	dev_ptr->active = NULL;	/* reset tail pointer */
    }	/* end of if */
    /* clean up pending cmd queue */
    if (queue_type != 0)
    {
	bp = dev_ptr->head_pend;
	while (bp != NULL)
	{
	    bp->status_validity = 0;
	    bp->bufstruct.b_flags |= B_ERROR;	/* set b_flags B_ERROR flag */
	    bp->bufstruct.b_error = err_type;	/* set b_flags B_ERROR flag */
	    /* point to next sc_buf in pending chain, if any */
	    dev_ptr->head_pend = (struct sc_buf *) bp->bufstruct.av_forw;
	    iodone((struct buf *) bp);
	    bp = dev_ptr->head_pend;
	}	/* endwhile */
	dev_ptr->tail_pend = NULL;	/* reset tail pointer */
    }
    if (dev_ptr->queue_state == STOPPING)
    {
	e_wakeup(&dev_ptr->stop_event);
    }
    else
    {
	dev_ptr->queue_state = HALTED;
    }
    TRACE_1 ("out fail", dev_ptr)
    i_enable(old_pri);
}  /* end psc_fail_cmd */

/*************************************************************************/
/*                                                                      */
/* NAME:        psc_delay                                               */
/*                                                                      */
/* FUNCTION:    provide variable delay in 1 usec increments             */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*     process or interrupt                                             */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*                                                                      */
/* INPUTS:                                                              */
/*      delay - number of usecs to delay                                */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*                                                                      */
/* ERROR DESCRIPTION:  The following errno values may be returned:      */
/*      none                                                            */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*                                                                      */
/************************************************************************/
int
psc_delay(int delay)
{
    caddr_t iocc_addr;
    int     i;
    uchar   no_val;

    iocc_addr = BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
    for (i = 0; i < delay; i++)
	/* GETCX due to hardware restriction */
	(void) BUS_GETCX((iocc_addr + 0xe0), &no_val);
    BUSIO_DET(iocc_addr);
}

/*************************************************************************/
/*                                                                      */
/* NAME:        psc_intr                                                */
/*                                                                      */
/* FUNCTION:    Adapter Driver Interrupt Handler                        */
/*                                                                      */
/*      This routine processes adapter interrupt conditions.            */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine runs on the interrupt level, therefore, it must    */
/*      only perform operations on pinned data.                         */
/*                                                                      */
/* NOTES:  This routine handles interrupts which are caused by scsi     */
/*     scripts processing.  This entails processing the interrupts      */
/*     handling error cleanup and logging, and issuing outstanding      */
/*     commands for other devices                                       */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      adapter_def - adapter unique data structure (one per adapter)   */
/*      sc_buf  - input/output request struct used between the adapter  */
/*                driver and the calling SCSI device driver             */
/*      dev_info - driver structure which holds information related to  */
/*                 a particular device and hence a command              */
/*      intr    - kernel interrupt information structure                */
/*                                                                      */
/* INPUTS:                                                              */
/*      handler - pointer to the intrpt handler structure               */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      INTR_FAIL    interrupt was not processed                        */
/*      INTR_SUCC    interrupt was processed                            */
/*                                                                      */
/* ERROR DESCRIPTION:  The following errno values may be returned:      */
/*      none                                                            */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*     w_start                                                          */
/*     w_clear             i_reset                                      */
/*     iodone              e_wakeup                                     */
/*     bzero               d_complete                                   */
/*                                                                      */
/************************************************************************/
int
psc_intr(struct intr * handler)
{
    struct sc_buf *bp;
    struct dev_info *dev_ptr;
    int     save_isr, save_dsr;	/* save status registers */
    int     save_interrupt, dma_stat;
    int     command_issued_flag, check_for_disconnect;
    int     issue_abrt_bdr;	/* are we doing abort/bdr */
    int     dev_index;
    int     dma_addr;
    int     start_new_job;	/* 1=start new job up */
    int     temp_dsp_val, temp_dsp_val2;
    uchar   save_stat, chip_abort_occurred;
    register int rc;

    DEBUG_0 ("Entering psc_intr routine\n")
    TRACE_1 ("in intr", 0)

    /* check and make sure its a scsi interrupt */
    rc = psc_read_reg(SCSIRS, SCSIRS_SIZE);
    DEBUG_1 ("psc_intr: SCSIRS = 0x%x\n", rc);
    DDHKWD1(HKWD_DD_SCSIDD, DD_ENTRY_INTR, 0, 0);
    /* channel check and associated interrupts */
    if (!((rc == REG_FAIL) || (rc & 0x000000F0)))
    {
	/* Check to see if this is our interrupt */
	if (!(rc & 0x00000001))
	{
	    /* stop the adapter's watchdog timer for aborts */
	    w_stop(&adp_str.adap_watchdog.dog);

	    /* Check to see if a command is active */
	    if (adp_str.DEVICE_ACTIVE_head != NULL)
	    {
		dma_addr = DMA_ADDR(adp_str.ddi.tcw_start_addr,
				    adp_str.STA_tcw + 1);
		rc = d_complete((int) adp_str.channel_id, DMA_TYPE,
				adp_str.SCRIPTS[0].script_ptr,
				(size_t) PAGESIZE,
				&adp_str.xmem_SCR, (char *) dma_addr);
		/* if we got a system dma error, we need to kill */
		/* all jobs on all devices.                      */
		if (rc == DMA_SUCC)
		{
		    /* read the ISTAT register to see which interrupt */
		    /* caused us to be called                         */
		    save_isr = psc_read_reg(ISTAT, ISTAT_SIZE);
		    DEBUG_1 ("ISTAT is 0x%x\n", save_isr);
		    if (save_isr != REG_FAIL)
		    {
			if (save_isr & 0x80)
			{	/* if a chip abort has occurred */
			    /* clear the interrupt */
			    chip_abort_occurred = TRUE;
			    if (psc_write_reg(ISTAT, ISTAT_SIZE,
				(uchar) 0x00) == REG_FAIL)
			    {
				TRACE_1 ("bad Wist", 0)
				TRACE_1 ("out intr", 0)
				psc_cleanup_reset(REG_ERR);
				i_reset(&(adp_str.intr_struct));
				DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_INTR, 0, 0);
				return (INTR_SUCC);
			    }
			}
			else
			{
			    chip_abort_occurred = FALSE;
			}
			/* initialize flags */
			start_new_job = FALSE;
			/********************************************/
			/* DMA INTERRUPT               */
			/********************************************/
			if ((save_isr & DIP) && (!(save_isr & SIP)))
			    /* DMA caused this interrupt */
			{
			    /* verify the interrupt and check for other
			       conditions */
			    rc = psc_read_reg(DSTAT, DSTAT_SIZE);
			    save_dsr = (rc & DIEN_CHK_MASK);
			    DEBUG_1 ("DSTAT is 0x%x\n", save_dsr);
			    if (rc == REG_FAIL)
			    {
				DEBUG_1 ("psc_intr: save_dsr = 0x%x\n",
					  save_dsr);
				TRACE_1 ("bad dsta", 0)
				TRACE_1 ("out intr", 0)
				psc_cleanup_reset(REG_ERR);
				i_reset(&(adp_str.intr_struct));
				DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_INTR, 0, 0);
				return (INTR_SUCC);
			    }
			    /* if a chip abort was set but another   */
			    /* interrupt won the race, clear the     */
			    /* stacked interrupt for which the abort */
			    /* has taken place.                      */
			    if ((chip_abort_occurred) && (!(save_dsr & DABRT)))
			    {
				(void) psc_read_reg(DSTAT, DSTAT_SIZE);
			    }
			    /**********************************************/
			    /*         PLANNED SCRIPT INTERRUPT           */
			    /**********************************************/
			    if (save_dsr & SIR)
			    {
				dma_stat = word_reverse(psc_read_reg(DSPS,
							DSPS_SIZE));
				DEBUG_1 ("PLANNED INTR is 0x%x\n", dma_stat);
				if (dma_stat == REG_FAIL)
				{
				    DEBUG_0 ("psc_intr: bad reg read DSPS\n")
				    TRACE_1 ("bad dsps", 0)
				    TRACE_1 ("out intr", 0)
				    psc_cleanup_reset(REG_ERR);
				    i_reset(&(adp_str.intr_struct));
				    DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_INTR,
					    0, 0);
				    return (INTR_SUCC);
				}
				TRACE_1 ("dma_stat", dma_stat)
				/* this contains the value of the inter. */
				/* which was set by us in the SCRIPTS to */
				/* to tell us the  type of interupt, the */
				/* location of the interrupt within the */
				/* the SCRIPT, and the device which this */
				/* interrupted SCRIPTS is running.      */
				/* top 16 bits are location of script   */
				/* bottom 16 bits are type of interrupt. */

				/* get top 16 bits which is index into  */
				/* array to get pointer to dev_info     */
				dev_index = ((0xffff0000 & dma_stat) >> 16);
				/* if we get a bad dev_ptr get out */
				dev_ptr =
				    adp_str.device_queue_hash[dev_index];
				if (dev_ptr == NULL)
				{
				    if (((dma_stat & 0x0000ffff) != A_io_wait)
					&& ((dma_stat & 0x0000ffff) !=
					    A_unknown_reselect_id)
					&& ((dma_stat & 0x0000ffff) !=
					    A_uninitialized_reselect))
				    {
					DEBUG_0 ("psc_intr: got bad dev_ptr\n")
					TRACE_1 ("bad dptr", 0)
					TRACE_1 ("out intr", 0)
					psc_logerr(ERRID_SCSI_ERR2,
						   ADAPTER_INT_ERROR,
						   85, 0, NULL, TRUE);
					i_reset(&(adp_str.intr_struct));
					DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_INTR,
						0, 0);
					return (INTR_SUCC);
				    }
				}
				else
				{
				    if ((adp_str.ABORT_BDR_head == dev_ptr) ||
					(dev_ptr->ABORT_BDR_bkwd != NULL))
				      /* there is an abort or bdr to queued */
					issue_abrt_bdr = TRUE;

				    else
				      /* there isn't an abort or bdr queued */
					issue_abrt_bdr = FALSE;
				}
				/* get bottom 16 bits ( interrupt type) so we
				   can switch it */
				save_interrupt = 0x0000ffff & dma_stat;
				/******************************************/
				/*       SWITCH ON TYPE OF INTERRUPT      */
				/******************************************/
				switch (save_interrupt)
				{	/* completed command */
				case A_check_next_io:
				    TRACE_1 ("chk next", dev_ptr)
				    case A_check_next_io_data:
				    /* the device is about to disc. from */
				    /* the bus. and a save DP has just   */
				    /* been sent. Move any data xferred  */
				    /* into the buffer and adjust the    */
				    /* data pointer within our buffer to */
				    /* show the buffer size has changed. */
				    TRACE_1 ("chk next dat", dev_ptr)
				    if (save_interrupt == A_check_next_io_data)
				    {
					dev_ptr->active->bufstruct.b_resid = 0;
				    }
				    else
					if (dev_ptr->flags & RESID_SUBTRACT)
					{
					    dev_ptr->active->bufstruct.b_resid
						-= dev_ptr->bytes_moved;
					}
				    /* Turn off resid subtract */
				    dev_ptr->flags &= ~RESID_SUBTRACT;
				    if (psc_write_reg(SXFER, SXFER_SIZE,
						      SYNC_VAL) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
				    }
				    else
				    {
					dev_ptr = NULL;
					start_new_job = TRUE;
				    }
				    break;

				case A_io_done:
				case A_io_done_after_data:
				    TRACE_1 ("A_iodone", dev_ptr)
				    DEBUG_0 ("psc_intr: Current I/O done\n")
				            w_stop(&dev_ptr->dev_watchdog.dog);
				    if (psc_write_reg(SXFER, SXFER_SIZE,
						      SYNC_VAL) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    /* reset completion flags */
				    if (save_interrupt == A_io_done_after_data)
				    {
					dev_ptr->active->bufstruct.b_resid = 0;
					/* resid = 0 */
				    }
				    else
					if (dev_ptr->flags & RESID_SUBTRACT)
					{
					    dev_ptr->active->bufstruct.b_resid
						-= dev_ptr->bytes_moved;
					    /* do subtraction */
					}
				    if (issue_abrt_bdr)
				    {
					TRACE_1 ("A_iodBDR", dev_ptr)
					        start_new_job =
						psc_issue_abort_bdr(dev_ptr,
								    TRUE);
				    }
				    else	/* issue a cmd */
				    {
					/* issue a command and complete
					   previous command */
					/* get status byte because issue will
					   destroy it */
					save_stat = GET_CMD_STATUS_BYTE(
						adp_str.SCRIPTS[dev_ptr->
						cmd_script_ptr].script_ptr);
					DEBUG_1 ("STATUS BUFFER = 0x%x\n",
						  save_stat);
					bp = dev_ptr->active;
					command_issued_flag = psc_issue_cmd();
					if (save_stat == SC_GOOD_STATUS)
					{	/* scsi status is ok */
					    if ((bp != NULL) &&
						(!(bp->bufstruct.b_flags &
						 B_ERROR)))
					    {
						TRACE_1 ("A_iodgds", dev_ptr)
						/* set scsi status in sc_buf */
						bp->bufstruct.b_error = 0;
						bp->status_validity = 0;
						bp->scsi_status =
							     SC_GOOD_STATUS;
						/* free resources */
						rc = psc_free_resources(
							     dev_ptr,TRUE);
						if (rc != PSC_NO_ERR)
						{
						    if (psc_fail_free_resources
							(dev_ptr, bp, rc))
							break;
						    if (dev_ptr->queue_state
							== STOPPING)
							e_wakeup(
							 &dev_ptr->stop_event);
						    else
							dev_ptr->queue_state
							  = HALTED;
						    break;
						}
						else
						{
						    iodone((struct buf *) bp);
						    if ((dev_ptr->queue_state
							 == STOPPING) &&
							(dev_ptr->head_pend
							 == NULL))
							e_wakeup(
							 &dev_ptr->stop_event);
						}
					    }
					    else
						if (bp != NULL)
						{	/* previous error */
						    TRACE_1 ("A_iodper",
							      dev_ptr)
						    (void) psc_free_resources(
							      dev_ptr, FALSE);
						    iodone((struct buf *)
							    dev_ptr->active);
						    if (dev_ptr->head_pend
							!= NULL)
							psc_fail_cmd(
							  dev_ptr, EIO, 1);
						    if (dev_ptr->queue_state
							== STOPPING)
							e_wakeup(
							 &dev_ptr->stop_event);
						    else
							dev_ptr->queue_state
							   = HALTED;
						}
					}
					else	/* status is not okay */
					{
					    if ((bp != NULL) &&
						(!(bp->bufstruct.b_flags
						   & B_ERROR)))
					    {
						TRACE_1 ("A_iodchk", dev_ptr)
						/* set scsi stat in sc_buf */
						bp->scsi_status = save_stat;
						bp->status_validity =
							      SC_SCSI_ERROR;
						bp->bufstruct.b_flags |=
							      B_ERROR;
						bp->bufstruct.b_error = EIO;
						rc = psc_free_resources(
							      dev_ptr, FALSE);
						if (rc != PSC_NO_ERR)
						{
						    if (psc_fail_free_resources
							(dev_ptr, bp, rc))
							break;
						}
						else
						    iodone((struct buf *)
							    dev_ptr->active);
					    }
					    else
						if (bp != NULL)
						{	/* previous error */
						    TRACE_1 ("A_iodpe2",
							      dev_ptr)
						    (void) psc_free_resources(
							     dev_ptr, FALSE);
						    iodone((struct buf *)
							    dev_ptr->active);
						}

					    if (dev_ptr->head_pend != NULL)
						psc_fail_cmd(dev_ptr, EIO, 1);
					    if (dev_ptr->queue_state ==
						STOPPING)
						e_wakeup(&dev_ptr->stop_event);
					    else
						dev_ptr->queue_state = HALTED;
					}	/* end else */
					dev_ptr->flags = RETRY_ERROR;
					dev_ptr->active = NULL;
					psc_deq_active(dev_ptr);
					psc_check_wait_queue(dev_ptr,
						       command_issued_flag);
				    }	/* end else  cmd */
				    break;

				case A_set_special_sync:
				    DEBUG_0 ("psc_intr: A_set_special_sync \n")
				    TRACE_1 ("spc sync", dev_ptr)
				    /* set xfer mode in chip to SYNC */
				    if (psc_write_reg(SXFER, SXFER_SIZE,
					dev_ptr->special_xfer_val) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }

				    /* and restart script */
				    rc = psc_read_reg(DSP, DSP_SIZE);
				    if (rc == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    if (psc_write_reg(DSP, DSP_SIZE, rc)
					 == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
				    }
				    break;

				case A_change_to_async:
				    DEBUG_0 ("psc_intr: A_change_to_async \n")
				    TRACE_1 ("chg asyn", dev_ptr)
				    /* set xfer mode in chip to ASYNC */
				    if (psc_write_reg(SXFER, SXFER_SIZE, 0x00)
					 == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }

				    /* and restart script */
				    rc = psc_read_reg(DSP, DSP_SIZE);
				    if (rc == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    if (psc_write_reg(DSP, DSP_SIZE, rc)
					 == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    break;

				case A_io_wait:
				    TRACE_1 ("io wait ", dev_ptr)
				    psc_patch_iowait_int_off();
				    rc = psc_write_reg(DSP, DSP_SIZE,
					  word_reverse(
					   adp_str.SCRIPTS[0].dma_ptr));
				    if (rc == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
				    }
				    break;

				case A_sync_neg_done:
				    TRACE_1 ("neg done", dev_ptr)
				    /* sync complete and successful */
				            dev_ptr->negotiate_flag = FALSE;
				    dev_ptr->async_device = FALSE;
				    if (psc_write_reg(SXFER, SXFER_SIZE,
					 SYNC_VAL) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    /* supports synch negotiation */
				    psc_verify_neg_answer(adp_str.SCRIPTS
					[dev_ptr->cmd_script_ptr].script_ptr,
					dev_ptr, dev_index);
				    DEBUG_0 ("psc_intr: SYNC_NEG_DONE \n")
				            w_stop(&dev_ptr->dev_watchdog.dog);

				    if (issue_abrt_bdr)
					/* something on abort que */
					start_new_job =
					  psc_issue_abort_bdr(dev_ptr, FALSE);
				    else	/* nothing on abort que */
				    {
					/* setup script to issue the command */
					/* already tagged to this device
					   queue  */
					if ((bp = dev_ptr->active) != NULL)
					{
					    psc_prep_main_script(
						adp_str.SCRIPTS[0].script_ptr,
						adp_str.SCRIPTS
						 [dev_ptr->cmd_script_ptr].
						 script_ptr,
						dev_ptr,
						adp_str.SCRIPTS[0].dma_ptr,
						dev_ptr->script_dma_addr);
					    ISSUE_MAIN_AFTER_NEG(
						dev_ptr->script_dma_addr);
					    dev_ptr->dev_watchdog.dog.restart =
						(ulong) bp->timeout_value + 1;
					    w_start(
						&dev_ptr->dev_watchdog.dog);
					    dev_ptr->cmd_activity_state =
						CMD_IN_PROGRESS;
					}
					else
					{
					    start_new_job = TRUE;
					}
				    }
				    break;

				case A_sync_msg_reject:
				    TRACE_1 ("sync rej", dev_ptr)
				    /* we tried to negotiate for synch mode,
				       but */
				    /* failed. ASYNC mode set in
				       psc_patch_async */
				    dev_ptr->negotiate_flag = FALSE;
				    dev_ptr->async_device = TRUE;
				    psc_prep_main_script(
					adp_str.SCRIPTS[0].script_ptr,
					adp_str.SCRIPTS
					 [dev_ptr->cmd_script_ptr].script_ptr,
					dev_ptr, adp_str.SCRIPTS[0].dma_ptr,
					dev_ptr->script_dma_addr);
				    dev_ptr->cmd_activity_state =
							    CMD_IN_PROGRESS;
				    psc_patch_async_switch_int(
					adp_str.SCRIPTS
					[dev_ptr->cmd_script_ptr].script_ptr,
					dev_index);
				    DEBUG_0 ("psc_intr: NEG_FAILED \n")
				    if (psc_write_reg(SXFER,
					 SXFER_SIZE, 0x00) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    /* and restart script */
				    rc = psc_read_reg(DSP, DSP_SIZE);
				    if (rc == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    rc = psc_write_reg(DSP, DSP_SIZE, rc);
				    if (rc == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
				    }
				    break;

				case A_ext_msg:
				case A_modify_data_ptr:
				case A_target_sync_sent:
				    DEBUG_0 ("psc_intr: mod. data ptr \n")
				    TRACE_1 ("ext msg ", dev_ptr)
				    if (psc_write_reg(SXFER, SXFER_SIZE,
					 SYNC_VAL) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    if (rc = psc_handle_extended_messages(
					      adp_str.SCRIPTS
					       [dev_ptr->cmd_script_ptr].
					       script_ptr,
					      dev_ptr->script_dma_addr,
					      dev_index, dev_ptr,
					      save_interrupt) == REG_FAIL)
				    {	/* register error */
					psc_cleanup_reset(REG_ERR);
				    }
				    break;

				case A_abort_io_complete:
				case A_bdr_io_complete:
				    DEBUG_0 ("psc_intr: ABRBDR_IO_COMPLETE \n")
				    TRACE_1 ("A/B done", dev_ptr)
				    w_stop(&dev_ptr->dev_watchdog.dog);
				    if (psc_write_reg(SXFER, SXFER_SIZE,
					 SYNC_VAL) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    /* remove device from active queue */
				    /* if there is a command tagged to this */
				    if ((bp = dev_ptr->active) != NULL)
				    {
					if (!(bp->bufstruct.b_flags & B_ERROR))
					{	/* no previous error */
					    bp->bufstruct.b_resid =
						     bp->bufstruct.b_bcount;
					    bp->bufstruct.b_flags |= B_ERROR;
					    bp->bufstruct.b_error = ENXIO;
					}
					rc = psc_free_resources(dev_ptr,
								FALSE);
					if (rc != PSC_NO_ERR)
					{
					    if (psc_fail_free_resources(
						 dev_ptr, bp, rc))
						break;
					}
					else
					    /* finish off active and pending */
					    /* commands */
					    iodone((struct sc_buf *) bp);
				    }
				    dev_ptr->flags = RETRY_ERROR;
				    dev_ptr->active = NULL;
				    /* iodone pending commands and go */
				    /* issue a new one */
				    if (dev_ptr->head_pend != NULL)
					psc_fail_cmd(dev_ptr, ENXIO, 1);
				    if (dev_ptr->queue_state == STOPPING)
					e_wakeup(&dev_ptr->stop_event);
				    else
					dev_ptr->queue_state = HALTED;
				    if (dev_ptr->ioctl_wakeup == TRUE)
				    {
					dev_ptr->ioctl_errno = PSC_NO_ERR;
					dev_ptr->ioctl_wakeup = FALSE;
					e_wakeup(&dev_ptr->ioctl_event);
				    }
				    if (dev_ptr->flags & SCSI_BDR)
				    {
					dev_ptr->negotiate_flag = TRUE;
					dev_ptr->restart_in_prog = TRUE;
					adp_str.restart_watchdog.dog.restart =
					    adp_str.ddi.cmd_delay + 1;
					w_start(
					   &(adp_str.restart_watchdog.dog));
				    }
				    psc_deq_active(dev_ptr);
				    start_new_job = TRUE;
				    break;

				case A_neg_select_failed:
				case A_cmd_select_atn_failed:
				    TRACE_1 ("Cneg bad", dev_ptr)
				    /* We got beaten to the bus, probably by */
				    /* a cmd completion.  Put the script onto*/
				    /* waiting to execute queue and go to the*/
				    /* WAIT SCRIPT. This isn't an error but a*/
				    /* change of status.                     */

				    DEBUG_0 ("psc_intr: select_atn_failed \n")
				    w_stop(&dev_ptr->dev_watchdog.dog);
				    if (psc_write_reg(SXFER, SXFER_SIZE,
					 SYNC_VAL) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    psc_deq_active(dev_ptr);
				    /* place this dev on front of WAIT Q */
				    if (adp_str.DEVICE_WAITING_head == NULL)
				    {
					dev_ptr->DEVICE_WAITING_fwd = NULL;
					adp_str.DEVICE_WAITING_head = dev_ptr;
					adp_str.DEVICE_WAITING_tail = dev_ptr;
				    }
				    else	/* list not empty */
				    {
					dev_ptr->DEVICE_WAITING_fwd =
					    adp_str.DEVICE_WAITING_head;
					adp_str.DEVICE_WAITING_head = dev_ptr;
				    }
				    /* if no commands are outstanding, the */
				    /* scsi bus is hung. Attempt a scsi bus */
				    /* reset to clear up this condition.   */
				    if (adp_str.DEVICE_ACTIVE_head == NULL)
				    {
					psc_command_reset_scsi_bus();
				    }
				    else
				    {
					/* restart IOWAIT routine */
					if (psc_write_reg(DSP, DSP_SIZE,
					    word_reverse(adp_str.SCRIPTS[0].
						      dma_ptr)) == REG_FAIL)
					    psc_cleanup_reset(REG_ERR);
				    }
				    break;

				case A_abort_select_failed:
				    /* someone beat us to the bus. */
				    /* So we have to */
				    /* remember to reissue the abort again */
				    /* when the bus frees up.  */

				case A_bdr_select_failed:
				    DEBUG_0 ("psc_intr: ABRTBDR_SEL_FAILED \n")
				    TRACE_1 ("Aneg bad", dev_ptr)
				    if (psc_write_reg(SXFER, SXFER_SIZE,
					 SYNC_VAL) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    /* if there's an active command this */
				    /* abort/bdr was issued for, continue */
				    /* to time the active command.       */
				    if (dev_ptr->flags & RETRY_ERROR)
				    {
					w_stop(&dev_ptr->dev_watchdog.dog);
					dev_ptr->dev_watchdog.dog.restart =
					    LONGWAIT;
					dev_ptr->flags &= ~RETRY_ERROR;
					dev_ptr->retry_count = 1;
					w_start(&dev_ptr->dev_watchdog.dog);
				    }
				    if (dev_ptr->active == NULL)
				    {	/* no active command */
					psc_deq_active(dev_ptr);
				    }
				    else
				    {	/* restore the iowait pointer for  */
					/* this device.                    */
					psc_restore_iowait_jump(
					      adp_str.SCRIPTS[0].script_ptr,
					 dev_ptr, dev_ptr->script_dma_addr);
				    }
				    /* put device on front of abort queue   */
				    if (adp_str.ABORT_BDR_head == NULL)
				    {
					dev_ptr->ABORT_BDR_fwd = NULL;
					dev_ptr->ABORT_BDR_bkwd = NULL;
					adp_str.ABORT_BDR_head = dev_ptr;
					adp_str.ABORT_BDR_tail = dev_ptr;
				    }
				    else	/* list not empty */
				    {
					adp_str.ABORT_BDR_head->ABORT_BDR_bkwd
					  = dev_ptr;
					dev_ptr->ABORT_BDR_bkwd = NULL;
					dev_ptr->ABORT_BDR_fwd =
					   adp_str.ABORT_BDR_head;
					adp_str.ABORT_BDR_head = dev_ptr;
				    }
				    /* if no commands are outstanding, the */
				    /* scsi bus is hung. Attempt a scsi bus */
				    /* reset to clear up this condition.   */
				    if ((adp_str.DEVICE_ACTIVE_head == NULL) ||
					(dev_ptr->retry_count >=
					 PSC_RETRY_COUNT))
				    {
					psc_command_reset_scsi_bus();
				    }
				    else
				    {
					/* restart IOWAIT routine */
					if (psc_write_reg(DSP, DSP_SIZE,
					    word_reverse(adp_str.SCRIPTS[0].
						      dma_ptr)) == REG_FAIL)
					    psc_cleanup_reset(REG_ERR);
				    }
				    break;

				    /* for any phase or unknown message */
				    /* error do an abort */
				case A_phase_error:
				    TRACE_1 ("Aphaserr", dev_ptr)
				case A_err_not_ext_msg:
				    TRACE_1 ("not ext ", dev_ptr)
				case A_unknown_msg:
				    DEBUG_0 ("psc_intr: MSG_ERR Phas_error\n")
				    TRACE_1 ("unk msg ", dev_ptr)
				    w_stop(&dev_ptr->dev_watchdog.dog);
				    bp = dev_ptr->active;
				    if ((bp != NULL) &&
					(!(bp->bufstruct.b_flags & B_ERROR)))
				    {	/* there's no previous error so log */
					/* this one.                        */
					switch (save_interrupt)
					{	/* for command logging */
					case A_phase_error:
					    psc_logerr(ERRID_SCSI_ERR2,
							PHASE_ERROR, 90, 0,
							dev_ptr, TRUE);
					    break;
					case A_err_not_ext_msg:
					    psc_logerr(ERRID_SCSI_ERR2,
							ERROR_NOT_EXTENT, 90,
							0, dev_ptr, TRUE);
					    break;
					case A_unknown_msg:
					default:
					    psc_logerr(ERRID_SCSI_ERR1,
							UNKNOWN_MESSAGE, 90,
							0, dev_ptr, TRUE);
					    break;
					}
				    }
				    if (psc_write_reg(SXFER, SXFER_SIZE,
					 SYNC_VAL) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }

				    if ((dev_ptr->cmd_activity_state
					 == ABORT_IN_PROGRESS) ||
					(dev_ptr->cmd_activity_state
					 == BDR_IN_PROGRESS))
				    {
					/* remove device from active queue */
					/* if there is a command tagged to
					   this */
					if (bp != NULL)
					{
					    if (!(bp->bufstruct.b_flags &
						  B_ERROR))
					    {	/* no previous error */
						bp->bufstruct.b_resid =
						     bp->bufstruct.b_bcount;
						bp->status_validity =
						     SC_ADAPTER_ERROR;
						bp->general_card_status =
						     SC_SCSI_BUS_FAULT;
						bp->bufstruct.b_flags |=
						     B_ERROR;
						bp->bufstruct.b_error = EIO;
					    }
					}
					psc_command_reset_scsi_bus();
					break;
				    }
				    /* update wait script */
				    psc_reset_iowait_jump(
					       adp_str.SCRIPTS[0].script_ptr,
					       dev_index);
				    /* issue an abort script call */
				    if (psc_issue_abort_script(adp_str.SCRIPTS
				       [dev_ptr->cmd_script_ptr].script_ptr,
					  dev_ptr->script_dma_addr, dev_ptr,
						     dev_index) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    dev_ptr->cmd_activity_state =
							  ABORT_IN_PROGRESS;
				    dev_ptr->dev_watchdog.dog.restart =
							  LONGWAIT;
				    dev_ptr->flags |= SCSI_ABORT;
				    dev_ptr->flags &= ~RETRY_ERROR;
				    dev_ptr->retry_count = 1;
				    w_start(&dev_ptr->dev_watchdog.dog);
				    /* log error */
				    if (bp != NULL)
				    {
					/* this error type requires an abort */
					/* to be sent to the offending device*/

					bp->bufstruct.b_resid =
						     bp->bufstruct.b_bcount;
					bp->status_validity = SC_ADAPTER_ERROR;
					bp->general_card_status =
						     SC_SCSI_BUS_FAULT;
					bp->bufstruct.b_flags |= B_ERROR;
					bp->bufstruct.b_error = EIO;
				    }
				    if (dev_ptr->ioctl_wakeup == TRUE)
				    {
					dev_ptr->ioctl_errno = EIO;
					dev_ptr->ioctl_wakeup = FALSE;
					e_wakeup(&dev_ptr->ioctl_event);
				    }
				    if (issue_abrt_bdr)
					psc_deq_abort_bdr(dev_ptr);
				    break;

				case A_bdr_msg_error:
				case A_abort_msg_error:
				    DEBUG_0 ("psc_intr:A_abort_err or A_bdr\n")
				    TRACE_1 ("Amsg err", dev_ptr)
				    w_stop(&dev_ptr->dev_watchdog.dog);
				    if (psc_write_reg(SXFER, SXFER_SIZE,
					 SYNC_VAL) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }

				    /* update wait script  */
				    psc_reset_iowait_jump(
					     adp_str.SCRIPTS[0].script_ptr,
					     dev_index);
				    /* log the appropriate error */
				    if (dev_ptr->flags & SCSI_ABORT)
					psc_logerr(ERRID_SCSI_ERR1,
						    ABORT_FAILED, 95, 0,
						    dev_ptr, TRUE);
				    else
					psc_logerr(ERRID_SCSI_ERR1, BDR_FAILED,
						   95, 0, dev_ptr, TRUE);

				    /* if abort failed issue a BDR */
				    bp = dev_ptr->active;
				    if ((dev_ptr->flags & SCSI_ABORT) &&
					(bp != NULL))
				    {
					dev_ptr->flags |= SCSI_BDR;
					/* issue a bdr script call */
					if (psc_issue_bdr_script(
					     adp_str.SCRIPTS
					      [dev_ptr->cmd_script_ptr].
					      script_ptr,
					     dev_ptr->script_dma_addr,
					     dev_index) == REG_FAIL)
					{
					    psc_cleanup_reset(REG_ERR);
					    break;
					}
					dev_ptr->cmd_activity_state =
							    BDR_IN_PROGRESS;
					/* the timer is restarted from it's */
					/* last point to eventually time    */
					/* out the cmd if all else fails.  */
					dev_ptr->dev_watchdog.dog.restart =
					    LONGWAIT;
					dev_ptr->flags &= ~RETRY_ERROR;
					dev_ptr->retry_count = 1;
					w_start(&dev_ptr->dev_watchdog.dog);
					if (!(bp->bufstruct.b_flags & B_ERROR))
					{
					    bp->bufstruct.b_resid =
						     bp->bufstruct.b_bcount;
					    bp->status_validity =
						     SC_ADAPTER_ERROR;
					    bp->general_card_status =
						     SC_SCSI_BUS_FAULT;
					    bp->scsi_status = SC_GOOD_STATUS;
					    bp->bufstruct.b_flags |= B_ERROR;
					    bp->bufstruct.b_error = EIO;
					}
					break;
				    }
				    if (bp != NULL)
				    {
					bp->bufstruct.b_resid =
						     bp->bufstruct.b_bcount;
					bp->status_validity =
						     SC_ADAPTER_ERROR;
					bp->general_card_status =
						     SC_SCSI_BUS_FAULT;
					bp->scsi_status = SC_GOOD_STATUS;
					bp->bufstruct.b_flags |= B_ERROR;
					bp->bufstruct.b_error = EIO;
					if ((rc = psc_free_resources(dev_ptr,
					      TRUE)) != PSC_NO_ERR)
					{
					    if (psc_fail_free_resources(
						 dev_ptr, bp, rc))
						break;
					}
					else
					    iodone((struct sc_buf *) bp);
				    }
				    dev_ptr->flags = RETRY_ERROR;
				    dev_ptr->active = NULL;
				    psc_deq_active(dev_ptr);
				    if (dev_ptr->head_pend != NULL)
					psc_fail_cmd(dev_ptr, EIO, 1);
				    if (dev_ptr->queue_state == STOPPING)
					e_wakeup(&dev_ptr->stop_event);
				    else
					dev_ptr->queue_state = HALTED;
				    if (dev_ptr->ioctl_wakeup == TRUE)
				    {
					dev_ptr->ioctl_errno = EIO;
					dev_ptr->ioctl_wakeup = FALSE;
					e_wakeup(&dev_ptr->ioctl_event);
				    }
				    if (issue_abrt_bdr)
					psc_deq_abort_bdr(dev_ptr);
				    start_new_job = TRUE;
				    break;

				case A_unexpected_status:
				    TRACE_1 ("unexp st", dev_ptr)
				    /* we received status back almost im- */
				    /* mediately. It is probably a hw or  */
				    /* protocol error.  Fail the job and  */
				    /* start a new one.                   */
				    DEBUG_0 ("psc_intr: A_unexpect_status \n")
				    w_stop(&dev_ptr->dev_watchdog.dog);
				    if (psc_write_reg(SXFER, SXFER_SIZE,
						       SYNC_VAL) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    psc_logerr(ERRID_SCSI_ERR2,
						UNEXPECTED_STATUS, 100, 0,
						dev_ptr, TRUE);
				    if ((bp = dev_ptr->active) != NULL)
				    {
					bp->scsi_status = GET_CMD_STATUS_BYTE(
							   adp_str.SCRIPTS
							    [dev_ptr->
							    cmd_script_ptr].
							    script_ptr);
					bp->bufstruct.b_resid =
						     bp->bufstruct.b_bcount;
					bp->status_validity =
						     SC_ADAPTER_ERROR;
					bp->general_card_status =
						     SC_SCSI_BUS_FAULT;
					bp->bufstruct.b_flags |= B_ERROR;
					bp->bufstruct.b_error = EIO;
					if ((rc = psc_free_resources(dev_ptr,
						   TRUE)) != PSC_NO_ERR)
					{
					    if (psc_fail_free_resources(
						 dev_ptr, bp, rc))
						break;
					}
					else
					    iodone((struct sc_buf *) bp);
				    }
				    dev_ptr->flags = RETRY_ERROR;
				    dev_ptr->active = NULL;
				    psc_deq_active(dev_ptr);
				    if (dev_ptr->head_pend != NULL)
					psc_fail_cmd(dev_ptr, EIO, 1);
				    if (dev_ptr->queue_state == STOPPING)
					e_wakeup(&dev_ptr->stop_event);
				    else
					dev_ptr->queue_state = HALTED;
				    if (dev_ptr->ioctl_wakeup == TRUE)
				    {
					dev_ptr->ioctl_errno = EIO;
					dev_ptr->ioctl_wakeup = FALSE;
					e_wakeup(&dev_ptr->ioctl_event);
				    }
				    if (issue_abrt_bdr)
					psc_deq_abort_bdr(dev_ptr);
				    start_new_job = TRUE;
				    break;

				case A_unknown_reselect_id:
				    TRACE_1 ("unkn id ", dev_ptr)
				    /* a device reselected us, but we can't*/
				    /* figure out who it really is.        */
				            psc_reselect_router();
				    break;

				case A_uninitialized_reselect:
				    /* an unconfigured device is trying to */
				    /* select us (spurious interrupt) or a */
				    /* device which was aborted or BDR'ed  */
				    /* is trying to reselect us and is to  */
				    /* trying finish off a cmd.            */
				    DEBUG_0 ("psc_intr: uninit_reselect.\n")
				    TRACE_1 ("bad resl", dev_ptr)
				    if (psc_write_reg(SXFER, SXFER_SIZE,
						       SYNC_VAL) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    psc_logerr(ERRID_SCSI_ERR2,
					       UNINITIALIZED_RESELECT, 105,
					       0, NULL, TRUE);
				    psc_command_reset_scsi_bus();
				    break;

				default:
				    /* unknown interrupt flag.  This should*/
				    /* not happen. If it does, we have a   */
				    /* fatal error which we must back out  */
				    /* of.                                 */
				    DEBUG_0 ("psc_intr: SCRIPT err dflt \n")
				    TRACE_1 ("default ", dev_ptr)
				    if (psc_write_reg(SXFER, SXFER_SIZE,
						      SYNC_VAL) == REG_FAIL)
				    {
					psc_cleanup_reset(REG_ERR);
					break;
				    }
				    psc_logerr(ERRID_SCSI_ERR2,
					       ADAPTER_INT_ERROR, 110, 0,
					       NULL, TRUE);
				    start_new_job = TRUE;
				    break;

				}	/* end script switch */
			    }	/* end script interrupt (SIR) */
			    /******************************************/
			    /*      ABORT CAUSED THIS INTERRUPT       */
			    /******************************************/
			    else
				if (save_dsr & DABRT)
				{
				    DEBUG_0 ("psc_intr: Abort caused int\n")
				    TRACE_1 ("abrt int", 0)
				    /* interrupt cleared up above */
				    rc = psc_read_reg(ISTAT, ISTAT_SIZE);
				    /* A read of the istat register verifies*/
				    /* that the scsi chip is currently not  */
				    /* busy on the scsi bus, so we can issue*/
				    /* the new command. */
				    if (!(rc & CONNECTED))
				    {
					/* Read the DSP register to make  */
					/* sure that the abort took place */
					/* while executing the iowait     */
					/* script. If this is the case, or*/
					/* a scsi abort/BDR command is    */
					/* waiting to be issued, go ahead */
					/* and issue the command.         */
					/* Else, restart the script.      */
					dma_stat = word_reverse(
					     psc_read_reg(DSP, DSP_SIZE));
#ifdef PSC_TRACE
					rc = psc_read_reg(DCMD, DCMD_SIZE);
					TRACE_2("DSP DCMD", dma_stat, rc)
#endif
					if (((dma_stat >=
					     adp_str.SCRIPTS[0].dma_ptr) &&
					   (dma_stat <
					    (adp_str.SCRIPTS[0].dma_ptr +
					     Ent_scripts_entry_point))) ||
					   (adp_str.ABORT_BDR_head != NULL))
                                        {
					    if (psc_write_reg(SXFER,
					         SXFER_SIZE, SYNC_VAL)
						 == REG_FAIL)
					    {
						psc_cleanup_reset(REG_ERR);
					    }
					    else
					    {
						dev_ptr = NULL;
						start_new_job = TRUE;
					    }
					}
					else /* restart script */
					{
					    psc_patch_iowait_int_off();
                                            dma_stat -= 8;
					    rc = psc_write_reg(DSP, DSP_SIZE,
						   word_reverse(dma_stat));
					    if (rc == REG_FAIL)
					    {
						psc_cleanup_reset(REG_ERR);
					    }
					}
				    }
				    else
				    {
					(void) psc_read_reg(DSTAT, DSTAT_SIZE);
					/* restart iowait script */
					psc_patch_iowait_int_off();
					dma_stat = word_reverse(
					     psc_read_reg(DSP, DSP_SIZE));
					dma_stat -= 8;
#ifdef PSC_TRACE
					rc = psc_read_reg(DCMD, DCMD_SIZE);
					TRACE_2("DSP DCMD", dma_stat, rc)
#endif
					rc = psc_write_reg(DSP, DSP_SIZE,
						 word_reverse(dma_stat));
					if (rc == REG_FAIL)
					{
					    psc_cleanup_reset(REG_ERR);
					}
				    }
				}
			    /**************************************/
			    /*      WATCH DOG TIMER INTERRUPT     */
			    /**************************************/
				else
				    if (save_dsr & WTD)
				    {
					DEBUG_0 ("psc_intr: Timeout int\n")
					TRACE_1 ("watch tm", 0)
					if (psc_write_reg(SXFER, SXFER_SIZE,
					     SYNC_VAL) == REG_FAIL)
					    psc_cleanup_reset(REG_ERR);
					psc_logerr(ERRID_SCSI_ERR2,
						   ADAPTER_INT_ERROR, 115, 0,
						   NULL, TRUE);
					dev_ptr = NULL;
					start_new_job = TRUE;
				    }
			    /**************************************/
			    /*    ILLEGAL INSTRUCTION DETECTED    */
			    /**************************************/
				    else
					if (save_dsr & OPC)
					{
					    DEBUG_0 ("psc_intr: Ig Inst int\n")
					    TRACE_1 ("igl inst", 0)
					/* A check is made here to read   */
					/* the command that is currently  */
					/* executing on the scsi chip. If */
					/* the command is a "wait for dis-*/
					/* conect command" (0x48), then a */
					/* reselect occured during this   */
					/* command.  Due to a chip fault, */
					/* this is reflected as an illegal*/
					/* instruction interrupt.  In this*/
					/* case, the DSP holds the address*/
					/* of the interrupt instruction   */
					/* that should have occurred. So  */
					/* the script is restarted to     */
					/* cause this interrupt to occur. */
					/* Otherwise, this interrupt is an*/
					/* error and is logged as such.   */
					    rc = psc_read_reg(DCMD,DCMD_SIZE);
					    if (rc == REG_FAIL)
						psc_cleanup_reset(REG_ERR);
					    else
					    {
						if (rc == 0x48)
						{
						    rc = psc_read_reg(DSP,
							       DSP_SIZE);
						    if (rc == REG_FAIL)
						    {
							psc_cleanup_reset
							  (REG_ERR);
						    }
						    else
						    {
							rc = psc_write_reg
							 (DSP, DSP_SIZE, rc);
							if (rc == REG_FAIL)
							{
							    psc_cleanup_reset
							      (REG_ERR);
							}
						    }
						}
						else
						{
						if (psc_write_reg(SXFER,
						    SXFER_SIZE, SYNC_VAL)
						    == REG_FAIL)
						    psc_cleanup_reset(REG_ERR);
						else
						{
						    dev_ptr = NULL;
						    start_new_job = TRUE;
						}
						psc_logerr(ERRID_SCSI_ERR2,
							 ADAPTER_INT_ERROR,
							 120, 0, NULL, TRUE);
					        }
					    }
					}
			    /**************************************/
			    /*       SINGLE STEP INTERRUPT        */
			    /**************************************/
					else
					    if (save_dsr & SSI)
					    {
						DEBUG_0 ("psc_intr S Sp Int\n")
						TRACE_1 ("sngl stp", 0)
						if (psc_write_reg(SXFER,
						     SXFER_SIZE, SYNC_VAL) ==
						     REG_FAIL)
						    psc_cleanup_reset(REG_ERR);
						psc_logerr(ERRID_SCSI_ERR2,
							   ADAPTER_INT_ERROR,
							   125, 0, NULL, TRUE);
						dev_ptr = NULL;
						start_new_job = TRUE;
					    }
			}	/* if ISR & DIP */
			/*******************************************/
			/*        THIS IS A SCSI INTERRUPT         */
			/*******************************************/
			/* in case of DIP and SIP will fall here */
			else
			    if (save_isr & SIP)
			    {
				/* read scsi status register 0 for cause int */
				rc = psc_read_reg(SSTAT0, SSTAT0_SIZE);
				save_interrupt = (rc & SIEN_MASK);
				DEBUG_1 ("CHIP INTR is 0x%x\n",
					  save_interrupt);
				if (rc == REG_FAIL)
				{
				    /* clean up errors and clear interrupts */
				    psc_cleanup_reset(REG_ERR);
				}
				else
				    if (psc_write_reg(SXFER, SXFER_SIZE,
					  SYNC_VAL) == REG_FAIL)
				    {
					/* clean up errors and clear ints */
					psc_cleanup_reset(REG_ERR);
				    }
				    else
				    {
					/* if a chip abort was set but    */
					/* another interrupt won the race,*/
					/* clear the stacked interrupt for*/
					/* which the abort has taken place*/
					if (chip_abort_occurred)
					    (void) psc_read_reg(DSTAT,
					      DSTAT_SIZE);
					dev_ptr = NULL;
					check_for_disconnect = TRUE;
				scsi_int:
					switch (save_interrupt)
					{

					case PHASE_MIS:
					    DEBUG_0 ("psc_intr: PHASE_MIS\n")
					    if ((dev_ptr = psc_find_devinfo(0))
						 != NULL)
					    {
						TRACE_1 ("phase ms", dev_ptr)
						rc = psc_update_dataptr(
							 adp_str.SCRIPTS
							  [dev_ptr->
							  cmd_script_ptr].
							 script_ptr,
							 dev_ptr);
						if (rc == REG_FAIL)
						{
						    psc_cleanup_reset(REG_ERR);
						    break;
						}
						/* clear DMA and SCSI FIFOs */
						if (psc_write_reg(DFIFO,
						     DFIFO_SIZE, 0x40) ==
						     REG_FAIL)
						{
						    psc_cleanup_reset(REG_ERR);
						    break;
						}
						ISSUE_MAIN_AFTER_NEG(
						   dev_ptr->script_dma_addr);
					    }
					    else
					    {
						TRACE_1 ("phase ms", 0)
						DEBUG_0 ("psc_intr: PH_MIS /n")
						psc_cleanup_reset(PHASE_ERR);
						psc_logerr(ERRID_SCSI_ERR2,
							   PHASE_ERROR, 130, 0,
							   NULL, TRUE);
					    }
					    break;

					case SCSI_SEL_FAIL:
					    DEBUG_1 ("psc_intr: sel f 0x%x\n",
						       save_interrupt);
					    if ((dev_ptr = psc_find_devinfo(0))
						 != NULL)
					    {
						TRACE_1 ("sel fail", dev_ptr)
						if (check_for_disconnect)
						    start_new_job =
							psc_scsi_parity_error(
							    dev_ptr,
							    save_interrupt,
							    TRUE);
						else
						    start_new_job =
							psc_scsi_parity_error(
							    dev_ptr,
							    save_interrupt,
							    FALSE);
					    }
					    else
					    {
						DEBUG_0 ("psc_intr:UNEXP_DC\n")
						TRACE_1 ("sel fail", 0)
						/* clean active queue */
						psc_cleanup_reset(DEVICE_ERR);
						psc_logerr(ERRID_SCSI_ERR2,
						  UNEXPECTED_SELECT_TIMEOUT,
							 135, 0, NULL, TRUE);
					    }
					    break;

					case SCSI_GROSS_ERR:
					    DEBUG_1 ("psc_intr:gross 0x%x\n",
						             save_interrupt);
					case SCSI_PARITY_ERR:
					    DEBUG_1 ("psc_intr:parity 0x%x\n",
						             save_interrupt);
					    if ((dev_ptr = psc_find_devinfo(0))
						 != NULL)
					    {
						TRACE_1 ("par err ", dev_ptr)
						start_new_job =
						    psc_scsi_parity_error(
							dev_ptr,
							save_interrupt, TRUE);
					    }
					    else
					    {
						DEBUG_0 ("psc_intr: no dev\n")
						TRACE_1 ("par err ", 0)
						/* clean active queue */
						if (save_interrupt ==
						     SCSI_GROSS_ERR)
						{
						    psc_cleanup_reset(
						      DEVICE_ERR);
						    psc_logerr(ERRID_SCSI_ERR2,
							       GROSS_ERROR,
							       140, 0, NULL,
							       TRUE);
						}
						else	/* parity error */
						{
						    psc_cleanup_reset(
						      HOST_ERR);
						    psc_logerr(ERRID_SCSI_ERR2,
							BAD_PARITY_DETECTED,
							140, 0, NULL, TRUE);
						}
					    }
					    break;

					case SCSI_UNEXP_DISC:
					    DEBUG_1 ("psc_intr: unexp 0x%x\n",
						             save_interrupt);
					    if ((dev_ptr = psc_find_devinfo(0))
						  != NULL)
					    {
						TRACE_1 ("uex dis ", dev_ptr)
						start_new_job =
						   psc_scsi_parity_error(
						     dev_ptr, save_interrupt,
						     FALSE);
					    }
					    else
					    {
						DEBUG_0 ("psc_intr:  no dev\n")
						TRACE_1 ("uex dis ", dev_ptr)
						/* clean active queue */
						psc_cleanup_reset(DISC_ERR);
						psc_logerr(ERRID_SCSI_ERR2,
							UNEXPECTED_BUS_FREE,
							140, 0, NULL, TRUE);
					    }
					    break;

					case SCSI_RST:
					    DEBUG_0 ("psc_intr: int reset.\n")
					    TRACE_1 ("RESET!!!", 0)
					    w_stop(
						&(adp_str.reset_watchdog.dog));
					    /* clean active queue */
					    psc_cleanup_reset_error();
					    psc_logerr(ERRID_SCSI_ERR2,
						       SCSI_BUS_RESET,
						       145, 0, NULL, TRUE);
					    break;

					case SCSI_COMP:
					    /* if we get here something is
					       wrong */
					    /* int. should be disabled */
					    DEBUG_0 ("psc_intr: SCSI_COMP.\n")
					    TRACE_1 ("scsi cmp", 0)
						    psc_logerr(ERRID_SCSI_ERR2,
							ADAPTER_INT_ERROR,
						        150, 0, NULL, TRUE);
					    break;

					case SCSI_SEL:
					    /* if we get here something is */
					    /* wrong int should be disabled*/
					    DEBUG_0 ("psc_intr: SCSI_SEL\n")
					    TRACE_1 ("scsi sel", 0)
					    psc_logerr(ERRID_SCSI_ERR2,
						       ADAPTER_INT_ERROR,
						       155, 0, NULL, TRUE);
					    break;

					default:
					    DEBUG_1 ("psc_intr:int dflt0x%x\n",
						             save_interrupt)
					    TRACE_1 ("scsi dft", 0)
					    /* if here we probably have 2  */
					    /* interrupt bits set so mask  */
					    /* them and carry on           */
					    if (save_interrupt & SCSI_RST)
					    {
						/* on reset throw away the */
						/* others and execute */
						save_interrupt = SCSI_RST;
						goto scsi_int;
					    }
					    if (save_interrupt &
						 SCSI_UNEXP_DISC)
					    {
						check_for_disconnect = FALSE;
						save_interrupt =
						    (save_interrupt &
						     ~SCSI_UNEXP_DISC);
						goto scsi_int;
					    }
					    if (save_interrupt &
						 SCSI_PARITY_ERR)
					    {   /* insure that a        */
						/* SCSI_PARITY_ERR gets */
						/* processed            */
						save_interrupt =
						   SCSI_PARITY_ERR;
						goto scsi_int;
					    }
					    if (save_interrupt &
						 SCSI_GROSS_ERR)
					    {   /* insure that a        */
						/* SCSI_GROSS_ERR gets  */
						/* processed the same as*/
						/* a SCSI_PARITY_ERROR  */
						save_interrupt =
						   SCSI_PARITY_ERR;
						goto scsi_int;
					    }
					    psc_logerr(ERRID_SCSI_ERR2,
						       UNKNOWN_ADAPTER_ERROR,
						       160, 0, NULL, TRUE);
					    start_new_job = TRUE;
					    break;
					}	/* end of scsi interrupt
						   switch */
				    }
			    }	/* end else if end of SCSI interrupt (SIP) */
			    /* if DIP or SIP bits of ISTAT are not set  */
			    /* this implies that this error caused by   */
			    /* spurious interrupt log soft error        */
			    else
			    {
				DEBUG_0 ("psc_intr: Spurious error \n")
				TRACE_1 ("spurious", 0)
				psc_logerr(ERRID_SCSI_ERR2, ADAPTER_INT_ERROR,
					   170, 0, NULL, TRUE);
				(void) psc_write_reg(SXFER, SXFER_SIZE,
						     SYNC_VAL);
			    }
			/*********************************************/
			if (start_new_job)
			{
			    command_issued_flag = psc_issue_cmd();
			    psc_check_wait_queue(dev_ptr,command_issued_flag);
			}
		    }
		    else
		    {	/* read of the ISTAT failed */
			TRACE_1 ("bad Rist", 0)
			TRACE_1 ("out intr", 0)
			psc_cleanup_reset(REG_ERR);
		    }
		}
		else
		{	/* DMA error */
		    DEBUG_0 ("Leaving psc_intr: DMA FAILED \n")
		    TRACE_1 ("dma err ", 0)
		    TRACE_1 ("out intr", 0)
		    psc_logerr(ERRID_SCSI_ERR2, DMA_ERROR, 175,
			       (uint) rc, NULL, TRUE);
		    (void) psc_read_reg(DSTAT, DSTAT_SIZE);
		    (void) psc_read_reg(SSTAT0, SSTAT0_SIZE);
		    psc_cleanup_reset(DMA_ERR);
		}
	    }
	    else
	    {	/* an interrupt was received with no command active */
		DEBUG_1 ("Leaving psc_intr: rc=%d\n", rc);
		TRACE_1 ("no cmd  ", 0)
		TRACE_1 ("out intr", 0)
		/* If an epow has occurred, a reset may have been sent */
		/* by the epow handler.  If there are commands this    */
		/* will be handled by the scsi reset cleanup.  If not, */
		/* it will hit this code and have the watchdog reset.  */
		save_isr = psc_read_reg(ISTAT, ISTAT_SIZE);
		(void) psc_write_reg(ISTAT, ISTAT_SIZE, 0x00);
		if (save_isr & DIP)
		{
		    (void) psc_read_reg(DSTAT, DSTAT_SIZE);
		}
		if (save_isr & SIP)
		{
		    save_stat = psc_read_reg(SSTAT0, SSTAT0_SIZE);
		    if (save_stat & SCSI_RST)
		    {
			w_stop(&(adp_str.reset_watchdog.dog));
			/* clean active queue */
			psc_cleanup_reset_error();
			psc_logerr(ERRID_SCSI_ERR2, SCSI_BUS_RESET,
				   177, 0, NULL, TRUE);
		    }
		}
	    }
	}
	else
	{	/* this is not our interrupt */
	    DEBUG_1 ("Leaving psc_intr: not our int. SCSIRS 0x%x\n", rc);
	    TRACE_1 ("not ours", 0)
	    TRACE_1 ("out intr", 0)
	    DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_INTR, 0, 0);
	    return (INTR_FAIL);
	}
    }
    else	/* register fail or channel check in SCSIRS */
    {
	DEBUG_1 ("Leaving psc_intr: SCSIRS error 0x%x\n", rc);
	TRACE_1 ("bad SCIR", 0)
	TRACE_1 ("out intr", 0)
	if (rc == REG_FAIL)
	    psc_logerr(ERRID_SCSI_ERR2, PIO_RD_DATA_ERR, 180,
		       SCSIRS, NULL, TRUE);
	else
	    psc_logerr(ERRID_SCSI_ERR2, CHANNEL_CHK_ERROR, 180,
		       SCSIRS, NULL, TRUE);
	(void) psc_write_reg(SCSIRS, SCSIRS_SIZE, (rc | 0x1000000));
	/* make sure the adapter timers are stopped */
	w_stop(&(adp_str.adap_watchdog.dog));
	psc_cleanup_reset(REG_ERR);
    }
    DEBUG_0 ("Leaving psc_intr\n")
    TRACE_1 ("out intr", 0);
    i_reset(&(adp_str.intr_struct));
    DDHKWD1(HKWD_DD_SCSIDD, DD_EXIT_INTR, 0, 0);
    return (INTR_SUCC);
}  /* end of psc_intr      */

/************************************************************************/
/*                                                                      */
/* NAME:        psc_issue_cmd                                           */
/*                                                                      */
/* FUNCTION:    issues a waiting command, or ABORT/BDR command          */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      Called by the interrupt handler                                 */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*                                                                      */
/* INPUTS:                                                              */
/*      none                                                            */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      int return value                                                */
/*      TRUE - command issued all is well                               */
/*      FALSE - command not issued all is not well                      */
/*                                                                      */
/* ERROR DESCRIPTION:  The following errno values may be returned:      */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      iodone                                                          */
/************************************************************************/
int
psc_issue_cmd()
{
    struct dev_info *dev_ptr;
    struct sc_buf *bp;
    int     rc;	/* return codes */
    int     dev_index;	/* index into device hash table */
    uchar   chk_disconnect;

    DEBUG_0 ("Entering psc_issue_cmd routine\n")
    TRACE_1 ("in issue", 0)

    /* if abort/bdr queue not empty then go to work  */
    dev_ptr = adp_str.ABORT_BDR_head;
    /* loop while Q not empty */
    while ((dev_ptr != NULL) && (adp_str.epow_state != EPOW_PENDING))
    {
	if (dev_ptr->flags & SCSI_ABORT)
	{
	    if (dev_ptr->active != NULL)
	    {
		/* the timer is started only for aborts  */
		/* that have not previously been issued. */
		/* Otherwise, the current timer is left  */
		/* running.                              */
		if (dev_ptr->flags & RETRY_ERROR)
		{
		    w_stop(&dev_ptr->dev_watchdog.dog);
		    dev_ptr->dev_watchdog.dog.restart = LONGWAIT;
		    dev_ptr->flags &= ~RETRY_ERROR;
		    dev_ptr->retry_count = 1;
		    w_start(&dev_ptr->dev_watchdog.dog);
		}
		else
		{
		    dev_ptr->retry_count++;
		}
		dev_index = INDEX(dev_ptr->scsi_id, dev_ptr->lun_id);
		/* update wait script */
		psc_reset_iowait_jump(adp_str.SCRIPTS[0].script_ptr,
				      dev_index);
		if (psc_issue_abort_script(adp_str.SCRIPTS
				       [dev_ptr->cmd_script_ptr].script_ptr,
					dev_ptr->script_dma_addr, dev_ptr,
				       dev_index) == REG_FAIL)
		{
		    psc_cleanup_reset(REG_ERR);
		    return (TRUE);
		}
		TRACE_1 ("issue ab", dev_ptr)
		psc_deq_abort_bdr(dev_ptr);
		dev_ptr->cmd_activity_state = ABORT_IN_PROGRESS;
		w_start(&dev_ptr->dev_watchdog.dog);
		DEBUG_0 ("psc_issue_cmd: leaving TRUE\n")
		TRACE_1 ("out issu", 0)
		return (TRUE);  /* cmds issued */
	    }
	    else  /* else don't issue abort one is already in progress */
	    {
		psc_deq_abort_bdr(dev_ptr);
		if (dev_ptr->ioctl_wakeup == TRUE)
		{
		    dev_ptr->cmd_activity_state = 0;
		    dev_ptr->ioctl_errno = PSC_NO_ERR;
		    dev_ptr->ioctl_wakeup = FALSE;
		    e_wakeup(&dev_ptr->ioctl_event);
		}
	    }
	}
	else  /* a BDR */
	{
	    dev_index = INDEX(dev_ptr->scsi_id, dev_ptr->lun_id);
	    /* update wait script */
	    psc_reset_iowait_jump(adp_str.SCRIPTS[0].script_ptr, dev_index);
	    if (psc_issue_bdr_script(adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].
				     script_ptr, dev_ptr->script_dma_addr,
				     dev_index) == REG_FAIL)
	    {
		psc_cleanup_reset(REG_ERR);
		TRACE_1 ("out issu", 0)
		return (TRUE);
	    }
	    TRACE_1 ("issue bd", dev_ptr)
	    psc_deq_abort_bdr(dev_ptr);
	    /* the timer is started only for aborts  */
	    /* that have not previously been issued. */
	    /* Otherwise, the current timer is left  */
	    /* running.                              */
	    if (dev_ptr->flags & RETRY_ERROR)
	    {
		w_stop(&dev_ptr->dev_watchdog.dog);
		dev_ptr->dev_watchdog.dog.restart = LONGWAIT;
		dev_ptr->flags &= ~RETRY_ERROR;
		dev_ptr->retry_count = 1;
		w_start(&dev_ptr->dev_watchdog.dog);
	    }
	    else
	    {
		dev_ptr->retry_count++;
	    }
	    if (dev_ptr->active == NULL)
	    {
		psc_enq_active(dev_ptr);
	    }
	    dev_ptr->cmd_activity_state = BDR_IN_PROGRESS;
	    /* the timer is restarted from it's last */
	    /* point to eventually time out the cmd  */
	    /* if all else fails.                    */
	    DEBUG_0 ("psc_issue_cmd: leaving TRUE\n")
	    TRACE_1 ("out issu", 0)
	    return (TRUE);      /* cmds issued */
	}
	dev_ptr = adp_str.ABORT_BDR_head;
    }	/* if abrt queue */
    /* is there something on the waiting queue */
    dev_ptr = adp_str.DEVICE_WAITING_head;
    /* loop while Q not empty */
    if ((dev_ptr != NULL) && (adp_str.epow_state != EPOW_PENDING))
    {
	/* if we haven't negotiated */
	bp = dev_ptr->active;
	if (!(dev_ptr->negotiate_flag))
	{
	    if (dev_ptr->async_device)
	    {
		rc = psc_write_reg(SXFER, SXFER_SIZE, 0x00);
		if (rc == REG_FAIL)
		{
		    psc_cleanup_reset(REG_ERR);
		    TRACE_1 ("out issu", 0)
		    return (TRUE);
		}
	    }

	    else
		if ((dev_ptr->agreed_xfer != DEFAULT_MIN_PHASE) ||
		    (dev_ptr->agreed_req_ack < DEFAULT_BYTE_BUF))
		{
		    rc = psc_write_reg(SXFER, SXFER_SIZE,
				       dev_ptr->special_xfer_val);
		    if (rc == REG_FAIL)
		    {
			psc_cleanup_reset(REG_ERR);
			TRACE_1 ("out issu", 0)
			return (TRUE);
		    }
		}
	    chk_disconnect = bp->scsi_command.flags & SC_NODISC;
	    if ((dev_ptr->disconnect_flag ^ chk_disconnect) != 0)
		psc_set_disconnect(dev_ptr, chk_disconnect);
	    if (!(dev_ptr->flags & PREP_MAIN_COMPLETE))
	    {
		psc_prep_main_script(adp_str.SCRIPTS[0].script_ptr,
			    adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].
			    script_ptr, dev_ptr, adp_str.SCRIPTS[0].dma_ptr,
			    dev_ptr->script_dma_addr);
	    }
	    psc_patch_iowait_int_off();
	    ISSUE_MAIN_TO_DEVICE(dev_ptr->script_dma_addr);
	    dev_ptr->cmd_activity_state = CMD_IN_PROGRESS;
	}
	else
	{
	    if (!(dev_ptr->active->scsi_command.flags & SC_ASYNC))
	    {
		/* setup chip to exec the negotiate script */
		psc_patch_iowait_int_off();
		chk_disconnect = bp->scsi_command.flags & SC_NODISC;
		if ((dev_ptr->disconnect_flag ^ chk_disconnect) != 0)
		    psc_set_disconnect(dev_ptr, chk_disconnect);
		ISSUE_NEGOTIATE_TO_DEVICE(dev_ptr->script_dma_addr);
		dev_ptr->cmd_activity_state = NEGOTIATE_IN_PROGRESS;
	    }
	    else
	    {
		/* set async mode */
		rc = psc_write_reg(SXFER, SXFER_SIZE, 0x00);
		if (rc == REG_FAIL)
		{
		    psc_cleanup_reset(REG_ERR);
		    TRACE_1 ("out issu", 0)
		    return (TRUE);
		}
		chk_disconnect = bp->scsi_command.flags & SC_NODISC;
		if ((dev_ptr->disconnect_flag ^ chk_disconnect) != 0)
		    psc_set_disconnect(dev_ptr, chk_disconnect);
		dev_index = INDEX(dev_ptr->scsi_id, dev_ptr->lun_id);
		psc_patch_async_switch_int(adp_str.SCRIPTS
			   [dev_ptr->cmd_script_ptr].script_ptr, dev_index);
		if (!(dev_ptr->flags & PREP_MAIN_COMPLETE))
		{
		    psc_prep_main_script(adp_str.SCRIPTS[0].script_ptr,
			    adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].
			    script_ptr, dev_ptr, adp_str.SCRIPTS[0].dma_ptr,
			    dev_ptr->script_dma_addr);
		}
		psc_patch_iowait_int_off();
		ISSUE_MAIN_TO_DEVICE(dev_ptr->script_dma_addr);
		dev_ptr->cmd_activity_state = CMD_IN_PROGRESS;
	    }
	}
	TRACE_1 ("issue cd", dev_ptr)
	psc_deq_wait(dev_ptr);
	psc_enq_active(dev_ptr);
	dev_ptr->dev_watchdog.dog.restart = (ulong) bp->timeout_value + 1;
	w_start(&dev_ptr->dev_watchdog.dog);
	DEBUG_0 ("Leaving psc_issue_cmd with TRUE\n")
	TRACE_1 ("out issu", 0)
	return (TRUE);  /* command issued */
    }	/* while */
    /* no command to issue.  restart io wait and return             */
    DEBUG_0 ("Leaving psc_issue_cmd with FALSE\n")
    return (FALSE);     /* command not issued */
}  /* psc_issue_cmd */

/************************************************************************/
/*                                                                      */
/* NAME:        psc_check_wait_queue                                    */
/*                                                                      */
/* FUNCTION:    issue a waiting command or Abort/Bdr command            */
/*                                                                      */
/* NOTES:                                                               */
/*  this is called to see if a cmd which is waiting for resources       */
/*  may be satisified                                                   */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*                                                                      */
/* INPUTS:                                                              */
/*      *dev_ptr           - pointer to device structure                */
/*      issue_command_flag - FALSE= not issued                          */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*                                                                      */
/* ERROR DESCRIPTION:  The following errno values may be returned:      */
/*      none                                                            */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      iodone                                                          */
/************************************************************************/
void
psc_check_wait_queue(struct dev_info * dev_ptr, uchar issue_command_flag)
{
    struct dev_info *next_ptr;
    struct sc_buf *bp;
    uchar   command_issued_flag;
    int     rc;
    int     ret_code;

    DEBUG_2 ("Entering psc_check_wait_queue: dev_ptr 0x%x flag=0x%x\n",
	      dev_ptr, issue_command_flag)
    TRACE_2 ("in chkwq", dev_ptr, issue_command_flag)
    ret_code = FALSE;   /* no command issued */
    command_issued_flag = issue_command_flag;
    next_ptr = adp_str.DEVICE_WAITING_FOR_RESOURCES_head;

    while (next_ptr != NULL)	/* loop while Q not empty */
    {
	/* acquire resources waiting command */
	if ((rc = psc_alloc_resources(next_ptr)) == PSC_NO_ERR)
	{
	    psc_deq_wait_resources(next_ptr);
	    psc_enq_wait(next_ptr);
	    if (command_issued_flag == FALSE)
		command_issued_flag = psc_issue_cmd();
	}
	else
	    /* couldn't alloc necessary resources so quit */
            break;
	next_ptr = adp_str.DEVICE_WAITING_FOR_RESOURCES_head;
    }	/* end while */
    if ((dev_ptr != NULL) && (dev_ptr->head_pend != NULL) &&
       (dev_ptr->active == NULL))
    {
	bp = dev_ptr->head_pend;
	/* if xfer count <= 4k or no devices waiting            */
	/* for resources then attempt to alloc resources        */
	/* for command sitting on  pending queue of this device */
	if ((bp->bufstruct.b_bcount <= 4096) ||
	    (adp_str.DEVICE_WAITING_FOR_RESOURCES_head == NULL))
	{
	    /* attempt to allocate resources  for dev. on pending queue */
	    if ((rc = psc_alloc_resources(dev_ptr)) == PSC_NO_ERR)
	    {
		psc_enq_wait(dev_ptr);
		if (command_issued_flag == FALSE)
		    command_issued_flag = psc_issue_cmd();
	    }
	    else
                psc_enq_wait_resources(dev_ptr);
	}
	else
	    psc_enq_wait_resources(dev_ptr);
    }
    if ((command_issued_flag == FALSE) &&
	(adp_str.DEVICE_ACTIVE_head != NULL))
    {
	psc_patch_iowait_int_off();
	psc_write_reg(DSP, DSP_SIZE, word_reverse(
					       adp_str.SCRIPTS[0].dma_ptr));
	DEBUG_0 ("psc_check_wait_queue: Issuing IOWAIT\n")
    }
    DEBUG_0 ("leaving psc_check_wait_queue\n")
    TRACE_1 ("out chkq", 0)
}  /* end of psc_check_wait_queue */

/************************************************************************/
/*                                                                      */
/* NAME:        psc_cleanup_reset                                       */
/*                                                                      */
/* FUNCTION:    Cleans up the active queues due to a register error or  */
/*              a dma error.                                            */
/*                                                                      */
/* NOTES:                                                               */
/*  This is called when an error occurrs that requires a bus reset to   */
/*  be issued.  All commands will be cleaned up upon completion of the  */
/*  reset by psc_cleanup_reset_error.                                   */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*                                                                      */
/* INPUTS:                                                              */
/*      unint  err_type                                                 */
/*            1=dma 2=register                                          */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      none                                                            */
/*                                                                      */
/* ERROR DESCRIPTION:  The following errno values may be returned:      */
/*      none                                                            */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*                                                                      */
/************************************************************************/
void
psc_cleanup_reset(int err_type)
{
    struct dev_info *dev_ptr;	/* device structure */
    struct sc_buf *bp;

    DEBUG_1 ("Entering psc_cleanup_reset routine error = 0x%x\n", err_type);
    TRACE_1 ("in clres", err_type)

    /* loop while devices are on Active queue */
    dev_ptr = adp_str.DEVICE_ACTIVE_head;
    while (dev_ptr != NULL)
    {
	/* this error may have been caused by a timeout or epow */
	/* or a glitch on the reset line                        */
	if ((bp = dev_ptr->active) != NULL)
	{
	    /* if active cmd status not already set */
	    if (!(bp->bufstruct.b_flags & B_ERROR))
	    {
		/* set scsi status in sc_buf to say adapter error */
		bp->status_validity = SC_ADAPTER_ERROR;
		switch (err_type)
		{
		case PHASE_ERR:
		case DISC_ERR:
		    bp->general_card_status = SC_SCSI_BUS_FAULT;
		    break;
		case DEVICE_ERR:
		    bp->general_card_status = SC_NO_DEVICE_RESPONSE;
		    break;
		case HOST_ERR:
		case REG_ERR:
		case DMA_ERR:
		default:
		    bp->general_card_status = SC_HOST_IO_BUS_ERR;
		    break;
		}
		bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
		bp->bufstruct.b_flags |= B_ERROR;
		bp->bufstruct.b_error = EIO;
	    }
	}
	if ((adp_str.ABORT_BDR_head == dev_ptr) ||
	    (dev_ptr->ABORT_BDR_bkwd != NULL))
	    psc_deq_abort_bdr(dev_ptr);
	if (dev_ptr->ioctl_wakeup == TRUE)
	{
	    dev_ptr->ioctl_errno = EIO;
	    dev_ptr->ioctl_wakeup = FALSE;
	    e_wakeup(&dev_ptr->ioctl_event);
	}
	dev_ptr = dev_ptr->DEVICE_ACTIVE_fwd;
    }	/* end of while loop */
    if (err_type == REG_ERR)
    {
	psc_chip_register_reset(PSC_RESET_CHIP | PSC_RESET_DMA);
	psc_command_reset_scsi_bus();
    }
    else
	if (err_type == DMA_ERR)
	{
	    psc_command_reset_scsi_bus();
	    psc_chip_register_reset(PSC_RESET_DMA);
	}
	else
	{
	    psc_command_reset_scsi_bus();
	}
    TRACE_1 ("out clre", err_type)
    DEBUG_0 ("leaving psc_cleanup_reset\n")
}  /* end psc_cleanup_reset */

/************************************************************************/
/*                                                                      */
/* NAME:        psc_cleanup_reset_error                                 */
/*                                                                      */
/* FUNCTION:    Cleans up the active queues after a scsi bus reset has  */
/*              occurred.                                               */
/*                                                                      */
/* NOTES:                                                               */
/*  After a scsi bus reset occurs, all active commands are cleaned up.  */
/*  In addition, all resources are freed and all queues are serached to */
/*  ensure command restart occurs without error.                        */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*                                                                      */
/* INPUTS:                                                              */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      none                                                            */
/*                                                                      */
/* ERROR DESCRIPTION:  The following errno values may be returned:      */
/*      none                                                            */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*                                                                      */
/************************************************************************/
void
psc_cleanup_reset_error()
{
    struct dev_info *dev_ptr;	/* device structure */
    struct dev_info *next_ptr;	/* device structure */
    struct sc_buf *bp;
    int     i, rc, waiting_queue, resource_queue;
    int     dev_index, save_error;

    DEBUG_0 ("Entering psc_cleanup_reset_error routine\n")
    TRACE_1 ("in rserr", 0)

    /* make sure the adapter timers are stopped */
    w_stop(&(adp_str.adap_watchdog.dog));
    /* loop while devices are on Active queue */
    dev_ptr = adp_str.DEVICE_ACTIVE_head;
    while (dev_ptr != NULL)
    {
	/* this error may have been caused by a timeout or epow */
	/* or a glitch on the reset line                        */
	w_stop(&dev_ptr->dev_watchdog.dog);
	if ((bp = dev_ptr->active) != NULL)
	{
	    /* free resources for this command sta's, tcw's */
	    (void) psc_free_resources(dev_ptr, FALSE);
	    /* if active cmd status not set */
	    if (!(bp->bufstruct.b_flags & B_ERROR))
	    {
		bp->status_validity = SC_ADAPTER_ERROR;
		bp->general_card_status = SC_SCSI_BUS_RESET;
		bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
		bp->bufstruct.b_flags |= B_ERROR;
		bp->bufstruct.b_error = EIO;
	    }
	    iodone((struct sc_buf *) bp);
	}
	dev_ptr->active = NULL;
	dev_ptr->flags = RETRY_ERROR;
	psc_deq_active(dev_ptr);
	if ((adp_str.ABORT_BDR_head == dev_ptr) ||
	    (dev_ptr->ABORT_BDR_bkwd != NULL))
	    psc_deq_abort_bdr(dev_ptr);
	if (dev_ptr->head_pend != NULL)
	    psc_fail_cmd(dev_ptr, EIO, 1); /* kill active and pending cmds */
	if (dev_ptr->queue_state == STOPPING)
	    e_wakeup(&dev_ptr->stop_event);
	else
	    dev_ptr->queue_state = HALTED;
	if (dev_ptr->ioctl_wakeup == TRUE)
	{
	    if (dev_ptr->flags & CAUSED_TIMEOUT)
	    {
		dev_ptr->ioctl_errno = ETIMEDOUT;
	    }
	    else
	    {
		dev_ptr->ioctl_errno = EIO;
	    }
	    dev_ptr->ioctl_wakeup = FALSE;
	    e_wakeup(&dev_ptr->ioctl_event);
	}
	dev_ptr = adp_str.DEVICE_ACTIVE_head;
    }	/* end of while loop */

    dev_ptr = adp_str.DEVICE_WAITING_head;
    while (dev_ptr != NULL)
    {
	psc_deq_wait(dev_ptr);
	(void) psc_free_resources(dev_ptr, FALSE);
	if (dev_ptr->head_pend == NULL)
	{	/* queue is empty  */
	    dev_ptr->head_pend = dev_ptr->active;
	    dev_ptr->tail_pend = dev_ptr->active;
	}
	else
	{	/* pending queue not empty */
	    /* point first cmd's av_forw at the new request */
	    dev_ptr->active->bufstruct.av_forw =
					  (struct buf *) dev_ptr->head_pend;
	    dev_ptr->head_pend = dev_ptr->active;
	}
	dev_ptr->active = NULL;
	dev_ptr = adp_str.DEVICE_WAITING_head;
    }

    dev_ptr = adp_str.DEVICE_WAITING_FOR_RESOURCES_head;
    while (dev_ptr != NULL)
    {
	psc_deq_wait_resources(dev_ptr);
	dev_ptr = adp_str.DEVICE_WAITING_FOR_RESOURCES_head;
    }

    /* clear the DMA and scsi fifo's on the chip */
    if (psc_write_reg(DFIFO, DFIFO_SIZE, 0x40) == REG_FAIL)
    {	/* if ther's a failure to clear the fifo, reset the chip */
	psc_chip_register_reset(PSC_RESET_CHIP | PSC_RESET_DMA);
    }
    /* loop through hash table setting the negotiation structs to TRUE */
    for (i = 0; i <= MAX_DEVICES; i++)
    {
	dev_ptr = adp_str.device_queue_hash[i];
	/* a check is made to determine if the device queue has */
	/* pending command that, due to the reset error, were   */
	/* moved to the waiting or waiting for resoures queues. */
	/* if so, they are taken care of here.                  */
	if (dev_ptr != NULL)
	{
	    dev_ptr->negotiate_flag = TRUE;
	    dev_ptr->restart_in_prog = TRUE;
	}
    }
    adp_str.restart_watchdog.dog.restart =
	adp_str.ddi.cmd_delay + 1;
    w_start(&(adp_str.restart_watchdog.dog));
    TRACE_1 ("out rser", 0)
    DEBUG_0 ("leaving psc_cleanup_reset_error\n")
}  /* end psc_cleanup_reset */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_epow                                                        */
/*                                                                        */
/* FUNCTION:  Scsi chip device driver suspend routine.                    */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine is called when an EPOW (Early Power Off Warning)  */
/*         has been detected.                                             */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*      adapter_def - adapter unique data structure (one per adapter)     */
/*      intr    -  kernel interrupt handler information structure         */
/*                                                                        */
/* INPUTS:                                                                */
/*      handler - pointer to the intrpt handler structure                 */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      INTR_SUCC - interrupt was not processed                           */
/*      INTR_SUCC - interrupt was processed                               */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*      i_disable  i_enable                                               */
/*                                                                        */
/**************************************************************************/
int
psc_epow(struct intr * handler)
{
    struct dev_info *dev_ptr;	/* device structure */
    struct dev_info *next_ptr;	/* device structure */
    int     i, rc;
    int     waiting_queue, resource_queue;
    int     old_pri1, old_pri2;
    int     issue_cmd_flag;

    DEBUG_0 ("Entering psc_epow routine.\n")
    if ((handler->flags & EPOW_SUSPEND) ||
	((handler->flags & EPOW_BATTERY) &&
	 (!(adp_str.ddi.battery_backed))))
    {
	DEBUG_0 ("psc_epow: suspend\n")
	adp_str.epow_state = EPOW_PENDING;
	psc_command_reset_scsi_bus();
    }	/* endif */
    else  /* restart again after an EPOW has occurred */
    {
	if (handler->flags & EPOW_RESUME)
	{
	    DEBUG_0 ("psc_epow: resume\n")
	    /* handle a resume command */
	    /* disable to int level during this */
	    old_pri1 = i_disable(adp_str.ddi.int_prior);

	    /* disable to close window around the test */
	    old_pri2 = i_disable(INTEPOW);

	    if (((!(handler->flags & EPOW_BATTERY)) &&
		 (!(handler->flags & EPOW_SUSPEND))) &&
		(adp_str.epow_state == EPOW_PENDING))
	    {
		adp_str.epow_state = 0;	/* reset epow state */
		i_enable(old_pri2);	/* return to lower priority */
	    }
	    else
	    {   /* either a SUSPEND has re-occurred, or this */
		/* adap was not put in epow pending state.   */
		/* for these cases--leave adapter as is      */
		i_enable(old_pri2);	/* return to lower priority */
	    }
	    i_enable(old_pri1);	/* re-enable */
	}	/* end if */
    }	/* end else */

    DEBUG_0 ("leaving epow: intr_succeeded\n")
    return (INTR_SUCC);
}  /* end psc_epow */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_command_watchdog                                            */
/*                                                                        */
/* FUNCTION:  Command timer routine.                                      */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine is called when a command has timed out.  A BDR is */
/*         issued in an attempt to cleanup the device This routines exe-  */
/*         cutes to reset the bus during normal command executtion.       */
/*                                                                        */
/* DATA STRUCTURES: None                                                  */
/*  w - gives you the address of the dev_info structure                   */
/*                                                                        */
/* INPUTS: None                                                           */
/*         w - pointer to watch dog timer structure which timed out       */
/*                                                                        */
/* RETURN VALUE DESCRIPTION: None                                         */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*        w_start  e_wakeup                                               */
/**************************************************************************/
void
psc_command_watchdog(struct watchdog * w)
{
    struct dev_info *dev_ptr;
    struct sc_buf *bp;
    struct timer *tdog;
    int     i, old_pri, dev_index;
    uchar   issue_abrt_bdr;
    uchar   istat_val;
    caddr_t iocc_addr;
    register int rc;

    DEBUG_0 ("Entering psc_command_watchdog\n")
    old_pri = i_disable(adp_str.ddi.int_prior);

    tdog = (struct timer *) w;
    TRACE_1 ("in watch", tdog->timer_id)
    if (tdog->timer_id == PSC_SIOP_TMR)
    {
	w_stop(&adp_str.adap_watchdog.dog);
	/* if we failed in causing a chip abort, the chip is hung */
	/* reset the chip and DMA before doing a scsi bus reset.  */
	psc_chip_register_reset(PSC_RESET_CHIP | PSC_RESET_DMA);
	psc_command_reset_scsi_bus();
    }
    else
	if (tdog->timer_id == PSC_RESET_TMR)
	{
	    w_stop(&adp_str.reset_watchdog.dog);
	    /* if we failed in causing the scsi bus to reset.  This   */
	    /* guarantees that the chip is hung, so reset the chip.   */
	    psc_chip_register_reset(PSC_RESET_CHIP | PSC_RESET_DMA);
	    psc_command_reset_scsi_bus();
	}
	else
	    if (tdog->timer_id == PSC_RESTART_TMR)
	    {
		w_stop(&adp_str.restart_watchdog.dog);
		if (adp_str.epow_state == EPOW_PENDING)
		{	/* an epow is still pending, restart timer */
		    adp_str.restart_watchdog.dog.restart =
			adp_str.ddi.cmd_delay + 1;
		    w_start(&(adp_str.restart_watchdog.dog));
		}
		else
		{
		    /* loop through hash table restarting each device */
		    for (i = 0; i <= MAX_DEVICES; i++)
		    {
			dev_ptr = adp_str.device_queue_hash[i];
			if ((dev_ptr != NULL) && (dev_ptr->restart_in_prog))
			{
			    dev_ptr->restart_in_prog = FALSE;
			    if (dev_ptr->head_pend != NULL)
				psc_start(dev_ptr);
			}
		    }
		}
	    }
	    else
	    {
		/* if here a fatal command timeout occurred */
		dev_ptr = (struct dev_info *) ((uint) (tdog));
		w_stop(&dev_ptr->dev_watchdog.dog);
		dev_ptr->flags |= CAUSED_TIMEOUT;
		if (dev_ptr->cmd_activity_state == 0)
		    return;
		if ((adp_str.ABORT_BDR_head == dev_ptr) ||
		    (dev_ptr->ABORT_BDR_bkwd != NULL))
		    issue_abrt_bdr = TRUE;
		else
		    issue_abrt_bdr = FALSE;
		/* is this is an abort or bdr */
		DEBUG_1 ("Command active = %d\n", dev_ptr->cmd_activity_state)
		DEBUG_1 ("issue_abrt_bdr = %d\n", issue_abrt_bdr)
		if ((dev_ptr->cmd_activity_state == ABORT_IN_PROGRESS) ||
		    (dev_ptr->cmd_activity_state == BDR_IN_PROGRESS) ||
		    (issue_abrt_bdr == TRUE))
		{
		    DEBUG_0 ("psc_command_watchdog: ABORT/BDR in PROGRESS\n")
		    if ((bp = dev_ptr->active) != NULL)
		    {
			bp->status_validity = SC_ADAPTER_ERROR;
			bp->general_card_status = SC_CMD_TIMEOUT;
			bp->scsi_status = SC_GOOD_STATUS;
			bp->bufstruct.b_flags |= B_ERROR;
			bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
			bp->bufstruct.b_error = EIO;
		    }
		    psc_command_reset_scsi_bus();	/* reset scsi bus */
		}
		else	/* regular command */
		{
		    DEBUG_0 ("psc_command_watchdog: regular cmd \n")
		    if (issue_abrt_bdr)
			psc_deq_abort_bdr(dev_ptr);
		    if (dev_ptr->ioctl_wakeup)
		    {
			e_wakeup(&dev_ptr->ioctl_event);
			dev_ptr->ioctl_errno = ETIMEDOUT;
			dev_ptr->ioctl_wakeup = FALSE;
		    }
		    /* set up error's will come in through BDR */
		    /* after BDR */
		    if ((bp = dev_ptr->active) != NULL)
		    {
			bp->status_validity = SC_ADAPTER_ERROR;
			bp->general_card_status = SC_CMD_TIMEOUT;
			bp->scsi_status = SC_GOOD_STATUS;
			bp->bufstruct.b_flags |= B_ERROR;
			bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
			bp->bufstruct.b_error = EIO;
		    }
		    dev_ptr->flags |= SCSI_BDR;
		    dev_ptr->flags &= ~RETRY_ERROR;
		    dev_ptr->retry_count = 1;
		    psc_enq_abort_bdr(dev_ptr);
		    /* if dev is not connected                          */
		    /* to the scsi bus, then SUSPEND processing by chip */
		    /* (use a timer to keep an eye on things)           */

		    /* see if chip is waiting for reselection */
		    istat_val = -1;
		    iocc_addr =
			BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
		    psc_patch_iowait_int_on();
		    for (i = 0; i < 3; i++)
		    {
			rc = BUS_GETCX(iocc_addr + ISTAT, &istat_val);
			if (!(istat_val & CONN_INT_TEST))
			{	/* The register read is successful */
			    rc |= BUS_PUTCX(iocc_addr + ISTAT, 0x80);
			    if (rc != REG_FAIL)
			    {
				/* Start a timer for the chip suspend */
				w_start(&(adp_str.adap_watchdog.dog));
				break;
			    }
			    else
			    {
				/* Failed during a register write */
				istat_val = rc;
			    }
			}
			if (istat_val != REG_FAIL)
			{	/* Can't issue abort so  */
			    break;
			}
		    }
		    if (istat_val == REG_FAIL)
		    {
			psc_command_reset_scsi_bus();
		    }
		    else
		    {
			dev_ptr->dev_watchdog.dog.restart = LONGWAIT;
			w_start(&dev_ptr->dev_watchdog.dog);
		    }
		}
	    }	/* else dev_ptr */
    TRACE_1 ("outwatch", 0)
    i_enable(old_pri);
    DEBUG_0 ("Leaving psc_command_watchdog\n")
}  /* psc_command_watchdog */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_command_reset_scsi_bus                                      */
/*                                                                        */
/* FUNCTION:  Resets the SCSI bus lines.                                  */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine will toggle the SCSI bus reset line for 30 micro- */
/*         seconds to assert a reset on the SCSI bus. This routines exe-  */
/*         cutes to reset the bus during normal command executtion.       */
/*                                                                        */
/* DATA STRUCTURES: None                                                  */
/*                                                                        */
/* INPUTS: None                                                           */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      PSC_NO_ERR - for good completion.                                 */
/*      ERRNO      - value otherwise                                      */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
void
psc_command_reset_scsi_bus()
{
    uchar   save, reset_chip;
    int     rc, i;
    int     old_pri;

    DEBUG_0 ("Entering psc_comand_reset_scsi_bus\n")
    TRACE_1 ("in busrs", 0)
    old_pri = i_disable(adp_str.ddi.int_prior);
    rc = psc_write_reg(ISTAT, ISTAT_SIZE, 0x80);  /* assert the SIOP abort */
    if (rc != REG_FAIL)
    {
	reset_chip = TRUE;
	/* check to see when abort has occurred */
	for (i = 0; i < 30; i++)
	{
	    psc_delay(1);
	    /* deassert abort cmd */
	    rc = psc_read_reg(ISTAT, ISTAT_SIZE);
	    if (!(rc & DIP))
	    {
		if (rc & SIP)
		    (void) psc_read_reg(SSTAT0, SSTAT0_SIZE);
		continue;
	    }
	    /* deassert abort cmd */
	    rc = psc_write_reg(ISTAT, ISTAT_SIZE, 0x00);
	    if (rc == REG_FAIL)
	    {
		break;
	    }

	    /* read the scsi and dma status registers to clear them */
	    rc = psc_read_reg(DSTAT, DSTAT_SIZE);
	    if (rc == REG_FAIL)
	    {
		break;
	    }

	    if (!(rc & DABRT))
	    {
		rc = psc_write_reg(ISTAT, ISTAT_SIZE, 0x80);
		continue;
	    }
	    reset_chip = FALSE;
	    break;
	}
    }
    /* handle fall through case */
    /* if we failed in causing a chip abort, the chip is hung. */
    /* reset the chip and DMA before doing a scsi bus reset.   */
    if ((reset_chip) || (rc == REG_FAIL))
    {
	psc_chip_register_reset(PSC_RESET_CHIP | PSC_RESET_DMA);
    }

    /* attempt to assert the reset */
    /* write SCNTL1 to assert scsi reset */
    /* assert the SCSI reset signal */
    (void) psc_write_reg(SCNTL1, SCNTL1_SIZE, (SCNTL1_INIT | 0x08));
    psc_delay(30);

    /* write SCNTL1 to deassert scsi reset */
    (void) psc_write_reg(SCNTL1, SCNTL1_SIZE, SCNTL1_INIT);

    w_start(&(adp_str.reset_watchdog.dog));
    TRACE_1 ("out busr", 0)
            i_enable(old_pri);
    DEBUG_0 ("LEAVING psc_command_reset_scsi_bus\n")
}  /* psc_command_reset_scsi_bus */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_find_devinfo                                                */
/*                                                                        */
/* FUNCTION:  Finds the dev_info whose script addr is in the DSP          */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      may be called by interrupt or process                             */
/*                                                                        */
/* DATA STRUCTURES: None                                                  */
/*                                                                        */
/* INPUTS: None                                                           */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      struct dev_info *dev_ptr                                          */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*      NULL    if address not found                                      */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/*                                                                        */
/**************************************************************************/
struct dev_info *
psc_find_devinfo(
		 int predefined_dsp)
{
    struct dev_info *dev_ptr;
    int     script_size;
    int     dsp;

    DEBUG_0 ("Entering psc_find_devinfo\n")

    /* theory is get script dma addr from DSP       */
    /* and then look at the dev_info active que     */
    /* and see if the address falls into this arena */
            script_size = INSTRUCTIONS * 8;
    if (predefined_dsp != 0)
	dsp = predefined_dsp;
    else
	dsp = word_reverse(psc_read_reg(DSP, DSP_SIZE));
    if (dsp == REG_FAIL)
	return (NULL);

    /* go through active queue and see if we can find our device */
    dev_ptr = adp_str.DEVICE_ACTIVE_head;
    while (dev_ptr != NULL)
    {
	if ((dsp >= (dev_ptr->script_dma_addr + Ent_scripts_entry_point)) &&
	    (dsp <= (dev_ptr->script_dma_addr + script_size)))
	{
	    DEBUG_1 ("leaving psc_find_devinfo: dev_ptr= 0x%x\n", dev_ptr);
	    return (dev_ptr);
	}
	dev_ptr = dev_ptr->DEVICE_ACTIVE_fwd;	/* go to next entry */
    }	/* while */
    return (NULL);
}  /* end of psc_find_devinfo */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_fail_free_resources                                         */
/*                                                                        */
/* FUNCTION:  Handles the case where free resources fails                 */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      may be called by interrupt or process                             */
/*                                                                        */
/* DATA STRUCTURES: None                                                  */
/*                                                                        */
/* INPUTS:                                                                */
/*      sruct dev_info, struct sc_buf, int err_code                       */
/* RETURN VALUE DESCRIPTION:                                              */
/*      none                                                              */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/*                                                                        */
/**************************************************************************/
int
psc_fail_free_resources(
			struct dev_info * dev_ptr,
			struct sc_buf * bp,
			int err_code)
{
    DEBUG_1 ("Entering psc_fail_free_resources error= 0x%x\n", err_code);
    TRACE_1 ("in failf", 0)
    if (err_code == DMA_SYSTEM)
    {
	psc_cleanup_reset(DMA_ERR);
	psc_logerr(ERRID_SCSI_ERR2, DMA_ERROR, 185,
		   0, dev_ptr, TRUE);
	TRACE_1 ("out faif", 0)
	return (TRUE);
    }
    else
	if (err_code == PSC_DMA_ERROR)
	{
	    if (!(bp->bufstruct.b_flags & B_ERROR))
	    {	/* no previous error */
		bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
		bp->status_validity = SC_ADAPTER_ERROR;
		bp->general_card_status = SC_HOST_IO_BUS_ERR;
		bp->scsi_status = SC_GOOD_STATUS;
		bp->bufstruct.b_flags |= B_ERROR;
		bp->bufstruct.b_error = EIO;
		psc_logerr(ERRID_SCSI_ERR2, DMA_ERROR, 185,
			   0, dev_ptr, TRUE);
	    }
	    iodone((struct buf *) bp);
	    /* fail he rest of the pending queue */
	    if (dev_ptr->head_pend != NULL)
		psc_fail_cmd(dev_ptr, ENXIO, 1);
	}
	else	/* PSC_COPY_ERROR */
	{
	    if (!(bp->bufstruct.b_flags & B_ERROR))
	    {	/* no previous error */
		bp->status_validity = 0;
		bp->bufstruct.b_error = EFAULT;
		bp->bufstruct.b_flags |= B_ERROR;
		bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
	    }
	    iodone((struct buf *) bp);
	    /* fail he rest of the pending queue */
	    if (dev_ptr->head_pend != NULL)
		psc_fail_cmd(dev_ptr, ENXIO, 1);
	}
    DEBUG_0 ("leaving psc_fail_free_resources\n")
    TRACE_1 ("out faif", 0)
    return (FALSE);
}  /* psc_fail_free_resources */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_issue_abort_bdr                                             */
/*                                                                        */
/* FUNCTION:  Handles the issueing of bdr/aborts for interrupts           */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      may be called by interrupt or process                             */
/*                                                                        */
/* DATA STRUCTURES: None                                                  */
/*                                                                        */
/* INPUTS:                                                                */
/*      struct dev_info, struct sc_buf, int err_code                      */
/*      int iodone_flag     if =1 halt queue                              */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      none                                                              */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_issue_abort_bdr(
		    struct dev_info * dev_ptr,
		    uint iodone_flag)
{
    struct sc_buf *bp;
    int     command_issued_flag;
    int     dev_index;
    uint    rc;

    DEBUG_0 ("Entering psc_issue_abort_bdr\n")
    TRACE_1 ("in issAB", dev_ptr)
    if (dev_ptr->flags & SCSI_BDR)
    {
	/* update wait script */
	dev_index = INDEX(dev_ptr->scsi_id, dev_ptr->lun_id);
	psc_reset_iowait_jump(adp_str.SCRIPTS[0].script_ptr, dev_index);
	/* set the status */
	bp = dev_ptr->active;
	if ((iodone_flag) && (!(bp->bufstruct.b_flags & B_ERROR)))
	{
	    bp->scsi_status = GET_CMD_STATUS_BYTE(
		adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].script_ptr);
	    if (bp->scsi_status != SC_GOOD_STATUS)
	    {
		bp->status_validity = SC_SCSI_ERROR;
		bp->bufstruct.b_error = EIO;
		bp->bufstruct.b_flags |= B_ERROR;
	    }
	    else
	    {
		bp->bufstruct.b_error = 0;
		bp->scsi_status = SC_GOOD_STATUS;
		bp->status_validity = 0;
	    }
	}
	else
	    if (!(iodone_flag) && (!(bp->bufstruct.b_flags & B_ERROR)))
	    {
		/* setup and issue the BDR */
		bp->status_validity = 0;
		bp->scsi_status = 0;
		bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
		bp->bufstruct.b_error = ENXIO;
		bp->bufstruct.b_flags |= B_ERROR;
	    }
	if (psc_issue_bdr_script(
			adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].script_ptr,
			dev_ptr->script_dma_addr, dev_index) == REG_FAIL)
	{
	    psc_cleanup_reset(REG_ERR);
	    TRACE_1 ("outissAB", 0)
	    return (FALSE);
	}
	dev_ptr->cmd_activity_state = BDR_IN_PROGRESS;
	dev_ptr->dev_watchdog.dog.restart = LONGWAIT;
	dev_ptr->flags &= ~RETRY_ERROR;
	dev_ptr->retry_count = 1;
	w_start(&dev_ptr->dev_watchdog.dog);
	psc_deq_abort_bdr(dev_ptr);
	TRACE_1 ("outissAB", 0)
	return (FALSE);
    }
    else	/* abort was requested */
    {
	psc_deq_abort_bdr(dev_ptr);
	/* remove device from ABRT/BDR queue */
	bp = dev_ptr->active;
	if ((iodone_flag) && (!(bp->bufstruct.b_flags & B_ERROR)))
	{
	    bp->scsi_status =
		GET_CMD_STATUS_BYTE(
		 adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].script_ptr);
	    if (bp->scsi_status != SC_GOOD_STATUS)
	    {
		bp->status_validity = SC_SCSI_ERROR;
		bp->bufstruct.b_flags |= B_ERROR;
		bp->bufstruct.b_error = EIO;
		rc = psc_free_resources(dev_ptr, FALSE);
		if ((rc != PSC_NO_ERR) &&
		    (psc_fail_free_resources(dev_ptr, bp, rc)))
		{
		    TRACE_1 ("outissAB", 0)
		            return (FALSE);
		}
	    }
	    else
	    {
		bp->bufstruct.b_error = 0;
		bp->scsi_status = SC_GOOD_STATUS;
		bp->status_validity = 0;
		rc = psc_free_resources(dev_ptr, TRUE);
		if ((rc != PSC_NO_ERR) &&
		    (psc_fail_free_resources(dev_ptr, bp, rc)))
		{
		    TRACE_1 ("outissAB", 0)
		            return (FALSE);
		}
	    }
	}
	else
	    if (!(iodone_flag) && (!(bp->bufstruct.b_flags & B_ERROR)))
	    {
		bp->status_validity = 0;
		bp->scsi_status = 0;
		bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
		bp->bufstruct.b_error = ENXIO;
		bp->bufstruct.b_flags |= B_ERROR;
		rc = psc_free_resources(dev_ptr, FALSE);
		if ((rc != PSC_NO_ERR) &&
		    (psc_fail_free_resources(dev_ptr, bp, rc)))
		{
		    TRACE_1 ("outissAB", 0)
		            return (FALSE);
		}
	    }
	iodone((struct buf *) bp);
	dev_ptr->flags = RETRY_ERROR;
	dev_ptr->active = NULL;
	psc_deq_active(dev_ptr);
	/* fail and iodone pending cmds with ENXIO */
	if (dev_ptr->head_pend != NULL)
	    psc_fail_cmd(dev_ptr, ENXIO, 1);

	if (dev_ptr->queue_state == STOPPING)
	    e_wakeup(&dev_ptr->stop_event);
	else
	    dev_ptr->queue_state = HALTED;
	/* if IOCTL wakeup flag is TRUE then call wakeup */
	/* and set IOCTL status to EIO                   */
	if (dev_ptr->ioctl_wakeup == TRUE)
	{
	    dev_ptr->ioctl_errno = EIO;
	    dev_ptr->ioctl_wakeup = FALSE;
	    e_wakeup(&dev_ptr->ioctl_event);
	}
	/* remove device from ACTIVE queue */
    }	/* end else  abort */
    DEBUG_0 ("Leaving psc_issue_abort_bdr\n")
    TRACE_1 ("outissAB", 0)
    return (TRUE);
}  /* psc_issue_abort_bdr */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_scsi_parity_error                                           */
/*                                                                        */
/* FUNCTION:  Handles the scsi parity error function                      */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      may be called by interrupt or process                             */
/*                                                                        */
/* DATA STRUCTURES: None                                                  */
/*                                                                        */
/* INPUTS:                                                                */
/*      struct dev_info, struct sc_buf, int err_code                      */
/*      int iodone_flag     if =1 halt queue                              */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:                                              */
/*      none                                                              */
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_scsi_parity_error(
		      struct dev_info * dev_ptr,
		      int save_interrupt,
		      int issue_abrt_bdr)
{
    struct sc_buf *bp;
    uchar   log_err, gen_err;
    int     i, istat_val, sstat_val;
    uint    dev_index;

    DEBUG_0 ("psc_scsi_parity_error: \n")
    TRACE_2 ("in perr ", dev_ptr, save_interrupt)
    w_stop(&dev_ptr->dev_watchdog.dog);
    switch (save_interrupt)
    {
    case SCSI_UNEXP_DISC:
	log_err = UNEXPECTED_BUS_FREE;
	gen_err = SC_SCSI_BUS_FAULT;
	break;
    case SCSI_GROSS_ERR:
	log_err = GROSS_ERROR;
	gen_err = SC_ADAPTER_HDW_FAILURE;
	break;
    case SCSI_PARITY_ERR:
	log_err = BAD_PARITY_DETECTED;
	gen_err = SC_HOST_IO_BUS_ERR;
	break;
    case SCSI_SEL_FAIL:
	dev_ptr->flags |= SELECT_TIMEOUT;
	log_err = SC_NO_DEVICE_RESPONSE;
	gen_err = SC_NO_DEVICE_RESPONSE;
	break;
    }
    /* for selection timeouts, the chip was designed to */
    /* return an unexpected disconnect after a selection */
    /* timeout.  So in the cases of a selection timeout */
    /* where the ISTAT shows another SIP interrupt wait- */
    /* ing, it can be assumed that this is the discon-  */
    /* nect that is expected.  An attempt is made to    */
    /* clear it here via a read loop of the istat for up */
    /* to 10 usecs.  If this fails, then exit the loop  */
    /* and issue an abort to the device.                */
    if (save_interrupt == SCSI_SEL_FAIL)
    {
	for (i = 0; i < 10; i++)
	{
	    psc_delay(1);
	    istat_val = psc_read_reg(ISTAT, ISTAT_SIZE);
	    if (istat_val == REG_FAIL)
	    {
		psc_cleanup_reset(REG_ERR);
		TRACE_1 ("out perr", 0)
		return (FALSE);
	    }
	    if (istat_val & SIP)
	    {
		sstat_val = psc_read_reg(SSTAT0, SSTAT0_SIZE);
		if (sstat_val == REG_FAIL)
		{
		    psc_cleanup_reset(REG_ERR);
		    TRACE_1 ("out perr", 0)
		    return (FALSE);
		}
		if (sstat_val & SCSI_UNEXP_DISC)
		    issue_abrt_bdr = FALSE;
	    }
	}
    }
    /* for unexpected disconnects, it is possible that a */
    /* selection timeout interrupt is stacked due to the */
    /* design of the chip.  Here, the interrupt is      */
    /* cleared and the error returned is set to repre-  */
    /* set a selection timeout.                         */
    if (save_interrupt == SCSI_UNEXP_DISC)
    {
	istat_val = psc_read_reg(ISTAT, ISTAT_SIZE);
	if (istat_val == REG_FAIL)
	{
	    psc_cleanup_reset(REG_ERR);
	    TRACE_1 ("out perr", 0)
	    return (FALSE);
	}
	if (istat_val & SIP)
	{
	    sstat_val = psc_read_reg(SSTAT0, SSTAT0_SIZE);
	    if (sstat_val == REG_FAIL)
	    {
		psc_cleanup_reset(REG_ERR);
		TRACE_1 ("out perr", 0)
		return (FALSE);
	    }
	    if (sstat_val & SCSI_SEL_FAIL)
	    {
		save_interrupt = SCSI_SEL_FAIL;
		dev_ptr->flags |= SELECT_TIMEOUT;
		log_err = SC_NO_DEVICE_RESPONSE;
		gen_err = SC_NO_DEVICE_RESPONSE;
		issue_abrt_bdr = FALSE;
	    }
	}
    }
    bp = dev_ptr->active;
    /* It is determined which machine caused this */
    /* issue an abort if one is not in progress */
    if (((dev_ptr->cmd_activity_state != ABORT_IN_PROGRESS) &&
	 (dev_ptr->cmd_activity_state != BDR_IN_PROGRESS)) &&
	(issue_abrt_bdr))
    {
	dev_index = INDEX(bp->scsi_command.scsi_id,
			  (bp->scsi_command.scsi_cmd.lun) >> 5);
	/* update wait script */
	psc_reset_iowait_jump(adp_str.SCRIPTS[0].script_ptr,
			      dev_index);
	if (psc_issue_abort_script(
		  adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].script_ptr,
		  dev_ptr->script_dma_addr, dev_ptr, dev_index) == REG_FAIL)
	{
	    psc_cleanup_reset(REG_ERR);
	    TRACE_1 ("out perr", 0)
	    return (FALSE);
	}
	dev_ptr->cmd_activity_state = ABORT_IN_PROGRESS;
	dev_ptr->dev_watchdog.dog.restart = LONGWAIT;
	dev_ptr->flags |= SCSI_ABORT;
	dev_ptr->flags &= ~RETRY_ERROR;
	dev_ptr->retry_count = 1;
	w_start(&dev_ptr->dev_watchdog.dog);
	if ((bp != NULL) &&
	    (!(bp->bufstruct.b_flags & B_ERROR)))
	{
	    /* set scsi stat in sc_buf and bytes   */
	    /* based on the error that got us here */
	    bp->status_validity = SC_ADAPTER_ERROR;
	    bp->general_card_status = gen_err;
	    bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
	    bp->bufstruct.b_flags |= B_ERROR;
	    bp->bufstruct.b_error = EIO;
	    if (save_interrupt != SCSI_SEL_FAIL)
		psc_logerr(ERRID_SCSI_ERR2, log_err,
			   190, 0, dev_ptr, TRUE);
	}
	TRACE_1 ("out perr", 0)
	return (FALSE);
    }
    else
    {	/* abort/bdr on command with no abort completed w/error */
	/* for command timeouts, a BDR is sent to clear the     */
	/* drive.  The expected result is an unexpected discon- */
	/* nect.  If this didn't happen or a selection timeout  */
	/* was recorded, reset the scsi bus.                    */
	if ((dev_ptr->flags & CAUSED_TIMEOUT) &&
	    ((save_interrupt != SCSI_UNEXP_DISC) ||
	     (dev_ptr->flags & SELECT_TIMEOUT)))
	{
	    psc_command_reset_scsi_bus();	/* reset scsi bus */
	    return (FALSE);
	}
	if (bp != NULL)
	{
	    if (!(bp->bufstruct.b_flags & B_ERROR))
	    {
		/* set scsi stat in sc_buf and bytes   */
		/* based on the error that got us here */
		bp->status_validity = SC_ADAPTER_ERROR;
		bp->general_card_status = gen_err;
		bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
		bp->bufstruct.b_flags |= B_ERROR;
		bp->bufstruct.b_error = EIO;
		/* For the case of a scsi selction failure, or */
		/* an unexpected disconnect (bus free) occur- */
		/* ing during negotiation, or an unexpected   */
		/* disconnect (bus free) occurring after a    */
		/* sync message reject, or a bus free occuring */
		/* after an abort or bdr is issued to a device */
		/* no log entry should be made, These are all */
		/* legal scsi phases.  An error, however,     */
		/* should be returned to the calling process. */
		if (!((save_interrupt == SCSI_SEL_FAIL) ||
		      ((save_interrupt == SCSI_UNEXP_DISC) &&
		       (dev_ptr->cmd_activity_state ==
			NEGOTIATE_IN_PROGRESS)) ||
		      ((save_interrupt == SCSI_UNEXP_DISC) &&
		       (dev_ptr->cmd_activity_state ==
			BDR_IN_PROGRESS)) ||
		      ((save_interrupt == SCSI_UNEXP_DISC) &&
		       (dev_ptr->cmd_activity_state ==
			ABORT_IN_PROGRESS))))
		{
		    psc_logerr(ERRID_SCSI_ERR2, log_err,
			       195, 0, dev_ptr, TRUE);
		}
	    }
	    /* status already set and error logged */
	    (void) psc_free_resources(dev_ptr, FALSE);
	    iodone((struct sc_buf *) bp);
	}
	else
	    if ((save_interrupt != SCSI_SEL_FAIL) &&
		(save_interrupt != SCSI_UNEXP_DISC))
	    {
		psc_logerr(ERRID_SCSI_ERR2, log_err,
			   200, 0, dev_ptr, TRUE);
	    }
	/* remove device from active queue */
	dev_ptr->flags = RETRY_ERROR;
	dev_ptr->active = NULL;
	if ((adp_str.ABORT_BDR_head == dev_ptr) ||
	    (dev_ptr->ABORT_BDR_bkwd != NULL))
	    /* there is an abort or bdr to queued */
	    psc_deq_abort_bdr(dev_ptr);
	psc_fail_cmd(dev_ptr, ENXIO, 1);
	if (dev_ptr->queue_state == STOPPING)
	    e_wakeup(&dev_ptr->stop_event);
	else
	    dev_ptr->queue_state = HALTED;
	if (dev_ptr->ioctl_wakeup == TRUE)
	{
	    if (save_interrupt != SCSI_SEL_FAIL)
		dev_ptr->ioctl_errno = PSC_NO_ERR;
	    else
		dev_ptr->ioctl_errno = EIO;
	    dev_ptr->ioctl_wakeup = FALSE;
	    e_wakeup(&dev_ptr->ioctl_event);
	}
	if (dev_ptr->cmd_activity_state == BDR_IN_PROGRESS)
	{
	    dev_ptr->negotiate_flag = TRUE;
	    dev_ptr->restart_in_prog = TRUE;
	    adp_str.restart_watchdog.dog.restart =
		adp_str.ddi.cmd_delay + 1;
	    w_start(&(adp_str.restart_watchdog.dog));
	}
	psc_deq_active(dev_ptr);
    }
    TRACE_1 ("out perr", 0)
    return (TRUE);
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_script_init                                         */
/*                                                                      */
/* FUNCTION:    Adapter Script Initialization Routine                   */
/*      This function will patch all interrupts with the half-word      */
/*      that makes up the hash index into the device structure hash     */
/*      table.  Next, it will patch all the target ids referenced       */
/*      within the script.  Then, it will patch the jump command of the */
/*      global IO_WAIT script to jump to this script when the target    */
/*      device does a reselection of the chip.  Finally, we prepare     */
/*      the negotiation, abort, and bdr scripts with the appropriate    */
/*      messages in case they are ever needed by this device.           */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      None.                                                           */
/*                                                                      */
/* INPUTS:                                                              */
/*      *iowait_vir_addr - the virtual address pointing to the          */
/*              IO_WAIT script that all the scripts will be dependent   */
/*              on as a router.                                         */
/*      *script_vir_addr - the virtual address of the script just       */
/*              created and copied into memory.  The script that needs  */
/*              to be initialized.                                      */
/*      dev_info_hash - the hash value needed by the interrupt handler  */
/*              to determine which device-script caused the programmed  */
/*              interrupt.                                              */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/*                                                                      */
/* EXTERNAL PROCEDURES CALLED:                                          */
/*      None.                                                           */
/*                                                                      */
/************************************************************************/
void
psc_script_init(
		uint * iowait_vir_addr,
		uint * script_vir_addr,
		int dev_info_hash,
		uint iowait_dma_addr,
		uint script_dma_addr)

{
    ulong   top_half_word, id_bit_mask;
    ulong  *target_script, *io_wait_ptr;
    ulong   word_buffer;
    uint    jump_address;
    int     i, scsi_id, offset;
    uchar   id_mask, initiator_mask;

    target_script = (ulong) script_vir_addr;

    /******************** BEGIN INTERRUPT PATCHES **********************/

    /* Entering psc_script_init */

    /* patch all interrupts' msb half-word with the hash value for the */
    /* table hash_dev_info[]. */
    top_half_word = word_reverse((((ulong) dev_info_hash) << 16));
    for (i = 0; i < 17; i++)
    {
	target_script[A_phase_error_Used[i]] &= 0xFFFF0000;
	target_script[A_phase_error_Used[i]] |= top_half_word;
    }
    target_script[A_io_done_Used[0]] &= 0xFFFF0000;
    target_script[A_io_done_Used[0]] |= top_half_word;

    target_script[A_io_done_after_data_Used[0]] &= 0xFFFF0000;
    target_script[A_io_done_after_data_Used[0]] |= top_half_word;

    for (i = 0; i < 2; i++)
    {
	target_script[A_unexpected_status_Used[i]] &= 0xFFFF0000;
	target_script[A_unexpected_status_Used[i]] |= top_half_word;

    }
    target_script[A_unknown_msg_Used[0]] &= 0xFFFF0000;
    target_script[A_unknown_msg_Used[0]] |= top_half_word;

    target_script[A_sync_msg_reject_Used[0]] &= 0xFFFF0000;
    target_script[A_sync_msg_reject_Used[0]] |= top_half_word;

    target_script[A_check_next_io_Used[0]] &= 0xFFFF0000;
    target_script[A_check_next_io_Used[0]] |= top_half_word;

    target_script[A_check_next_io_data_Used[0]] &= 0xFFFF0000;
    target_script[A_check_next_io_data_Used[0]] |= top_half_word;

    target_script[A_ext_msg_Used[0]] &= 0xFFFF0000;
    target_script[A_ext_msg_Used[0]] |= top_half_word;

    target_script[A_cmd_select_atn_failed_Used[0]] &= 0xFFFF0000;
    target_script[A_cmd_select_atn_failed_Used[0]] |= top_half_word;

    target_script[A_err_not_ext_msg_Used[0]] &= 0xFFFF0000;
    target_script[A_err_not_ext_msg_Used[0]] |= top_half_word;

    target_script[A_sync_neg_done_Used[0]] &= 0xFFFF0000;
    target_script[A_sync_neg_done_Used[0]] |= top_half_word;

    target_script[A_abort_msg_error_Used[0]] &= 0xFFFF0000;
    target_script[A_abort_msg_error_Used[0]] |= top_half_word;

    target_script[A_neg_select_failed_Used[0]] &= 0xFFFF0000;
    target_script[A_neg_select_failed_Used[0]] |= top_half_word;

    target_script[A_abort_select_failed_Used[0]] &= 0xFFFF0000;
    target_script[A_abort_select_failed_Used[0]] |= top_half_word;

    target_script[A_abort_io_complete_Used[0]] &= 0xFFFF0000;
    target_script[A_abort_io_complete_Used[0]] |= top_half_word;

    target_script[R_dummy_int_Used[0] - 1] = word_reverse(0x80000000);
    target_script[R_dummy_int_Used[0]] = 0x00000000;

    /*********************************************************************/
    /* The three interrupts, illegal_target_id, unknown_reselect_id, and */
    /* uninitialized_reselect will not be patched over with the          */
    /* top_half_word because these cases represent ones in which we do   */
    /* not have a script associated with a device.  Since we don't       */
    /* recognize the device, it will not exist in the dev_info hash table*/
    /* that the top_half_word is used for.  Thus, it is not included     */
    /* because it is not a valid assignment-patch to the interrupts.     */
    /*********************************************************************/

    /*********************************************************************/
    /* We will patch all the jump values with the virtual address where  */
    /* we want the jump to go to.  We take the relative jump value       */
    /* (in bytes) sitting at the jump address and add it to the base,    */
    /* virtual address and write that back into the jump address         */
    /* location.  We must be sure to take the WORD REVERSE value before  */
    /* and after the addition of the base address.                       */
    /*********************************************************************/
    DEBUG_1 ("script dma address is >>> %x\n", script_dma_addr);
    for (i = 0; i < PATCHES; i++)
    {
	jump_address = word_reverse(PSC_SCRIPT[LABELPATCHES[i]]);
	jump_address += script_dma_addr;
	target_script[LABELPATCHES[i]] = word_reverse(jump_address);
    }
    /*********************************************************************/
    /* Patch the interrupt for the case where we have been reselected by */
    /* a target but we don't have a valid, set jump point.  Since the    */
    /* top_half_word represents a recognized, distinct scsi id, we must  */
    /* patch this interrupt individually everytime a new script is       */
    /* generated and this function is called to intialize the new script */
    /* for its new device.                                               */
    /*********************************************************************/

    io_wait_ptr = (ulong) iowait_vir_addr;
    /* We only want to do this once, at the first open of the device */
    if ((io_wait_ptr == target_script) && (!(adp_str.iowait_inited)))
    {
	/***************************************************************/
	/* Find out what the initiator's SCSI ID is, then make this an */
	/* illegal target id.  If a target ever selects the script     */
	/* with this id, we will hit the illegal_target_id interupt    */
	/***************************************************************/
	initiator_mask = (0x01 << adp_str.ddi.card_scsi_id);
	id_mask = 0x01;
	for (i = 4; i < 20; i += 2)
	{
	    if (initiator_mask == id_mask)
	    {
		io_wait_ptr[i] = word_reverse(0x800C00FF);
	    }
	    else
	    {
		io_wait_ptr[i] = word_reverse(0x800C0000 |
				      ((ulong) (id_mask | initiator_mask)));
	    }
	    id_mask = (id_mask << 1);
	}
	for (i = 0; i < 8; i++)
	{
	    io_wait_ptr[scsi_0_lun_Used[i]] =
		word_reverse(((A_uninitialized_reselect_Used[0]
			       - 1) * 4 + script_dma_addr));
	    io_wait_ptr[scsi_1_lun_Used[i]] =
		word_reverse(((A_uninitialized_reselect_Used[1]
			       - 1) * 4 + script_dma_addr));
	    io_wait_ptr[scsi_2_lun_Used[i]] =
		word_reverse(((A_uninitialized_reselect_Used[2]
			       - 1) * 4 + script_dma_addr));
	    io_wait_ptr[scsi_3_lun_Used[i]] =
		word_reverse(((A_uninitialized_reselect_Used[3]
			       - 1) * 4 + script_dma_addr));
	    io_wait_ptr[scsi_4_lun_Used[i]] =
		word_reverse(((A_uninitialized_reselect_Used[4]
			       - 1) * 4 + script_dma_addr));
	    io_wait_ptr[scsi_5_lun_Used[i]] =
		word_reverse(((A_uninitialized_reselect_Used[5]
			       - 1) * 4 + script_dma_addr));
	    io_wait_ptr[scsi_6_lun_Used[i]] =
		word_reverse(((A_uninitialized_reselect_Used[6]
			       - 1) * 4 + script_dma_addr));
	    io_wait_ptr[scsi_7_lun_Used[i]] =
		word_reverse(((A_uninitialized_reselect_Used[7]
			       - 1) * 4 + script_dma_addr));
	    io_wait_ptr[lun_msg_addr_Used[i]] =
		word_reverse(iowait_dma_addr + Ent_lun_msg_buf);
	}

	io_wait_ptr[(Ent_lun_msg_buf / 4)] = 0x00000000;
	io_wait_ptr[(Ent_lun_msg_buf / 4) + 1] = 0x00000000;

	adp_str.iowait_inited = TRUE; /* only do this once - ever */
    }
    /*********************** END INTERRUPT PATCHES ***********************/
    /*********************** BEGIN ID PATCHES ****************************/

    /*********************************************************************/
    /* We need to set the SCSI ID bits in the format required by SCSI    */
    /* protocol and use the values of the id_bit_mask set in the previous*/
    /* switch-case code.  These ID patches are used in all the SELECT ATN*/
    /* commands to start the SCSI protocol with a target device.         */
    /*********************************************************************/
    scsi_id = SID(dev_info_hash);
    switch (scsi_id)
    {
    case 0:
	id_bit_mask = 0x00010000;
	break;
    case 1:
	id_bit_mask = 0x00020000;
	break;
    case 2:
	id_bit_mask = 0x00040000;
	break;
    case 3:
	id_bit_mask = 0x00080000;
	break;
    case 4:
	id_bit_mask = 0x00100000;
	break;
    case 5:
	id_bit_mask = 0x00200000;
	break;
    case 6:
	id_bit_mask = 0x00400000;
	break;
    case 7:
	id_bit_mask = 0x00800000;
	break;
    default:
	break;
    }
    for (i = 0; i < 4; i++)
    {

	word_buffer = word_reverse(target_script[R_target_id_Used[i]]);
	word_buffer = ((word_buffer & 0xFF00FFFF) | id_bit_mask);
	target_script[R_target_id_Used[i]] = word_reverse(word_buffer);
    }
    /**********************************************************************/
    /* point to the location in memory where our identify_msg_buf resides */
    /* This is used for the regular command, sync, abort, or bdr cases.   */
    /**********************************************************************/
    target_script[identify_msg_addr_Used[0]] =
	word_reverse(script_dma_addr + Ent_identify_msg_buf);
    /**********************************************************************/
    /* initialize the identify_msg_buffer to the lun id of the device     */
    /* this script is associated with.  Bit 7 is set to show that it is   */
    /* an identify message.  Bit 6 is set to show we allow disconnections.*/
    /* The lun pattern of (0-8) is held in Bits 2-0.                      */
    /**********************************************************************/
    id_bit_mask = 0xC0000000;
    id_bit_mask |= ((ulong) LUN(dev_info_hash) << 24);
    target_script[(Ent_identify_msg_buf / 4)] = id_bit_mask;
    target_script[(Ent_identify_msg_buf / 4) + 1] = 0x00000000;

    /**************** END ID PATCHES **********************/
    /**************** BEGIN BUFFER PATCHES ****************/
    /* point to the location in memory where our cmd_msg_in_buf resides. */
    for (i = 0; i < 4; i++)
    {
	target_script[cmd_msg_in_addr_Used[i]] =
	    word_reverse(script_dma_addr + Ent_cmd_msg_in_buf);
    }
    /* Clear out the message buffers used in synchronous negotiation.    */
    target_script[(Ent_cmd_msg_in_buf / 4)] = 0x00000000;
    target_script[(Ent_cmd_msg_in_buf / 4) + 1] = 0x00000000;
    for (i = 0; i < 4; i++)
    {
	target_script[status_addr_Used[i]] =
	    word_reverse(script_dma_addr + Ent_status_buf);
    }
    target_script[reject_msg_addr_Used[0]] =
	word_reverse(script_dma_addr + Ent_reject_msg_buf);
    target_script[(Ent_reject_msg_buf / 4)] = 0x07000000;
    target_script[(Ent_reject_msg_buf / 4) + 1] = 0x00000000;
    /* point to the location in memory where our sync_msg_out_buf resides. */
    target_script[sync_msg_out_addr_Used[0]] =
	word_reverse(script_dma_addr + Ent_sync_msg_out_buf);
    target_script[sync_msg_out_addr2_Used[0]] =
	word_reverse(script_dma_addr + Ent_sync_msg_out_buf2);

    /* Patch the messages needed to do synchronous negotiation.           */
    id_bit_mask = 0xC0000000;
    id_bit_mask |= ((ulong) LUN(dev_info_hash) << 24);
    target_script[(Ent_sync_msg_out_buf / 4)] = 0x00010301 | id_bit_mask;
    word_buffer = 0;
    word_buffer = DEFAULT_MIN_PHASE;
    word_buffer = (word_buffer << 8) | DEFAULT_BYTE_BUF;
    word_buffer = (word_buffer << 16);
    target_script[(Ent_sync_msg_out_buf / 4) + 1] = word_buffer;

    /* Patch the messages needed to do synchronous negotiation.           */
    target_script[(Ent_sync_msg_out_buf2 / 4)] = (0x01030100 |
						  (ulong) DEFAULT_MIN_PHASE);
    target_script[(Ent_sync_msg_out_buf2 / 4) + 1] =
	((ulong) DEFAULT_BYTE_BUF << 24);

    /* point to the location in memory where our cmd_buf resides.         */
    target_script[cmd_addr_Used[0]] =
	word_reverse(script_dma_addr + Ent_cmd_buf);

    /* Clear out the buffers to be used to hold the SCSI commands         */
    target_script[(Ent_cmd_buf / 4)] = 0x00000000;
    target_script[(Ent_cmd_buf / 4) + 1] = 0x00000000;
    target_script[(Ent_cmd_buf / 4) + 2] = 0x00000000;
    target_script[(Ent_cmd_buf / 4) + 3] = 0x00000000;

    /* Clear out the status buffers                                       */
    target_script[(Ent_status_buf / 4)] = 0x00000000;
    target_script[(Ent_status_buf / 4) + 1] = 0x00000000;

    /* point to the location in memory where our extended_msg_buf resides. */
    for (i = 0; i < 5; i++)
	target_script[extended_msg_addr_Used[i]] =
	    word_reverse(script_dma_addr + Ent_extended_msg_buf);

    /* Clear out the extended msg buffers                                 */
    target_script[(Ent_extended_msg_buf / 4)] = 0x00000000;
    target_script[(Ent_extended_msg_buf / 4) + 1] = 0x00000000;

    /* point to the location in memory where our abort_msg_out_buf resides */
    for (i = 0; i < 2; i++)
	target_script[abort_bdr_msg_out_addr_Used[i]] =
	    word_reverse(script_dma_addr + Ent_abort_bdr_msg_out_buf);

    /* Clear out the abort_msg_out msg buffers and set the abort message  */
    target_script[(Ent_abort_bdr_msg_out_buf / 4)] = word_reverse(0x0C000000);
    target_script[(Ent_abort_bdr_msg_out_buf / 4) + 1] = 0x00000000;

    /* point to location in memory where our abort_bdr_msg_in_buf resides. */

    for (i = 0; i < 2; i++)
	target_script[abort_bdr_msg_in_addr_Used[i]] =
	    word_reverse(script_dma_addr + Ent_abort_bdr_msg_in_buf);

    /* Clear out the abort_msg_in msg buffers                             */
    target_script[(Ent_abort_bdr_msg_in_buf / 4)] = 0x00000000;
    target_script[(Ent_abort_bdr_msg_in_buf / 4) + 1] = 0x00000000;
    /**************** END BUFFER PATCHES ******************/
    return;
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_prep_main_script                                    */
/*                                                                      */
/* FUNCTION:    Adapter Script Command Loading Routine                  */
/*      This will patch in the SCSI command bytes, found by looking     */
/*      at the sc_buf to be executed.  Then, it will patch in the       */
/*      number of bytes to be transfered (if any).  In addition, it     */
/*      will patch the jump within the IO_WAIT script to jump to        */
/*      this particular script when the target reselects the chip.      */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      dev_info structure                                              */
/*      sc_buf structure                                                */
/*                                                                      */
/* INPUTS:                                                              */
/*      *iowait_vir_addr - the virtual address pointing to the          */
/*              IO_WAIT script that all the scripts will be dependent   */
/*              on as a router.                                         */
/*      *script_vir_addr - the virtual address of the script just       */
/*              created and copied into memory.  The script that needs  */
/*              to be loaded with the actual SCSI command to be sent to */
/*              the device.  This is the pointer to the TOP             */
/*              of the script, not to the "main" portion of the script. */
/*      *dev_ptr - pointer to the dev_info structure which holds        */
/*              the sc_buf which contains the SCSI command given to us  */
/*              by the driver head.                                     */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/*                                                                      */
/************************************************************************/
void
psc_prep_main_script(
		     uint * iowait_vir_addr,
		     uint * script_vir_addr,
		     struct dev_info * dev_ptr,
		     uint iowait_dma_addr,
		     uint script_dma_addr)
{

    ulong  *target_script, *io_wait_ptr;
    char   *script_byte_ptr;
    uchar   scsi_command, lun, *tmp_ptr;
    ulong   temp_addr, cmd_byte_length, word_buffer;
    int     i, offset_to_jump, rel_jump_in_bytes;
    struct sc_buf *bp;

    /* Entering psc_prep_main_script */
    script_byte_ptr = (char *) script_vir_addr;
    target_script = (ulong *) script_vir_addr;
    io_wait_ptr = (ulong *) iowait_vir_addr;
    bp = dev_ptr->active;

    DEBUG_1 ("psc_prep_main: dma_addr= 0x%x\n", dev_ptr->dma_addr);
    TRACE_1 ("in prepm", 0)

    /* Load the byte length for the scsi cmds.                            */
    cmd_byte_length = (ulong) bp->scsi_command.scsi_length;
    word_buffer = word_reverse(target_script[R_cmd_bytes_out_count_Used[0]]);
    word_buffer = ((word_buffer & 0xFF000000) | (cmd_byte_length));
    target_script[R_cmd_bytes_out_count_Used[0]] = word_reverse(word_buffer);


    /* Load the scsi command into the command buffer                      */
    scsi_command = bp->scsi_command.scsi_cmd.scsi_op_code;
    script_byte_ptr[Ent_cmd_buf] = scsi_command;
    script_byte_ptr[Ent_cmd_buf + 1] = bp->scsi_command.scsi_cmd.lun;
    for (i = 2; i < cmd_byte_length; i++)
	script_byte_ptr[Ent_cmd_buf + i] =
	    bp->scsi_command.scsi_cmd.scsi_bytes[i - 2];

    DEBUG_0 ("SCSI COMMAND ^^^^^^^^^^^^^^^^^^^^^^^^^^^ SCSI COMMAND\n")
    DEBUG_2 ("op_code is       0x%x  for ID  0x%x\n",
	      bp->scsi_command.scsi_cmd.scsi_op_code,
	      bp->scsi_command.scsi_id);
    DEBUG_0 ("SCSI COMMAND ^^^^^^^^^^^^^^^^^^^^^^^^^^^ SCSI COMMAND\n")
    if (dev_ptr->active == NULL)
    {
	dev_ptr->max_disconnect = 0;
	DEBUG_0 ("ERROR, ERROR, ERROR, ERROR, ERROR, ERROR\n")
    }
    else
    {
	if (bp->bufstruct.b_bcount == 0)
	    dev_ptr->max_disconnect = 0x00100000;
	else
	    dev_ptr->max_disconnect = bp->bufstruct.b_bcount;
    }

    DEBUG_1 ("The value set in max_disconnect is %x\n",
	      dev_ptr->max_disconnect);
    for (i = 0; i < 2; i++)
    {
	target_script[R_data_byte_count_Used[i]] &= word_reverse(0xff000000);
	target_script[R_data_byte_count_Used[i]] |= word_reverse(
						   dev_ptr->max_disconnect);
	/* Patch the dma address where the data in or out gets buffered. */
	target_script[data_addr_Used[i]] = word_reverse(dev_ptr->dma_addr);
    }
    /*********************************************************************/
    /* Almost every define and address label in the SCRIPTS is in terms  */
    /* of word offsets except those used in JUMP commands.  These are in */
    /* terms of bytes.  All offsets are relative to the beginning address*/
    /* of the scripts, where the IO_WAIT script also resides.            */
    /*********************************************************************/
    lun = bp->scsi_command.scsi_cmd.lun >> 5;
    switch (bp->scsi_command.scsi_id)
    {
    case 0:
	offset_to_jump = scsi_0_lun_Used[lun];
	break;
    case 1:
	offset_to_jump = scsi_1_lun_Used[lun];
	break;
    case 2:
	offset_to_jump = scsi_2_lun_Used[lun];
	break;
    case 3:
	offset_to_jump = scsi_3_lun_Used[lun];
	break;
    case 4:
	offset_to_jump = scsi_4_lun_Used[lun];
	break;
    case 5:
	offset_to_jump = scsi_5_lun_Used[lun];
	break;
    case 6:
	offset_to_jump = scsi_6_lun_Used[lun];
	break;
    case 7:
	offset_to_jump = scsi_7_lun_Used[lun];
	break;
    default:
	break;
    }

    io_wait_ptr[offset_to_jump] = word_reverse(script_dma_addr +
					       Ent_script_reconnect_point);
    /* Turn off resid subract flag */
    dev_ptr->flags &= ~RESID_SUBTRACT;
    dev_ptr->bytes_moved = bp->bufstruct.b_bcount;
    bp->bufstruct.b_resid = bp->bufstruct.b_bcount;
    dev_ptr->flags |= PREP_MAIN_COMPLETE;
    TRACE_1 ("outprepm", 0)
    return;
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_patch_async_switch_int                              */
/*                                                                      */
/* FUNCTION:    Adapter Script Initialization Routine                   */
/*      This function goes into the main script and patches an interrupt*/
/*      command and the interrupt value A_change_to_async into the      */
/*      Scripts.  If this reserved memory location is not patched       */
/*      (because we have a synchronous device), then it does a simple   */
/*      NOP and drops into the portion of the main script that          */
/*      continues the SCSI command processing.                          */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      None.                                                           */
/*                                                                      */
/* INPUTS:                                                              */
/*      *script_vir_addr - the virtual address of the script that needs */
/*              to be altered with an interrupt we are patching in      */
/*              ourselves.  This can be considered a form of code       */
/*              self-modification.                                      */
/*      dev_info_hash - the hash value needed by the interrupt handler  */
/*              to determine which device-script caused the programmed  */
/*              interrupt.                                              */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/*                                                                      */
/************************************************************************/
void
psc_patch_async_switch_int(
			   uint * script_vir_addr,
			   int dev_info_hash)
{
    ulong  *script_ptr;

    script_ptr = (ulong) script_vir_addr;
    script_ptr[R_dummy_int_Used[0] - 1] = word_reverse(0x98080000);
    /* write in the int op code */
    script_ptr[R_dummy_int_Used[0]] = word_reverse((A_change_to_async |
					    ((ulong) dev_info_hash << 16)));
    return;
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_set_disconnect                                      */
/*                                                                      */
/* FUNCTION:    Adapter Script Initialization Routine                   */
/*      This function goes into the main script and patches the identify*/
/*      message to have the disconnect bit either off or on depending   */
/*      on the flags value from the command of SC_NODISC.               */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      None.                                                           */
/*                                                                      */
/* INPUTS:                                                              */
/*      dev_ptr - pointer to the device information structure for       */
/*              the particular script that needs to be altered.         */
/*                                                                      */
/*      chk_disconnect - flag used to determine if the device wants no  */
/*              disconnect to take place.                               */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/*                                                                      */
/************************************************************************/
void
psc_set_disconnect(
		   struct dev_info * dev_ptr,
		   uchar chk_disconnect)
{
    ulong  *script_ptr;
    ulong   id_bit_mask;

    script_ptr = (ulong) adp_str.SCRIPTS[dev_ptr->cmd_script_ptr].script_ptr;
    dev_ptr->disconnect_flag = chk_disconnect;
    if (chk_disconnect & SC_NODISC)
	id_bit_mask = 0x80000000;
    else
	id_bit_mask = 0xC0000000;
    id_bit_mask |= (((ulong) dev_ptr->lun_id) << 24);
    script_ptr[(Ent_identify_msg_buf / 4)] = id_bit_mask;
    script_ptr[(Ent_identify_msg_buf / 4) + 1] = 0x00000000;
    return;
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_patch_iowait_int_on                                 */
/*                                                                      */
/* FUNCTION:    Adapter Script Initialization Routine                   */
/*      This function goes into the main script and patches an interrupt*/
/*      command and the interrupt value A_iowait_int into the           */
/*      Scripts.                                                        */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      None.                                                           */
/*                                                                      */
/* INPUTS:                                                              */
/*      *script_vir_addr - the virtual address of the script that needs */
/*              to be altered with an interrupt we are patching in      */
/*              ourselves.  This can be considered a form of code       */
/*              self-modification.                                      */
/*      dev_info_hash - the hash value needed by the interrupt handler  */
/*              to determine which device-script caused the programmed  */
/*              interrupt.                                              */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/*                                                                      */
/************************************************************************/
void
psc_patch_iowait_int_on()
{
    ulong  *script_ptr;

    script_ptr = (ulong) adp_str.SCRIPTS[0].script_ptr;
    script_ptr[A_io_wait_Used[0] - 1] = word_reverse(0x98080000);
    return;

}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_patch_iowait_int_off                                */
/*                                                                      */
/* FUNCTION:    Adapter Script Initialization Routine                   */
/*      This function goes into the main script and patches an interrupt*/
/*      command and the interrupt value A_iowait_int into the           */
/*      Scripts.                                                        */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      None.                                                           */
/*                                                                      */
/* INPUTS:                                                              */
/*      dev_info_hash - the hash value needed by the interrupt handler  */
/*              to determine which device-script caused the programmed  */
/*              interrupt.                                              */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/*                                                                      */
/************************************************************************/
void
psc_patch_iowait_int_off()
{
    ulong  *script_ptr;

    script_ptr = (ulong) adp_str.SCRIPTS[0].script_ptr;
    script_ptr[A_io_wait_Used[0] - 1] = word_reverse(0x80000000);
    return;
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_reselect_router                                     */
/*                                                                      */
/* FUNCTION:    Reselect router after an abort                          */
/*      This function routes the command to the proper place when an    */
/*      abort has taken place and we can't issue a new command.         */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine only be caused by the interrupt handler and        */
/*      can not page fault.                                             */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      None.                                                           */
/*                                                                      */
/* INPUTS:                                                              */
/*      None.                                                           */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/*                                                                      */
/************************************************************************/
void
psc_reselect_router()
{
    uchar   targ_scsi_id;

    targ_scsi_id = psc_read_reg(SDID, SDID_SIZE);
    if (targ_scsi_id != 0xff)
    {
	switch (targ_scsi_id)
	{
	case 0x80:
	    targ_scsi_id = psc_write_reg(DSP, DSP_SIZE,
					 word_reverse(
		      ((uint) adp_str.SCRIPTS[0].dma_ptr + Ent_scsi_id_0)));
	    break;
	case 0x40:
	    targ_scsi_id = psc_write_reg(DSP, DSP_SIZE,
					 word_reverse(
		      ((uint) adp_str.SCRIPTS[0].dma_ptr + Ent_scsi_id_1)));
	    break;
	case 0x20:
	    targ_scsi_id = psc_write_reg(DSP, DSP_SIZE,
					 word_reverse(
		      ((uint) adp_str.SCRIPTS[0].dma_ptr + Ent_scsi_id_2)));
	    break;
	case 0x10:
	    targ_scsi_id = psc_write_reg(DSP, DSP_SIZE,
					 word_reverse(
		      ((uint) adp_str.SCRIPTS[0].dma_ptr + Ent_scsi_id_3)));
	    break;
	case 0x08:
	    targ_scsi_id = psc_write_reg(DSP, DSP_SIZE,
					 word_reverse(
		      ((uint) adp_str.SCRIPTS[0].dma_ptr + Ent_scsi_id_4)));
	    break;
	case 0x04:
	    targ_scsi_id = psc_write_reg(DSP, DSP_SIZE,
					 word_reverse(
		      ((uint) adp_str.SCRIPTS[0].dma_ptr + Ent_scsi_id_5)));
	    break;
	case 0x02:
	    targ_scsi_id = psc_write_reg(DSP, DSP_SIZE,
					 word_reverse(
		      ((uint) adp_str.SCRIPTS[0].dma_ptr + Ent_scsi_id_6)));
	    break;
	case 0x01:
	    targ_scsi_id = psc_write_reg(DSP, DSP_SIZE,
					 word_reverse(
		      ((uint) adp_str.SCRIPTS[0].dma_ptr + Ent_scsi_id_7)));
	    break;
	default:
	    targ_scsi_id = 0;
	    psc_logerr(ERRID_SCSI_ERR2, UNKNOWN_RESELECT_ID,
		       205, 0, NULL, TRUE);
	    psc_command_reset_scsi_bus();
	    break;
	}
    }
    if (targ_scsi_id == 0xff)
    {
	psc_cleanup_reset(REG_ERR);
    }
    return;
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_verify_neg_answer 					*/
/*                                                                      */
/* FUNCTION:    Adapter Script Initialization Routine                   */
/*      This function goes into the main script and patches an interrupt*/
/*      command and the interrupt value A_change_to_async into the      */
/*      Scripts.  If this reserved memory location is not patched       */
/*      (because we have a synchronous device), then it does a simple   */
/*      NOP and drops into the portion of the main script that          */
/*      continues the SCSI command processing.                          */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      None.                                                           */
/*                                                                      */
/* INPUTS:                                                              */
/*      *script_vir_addr - the virtual address of the script that needs */
/*              to be altered with an interrupt we are patching in      */
/*              ourselves.  This can be considered a form of code       */
/*              self-modification.                                      */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/*                                                                      */
/************************************************************************/
void
psc_verify_neg_answer(
		      uint * script_vir_addr,
		      struct dev_info * dev_ptr,
		      int dev_info_hash)

{
    uchar   xfer_period, req_ack_offset;
    int     i, j;

    /* Entering psc_verify_neg_answer */
    xfer_period = (uchar) (script_vir_addr[Ent_extended_msg_buf / 4] >> 8);
    req_ack_offset = (uchar) (script_vir_addr[Ent_extended_msg_buf / 4]);
    if ((xfer_period != DEFAULT_MIN_PHASE) ||
	(req_ack_offset != DEFAULT_BYTE_BUF))
    {
	dev_ptr->agreed_xfer = xfer_period;
	dev_ptr->agreed_req_ack = req_ack_offset;
	if (req_ack_offset == 0)
	{
	    /* if req_ack_offset is zero, we must treat this like */
	    /* an asynchronous disk				      */
	    dev_ptr->async_device = TRUE;

	    script_vir_addr[R_dummy_int_Used[0] - 1] =
		word_reverse(0x98080000);
	    /* write in the int op code */
	    script_vir_addr[R_dummy_int_Used[0]] =
		word_reverse((A_change_to_async |
			      ((ulong) dev_info_hash << 16)));

	    /* set the chip to transfer data in async mode */
	    psc_write_reg((uint) SXFER, (char) SXFER_SIZE, 0x00);
	    dev_ptr->special_xfer_val = 0x00;
	}
	else
	{
	    /* the special_xfer_val has to hold the FIFO byte size as */
	    /* well as the variable which controls the transfer rate  */
	    for (i = 0; i < 8; i++)
		if (xfer_period <= adp_str.xfer_max[i])
		{
		    dev_ptr->special_xfer_val =
			(uchar) ((i << 4) | req_ack_offset);
		    break;
		}
	    psc_write_reg((uint) SXFER, (char) SXFER_SIZE,
			  (uchar) dev_ptr->special_xfer_val);

	    /* patch in special interrupt for sync devices */
	    script_vir_addr[R_dummy_int_Used[0] - 1] =
		word_reverse(0x98080000);
	    /* write in the int op code */
	    script_vir_addr[R_dummy_int_Used[0]] =
		word_reverse((A_set_special_sync |
			      ((ulong) dev_info_hash << 16)));

	}
    }
    else
    {
	dev_ptr->agreed_xfer = DEFAULT_MIN_PHASE;
	dev_ptr->agreed_req_ack = DEFAULT_BYTE_BUF;
    }

    /* Leaving psc_verify_neg_answer */

    return;

}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_reset_iowait_jump                                   */
/*                                                                      */
/* FUNCTION:    Adapter Script Jump Reset Routine                       */
/*      This function does a patch of the jump instruction within the   */
/*      IO_WAIT script.  It resets the jump back to the original default*/
/*      jump that existed before it was initialized to run a script.    */
/*      This function should be called before issuing an abort or bdr   */
/*      Script.  This is action and function acts as a safety net       */
/*      against an illegal or spurious interrupt by a target device     */
/*      that is supposed to be inactive.                                */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      None.                                                           */
/*                                                                      */
/* INPUTS:                                                              */
/*      *iowait_vir_addr - the virtual address pointing to the          */
/*              IO_WAIT script that all the scripts will be dependent   */
/*              on as a router.                                         */
/*      dev_info_hash - the hash value needed by the interrupt handler  */
/*              to determine which device-script caused the programmed  */
/*              interrupt.                                              */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/*                                                                      */
/************************************************************************/
void
psc_reset_iowait_jump(
		      uint * iowait_vir_addr,
		      int dev_info_hash)
{
    int     scsi_id, offset_to_jump, orig_byte_jump_offset;
    ulong  *io_wait_ptr;
    uchar   lun;

    io_wait_ptr = (ulong *) iowait_vir_addr,
    lun = LUN(dev_info_hash);
    scsi_id = SID(dev_info_hash);
    switch (scsi_id)
    {
    case 0:
	orig_byte_jump_offset = A_uninitialized_reselect_Used[0] - 1;
	offset_to_jump = scsi_0_lun_Used[lun];
	break;
    case 1:
	orig_byte_jump_offset = A_uninitialized_reselect_Used[1] - 1;
	offset_to_jump = scsi_1_lun_Used[lun];
	break;
    case 2:
	orig_byte_jump_offset = A_uninitialized_reselect_Used[2] - 1;
	offset_to_jump = scsi_2_lun_Used[lun];
	break;
    case 3:
	orig_byte_jump_offset = A_uninitialized_reselect_Used[3] - 1;
	offset_to_jump = scsi_3_lun_Used[lun];
	break;
    case 4:
	orig_byte_jump_offset = A_uninitialized_reselect_Used[4] - 1;
	offset_to_jump = scsi_4_lun_Used[lun];
	break;
    case 5:
	orig_byte_jump_offset = A_uninitialized_reselect_Used[5] - 1;
	offset_to_jump = scsi_5_lun_Used[lun];
	break;
    case 6:
	orig_byte_jump_offset = A_uninitialized_reselect_Used[6] - 1;
	offset_to_jump = scsi_6_lun_Used[lun];
	break;
    case 7:
	orig_byte_jump_offset = A_uninitialized_reselect_Used[7] - 1;
	offset_to_jump = scsi_7_lun_Used[lun];
	break;
    default:
	break;

    }

    io_wait_ptr[offset_to_jump] =
	word_reverse((orig_byte_jump_offset * 4) + iowait_vir_addr);
    return;
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_restore_iowait_jump                                 */
/*                                                                      */
/* FUNCTION:    Adapter Script Command Loading Routine                  */
/*      This will patch the iowait script with the address required     */
/*      for the particular device.  This routine is called after a      */
/*      selection failure when trying to issue an abort or bdr to a     */
/*      device.                                                         */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      dev_info structure                                              */
/*      sc_buf structure                                                */
/*                                                                      */
/* INPUTS:                                                              */
/*      *iowait_vir_addr - the virtual address pointing to the          */
/*              IO_WAIT script that all the scripts will be dependent   */
/*              on as a router.                                         */
/*      script_dma_addr - the dma address of the script just            */
/*              created and copied into memory.                         */
/*      *dev_ptr - pointer to the dev_info structure which holds        */
/*              the sc_buf which contains the SCSI command given to us  */
/*              by the driver head.                                     */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/*                                                                      */
/************************************************************************/
void
psc_restore_iowait_jump(
			uint * iowait_vir_addr,
			struct dev_info * dev_ptr,
			uint script_dma_addr)
{
    struct sc_buf *bp;
    ulong  *io_wait_ptr;
    int     offset_to_jump;
    uchar   lun;

    io_wait_ptr = (ulong *) iowait_vir_addr;
    bp = dev_ptr->active;
    /*********************************************************************/
    /* Almost every define and address label in the SCRIPTS is in terms  */
    /* of word offsets except those used in JUMP commands.  These are in */
    /* terms of bytes.  All offsets are relative to the beginning address*/
    /* of the scripts, where the IO_WAIT script also resides.            */
    /*********************************************************************/
    lun = bp->scsi_command.scsi_cmd.lun >> 5;
    switch (bp->scsi_command.scsi_id)
    {
    case 0:
	offset_to_jump = scsi_0_lun_Used[lun];
	break;
    case 1:
	offset_to_jump = scsi_1_lun_Used[lun];
	break;
    case 2:
	offset_to_jump = scsi_2_lun_Used[lun];
	break;
    case 3:
	offset_to_jump = scsi_3_lun_Used[lun];
	break;
    case 4:
	offset_to_jump = scsi_4_lun_Used[lun];
	break;
    case 5:
	offset_to_jump = scsi_5_lun_Used[lun];
	break;
    case 6:
	offset_to_jump = scsi_6_lun_Used[lun];
	break;
    case 7:
	offset_to_jump = scsi_7_lun_Used[lun];
	break;
    default:
	break;
    }
    io_wait_ptr[offset_to_jump] = word_reverse(script_dma_addr +
					       Ent_script_reconnect_point);
    return;
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_update_dataptr                                      */
/*                                                                      */
/* FUNCTION:    Adapter Script Save Data Pointer                        */
/*      This routine is called when the chip signals us that to save    */
/*      our data pointers before it disconnects.  Usually, this means   */
/*      that the device is going to take a break from the data transfer */
/*      to get more more data from its cache.  Since our data buffer    */
/*      (where the dma'd data is going) is still pointing to the        */
/*      same place, we need to adjust it so that when the device        */
/*      eventually comes back with the rest of the data, we do not      */
/*      overwrite what we have already received from the device.        */
/*      The DBC registers hold how many bytes we have yet to transfer   */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      None.                                                           */
/*                                                                      */
/* INPUTS:                                                              */
/*      *script_vir_addr - the virtual address of the script that needs */
/*              to be patched so that the dma'd information coming from */
/*              from the target device will go to a new memory location.*/
/*      *dev_ptr - pointer to the dev_info structure which holds        */
/*              the sc_buf which contains the SCSI command given to us  */
/*              by the driver head.                                     */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/*                                                                      */
/************************************************************************/
int
psc_update_dataptr(
		   ulong * script_vir_addr,
		   struct dev_info * dev_ptr)
{
    int     i, rc;
    ulong   orig_byte_transfer_request;
    ulong   dma_fifo_bytes, dbc_bytes, left_over_bytes;
    ulong  *target_script, new_dma_offset;
    struct sc_buf *bp;
    register uchar sstat1_sav;

    DEBUG_0 ("Entering psc_update_dataptr\n")

    target_script = (ulong) script_vir_addr;
    dbc_bytes = (word_reverse((ulong) psc_read_reg((uint) DBC,
						   (char) DBC_SIZE)) >> 8);
    orig_byte_transfer_request = (word_reverse(target_script
						[R_data_byte_count_Used[0]])
						 & 0x00FFFFFF);
    bp = dev_ptr->active;
    if (!(bp->bufstruct.b_flags & B_READ))
    {	/* must be a send data command */
	dma_fifo_bytes = (((psc_read_reg((uint) DFIFO, (char) DFIFO_SIZE) &
			    0x3F) - (dbc_bytes & 0x3F)) & 0x3F);
	left_over_bytes = dma_fifo_bytes + dbc_bytes;
	sstat1_sav = psc_read_reg((uint) SSTAT1, (char) SSTAT1_SIZE);
	if (sstat1_sav & 0x20)
	    left_over_bytes++;
	if ((sstat1_sav & 0x40) && !(dev_ptr->async_device))
	    left_over_bytes++;
    }
    else
	left_over_bytes = dbc_bytes;
    new_dma_offset = bp->bufstruct.b_bcount - left_over_bytes;

    DEBUG_0 ("************************************************\n")
    DEBUG_1 ("orig req = 0x%x\n", orig_byte_transfer_request);
    DEBUG_1 ("left over = 0x%x\n", left_over_bytes);
    DEBUG_1 ("new dma off = 0x%x\n", new_dma_offset);
    DEBUG_1 ("max_disconnect BEFORE change = 0x%x\n", dev_ptr->max_disconnect);

    if ((dev_ptr->max_disconnect > left_over_bytes) &&
	(left_over_bytes > 0))
	dev_ptr->max_disconnect = left_over_bytes;

    DEBUG_1 ("max_disconnect AFTER change = 0x%x\n", dev_ptr->max_disconnect);
    DEBUG_0 ("************************************************\n")

    for     (i = 0; i < 2; i++)
    {
	target_script[R_data_byte_count_Used[i]] &= word_reverse(0xff000000);
	target_script[R_data_byte_count_Used[i]] |=
				       word_reverse(dev_ptr->max_disconnect);
	target_script[data_addr_Used[i]] = word_reverse(
				       dev_ptr->dma_addr + new_dma_offset);

    }
    dev_ptr->bytes_moved = orig_byte_transfer_request - left_over_bytes;
    if (bp->scsi_command.flags & SC_NODISC)
    {
	/* This is a no disconnect device that needs the byte  */
	/* count subtraction done here.  In addition, the      */
	/* SXFER reg should be set back to it's original value */
	if (dev_ptr->async_device)
	{
	    rc = psc_write_reg(SXFER, SXFER_SIZE, 0x00);
	    if (rc == REG_FAIL)
		return (rc);
	}
	else
	    if ((dev_ptr->agreed_xfer != DEFAULT_MIN_PHASE) ||
		(dev_ptr->agreed_req_ack < DEFAULT_BYTE_BUF))
	    {
		rc = psc_write_reg(SXFER, SXFER_SIZE,
				   dev_ptr->special_xfer_val);
		if (rc == REG_FAIL)
		    return (rc);
	    }
	bp->bufstruct.b_resid -= dev_ptr->bytes_moved;
    }
    else
    {
	dev_ptr->flags |= RESID_SUBTRACT;
    }
    DEBUG_1 ("NEW dma address is 0x%x\n",
	      word_reverse(target_script[data_addr_Used[0]]));
    DEBUG_0 ("Leaving psc_update_dataptr\n")
    return (PSC_NO_ERR);
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_handle_extended_messages                            */
/*                                                                      */
/* FUNCTION:    Adpater Script to Read and Process Extended Messages    */
/*      This function does is called for two cases:                     */
/*      First, an extended message was received before a command was    */
/*      sent, after a command was sent, or after data was received.     */
/*      We must read in the variable length extended message.  Thus     */
/*      we set up the byte count of how many message bytes be read      */
/*      in and complete the reading in of the messages.                 */
/*      Second, if the extended message is a command to modify the      */
/*      data pointer, then we must modify the dma address that the      */
/*      script will dma to when the script continues.                   */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      None.                                                           */
/*                                                                      */
/* INPUTS:                                                              */
/*      *script_vir_addr - the virtual address of the script that needs */
/*              to be patched to read in the variable lengthed extended */
/*              messages coming in from the target.                     */
/*      *script_dma_addr - the dma address of the script that needs to  */
/*              be fed into the DSP register so that the chip will dma  */
/*              its instructions to itself.                             */
/*      dev_info_hash - the hash value needed by the interrupt handler  */
/*              to determine which device-script caused the programmed  */
/*              interrupt.                                              */
/*      *dev_ptr - pointer to the dev_info structure which holds        */
/*              the sc_buf which contains the SCSI command given to us  */
/*              by the driver head.                                     */
/*      interrupt_flag - Interrupt that resulted in this call by the    */
/*              interrupt handler.                                      */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/************************************************************************/
int
psc_handle_extended_messages(
			     uint * script_vir_addr,
			     uint * script_dma_addr,
			     int dev_info_hash,
			     struct dev_info * dev_ptr,
			     int interrupt_flag)
{
    uint   *target_script;
    uchar   ext_msg_opcode;
    ulong   ext_msg_length;
    int     data_ptr_offset;
    int     i, rc, dsp_val;

    DEBUG_0 ("Entering psc_handle_extended_messages\n")
    target_script = script_vir_addr;
    ext_msg_length = (target_script[Ent_extended_msg_buf / 4] >> 24);
    ext_msg_opcode = (target_script[Ent_extended_msg_buf / 4] >> 16);
    target_script[R_ext_msg_size_Used[0]] &= 0x000000FF;
    target_script[R_ext_msg_size_Used[0]] |= word_reverse(ext_msg_length - 1);

    if (interrupt_flag != A_modify_data_ptr)
    {
	if ((ext_msg_opcode == 0x01) ||
	    (ext_msg_opcode == 0x02) ||
	    (interrupt_flag == A_target_sync_sent))
	{
	    if ((ext_msg_opcode == 0x01) &&
		(interrupt_flag != A_target_sync_sent))
	    {
		if (dev_ptr->active->scsi_command.flags &
		    SC_ASYNC)
		{
		    /* The target is trying to do sync negotiation */
		    /* but we do not want to do sync neg.  Thus    */
		    /* we will send a reject msg out and force it  */
		    /* into ASYNC mode                             */
		    dev_ptr->async_device = TRUE;

		    /* patch the int opcode for async     */
		    target_script[R_dummy_int_Used[0] - 1] =
			word_reverse(0x98080000);
		    target_script[R_dummy_int_Used[0]] =
			word_reverse((A_change_to_async |
				      ((ulong) dev_info_hash << 16)));

		    /* set the chip to async mode */
		    psc_write_reg((uint) SXFER,
				  (char) SXFER_SIZE, 0x00);
		    dev_ptr->special_xfer_val = 0x00;

		    /* go to script that sends msg reject */
		    /* for sync and then continues on     */
		    rc = psc_write_reg((uint) DSP,
				       (char) DSP_SIZE,
				       word_reverse(script_dma_addr +
					     (Ent_reject_target_sync / 4)));
		    return (rc);
		}
		else
		{
		    /* The target is trying to do sync negotiation */
		    /* Since we want to talk in sync mode, we now  */
		    /* set up the scripts in interrupt back into   */
		    /* this function after it has read in the      */
		    /* target's sync message.                      */
		    target_script[Ent_ext_msg_patch / 4] =
			word_reverse(0x98080000);
		    target_script[(Ent_ext_msg_patch / 4) + 1] =
			word_reverse(A_target_sync_sent |
				     ((ulong) dev_info_hash << 16));
		}
	    }
	    else
		if (interrupt_flag == A_target_sync_sent)
		{
		    /* being here tells us that we have previously patched */
		    /* this interrupt in order to move in the target's     */
		    /* sync request.  Now, we must examine the request     */
		    /* to determine if it is higher, lower, or equal to    */
		    /* what we wish to be the sync transfer rate. In this  */
		    /* case the ext_msg_length is really the xfer period   */
		    /* and the ext_msg_opcode is really the rec/ack offset */

		    /* cleanup up patched interrupt.  Make it go */
		    /* back to being a NOP instruction           */
		    target_script[Ent_ext_msg_patch / 4] =
			word_reverse(0x80000000);
		    target_script[(Ent_ext_msg_patch / 4) + 1] =
			word_reverse(script_dma_addr + Ent_ext_msg_patch);

		    /* cleanup up any special or asynch inter- */
		    /* rupt that may have been set for this    */
		    /* device.                                 */
		    target_script[R_dummy_int_Used[0] - 1] =
			word_reverse(0x80000000);
		    target_script[R_dummy_int_Used[0]] =
			0x00000000;

		    /* make sure that the asynch flag is turned off */
		    dev_ptr->async_device = FALSE;

		    if ((ext_msg_length != DEFAULT_MIN_PHASE) ||
			(ext_msg_opcode != DEFAULT_BYTE_BUF))
		    {
			/* disagreement in xfer rate.  If value sent  */
			/* by the target is smaller than what we want */
			/* we take that value.  If it is larger, then */
			/* we send our own smaller value and the      */
			/* target must accept that value.	      */
			if ((ext_msg_length > DEFAULT_MIN_PHASE) ||
			    (ext_msg_opcode > DEFAULT_BYTE_BUF))
			{
			    /* one of target's values is bigger   */
			    /* must go through negotiation.       */
			    dev_ptr->agreed_xfer = DEFAULT_MIN_PHASE;
			    dev_ptr->agreed_req_ack = DEFAULT_BYTE_BUF;
			    /* Patch negotiation msgs */
			    target_script
				[(Ent_sync_msg_out_buf2 / 4)] =
				(0x01030100 |
				 (ulong) DEFAULT_MIN_PHASE);
			    target_script
				[(Ent_sync_msg_out_buf2 / 4) + 1] =
				((ulong) DEFAULT_BYTE_BUF << 24);
			    psc_write_reg(SXFER, SXFER_SIZE,
					  SYNC_VAL);
			    rc = psc_write_reg((uint) DSP,
					       (char) DSP_SIZE,
					       word_reverse(
						script_dma_addr +
					       (Ent_renegotiate_sync / 4)));
			    dev_ptr->flags |= NEG_PHASE_2;
			    return (rc);
			}
			else
			{
			    /* target's values are smaller than   */
			    /* ours.  We accept their values.     */
			    for (i = 0; i < 8; i++)
				if (ext_msg_length <= adp_str.xfer_max[i])
				{
				    dev_ptr->special_xfer_val =
					(uchar) ((i << 4) |
						 ext_msg_opcode);
				    break;
				}
			    dev_ptr->agreed_xfer = (uchar) ext_msg_length;
			    dev_ptr->agreed_req_ack = ext_msg_opcode;
			    rc = psc_write_reg(SXFER, SXFER_SIZE,
					       dev_ptr->special_xfer_val);
			    /* patch in special interrupt for  */
			    /* sync devices                    */
			    script_vir_addr[R_dummy_int_Used[0] - 1] =
						  word_reverse(0x98080000);
			    /* write in the int op code */
			    script_vir_addr[R_dummy_int_Used[0]] =
				  word_reverse((A_set_special_sync |
					      ((ulong) dev_info_hash
					       << 16)));
			    target_script[(Ent_sync_msg_out_buf2 / 4)] =
				     (0x01030100 | (ulong) ext_msg_length);
			    target_script[(Ent_sync_msg_out_buf2 / 4) + 1] =
				     ((ulong) ext_msg_opcode << 24);
			    rc = psc_write_reg((uint) DSP, (char) DSP_SIZE,
					       word_reverse(script_dma_addr +
					       (Ent_renegotiate_sync / 4)));
			    dev_ptr->flags |= NEG_PHASE_2;
			    return (rc);
			}
		    }
		    else
			if (dev_ptr->flags & NEG_PHASE_2)
			{	/* value is the same as our defaults */
			    dev_ptr->agreed_xfer = DEFAULT_MIN_PHASE;
			    dev_ptr->agreed_req_ack = DEFAULT_BYTE_BUF;
			    psc_write_reg(SXFER, SXFER_SIZE, SYNC_VAL);
			    dsp_val = psc_read_reg((uint) DSP,
						   (char) DSP_SIZE);
			    if (dsp_val != REG_FAIL)
				rc = psc_write_reg((uint) DSP, (char) DSP_SIZE,
						   dsp_val);
			    else
				rc = REG_FAIL;
			    dev_ptr->flags &= ~NEG_PHASE_2;
			}
			else
			{	/* device is negotiating at dflt values */
			    dev_ptr->agreed_xfer = DEFAULT_MIN_PHASE;
			    dev_ptr->agreed_req_ack = DEFAULT_BYTE_BUF;
			    /* Patch negotiation msgs */
			    target_script[(Ent_sync_msg_out_buf2 / 4)] =
				  (0x01030100 | (ulong) DEFAULT_MIN_PHASE);
			    target_script[(Ent_sync_msg_out_buf2 / 4) + 1] =
				  ((ulong) DEFAULT_BYTE_BUF << 24);
			    psc_write_reg(SXFER, SXFER_SIZE, SYNC_VAL);
			    rc = psc_write_reg((uint) DSP, (char) DSP_SIZE,
					       word_reverse(script_dma_addr +
					       (Ent_renegotiate_sync / 4)));
			    dev_ptr->flags |= NEG_PHASE_2;
			    return (rc);
			}
		}
	    DEBUG_0 ("Extended msg was not sync negotiation\n")
	}
	else
	    if (ext_msg_opcode == 0x00)
	    {
		/******************************************************/
		/* continue execution of script to read in messages   */
		/* telling us about how they want the data pointers   */
		/* modified.  Then, we hit the interrupt we planted   */
		/* ourselves so we can do data pointer modification.  */
		/******************************************************/
		target_script[Ent_ext_msg_patch / 4] =
						  word_reverse(0x98080000);
		target_script[(Ent_ext_msg_patch / 4) + 1] =
		    word_reverse(A_modify_data_ptr |
				 ((ulong) dev_info_hash << 16));
	    }
	    else	/* unknown opcode */
	    {
		target_script[Ent_ext_msg_patch / 4] =
						 word_reverse(0x98080000);
		target_script[(Ent_ext_msg_patch / 4) + 1] =
		    word_reverse(A_phase_error |
				 ((ulong) dev_info_hash << 16));
	    }
	rc = psc_write_reg((uint) DSP, (char) DSP_SIZE,
		word_reverse(script_dma_addr + (Ent_complete_ext_msg / 4)));
    }
    else
	/*******************************************************************/
	/* Being in this portion of the code implies that before, we had   */
	/* received notification of an extended message.  We then deter-   */
	/* mined that the extended message was a modify_data_ptr command.  */
	/* We read in the fourth byte offset and set up an interrupt to    */
	/* ourselves known as A_modify_data_ptr.  Now, we read what the    */
	/* data offset is that was sent by the target device and we add it */
	/* to our dma data pointer within the dev_struct.  Then, we con-   */
	/* tinue on with the script as before.                             */
	/*******************************************************************/
    {
	/* cleanup up patched interrupt.  Make it go  */
	/* back to being a NOP instruction	      */
	target_script[Ent_ext_msg_patch / 4] =
	    word_reverse(0x80000000);
	target_script[(Ent_ext_msg_patch / 4) + 1] =
	    word_reverse(script_dma_addr +
			 Ent_ext_msg_patch);
	data_ptr_offset = word_reverse(target_script
					[Ent_extended_msg_buf / 4]);
	dev_ptr->dma_addr += data_ptr_offset;
	rc = psc_write_reg((uint) DSP, (char) DSP_SIZE,
		   word_reverse(script_dma_addr + (Ent_cleanup_phase / 4)));
    }
    return (rc);
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_issue_abort_script                                  */
/*                                                                      */
/* FUNCTION:                                                            */
/*      This function will patch in the ABORT message into the message  */
/*      buffer and patch all interrupts within the abort_bdr script     */
/*      to reflect that this was an abort script executing if it ever   */
/*      hits an interrupt because of an error.                          */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      None.                                                           */
/*                                                                      */
/* INPUTS:                                                              */
/*      *script_vir_addr - the virtual address of the script that needs */
/*              to be altered to show that this shared abort/bdr script */
/*              is being used as an abort script this time.             */
/*      *script_dma_addr - the dma address of the script that needs to  */
/*              be fed into the DSP register so that the chip will dma  */
/*              its instructions to itself.                             */
/*      dev_info_hash - the hash value needed by the interrupt handler  */
/*              to determine which device-script caused the programmed  */
/*              interrupt.                                              */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/************************************************************************/
int
psc_issue_abort_script(
		       uint * script_vir_addr,
		       uint * script_dma_addr,
		       struct dev_info * dev_ptr,
		       int dev_info_hash)
{
    ulong  *target_script, top_half_word;
    ulong   id_bit_mask;
    int     rc;

    target_script = (ulong) script_vir_addr;
    top_half_word = ((ulong) dev_info_hash << 16);

    if (dev_ptr->disconnect_flag & SC_NODISC)
	id_bit_mask = 0x80000000;
    else
	id_bit_mask = 0xC0000000;
    id_bit_mask |= 0x00060000;
    target_script[(Ent_abort_bdr_msg_out_buf / 4)] = id_bit_mask;
    target_script[A_abort_msg_error_Used[0]] =
	word_reverse(A_abort_msg_error | top_half_word);
    target_script[A_abort_io_complete_Used[0]] =
	word_reverse(A_abort_io_complete | top_half_word);
    target_script[A_abort_select_failed_Used[0]] =
	word_reverse(A_abort_select_failed | top_half_word);
    rc = psc_write_reg((uint) DSP, (char) DSP_SIZE,
		  word_reverse(script_dma_addr + (Ent_abort_sequence / 4)));
    return (rc);
}

/************************************************************************/
/*                                                                      */
/* NAME:        psc_issue_bdr_script                                    */
/*                                                                      */
/* FUNCTION:                                                            */
/*      This function will patch in the BDR message into the message    */
/*      buffer and patch all interrupts within the abort_bdr script     */
/*      to reflect that this was a bdr script executing if it ever      */
/*      hits an interrupt because of an error.                          */
/*                                                                      */
/*                                                                      */
/* EXECUTION ENVIRONMENT:                                               */
/*      This routine can be called by a process.                        */
/*      It can page fault only if called under a process                */
/*      and the stack is not pinned.                                    */
/*                                                                      */
/* DATA STRUCTURES:                                                     */
/*      None.                                                           */
/*                                                                      */
/* INPUTS:                                                              */
/*      *script_vir_addr - the virtual address of the script that needs */
/*              to be altered to show that this shared abort/bdr script */
/*              is being used as a bdr script this time.                */
/*      *script_dma_addr - the dma address of the script that needs to  */
/*              be fed into the DSP register so that the chip will dma  */
/*              its instructions to itself.                             */
/*      dev_info_hash - the hash value needed by the interrupt handler  */
/*              to determine which device-script caused the programmed  */
/*              interrupt.                                              */
/*                                                                      */
/* RETURN VALUE DESCRIPTION:                                            */
/*      None.                                                           */
/************************************************************************/
int
psc_issue_bdr_script(
		     uint * script_vir_addr,
		     uint * script_dma_addr,
		     int dev_info_hash)
{

    ulong  *target_script, top_half_word;
    int     rc;

    target_script = (ulong) script_vir_addr;
    top_half_word = ((ulong) dev_info_hash << 16);

    target_script[(Ent_abort_bdr_msg_out_buf / 4)] = 0x0C000000;
    /* patch in a bdr interrupt */
    target_script[A_abort_msg_error_Used[0]] =
	word_reverse(A_bdr_msg_error | top_half_word);
    target_script[A_abort_io_complete_Used[0]] =
	word_reverse(A_bdr_io_complete | top_half_word);
    target_script[A_abort_select_failed_Used[0]] =
	word_reverse(A_bdr_select_failed | top_half_word);
    rc = psc_write_reg((uint) DSP, (char) DSP_SIZE,
		    word_reverse(script_dma_addr + (Ent_bdr_sequence / 4)));
    return (rc);
}

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_diagnostic                                                  */
/*                                                                        */
/* FUNCTION:  Runs chip diagnostics.                                      */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine causes diagnostics to be run.                     */
/*         The calling process is responsible for NOT calling this rou-   */
/*         tine if the SCIOSTART failed.  Such a failure would indicate   */
/*         that another process has this device open and interference     */
/*         could cause improper error reporting.                          */
/*                                                                        */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    scsi chip structure - scsi chip information.                        */
/*    diagnostic structure - diagnostic information.                      */
/*                                                                        */
/* INPUTS:                                                                */
/*    cmd       - command arguments.                                      */
/*    devflag   - device flag.                                            */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:  0 for good completion, ERRNO value otherwise*/
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_diagnostic(
	       int cmd,
	       ulong devflag)
{
    int     rc, ret_code;
    struct sc_card_diag *diag_ptr;
    struct sc_card_diag diag;
    caddr_t iocc_addr;

    DEBUG_0 ("Entering psc_diagnostic\n")
    if (!(devflag & DKERNEL))
    {	/* handle user process */
	rc = copyin((char *) cmd, &diag, sizeof(struct sc_card_diag));
	if (rc != 0)
	{
	    return (EFAULT);
	}
    }
    else
    {
	/* handle kernel process */
	bcopy((char *) cmd, &diag, sizeof(struct sc_card_diag));
    }

    rc = 0;
    diag_ptr = &diag;
    bzero(&diag_ptr->diag_rc, sizeof(struct rc));
    if (adp_str.opened != TRUE)
    {
	DEBUG_0 ("psc_diag: adapter not opened\n")
	return EINVAL;
    }
    if (adp_str.open_mode == NORMAL_MODE)
    {
	DEBUG_0 ("psc_diag: adapter not opened in diag mode\n")
	return EACCES;
    }
    switch (diag_ptr->option)
    {
    case SC_CARD_DIAGNOSTICS:
	rc = psc_run_diagnostics(diag_ptr);
	break;
    case SC_RESUME_DIAGNOSTICS:
	rc = psc_run_diagnostics(diag_ptr);
	break;
    case SC_CARD_SCSI_WRAP:
	rc = psc_loop_test_diagnostics(diag_ptr);
	break;
    case SC_CARD_REGS_TEST:
	rc = psc_register_test(diag_ptr);
	if (rc != 0)
	{
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_MISCMPR;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	}
	break;
    case SC_CARD_POS_TEST:
	rc = psc_pos_register_test(diag_ptr);
	if (rc != 0)
	{
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_MISCMPR;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	}
	break;
    case SC_SCSI_BUS_RESET:
	rc = psc_diag_reset_scsi_bus(diag_ptr);
	break;
    default:
	DEBUG_0 ("psc_diag: Invalid command arg\n")
	rc = EINVAL;
	break;
    }
    /* Copy the status bytes to the user space provided. */
    if (!(devflag & DKERNEL))
    {	/* handle user process */
	ret_code = copyout(diag_ptr, (char *) cmd,
			   sizeof(struct sc_card_diag));
	if (ret_code != 0)
	    rc = EFAULT;
    }
    else
    {	/* handle kernel process */
	bcopy(diag_ptr, (char *) cmd, sizeof(struct sc_card_diag));
    }
    /* Assert a chip reset */
    iocc_addr = BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
    BUS_PUTCX((iocc_addr + DCNTL), 0x01);
    psc_delay(1);
    BUS_PUTCX((iocc_addr + DCNTL), 0x00);
    /* chip register init */
    BUS_PUTCX((iocc_addr + SCNTL0), SCNTL0_INIT);
    BUS_PUTCX((iocc_addr + SCNTL1), SCNTL1_INIT);
    BUS_PUTCX((iocc_addr + SIEN), SIEN_MASK);
    BUS_PUTCX((iocc_addr + SCID), (0x01 << adp_str.ddi.card_scsi_id));
    BUS_PUTCX((iocc_addr + SXFER), SYNC_VAL);
    BUS_PUTCX((iocc_addr + DMODE), DMODE_INIT);
    BUS_PUTCX((iocc_addr + DIEN), DIEN_MASK);
    BUS_PUTCX((iocc_addr + DCNTL), DCNTL_INIT);

    BUSIO_DET(iocc_addr);
    DEBUG_0 ("Leaving psc_diagnostic\n");
    return rc;
}  /* end psc_diagnostic */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_run_diagnostics                                             */
/*                                                                        */
/* FUNCTION:  Runs a test of the chip DMA and SCSI FIFOs.                 */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine causes chip diagnostics to be run.  This involves */
/*         doing write, read, and compare of the DMA ans SCSI FIFO on the */
/*         chip.  Parity is also tested during this phase.  In addition, a*/
/*         test of the chip parity logic is performed.                    */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    scsi chip structure - scsi chip information.                        */
/*    diagnostic structure - diagnostic information.                      */
/*                                                                        */
/* INPUTS:                                                                */
/*    diag_ptr - diagnostic structure pointer                             */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:  0 for good completion, ERRNO value otherwise*/
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_run_diagnostics(
		    struct sc_card_diag * diag_ptr)

{
    int     rc, count1, count2;
    char    pio_rc;
    caddr_t iocc_addr;

    DEBUG_0 ("Entering psc_run_diagnostics\n")
    TRACE_1 ("in run_diag", diag_ptr);
    diag_ptr->diag_rc.diag_stat = 0;
    diag_ptr->diag_rc.diag_validity = 0;
    diag_ptr->diag_rc.ahs_validity = 0;
    diag_ptr->diag_rc.ahs = 0;
    iocc_addr = BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
    /* The proceeding section of code writes the entire chip DMA */
    /* FIFO and then performs a readback check.                  */
    for (count1 = 4; count1 < 8; count1++)
    {
	/* enable the the DMA FIFO lane. */
	if ((BUS_PUTCX((iocc_addr + CTEST4), count1)) != 0)
	{
	    DEBUG_0 ("psc_run_diag: error enabling DMA FIFO lane\n");
	    TRACE_1 ("failed C4", 0);
	    TRACE_1 ("out run_diag", EFAULT);
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}
	for (count2 = 1; count2 < 9; count2++)
	{
	    /* force even parity. */
	    if ((BUS_PUTCX((iocc_addr + CTEST7), 0x00)) != 0)
	    {
		DEBUG_0 ("psc_run_diag: error writing CTEST7\n");
		TRACE_1 ("failed C7", 0);
		TRACE_1 ("out run_diag", EFAULT);
		BUSIO_DET(iocc_addr);
		return EFAULT;
	    }
	    /* write a data pattern with even parity to the DMA FIFO. */
	    if ((BUS_PUTCX((iocc_addr + CTEST6), 0xAA)) != 0)
	    {
		DEBUG_0 ("psc_run_diag: error writing CTEST6\n");
		TRACE_1 ("failed C6", 0);
		TRACE_1 ("out run_diag", EFAULT);
		BUSIO_DET(iocc_addr);
		return EFAULT;
	    }
	}
    }
    /* Read DSTAT reg to check DMA FIFO Empty bit. */
    BUS_GETCX((iocc_addr + DSTAT), &pio_rc);
    pio_rc &= 0x80;
    if (pio_rc == 1)
    {
	DEBUG_0 ("psc_run_diag: DMA FIFO empty\n");
	TRACE_1 ("dfifo empty", 0);
	TRACE_1 ("out run_diag", EFAULT);
	(void) BUS_PUTCX((iocc_addr + CTEST4), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    for (count1 = 4; count1 < 8; count1++)
    {
	/* enable the the DMA FIFO lane. */
	if ((BUS_PUTCX((iocc_addr + CTEST4), count1)) != 0)
	{
	    DEBUG_0 ("psc_run_diag: error enabling DMA FIFO lane\n");
	    TRACE_1 ("failed CT4", 0);
	    TRACE_1 ("out run_diag", EFAULT);
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}
	for (count2 = 1; count2 < 9; count2++)
	{
	    BUS_GETCX((iocc_addr + CTEST6), &pio_rc);
	    if (pio_rc != 0xAA)
	    {
		DEBUG_0 ("psc_run_diag: compare error CTEST6\n");
		TRACE_1 ("failed CT6", 0);
		TRACE_1 ("out run_diag", EFAULT);
		BUS_PUTCX((iocc_addr + CTEST4), 0x00);
		diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
		diag_ptr->diag_rc.diag_validity = 0x01;
		diag_ptr->diag_rc.ahs_validity = 0x01;
		diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
		BUSIO_DET(iocc_addr);
		return EFAULT;
	    }
	    BUS_GETCX((iocc_addr + CTEST2), &pio_rc);
	    pio_rc &= 0x08;
	    if (pio_rc != 0)
	    {
		DEBUG_0 ("psc_run_diag: Even parity error in CTEST2\n");
		TRACE_1 ("even par err", 0xc2);
		TRACE_1 ("out run_diag", EFAULT);
		diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
		diag_ptr->diag_rc.diag_validity = 0x01;
		diag_ptr->diag_rc.ahs_validity = 0x01;
		diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
		BUS_PUTCX((iocc_addr + CTEST4), 0x00);
		BUSIO_DET(iocc_addr);
		return EFAULT;
	    }
	}	/* end count2 loop */
    }	/* end count1 loop */
    BUS_GETCX((iocc_addr + DSTAT), &pio_rc);
    pio_rc &= 0x80;
    if (pio_rc != 0x80)
    {
	DEBUG_0 ("psc_run_diag: DMA FIFO not empty\n");
	BUS_PUTCX((iocc_addr + CTEST4), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    for (count1 = 4; count1 < 8; count1++)
    {
	/* enable the the DMA FIFO lane. */
	if ((BUS_PUTCX((iocc_addr + CTEST4), count1)) != 0)
	{
	    DEBUG_0 ("psc_run_diag: error enabling DMA FIFO lane\n");
	    TRACE_1 ("failed CTS4", 0);
	    TRACE_1 ("out run_diag", EFAULT);
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}
	for (count2 = 1; count2 < 9; count2++)
	{
	    /* force odd parity. */
	    if ((BUS_PUTCX((iocc_addr + CTEST7), 0x08)) != 0)
	    {
		DEBUG_0 ("psc_run_diag: error writing CTEST7\n");
		TRACE_1 ("failed CTS7", 0);
		TRACE_1 ("out run_diag", EFAULT);
		diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
		diag_ptr->diag_rc.diag_validity = 0x01;
		diag_ptr->diag_rc.ahs_validity = 0x01;
		diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
		BUSIO_DET(iocc_addr);
		return EFAULT;
	    }
	    /* write a data pattern with odd parity to the DMA FIFO. */
	    if ((BUS_PUTCX((iocc_addr + CTEST6), 0x55)) != 0)
	    {
		DEBUG_0 ("run_diag:error CTEST6 odd parity\n");
		TRACE_1 ("failed CTS6", 0);
		TRACE_1 ("out run_diag", EFAULT);
		diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
		diag_ptr->diag_rc.diag_validity = 0x01;
		diag_ptr->diag_rc.ahs_validity = 0x01;
		diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
		BUSIO_DET(iocc_addr);
		return EFAULT;
	    }
	}	/* end count2 loop */
    }	/* end count1 loop */

    /* Read DSTAT reg to check DMA FIFO Empty bit. */
    rc = psc_read_reg(DSTAT, DSTAT_SIZE);
    BUS_GETCX((iocc_addr + DSTAT), &pio_rc);
    pio_rc &= 0x80;
    if (pio_rc == 0x80)
    {
	DEBUG_0 ("psc_run_diag: DMA FIFO empty\n");
	TRACE_1 ("dfifo empty", 0);
	TRACE_1 ("out run_diag", EFAULT);
	BUS_PUTCX((iocc_addr + CTEST4), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    for (count1 = 4; count1 < 8; count1++)
    {
	/* enable the the DMA FIFO lane. */
	if ((BUS_PUTCX((iocc_addr + CTEST4), count1)) != 0)
	{
	    DEBUG_0 ("psc_run_diag: error enabling DMA FIFO lane\n");
	    TRACE_1 ("failed CTST4", 0);
	    TRACE_1 ("out run_diag", EFAULT);
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}
	for (count2 = 1; count2 < 9; count2++)
	{
	    BUS_GETCX((iocc_addr + CTEST6), &pio_rc);
	    if (pio_rc != 0x55)
	    {
		DEBUG_0 ("psc_run_diag: error comparing CTEST6 odd parity\n");
		TRACE_1 ("failed CTST6", 0);
		TRACE_1 ("out run_diag", EFAULT);
		diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
		diag_ptr->diag_rc.diag_validity = 0x01;
		diag_ptr->diag_rc.ahs_validity = 0x01;
		diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
		BUS_PUTCX((iocc_addr + CTEST4), 0x00);
		BUSIO_DET(iocc_addr);
		return EFAULT;
	    }
	    BUS_GETCX((iocc_addr + CTEST2), &pio_rc);
	    pio_rc &= 0x08;
	    if (pio_rc != 0x08)
	    {
		DEBUG_0 ("psc_run_diag:Odd parity err CTEST2\n");
		TRACE_1 ("odd par err", 0);
		TRACE_1 ("out run_diag", EFAULT);
		diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
		diag_ptr->diag_rc.diag_validity = 0x01;
		diag_ptr->diag_rc.ahs_validity = 0x01;
		diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
		BUS_PUTCX((iocc_addr + CTEST4), 0x00);
		BUSIO_DET(iocc_addr);
		return EFAULT;
	    }
	}	/* end count2 loop */
    }	/* end count1 loop */

    BUS_GETCX((iocc_addr + DSTAT), &pio_rc);
    pio_rc &= 0x80;
    if (pio_rc != 0x80)
    {
	DEBUG_0 ("psc_run_diag: DMA FIFO not empty\n");
	TRACE_1 ("dfifo !empty", 0);
	TRACE_1 ("out run_diag", EFAULT);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	BUS_PUTCX((iocc_addr + CTEST4), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    /* The proceeding section of code writes the entire chip SCSI */
    /* FIFO and then performs a readback check.                   */
    if ((BUS_PUTCX((iocc_addr + SCNTL0), 0x04)) != 0)
    {
	DEBUG_0 ("psc_run_diag: error enabling parity in SCNTL0\n");
	TRACE_1 ("enable par err", 0);
	TRACE_1 ("out run_diag", EFAULT);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    /* start check for even parity logic */
    if ((BUS_PUTCX((iocc_addr + SCNTL1), 0x04)) != 0)
    {
	DEBUG_0 ("psc_run_diag: error forcing even parity in SCNTL1\n");
	TRACE_1 ("even par sctnl1", 0);
	TRACE_1 ("out run_diag", EFAULT);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    if ((BUS_PUTCX((iocc_addr + CTEST4), 0x08)) != 0)
    {
	DEBUG_0 ("psc_run_diag: error enabling SCSI FIFO in CTEST4\n");
	TRACE_1 ("failed en sfifo", 0);
	TRACE_1 ("out run_diag", EFAULT);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    for (count1 = 1; count1 < 9; count1++)
    {
	if ((BUS_PUTCX((iocc_addr + SODL), 0xAA)) != 0)
	{
	    DEBUG_0 ("psc_run_diag: error writing SODL for even parity\n");
	    TRACE_1 ("failed sodl", 0);
	    TRACE_1 ("out run_diag", EFAULT);
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}
    }
    for (count1 = 1; count1 < 9; count1++)
    {
	rc = psc_read_reg(CTEST3, CTEST3_SIZE);
	BUS_GETCX((iocc_addr + CTEST3), &pio_rc);
	if (pio_rc != 0xAA)
	{
	    DEBUG_0 ("psc_run_diag: error comparing CTEST3\n");
	    TRACE_1 ("failed c3", 0);
	    TRACE_1 ("out run_diag", EFAULT);
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	    BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	    BUS_PUTCX((iocc_addr + SCNTL1), 0x00);
	    BUS_PUTCX((iocc_addr + CTEST4), 0x00);
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}
	BUS_GETCX((iocc_addr + CTEST2), &pio_rc);
	pio_rc &= 0x10;
	if (pio_rc != 0)
	{
	    DEBUG_0 ("psc_run_diag: error checking parity in CTEST2\n");
	    TRACE_1 ("failed c2", pio_rc);
	    TRACE_1 ("out run_diag", EFAULT);
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	    BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	    BUS_PUTCX((iocc_addr + SCNTL1), 0x00);
	    BUS_PUTCX((iocc_addr + CTEST4), 0x00);
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}
    }	/* end count1 loop */

    /* Now check the odd parity logic */
    if ((BUS_PUTCX((iocc_addr + SCNTL1), 0x00)) != 0)
    {
	DEBUG_0 ("psc_run_diag: error forcing even parity in SCNTL1\n");
	TRACE_1 ("failed scntl1", 0);
	TRACE_1 ("out run_diag", EFAULT);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    for (count1 = 1; count1 < 9; count1++)
    {
	if ((BUS_PUTCX((iocc_addr + SODL), 0x55)) != 0)
	{
	    DEBUG_0 ("run_diag: error writing SODL odd parity\n");
	    TRACE_1 ("failed odd par", SODL);
	    TRACE_1 ("out run_diag", EFAULT);
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}
    }
    for (count1 = 1; count1 < 9; count1++)
    {
	BUS_GETCX((iocc_addr + CTEST3), &pio_rc);
	if (pio_rc != 0x55)
	{
	    DEBUG_0 ("run_diag: compare error CTEST3 odd parity\n");
	    TRACE_1 ("failed compare", 0);
	    TRACE_1 ("out run_diag", EFAULT);
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	    BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	    BUS_PUTCX((iocc_addr + SCNTL1), 0x00);
	    BUS_PUTCX((iocc_addr + CTEST4), 0x00);
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}
	BUS_GETCX((iocc_addr + CTEST2), &pio_rc);
	pio_rc &= 0x10;
	if (pio_rc != 0x10)
	{
	    DEBUG_0 ("psc_run_diag: error checking parity in CTEST2\n");
	    TRACE_1 ("failed par chk", 0);
	    TRACE_1 ("out run_diag", EFAULT);
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	    BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	    BUS_PUTCX((iocc_addr + SCNTL1), 0x00);
	    BUS_PUTCX((iocc_addr + CTEST4), 0x00);
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}
    }	/* end count1 loop */

    BUS_GETCX((iocc_addr + CTEST3), &pio_rc);
    BUS_GETCX((iocc_addr + SSTAT0), &pio_rc);
    pio_rc &= 0x08;
    if (pio_rc != 0x08)
    {
	DEBUG_0 ("psc_run_diag: error checking Gross error in SSTAT0\n");
	TRACE_1 ("failed gross chk", 0);
	TRACE_1 ("out run_diag", EFAULT);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	BUS_PUTCX((iocc_addr + SCNTL1), 0x00);
	BUS_PUTCX((iocc_addr + CTEST4), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    if ((BUS_PUTCX((iocc_addr + CTEST4), 0x10)) != 0)
    {
	DEBUG_0 ("psc_run_diag: error enabling loopback mode\n");
	TRACE_1 ("CTEST4 error", 0);
	TRACE_1 ("out run_diag", EFAULT);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    if ((BUS_PUTCX((iocc_addr + SCNTL0), 0x0c)) != 0)
    {
	DEBUG_0 ("psc_run_diag: error enabling parity in SCNTL0\n");
	TRACE_1 ("failed scntl0", 0);
	TRACE_1 ("out run_diag", EFAULT);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    if ((BUS_PUTCX((iocc_addr + SCNTL1), 0x04)) != 0)
    {
	DEBUG_0 ("psc_run_diag: error forcing even parity in SCNTL1\n")
	TRACE_1 ("failed scntl1 par", 0);
	TRACE_1 ("out run_diag", EFAULT);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    if ((BUS_PUTCX((iocc_addr + CTEST7), 0x04)) != 0)
    {
	DEBUG_0 ("psc_run_diag: error forcing even parity in CTEST7\n");
	TRACE_1 ("failed ct7 par", 0);
	TRACE_1 ("out run_diag", EFAULT);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    if ((BUS_PUTCX((iocc_addr + SODL), 0x07)) != 0)
    {
	DEBUG_0 ("psc_run_diag: error writing data to SODL\n");
	TRACE_1 ("failed sodl data   ", 0);
	TRACE_1 ("out run_diag", EFAULT);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    if ((BUS_PUTCX((iocc_addr + SCNTL1), 0x44)) != 0)
    {
	DEBUG_0 ("psc_run_diag: error asserting data in SCNTL1\n");
	TRACE_1 ("scntl1 data assert  ", 0);
	TRACE_1 ("out run_diag", EFAULT);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    rc = psc_read_reg(SSTAT0, SSTAT0_SIZE);
    BUS_GETCX((iocc_addr + SSTAT0), &pio_rc);
    pio_rc &= 0x01;
    if (pio_rc != 0)
    {
	DEBUG_0 ("psc_run_diag: error checking parity error in SSTAT0\n")
	TRACE_1 ("par chk err    ", 0);
	TRACE_1 ("out run_diag", EFAULT);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	BUS_PUTCX((iocc_addr + SCNTL1), 0x00);
	BUS_PUTCX((iocc_addr + CTEST4), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    DEBUG_0 ("Leaving psc_run_diagnostics\n");
    BUSIO_DET(iocc_addr);
    return (0);
}  /* end psc_run_diagnostics */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_loop_test_diagnotics                                        */
/*                                                                        */
/* FUNCTION:  Runs a loop test of the chip SCSI buss.                     */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine runs a loop test of the chip scsi bus.  Signals   */
/*         and some register testing is done here.                        */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    scsi chip structure - scsi chip information.                        */
/*    diagnostic structure - diagnostic information.                      */
/*                                                                        */
/* INPUTS:                                                                */
/*    diag_ptr - diagnostic structure pointer                             */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:  0 for good completion, ERRNO value otherwise*/
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_loop_test_diagnostics(
			  struct sc_card_diag * diag_ptr)
{
    int     count1, rc;
    char    pio_rc, count;
    caddr_t iocc_addr;

    DEBUG_0 ("Entering psc_loop_test_diagnostics\n")
    TRACE_1 ("in test ", 0);
    iocc_addr = BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
    rc = BUS_PUTCX((iocc_addr + DCNTL), 0x81);
    if (rc != 0)
    {
	DEBUG_0 ("error writing to chip reset in DCNTL\n")
	BUS_PUTCX((iocc_addr + DCNTL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    rc = BUS_PUTCX((iocc_addr + DCNTL), 0x80);
    if (rc != 0)
    {
	DEBUG_0 ("error clearing chip reset in DCNTL\n")
	BUS_PUTCX((iocc_addr + DCNTL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    TRACE_1 ("arb mode ", 0);
    BUS_PUTCX((iocc_addr + SCNTL0), 0xc0);
    /* Arbitrate for the bus */
    TRACE_1 ("lpback mode ", 0);
    rc = BUS_PUTCX((iocc_addr + CTEST4), 0x10);
    if (rc != 0)
    {
	DEBUG_0 ("error writing to loopback mode in CTEST4\n")
	BUS_PUTCX((iocc_addr + CTEST4), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    TRACE_1 ("arb mode ", 0);
    rc = BUS_PUTCX((iocc_addr + SCNTL0), 0xc0);
    if (rc != 0)
    {
	DEBUG_0 ("error writing to arbitration mode in SCNTL0\n")
	BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    /* Set low level mode */
    rc = BUS_PUTCX((iocc_addr + DCNTL), 0x80);
    if (rc != 0)
    {
	DEBUG_0 ("error writing to low level mode in DCNTL\n")
	BUS_PUTCX((iocc_addr + DCNTL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    /* Set destination ID */
    rc = BUS_PUTCX((iocc_addr + SDID), 0x01);
    if (rc != 0)
    {
	DEBUG_0 ("error writing dest. id in SDID\n")
	BUS_PUTCX((iocc_addr + SDID), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    /* Set card IDs */
    rc = BUS_PUTCX((iocc_addr + SCID), 0x81);
    if (rc != 0)
    {
	DEBUG_0 ("error writing chip id in SCID\n")
	BUS_PUTCX((iocc_addr + SCID), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    TRACE_1 ("start seq ", 0);
    rc = BUS_PUTCX((iocc_addr + SCNTL0), 0xe0);
    if (rc != 0)
    {
	DEBUG_0 ("error writing to start sequence bit in SCNTL0\n")
	BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    /* read ISTAT to see if we are connect to the bus to */
    /* tell if we won arbitration                        */
    count = 5;
    pio_rc = 0;
    TRACE_1 ("arb check ", 0);
    while ((count > 0) && (pio_rc != 0x08))
    {
	BUS_GETCX((iocc_addr + ISTAT), &pio_rc);
	pio_rc &= 0x08;
        TRACE_1 ("ISTAT     ", pio_rc);
	count--;
    }
    if (pio_rc != 0x08)
    {
	DEBUG_0 ("Unsuccessful arbitration phase\n")
	BUS_PUTCX((iocc_addr + SOCL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    TRACE_1 ("sel phase ", 0);
    /* Go to selection phase */
    BUS_GETCX((iocc_addr + SBCL), &pio_rc);
    if (pio_rc < 0)
    {
	DEBUG_0 ("error reading SBCL\n")
	BUS_PUTCX((iocc_addr + SOCL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    pio_rc |= 0x20;
    TRACE_1 ("mid sel  ", rc);
    rc = BUS_PUTCX((iocc_addr + SOCL), pio_rc);
    if (rc != 0)
    {
	DEBUG_0 ("error writing to selection phase in SOCL\n")
	BUS_PUTCX((iocc_addr + SOCL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    count = 5;
    pio_rc = 0;
    while ((count > 0) && (pio_rc != 0x40))
    {
	BUS_GETCX((iocc_addr + SSTAT0), &pio_rc);
	pio_rc &= 0x40;
	count--;
    }
    TRACE_1 ("end sel  ", (int) pio_rc);
    if (pio_rc != 0x40)
    {
	DEBUG_0 ("Unsuccessful selection phase\n")
	BUS_PUTCX((iocc_addr + SOCL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    /* Successful selection phase */
    /* Get to data out phase by asserting REQ */
    TRACE_1 ("strt data out ", 0);
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    pio_rc &= 0xF8;
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc | 0x80));
    for (count = 0; count < 16; count++)
    {
	/* Assert REQ to request target to receive data */
	BUS_GETCX((iocc_addr + SOCL), &pio_rc);
	BUS_PUTCX((iocc_addr + SOCL), (pio_rc | 0x80));
	/* change to target mode to receive data */
	rc = BUS_PUTCX((iocc_addr + SCNTL0), 0xC1);
	if (rc != 0)
	{
	    DEBUG_0 ("error changing to target in SCNTL0\n")
	    BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}

	/* write data onto bus */
	BUS_PUTCX((iocc_addr + SODL), count);
	BUS_GETCX((iocc_addr + SCNTL1), &pio_rc);
	BUS_PUTCX((iocc_addr + SCNTL1), (pio_rc | 0x40));

	/* assert ACK to acknowledge data received */
	BUS_GETCX((iocc_addr + SOCL), &pio_rc);
	BUS_PUTCX((iocc_addr + SOCL), (pio_rc | 0x40));

	/* compare data received */
	BUS_GETCX((iocc_addr + SBDL), &pio_rc);
	if (pio_rc != count)
	{
	    DEBUG_0 ("error comparing data out of SBDL\n")
	    TRACE_1 ("cmp error  ", count);
	    BUS_PUTCX((iocc_addr + SOCL), 0x00);
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}
	/* deassert REQ */
	BUS_GETCX((iocc_addr + SOCL), &pio_rc);
	BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0x7F));

	/* deassert ACK */
	BUS_GETCX((iocc_addr + SOCL), &pio_rc);
	BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0xBF));
	BUS_PUTCX((iocc_addr + SODL), 0x00);
    }	/* end for loop */
    /* Write the bus phases in the SOCL reg to bring the chip */
    /* to the command completion phase                        */

    /* deassert ATN */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0xF7));

    /* change to target mode to receive data */
    rc = BUS_PUTCX((iocc_addr + SCNTL0), 0xC1);
    if (rc != 0)
    {
	DEBUG_0 ("error changing to target in SCNTL0\n")
	BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }

    /* Get to msg in phase (assert I/O,MSG & C/D) */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc | 0x07));

    /* assert REQ */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc | 0x80));

    /* indicate command completion */
    BUS_PUTCX((iocc_addr + SODL), 0x00);

    /* assert ACK */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc | 0x40));

    /* deassert ATN */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0xF7));

    /* deassert ACK */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0xBF));

    /* deassert MSG & C/D (take out of msg in phase) */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0xF8));
    if ((BUS_PUTCX((iocc_addr + SCNTL1), 0x08)) != 0)
    {
	DEBUG_0 ("Error writing to SCNTL1\n");
	BUS_PUTCX((iocc_addr + SCNTL1), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    delay(1 * HZ);
    BUS_PUTCX((iocc_addr + SCNTL1), 0x00);

    /* change to initiator mode to free bus */
    TRACE_1("initiator    ",0);
    rc = BUS_PUTCX((iocc_addr + SCNTL0), 0xC0);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0x0));

    /* bring the chip to the data in phase */

    rc = BUS_PUTCX((iocc_addr + DCNTL), 0x81);
    if (rc != 0)
    {
	DEBUG_0 ("error writing to chip reset in DCNTL\n")
	BUS_PUTCX((iocc_addr + DCNTL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    rc = BUS_PUTCX((iocc_addr + DCNTL), 0x80);
    if (rc != 0)
    {
	DEBUG_0 ("error clearing chip reset in DCNTL\n")
	BUS_PUTCX((iocc_addr + DCNTL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    TRACE_1 ("arb mode ", 0);
    BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
    /* Arbitrate for the bus */
    TRACE_1 ("lpback mode ", 0);
    rc = BUS_PUTCX((iocc_addr + CTEST4), 0x10);
    if (rc != 0)
    {
	DEBUG_0 ("error writing to loopback mode in CTEST4\n")
	        BUS_PUTCX((iocc_addr + CTEST4), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    TRACE_1 ("arb mode ", 0);
    rc = BUS_PUTCX((iocc_addr + SCNTL0), 0xc1);
    if (rc != 0)
    {
	DEBUG_0 ("error writing to arbitration mode in SCNTL0\n")
	BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    /* Set low level mode */
    rc = BUS_PUTCX((iocc_addr + DCNTL), 0x80);
    if (rc != 0)
    {
	DEBUG_0 ("error writing to low level mode in DCNTL\n")
	BUS_PUTCX((iocc_addr + DCNTL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    /* Set destination ID */
    rc = BUS_PUTCX((iocc_addr + SDID), 0x01);
    if (rc != 0)
    {
	DEBUG_0 ("error writing dest. id in SDID\n")
	BUS_PUTCX((iocc_addr + SDID), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    /* Set card IDs */
    rc = BUS_PUTCX((iocc_addr + SCID), 0x81);
    if (rc != 0)
    {
	DEBUG_0 ("error writing chip id in SCID\n")
	BUS_PUTCX((iocc_addr + SCID), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    TRACE_1 ("start seq ", 0);
    rc = BUS_PUTCX((iocc_addr + SCNTL0), 0xe1);
    if (rc != 0)
    {
	DEBUG_0 ("error writing to start sequence bit in SCNTL0\n")
	BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    /* read SBCL sel active and bsy inactive indicating */
    /* we won arbitration                               */
    count = 5;
    pio_rc = 0;
    TRACE_1 ("arb check ", 0);
    while ((count > 0) && (pio_rc != 0x08))
    {
	BUS_GETCX((iocc_addr + ISTAT), &pio_rc);
	pio_rc &= 0x08;
        TRACE_1 ("ISTAT     ", pio_rc);
	count--;
    }
    if (pio_rc != 0x08)
    {
	DEBUG_0 ("Unsuccessful arbitration phase\n")
	BUS_PUTCX((iocc_addr + SOCL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    TRACE_1 ("sel phase ", 0);
    /* Go to selection phase */
    BUS_GETCX((iocc_addr + SBCL), &pio_rc);
    if (pio_rc < 0)
    {
	DEBUG_0 ("error reading SBCL\n")
	BUS_PUTCX((iocc_addr + SOCL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    pio_rc |= 0x20;
    TRACE_1 ("mid sel  ", (int) pio_rc);
    rc = BUS_PUTCX((iocc_addr + SOCL), pio_rc);
    if (rc != 0)
    {
	DEBUG_0 ("error writing to selection phase in SOCL\n")
	BUS_PUTCX((iocc_addr + SOCL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    count = 5;
    pio_rc = 0;
    while ((count > 0) && (pio_rc != 0x40))
    {
	BUS_GETCX((iocc_addr + SSTAT0), &pio_rc);
	pio_rc &= 0x40;
	count--;
    }
    TRACE_1 ("end sel  ", (int) pio_rc);
    if (pio_rc != 0x40)
    {
	DEBUG_0 ("Unsuccessful selection phase\n")
	BUS_PUTCX((iocc_addr + SOCL), 0x00);
	diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	diag_ptr->diag_rc.diag_validity = 0x01;
	diag_ptr->diag_rc.ahs_validity = 0x01;
	diag_ptr->diag_rc.ahs = PIO_RD_OTHR_ERR;
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }

    /* Put into initiator mode */
    BUS_PUTCX((iocc_addr + SCNTL0), 0xc0);

    /* write data onto bus */
    BUS_PUTCX((iocc_addr + SODL), 0x01);

    BUS_GETCX((iocc_addr + SCNTL1), &pio_rc);
    BUS_PUTCX((iocc_addr + SCNTL1), (pio_rc | 0x40));

    /* Put into target mode */
    BUS_PUTCX((iocc_addr + SCNTL0), 0xc1);
    /* assert ACK */

    /* Put into initiator mode */
    BUS_PUTCX((iocc_addr + SCNTL0), 0xc0);
    TRACE_1 ("strt data in ", 0);
    for (count = 0; count < 16; count++)
    {
	/* write data onto bus */
	BUS_PUTCX((iocc_addr + SODL), count);
	BUS_GETCX((iocc_addr + SCNTL1), &pio_rc);
	BUS_PUTCX((iocc_addr + SCNTL1), (pio_rc | 0x40));

	/* Put into initiator mode */
	BUS_PUTCX((iocc_addr + SCNTL0), 0xc0);

	/* Change to data in */
	BUS_PUTCX((iocc_addr + SOCL), 0x21);

	/* assert REQ */
	BUS_GETCX((iocc_addr + SOCL), &pio_rc);
	BUS_PUTCX((iocc_addr + SOCL), (pio_rc | 0x80));

	/* deassert REQ */
	BUS_GETCX((iocc_addr + SOCL), &pio_rc);
	BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0x7F));

	/* assert ACK */
	BUS_GETCX((iocc_addr + SOCL), &pio_rc);
	BUS_PUTCX((iocc_addr + SOCL), (pio_rc | 0x40));

	/* compare data received */
	BUS_GETCX((iocc_addr + SBDL), &pio_rc);
	if (pio_rc != count)
	{
	    DEBUG_0 ("error comparing data out of SBDL\n")
	    TRACE_1 ("cmp err2     ", count);
	    BUS_PUTCX((iocc_addr + SOCL), 0x00);
	    diag_ptr->diag_rc.diag_stat = SC_DIAG_UNRECOV_PIO;
	    diag_ptr->diag_rc.diag_validity = 0x01;
	    diag_ptr->diag_rc.ahs_validity = 0x01;
	    diag_ptr->diag_rc.ahs = PIO_WR_OTHR_ERR;
	    BUSIO_DET(iocc_addr);
	    return EFAULT;
	}

	/* deassert ACK */
	BUS_GETCX((iocc_addr + SOCL), &pio_rc);
	BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0xBF));

	BUS_PUTCX((iocc_addr + SODL), 0x00);
    }	/* end for loop */
    TRACE_1 ("end data in  ", 0);

    /* Write the bus phases in the SOCL reg to bring the chip */
    /* to the command completion phase                        */

    /* assert ATN */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc | 0x08));

    /* Get to msg out phase (assert MSG & C/D) */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc | 0x60));

    /* assert REQ */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc | 0x80));

    /* indicate command completion */
    BUS_PUTCX((iocc_addr + SODL), 0x00);

    /* assert ACK */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc | 0x40));

    /* deassert REQ */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0x7F));

    /* deassert ACK */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0xBF));

    /* deassert ATN */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0xF7));

    BUS_PUTCX((iocc_addr + SODL), 0x00);
    /* deassert MSG & C/D (take out of msg out phase) */
    BUS_GETCX((iocc_addr + SOCL), &pio_rc);
    BUS_PUTCX((iocc_addr + SOCL), (pio_rc & 0xF9));

    /* Go to bus free (deassert all signals) */
    BUS_PUTCX((iocc_addr + SOCL), 0x00);
    TRACE_1 ("out test", 0);
    DEBUG_0 ("Leaving psc_loop_test_diagnostics\n")
            BUSIO_DET(iocc_addr);
    return (PSC_NO_ERR);
}  /* end psc_loop_test_diagnostics */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_register_test                                               */
/*                                                                        */
/* FUNCTION:  Runs a test of the chip register logic circuits.            */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine runs test of chip register logic circuitry.  Data */
/*         patterns are written to each applicable register and read back */
/*         for compare purposes.                                          */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    scsi chip structure - scsi chip information.                        */
/*    diagnostic structure - diagnostic information.                      */
/*                                                                        */
/* INPUTS:                                                                */
/*    diag_ptr - diagnostic structure pointer                             */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:  0 for good completion, ERRNO value otherwise*/
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_register_test(
		  struct sc_card_diag * diag_ptr)
{
    int     rc;
    char    pio_rc;
    caddr_t iocc_addr;

    DEBUG_0 ("Entering psc_register_test\n")
    TRACE_1 ("in reg_test       ", diag_ptr);
    iocc_addr = BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
    BUS_PUTCX((iocc_addr + SCNTL0), 0xCA);
    BUS_GETCX((iocc_addr + SCNTL0), &pio_rc);
    if (pio_rc != 0xCA)
    {
	DEBUG_0 ("write to SCNTL0 of 0xCA failed\n");
	TRACE_1 ("SCNTL0 failed", pio_rc);
	BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + SCNTL0), 0x55);
    BUS_GETCX((iocc_addr + SCNTL0), &pio_rc);
    if (pio_rc != 0x55)
    {
	DEBUG_0 ("write to SCNTL0 of 0x55 failed\n");
	TRACE_1 ("E_scntl0 55    ", pio_rc);
	BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + SCNTL0), 0x00);
    BUS_PUTCX((iocc_addr + SCNTL1), 0xA0);
    BUS_GETCX((iocc_addr + SCNTL1), &pio_rc);
    if (pio_rc != 0xA0)
    {
	DEBUG_0 ("write to SCNTL1 of 0xA0 failed\n");
	TRACE_1 ("E_scntl1 A0    ", pio_rc);
	BUS_PUTCX((iocc_addr + SCNTL1), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + SCNTL1), 0x04);
    BUS_GETCX((iocc_addr + SCNTL1), &pio_rc);
    if (pio_rc != 0x04)
    {
	DEBUG_0 ("write to SCNTL1 of 0x04 failed\n");
	TRACE_1 ("E_scntl1 04    ", pio_rc);
	BUS_PUTCX((iocc_addr + SCNTL1), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + SCNTL1), 0x00);
    BUS_PUTCX((iocc_addr + SDID), 0xAA);
    BUS_GETCX((iocc_addr + SDID), &pio_rc);
    if (pio_rc != 0xAA)
    {
	DEBUG_0 ("write to SDID of 0xAA failed\n");
	TRACE_1 ("E_sdid AA    ", pio_rc);
	BUS_PUTCX((iocc_addr + SDID), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + SDID), 0x55);
    BUS_GETCX((iocc_addr + SDID), &pio_rc);
    if (pio_rc != 0x55)
    {
	DEBUG_0 ("write to SDID of 0x55 failed\n")
	TRACE_1 ("E_sdid 55    ", pio_rc);
	BUS_PUTCX((iocc_addr + SDID), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + SDID), 0x00);
    BUS_PUTCX((iocc_addr + SIEN), 0xAA);
    BUS_GETCX((iocc_addr + SIEN), &pio_rc);
    if (pio_rc != 0xAA)
    {
	DEBUG_0 ("write to SIEN of 0xAA failed\n");
	TRACE_1 ("E_sien AA    ", pio_rc);
	BUS_PUTCX((iocc_addr + SIEN), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + SIEN), 0x55);
    BUS_GETCX((iocc_addr + SIEN), &pio_rc);
    if (pio_rc != 0x55)
    {
	DEBUG_0 ("write to SIEN of 0x55 failed\n");
	TRACE_1 ("E_sien 55    ", pio_rc);
	BUS_PUTCX((iocc_addr + SIEN), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + SIEN), 0x00);
    BUS_PUTCX((iocc_addr + SCID), 0xAA);
    BUS_GETCX((iocc_addr + SCID), &pio_rc);
    if (pio_rc != 0xAA)
    {
	DEBUG_0 ("write to SCID of 0xAA failed\n")
	TRACE_1 ("E_scid AA    ", pio_rc);
	BUS_PUTCX((iocc_addr + SCID), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + SCID), 0x55);
    BUS_GETCX((iocc_addr + SCID), &pio_rc);
    if (pio_rc != 0x55)
    {
	DEBUG_0 ("write to SCID of 0x55 failed\n")
	TRACE_1 ("E_scid 55    ", pio_rc);
	BUS_PUTCX((iocc_addr + SCID), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + SCID), 0x00);
    BUS_PUTCX((iocc_addr + SXFER), 0xAA);
    BUS_GETCX((iocc_addr + SXFER), &pio_rc);
    if (pio_rc != 0xAA)
    {
	DEBUG_0 ("write to SXFER of 0xAA failed\n");
	TRACE_1 ("E_sxfer AA    ", pio_rc);
	BUS_PUTCX((iocc_addr + SXFER), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + SXFER), 0x55);
    BUS_GETCX((iocc_addr + SXFER), &pio_rc);
    if (pio_rc != 0x55)
    {
	DEBUG_0 ("write to SXFER of 0x55 failed\n")
	TRACE_1 ("E_sxfer 55    ", pio_rc);
	BUS_PUTCX((iocc_addr + SXFER), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + SXFER), 0x00);
    BUS_PUTCX((iocc_addr + DCMD), 0xAA);
    BUS_GETCX((iocc_addr + DCMD), &pio_rc);
    if (pio_rc != 0xAA)
    {
	DEBUG_0 ("write to DCMD of 0xAA failed\n")
	TRACE_1 ("E_dcmd AA    ", pio_rc);
	BUS_PUTCX((iocc_addr + DCMD), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + DCMD), 0x55);
    BUS_GETCX((iocc_addr + DCMD), &pio_rc);
    if (pio_rc != 0x55)
    {
	DEBUG_0 ("write to DCMD of 0x55 failed\n");
	TRACE_1 ("E_dcmd 55    ", pio_rc);
	BUS_PUTCX((iocc_addr + DCMD), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + DCMD), 0x00);
    BUS_PUTLX((iocc_addr + DNAD), 0xAA55AA55);
    BUS_GETLX((iocc_addr + DNAD), &rc);
    if (rc != 0xAA55AA55)
    {
	DEBUG_0 ("write to DNAD of 0xAA55AA55 failed\n");
	TRACE_1 ("E_dnad AA55AA55   ", rc);
	BUS_PUTLX((iocc_addr + DNAD), 0x00000000);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTLX((iocc_addr + DNAD), 0x55AA55AA);
    BUS_GETLX((iocc_addr + DNAD), &rc);
    if (rc != 0x55AA55AA)
    {
	DEBUG_0 ("write to DNAD of 0x55AA55AA failed\n");
	TRACE_1 ("E_dnad 55AA55AA   ", rc);
	BUS_PUTLX((iocc_addr + DNAD), 0x00000000);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTLX((iocc_addr + DNAD), 0x00000000);
    BUS_PUTCX((iocc_addr + DMODE), 0x01);
    BUS_PUTLX((iocc_addr + DSP), 0xAA55AA55);
    BUS_GETLX((iocc_addr + DSP), &rc);
    if (rc != 0xAA55AA55)
    {
	DEBUG_0 ("write to DSP of 0xAA55AA55 failed\n");
	TRACE_1 ("E_dsp AA55AA55   ", rc);
	BUS_PUTLX((iocc_addr + DSP), 0x00000000);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTLX((iocc_addr + DSP), 0x55AA55AA);
    BUS_GETLX((iocc_addr + DSP), &rc);
    if (rc != 0x55AA55AA)
    {
	DEBUG_0 ("write to DSP of 0x55AA55AA failed\n")
	TRACE_1 ("E_dsp 55AA55AA   ", rc);
	BUS_PUTLX((iocc_addr + DSP), 0x00000000);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTLX((iocc_addr + DSP), 0x00000000);
    BUS_PUTLX((iocc_addr + DSPS), 0xAA55AA55);
    BUS_GETLX((iocc_addr + DSPS), &rc);
    if (rc != 0xAA55AA55)
    {
	DEBUG_0 ("write to DSPS of 0xAA55AA55 failed\n");
	TRACE_1 ("E_dsps AA55AA55   ", rc);
	BUS_PUTLX((iocc_addr + DSPS), 0x00000000);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTLX((iocc_addr + DSPS), 0x55AA55AA);
    BUS_GETLX((iocc_addr + DSPS), &rc);
    if (rc != 0x55AA55AA)
    {
	DEBUG_0 ("write to DSPS of 0x55AA55AA failed\n")
	TRACE_1 ("E_dsps 55AA55AA   ", rc);
	BUS_PUTLX((iocc_addr + DSPS), 0x00000000);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTLX((iocc_addr + DSPS), 0x00000000);
    BUS_PUTCX((iocc_addr + DMODE), 0xAA);
    BUS_GETCX((iocc_addr + DMODE), &pio_rc);
    if (pio_rc != 0xAA)
    {
	DEBUG_0 ("write to DMODE of 0xAA failed\n")
	TRACE_1 ("E_dmode AA    ", pio_rc);
	BUS_PUTCX((iocc_addr + DMODE), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + DMODE), 0x00);
    BUS_PUTCX((iocc_addr + DIEN), 0xAA);
    BUS_GETCX((iocc_addr + DIEN), &pio_rc);
    if (pio_rc != 0xAA)
    {
	DEBUG_0 ("write to DIEN of 0xAA failed\n")
	TRACE_1 ("E_dien AA    ", pio_rc);
	BUS_PUTCX((iocc_addr + DIEN), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + DIEN), 0x15);
    BUS_GETCX((iocc_addr + DIEN), &pio_rc);
    if (pio_rc != 0x15)
    {
	DEBUG_0 ("write to DIEN of 0x15 failed\n")
	TRACE_1 ("E_dien 15    ", pio_rc);
	BUS_PUTCX((iocc_addr + DIEN), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + DIEN), 0x00);
    BUS_PUTCX((iocc_addr + DWT), 0xAA);
    BUS_GETCX((iocc_addr + DWT), &pio_rc);
    if (pio_rc != 0xAA)
    {
	DEBUG_0 ("write to DWT of 0xAA failed\n")
	TRACE_1 ("E_dwt AA    ", pio_rc);
	BUS_PUTCX((iocc_addr + DWT), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + DWT), 0x55);
    BUS_GETCX((iocc_addr + DWT), &pio_rc);
    if (pio_rc != 0x55)
    {
	DEBUG_0 ("write to DWT of 0x55 failed\n")
	TRACE_1 ("E_dwt 55    ", pio_rc);
	BUS_PUTCX((iocc_addr + DWT), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + DWT), 0x00);
    BUS_PUTCX((iocc_addr + DCNTL), 0xA0);
    BUS_GETCX((iocc_addr + DCNTL), &pio_rc);
    if (pio_rc != 0xA0)
    {
	DEBUG_0 ("write to DCNTL of 0xA0 failed\n")
	TRACE_1 ("E_sdid A0    ", pio_rc);
	BUS_PUTCX((iocc_addr + DCNTL), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + DCNTL), 0x50);
    BUS_GETCX((iocc_addr + DCNTL), &pio_rc);
    if (pio_rc != 0x50)
    {
	DEBUG_0 ("write to DCNTL of 0x50 failed\n")
	TRACE_1 ("E_sdid 50    ", pio_rc);
	BUS_PUTCX((iocc_addr + DCNTL), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + DCNTL), 0x00);
    DEBUG_0 ("Leaving psc_register_test\n")
    TRACE_1 ("out reg test   ", 0);
    BUSIO_DET(iocc_addr);
    return (0);
}  /* end psc_register_test */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_pos_register_test                                           */
/*                                                                        */
/* FUNCTION:  Runs a test of the POS register logic circuits.             */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine runs a test of the POS registers by first writing */
/*         a data pattern and then reading it for compare.                */
/*                                                                        */
/* DATA STRUCTURES:                                                       */
/*    scsi chip structure - scsi chip information.                        */
/*    diagnostic structure - diagnostic information.                      */
/*                                                                        */
/* INPUTS:                                                                */
/*    diag_ptr - diagnostic structure pointer                             */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:  0 for good completion, ERRNO value otherwise*/
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_pos_register_test(
		      struct sc_card_diag * diag_ptr)
{
    caddr_t iocc_addr;
    char    pio_rc, pos_save;

    DEBUG_0 ("Entering psc_pos_register_test\n")
    iocc_addr = IOCC_ATT(adp_str.ddi.bus_id, 0);
    BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) + POS0), &pio_rc);
    if (pio_rc != 0xF4)
    {
	DEBUG_0 ("Error reading POS0\n");
	TRACE_1 ("err pos0", pio_rc);
	return EFAULT;
    }
    BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) + POS1), &pio_rc);
    if (pio_rc != 0x8F)
    {
	DEBUG_0 ("Error reading POS1\n");
	TRACE_1 ("err pos1", pio_rc);
	return EFAULT;
    }
    BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) + POS4), &pos_save);
    BUS_PUTCX((iocc_addr + (adp_str.ddi.slot << 16) + POS4), 0x0A);
    BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) + POS4), &pio_rc);
    pio_rc &= 0x0A;
    if (pio_rc != 0x0A)
    {
	DEBUG_0 ("Error reading POS4\n");
	TRACE_1 ("err1pos4", pio_rc);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + (adp_str.ddi.slot << 16) + POS4), 0x15);
    BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) + POS4), &pio_rc);
    pio_rc &= 0x15;
    if (pio_rc != 0x15)
    {
	DEBUG_0 ("Error reading POS4\n");
	TRACE_1 ("err2pos4", pio_rc);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + (adp_str.ddi.slot << 16) + POS4), pos_save);
    BUS_PUTCX((iocc_addr + (adp_str.ddi.slot << 16) + POS6), 0xAA);
    BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) + POS6), &pio_rc);
    if (pio_rc != 0xAA)
    {
	DEBUG_0 ("Error reading POS6\n");
	TRACE_1 ("err1pos6", pio_rc);
	return EFAULT;
    }
    BUS_PUTCX((iocc_addr + (adp_str.ddi.slot << 16) + POS6), 0x55);
    BUS_GETCX((iocc_addr + (adp_str.ddi.slot << 16) + POS6), &pio_rc);
    if (pio_rc != 0x55)
    {
	DEBUG_0 ("Error reading POS6\n");
	TRACE_1 ("err2pos6", pio_rc);
	return EFAULT;
    }
    IOCC_DET(iocc_addr);
    DEBUG_0 ("Leaving psc_pos_register_test\n");
    return (0);
}  /* end psc_pos_register_test */

/**************************************************************************/
/*                                                                        */
/* NAME:  psc_diag_reset_scsi_bus                                         */
/*                                                                        */
/* FUNCTION:  Resets the SCSI bus lines.                                  */
/*                                                                        */
/* EXECUTION ENVIRONMENT:                                                 */
/*      This routine can be called by a process.                          */
/*                                                                        */
/* NOTES:  This routine will toggle the SCSI bus reset line for 30 micro- */
/*         seconds to assert a reset on the SCSI bus. This routines exe-  */
/*         cutes to reset the bus during diagnostic execution.            */
/*                                                                        */
/* DATA STRUCTURES: None                                                  */
/*                                                                        */
/* INPUTS: None                                                           */
/*                                                                        */
/* RETURN VALUE DESCRIPTION:  0 for good completion, ERRNO value otherwise*/
/*                                                                        */
/* ERROR DESCRIPTION:                                                     */
/*                                                                        */
/* EXTERNAL PROCEDURES CALLED:                                            */
/*                                                                        */
/**************************************************************************/
int
psc_diag_reset_scsi_bus(
			struct sc_card_diag * diag_ptr)
{
    caddr_t iocc_addr;

    DEBUG_0 ("Entering psc_diag_reset_scsi_bus\n")
    iocc_addr = BUSIO_ATT((adp_str.ddi.bus_id | 0x000c0060), 0);
    if ((BUS_PUTCX((iocc_addr + SCNTL1), 0x08)) != 0)
    {
	DEBUG_0 ("Error writing to SCNTL1\n");
	BUS_PUTCX((iocc_addr + SCNTL1), 0x00);
	BUSIO_DET(iocc_addr);
	return EFAULT;
    }
    delay(1 * HZ);
    BUS_PUTCX((iocc_addr + SCNTL1), 0x00);
    BUSIO_DET(iocc_addr);
    DEBUG_0 ("Leaving psc_diag_reset_scsi_bus\n")
    return (PSC_NO_ERR);
}  /* end psc_diag_reset_scsi_bus */

#ifdef PSC_TRACE
void
psc_trace_1(char *string, int data)
{
    int     i;

    if (adp_str.current_trace_line > 0xFF0)
    {
	adp_str.current_trace_line = 1;
    }
    for (i = 0; i < 13; i++)
	adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.one_val.header1[i] = *(string + i);
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.one_val.data = data;
    adp_str.current_trace_line++;

    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[0] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[1] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[2] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[3] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[4] = 'P';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[5] = 'O';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[6] = 'I';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[7] = 'N';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[8] = 'T';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[9] = 'E';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[10] = 'R';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[11] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[12] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[13] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[14] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[15] = '*';
}
#endif

#ifdef PSC_TRACE
void
psc_trace_2(char *string, int val1, int val2)
{
    int     i;

    if (adp_str.current_trace_line > 0xFF0)
    {
	adp_str.current_trace_line = 1;
    }
    for (i = 0; i < 9; i++)
	adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.two_vals.header2[i] = *(string + i);
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.two_vals.data1 = val1;
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.two_vals.data2 = val2;
    adp_str.current_trace_line++;
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[0] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[1] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[2] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[3] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[4] = 'P';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[5] = 'O';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[6] = 'I';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[7] = 'N';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[8] = 'T';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[9] = 'E';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[10] = 'R';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[11] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[12] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[13] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[14] = '*';
    adp_str.trace_ptr->trace_buffer
	 [adp_str.current_trace_line].un.header[15] = '*';
}
#endif

#ifdef PSC_TRACE
void
psc_trace_3(char *string, int data1, int data2, int data3)
{
    int     i;

    if (adp_str.current_trace_line > 0xFF0)
    {
	adp_str.current_trace_line = 1;
    }
    for (i = 0; i < 5; i++)
	adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.three_vals.header3[i] = *(string + i);
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.three_vals.val1 = data1;
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.three_vals.val2 = data2;
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.three_vals.val3 = data3;
    adp_str.current_trace_line++;
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[0] = '*';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[1] = '*';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[2] = '*';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[3] = '*';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[4] = 'P';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[5] = 'O';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[6] = 'I';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[7] = 'N';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[8] = 'T';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[9] = 'E';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[10] = 'R';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[11] = '*';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[12] = '*';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[13] = '*';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[14] = '*';
    adp_str.trace_ptr->trace_buffer
	[adp_str.current_trace_line].un.header[15] = '*';
}
#endif
